/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#include "BitArray.h"


static BitArray                 gbaCache;
static QWORD                    gqwTargetSize;
static DWORD                    gdwBlkSize;
static UINT                     gnBitCnt;

void Usages()
{
    log_wprintf(L"\nUsages:\n");
    log_wprintf(L"\tCreateCache c cache_file target_size blk_size\n");
    log_wprintf(L"\tCreateCache m cache_file target_size blk_size part_bmfile part_off part_len clu_size(B)\n");
}

bool CheckArgs(int argc, char *argv[])
{
    if ( argc < 2 )
    {
        return false;
    }

    if ( (argv[1][0] == 'c') && (argc == 5) )
    {
        return true;
    }

    if ( (argv[1][0] == 'm') && (argc == 9) )
    {
        return true;
    }

    return false;
}

UINT CalcBitCount(QWORD qwTargetSize, DWORD dwBlkSize)
{
    return (UINT)((qwTargetSize + dwBlkSize - 1) / (QWORD)dwBlkSize);
}

UINT Pos2Bit(QWORD qwPos)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return (UINT)(qwPos / gdwBlkSize);
}

QWORD AllignPosUp(QWORD qwPos)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return ((qwPos + gdwBlkSize - 1) / gdwBlkSize * gdwBlkSize);
}

QWORD AllignPosDown(QWORD qwPos)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return ((qwPos + 1) / gdwBlkSize * gdwBlkSize - 1);
}

void MarkCache(QWORD qwStartPos, QWORD qwEndPos, bool bClear)
{
    UINT  nStartBit, nEndBit, i;

    DBG(D_VERB, (L"%s: pos: %lld -> %lld\n", FULL_FUNC_NAME, qwStartPos, qwEndPos));

    nStartBit = Pos2Bit(qwStartPos);
    nEndBit = Pos2Bit(qwEndPos);

    DBG(D_VERB, (L"%s: bit: %d -> %d\n", FULL_FUNC_NAME, nStartBit, nEndBit));
    for ( i = nStartBit; i <= nEndBit; ++i )
    {
        bClear ? gbaCache.ClearBit(i) : gbaCache.SetBit(i);
    }
}

void InitPartCache(QWORD qwStartPos, QWORD qwEndPos)
{
    QWORD  qwStartPos2, qwEndPos2;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    qwStartPos2 = AllignPosUp(qwStartPos);
    qwEndPos2 = AllignPosDown(qwEndPos);

    log_wprintf(L"mark cache: %lld(%lld) -> %lld(%lld)\n", qwStartPos2, qwStartPos, qwEndPos2, qwEndPos);
    if ( qwStartPos2 < qwEndPos2 )
    {
        MarkCache(qwStartPos2, qwEndPos2, false);
    }
}

DWORD MergeCache(const char *szBmFile, QWORD qwPartOff, QWORD qwPartLen, DWORD dwCluSize)
{
    DWORD  dwRet;
    BitArray  baPart;
    UINT  nCluCnt, i;
    QWORD  qwPartEndPos;
    QWORD  qwStartPos, qwEndPos;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( qwPartLen == 0 )
    {
        log_wprintf(L"part length is zero, skip it\n");
        return ERROR_SUCCESS;
    }

    qwPartEndPos = qwPartOff + qwPartLen - 1;
    if ( (qwPartOff >= gqwTargetSize) || ( qwPartEndPos >= gqwTargetSize) )
    {
        log_wprintf(L"part offset or length invalid\n");
        assert(0);
        return ERROR_INVALID_PARAMETER;
    }

    nCluCnt = CalcBitCount(qwPartLen, dwCluSize);

    dwRet = baPart.Init(nCluCnt, szBmFile, false);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init bitmap file(%hs) failed: %d\n", szBmFile, dwRet);
        goto err1;
    }

    InitPartCache(qwPartOff, qwPartEndPos);

    qwStartPos = qwPartOff;
    qwEndPos = qwStartPos + dwCluSize - 1;
    for ( i = 0; i < nCluCnt; ++i )
    {
        if ( baPart.IsBitSet(i) )
        {
            MarkCache(qwStartPos, qwEndPos, true);
        }
        qwStartPos += dwCluSize;
        qwEndPos += dwCluSize;
    }

    baPart.Fini();

err1:
    return dwRet;
}

int main(int argc, char *argv[])
{
    DWORD  dwRet;

    if ( !CheckArgs(argc, argv) )
    {
        Usages();
        return ERROR_INVALID_PARAMETER;
    }

    gqwTargetSize = _atoi64(argv[3]);
    gdwBlkSize = atoi(argv[4]);
    gnBitCnt = CalcBitCount(gqwTargetSize, gdwBlkSize);
    log_wprintf(L"file: %lldMB, blk: %dKB, bits: %d\n", gqwTargetSize>>20, gdwBlkSize>>10, gnBitCnt);

    dwRet = gbaCache.Init(gnBitCnt, argv[2], false);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init cache file(%hs) failed: %d\n", argv[2], dwRet);
        goto err1;
    }

    switch ( argv[1][0] )
    {
        case 'c':
            log_wprintf(L"create (%hs) and zero(%d bytes) cache file\n", argv[2], gbaCache.GetBytesCount());
            gbaCache.ReInit(false);
            break;

        case 'm':
            dwRet = MergeCache(argv[5], _atoi64(argv[6]), _atoi64(argv[7]), atoi(argv[8]));
            break;
    }

    gbaCache.Fini();

err1:
    return dwRet;
}
