/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#include "BufferItem.h"
#include "SHA1Hash.h"


DWORD ReadBlock(FILE *hFile, BufferItem *pbiData, DWORD dwBlkSize)
{
    DWORD  dwRet = ERROR_SUCCESS;
    DWORD  dwSize;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(pbiData->tail_size() >= dwBlkSize);

    while ( dwBlkSize )
    {
        dwSize = (DWORD)fread(pbiData->tail(), 1, dwBlkSize, hFile);
        if ( dwSize == 0 )
        {
            if ( !feof(hFile) )
            {
                dwRet = GetLastError();
                DBG(D_ERROR, (L"%s: read file failed: %d\n", FULL_FUNC_NAME, dwRet));
            }
            break;
        }

        pbiData->Put(dwSize);
        dwBlkSize -= dwSize;
    }

    return dwRet;
}

DWORD HashBlock(BufferItem *pbiData, PSHA1_VAL pValSHA1)
{
    DWORD  dwRet;
    SHA1Hash  sha1;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    dwRet = sha1.BeginHash();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_BeginHash failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    dwRet = sha1.DoHash(pbiData->data(), pbiData->data_size());
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_DoHash failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    dwRet = sha1.GetHashVal(pValSHA1);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_GetHashVal failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    dwRet = ERROR_SUCCESS;

err2:
    sha1.EndHash();

err1:
    return dwRet;
}

DWORD HashFile(FILE *hFile, PSEED_HDR pSeedHdr)
{
    DWORD  dwRet = ERROR_INVALID_PARAMETER;
    DWORD  dwCnt, dwBlkSize, dwBlkNum;
    BufferItem  biData;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    fseeko(hFile, 0, SEEK_SET);

    dwBlkSize = pSeedHdr->dwBlockSize;
    dwBlkNum = pSeedHdr->dwSHA1Num;
    biData.Alloc(dwBlkSize);
    for ( dwCnt = 0; dwCnt < dwBlkNum; ++dwCnt )
    {
        biData.Reset();
        dwRet = ReadBlock(hFile, &biData, dwBlkSize);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: ReadBlock failed: %d\n", FULL_FUNC_NAME, dwRet));
            break;
        }

        log_wprintf(L"Hashing Block[%d / %d]\r", dwCnt, dwBlkNum);
        dwRet = HashBlock(&biData, &pSeedHdr->SHA1ValBlk[dwCnt]);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: HashBlock failed: %d\n", FULL_FUNC_NAME, dwRet));
            break;
        }
    }
    log_wprintf(L"\n");

    return dwRet;
}

DWORD HashInfo(const char *szFile, PSEED_HDR pSeedHdr)
{
    DWORD  dwRet;
    QWORD  qwPerfCnt;
    SHA1Hash  sha1;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    dwRet = sha1.BeginHash();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_BeginHash failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    // see issues_seed.txt
    qwPerfCnt = GetPerformanceCounter();
    dwRet = sha1.DoHash((PBYTE)&qwPerfCnt, sizeof(qwPerfCnt));
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_DoHash failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    // [0] always exist
    dwRet = sha1.GetHashVal(&pSeedHdr->SHA1ValBlk[0]);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_GetHashVal failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    dwRet = ERROR_SUCCESS;

err2:
    sha1.EndHash();

err1:
    return dwRet;
}

DWORD HashSeed(PSEED_HDR pSeedHdr)
{
    DWORD  dwRet;
    SHA1Hash  sha1;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    dwRet = sha1.BeginHash();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_BeginHash failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    for ( DWORD i = 0; i < pSeedHdr->dwSHA1Num; ++i )
    {
        dwRet = sha1.DoHash(pSeedHdr->SHA1ValBlk[i].bVal, sizeof(SHA1_VAL));
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: SHA1_DoHash failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto err2;
        }
    }

    dwRet = sha1.GetHashVal(&pSeedHdr->SHA1ValSeed);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_GetHashVal failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    dwRet = ERROR_SUCCESS;

err2:
    sha1.EndHash();

err1:
    return dwRet;
}

int main(int argc, char *argv[])
{
    DWORD  dwRet;
    FILE  *hOrgFile, *hSeedFile;
    BufferItem  biSeedHdr;
    PSEED_HDR  pSeedHdr;
    QWORD  qwFileSize;
    bool  bFullHash;
    char  *szOrgFile, *szSeedFile;
    DWORD  dwBlkSizeMB, dwBlkSize, dwBlkNum, dwSeedSize;

    if ( argc < 4 )
    {
        log_wprintf(L"CreateSeed.exe <org_file> <seed_file> <blk_size(MB)> [1=full_hash]\n");
        return ERROR_INVALID_PARAMETER;
    }

    szOrgFile = argv[1];
    szSeedFile = argv[2];
    dwBlkSizeMB = atoi(argv[3]);
    bFullHash = (argc > 4) ? (argv[4][0] == '1') : false;

    if ( dwBlkSizeMB > 4095 )
    {
        log_wprintf(L"blk_size(MB) is too large\n");
        return ERROR_INVALID_PARAMETER;
    }

    dwRet = SHA1Hash::Startup();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_InitCtx failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    hOrgFile = fopen(szOrgFile, "rb");
    if ( !hOrgFile )
    {
        dwRet = GetLastError();
        log_wprintf(L"open original file(%hs) failed: %d\n", szOrgFile, dwRet);
        goto err2;
    }

    hSeedFile = fopen(szSeedFile, "wb");
    if ( !hSeedFile )
    {
        dwRet = GetLastError();
        log_wprintf(L"open seed file(%hs) failed: %d\n", szSeedFile, dwRet);
        goto err3;
    }

    if ( fseeko(hOrgFile, 0, SEEK_END) )
    {
        dwRet = GetLastError();
        log_wprintf(L"seek original file(%hs) failed: %d\n", szOrgFile, dwRet);
        goto err4;
    }

    qwFileSize = ftello(hOrgFile);
    dwBlkSize = dwBlkSizeMB * 1024 * 1024;
    dwBlkNum = (DWORD)((qwFileSize + dwBlkSize - 1) / (QWORD)dwBlkSize);

    dwSeedSize = sizeof(SEED_HDR) + (dwBlkNum - 1) * sizeof(SHA1_VAL);
    biSeedHdr.Alloc(dwSeedSize);
    pSeedHdr = (PSEED_HDR)biSeedHdr.Put(dwSeedSize);
    memset(pSeedHdr, 0, dwSeedSize);

    pSeedHdr->dwMagic = SEED_MAGIC_NUM;
    GetFileName(szOrgFile, pSeedHdr->szFileName, MAX_FILE_NAME_LEN);
    pSeedHdr->qwFileSize = qwFileSize;
    pSeedHdr->dwBlockSize = dwBlkSize;
    pSeedHdr->dwSHA1Num = dwBlkNum;

    dwRet = bFullHash ? HashFile(hOrgFile, pSeedHdr) : HashInfo(szOrgFile, pSeedHdr);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"hash(SHA1) file(%hs) failed: %d\n", szOrgFile, dwRet);
        goto err4;
    }

    dwRet = HashSeed(pSeedHdr);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"hash(SHA1) seed failed: %d\n", dwRet);
        goto err4;
    }

    if ( fwrite(pSeedHdr, 1, dwSeedSize, hSeedFile) != dwSeedSize )
    {
        // see misc_fio.txt
        dwRet = GetLastError();
        log_wprintf(L"write seed file(%hs) failed: %d\n", szSeedFile, dwRet);
        goto err4;
    }

    log_wprintf(L"create seed file(%hs) successfully:\n", szSeedFile);
    log_wprintf(L"    seed file size = %d\n", dwSeedSize);
    log_wprintf(L"    org file size = %lld\n", qwFileSize);
    log_wprintf(L"    block size = %dMB\n", dwBlkSizeMB);
    log_wprintf(L"    block number = %d\n", dwBlkNum);

    if ( ((dwBlkNum + 7 ) / 8) > (LP2P_UPDOWN_DATABSIZE - PDU_HEAD_RESERVE - sizeof(RET_GET_SEEDTARGET_BITMASK)) )
    {
        log_wprintf(L" too many blocks: %d, may exceed the send buf size\n", dwBlkNum);
    }

    dwRet = ERROR_SUCCESS;

err4:
    fclose(hSeedFile);

err3:
    fclose(hOrgFile);

err2:
    SHA1Hash::Cleanup();

err1:
    return dwRet;
}
