/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#include "BitArray.h"


DWORD WriteBuffer(int fd, PBYTE buf, ssize_t len)
{
    DWORD  dwRet = ERROR_SUCCESS;
    ssize_t  bytes;

    while ( len > 0 )
    {
        bytes = write(fd, buf, (UINT)len);
        if ( bytes < 0 )
        {
            dwRet = GetLastError();
            break;
        }

        buf += bytes;
        len -= bytes;
    }

    return dwRet;    
}

UINT CalcBitCount(QWORD qwFileSize, DWORD dwBlkSize)
{
    return (UINT)((qwFileSize + dwBlkSize - 1) / (QWORD)dwBlkSize);
}

DWORD CatBlock(int hSrc, int hDst, UINT nBlkIdx, DWORD dwBlkSize)
{
    DWORD  dwRet = ERROR_SUCCESS;
    QWORD  qwOffset;
    ssize_t  nBytes, nLeft;
    BYTE  buf[512];

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    qwOffset = (QWORD)dwBlkSize * (QWORD)nBlkIdx;
    if ( lseek(hSrc, qwOffset, SEEK_SET) < 0 )
    {
        dwRet = GetLastError();
        err_wprintf(L"seek src file to %lld failed: %d\n", qwOffset, dwRet);
        return dwRet;
    }

    nLeft = dwBlkSize;
    while ( nLeft > 0 )
    {
        nBytes = min(sizeof(buf), (size_t)nLeft);

        nBytes = read(hSrc, buf, (UINT)nBytes);
        if ( nBytes < 0 )
        {
            dwRet = GetLastError();
            err_wprintf(L"read from src file failed: %d\n", dwRet);
            break;
        }

        if ( nBytes == 0 )
        {
            break;
        }

        dwRet = WriteBuffer(hDst, buf, nBytes);
        if ( dwRet != ERROR_SUCCESS )
        {
            err_wprintf(L"write to dst file failed: %d\n", dwRet);
            break;
        }

        nLeft -= nBytes;
    }

    return dwRet;
}

DWORD SkipBlock(int hDst, DWORD dwBlkSize)
{
    DWORD  dwRet = ERROR_SUCCESS;
    ssize_t  nBytes, nLeft;
    BYTE  buf[512];

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    memset(buf, 0, sizeof(buf));

    nLeft = dwBlkSize;
    while ( nLeft > 0 )
    {
        nBytes = min(sizeof(buf), (size_t)nLeft);

        dwRet = WriteBuffer(hDst, buf, nBytes);
        if ( dwRet != ERROR_SUCCESS )
        {
            err_wprintf(L"write to dst file failed: %d\n", dwRet);
            break;
        }

        nLeft -= nBytes;
    }

    return dwRet;
}

int main(int argc, char *argv[])
{
    DWORD  dwRet;
    BitArray  gbaCache;
    QWORD  qwFileSize;
    DWORD  dwBlkSize;
    UINT  nBitCnt;
    int  hSrcFile, hDstFile;

    if ( argc < 5 )
    {
        err_wprintf(L"FastCat.exe <file_name> <file_size> <cache_file> <blk_size>\n");
        return ERROR_INVALID_PARAMETER;
    }

    qwFileSize = _atoi64(argv[2]);
    dwBlkSize = atoi(argv[4]);
    nBitCnt = CalcBitCount(qwFileSize, dwBlkSize);
    err_wprintf(L"file: %lldMB, blk: %dKB, bits: %d\n", qwFileSize>>20, dwBlkSize>>10, nBitCnt);

    dwRet = gbaCache.Init(nBitCnt, argv[3], false);
    if ( dwRet != ERROR_SUCCESS )
    {
        err_wprintf(L"init cache file(%hs) failed: %d\n", argv[3], dwRet);
        goto err1;
    }

    hSrcFile = open(argv[1], O_RDONLY);
    if ( hSrcFile < 0 )
    {
        dwRet = GetLastError();
        err_wprintf(L"open file(%hs) failed: %d\n", argv[1], dwRet);
        goto err2;
    }

    hDstFile = STDOUT_FILENO;

    err_wprintf(L"will skip %d blocks\n", gbaCache.GetSetsCount());

    for ( UINT i = 0; i < nBitCnt; ++i )
    {
        dwRet = gbaCache.IsBitSet(i) ? SkipBlock(hDstFile, dwBlkSize) : CatBlock(hSrcFile, hDstFile, i, dwBlkSize);
        if ( dwRet != ERROR_SUCCESS )
        {
            err_wprintf(L"cat block[%d] failed: %d\n", i, dwRet);
            break;
        }
    }

    close(hSrcFile);

err2:
    gbaCache.Fini();

err1:
    return dwRet;
}
