/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#include "BitArray.h"


static BitArray                 gbaCache;
static QWORD                    gqwFileSize;
static DWORD                    gdwBlkSize;
static UINT                     gnBitCnt;

static QWORD                    gqwCurOff;


UINT CalcBitCount(QWORD qwFileSize, DWORD dwBlkSize)
{
    return (UINT)((qwFileSize + dwBlkSize - 1) / (QWORD)dwBlkSize);
}

DWORD CopyBlock(FILE *hSrc, FILE *hDst, UINT nBlkIdx)
{
    DWORD  dwRet = ERROR_SUCCESS;
    QWORD  qwOffset;
    size_t  nBytes, nLeft;
    BYTE  buf[512];

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    qwOffset = (QWORD)gdwBlkSize * (QWORD)nBlkIdx;

    if ( qwOffset != gqwCurOff )
    {
        if ( fseeko(hSrc, qwOffset, SEEK_SET) )
        {
            dwRet = GetLastError();
            log_wprintf(L"seek src file to %lld failed: %d\n", qwOffset, dwRet);
            return dwRet;
        }

        if ( fseeko(hDst, qwOffset, SEEK_SET) )
        {
            dwRet = GetLastError();
            log_wprintf(L"seek dst file to %lld failed: %d\n", qwOffset, dwRet);
            return dwRet;
        }

        gqwCurOff = qwOffset;
    }

    nLeft = gdwBlkSize;
    while ( nLeft > 0 )
    {
        nBytes = min(sizeof(buf), nLeft);

        nBytes = fread(buf, 1, nBytes, hSrc);
        if ( nBytes == 0 )
        {
            if ( ferror(hSrc) )
            {
                dwRet = GetLastError();
                log_wprintf(L"read from src file failed: %d\n", dwRet);
            }
            break;
        }

        if ( fwrite(buf, 1, nBytes, hDst) != nBytes )
        {
            // see misc_fio.txt
            dwRet = GetLastError();
            log_wprintf(L"write to dst file failed: %d\n", dwRet);
            break;
        }

        nLeft -= nBytes;
        gqwCurOff += nBytes;
    }

    return dwRet;
}

int main(int argc, char *argv[])
{
    DWORD  dwRet;
    FILE  *hSrcFile, *hDstFile;
    UINT  nOldPer, nCurPer;

    if ( argc < 6 )
    {
        log_wprintf(L"FastCopy.exe <src_file> <dst_file> <file_size> <cache_file> <blk_size>\n");
        return ERROR_INVALID_PARAMETER;
    }

    // GetFileLength(/dev/sda) to calc <file_size> is not expected
    gqwFileSize = _atoi64(argv[3]);
    gdwBlkSize = atoi(argv[5]);
    gnBitCnt = CalcBitCount(gqwFileSize, gdwBlkSize);
    log_wprintf(L"file: %lldMB, blk: %dKB, bits: %d\n", gqwFileSize>>20, gdwBlkSize>>10, gnBitCnt);

    dwRet = gbaCache.Init(gnBitCnt, argv[4], false);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init cache file(%hs) failed: %d\n", argv[4], dwRet);
        goto err1;
    }

    hSrcFile = fopen(argv[1], "rb");
    if ( !hSrcFile )
    {
        dwRet = GetLastError();
        log_wprintf(L"open src file(%hs) failed: %d\n", argv[1], dwRet);
        goto err2;
    }

    hDstFile = fopen(argv[2], "wb");
    if ( !hDstFile )
    {
        dwRet = GetLastError();
        log_wprintf(L"open dst file(%hs) failed: %d\n", argv[2], dwRet);
        goto err3;
    }

    log_wprintf(L"will skip %d blocks\n", gbaCache.GetSetsCount());
    gqwCurOff = 0;
    nOldPer = 0;
    for ( UINT i = 0; i < gnBitCnt; ++i )
    {
        if ( !gbaCache.IsBitSet(i) )
        {
            dwRet = CopyBlock(hSrcFile, hDstFile, i);
            if ( dwRet != ERROR_SUCCESS )
            {
                log_wprintf(L"copy block[%d] failed: %d\n", i, dwRet);
                break;
            }

            nCurPer = (i + 1) * 100 / gnBitCnt;
            if ( nCurPer != nOldPer )
            {
                log_wprintf(L"progress: %d%%\r", nCurPer);
                nOldPer = nCurPer;
            }
        }
    }
    log_wprintf(L"\ndone.\n");

    // dst file size may > src, because cache is 4KB alligned
    // dst file size may < src, because the last block may invalid
    log_wprintf(L"set dst file(%hs) length to %lld\n", argv[2], gqwFileSize);    
    dwRet = SetFileLength(hDstFile, gqwFileSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"set dst file(%hs) length failed: %d\n", argv[2], dwRet);
        goto err4;
    }

err4:
    fclose(hDstFile);

err3:
    fclose(hSrcFile);

err2:
    gbaCache.Fini();

err1:
    return dwRet;
}
