/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "BitArray.h"

BitArray::BitArray()
    : _nBitsCnt(0)
    , _nBytesCnt(0)
    , _pByteArray(NULL)
    , _hCacheFile(NULL)
    , _pCacheMapping(NULL)
    , _nSetsCnt(0)
    , _bCalcMask(false)
    , _MaskArray(0)
    , _qwMaskAnd(0)
    , _qwMaskOr(0)
    , _nRemain(0)
    , _nRightH(0)
    , _nLeftH(0)
    , _nDivide(0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

BitArray::~BitArray()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(!_pByteArray);
}

DWORD BitArray::Init(UINT nBitsCnt, bool bSetAll, bool bCalcMask)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nBitsCnt));

    return DoInit(nBitsCnt, true, bSetAll, bCalcMask);
}

DWORD BitArray::Init(UINT nBitsCnt, const PBYTE pByteData, UINT nSize, bool bCalcMask)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nBitsCnt));

    dwRet = DoInit(nBitsCnt, true, false, bCalcMask);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: DoInit(%d) failed: %d\n", FULL_FUNC_NAME, nBitsCnt, dwRet));
        goto err1;
    }

    assert(nSize ==_nBytesCnt);
    memcpy(_pByteArray, pByteData, min(nSize, _nBytesCnt));

    for ( UINT i = 0; i < nBitsCnt; ++i )
    {
        if ( IsBitSet(i) )
        {
            ++_nSetsCnt;
            OnSetBit(i);
        }
    }

    return ERROR_SUCCESS;

err1:
    return dwRet;
}

DWORD BitArray::Init(UINT nBitsCnt, const char *szCacheFile, bool bCalcMask)
{
    DWORD  dwRet;
    QWORD  qwSize;

    DBG(D_INFO, (L"%s: %d, %hs\n", FULL_FUNC_NAME, nBitsCnt, szCacheFile));

    assert(!_hCacheFile);

    dwRet = DoInit(nBitsCnt, false, false, bCalcMask);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: DoInit(%d) failed: %d\n", FULL_FUNC_NAME, nBitsCnt, dwRet));
        goto err1;
    }

    _hCacheFile = fopen(szCacheFile, "r+b");
    if ( !_hCacheFile )
    {
        DBG(D_INFO, (L"%s: cache file(%hs) not exist, create it\n", FULL_FUNC_NAME, szCacheFile));
        _hCacheFile = fopen(szCacheFile, "w+b");
        if ( !_hCacheFile )
        {
            dwRet = GetLastError();
            DBG(D_WARN, (L"%s: create cache file(%hs) failed: %d\n", FULL_FUNC_NAME, szCacheFile, dwRet));
            goto err2;
        }
    }

    dwRet = GetFileLength(_hCacheFile, qwSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: GetFileLength(%hs) failed: %d\n", FULL_FUNC_NAME, szCacheFile, dwRet));
        goto err3;
    }

    if ( (UINT)qwSize != _nBytesCnt )
    {
        DBG(D_INFO, (L"%s: file size: %lld, need: %d\n", FULL_FUNC_NAME, qwSize, _nBytesCnt));
        dwRet = SetFileLength(_hCacheFile, _nBytesCnt);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: SetFileLength(%hs) failed: %d\n", FULL_FUNC_NAME, szCacheFile, dwRet));
            goto err3;
        }
    }

    _pCacheMapping = new(std::nothrow) FileMapping();
    if ( !_pCacheMapping )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc _pCacheMapping failed\n", FULL_FUNC_NAME));
        goto err3;
    }

    dwRet = _pCacheMapping->Init(_hCacheFile, _nBytesCnt, false);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: _pCacheMapping init(%d) failed: %d\n", FULL_FUNC_NAME, _nBytesCnt, dwRet));
        goto err4;
    }

    dwRet = _pCacheMapping->Map(0, _nBytesCnt, true, true, _pByteArray);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: _pCacheMapping map(%d) failed: %d\n", FULL_FUNC_NAME, _nBytesCnt, dwRet));
        goto err5;
    }

    // see issues_ntfs_bm.txt
//    if ( (UINT)qwSize != _nBytesCnt )
    if ( (UINT)qwSize < _nBytesCnt )
    {
        DBG(D_WARN, (L"%s: cache file(%hs) invalid, zero it\n", FULL_FUNC_NAME, szCacheFile));
        memset(_pByteArray, 0, _nBytesCnt);
        _pCacheMapping->Flush(_pByteArray, _nBytesCnt);
    }

    for ( UINT i = 0; i < nBitsCnt; ++i )
    {
        if ( IsBitSet(i) )
        {
            ++_nSetsCnt;
            OnSetBit(i);
        }
    }

    return ERROR_SUCCESS;

err5:
    _pCacheMapping->Fini();

err4:
    delete _pCacheMapping;
    _pCacheMapping = NULL;

err3:
    fclose(_hCacheFile);
    _hCacheFile = NULL;

err2:
    Fini();

err1:
    return dwRet;
}

void BitArray::ReInit(bool bSetAll)
{
    DBG(D_INFO, (L"%s: bSetAll=%d\n", FULL_FUNC_NAME, bSetAll));

    memset(_pByteArray, (bSetAll ? 0xFF : 0), _nBytesCnt);
    if ( _pCacheMapping )
    {
        _pCacheMapping->Flush(_pByteArray, _nBytesCnt);
    }

    _nSetsCnt = bSetAll ? _nBitsCnt : 0;

    if ( _bCalcMask )
    {
        InitMatrixParam(bSetAll);
    }
}

void BitArray::Fini()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _hCacheFile )
    {
        assert(_pByteArray);
        assert(_pCacheMapping);
        _pCacheMapping->Unmap(_pByteArray, _nBytesCnt);

        _pCacheMapping->Fini();
        delete _pCacheMapping;
        _pCacheMapping = NULL;

        fclose(_hCacheFile);
        _hCacheFile = NULL;
    }
    else
    {
        delete []_pByteArray;
    }

    _pByteArray = NULL;
}

UINT BitArray::GetBitsCount() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return _nBitsCnt;
}

UINT BitArray::GetBytesCount() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return _nBytesCnt;
}

void BitArray::SetBit(UINT nPos)
{
    UINT  nIdx;
    BYTE  bMask;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nPos));

    Pos2ArrayMask(nPos, nIdx, bMask);

    if ( !ISBITSET(nIdx, bMask) )
    {
        assert(_nSetsCnt < _nBitsCnt);
        ++_nSetsCnt;

        SETBIT(nIdx, bMask);
        OnSetBit(nPos);
    }
}

void BitArray::SetBit(UINT nPos, bool &rbMaskUpd)
{
    QWORD  qwAndOld, qwOrOld;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nPos));

    qwAndOld = _qwMaskAnd;
    qwOrOld = _qwMaskOr;

    SetBit(nPos);

    rbMaskUpd = (qwAndOld != _qwMaskAnd) || (qwOrOld != _qwMaskOr);
}

void BitArray::ClearBit(UINT nPos)
{
    UINT  nIdx;
    BYTE  bMask;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nPos));

    Pos2ArrayMask(nPos, nIdx, bMask);

    if ( ISBITSET(nIdx, bMask) )
    {
        assert(_nSetsCnt > 0);
        --_nSetsCnt;

        CLEARBIT(nIdx, bMask);
        OnClearBit(nPos);
    }
}

void BitArray::ClearBit(UINT nPos, bool &rbMaskUpd)
{
    QWORD  qwAndOld, qwOrOld;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nPos));

    qwAndOld = _qwMaskAnd;
    qwOrOld = _qwMaskOr;

    ClearBit(nPos);

    rbMaskUpd = (qwAndOld != _qwMaskAnd) || (qwOrOld != _qwMaskOr);
}

bool BitArray::TestAndSetBit(UINT nPos)
{
    UINT  nIdx;
    BYTE  bMask;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nPos));

    Pos2ArrayMask(nPos, nIdx, bMask);

    if ( ISBITSET(nIdx, bMask) )
    {
        return false;
    }

    assert(_nSetsCnt < _nBitsCnt);
    ++_nSetsCnt;

    SETBIT(nIdx, bMask);
    OnSetBit(nPos);

    return true;
}

bool BitArray::TestAndClearBit(UINT nPos)
{
    UINT  nIdx;
    BYTE  bMask;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nPos));

    Pos2ArrayMask(nPos, nIdx, bMask);

    if ( !ISBITSET(nIdx, bMask) )
    {
        return false;
    }

    assert(_nSetsCnt > 0);
    --_nSetsCnt;

    CLEARBIT(nIdx, bMask);
    OnClearBit(nPos);

    return true;
}

bool BitArray::IsBitSet(UINT nPos) const
{
    UINT  nIdx;
    BYTE  bMask;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nPos));

    Pos2ArrayMask(nPos, nIdx, bMask);

    return ISBITSET(nIdx, bMask);
}

UINT BitArray::GetSetsCount() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return _nSetsCnt;
}

bool BitArray::IsAllSet() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return (_nSetsCnt == _nBitsCnt);
}

bool BitArray::IsAllClear() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return (!_nSetsCnt);
}

UINT BitArray::GetDegree() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_nSetsCnt <= _nBitsCnt);

    return (_nSetsCnt * 100 / _nBitsCnt);
}

void BitArray::GetCompactMask(QWORD &rqwMaskAnd, QWORD &rqwMaskOr) const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_bCalcMask);

    rqwMaskAnd = _qwMaskAnd;
    rqwMaskOr = _qwMaskOr;
}

DWORD BitArray::GetFullMask(PBYTE pByteBuf, UINT &rnSize) const
{
    UINT  nBufSize;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    nBufSize = rnSize;
    rnSize = _nBytesCnt;

    if ( nBufSize < _nBytesCnt )
    {
        DBG(D_WARN, (L"%s: need %d\n", FULL_FUNC_NAME, _nBytesCnt));
        assert(0);
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    memcpy(pByteBuf, _pByteArray, _nBytesCnt);

    return ERROR_SUCCESS;
}

DWORD BitArray::DoInit(UINT nBitsCnt, bool bAlloc, bool bSetAll, bool bCalcMask)
{
    DBG(D_INFO, (L"%s: nBitsCnt=%d, bSetAll=%d\n", FULL_FUNC_NAME, nBitsCnt, bSetAll));

    assert(!_pByteArray);

    _nBitsCnt = nBitsCnt;
    _nBytesCnt = (nBitsCnt + bitsizeof(BYTE) - 1) / bitsizeof(BYTE);

    if ( bAlloc )
    {
        _pByteArray = new(std::nothrow) BYTE[_nBytesCnt];
        if ( !_pByteArray )
        {
            DBG(D_WARN, (L"%s: alloc _pByteArray[%d] failed\n", FULL_FUNC_NAME, _nBytesCnt));
            return ERROR_OUTOFMEMORY;
        }
        memset(_pByteArray, (bSetAll ? 0xFF : 0), _nBytesCnt);
    }

    _nSetsCnt = bSetAll ? nBitsCnt : 0;

    _bCalcMask = bCalcMask;
    if ( bCalcMask )
    {
        InitMatrixParam(bSetAll);
    }

    return ERROR_SUCCESS;
}

void BitArray::SETBIT(UINT nIdx, BYTE bMask)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(nIdx < _nBytesCnt);

    _pByteArray[nIdx] |= bMask;
}

void BitArray::CLEARBIT(UINT nIdx, BYTE bMask)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(nIdx < _nBytesCnt);

    _pByteArray[nIdx] &= (~bMask);
}

bool BitArray::ISBITSET(UINT nIdx, BYTE bMask) const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(nIdx < _nBytesCnt);

    return ((_pByteArray[nIdx] & bMask) == bMask);
}

void BitArray::OnSetBit(UINT nPos)
{
    UINT  nRow;
    QWORD  qwMask;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _bCalcMask )
    {
        Pos2MatrixMask(nPos, nRow, qwMask);

        _MaskArray[nRow] |= qwMask;

        _qwMaskOr |= qwMask;

        _qwMaskAnd = ~0ULL;
        if ( _nRemain )
        {
            _MaskArray[_MaskArray.size() - 1] |= ((~0ULL) << _nRemain);
        }
        for ( size_t i = 0, nRowCnt = _MaskArray.size(); i < nRowCnt; ++i )
        {
            _qwMaskAnd &= _MaskArray[i];
        }
    }
}

void BitArray::OnClearBit(UINT nPos)
{
    UINT  nRow;
    QWORD  qwMask;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _bCalcMask )
    {
        Pos2MatrixMask(nPos, nRow, qwMask);
        qwMask = ~qwMask;

        _MaskArray[nRow] &= qwMask;

        _qwMaskAnd &= qwMask;

        _qwMaskOr = 0;
        if ( _nRemain )
        {
            _MaskArray[_MaskArray.size() - 1] &= ((~0ULL) >> (bitsizeof(QWORD) - _nRemain));
        }
        for ( size_t i = 0, nRowCnt = _MaskArray.size(); i < nRowCnt; ++i )
        {
            _qwMaskOr |= _MaskArray[i];
        }
    }
}

void BitArray::Pos2ArrayMask(UINT nPos, UINT &rnIdx, BYTE &rbMask) const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(nPos < _nBitsCnt);

    rnIdx = nPos / 8;
    rbMask = 1 << (nPos % 8);
}

void BitArray::InitMatrixParam(bool bSetAll)
{
    UINT  nRow, nRemain;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    nRow = (_nBitsCnt + bitsizeof(QWORD) - 1) / bitsizeof(QWORD);
    nRemain = _nBitsCnt % bitsizeof(QWORD);

    _MaskArray.resize(nRow, (bSetAll ? (~0ULL) : 0));

    _qwMaskAnd = bSetAll ? (~0ULL) : 0;
    _qwMaskOr = _qwMaskAnd;

    _nRemain = nRemain;
    _nRightH = (_nBitsCnt + bitsizeof(QWORD)) / bitsizeof(QWORD);
    _nLeftH = _nRightH - 1;
    _nDivide = _nRemain * _nRightH;
}

void BitArray::Pos2MatrixMask(UINT nPos, UINT &rnRow, QWORD &rqwMask)
{
    UINT  nCol;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(nPos < _nBitsCnt);

    if ( nPos < _nDivide )
    {
        rnRow = nPos % _nRightH;
        nCol = nPos / _nRightH;
    }
    else
    {
        nPos -= _nDivide;
        rnRow = nPos % _nLeftH;
        nCol = nPos / _nLeftH + _nRemain;
    }

    rqwMask = 1ULL << nCol;
}
