/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "ExecutorPool.h"


typedef struct _THREAD_ARG
{
    UINT                        nIdx;
    ExecutorPool              * pThis;
}
THREAD_ARG, *PTHREAD_ARG;


Runnable::Runnable()
    : _pEvent(NULL)
    , _nSigIdx(INVALID_PCOND_IDX)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

Runnable::~Runnable()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

void Runnable::Bind(PCondEvent *pEvent, UINT nSigIdx)
{
    DBG(D_INFO, (L"%s: idx(%d)\n", FULL_FUNC_NAME, nSigIdx));

    assert(!_pEvent);
    _pEvent = pEvent;

    assert(_nSigIdx == INVALID_PCOND_IDX);
    _nSigIdx = nSigIdx;
}

void Runnable::Unbind()
{
    assert(_nSigIdx != INVALID_PCOND_IDX);
    _nSigIdx = INVALID_PCOND_IDX;

    assert(_pEvent);
    _pEvent = NULL;
}

void Runnable::ReqExec()
{
    DBG(D_VERB, (L"%s: idx(%d)\n", FULL_FUNC_NAME, _nSigIdx));

    assert(_pEvent);
    assert(_nSigIdx != INVALID_PCOND_IDX);
    _pEvent->Signal(_nSigIdx);
}


ExecutorPool::ExecutorPool()
    : _pEvent(NULL)
    , _bQuit(false)
    , _hWorkThreads(0)
    , _RunnList(0)
    , _RunnListMate(true)
    , _dwRunnMagic(0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

ExecutorPool::~ExecutorPool()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

DWORD ExecutorPool::Load(UINT nMaxRunnNum, UINT nExTaskNum)
{
    DWORD  dwRet;
    UINT  i;
    PRUNNABLE_REC  pRunnRec;
    pthread_t  hThread;
    PTHREAD_ARG  pThreadArg;

    DBG(D_LOG, (L"%s: MaxRunnNum=%d, ExTaskNum=%d\n", FULL_FUNC_NAME, nMaxRunnNum, nExTaskNum));

    nExTaskNum += 1;

    _pEvent = new PCondEvent(nMaxRunnNum);

    _bQuit = false;
    _hWorkThreads.resize(nExTaskNum);

    _RunnList.resize(nMaxRunnNum);
    _RunnListMate.Init(nMaxRunnNum);

    for ( i = 0; i < nMaxRunnNum; ++i )
    {
        pRunnRec = &_RunnList.at(i);
        DO_IF_DBG(pRunnRec->bUsed = false);
        pRunnRec->bValid = false;
    }

    for ( i = 0; i < nExTaskNum; ++i )
    {
        SET_PTH_INVALID(_hWorkThreads.at(i));
    }

    for ( i = 0; i < nExTaskNum; ++i )
    {
        pThreadArg = new(std::nothrow) THREAD_ARG;
        if ( !pThreadArg )
        {
            dwRet = ERROR_OUTOFMEMORY;
            DBG(D_ERROR, (L"%s: alloc pThreadArg failed\n", FULL_FUNC_NAME));
            goto err;
        }

        pThreadArg->nIdx = i;
        pThreadArg->pThis = this;
        if ( pthread_create(&hThread, NULL, WorkThreadProc, pThreadArg) )
        {
            dwRet = GetLastError();
            DBG(D_ERROR, (L"%s: create work thread[#%d] failed: %d\n", FULL_FUNC_NAME, i, dwRet));
            delete pThreadArg;
            goto err;
        }

        _hWorkThreads.at(i) = hThread;
    }

    return ERROR_SUCCESS;

err:
    Unload();

    return dwRet;
}

void ExecutorPool::Unload()
{
    pthread_t  *phThread;
    UINT  i, nSize;

    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    _bQuit = true;
    _pEvent->Release();

    for ( i = 0, nSize = (UINT)_hWorkThreads.size(); i < nSize; ++i )
    {
        phThread = &_hWorkThreads.at(i);
        if ( IS_PTH_VALID(*phThread) )
        {
            pthread_join(*phThread, NULL);
            SET_PTH_INVALID(*phThread);
        }
    }
    _hWorkThreads.resize(0);

    _RunnList.resize(0);
    _RunnListMate.Fini();

    delete _pEvent;
}

DWORD ExecutorPool::RegisterRunnable(Runnable *pRunnable, DWORD &rdwRunnId)
{
    DWORD  dwRet;
    UINT  nIdx;
    DWORD  dwMagic;
    PRUNNABLE_REC  pRunnRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rdwRunnId = INVALID_RUNNABLE_ID;

    // _RunnListMate has internal lock
    dwRet = _RunnListMate.GetFreeIndex(nIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: reach the max exec limit\n", FULL_FUNC_NAME));
        return dwRet;
    }

    pRunnRec = &_RunnList.at(nIdx);
    assert(!pRunnRec->bUsed);
    DO_IF_DBG(pRunnRec->bUsed = true);

    dwMagic = InterlockedIncrement((LONG*)&_dwRunnMagic);

    pRunnRec->wMagic = ADJUST_WMAGICNUM(dwMagic);
    pRunnRec->CtlReq = RUNN_REQ_NONE;
    pRunnRec->nSlotIdx = INVALID_INDEX_VALUE;
    pRunnRec->pRunnable = pRunnable;
    pRunnRec->bValid = true;

    // see issues_connid.txt
    rdwRunnId = MAKE_DWID(nIdx, pRunnRec->wMagic);

    pRunnable->Bind(_pEvent, nIdx);
    DBG(D_INFO, (L"%s: idx(%d)\n", FULL_FUNC_NAME, nIdx));

    return ERROR_SUCCESS;
}

DWORD ExecutorPool::UnregisterRunnable(DWORD dwRunnId)
{
    UINT  nRunnIdx;
    WORD  wMagic;
    PRUNNABLE_REC  pRunnRec;

    nRunnIdx = WIDX_FROM_DWID(dwRunnId);
    wMagic = WMAGICNUM_FROM_DWID(dwRunnId);

    DBG(D_INFO, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nRunnIdx, wMagic));

    if ( nRunnIdx >= _RunnList.size() )
    {
        DBG(D_WARN, (L"%s: invalid runn id\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_INVALID_PARAMETER;
    }

    pRunnRec = &_RunnList.at(nRunnIdx);
    if ( wMagic != pRunnRec->wMagic )
    {
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nRunnIdx));
        assert(0);
        return ERROR_NOT_FOUND;
    }
    assert(pRunnRec->bUsed && pRunnRec->bValid);

    pRunnRec->pRunnable->Unbind();

    pRunnRec->CtlReq = RUNN_REQ_REMOVE;
    _pEvent->Signal(nRunnIdx);
    while ( pRunnRec->CtlReq != RUNN_REQ_DONE )
    {
        DBG(D_INFO, (L"%s: waiting idx(%d) ...\n", FULL_FUNC_NAME, nRunnIdx));
        Sleep(10);
    }

    // see issues_executor3.txt
    assert(pRunnRec->nSlotIdx != INVALID_INDEX_VALUE);
    _pEvent->Reset(nRunnIdx, pRunnRec->nSlotIdx);

    pRunnRec->bValid = false;

    DO_IF_DBG(pRunnRec->bUsed = false);
    _RunnListMate.PutFreeIndex(nRunnIdx);

    return ERROR_SUCCESS;
}

DWORD ExecutorPool::WorkThread(UINT nThreadIdx)
{
    UINT  nSigIdx;
    PRUNNABLE_REC  pRunnRec;

    DBG(D_LOG, (L"%s[#%d]: enter\n", FULL_FUNC_NAME, nThreadIdx));

    while ( !_bQuit )
    {
        nSigIdx = _pEvent->Wait(nThreadIdx);

        if ( _bQuit )
        {
            break;
        }

        assert(nSigIdx != INVALID_PCOND_IDX);
        pRunnRec = &_RunnList.at(nSigIdx);
        assert(pRunnRec->bUsed && pRunnRec->bValid);

        while ( true )
        {
            // see issues_executor3.txt
            if ( pRunnRec->CtlReq == RUNN_REQ_REMOVE )
            {
                pRunnRec->nSlotIdx = nThreadIdx;
                DBG(D_INFO, (L"%s: idx(%d): REMOVE -> DONE\n", FULL_FUNC_NAME, nSigIdx));
                pRunnRec->CtlReq = RUNN_REQ_DONE;
                break;
            }

            assert(pRunnRec->pRunnable);
            pRunnRec->pRunnable->Execute();

            if ( _pEvent->Finish(nSigIdx, nThreadIdx) )
            {
                break;
            }
        }
    }

    DBG(D_LOG, (L"%s[#%d]: leave\n", FULL_FUNC_NAME, nThreadIdx));

    return ERROR_SUCCESS;
}

void *ExecutorPool::WorkThreadProc(void *arg)
{
    PTHREAD_ARG  pArg = reinterpret_cast<PTHREAD_ARG>(arg);

    pArg->pThis->WorkThread(pArg->nIdx);
    delete pArg;

    return NULL;
}


ExecutorCenter *ExecutorCenter::_instance = NULL;

ExecutorCenter *ExecutorCenter::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new ExecutorCenter();
    }
    return _instance;
}

void ExecutorCenter::Destroy()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

ExecutorCenter::ExecutorCenter()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

ExecutorCenter::~ExecutorCenter()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}
