/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "FileMapping.h"


FileMapping::FileMapping()
    : _hFile(NULL)
#ifdef _WIN32_
    , _hMapFile(NULL)
#endif
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

FileMapping::~FileMapping()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

#ifdef _WIN32_
    if ( _hMapFile )
    {
        assert(0);
        CloseHandle(_hMapFile);
    }
#endif

    assert(!_hFile);
}

DWORD FileMapping::Init(FILE *hFile, QWORD qwFileSize, bool bReadOnly)
{
    DBG(D_INFO, (L"%s: %lld\n", FULL_FUNC_NAME, qwFileSize));

    assert(!_hFile);
    _hFile = hFile;

#ifdef _WIN32_
    DWORD  dwRet;
    HANDLE  hWinFile = (HANDLE)_get_osfhandle(_fileno(hFile));
    DWORD  flProtect = bReadOnly ? PAGE_READONLY : PAGE_READWRITE;

    assert(!_hMapFile);
    // dwMaximumSize MUST be 0, means the whole file CAN be mapped
    _hMapFile = CreateFileMapping(hWinFile, NULL, flProtect, HIDWORD(qwFileSize), LODWORD(qwFileSize), NULL);
    if ( !_hMapFile )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: CreateFileMapping failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }
#endif

    return ERROR_SUCCESS;
}

void FileMapping::Fini()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

#ifdef _WIN32_
    if ( _hMapFile )
    {
        CloseHandle(_hMapFile);
        _hMapFile = NULL;
    }
#endif

    _hFile = NULL;
}

DWORD FileMapping::Map(QWORD qwOffset, DWORD dwSize, bool bRead, bool bWrite, PBYTE &rpBuf)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %#.8x @ %#.16llx\n", FULL_FUNC_NAME, dwSize, qwOffset));

#ifdef _WIN32_
    DWORD  dwFlags = 0;
    if ( bRead ) dwFlags |= FILE_MAP_READ;
    if ( bWrite ) dwFlags |= FILE_MAP_WRITE;
    rpBuf = (PBYTE)MapViewOfFile(_hMapFile, dwFlags, HIDWORD(qwOffset), LODWORD(qwOffset), dwSize);
    if ( !rpBuf )
#else
    int  prot = 0;
    if ( bRead ) prot |= PROT_READ;
    if ( bWrite ) prot |= PROT_WRITE;
    // MAP_PRIVATE: Create a private copy-on-write mapping.
    rpBuf = (PBYTE)mmap64(NULL, dwSize, prot, MAP_SHARED/*MAP_PRIVATE*/, _fileno(_hFile), qwOffset);
    if ( rpBuf == MAP_FAILED )
#endif
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: (%#.8x @ %#.16llx) failed: %d\n", FULL_FUNC_NAME, dwSize, qwOffset, dwRet));
        goto err;
    }

    return ERROR_SUCCESS;

err:
    rpBuf = NULL;

    return dwRet;
}

DWORD FileMapping::Unmap(PBYTE pBuf, DWORD dwSize)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %#.8x @ 0x%p\n", FULL_FUNC_NAME, dwSize, pBuf));

#ifdef _WIN32_
    if ( !UnmapViewOfFile(pBuf) )
#else
    if ( munmap(pBuf, dwSize) )
#endif
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: (%#.8x @ 0x%p) failed: %d\n", FULL_FUNC_NAME, dwSize, pBuf, dwRet));
        return dwRet;
    }

    return ERROR_SUCCESS;
}

DWORD FileMapping::Flush(PBYTE pBuf, DWORD dwSize, bool bSync)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %#.8x @ 0x%p\n", FULL_FUNC_NAME, dwSize, pBuf));

#ifdef _WIN32_
    if ( !FlushViewOfFile(pBuf, dwSize) )
#else
    if ( msync(pBuf, dwSize, (bSync ? MS_SYNC : MS_ASYNC)) )
#endif
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: (%#.8x @ 0x%p) failed: %d\n", FULL_FUNC_NAME, dwSize, pBuf, dwRet));
        return dwRet;
    }

    return ERROR_SUCCESS;

}
