/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#ifdef _LINUX_
#include <termios.h>
#endif
#include "MiscUtils.h"


#if defined(_ANDROID_)
#include <android/log.h>
void LOGI(const wchar_t *fmt, ...)
{
    va_list  arg;
    wchar_t  wcs[200];
    char  mbs[200];

    va_start(arg, fmt);
    vswprintf(wcs, _countof(wcs), fmt, arg);
    va_end(arg);

    wcstombs(mbs, wcs, _countof(mbs));

    __android_log_print(ANDROID_LOG_INFO, TAG, mbs);
}
#endif

#ifdef _WIN32_
#ifdef _NO_STD_DBG_
void TraceW(const wchar_t *fmt, ...)
{
    va_list  argptr;
    wchar_t  buf[1000];

    va_start(argptr, fmt);
    vswprintf_s(buf, _countof(buf), fmt, argptr);
    va_end(argptr);

    OutputDebugString(buf);
}
#endif
#endif

DWORD Round2Power(DWORD dwVal)
{
    dwVal--;
    dwVal |= (dwVal >> 1);
    dwVal |= (dwVal >> 2);
    dwVal |= (dwVal >> 4);
    dwVal |= (dwVal >> 8);
    dwVal |= (dwVal >> 16);
    dwVal++;

    return dwVal;
}

bool IsMCastAddr(const PSOCK_ADDR pAddr)
{
    return ((pAddr->IPv4Addr.dwIP & 0xE0) == 0xE0);
}

void addr2str(const PSOCK_ADDR pAddr, char *szBuf, UINT nSize)
{
    _snprintf(szBuf, nSize, "%d.%d.%d.%d:%d",
        LOBYTE(LOWORD(pAddr->IPv4Addr.dwIP)),
        HIBYTE(LOWORD(pAddr->IPv4Addr.dwIP)),
        LOBYTE(HIWORD(pAddr->IPv4Addr.dwIP)),
        HIBYTE(HIWORD(pAddr->IPv4Addr.dwIP)),
        NTOHS(pAddr->IPv4Addr.wPort));
}

bool operator == (const SOCK_ADDR &lhs, const SOCK_ADDR &rhs)
{
    return ((lhs.IPv4Addr.dwIP == rhs.IPv4Addr.dwIP) && (lhs.IPv4Addr.wPort == rhs.IPv4Addr.wPort));
}

bool operator != (const SOCK_ADDR &lhs, const SOCK_ADDR &rhs)
{
    return ((lhs.IPv4Addr.dwIP != rhs.IPv4Addr.dwIP) || (lhs.IPv4Addr.wPort != rhs.IPv4Addr.wPort));
}

bool operator == (const SHA1_VAL &lhs, const SHA1_VAL &rhs)
{
    for ( UINT i = 0; i < SHA1_VAL_LEN; ++i )
    {
        if ( lhs.bVal[i] != rhs.bVal[i] )
        {
            return false;
        }
    }
    return true;
}

bool operator != (const SHA1_VAL &lhs, const SHA1_VAL &rhs)
{
    for ( UINT i = 0; i < SHA1_VAL_LEN; ++i )
    {
        if ( lhs.bVal[i] != rhs.bVal[i] )
        {
            return true;
        }
    }
    return false;
}

// basename@linux, PathFindFileName@windows
bool PathHasFileName(const char *path)
{
    size_t  len;
    char  ch;

    len = strlen(path);
    if ( !len )
    {
        return false;
    }

    ch = path[--len];

    if ( (ch == '/') || (ch == '\\') )
    {
        return false;
    }

    return true;
}

void GetFileName(const char *path, char *buf, int size)
{
    const char  *p;

    p = strrchr(path, '\\');
    if ( !p )
    {
        p = strrchr(path, '/');
        if ( !p )
        {
            p = path;
        }
    }
    if ( (*p == '/') || (*p == '\\') )
    {
        ++p;
    }

    strncpy(buf, p, size);
}

// dirname@linux, PathFindFileName@windows
void GetFilePath(const char *path, char *buf, int size)
{
    const char  *p;
    int  len;

    p = strrchr(path, '\\');
    if ( !p )
    {
        p = strrchr(path, '/');
        if ( !p )
        {
            p = path;
        }
    }
    if ( (*p == '/') || (*p == '\\') )
    {
        ++p;
    }

    len = min((int)(p - path), size - 1);
    memcpy(buf, path, len);
    buf[len] = 0;
}

DWORD IsRegularFile(FILE *hFile, bool &rbRegular)
{
    DWORD  dwRet;
    int  fd;
    struct stat  buf;

    fd = _fileno(hFile);
    if ( fd < 0 )
    {
        dwRet = GetLastError();
        return dwRet;
    }

    if ( fstat(fd, &buf) != 0 )
    {
        dwRet = GetLastError();
        return dwRet;
    }

    rbRegular = S_ISREG(buf.st_mode);

    return ERROR_SUCCESS;
}

DWORD GetFileLength(FILE *hFile, QWORD &rqwSize)
{
    DWORD  dwRet;
    off64_t  lCurPos, lSize;

    lCurPos = ftello(hFile);
    if ( lCurPos < 0 )
    {
        dwRet = GetLastError();
        return dwRet;
    }

    if ( fseeko(hFile, 0, SEEK_END) )
    {
        dwRet = GetLastError();
        return dwRet;
    }

    lSize = ftello(hFile);
    if ( lSize < 0 )
    {
        dwRet = GetLastError();
        return dwRet;
    }

    if ( fseeko(hFile, lCurPos, SEEK_SET) )
    {
        dwRet = GetLastError();
        return dwRet;
    }

    rqwSize = lSize;

    return ERROR_SUCCESS;
}

DWORD SetFileLength(FILE *hFile, QWORD qwSize)
{
    DWORD  dwRet = ERROR_SUCCESS;
    int  fd;

    fd = _fileno(hFile);
    if ( fd < 0 )
    {
        dwRet = GetLastError();
        return dwRet;
    }

    // see issues_ftrunc.txt
    fflush(hFile);

#ifdef _WIN32_
//    dwRet = _chsize_s(fd, qwSize);
// see issues_mmap.txt
    DWORD  dwTemp;
    LONG  lOffLo, lOffHi;
    HANDLE  hWinFile = (HANDLE)_get_osfhandle(fd);
    DeviceIoControl(hWinFile, FSCTL_SET_SPARSE, NULL, 0, NULL, 0, &dwTemp, NULL);
    lOffLo = (LONG)LODWORD(qwSize);
    lOffHi = (LONG)HIDWORD(qwSize);
    if ( SetFilePointer(hWinFile, lOffLo, &lOffHi, FILE_BEGIN) == INVALID_SET_FILE_POINTER )
    {
        dwRet = GetLastError();
    }
#else
    // see issues_ftrunc.txt
    if ( ftruncate(fd, qwSize) )
    {
        dwRet = GetLastError();
    }
#endif
    if ( dwRet != ERROR_SUCCESS )
    {
        return dwRet;
    }

    if ( fflush(hFile) )
    {
        dwRet = GetLastError();
        return dwRet;
    }

    return ERROR_SUCCESS;
}

#if defined(_ANDROID_) || defined(_LINUX_)
DWORD GetTickCount()
{
    struct timespec  ts;

    clock_gettime(CLOCK_MONOTONIC, &ts);

    return  ts.tv_sec * 1000 + ts.tv_nsec / 1000000;
}
#endif

#if defined(_WINCE_) || defined(_WIN32_)
int clock_gettime(int, struct timespec *tp)
{
    QWORD  qwCurTime;

    GetSystemTimeAsFileTime((FILETIME*)&qwCurTime);

    qwCurTime -= 116444736000000000i64;  //1jan1601 to 1jan1970
    tp->tv_sec = qwCurTime / 10000000i64;
    tp->tv_nsec = qwCurTime % 10000000i64 *100;
 
    return 0;
}
#endif

DWORD GetSecondCount()
{
    struct timespec  ts;

    clock_gettime(CLOCK_MONOTONIC, &ts);

    return  (DWORD)ts.tv_sec;
}

QWORD GetPerformanceCounter()
{
#ifdef _WIN32_
    LARGE_INTEGER  liCnt;
    QueryPerformanceCounter(&liCnt);
    return liCnt.QuadPart;
#else
    struct timespec  ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    return  ((QWORD)ts.tv_sec*1000000000ULL + (QWORD)ts.tv_nsec);
#endif
}

QWORD GetPerformanceFrequency()
{
#ifdef _WIN32_
    LARGE_INTEGER  liFreq;
    QueryPerformanceFrequency(&liFreq);
    return liFreq.QuadPart;
#else
    return 1000000000ULL;
#endif
}

// -1: ts1<ts2, 0: ts1=ts2, 1: ts1>ts2
int TimeCompare(const timespec &ts1, const timespec &ts2)
{
    if ( ts1.tv_sec < ts2.tv_sec )
    {
        return -1;
    }
    else if (ts1.tv_sec > ts2.tv_sec)
    {
        return 1;
    }

    if ( ts1.tv_nsec < ts2.tv_nsec )
    {
        return -1;
    }
    else if (ts1.tv_nsec > ts2.tv_nsec)
    {
        return 1;
    }

    return 0;

}

void TimeAdd(timespec &ts, DWORD dwMilliseconds)
{
    ts.tv_sec += (dwMilliseconds / 1000);
    ts.tv_nsec += (dwMilliseconds % 1000) * 1000000;

    if ( ts.tv_nsec > 1000000000 )
    {
        ts.tv_nsec -= 1000000000;
        ts.tv_sec += 1;
    }
}

void InitStdIn()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

#ifdef _WIN32_
    SetConsoleMode(GetStdHandle(STD_INPUT_HANDLE), ENABLE_ECHO_INPUT|ENABLE_INSERT_MODE|ENABLE_LINE_INPUT);
#else
    struct termios  t;
    tcgetattr(STDIN_FILENO, &t);
    t.c_lflag &= ~ICANON;
    t.c_cc[VMIN] = 1;
    tcsetattr(STDIN_FILENO, TCSANOW, &t);
#endif
}

DWORD ReadStdIn(char *pBuffer, UINT nBufSize, DWORD dwMilliseconds)
{
#ifdef _WIN32_
    DWORD  dwRet = ERROR_SUCCESS;
    UINT  nPos;
    char  ch;
    DWORD  dwTimeout;
    HANDLE  hStdIn = GetStdHandle(STD_INPUT_HANDLE);
    INPUT_RECORD  irInBuf; 
    DWORD  dwNumRead;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    nPos = 0;
    dwTimeout = dwMilliseconds;
    while ( true )
    {
        if ( nPos >= nBufSize )
        {
            dwRet = ERROR_NOT_ENOUGH_MEMORY;
            break;
        }

        dwRet = WaitForSingleObject(hStdIn, dwTimeout);

        if ( dwRet == WAIT_FAILED )
        {
            dwRet = GetLastError();
            DBG(D_WARN, (L"%s: WaitForSingleObject failed: %d\n", FULL_FUNC_NAME, dwRet));
            break;
        }

        if ( dwRet == WAIT_TIMEOUT )
        {
            break;
        }

        ReadConsoleInput(hStdIn, &irInBuf, 1, &dwNumRead);
        if ( (irInBuf.EventType == KEY_EVENT) && irInBuf.Event.KeyEvent.bKeyDown )
        {
            ch = irInBuf.Event.KeyEvent.uChar.AsciiChar;
            if ( !ch )
            {
                continue;
            }

            putchar(ch);

            if ( ch == '\r' )
            {
                pBuffer[nPos] = 0;
                break;
            }

            pBuffer[nPos++] = ch;

            dwTimeout = INFINITE;
        }
    }

    return dwRet;
#else
#if 1
    DWORD  dwRet;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    dwRet = WaitStdIn(dwMilliseconds);
    if ( dwRet != ERROR_SUCCESS )
    {
        goto out;
    }

    // see misc_stdin.txt
    if ( !fgets(pBuffer, nBufSize, stdin) )
    {
        dwRet = feof(stdin) ? ERROR_INVALID_STATE : GetLastError();
        DBG(D_WARN, (L"%s: fgets failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto out;
    }

out:
    return dwRet;
#else
    DWORD  dwRet = ERROR_SUCCESS;
    int  ret;
    UINT  nPos;
    char  ch;
    DWORD  dwTimeout;
    fd_set  rd_fds;
    timeval  tv;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    rnReadSize = 0;
    nPos = 0;
    dwTimeout = dwMilliseconds;
    while ( true )
    {
        if ( nPos >= nBufSize )
        {
            dwRet = ERROR_NOT_ENOUGH_MEMORY;
            DBG(D_WARN, (L"%s: not enough room\n", FULL_FUNC_NAME));
            break;
        }

        FD_ZERO(&rd_fds);
        FD_SET(STDIN_FILENO, &rd_fds);
        tv.tv_sec = dwTimeout / 1000;
        tv.tv_usec = (dwTimeout % 1000) * 1000;
        ret = select(STDIN_FILENO + 1, &rd_fds, NULL, NULL, &tv);
        if ( ret == SOCKET_ERROR )
        {
            dwRet = GetLastError();
            DBG(D_WARN, (L"%s: select failed: %d\n", FULL_FUNC_NAME, dwRet));
            break;
        }

        if ( ret == 0 )
        {
            dwRet = WAIT_TIMEOUT;
            break;
        }

        ch = getchar();

        if ( ch == '\n' )
        {
            pBuffer[nPos] = 0;
            break;
        }

        pBuffer[nPos++] = ch;
        ++rnReadSize;

        dwTimeout = INFINITE;
    }

    return dwRet;
#endif
#endif
}

DWORD WaitStdIn(DWORD dwMilliseconds)
{
#ifdef _WIN32_
    DWORD  dwRet;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    dwRet = WaitForSingleObject(GetStdHandle(STD_INPUT_HANDLE), dwMilliseconds);
    if ( dwRet == WAIT_FAILED )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: WaitForSingleObject failed: %d\n", FULL_FUNC_NAME, dwRet));
    }
    
    return dwRet;
#else
    DWORD  dwRet;
    int  ret;
    fd_set  rd_fds;
    timeval  tv;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    FD_ZERO(&rd_fds);
    FD_SET(STDIN_FILENO, &rd_fds);
    tv.tv_sec = dwMilliseconds / 1000;
    tv.tv_usec = (dwMilliseconds % 1000) * 1000;
    ret = select(STDIN_FILENO + 1, &rd_fds, NULL, NULL, &tv);
    if ( ret == SOCKET_ERROR )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: select failed: %d\n", FULL_FUNC_NAME, dwRet));
    }
    else
    {
        dwRet = (ret == 0) ? WAIT_TIMEOUT : ERROR_SUCCESS;
    }
    
    return dwRet;
#endif
}

void InitRand()
{
//    random_device  rd;

    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

//    srand(rd());
//    srand(time(NULL));
    srand(GetTickCount());
}

#ifdef _WIN32_
UINT RAND_BETWEEN(UINT nMin, UINT nMax)
{
    errno_t  err;
    UINT  nResult, nNumber;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    err = rand_s(&nNumber);
    if ( err )
    {
        nResult = (UINT)((double)rand() / (RAND_MAX+1.0) * (nMax - nMin) + nMin);
    }
    else
    {
        nResult = (UINT)((double)nNumber / (UINT_MAX+1.0) * (nMax - nMin) + nMin); 
    }

    return nResult;
}
#endif
