/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "PCondEvent.h"


PCondEvent::PCondEvent(UINT nSigNum)
    : _bQuit(false)
    , _SigSlotMap(nSigNum)
    , _SigPipe(nSigNum, false)
{
    PPCOND_SIGMAP_ITEM  pItem;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_init(&_hLock, NULL);
    pthread_cond_init(&_hCond, NULL);

    for ( UINT i = 0; i < nSigNum; ++i )
    {
        pItem = &_SigSlotMap.at(i);

        pItem->state = SIG_STATE_NONE;
        pItem->bSignaled = false;
        pItem->nSlotIdx = INVALID_PCOND_IDX;
    }
}

PCondEvent::~PCondEvent()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_cond_destroy(&_hCond);
    pthread_mutex_destroy(&_hLock);
}

void PCondEvent::Signal(UINT nSigIdx)
{
    volatile PCOND_SIGMAP_ITEM  *pItem;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(nSigIdx < _SigSlotMap.size());
    pItem = &_SigSlotMap.at(nSigIdx);

    pthread_mutex_lock(&_hLock);
    pItem->bSignaled = true;
    if ( pItem->state == SIG_STATE_NONE )
    {
        assert(pItem->nSlotIdx == INVALID_PCOND_IDX);

        // set state first, because once call _SigPipe.PushAtBack,
        // Wait will break the loop, assert pItem->state will failed
        pItem->state = SIG_STATE_SUBMITED;

        _SigPipe.PushAtBack(nSigIdx);
        pthread_cond_signal(&_hCond);
    }
    pthread_mutex_unlock(&_hLock);
}

UINT PCondEvent::Wait(UINT nSlotIdx, bool bNonBlock)
{
    volatile PCOND_SIGMAP_ITEM  *pItem;
    UINT  nSigIdx;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    nSigIdx = INVALID_PCOND_IDX;

    pthread_mutex_lock(&_hLock);

    if ( !bNonBlock )
    {
        while ( (!_bQuit) && _SigPipe.IsEmpty() )
        {
            pthread_cond_wait(&_hCond, &_hLock);
        }
    }

    if ( !_SigPipe.IsEmpty() )
    {
        nSigIdx = _SigPipe.PopAtFront();
    }

    pthread_mutex_unlock(&_hLock);

    // no test & set, so no need lock
    if ( nSigIdx != INVALID_PCOND_IDX )
    {
        assert(nSigIdx < _SigSlotMap.size());
        pItem = &_SigSlotMap.at(nSigIdx);

        assert(pItem->state == SIG_STATE_SUBMITED);
        assert(pItem->bSignaled);
        assert(pItem->nSlotIdx == INVALID_PCOND_IDX);

        pItem->state = SIG_STATE_PROCESSING;
        pItem->bSignaled = false;
        pItem->nSlotIdx = nSlotIdx;
    }

    return nSigIdx;
}

bool PCondEvent::Finish(UINT nSigIdx, UINT nSlotIdx)
{
    volatile PCOND_SIGMAP_ITEM  *pItem;
    bool  bFinished;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(nSigIdx < _SigSlotMap.size());
    pItem = &_SigSlotMap.at(nSigIdx);

    assert(pItem->state == SIG_STATE_PROCESSING);
    assert(pItem->nSlotIdx == nSlotIdx);

    pthread_mutex_lock(&_hLock);
    bFinished = !pItem->bSignaled;
    if ( bFinished )
    {
        pItem->state = SIG_STATE_DONE;
        pItem->nSlotIdx = INVALID_PCOND_IDX;
    }
    else
    {
        pItem->bSignaled = false;
    }
    pthread_mutex_unlock(&_hLock);

    return bFinished;
}

void PCondEvent::Reset(UINT nSigIdx, UINT nSlotIdx)
{
    PPCOND_SIGMAP_ITEM  pItem;

    DBG(D_INFO, (L"%s: sig: %d, slot: %d\n", FULL_FUNC_NAME, nSigIdx, nSlotIdx));

    assert(nSigIdx < _SigSlotMap.size());
    pItem = &_SigSlotMap.at(nSigIdx);

    assert(pItem->nSlotIdx == nSlotIdx);

    // MUST guarantee no one call "Signal" during "Reset"
    pItem->state = SIG_STATE_NONE;
    pItem->bSignaled = false;
    pItem->nSlotIdx = INVALID_PCOND_IDX;
}

void PCondEvent::Release()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_lock(&_hLock);

    _bQuit = true;

    pthread_cond_broadcast(&_hCond);

    pthread_mutex_unlock(&_hLock);
}
