/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "ProducerQueue.h"


ProducerQueue::ProducerQueue(UINT nQueueSize, UINT nBufferSize, bool bEnableLock)
    : _nQueueSize(nQueueSize)
    , _nBufferSize(nBufferSize)
    , _bEnableLock(bEnableLock)
    , _FreeList(nQueueSize, false)
    , _DataList(nQueueSize, false)
{
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _bEnableLock )
    {
        pthread_mutex_init(&_hLock, NULL);
    }

    if ( nBufferSize )
    {
        for ( UINT i = 0; i < nQueueSize; ++i )
        {
            pbi = new BufferItem(nBufferSize);
            _FreeList.PushAtBack(pbi);
        }
    }
}

ProducerQueue::~ProducerQueue()
{
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    while ( !_FreeList.IsEmpty() )
    {
        pbi = _FreeList.PopAtFront();
        delete pbi;
    }

    while ( !_DataList.IsEmpty() )
    {
        pbi = _DataList.PopAtFront();
        delete pbi;
    }

    if ( _bEnableLock )
    {
        pthread_mutex_destroy(&_hLock);
    }
}

UINT ProducerQueue::GetQueueSize() const
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return _nQueueSize;
}

UINT ProducerQueue::GetBufferSize() const
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return _nBufferSize;
}

bool ProducerQueue::HasFreeItem() const
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return (!_FreeList.IsEmpty());
}

bool ProducerQueue::HasDataItem() const
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return (!_DataList.IsEmpty());
}

BufferItem *ProducerQueue::FreeItem(UINT nReserveSize)
{
    BufferItem  *pbi = NULL;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_FreeList.IsEmpty() )
    {
        pbi = _FreeList.Front();
        pbi->Reset(nReserveSize);
    }

    return pbi;
}

void ProducerQueue::PopFreeItem()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    assert(!_FreeList.IsEmpty());

    _FreeList.Pop();
}

BufferItem *ProducerQueue::GetFreeItem(UINT nReserveSize, bool bLock)
{
    BufferItem  *pbi = NULL;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( bLock )
    {
        assert(_bEnableLock);
        pthread_mutex_lock(&_hLock);
    }

    if ( !_FreeList.IsEmpty() )
    {
        pbi = _FreeList.PopAtFront();
        pbi->Reset(nReserveSize);
    }

    if ( bLock )
    {
        pthread_mutex_unlock(&_hLock);
    }

    return pbi;
}

void ProducerQueue::PutFreeItem(BufferItem *pbi, bool bLock)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( bLock )
    {
        assert(_bEnableLock);
        pthread_mutex_lock(&_hLock);
    }

    assert(!_FreeList.IsFull());
    _FreeList.PushAtBack(pbi);

    if ( bLock )
    {
        pthread_mutex_unlock(&_hLock);
    }
}

BufferItem *ProducerQueue::DataItem()
{
    BufferItem  *pbi = NULL;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_DataList.IsEmpty() )
    {
        pbi = _DataList.Front();
    }

    return pbi;
}

void ProducerQueue::PopDataItem()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    assert(!_DataList.IsEmpty());

    _DataList.Pop();
}

BufferItem *ProducerQueue::GetDataItem(bool bLock)
{
    BufferItem  *pbi = NULL;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( bLock )
    {
        assert(_bEnableLock);
        pthread_mutex_lock(&_hLock);
    }

    if ( !_DataList.IsEmpty() )
    {
        pbi = _DataList.PopAtFront();
    }

    if ( bLock )
    {
        pthread_mutex_unlock(&_hLock);
    }

    return pbi;
}

void ProducerQueue::PutDataItem(BufferItem *pbi, bool bLock)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( bLock )
    {
        assert(_bEnableLock);
        pthread_mutex_lock(&_hLock);
    }

    assert(!_DataList.IsFull());
    _DataList.PushAtBack(pbi);

    if ( bLock )
    {
        pthread_mutex_unlock(&_hLock);
    }
}