/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "SignalPipe.h"


#if defined(_ANDROID_) || defined(_LINUX_)
DWORD CreateSignalPipe(SOCKET *phSend, SOCKET *phRecv)
{
    DWORD  dwRet;
    int  fd[2];

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( pipe(fd)/*pipe2(fd, O_NONBLOCK)*/ )
    {
        dwRet = GetLastError();
        DBG(D_ERROR, (L"%s: create pipe failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    *phSend = fd[1];
    *phRecv = fd[0];

    return ERROR_SUCCESS;
}

DWORD PipeSendSignal(SOCKET hSend)
{
    DWORD  dwRet;
    char  ch = 's';

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( write(hSend, &ch, 1) < 0 )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: send signal failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    return ERROR_SUCCESS;
}

DWORD PipeRecvSignal(SOCKET hRecv)
{
    DWORD  dwRet;
    ssize_t  nBytes;
    char  buf[100];

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    nBytes = read(hRecv, buf, sizeof(buf));
    if ( nBytes <= 0 )
    {
        dwRet = (nBytes < 0) ? GetLastError() : ERROR_NOT_CONNECTED;
        DBG(D_WARN, (L"%s: recv signal failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    return ERROR_SUCCESS;
}

DWORD PipeSendData(SOCKET hSend, PBYTE pData, UINT nSize)
{
    DWORD  dwRet;
    ssize_t  nBytes;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    nBytes = write(hSend, pData, nSize);
    if ( nBytes < 0 )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: send data failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    if ( (UINT)nBytes < nSize )
    {
        DBG(D_WARN, (L"%s: partial: %zd of %d\n", FULL_FUNC_NAME, nBytes, nSize));
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    return ERROR_SUCCESS;
}

DWORD PipeRecvData(SOCKET hSend, PBYTE pData, UINT &rnSize)
{
    DWORD  dwRet;
    ssize_t  nBytes;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    nBytes = read(hSend, pData, rnSize);
    if ( nBytes <= 0 )
    {
        dwRet = (nBytes < 0) ? GetLastError() : ERROR_NOT_CONNECTED;
        DBG(D_WARN, (L"%s: recv data failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    rnSize = nBytes;

    return ERROR_SUCCESS;
}

void CloseSignalPipe(SOCKET hPipe)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    close(hPipe);
}
#else
DWORD CreateSignalPipe(SOCKET *phSend, SOCKET *phRecv)
{
    DWORD  dwRet;
    SOCKET  hSend, hRecv;
    struct sockaddr_in  addr;
    socklen_t  addrlen;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    hSend = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if ( hSend == INVALID_SOCKET )
    {
        dwRet = WSAGetLastError();
        DBG(D_ERROR, (L"%s: create socket failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    hRecv = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if ( hRecv == INVALID_SOCKET )
    {
        dwRet = WSAGetLastError();
        DBG(D_ERROR, (L"%s: create socket failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    addr.sin_port = htons(0);
    if ( ::bind(hRecv, (struct sockaddr*)&addr, sizeof(addr)) == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_ERROR, (L"%s: bind socket failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err3;
    }

    addrlen = sizeof(addr);
    if (getsockname(hRecv, (struct sockaddr*)&addr, &addrlen) == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_ERROR, (L"%s: get socket addr failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err3;
    }

    if ( connect(hSend, (struct sockaddr*)&addr, sizeof(addr)) == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_ERROR, (L"%s: connect socket failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err3;
    }

    DBG(D_VERB, (L"%s: pipe create at port(%d)\n", FULL_FUNC_NAME, ntohs(addr.sin_port)));

    *phSend = hSend;
    *phRecv = hRecv;

    return ERROR_SUCCESS;

err3:
    closesocket(hRecv);

err2:
    closesocket(hSend);

err1:
    return dwRet;
}

DWORD PipeSendSignal(SOCKET hSend)
{
    DWORD  dwRet;
    char  ch = 's';

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( send(hSend, &ch, 1, MSG_NOSIGNAL) == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_WARN, (L"%s: send signal failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    return ERROR_SUCCESS;
}

DWORD PipeRecvSignal(SOCKET hRecv)
{
    DWORD  dwRet;
    char  buf[100];
    int  len;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    len = recv(hRecv, buf, sizeof(buf), 0);
    if ( len <= 0 )
    {
        dwRet = (len < 0) ? WSAGetLastError() : ERROR_NOT_CONNECTED;
        DBG(D_WARN, (L"%s: recv signal failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    return ERROR_SUCCESS;
}

void CloseSignalPipe(SOCKET hPipe)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    closesocket(hPipe);
}
#endif
