/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "TimerCenter.h"


TimerCenter *TimerCenter::_instance = NULL;

TimerCenter *TimerCenter::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new TimerCenter();
    }
    return _instance;
}

void TimerCenter::Destroy()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

TimerCenter::TimerCenter()
    : _bQuit(false)
    , _TimerList(0)
    , _TimerListMate(true)
    , _dwMagicBase(1)
    , _pEntryPool(NULL)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    sem_init(&_hEvent, 0, 0);
    SET_PTH_INVALID(_hTimerThread);
    pthread_mutex_init(&_hTimerDelLock, NULL);
    pthread_mutex_init(&_hEntryLock, NULL);
}

TimerCenter::~TimerCenter()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_destroy(&_hEntryLock);
    pthread_mutex_destroy(&_hTimerDelLock);
    sem_destroy(&_hEvent);
}

DWORD TimerCenter::Load(UINT nMaxTimerNum)
{
    DWORD  dwRet;
    PTIMER_REC  pTimerRec;
    PTIMER_ENTRY  pTimerEntry;
    UINT  i;

    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    _TimerList.resize(nMaxTimerNum);
    _TimerListMate.Init(nMaxTimerNum);

    _bQuit = false;

    for ( i = 0; i < nMaxTimerNum; ++i )
    {
        pTimerRec = &_TimerList.at(i);
        memset(pTimerRec, 0, sizeof(TIMER_REC));
    }

    _pEntryPool = new(std::nothrow) RingBuffer<PTIMER_ENTRY>(nMaxTimerNum, true);
    if ( !_pEntryPool )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_ERROR, (L"%s: alloc _pEntryPool failed\n", FULL_FUNC_NAME));
        goto err;
    }

    for ( i = 0; i < nMaxTimerNum; ++i )
    {
        pTimerEntry = new(std::nothrow) TIMER_ENTRY;
        if ( !pTimerEntry )
        {
            dwRet = ERROR_OUTOFMEMORY;
            DBG(D_ERROR, (L"%s: alloc pTimerEntry failed\n", FULL_FUNC_NAME));
            goto err;
        }

        _pEntryPool->PushAtBack(pTimerEntry);
    }

    if ( pthread_create(&_hTimerThread, NULL, TimerThreadProc, this) )
    {
        dwRet = GetLastError();
        DBG(D_ERROR, (L"%s: create timer thread failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err;
    }

    return ERROR_SUCCESS;

err:
    Unload();

    return dwRet;
}

void TimerCenter::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    if ( IS_PTH_VALID(_hTimerThread) )
    {
        _bQuit = true;
        sem_post(&_hEvent);

        pthread_join(_hTimerThread, NULL);
        SET_PTH_INVALID(_hTimerThread);
    }

    _TimerList.resize(0);

    if ( _pEntryPool )
    {
        while ( !_pEntryPool->IsEmpty() )
        {
            delete _pEntryPool->PopAtFront();
        }
        delete _pEntryPool;
        _pEntryPool = NULL;
    }

    while ( !_EntryQueue.empty() )
    {
        delete _EntryQueue.top();
        _EntryQueue.pop();
    }
}

DWORD TimerCenter::CreateTimer(TIMERFUNC pTimerFunc, HANDLE hTimerArg, DWORD &rdwTimerId)
{
    DWORD  dwRet;
    UINT  nIdx;
    DWORD  dwMagic;
    PTIMER_REC  pTimerRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rdwTimerId = INVALID_TIMER_ID;

    dwRet = _TimerListMate.GetFreeIndex(nIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: no free timer\n", FULL_FUNC_NAME));
        return ERROR_ACCESS_DENIED;
    }

    pTimerRec = &_TimerList.at(nIdx);
    assert(!pTimerRec->bUsed);
    DO_IF_DBG(pTimerRec->bUsed = true);

    dwMagic = InterlockedIncrement((LONG*)&_dwMagicBase);

    pTimerRec->bReqDel = false;
    pTimerRec->bBusy = false;
    pTimerRec->wMagic = ADJUST_WMAGICNUM(dwMagic);
    pTimerRec->dwEntryMagic = 0;
    pTimerRec->pTimerFunc = pTimerFunc;
    pTimerRec->hTimerArg = hTimerArg;

    rdwTimerId = MAKE_DWID(nIdx, pTimerRec->wMagic);
    DBG(D_INFO, (L"%s: idx(%d), id(%x)\n", FULL_FUNC_NAME, nIdx, rdwTimerId));

    return ERROR_SUCCESS;
}

DWORD TimerCenter::DestroyTimer(DWORD dwTimerId)
{
    UINT  nIdx;
    WORD  wMagic;
    PTIMER_REC  pTimerRec;
    pthread_t  hCurThread;

    nIdx = WIDX_FROM_DWID(dwTimerId);
    wMagic = WMAGICNUM_FROM_DWID(dwTimerId);

    DBG(D_INFO, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nIdx, wMagic));

    if ( nIdx >= _TimerList.size() )
    {
        DBG(D_WARN, (L"%s: invalid timer id\n", FULL_FUNC_NAME));
        return ERROR_INVALID_PARAMETER;
    }

    pTimerRec = &_TimerList.at(nIdx);
    if ( wMagic != pTimerRec->wMagic )
    {
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nIdx));
        return ERROR_NOT_FOUND;
    }
    assert(pTimerRec->bUsed);

    // now no one can use it from external
    pTimerRec->wMagic = ~wMagic;

    hCurThread = pthread_self();
    if ( IS_PTH_EQUAL(hCurThread, _hTimerThread) )
    {
        pTimerRec->bReqDel = true;
        return ERROR_SUCCESS;
    }

    pthread_mutex_lock(&_hTimerDelLock);
    pTimerRec->dwEntryMagic = 0;
    while ( pTimerRec->bBusy )
    {
        pthread_mutex_unlock(&_hTimerDelLock);
        DBG(D_INFO, (L"%s: waiting idx(%d) ...\n", FULL_FUNC_NAME, nIdx));
        Sleep(100);
        pthread_mutex_lock(&_hTimerDelLock);
    }
    pthread_mutex_unlock(&_hTimerDelLock);

    DO_IF_DBG(pTimerRec->bUsed = false);
    _TimerListMate.PutFreeIndex(nIdx);

    return ERROR_SUCCESS;
}

DWORD TimerCenter::SetTimer(DWORD dwTimerId, DWORD dwMilliseconds)
{
    UINT  nIdx;
    WORD  wMagic;
    PTIMER_REC  pTimerRec;
    PTIMER_ENTRY pTimerEntry;

    nIdx = WIDX_FROM_DWID(dwTimerId);
    wMagic = WMAGICNUM_FROM_DWID(dwTimerId);

    DBG(D_VERB, (L"%s: idx(%d): ms(%d)\n", FULL_FUNC_NAME, nIdx, dwMilliseconds));

    if ( nIdx >= _TimerList.size() )
    {
        DBG(D_WARN, (L"%s: invalid timer id\n", FULL_FUNC_NAME));
        return ERROR_INVALID_PARAMETER;
    }

    pTimerRec = &_TimerList.at(nIdx);
    if ( wMagic != pTimerRec->wMagic )
    {
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nIdx));
        return ERROR_NOT_FOUND;
    }
    assert(pTimerRec->bUsed);

    if ( !_pEntryPool->PopAtFrontSafe(pTimerEntry) )
    {
        DBG(D_WARN, (L"%s: idx(%d): no free entry\n", FULL_FUNC_NAME, nIdx));
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    pTimerRec->dwEntryMagic = InterlockedIncrement((LONG*)&_dwMagicBase);

    pTimerEntry->nTimerIdx = nIdx;
    pTimerEntry->dwMagic = pTimerRec->dwEntryMagic;
    clock_gettime(CLOCK_REALTIME, &pTimerEntry->tsTime);
    TimeAdd(pTimerEntry->tsTime, dwMilliseconds);

    pthread_mutex_lock(&_hEntryLock);
    _EntryQueue.push(pTimerEntry);
    pthread_mutex_unlock(&_hEntryLock);

    sem_post(&_hEvent);

    return ERROR_SUCCESS;
}

void TimerCenter::DoTimerEntry(PTIMER_ENTRY pTimerEntry)
{
    PTIMER_REC  pTimerRec;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    pTimerRec = &_TimerList.at(pTimerEntry->nTimerIdx);

    pthread_mutex_lock(&_hTimerDelLock);
    if ( pTimerEntry->dwMagic == pTimerRec->dwEntryMagic )
    {
        assert(pTimerRec->bUsed);
        assert(!pTimerRec->bBusy);
        pTimerRec->bBusy = true;
    }
    else
    {
        pTimerRec = NULL;
    }
    pthread_mutex_unlock(&_hTimerDelLock);

    if ( !pTimerRec )
    {
        goto out;
    }

    pTimerRec->pTimerFunc(pTimerRec->hTimerArg);

    if ( pTimerRec->bReqDel )
    {
        DO_IF_DBG(pTimerRec->bUsed = false);
        _TimerListMate.PutFreeIndex(pTimerEntry->nTimerIdx);
        goto out;
    }

    pTimerRec->bBusy = false;

out:
    _pEntryPool->PushAtBack(pTimerEntry);
}

DWORD TimerCenter::TimerThread()
{
    PTIMER_ENTRY  pTimerEntry;
    timespec  tsWait, tsCur;

    DBG(D_LOG, (L"%s: enter\n", FULL_FUNC_NAME));

    clock_gettime(CLOCK_REALTIME, &tsWait);
    tsWait.tv_sec += 31536000;  // one year

    while ( !_bQuit )
    {
        sem_timedwait(&_hEvent, &tsWait);

        if ( _bQuit )
        {
            break;
        }

        clock_gettime(CLOCK_REALTIME, &tsWait);
        tsWait.tv_sec += 31536000;  // one year

        pthread_mutex_lock(&_hEntryLock);
        while ( !_EntryQueue.empty() )
        {
            clock_gettime(CLOCK_REALTIME, &tsCur);
            pTimerEntry = _EntryQueue.top();
            if ( TimeCompare(tsCur, pTimerEntry->tsTime) == -1 )
            {
                tsWait = pTimerEntry->tsTime;
                break;
            }

            _EntryQueue.pop();

            pthread_mutex_unlock(&_hEntryLock);

            DoTimerEntry(pTimerEntry);

            pthread_mutex_lock(&_hEntryLock);
        }
        pthread_mutex_unlock(&_hEntryLock);
    }

    DBG(D_LOG, (L"%s: leave\n", FULL_FUNC_NAME));

    return ERROR_SUCCESS;
}

void *TimerCenter::TimerThreadProc(void *arg)
{
    TimerCenter  *pThis = static_cast<TimerCenter*>(arg);

    pThis->TimerThread();

    return NULL;
}

