/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#ifndef _RING_BUFFER_H
#define _RING_BUFFER_H


template<class T> class RingBuffer
{
public:
    RingBuffer(UINT nSize, bool bEnableLock)
        : _bEnableLock(bEnableLock)
    {
        DBG(D_FUNC, (L"%s: size(%d)\n", FULL_FUNC_NAME, nSize));

        if ( _bEnableLock )
        {
            pthread_mutex_init(&_hLock, NULL);
        }

        _nPushPos = 0;
        _nPopPos = 0;
        _nSize = nSize + 1;
        _pRingBuf = new T[_nSize];
    }

    ~RingBuffer()
    {
        DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

        delete []_pRingBuf;

        if ( _bEnableLock )
        {
            pthread_mutex_destroy(&_hLock);
        }
    }

    void Reset()
    {
        DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

        _nPushPos = 0;
        _nPopPos = 0;
    }

    bool IsEmpty() const
    {
        DBG(D_FUNC, (L"%s: push(%d), pop(%d), slot(%d)\n", FULL_FUNC_NAME, _nPushPos, _nPopPos, _nSize));

        return (_nPushPos == _nPopPos);
    }

    bool IsFull() const
    {
        DBG(D_FUNC, (L"%s: push(%d), pop(%d), slot(%d)\n", FULL_FUNC_NAME, _nPushPos, _nPopPos, _nSize));

        return (((_nPushPos + 1) % _nSize) == _nPopPos);
    }

    void PushAtBack(T item)
    {
        DBG(D_FUNC, (L"%s: push(%d), pop(%d), slot(%d)\n", FULL_FUNC_NAME, _nPushPos, _nPopPos, _nSize));

        assert(!IsFull());

        _pRingBuf[_nPushPos] = item;

        MemoryBarrier();

        _nPushPos = (_nPushPos + 1) % _nSize;

        MemoryBarrier();
    }

    T PopAtFront()
    {
        T  item;

        DBG(D_FUNC, (L"%s: push(%d), pop(%d), slot(%d)\n", FULL_FUNC_NAME, _nPushPos, _nPopPos, _nSize));

        assert(!IsEmpty());

        item = _pRingBuf[_nPopPos];

        MemoryBarrier();

        _nPopPos = (_nPopPos + 1) % _nSize;

        MemoryBarrier();

        return item;
    }

    T Front()
    {
        DBG(D_FUNC, (L"%s: push(%d), pop(%d), slot(%d)\n", FULL_FUNC_NAME, _nPushPos, _nPopPos, _nSize));

        assert(!IsEmpty());

        return _pRingBuf[_nPopPos];
    }

    void Pop()
    {
        DBG(D_FUNC, (L"%s: push(%d), pop(%d), slot(%d)\n", FULL_FUNC_NAME, _nPushPos, _nPopPos, _nSize));

        assert(!IsEmpty());

        _nPopPos = (_nPopPos + 1) % _nSize;

        MemoryBarrier();
    }

    bool PushAtBackSafe(T item)
    {
        bool  ret;

        DBG(D_FUNC, (L"%s: push(%d), pop(%d), slot(%d)\n", FULL_FUNC_NAME, _nPushPos, _nPopPos, _nSize));

        assert(_bEnableLock);

        pthread_mutex_lock(&_hLock);
        ret = !IsFull();
        if ( ret )
        {
            PushAtBack(item);
        }
        pthread_mutex_unlock(&_hLock);

        return ret;
    }

    bool PopAtFrontSafe(T &item)
    {
        bool  ret;

        DBG(D_FUNC, (L"%s: push(%d), pop(%d), slot(%d)\n", FULL_FUNC_NAME, _nPushPos, _nPopPos, _nSize));

        assert(_bEnableLock);

        pthread_mutex_lock(&_hLock);
        ret = !IsEmpty();
        if ( ret )
        {
            item = PopAtFront();
        }
        pthread_mutex_unlock(&_hLock);

        return ret;
    }

private:
    bool                        _bEnableLock;
    pthread_mutex_t             _hLock;

    volatile UINT               _nPushPos;
    volatile UINT               _nPopPos;
    UINT                        _nSize;
    T                         * _pRingBuf;
};


#endif
