/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#ifndef _SEED_TARGET_H
#define _SEED_TARGET_H


#include "PObject.h"
#include "BitArray.h"
#include "BufferItem.h"
#include "VectorMate.h"
#include "FileMapping.h"
#include "Runnable.h"
#include "MeasElapseTime.h"

class SHA1Hash;
class SeedTarget;
class SeedTargetObserver;


typedef struct _ST_OBS_REC
{
    bool                        bNewOne;
    SeedTargetObserver        * pObserver;
    DWORD                       dwCookie;
}
ST_OBS_REC, *PST_OBS_REC;

typedef struct _ST_RWS_REC
{
#ifndef NDEBUG
    volatile bool               bUsed;
#endif

    volatile WORD               wMagic;

    MeasElapseTime              MesaRdWr;

    UINT                        nBlkIdx;

    QWORD                       qwBlkOffset;
    DWORD                       dwBlkSize;
    DWORD                       dwOffInBlk;
    DWORD                       dwLeftInBlk;

    PBYTE                       pBlkBuf;

#ifdef USE_NT_FIO
    HANDLE                      hFile;
    QWORD                       qwCurOffset;
#else
#if defined(CONFIG_ST_BLK_RD_FIO) || defined(CONFIG_ST_BLK_WR_FIO)
    FILE                      * hFile;
    QWORD                       qwCurOffset;
#if defined(CONFIG_ST_BLK_RD_BUF) || defined(CONFIG_ST_BLK_WR_BUF)
    char                      * pFIOBuf;
#endif
#endif
#endif

#ifdef CONFIG_ST_BLK_WR_VERIFY
    SHA1Hash                  * sha1;
#endif
}
ST_RWS_REC, *PST_RWS_REC;


class SeedTargetObserver
{
public:
    virtual void OnStateUpdated(SeedTarget *pSeedTarget, DWORD dwCookie, UINT nDegree, QWORD qwMaskAnd, QWORD qwMaskOr) = 0;
};


class SeedTarget : public PObject, protected Runnable
{
public:
    SeedTarget(UINT nMaxRdNum, UINT nMaxWrNum);

protected:
    virtual ~SeedTarget();

public:
    DWORD Init(const char *szSeedFile, const char *szTargetPath, bool bReadOnly, bool bAllowVerify, const char *szCacheFile = NULL);
    void Fini();

    DWORD InitRWSEnv(const char *szTargetFile);
    void FiniRWSEnv();

#ifdef CONFIG_ST_BLK_WR_VERIFY
    DWORD InitHashEnv();
    void FiniHashEnv();
#endif

    DWORD RegisterObserver(SeedTargetObserver *pObserver, DWORD dwCookie);
    void UnregisterObserver(SeedTargetObserver *pObserver);

    bool IsMatch(const SHA1_VAL &SHA1ValSeed) const;

    const char *GetName() const;
    void GetSeedInfo(PSEED_INFO pInfo) const;

    bool IsFinished() const;
    bool IsAllBusy() const;

    UINT GetMaxReadNum() const;
    UINT GetMaxWriteNum() const;

    UINT GetBlockReadCount() const;
    UINT GetBlockWriteCount() const;

    const BitArray *GetProgressMask() const;

    UINT GetBlockCount() const;
    DWORD GetBlockSize() const;

    DWORD BeginSession(bool bReadOrWrite, DWORD &rdwSessionId);
    DWORD EndSession(DWORD dwSessionId);

    DWORD OpenBlock(DWORD dwSessionId, UINT nBlkIdx, DWORD &rdwBlkId);
    DWORD ReadBlock(DWORD dwBlkId, PBYTE pBuf, DWORD &rdwSize, bool &rbEndOfBlk);
    DWORD WriteBlock(DWORD dwBlkId, PBYTE pBuf, DWORD dwSize, bool &rbEndOfBlk);
    DWORD MapBlock(DWORD dwBlkId, PBYTE &rpBuf, DWORD &rdwSize);
    DWORD UnmapBlock(DWORD dwBlkId, PBYTE pBuf, DWORD dwSizeUsed);
    DWORD CloseBlock(DWORD dwBlkId);

    virtual void Execute();

private:
    DWORD ReadSeedFile(const char *szSeedFile);
    DWORD VerifySeedHdr();

    DWORD OpenTargetFile(const char *szTargetFile, bool bReadOnly, bool &rbNeedVerify);
    DWORD VerifyTargetFile(bool bOneOnly);
    DWORD VerifyTargetBlock(UINT nBlkIdx, bool &rbValid);

#if defined(CONFIG_ST_BLK_RD_FIO) || defined(USE_NT_FIO)
    DWORD ReadTarget(PST_RWS_REC pRWSRec, QWORD qwOffset, PBYTE pBuf, DWORD dwSize);
#endif
#if defined(CONFIG_ST_BLK_WR_FIO) || defined(USE_NT_FIO)
    DWORD WriteTarget(PST_RWS_REC pRWSRec, QWORD qwOffset, PBYTE pBuf, DWORD dwSize);
#endif

    DWORD VerifySessionId(DWORD dwSessionId, UINT &rnRWSIdx, PST_RWS_REC &rpRWSRec);
    DWORD VerifyReadWrite(bool bReadOrWrite, UINT nRWSIdx, PST_RWS_REC pRWSRec);

private:
    DWORD                       _dwRunnId;

    BufferItem                  _biSeedHdr;
    PSEED_HDR                   _pSeedHdr;

#ifdef USE_NT_FIO
    HANDLE                      _hTargetFile;
#else
    FILE                      * _hTargetFile;

#if defined(CONFIG_ST_BLK_RD_MMAP) || defined(CONFIG_ST_BLK_WR_MMAP)
    FileMapping                 _TargetMapping;
#endif
#endif

    pthread_mutex_t             _hStateLock;
    volatile bool               _bObsAdded;
    volatile bool               _bMaskUpdated;

    pthread_mutex_t             _hObsLock;
    list<PST_OBS_REC>           _ObsList;

    LONG                        _lBlkRdCnt;
    LONG                        _lBlkWrCnt;

    BitArray                    _baBlkProg;
    BitArray                    _baBlkBusy;

    vector<ST_RWS_REC>          _RWSList;
    UINT                        _nRWSDivide;
    VectorMate                  _RdSListMate;
    VectorMate                  _WrSListMate;
    DWORD                       _dwRWSMagic;
};


#endif
