/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "SocketConnection.h"
#include "UdpConnectionIn.h"
#include "MCastConnection.h"
#include "ProxyFilterAgent.h"
#ifdef CONFIG_TLS_SUPPORT
#include "SecureFilterAgent.h"
#endif
#include "ConnectCenter.h"


ConnectCenter *ConnectCenter::_instance = NULL;

ConnectCenter *ConnectCenter::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new ConnectCenter();
    }
    return _instance;
}

void ConnectCenter::Destroy()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

ConnectCenter::ConnectCenter()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));
}

ConnectCenter::~ConnectCenter()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));
}

DWORD ConnectCenter::Load()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    return ERROR_SUCCESS;
}

void ConnectCenter::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));
}

DWORD ConnectCenter::CreateListener(PSOCK_ADDR pLocalAddr, UINT nBacklogNum, Connection **ppListenConnection)
{
    DWORD  dwRet;
    SOCKET  s;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    dwRet = CreateAndBind(pLocalAddr, false, false, s);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: CreateAndBind failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    if ( pLocalAddr->IPv4Addr.bStream )
    {
        if ( listen(s, nBacklogNum) == SOCKET_ERROR )
        {
            dwRet = WSAGetLastError();
            DBG(D_ERROR, (L"%s: listen failed: %d\n", FULL_FUNC_NAME,dwRet));
            goto err2;
        }
    }

    dwRet = AllocConnection(s, true, false, false, NULL, ppListenConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: AllocConnection failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    closesocket(s);

err1:
    return dwRet;
}

DWORD ConnectCenter::AcceptConnection(Connection *pListenConnection, bool bAsync, Connection **ppConnection)
{
    DWORD  dwRet;
    SOCKET  s_new;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    s_new = accept(dynamic_cast<SocketConnection*>(pListenConnection)->_s, NULL, NULL);
    if ( s_new == INVALID_SOCKET )
    {
        dwRet = WSAGetLastError();
        DBG(D_ERROR, (L"%s: accept failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    if ( bAsync )
    {
#ifdef _WIN32_
        u_long  iMode = 1;
        ioctlsocket(s_new, FIONBIO, &iMode);
#else
        int  nMode = fcntl(s_new, F_GETFL, 0);
        fcntl(s_new, F_SETFL, (nMode | O_NONBLOCK));
#endif
    }

    dwRet = AllocConnection(s_new, false, false, false, NULL, ppConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: AllocConnection failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    (*ppConnection)->EnableKeepAlive(KEEP_ALIVE_TIMEOUT);

    return ERROR_SUCCESS;
    
err2:
    closesocket(s_new);

err1:
    return dwRet;
}

DWORD ConnectCenter::ConnectIn(PSOCK_ADDR pLocalAddr, bool bAsync, Connection **ppConnection)
{
    DWORD  dwRet;
    bool  bMCast;
    DWORD  dwMCastIP = INADDR_ANY;
    struct ip_mreq  imr;
    SOCKET  s;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( pLocalAddr->IPv4Addr.bStream )
    {
        DBG(D_ERROR, (L"%s: only support UDP protocol\n", FULL_FUNC_NAME));
        return ERROR_NOT_SUPPORTED;
    }

    bMCast = IsMCastAddr(pLocalAddr);
    if ( bMCast )
    {
        dwMCastIP = pLocalAddr->IPv4Addr.dwIP;
        pLocalAddr->IPv4Addr.dwIP = INADDR_ANY;
    }

    dwRet = CreateAndBind(pLocalAddr, bAsync, false, s);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: CreateAndBind failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    if ( bMCast )
    {
        pLocalAddr->IPv4Addr.dwIP = dwMCastIP;

        imr.imr_multiaddr.s_addr = dwMCastIP;
        imr.imr_interface.s_addr = INADDR_ANY;
        if ( setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*)&imr, sizeof(imr)) == SOCKET_ERROR )
        {
            dwRet = GetLastError();
            DBG(D_ERROR, (L"%s: IP_ADD_MEMBERSHIP failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto err2;
        }
    }

    dwRet = AllocConnection(s, false, false, true, bMCast ? pLocalAddr : NULL, ppConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: AllocConnection failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    closesocket(s);

err1:
    return dwRet;
}

DWORD ConnectCenter::ConnectOut(PSOCK_ADDR pLocalAddr, const PSOCK_ADDR pRemoteAddr, bool bAsync, Connection **ppConnection)
{
    DWORD  dwRet;
    bool  bMCast;
    SOCKET  s;
    struct sockaddr_in  addr;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    bMCast = IsMCastAddr(pRemoteAddr);

    dwRet = CreateAndBind(pLocalAddr, bAsync, bMCast, s);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: CreateAndBind failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    if ( pLocalAddr->IPv4Addr.bStream )
    {
        // see issues_99stuck.txt
#ifdef _LINUX_
        int  iVal = 1;
        setsockopt(s, IPPROTO_TCP, TCP_SYNCNT, &iVal, sizeof(iVal));
#endif
    }

    if ( !bMCast )
    {
        addr.sin_family = AF_INET;
        addr.sin_addr.s_addr = pRemoteAddr->IPv4Addr.dwIP;
        addr.sin_port = pRemoteAddr->IPv4Addr.wPort;
        if ( connect(s, (struct sockaddr*)&addr, sizeof(addr)) == SOCKET_ERROR )
        {
            dwRet = WSAGetLastError();
            if ( (!bAsync) || (dwRet != WSAEWOULDBLOCK) )
            {
                DBG(D_ERROR, (L"%s: connect failed: %d\n", FULL_FUNC_NAME, dwRet));
                goto err2;
            }
        }
    }

    dwRet = AllocConnection(s, false, bAsync, false, bMCast ? pRemoteAddr : NULL, ppConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: AllocConnection failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    if ( pLocalAddr->IPv4Addr.bStream )
    {
        (*ppConnection)->EnableKeepAlive(KEEP_ALIVE_TIMEOUT);
    }

    return ERROR_SUCCESS;

err2:
    closesocket(s);

err1:
    return dwRet;
}

DWORD ConnectCenter::ConnectOut(PSOCK_ADDR pLocalAddr, const PSOCK_ADDR pRemoteAddr, const PSOCK_ADDR pProxyAddr, bool bAsync, Connection **ppConnection)
{
    DWORD  dwRet;
    Connection  *pConnection;
    FilterProtocol  *pProxyFilter;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    dwRet = ConnectOut(pLocalAddr, pProxyAddr, bAsync, &pConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: connect proxy(%#.8x/%d) failed: %d\n", FULL_FUNC_NAME, pProxyAddr->IPv4Addr.dwIP, ntohs(pProxyAddr->IPv4Addr.wPort), dwRet));
        goto err1;
    }

    pProxyFilter = new(std::nothrow) ProxyFilterAgent(pConnection, pRemoteAddr);
    if ( pProxyFilter == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc ProxyFilterAgent failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    *ppConnection = pProxyFilter;

    return ERROR_SUCCESS;

err2:
    FinishConnection(pConnection);

err1:
    return dwRet;
}

#ifdef CONFIG_TLS_SUPPORT
DWORD ConnectCenter::AcceptConnectionTLS(Connection *pListenConnection, bool bAsync, Connection **ppConnection)
{
    DWORD  dwRet;
    Connection  *pConnection;
    FilterProtocol  *pSecureFilter;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    dwRet = AcceptConnection(pListenConnection, bAsync, &pConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: AcceptConnection failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    pSecureFilter = new(std::nothrow) SecureFilterAgent(pConnection, true);
    if ( pSecureFilter == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc SecureFilterAgent failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    *ppConnection = pSecureFilter;

    return ERROR_SUCCESS;

err2:
    FinishConnection(pConnection);

err1:
    return dwRet;
}

DWORD ConnectCenter::ConnectOutTLS(PSOCK_ADDR pLocalAddr, const PSOCK_ADDR pRemoteAddr, bool bAsync, Connection **ppConnection)
{
    DWORD  dwRet;
    Connection  *pConnection;
    FilterProtocol  *pSecureFilter;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    dwRet = ConnectOut(pLocalAddr, pRemoteAddr, bAsync, &pConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: connect %#.8x/%d failed: %d\n", FULL_FUNC_NAME, pRemoteAddr->IPv4Addr.dwIP, ntohs(pRemoteAddr->IPv4Addr.wPort), dwRet));
        goto err1;
    }

    pSecureFilter = new(std::nothrow) SecureFilterAgent(pConnection, false);
    if ( pSecureFilter == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc SecureFilterAgent failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    *ppConnection = pSecureFilter;

    return ERROR_SUCCESS;

err2:
    FinishConnection(pConnection);

err1:
    return dwRet;
}

DWORD ConnectCenter::ConnectOutTLS(PSOCK_ADDR pLocalAddr, const PSOCK_ADDR pRemoteAddr, const PSOCK_ADDR pProxyAddr, bool bAsync, Connection **ppConnection)
{
    DWORD  dwRet;
    Connection  *pConnection;
    FilterProtocol  *pSecureFilter;
    FilterProtocol  *pProxyFilter;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    dwRet = ConnectOut(pLocalAddr, pProxyAddr, bAsync, &pConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: connect proxy(%#.8x/%d) failed: %d\n", FULL_FUNC_NAME, pProxyAddr->IPv4Addr.dwIP, ntohs(pProxyAddr->IPv4Addr.wPort), dwRet));
        goto err1;
    }

    pSecureFilter = new(std::nothrow) SecureFilterAgent(pConnection, false);
    if (pSecureFilter == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc SecureFilterAgent failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    pProxyFilter = new(std::nothrow) ProxyFilterAgent(pSecureFilter, pRemoteAddr);
    if ( pProxyFilter == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc ProxyFilterAgent failed\n", FULL_FUNC_NAME));
        goto err3;
    }

    *ppConnection = pProxyFilter;

    return ERROR_SUCCESS;

err3:
    FinishConnection(pSecureFilter);

err2:
    FinishConnection(pConnection);

err1:
    return dwRet;
}
#endif

void ConnectCenter::FinishConnection(Connection *pConnection)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    delete pConnection;
}

DWORD ConnectCenter::CreateAndBind(PSOCK_ADDR pLocalAddr, bool bAsync, bool bMCast, SOCKET &rs)
{
    DWORD  dwRet;
    socklen_t  addrlen;
    struct sockaddr_in  addr;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( pLocalAddr->bType != ADDR_TYPE_IPV4 )
    {
        DBG(D_WARN, (L"%s: type(%d) not support\n", FULL_FUNC_NAME, pLocalAddr->bType));
        return ERROR_NOT_SUPPORTED;
    }

    if ( pLocalAddr->IPv4Addr.bStream )
    {
        rs = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    }
    else
    {
        rs = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    }

    if ( rs == INVALID_SOCKET )
    {
        dwRet = WSAGetLastError();
        DBG(D_WARN, (L"%s: create socket failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = pLocalAddr->IPv4Addr.dwIP;
    addr.sin_port = pLocalAddr->IPv4Addr.wPort;
    if ( ::bind(rs, (struct sockaddr*)&addr, sizeof(addr)) == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_WARN, (L"%s: bind socket failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    if ( !addr.sin_port )
    {
        addrlen = sizeof(addr);
        getsockname(rs, (struct sockaddr*)&addr, &addrlen);
        pLocalAddr->IPv4Addr.wPort = addr.sin_port;
    }

    if ( bAsync )
    {
#ifdef _WIN32_
        u_long  iMode = 1;
        ioctlsocket(rs, FIONBIO, &iMode);
#else
        int  nMode = fcntl(rs, F_GETFL, 0);
        fcntl(rs, F_SETFL, (nMode | O_NONBLOCK));
#endif
    }

    if ( bMCast )
    {
#ifdef _WIN32_
#else
        char  cloop = 0;
        setsockopt(rs, IPPROTO_IP, IP_MULTICAST_LOOP, (char*)&cloop, sizeof(cloop));
#endif
    }

    return ERROR_SUCCESS;

err2:
    closesocket(rs);

err1:
    return dwRet;
}

DWORD ConnectCenter::AllocConnection(SOCKET s, bool bListener, bool bAsyncConnect, bool bUdpConnectIn, const PSOCK_ADDR pMCastAddr, Connection **ppConnection)
{
    Connection  *pConnection;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( pMCastAddr )
    {
        pConnection = new(std::nothrow) MCastConnection(s, pMCastAddr);
    }
    else if ( bUdpConnectIn )
    {
        pConnection = new(std::nothrow) UdpConnectionIn(s);
    }
    else
    {
        pConnection = new(std::nothrow) SocketConnection(s, bListener, bAsyncConnect);
    }

    if ( !pConnection )
    {
        DBG(D_WARN, (L"%s: alloc Connection failed\n", FULL_FUNC_NAME));
        return ERROR_OUTOFMEMORY;
    }

    *ppConnection = pConnection;

    return ERROR_SUCCESS;
}

