/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "ConnectCenter.h"
#include "ListenerHandler.h"
#include "Endpoint.h"


typedef struct _THREAD_ARG
{
    UINT                        nIdx;
    Endpoint                  * pThis;
}
THREAD_ARG, *PTHREAD_ARG;


Endpoint::Endpoint()
    : _nEpSigBase(0)
    , _pEvent(NULL)
    , _bQuit(false)
    , _hWorkThreads(0)
    , _ConnList(0)
    , _ConnListMate(true)
    , _dwConnMagic(0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_rwlock_init(&_hConnValidLock, NULL);
    pthread_mutex_init(&_hConnCmdLock, NULL);
}

Endpoint::~Endpoint()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_destroy(&_hConnCmdLock);
    pthread_rwlock_destroy(&_hConnValidLock);
}

DWORD Endpoint::Load(UINT nMaxConnNum, UINT nExTaskNum, UINT nEpSigNum)
{
    DWORD  dwRet;
    UINT  i, nSigNum;
    PCONN_REC  pConnRec;
    pthread_t  hThread;
    PTHREAD_ARG  pThreadArg;

    DBG(D_LOG, (L"%s: MaxConnNum=%d, ExTaskNum=%d, EpSigNum=%d\n", FULL_FUNC_NAME, nMaxConnNum, nExTaskNum, nEpSigNum));

    // see issues_maxconn.txt
    nMaxConnNum += nExTaskNum;
    nExTaskNum += 1;

    _nEpSigBase = nMaxConnNum;
    _nEpSigNum = nEpSigNum;
    nSigNum = _nEpSigBase + _nEpSigNum;

    _pEvent = new(std::nothrow) PCondEvent(nSigNum);
    if ( !_pEvent )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_ERROR, (L"%s: alloc PCondEvent(%d) failed\n", FULL_FUNC_NAME, nSigNum));
        goto err1;
    }

    _bQuit = false;
    _hWorkThreads.resize(nExTaskNum);

    _ConnList.resize(nMaxConnNum);
    dwRet = _ConnListMate.Init(nMaxConnNum);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: init VectorMate failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    for ( i = 0; i < nMaxConnNum; ++i )
    {
        pConnRec = &_ConnList.at(i);
        DO_IF_DBG(pConnRec->bUsed = false);
        pConnRec->bValid = false;
    }

    for ( i = 0; i < nExTaskNum; ++i )
    {
        SET_PTH_INVALID(_hWorkThreads.at(i));
    }

    for ( i = 0; i < nExTaskNum; ++i )
    {
        pThreadArg = new(std::nothrow) THREAD_ARG;
        if ( !pThreadArg )
        {
            dwRet = ERROR_OUTOFMEMORY;
            DBG(D_ERROR, (L"%s: alloc pThreadArg failed\n", FULL_FUNC_NAME));
            goto err3;
        }

        pThreadArg->nIdx = i;
        pThreadArg->pThis = this;
        if ( pthread_create(&hThread, NULL, WorkThreadProc, pThreadArg) )
        {
            dwRet = GetLastError();
            DBG(D_ERROR, (L"%s: create work thread[#%d] failed: %d\n", FULL_FUNC_NAME, i, dwRet));
            delete pThreadArg;
            goto err3;
        }

        _hWorkThreads.at(i) = hThread;
    }

    return ERROR_SUCCESS;

err3:
    UnloadThreads();
    _ConnList.resize(0);
    _ConnListMate.Fini();

err2:
    delete _pEvent;
    _pEvent = NULL;

err1:
    return dwRet;
}

void Endpoint::Unload()
{
    PCONN_REC  pConnRec;
    DWORD  dwConnId;
    UINT  i, nSize;

    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    UnloadThreads();

    for ( i = 0, nSize = (UINT)_ConnList.size(); i < nSize; ++i )
    {
        pConnRec = &_ConnList.at(i);
        if ( pConnRec->bValid )
        {
            dwConnId = MAKE_DWID(i, pConnRec->wMagic);
            ReleaseConn(dwConnId, pConnRec);
        }
    }
    _ConnList.resize(0);
    _ConnListMate.Fini();

    delete _pEvent;
    _pEvent = NULL;
}

DWORD Endpoint::AddListener(PSOCK_ADDR pLocalAddr, bool bOneShot, DWORD &rhListener, UINT nBacklogNum)
{
    DWORD  dwRet;
    UINT  nConnIdx;
    Connection  *pConnection;
    EpProtocolHandler  *pHandler;

    DBG(D_INFO, (L"%s: port(%d)\n", FULL_FUNC_NAME, ntohs(pLocalAddr->IPv4Addr.wPort)));

    dwRet = _ConnListMate.GetFreeIndex(nConnIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: reach the max conn limit\n", FULL_FUNC_NAME));
        goto err1;
    }

    dwRet = ConnectCenter::Instance()->CreateListener(pLocalAddr, nBacklogNum, &pConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: CreateConnection failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    pHandler = new(std::nothrow) ListenerHandler(bOneShot);
    if ( !pHandler )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc ListenerHandler failed\n", FULL_FUNC_NAME));
        goto err3;
    }

    rhListener = MakeConnId(nConnIdx);

    AddToConnPool(T_CONN_LISTENER, pConnection, pHandler, rhListener);

    return ERROR_SUCCESS;

err3:
    ConnectCenter::Instance()->FinishConnection(pConnection);

err2:
    _ConnListMate.PutFreeIndex(nConnIdx);

err1:
    rhListener = INVALID_LISTENER_ID;

    return dwRet;
}

void Endpoint::RemoveListener(DWORD hListener)
{
    DBG(D_INFO, (L"%s: id(%d)\n", FULL_FUNC_NAME, hListener));

    DisconnectConn(hListener);
}

DWORD Endpoint::ConnectIn(PSOCK_ADDR pLocalAddr, EpProtocolHandler *pHandler, DWORD &rdwConnId)
{
    DWORD  dwRet;
    UINT  nConnIdx;
    Connection  *pConnection;

    DBG(D_INFO, (L"%s: -> %#.8x/%d\n", FULL_FUNC_NAME, pLocalAddr->IPv4Addr.dwIP, ntohs(pLocalAddr->IPv4Addr.wPort)));

    dwRet = _ConnListMate.GetFreeIndex(nConnIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: reach the max conn limit\n", FULL_FUNC_NAME));
        goto err1;
    }

    dwRet = ConnectCenter::Instance()->ConnectIn(pLocalAddr, true, &pConnection);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: -> %#.8x/%d failed: %d\n", FULL_FUNC_NAME, pLocalAddr->IPv4Addr.dwIP, ntohs(pLocalAddr->IPv4Addr.wPort), dwRet));
        goto err2;
    }

    rdwConnId = MakeConnId(nConnIdx);

    AddToConnPool(T_CONN_CONNECT_IN, pConnection, pHandler, rdwConnId);

    return ERROR_SUCCESS;

err2:
    _ConnListMate.PutFreeIndex(nConnIdx);

err1:
    rdwConnId = INVALID_EP_CONN_ID;

    return dwRet;
}

DWORD Endpoint::ConnectOut(PSOCK_ADDR pLocalAddr, const PSOCK_ADDR pRemoteAddr, EpProtocolHandler *pHandler, DWORD &rdwConnId, const PSOCK_ADDR pProxyAddr)
{
    DWORD  dwRet;
    UINT  nConnIdx;
    Connection  *pConnection;

    DBG(D_LOG, (L"%s: -> %#.8x/%d\n", FULL_FUNC_NAME, pRemoteAddr->IPv4Addr.dwIP, ntohs(pRemoteAddr->IPv4Addr.wPort)));

    dwRet = _ConnListMate.GetFreeIndex(nConnIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: reach the max conn limit\n", FULL_FUNC_NAME));
        goto err1;
    }

    dwRet = (pProxyAddr == NULL) ?
#ifdef CONFIG_TLS_SUPPORT
        ConnectCenter::Instance()->ConnectOutTLS(pLocalAddr, pRemoteAddr, true, &pConnection)
      : ConnectCenter::Instance()->ConnectOutTLS(pLocalAddr, pRemoteAddr, pProxyAddr, true, &pConnection);
#else
        ConnectCenter::Instance()->ConnectOut(pLocalAddr, pRemoteAddr, true, &pConnection)
      : ConnectCenter::Instance()->ConnectOut(pLocalAddr, pRemoteAddr, pProxyAddr, true, &pConnection);
#endif
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: -> %#.8x/%d failed: %d\n", FULL_FUNC_NAME, pRemoteAddr->IPv4Addr.dwIP, ntohs(pRemoteAddr->IPv4Addr.wPort), dwRet));
        goto err2;
    }

    rdwConnId = MakeConnId(nConnIdx);

    AddToConnPool(T_CONN_CONNECT_OUT, pConnection, pHandler, rdwConnId);

    return ERROR_SUCCESS;

err2:
    _ConnListMate.PutFreeIndex(nConnIdx);

err1:
    rdwConnId = INVALID_EP_CONN_ID;

    return dwRet;
}

DWORD Endpoint::QueueConnCmd(DWORD dwConnId, DWORD dwCmdMask, BufferItem *pbiCmd)
{
    DWORD  dwRet;
    UINT  nConnIdx;
    WORD  wMagic;
    PCONN_REC  pConnRec;

    nConnIdx = WIDX_FROM_DWID(dwConnId);
    wMagic = WMAGICNUM_FROM_DWID(dwConnId);

    DBG(D_VERB, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nConnIdx, wMagic));

    if ( nConnIdx >= _ConnList.size() )
    {
        DBG(D_WARN, (L"%s: invalid conn id\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_INVALID_PARAMETER;
    }

    pConnRec = &_ConnList.at(nConnIdx);

    pthread_rwlock_rdlock(&_hConnValidLock);

    if ( (!pConnRec->bValid) || (wMagic != pConnRec->wMagic) )
    {
        dwRet = ERROR_NOT_FOUND;
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nConnIdx));
        goto out;
    }
    assert(pConnRec->bUsed);

    // because using rdlock, use lock here
    pthread_mutex_lock(&_hConnCmdLock);
    pConnRec->pHandler->QueueCmd(dwCmdMask, pbiCmd);
    pConnRec->dwCmdMask |= dwCmdMask;
    pthread_mutex_unlock(&_hConnCmdLock);

    // see issues_pcond.txt
    _pEvent->Signal(nConnIdx);

    dwRet = ERROR_SUCCESS;

out:
    pthread_rwlock_unlock(&_hConnValidLock);

    return dwRet;
}

DWORD Endpoint::DisconnectConn(DWORD dwConnId)
{
    DWORD  dwRet;
    UINT  nConnIdx;
    WORD  wMagic;
    PCONN_REC  pConnRec;

    nConnIdx = WIDX_FROM_DWID(dwConnId);
    wMagic = WMAGICNUM_FROM_DWID(dwConnId);

    DBG(D_INFO, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nConnIdx, wMagic));

    if ( nConnIdx >= _ConnList.size() )
    {
        DBG(D_WARN, (L"%s: invalid conn id\n", FULL_FUNC_NAME));
        return ERROR_INVALID_PARAMETER;
    }

    pConnRec = &_ConnList.at(nConnIdx);

    pthread_rwlock_rdlock(&_hConnValidLock);

    if ( (!pConnRec->bValid) || (wMagic != pConnRec->wMagic) )
    {
        dwRet = ERROR_NOT_FOUND;
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nConnIdx));
        goto out;
    }
    assert(pConnRec->bUsed);

    pConnRec->bCmdDisconn = true;

    // see issues_pcond.txt
    _pEvent->Signal(nConnIdx);

    dwRet = ERROR_SUCCESS;

out:
    pthread_rwlock_unlock(&_hConnValidLock);

    return dwRet;
}

void Endpoint::RequestEpWork(UINT nSigIdx)
{
    DBG(D_VERB, (L"%s: sig(%d)\n", FULL_FUNC_NAME, nSigIdx));

    assert(_pEvent);
    assert(nSigIdx < _nEpSigNum);

    _pEvent->Signal(nSigIdx + _nEpSigBase);
}

void Endpoint::UnloadThreads()
{
    pthread_t  *phThread;
    UINT  i, nSize;

    DBG(D_INFO, (L"%s: ...\n", FULL_FUNC_NAME));

    _bQuit = true;

    _pEvent->Release();

    for ( i = 0, nSize = (UINT)_hWorkThreads.size(); i < nSize; ++i )
    {
        phThread = &_hWorkThreads.at(i);
        if ( IS_PTH_VALID(*phThread) )
        {
            pthread_join(*phThread, NULL);
            SET_PTH_INVALID(*phThread);
        }
    }

    _hWorkThreads.resize(0);

    DBG(D_INFO, (L"%s: done\n", FULL_FUNC_NAME));
}

DWORD Endpoint::MakeConnId(UINT nConnIdx)
{
    DWORD  dwMagic;
    WORD  wMagic;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nConnIdx));

    dwMagic = InterlockedIncrement((LONG*)&_dwConnMagic);
    wMagic = ADJUST_WMAGICNUM(dwMagic);

    return MAKE_DWID(nConnIdx, wMagic);
}

void Endpoint::AddToConnPool(CONN_TYPE ConnType, Connection *pConnection, EpProtocolHandler *pHandler, DWORD dwConnId)
{
    UINT  nConnIdx;
    WORD  wMagic;
    PCONN_REC  pConnRec;

    DBG(D_VERB, (L"%s: %x\n", FULL_FUNC_NAME, dwConnId));

    nConnIdx = WIDX_FROM_DWID(dwConnId);
    wMagic = WMAGICNUM_FROM_DWID(dwConnId);

    assert(nConnIdx < _ConnList.size());
    pConnRec = &_ConnList.at(nConnIdx);
    assert(!pConnRec->bUsed);
    DO_IF_DBG(pConnRec->bUsed = true);

    pConnRec->wMagic = wMagic;
    pConnRec->ConnType = ConnType;
    pConnRec->bBounded = false;
    pConnRec->bCmdDisconn = false;
    pConnRec->pConnection = pConnection;
    pConnRec->pHandler = pHandler;
    pConnRec->dwCmdMask = 0;

    pConnRec->bValid = true;

    _pEvent->Signal(nConnIdx);
}

void Endpoint::OnNewConnection(DWORD hListener, Connection *pListenConnection)
{
    DWORD  dwRet;
    UINT  nConnIdx;
    DWORD  dwConnId;
    Connection  *pConnection;
    EpProtocolHandler  *pHandler;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    dwRet = _ConnListMate.GetFreeIndex(nConnIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: reach the max conn limit\n", FULL_FUNC_NAME));
        goto err1;
    }

#ifdef CONFIG_TLS_SUPPORT
    dwRet = ConnectCenter::Instance()->AcceptConnectionTLS(pListenConnection, true, &pConnection);
#else
    dwRet = ConnectCenter::Instance()->AcceptConnection(pListenConnection, true, &pConnection);
#endif
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: AcceptConnection failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    dwConnId = MakeConnId(nConnIdx);

    pHandler = OnNewHandler(hListener, pConnection, dwConnId);
    if ( !pHandler )
    {
        DBG(D_WARN, (L"%s: OnNewHandler: NULL\n", FULL_FUNC_NAME));
        goto err3;
    }

    AddToConnPool(T_CONN_CONNECT_ACCEPT, pConnection, pHandler, dwConnId);

    return;

err3:
    ConnectCenter::Instance()->FinishConnection(pConnection);

err2:
    _ConnListMate.PutFreeIndex(nConnIdx);

err1:
    return;
}

void Endpoint::ReleaseConn(DWORD dwConnId, PCONN_REC pConnRec)
{
    DBG(D_LOG, (L"%s: id(%#.8x)\n", FULL_FUNC_NAME, dwConnId));

    assert(pConnRec->pHandler);
    if ( pConnRec->bBounded )
    {
        pConnRec->pHandler->Unbind();
    }

    if ( pConnRec->ConnType == T_CONN_LISTENER )
    {
        OnListenerEnded(dwConnId);
        delete pConnRec->pHandler;
    }
    else
    {
        OnDisconnected(dwConnId, pConnRec->pHandler);
    }
    pConnRec->pHandler = NULL;

    assert(pConnRec->pConnection);
    ConnectCenter::Instance()->FinishConnection(pConnRec->pConnection);
    pConnRec->pConnection = NULL;
}

bool Endpoint::ProcessConnWork(UINT nSigIdx, DWORD dwConnId, PCONN_REC pConnRec)
{
    DWORD  dwRet;
    EpProtocolHandler  *pHandler;
    DWORD  dwFlags;
    BufferItem  *pbi;
    PEP_MSG_HDR  pMsgHdr;
    DWORD  dwCmdMask;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( pConnRec->bCmdDisconn )
    {
        // no matter the value of bBounded
        return true;
    }

    pHandler = pConnRec->pHandler;

    if ( !pConnRec->bBounded )
    {
        dwRet = pHandler->Bind(pConnRec->pConnection, _pEvent, nSigIdx);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: conn(id=%#.8x) bind failed: %d\n", FULL_FUNC_NAME, dwConnId, dwRet));
            return true;
        }

        // under conn's work thread, serialization and synchronization.
        // Bind() can do everything even EndSession(), 
        pConnRec->bBounded = true;
    }

    // see issues_getflags.txt
    dwFlags = pHandler->GetFlags();

    if ( (dwFlags == F_EPH_NONE) && (!pConnRec->dwCmdMask) )
    {
        return false;
    }

    if ( dwFlags & F_EPH_RECV )
    {
        pHandler->ProcessRecv();
    }

    if ( dwFlags & F_EPH_EVENT )
    {
        while ( (pbi = pHandler->GetEvent()) )
        {
            assert(pbi->data_size() >= sizeof(EP_MSG_HDR));
            pMsgHdr = (PEP_MSG_HDR)pbi->data();
            pbi->Pull(sizeof(EP_MSG_HDR));

            OnConnEvent(dwConnId, pHandler, pMsgHdr, pbi);
            pHandler->ReleaseEvent(pbi);
        }
    }

    if ( dwFlags & F_EPH_SESSIONEND )
    {
        return true;
    }

    if ( pConnRec->dwCmdMask )
    {
        pthread_mutex_lock(&_hConnCmdLock);
        dwCmdMask = pConnRec->dwCmdMask;
        pConnRec->dwCmdMask = 0;
        pthread_mutex_unlock(&_hConnCmdLock);

        pHandler->ProcessCmd(dwCmdMask);
    }

    if ( dwFlags & F_EPH_NEWCONNECTION )
    {
        OnNewConnection(dwConnId, pConnRec->pConnection);
        pHandler->ProcessCmd(C_CONN_RESUME_LISTEN);
    }
    else if ( dwFlags & F_EPH_ASYNCCONNECTED )
    {
        OnAsyncConnected(dwConnId, pHandler);
    }

    if ( dwFlags & F_EPH_SEND )
    {
        pHandler->PrepareSend();
    }

    return false;
}

DWORD Endpoint::WorkThread(UINT nThreadIdx)
{
    UINT  nSigIdx, nEpSigIdx;
    PCONN_REC  pConnRec;
    DWORD  dwConnId;

    DBG(D_LOG, (L"%s[#%d]: enter\n", FULL_FUNC_NAME, nThreadIdx));

    while ( !_bQuit )
    {
        nSigIdx = _pEvent->Wait(nThreadIdx);

        if ( _bQuit )
        {
            break;
        }

        assert(nSigIdx != INVALID_PCOND_IDX);

        if ( nSigIdx < _nEpSigBase )
        {
            pConnRec = &_ConnList.at(nSigIdx);
            assert(pConnRec->bUsed);
            assert(pConnRec->bValid);

            dwConnId = MAKE_DWID(nSigIdx, pConnRec->wMagic);

            while ( true )
            {
                // only below here remove "this" conn, so no need to lock
                if ( ProcessConnWork(nSigIdx, dwConnId, pConnRec) )
                {
                    // use "bValid" to guarantee QueueConnCmd/DisconnectConn
                    pthread_rwlock_wrlock(&_hConnValidLock);
                    pConnRec->bValid = false;
                    pthread_rwlock_unlock(&_hConnValidLock);

                    // to guarantee "old conn" not call "Signal" during "Reset"
                    ReleaseConn(dwConnId, pConnRec);

                    _pEvent->Reset(nSigIdx, nThreadIdx);

                    // set bUsed at last to guarantee "new conn" not call "Signal" during "Reset"
                    DO_IF_DBG(pConnRec->bUsed = false);
                    _ConnListMate.PutFreeIndex(nSigIdx);

                    break;
                }

                if ( _pEvent->Finish(nSigIdx, nThreadIdx) )
                {
                    break;
                }
            }
        }
        else
        {
            nEpSigIdx = nSigIdx - _nEpSigBase;

            while ( true )
            {
                ProcessEpWork(nEpSigIdx);

                if ( _pEvent->Finish(nSigIdx, nThreadIdx) )
                {
                    break;
                }
            }
        }
    }

    DBG(D_LOG, (L"%s[#%d]: leave\n", FULL_FUNC_NAME, nThreadIdx));

    return ERROR_SUCCESS;
}

void *Endpoint::WorkThreadProc(void *arg)
{
    PTHREAD_ARG  pArg = reinterpret_cast<PTHREAD_ARG>(arg);

    pArg->pThis->WorkThread(pArg->nIdx);
    delete pArg;

    return NULL;
}

