/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "ConnectCenter.h"
#include "EpBridgeHandler.h"


EpBridgeHandler::EpBridgeHandler()
    : EpPDUHandler(0, 0, 0, 0, 0, 0)
    , _CurState(S_BRIDGE_CMD)
    , _bReqConnect(false)
    , _pOutConnection(NULL)
    , _ForwardQueue(BRIDGE_DATAQSIZE, BRIDGE_DATABSIZE, false)
    , _ReplyQueue(BRIDGE_DATAQSIZE, BRIDGE_DATABSIZE, false)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

EpBridgeHandler::~EpBridgeHandler()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_CurState == S_BRIDGE_CMD);
    assert(_pOutConnection == NULL);
}

DWORD EpBridgeHandler::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    dwRet = EpPDUHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    pConnection->ResumeRecv(_ForwardQueue.GetQueueSize());

    return ERROR_SUCCESS;

err1:
    return dwRet;
}

void EpBridgeHandler::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    // detach _pOutConnection but not release it
    // because _pConnection's OnXXXReady may call _pOutConnection's ResumeXXX
    // after _pOutConnection's detach, none will call _pConnection's ResumeXXX
    if ( _pOutConnection != NULL )
    {
        _pOutConnection->Detach();
    }

    // Unbind will detach _pConnection first
    // after detach no OnXXX functions will be called
    // in other words OnXXX functions call safely use _pConnection/_pOutConnection
    // now it is safe to release _pConnection
    EpPDUHandler::Unbind();

    // after _pConnection's detach, none will call _pOutConnection's ResumeXXX
    // so it is safe to release _pOutConnection
    if ( _pOutConnection != NULL )
    {
        ConnectCenter::Instance()->FinishConnection(_pOutConnection);
        _pOutConnection = NULL;
    }
    _bReqConnect = false;

    _CurState = S_BRIDGE_CMD;
}

void EpBridgeHandler::OnAsyncConnected(Connection *pConnection)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    assert(pConnection == _pOutConnection);

    EnterBridgeMode();
}

void EpBridgeHandler::OnDisconnectedOrError(Connection *pConnection)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    ReqEndSession();
}

UINT EpBridgeHandler::OnRecvReady(Connection *pConnection)
{
    DWORD  dwRet;
    BufferItem  *pbiData;
    ProducerQueue  *pRecvQueue;
    Connection  *pSendConnection;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _CurState == S_BRIDGE_END )
    {
        DBG(D_VERB, (L"%s: bridge end\n", FULL_FUNC_NAME));
        return ABORT;
    }

    if ( pConnection == _pConnection )
    {
        pRecvQueue = &_ForwardQueue;
        pSendConnection = _pOutConnection;
    }
    else
    {
        assert(pConnection == _pOutConnection);
        pRecvQueue = &_ReplyQueue;
        pSendConnection = _pConnection;
    }

    pbiData = pRecvQueue->FreeItem(PDU_HEAD_RESERVE);
    if ( pbiData == NULL )
    {
        DBG(D_WARN, (L"%s: no free item\n", FULL_FUNC_NAME));
        assert(0);
        return ABORT;
    }

    dwRet = pConnection->Recv(pbiData);
    if ( dwRet != ERROR_SUCCESS )
    {
        if ( (dwRet == EAGAIN) || (dwRet == EWOULDBLOCK) )
        {
            return AGAIN;
        }

        DBG(D_INFO, (L"%s: recv failed: %d\n", FULL_FUNC_NAME, dwRet));
        ReqEndSession();
        return ABORT;
    }

    if ( _CurState != S_BRIDGE_ENABLED )
    {
        DBG(D_INFO, (L"%s: cmd mode\n", FULL_FUNC_NAME));
        assert(pConnection == _pConnection);

        ProcessRecvData(pbiData);

        // no pop and return "1", so ...
        // why not remove ResumeRecv and return "0"?
        // see issues_OnRecvReady.txt
        // pConnection->ResumeRecv(1);
        // now return "0" has new meaning
        return 0;
    }

    pRecvQueue->PopFreeItem();
    pRecvQueue->PutDataItem(pbiData);
    pSendConnection->ResumeSend(1);

    return 1;
}

UINT EpBridgeHandler::OnSendReady(Connection *pConnection)
{
    DWORD  dwRet;
    BufferItem  *pbiData;
    ProducerQueue  *pSendQueue;
    Connection  *pRecvConnection;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _CurState == S_BRIDGE_END )
    {
        DBG(D_VERB, (L"%s: bridge end\n", FULL_FUNC_NAME));
        return ABORT;
    }

    if ( pConnection == _pConnection )
    {
        pSendQueue = &_ReplyQueue;
        pRecvConnection = _pOutConnection;
    }
    else
    {
        assert(pConnection == _pOutConnection);
        pSendQueue = &_ForwardQueue;
        pRecvConnection = _pConnection;
    }

    pbiData = pSendQueue->DataItem();
    if ( pbiData == NULL )
    {
        DBG(D_WARN, (L"%s: no data item\n", FULL_FUNC_NAME));
        assert(0);
        return ABORT;
    }

    dwRet = pConnection->Send(pbiData);
    if ( dwRet != ERROR_SUCCESS )
    {
        if ( (dwRet == EAGAIN) || (dwRet == EWOULDBLOCK) )
        {
            return AGAIN;
        }

        DBG(D_INFO, (L"%s: send failed: %d\n", FULL_FUNC_NAME, dwRet));
        ReqEndSession();
        return ABORT;
    }

    if ( pbiData->data_size() )
    {
        DBG(D_VERB, (L"%s: partial data\n", FULL_FUNC_NAME));
        return 0;
    }

    pSendQueue->PopDataItem();

    pSendQueue->PutFreeItem(pbiData);
    pRecvConnection->ResumeRecv(1);

    return 1;
}

DWORD EpBridgeHandler::GetFlags()
{
    DWORD  dwRet;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( _bReqConnect )
    {
        _bReqConnect = false;
        dwRet = OpenConnection(&_RemoteAddr);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_INFO, (L"%s: OpenConnection failed: %d\n", FULL_FUNC_NAME, dwRet));
            ReqEndSession();
        }
    }

    return EpPDUHandler::GetFlags();
}

void EpBridgeHandler::PrepareSend()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

void EpBridgeHandler::ReqEndSession()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    _CurState = S_BRIDGE_END;

    EpPDUHandler::ReqEndSession();
}

void EpBridgeHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    assert(bReq);

    switch ( bCode )
    {
        case C_BRIDGE_OPEN:
            assert(pbiBody->data_size() == sizeof(REQ_BRIDGE_OPEN));
            OnOpenBridgeReq((PREQ_BRIDGE_OPEN)pbiBody->data());
            break;

        default:
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, bCode));
            assert(0);
            ReqEndSession();
            break;
    }
}

void EpBridgeHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    DBG(D_WARN, (L"%s: stream data(code=%d)\n", FULL_FUNC_NAME, bCode));
    assert(0);
    ReqEndSession();
}

void EpBridgeHandler::EnterBridgeMode()
{
    UINT  nBufCnt;
    BufferItem  *pbi;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    // may OnDisconnected(_pConnection) and call ReqEndSession
    // then OnAsyncConnected(_pOutConnection) occur
//    assert(_CurState == S_BRIDGE_CMD);
    // set state from END to BRIDGE is no harm
    _CurState = S_BRIDGE_ENABLED;

    // no data item, nBufCnt==QueueSize
    assert(!_ReplyQueue.HasDataItem());
    nBufCnt = _ReplyQueue.GetQueueSize();

    pbi = _ReplyQueue.GetFreeItem(PDU_HEAD_RESERVE);
    assert(pbi != NULL);
    --nBufCnt;

    BuildPDU(C_BRIDGE_OPEN, false, true, ERROR_SUCCESS, pbi);

    _ReplyQueue.PutDataItem(pbi);
    _pConnection->ResumeSend(1);

    // until now OnRecvReady@_pOutConnection will be called
    // "-1" because "_OutRecvQueue.GetFreeItem" will cause FreeSize-1
    // but _pOutConnection's _RecvCounter not know it
    // _RecvCounter must equ FreeSize
    _pOutConnection->ResumeRecv(nBufCnt);
}

void EpBridgeHandler::OnOpenBridgeReq(PREQ_BRIDGE_OPEN pOpenReq)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    _RemoteAddr = pOpenReq->addr;

    _bReqConnect = true;
    SignalEp();
}

DWORD EpBridgeHandler::OpenConnection(const PSOCK_ADDR pRemoteAddr)
{
    DWORD  dwRet;
    SOCK_ADDR  LocalAddr;

    DBG(D_INFO, (L"%s: -> %#.8x/%d\n", FULL_FUNC_NAME, pRemoteAddr->IPv4Addr.dwIP, ntohs(pRemoteAddr->IPv4Addr.wPort)));

    // OpenConnection & Unbind called under ConnThread,
    // so only create or destroy _pOutConnection,
    // don't access Recv/SendBuf and ResumeXXX

    // OnXXX functions called under MuxThread
    // so only access Recv/SendBuf and ResumeXXX
    // don't create or destroy _pOutConnection

    // no need to check state(cmd mode) then do connect
    // because state may be changed during check
    // unlikely to happen the state is not cmd mode
    // enen state is not cmd mode and do connect, no harm
    if ( _pOutConnection != NULL )
    {
        dwRet = ERROR_INVALID_STATE;
        DBG(D_WARN, (L"%s: out connection already exist\n", FULL_FUNC_NAME));
        assert(0);
        goto err1;
    }

    LocalAddr.bType = pRemoteAddr->bType;
    LocalAddr.IPv4Addr.bStream = pRemoteAddr->IPv4Addr.bStream;
    LocalAddr.IPv4Addr.dwIP = INADDR_ANY;
    LocalAddr.IPv4Addr.wPort = 0;
#ifdef CONFIG_TLS_SUPPORT
    dwRet = ConnectCenter::Instance()->ConnectOutTLS(&LocalAddr, pRemoteAddr, true, &_pOutConnection);
#else
    dwRet = ConnectCenter::Instance()->ConnectOut(&LocalAddr, pRemoteAddr, true, &_pOutConnection);
#endif
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ConnectOut failed: %d\n", FULL_FUNC_NAME, dwRet));
        // socket@CreateAndBind may fail with EMFILE(Too many open files)
//        assert(0);
        goto err1;
    }

    // "0,0" means no On{Recv/Send}Ready on _pOutConnection
    // until _pOutConnection's OnAsyncConnected()
    dwRet = _pOutConnection->Attach(this, 0, 0);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: Attach failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    ConnectCenter::Instance()->FinishConnection(_pOutConnection);
    _pOutConnection = NULL;

err1:
    return dwRet;
}
