/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "EpProtocolHandler.h"


void EpBuildMsg(UINT nCode, DWORD dwCookie, DWORD dwParam, BufferItem *pbiBody)
{
    PEP_MSG_HDR  pMsgHdr;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(pbiBody->head_size() >= sizeof(EP_MSG_HDR));
    pMsgHdr = (PEP_MSG_HDR)pbiBody->Push(sizeof(EP_MSG_HDR));

    pMsgHdr->nCode = nCode;
    pMsgHdr->dwCookie = dwCookie;
    pMsgHdr->dwParam = dwParam;
}


EpProtocolHandler::EpProtocolHandler
        (
            UINT nRecvQSize, UINT nRecvBSize,
            UINT nSendQSize, UINT nSendBSize,
            UINT nEvtQSize, UINT nEvtBSize, bool bEvtLock
        )
    : _pConnection(NULL)
    , _pEvent(NULL)
    , _nSigIdx(INVALID_PCOND_IDX)
    , _bSessionEnd(false)
    , _bAsyncConnected(false)
    , _bReqSendProc(false)
    , _RecvQueue(nRecvQSize, nRecvBSize?(nRecvBSize+PDU_HEAD_RESERVE):0, false)
    , _SendQueue(nSendQSize, nSendBSize?(nSendBSize+PDU_HEAD_RESERVE):0, false)
    , _EventQueue(nEvtQSize, nEvtBSize, bEvtLock)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _SendCounter.Reset(nSendQSize);
}

EpProtocolHandler::~EpProtocolHandler()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pConnection == NULL);
}

DWORD EpProtocolHandler::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s: SigIdx(%d)\n", FULL_FUNC_NAME, nSigIdx));

    _MesaRecv.Reset();

    _pEvent = pEvent;
    _nSigIdx = nSigIdx;

    _bSessionEnd = false;

    assert(!_pConnection);
    _pConnection = pConnection;

    dwRet = _pConnection->Attach(this, _RecvQueue.GetQueueSize(), 0);
    if ( dwRet != ERROR_SUCCESS)
    {
        DBG(D_WARN, (L"%s: Attach failed: %d\n", FULL_FUNC_NAME, dwRet));
        _pConnection = NULL;
    }

    return dwRet;
}

void EpProtocolHandler::Unbind()
{
    BufferItem  *pbi;

    DBG(D_INFO, (L"%s: SigIdx(%d)\n", FULL_FUNC_NAME, _nSigIdx));

    assert(_pConnection);
    _pConnection->Detach();
    _pConnection = NULL;

    _bSessionEnd = true;
    _bAsyncConnected = false;
    _bReqSendProc = false;

    // after pConnection->Detach,
    // otherwise PutEvent@ListenerHandler::OnRecvReady may crash
    _pEvent = NULL;
    _nSigIdx = INVALID_PCOND_IDX;

    while ( (pbi = _RecvQueue.GetDataItem()) )
    {
        _RecvQueue.PutFreeItem(pbi);
    }
    _RecvCounter.Reset(0);

    while ( (pbi = _SendQueue.GetDataItem()) )
    {
        ReleaseSendData(pbi);
    }

    while ( (pbi = _EventQueue.GetDataItem()) )
    {
        ReleaseEvent(pbi);
    }

    _MesaRecv.Print("EPH_RECV");
}

void EpProtocolHandler::OnAsyncConnected(Connection *pConnection)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    assert(pConnection == _pConnection);

    _bAsyncConnected = true;

    SignalEp();
}

void EpProtocolHandler::OnDisconnectedOrError(Connection *pConnection)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    assert(pConnection == _pConnection);

    ReqEndSession();
}

UINT EpProtocolHandler::OnRecvReady(Connection *pConnection)
{
    DWORD  dwRet;
    BufferItem  *pbiData;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(pConnection == _pConnection);

    if ( _bSessionEnd )
    {
        DBG(D_VERB, (L"%s: session end\n", FULL_FUNC_NAME));
        return ABORT;
    }

    pbiData = _RecvQueue.FreeItem(PDU_HEAD_RESERVE);
    if ( !pbiData )
    {
        // should never happen
        DBG(D_WARN, (L"%s: no free item\n", FULL_FUNC_NAME));
        assert(0);
        return ABORT;
    }

    dwRet = pConnection->Recv(pbiData);
    if ( dwRet != ERROR_SUCCESS )
    {
        if ( (dwRet == EAGAIN) || (dwRet == EWOULDBLOCK) )
        {
            // see issues_mux_eagain.txt
            // after select/epoll, there still may EAGAIN
            return AGAIN;
        }

        DBG(D_INFO, (L"%s: recv failed: %d\n", FULL_FUNC_NAME, dwRet));
        ReqEndSession();
        return ABORT;
    }

    _RecvQueue.PopFreeItem();

    PutRecvData(pbiData);

    return 1;
}

UINT EpProtocolHandler::OnSendReady(Connection *pConnection)
{
    DWORD  dwRet;
    BufferItem  *pbiData;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(pConnection == _pConnection);

    if ( _bSessionEnd )
    {
        DBG(D_INFO, (L"%s: session end\n", FULL_FUNC_NAME));
        return ABORT;
    }

    pbiData = _SendQueue.DataItem();
    if ( !pbiData )
    {
        // should never happen
        DBG(D_WARN, (L"%s: no data item\n", FULL_FUNC_NAME));
        assert(0);
        return ABORT;
    }

    dwRet = pConnection->Send(pbiData);
    if ( dwRet != ERROR_SUCCESS )
    {
        if ( (dwRet == EAGAIN) || (dwRet == EWOULDBLOCK) )
        {
            // see issues_mux_eagain.txt
            // after select/epoll, there still may EAGAIN
            return AGAIN;
        }

        DBG(D_INFO, (L"%s: send failed: %d\n", FULL_FUNC_NAME, dwRet));
        ReqEndSession();
        return ABORT;
    }

    if ( pbiData->data_size() )
    {
        DBG(D_VERB, (L"%s: partial data\n", FULL_FUNC_NAME));
        return 0;
    }

    _SendQueue.PopDataItem();
    if ( ReleaseSendData(pbiData) )
    {
        ReqSendProc(false);
    }

    return 1;
}

void EpProtocolHandler::QueueCmd(DWORD dwCmdMask, BufferItem *pbiCmd)
{
    DBG(D_VERB, (L"%s: mask(%x)\n", FULL_FUNC_NAME, dwCmdMask));
}

void EpProtocolHandler::ProcessCmd(DWORD dwCmdMask)
{
    DBG(D_VERB, (L"%s: mask(%d)\n", FULL_FUNC_NAME, dwCmdMask));
}

DWORD EpProtocolHandler::GetFlags()
{
    DWORD  dwFlags = F_EPH_NONE;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( _bSessionEnd )
    {
        dwFlags |= F_EPH_SESSIONEND;
    }

    if ( _bAsyncConnected )
    {
        _bAsyncConnected = false;
        dwFlags |= F_EPH_ASYNCCONNECTED;
    }

    // not use _RecvQueue.HasDataItem()
    // see issues_rscounter.txt
    if ( !_RecvCounter.IsEmpty() )
    {
        dwFlags |= F_EPH_RECV;
    }

    // not use _SendQueue.HasFreeItem()
    // see issues_rscounter.txt
    if ( _bReqSendProc && (!_SendCounter.IsEmpty()) )
    {
        dwFlags |= F_EPH_SEND;
    }

    if ( _EventQueue.HasDataItem() )
    {
        dwFlags |= F_EPH_EVENT;
    }

    return dwFlags;
}

void EpProtocolHandler::ProcessRecv()
{
    BufferItem  *pbiData;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _MesaRecv.Start();

    while ( !_RecvCounter.IsEmpty() )
    {
        pbiData = _RecvQueue.GetDataItem();
        _RecvCounter.Consume(1);

        _MesaRecv.Update(pbiData->data_size());

        if ( ProcessRecvData(pbiData) )
        {
            ReleaseRecvData(pbiData);
        }
    }

    _MesaRecv.Finish();
}

BufferItem *EpProtocolHandler::GetEvent()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return _EventQueue.GetDataItem();
}

void EpProtocolHandler::ReleaseEvent(BufferItem *pbiEvt)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _EventQueue.PutFreeItem(pbiEvt);
}

void EpProtocolHandler::SignalEp()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pEvent);
    _pEvent->Signal(_nSigIdx);
}

void EpProtocolHandler::ReqEndSession()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    _bSessionEnd = true;

    SignalEp();
}

void EpProtocolHandler::ReqSendProc(bool bSetReq)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( bSetReq )
    {
        _bReqSendProc = true;
    }

    if ( _bReqSendProc )
    {
        SignalEp();
    }
}

BufferItem *EpProtocolHandler::AllocEvent(bool bSync)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return _EventQueue.GetFreeItem(MSG_HEAD_RESERVE, bSync);
}

void EpProtocolHandler::PutEvent(BufferItem *pbiEvt, bool bSync)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    _EventQueue.PutDataItem(pbiEvt, bSync);

    SignalEp();
}

void EpProtocolHandler::PutRecvData(BufferItem *pbiData)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _RecvQueue.PutDataItem(pbiData);

    if ( _RecvCounter.Produce(1) )
    {
        SignalEp();
    }
}

void EpProtocolHandler::ReleaseRecvData(BufferItem *pbiData)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _RecvQueue.PutFreeItem(pbiData);

    if ( !_bSessionEnd )
    {
        _pConnection->ResumeRecv(1);
    }
}

BufferItem *EpProtocolHandler::GetSendBuf()
{
    BufferItem  *pbi = NULL;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( !_SendCounter.IsEmpty() )
    {
        pbi = _SendQueue.GetFreeItem(PDU_HEAD_RESERVE);
        _SendCounter.Consume(1);
    }

    return pbi;
}

void EpProtocolHandler::PutSendData(BufferItem *pbiData, bool bLastOne)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _bReqSendProc = !bLastOne;

    _SendQueue.PutDataItem(pbiData);

    _pConnection->ResumeSend(1);
}

bool EpProtocolHandler::ReleaseSendData(BufferItem *pbiData)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _SendQueue.PutFreeItem(pbiData);

    return _SendCounter.Produce(1);
}

