/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "ProxyFilterAgent.h"


ProxyFilterAgent::ProxyFilterAgent(Connection *pConnection, const PSOCK_ADDR pRemoteAddr)
    : FilterProtocol(pConnection)
    , _RemoteAddr(*pRemoteAddr)
    , _CurState(S_BRIDGE_CMD)
    , _nRecvBufCnt(0)
    , _nSendItemCnt(0)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_init(&_hStateLock, NULL);

    _biSendBuf.Alloc(PDU_HEAD_RESERVE + sizeof(REQ_BRIDGE_OPEN));
    _biRecvBuf.Alloc(sizeof(PDU_COMM_HDR));
}

ProxyFilterAgent::~ProxyFilterAgent()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_destroy(&_hStateLock);
}

DWORD ProxyFilterAgent::Attach(ProtocolHandler *pHandler, UINT nRecvBufCnt, UINT nSendItemCnt)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _nRecvBufCnt = nRecvBufCnt;
    _nSendItemCnt = nSendItemCnt;

    // (1,1) means will send & recv one PDU
    return FilterProtocol::Attach(pHandler, 1, 1);
}

void ProxyFilterAgent::ResumeRecv(UINT nBufCnt)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nBufCnt));

    if ( _CurState == S_BRIDGE_ENABLED )
    {
        _pConnection->ResumeRecv(nBufCnt);
    }
    else
    {
        pthread_mutex_lock(&_hStateLock);
        if ( _CurState == S_BRIDGE_ENABLED )
        {
            _pConnection->ResumeRecv(nBufCnt);
        }
        else
        {
            _nRecvBufCnt += nBufCnt;
        }        
        pthread_mutex_unlock(&_hStateLock);
    }
}

void ProxyFilterAgent::ResumeSend(UINT nItemCnt)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nItemCnt));

    if ( _CurState == S_BRIDGE_ENABLED )
    {
        _pConnection->ResumeSend(nItemCnt);
    }
    else
    {
        pthread_mutex_lock(&_hStateLock);
        if ( _CurState == S_BRIDGE_ENABLED )
        {
	        _pConnection->ResumeSend(nItemCnt);
        }
        else
        {
            _nSendItemCnt += nItemCnt;
        }        
        pthread_mutex_unlock(&_hStateLock);
    }
}

UINT ProxyFilterAgent::OnRecvReady(Connection *pConnection)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(pConnection == _pConnection);

    if ( _CurState == S_BRIDGE_ENABLED )
    {
        return _pHandler->OnRecvReady(this);
    }

    return OnOpenBridgeRep();
}

UINT ProxyFilterAgent::OnSendReady(Connection *pConnection)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(pConnection == _pConnection);

    if ( _CurState == S_BRIDGE_ENABLED )
    {
        return _pHandler->OnSendReady(this);
    }

    return SendOpenBridgeReq();
}

UINT ProxyFilterAgent::SendOpenBridgeReq()
{
    DWORD  dwRet;
    PREQ_BRIDGE_OPEN  pOpenReq;
    PPDU_COMM_HDR  pHdr;

    DBG(D_INFO, (L"%s: -> %#.8x/%d\n", FULL_FUNC_NAME, _RemoteAddr.IPv4Addr.dwIP, ntohs(_RemoteAddr.IPv4Addr.wPort)));

    _biSendBuf.Reset(PDU_HEAD_RESERVE);
    _biRecvBuf.Reset();

    pOpenReq = (PREQ_BRIDGE_OPEN)_biSendBuf.Put(sizeof(REQ_BRIDGE_OPEN));
    pOpenReq->addr = _RemoteAddr;

    pHdr = (PPDU_COMM_HDR)_biSendBuf.Push(sizeof(PDU_COMM_HDR));
    pHdr->dwMagic = htonl(PDU_MAGIC_NUM);
    pHdr->wBodySize = htons(sizeof(REQ_BRIDGE_OPEN));
    pHdr->bCode = C_BRIDGE_OPEN;
    pHdr->bReq = 1;
    pHdr->bBodyPDU = 1;
    pHdr->dwParam = 0;

    dwRet = _pConnection->Send(&_biSendBuf);
    assert((dwRet != EAGAIN) && (dwRet != EWOULDBLOCK));
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: send failed: %d\n", FULL_FUNC_NAME, dwRet));
        OnDisconnectedOrError(_pConnection);
        return ABORT;
    }

    // Attach(1,1), OnSendReady will not be called
    return 1;
}

UINT ProxyFilterAgent::OnOpenBridgeRep()
{
    DWORD  dwRet;
    PPDU_COMM_HDR  pHdr;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    assert(_biRecvBuf.data_size() < sizeof(PDU_COMM_HDR));

    dwRet = _pConnection->Recv(&_biRecvBuf);
    if ( (dwRet == EAGAIN) || (dwRet == EWOULDBLOCK) )
    {
        DBG(D_INFO, (L"%s: EAGAIN\n", FULL_FUNC_NAME));
        return AGAIN;
    }

    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ReadOpenBridgeRep failed: %d\n", FULL_FUNC_NAME, dwRet));
        OnDisconnectedOrError(_pConnection);
        return ABORT;
    }

    if ( _biRecvBuf.data_size() < sizeof(PDU_COMM_HDR) )
    {
        return 0;
    }

    pHdr = (PPDU_COMM_HDR)_biRecvBuf.data();
    if ( (pHdr->dwMagic != htonl(PDU_MAGIC_NUM)) || (pHdr->bCode != C_BRIDGE_OPEN) || (pHdr->dwParam != ERROR_SUCCESS) )
    {
        DBG(D_WARN, (L"%s: wrong pdu or failed: %d\n", FULL_FUNC_NAME, dwRet));
        OnDisconnectedOrError(_pConnection);
        return ABORT;
    }

    EnterBridgeMode();

    // see issues_OnRecvReady.txt
    // EnterBridgeMode will ResetRecv
    // now protocol may call ResumeRecv at the same time
    // so "+1" in EnterBridgeMode
    // _pConnection->ResumeRecv(1);
    // now "0" has new meaning
    return 0;
}

void ProxyFilterAgent::EnterBridgeMode()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_lock(&_hStateLock);

    // correct _RSCnter because Attach(1,1)@Attach,
    // the below ResumeXXX will cause resume polling
    // "+1" because ProcessRecvReady will return "1" after EnterBridgeMode
    // use ResetXXX so no need "+1"
    // "+1" to avoid conflict do at OnOpenBridgeRep
    // OnOpenBridgeRep return 0, so no need "_nRecvBufCnt+1"
    _pConnection->ResetRecv(_nRecvBufCnt);
    _pConnection->ResetSend(_nSendItemCnt);

    // ResumeXXX will not lock if S_BRIDGE_ENABLED state
    // so put at last
    _CurState = S_BRIDGE_ENABLED;

    pthread_mutex_unlock(&_hStateLock);
}
