/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "SignalPipe.h"
#include "SocketConnection.h"
#include "ReactorEPoll.h"


#ifdef CONFIG_REACTOR_EPOLL

ReactorEPoll *ReactorEPoll::_instance = NULL;

ReactorEPoll *ReactorEPoll::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new ReactorEPoll();
    }
    return _instance;
}

void ReactorEPoll::Destroy()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

ReactorEPoll::ReactorEPoll()
    : _hEPoll(INVALID_FD)
    , _hPipeSend(INVALID_SOCKET)
    , _hPipeRecv(INVALID_SOCKET)
    , _bQuit(false)
    , _MuxList(0)
    , _MuxListMate(true)
    , _dwMuxMagic(0)
    , _YieldList(EPOLL_EVT_NUM+EPOLL_MSG_NUM+1, false)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    SET_PTH_INVALID(_hMuxThread);
}

ReactorEPoll::~ReactorEPoll()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(!IS_PTH_VALID(_hMuxThread));
}

DWORD ReactorEPoll::Load(UINT nMaxEPollNum)
{
    DWORD  dwRet;
    PEPOLL_REC  pEPRec;

    DBG(D_LOG, (L"%s: MaxEPollNum=%d\n", FULL_FUNC_NAME, nMaxEPollNum));

    DO_IF_DBG(memset(&_stats, 0, sizeof(_stats)));

    _hEPoll = epoll_create(nMaxEPollNum);
    if ( _hEPoll == INVALID_FD )
    {
        dwRet = GetLastError();
        DBG(D_ERROR, (L"%s: epoll_create failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    dwRet = CreateSignalPipe(&_hPipeSend, &_hPipeRecv);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: create signal pipe failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    dwRet = AddEvent(_hPipeRecv, 0, true, false, false);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: AddEvent failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err3;
    }

    // [0] for SignalPipe
    _MuxList.resize(nMaxEPollNum + 1);
    dwRet = _MuxListMate.Init(nMaxEPollNum, 1);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: init VectorMate failed\n", FULL_FUNC_NAME));
        goto err4;
    }

    for ( UINT i = 0; i < nMaxEPollNum; ++i )
    {
        pEPRec = &_MuxList.at(i);
        DO_IF_DBG(pEPRec->bUsed = false);
        pEPRec->bValid = false;
    }

    _bQuit = false;
    if ( pthread_create(&_hMuxThread, NULL, MuxThreadProc, this) )
    {
        dwRet = GetLastError();
        DBG(D_ERROR, (L"%s: create mux thread failed: %d\n", FULL_FUNC_NAME, dwRet));
        SET_PTH_INVALID(_hMuxThread);
        goto err5;
    }

    return ERROR_SUCCESS;

err5:
    _MuxList.resize(0);
    _MuxListMate.Fini();

err4:
    RemoveEvent(_hPipeRecv);

err3:
    CloseSignalPipe(_hPipeSend);
    _hPipeSend = INVALID_SOCKET;
    CloseSignalPipe(_hPipeRecv);
    _hPipeRecv = INVALID_SOCKET;

err2:
    close(_hEPoll);
    _hEPoll = INVALID_FD;

err1:
    return dwRet;
}

void ReactorEPoll::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    _bQuit = true;

    if ( _hPipeSend != INVALID_SOCKET )
    {
        // will cause epoll_wait return
        CloseSignalPipe(_hPipeSend);
        _hPipeSend = INVALID_SOCKET;
    }

    if ( IS_PTH_VALID(_hMuxThread) )
    {
        pthread_join(_hMuxThread, NULL);
        SET_PTH_INVALID(_hMuxThread);
    }

    assert(_YieldList.IsEmpty());

    _MuxList.resize(0);
    _MuxListMate.Fini();

    if ( _hPipeRecv != INVALID_SOCKET )
    {
        RemoveEvent(_hPipeRecv);

        CloseSignalPipe(_hPipeRecv);
        _hPipeRecv = INVALID_SOCKET;
    }

    if ( _hEPoll != INVALID_FD )
    {
        close(_hEPoll);
        _hEPoll = INVALID_FD;
    }

#ifndef NDEBUG
    DBG(D_LOG, (L"\n***** epoll statistics data *****\n"));
    DBG(D_LOG, (L"poll count: %lld\n", _stats.llPollCnt));
    DBG(D_LOG, (L"modify count: %lld\n", _stats.llModifyCnt));
    DBG(D_LOG, (L"yield count: %lld\n", _stats.llYieldCnt));
    DBG(D_LOG, (L"msg call count: %lld\n", _stats.llMsgCallCnt));
    DBG(D_LOG, (L"msg send count: %lld\n", _stats.llMsgSendCnt));
    DBG(D_LOG, (L"***** epoll statistics data *****\n\n"));
#endif
}

DWORD ReactorEPoll::Register(SocketConnection *pSocketConn, DWORD &rdwMuxId)
{
    DWORD  dwRet;
    UINT  nIdx;
    DWORD  dwMagic;
    PEPOLL_REC  pEPRec;

    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, (int)pSocketConn->_s));

    dwRet = _MuxListMate.GetFreeIndex(nIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: reach the max epoll limit\n", FULL_FUNC_NAME));
        assert(0);
        goto err1;
    }

    pEPRec = &_MuxList.at(nIdx);
    assert(!pEPRec->bUsed);
    DO_IF_DBG(pEPRec->bUsed = true);

    dwMagic = InterlockedIncrement((LONG*)&_dwMuxMagic);

    pEPRec->wMagic = ADJUST_WMAGICNUM(dwMagic);
    pEPRec->pSocketConn = pSocketConn;
    pEPRec->CtlReq = EPOLL_REQ_NONE;
    pEPRec->bMsgSend = false;
    pEPRec->bRemoved = false;
    pEPRec->bRecvInd = pSocketConn->ReqRecvInd();
    pEPRec->bRecvReady = false;
    pEPRec->bSendInd = pSocketConn->ReqSendInd() || pSocketConn->_bAsyncConnect;
    pEPRec->bSendReady = false;
    pEPRec->bValid = true;

    // see issues_connid.txt
    rdwMuxId = MAKE_DWID(nIdx, pEPRec->wMagic);

#ifdef CONFIG_EPOLL_EDGETRIGGER
    // see issues_epoll_eagain.txt
    dwRet = AddEvent(pSocketConn->_s, rdwMuxId, true, !pSocketConn->_bListener, !pSocketConn->_bListener);
#else
    dwRet = AddEvent(pSocketConn->_s, rdwMuxId, pEPRec->bRecvInd, pEPRec->bSendInd, false);
#endif
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: AddEvent failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    DBG(D_INFO, (L"%s: socket(%d), idx(%d)\n", FULL_FUNC_NAME, pSocketConn->_s, nIdx));

    return ERROR_SUCCESS;

err2:
    pEPRec->bValid = false;
    DO_IF_DBG(pEPRec->bUsed = false);
    _MuxListMate.PutFreeIndex(nIdx);

err1:
    rdwMuxId = INVALID_MUX_ID;

    return dwRet;
}

DWORD ReactorEPoll::Unregister(DWORD dwMuxId)
{
    UINT  nMuxIdx;
    WORD  wMagic;
    PEPOLL_REC  pEPRec;

    nMuxIdx = WIDX_FROM_DWID(dwMuxId);
    wMagic = WMAGICNUM_FROM_DWID(dwMuxId);

    DBG(D_INFO, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nMuxIdx, wMagic));

    if ( nMuxIdx >= _MuxList.size() )
    {
        DBG(D_WARN, (L"%s: idx(%d) invalid\n", FULL_FUNC_NAME, nMuxIdx));
        assert(0);
        return ERROR_INVALID_PARAMETER;
    }

    pEPRec = &_MuxList.at(nMuxIdx);
    if ( wMagic != pEPRec->wMagic )
    {
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nMuxIdx));
        assert(0);
        return ERROR_NOT_FOUND;
    }
    assert(pEPRec->bUsed && pEPRec->bValid);

    pEPRec->CtlReq = EPOLL_REQ_REMOVE;
    // see issues_mb3.txt
    MemoryBarrier();
    SendMuxMsg(dwMuxId, pEPRec);
    Sleep(10);
    while ( pEPRec->CtlReq != EPOLL_REQ_DONE )
    {
        DBG(D_INFO, (L"%s: wait MuxId(%x) remove\n", FULL_FUNC_NAME, dwMuxId));
        Sleep(100);
    }

    // see issues_epoll2.txt
    // see issues_epoll3.txt
    pEPRec->bValid = false;

    DO_IF_DBG(pEPRec->bUsed = false);
    _MuxListMate.PutFreeIndex(nMuxIdx);

    return ERROR_SUCCESS;
}

void ReactorEPoll::ResumePolling(DWORD dwMuxId)
{
    UINT  nMuxIdx;
    WORD  wMagic;
    PEPOLL_REC  pEPRec;

    nMuxIdx = WIDX_FROM_DWID(dwMuxId);
    wMagic = WMAGICNUM_FROM_DWID(dwMuxId);

    DBG(D_VERB, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nMuxIdx, wMagic));

    if ( nMuxIdx >= _MuxList.size() )
    {
        DBG(D_WARN, (L"%s: idx(%d) invalid\n", FULL_FUNC_NAME, nMuxIdx));
        assert(0);
        return;
    }

    pEPRec = &_MuxList.at(nMuxIdx);
    if ( wMagic != pEPRec->wMagic )
    {
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nMuxIdx));
        assert(0);
        return;
    }

    // see issues_epoll3.txt
    // test for call resume before Unregister
    assert(pEPRec->bUsed && pEPRec->bValid);

    SendMuxMsg(dwMuxId, pEPRec);
}

DWORD ReactorEPoll::AddEvent(SOCKET s, DWORD dwMagic, bool bRecvInd, bool bSendInd, bool bEdgeTrigger)
{
    DWORD  dwRet;
    struct epoll_event  epevt;

    DBG(D_INFO, (L"%s: socket(%d), magic(%#.8x), r/s(%d/%d), et(%d)\n", FULL_FUNC_NAME, s, dwMagic, bRecvInd, bSendInd, bEdgeTrigger));

    epevt.events = EPOLLRDHUP;
#ifdef CONFIG_EPOLL_EDGETRIGGER
    if ( bEdgeTrigger )
    {
        epevt.events |= EPOLLET;
    }
#endif
    if ( bRecvInd )
    {
        epevt.events |= EPOLLIN;
    }
    if ( bSendInd )
    {
        epevt.events |= EPOLLOUT;
    }
    epevt.data.u32 = dwMagic;

    if ( epoll_ctl(_hEPoll, EPOLL_CTL_ADD, s, &epevt) )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: epoll_ctl failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    return ERROR_SUCCESS;
}

DWORD ReactorEPoll::ModifyEvent(SOCKET s, DWORD dwMagic, bool bRecvInd, bool bSendInd, bool bEdgeTrigger)
{
    DWORD  dwRet;
    struct epoll_event  epevt;

#ifdef CONFIG_EPOLL_EDGETRIGGER
    DBG(D_INFO, (L"%s: socket(%d), magic(%#.8x), r/s(%d/%d), et(%d)\n", FULL_FUNC_NAME, s, dwMagic, bRecvInd, bSendInd, bEdgeTrigger));
#else
    DBG(D_VERB, (L"%s: socket(%d), magic(%#.8x), r/s(%d/%d), et(%d)\n", FULL_FUNC_NAME, s, dwMagic, bRecvInd, bSendInd, bEdgeTrigger));
#endif

    epevt.events = EPOLLRDHUP;
#ifdef CONFIG_EPOLL_EDGETRIGGER
    if ( bEdgeTrigger )
    {
        epevt.events |= EPOLLET;
    }
#endif
    if ( bRecvInd )
    {
        epevt.events |= EPOLLIN;
    }
    if ( bSendInd )
    {
        epevt.events |= EPOLLOUT;
    }
    epevt.data.u32 = dwMagic;

    DO_IF_DBG(++_stats.llModifyCnt);
    if ( epoll_ctl(_hEPoll, EPOLL_CTL_MOD, s, &epevt) )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: epoll_ctl failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    return ERROR_SUCCESS;
}

void ReactorEPoll::RemoveEvent(SOCKET s)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: socket(%d)\n", FULL_FUNC_NAME, s));

    if ( epoll_ctl(_hEPoll, EPOLL_CTL_DEL, s, NULL) )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: epoll_ctl failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
    }
}

DWORD ReactorEPoll::SendMuxMsg(DWORD dwMuxId, PEPOLL_REC pEPRec)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %#.8x\n", FULL_FUNC_NAME, dwMuxId));

    DO_IF_DBG(InterlockedIncrement64(&_stats.llMsgCallCnt));

    if ( pEPRec->bMsgSend )
    {
        return ERROR_SUCCESS;
    }

    DO_IF_DBG(InterlockedIncrement64(&_stats.llMsgSendCnt));

    pEPRec->bMsgSend = true;
    // see misc_pipe.txt
    dwRet = PipeSendData(_hPipeSend, (PBYTE)&dwMuxId, sizeof(dwMuxId));
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: PipeSendData failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
    }

    return dwRet;
}

void ReactorEPoll::ProcessMuxMsg(DWORD dwMuxId)
{
    UINT  nMuxIdx;
    WORD  wMagic;
    PEPOLL_REC  pEPRec;
    SocketConnection  *pSocketConn;
#ifdef CONFIG_EPOLL_EDGETRIGGER
    bool  bEAgain, bRecvYield, bSendYield;
#else
    bool  bUpdated;
#endif

    nMuxIdx = WIDX_FROM_DWID(dwMuxId);
    wMagic = WMAGICNUM_FROM_DWID(dwMuxId);

    DBG(D_VERB, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nMuxIdx, wMagic));

    assert(nMuxIdx > 0);
    assert(nMuxIdx < _MuxList.size());

    pEPRec = &_MuxList.at(nMuxIdx);
    if ( (!pEPRec->bValid) || (wMagic != pEPRec->wMagic) )
    {
        // see issues_epoll2.txt
        return;
    }

    pSocketConn = pEPRec->pSocketConn;

    pEPRec->bMsgSend = false;
    // see issues_mb3.txt
    MemoryBarrier();
    if ( pEPRec->CtlReq == EPOLL_REQ_REMOVE )
    {
        if ( !pEPRec->bRemoved )
        {
            RemoveEvent(pSocketConn->_s);
        }
        // see issues_epoll2.txt
        pEPRec->wMagic = ~wMagic;
        DBG(D_INFO, (L"%s: idx(%d): REMOVE -> DONE\n", FULL_FUNC_NAME, nMuxIdx));
        pEPRec->CtlReq = EPOLL_REQ_DONE;
        // Unregister will see "DONE" at the first time
        MemoryBarrier();
        return;
    }

    if ( pEPRec->bRemoved )
    {
        return;
    }

    if ( pSocketConn->_bListener )
    {
        if ( (!pEPRec->bRecvInd) && pSocketConn->ReqRecvInd() )
        {
            pEPRec->bRecvInd = true;
            ModifyEvent(pSocketConn->_s, dwMuxId, true, false, false);
        }
        return;
    }

#ifdef CONFIG_EPOLL_EDGETRIGGER
    bRecvYield = false;
    if ( pSocketConn->ReqRecvInd() )
    {
        if ( pEPRec->bRecvReady )
        {
            pSocketConn->OnRecvReady(&bEAgain, &bRecvYield);
            pEPRec->bRecvInd = bEAgain || bRecvYield;
            pEPRec->bRecvReady = !bEAgain;
        }
        else
        {
            pEPRec->bRecvInd = true;
        }
    }

    bSendYield = false;
    if ( pSocketConn->ReqSendInd() )
    {
        if ( pEPRec->bSendReady )
        {
            pSocketConn->OnSendReady(&bEAgain, &bSendYield);
            pEPRec->bSendInd = bEAgain || bSendYield;
            pEPRec->bSendReady = !bEAgain;
        }
        else
        {
            pEPRec->bSendInd = true;
        }
    }

    if ( bRecvYield || bSendYield )
    {
        DBG(D_VERB, (L"%s: yield\n", FULL_FUNC_NAME));
        DO_IF_DBG(++_stats.llYieldCnt);
        _YieldList.PushAtBack(dwMuxId);
    }
#else
    bUpdated = false;

    if ( (!pEPRec->bRecvInd) && pSocketConn->ReqRecvInd() )
    {
        if ( pEPRec->bRecvReady )
        {
            pEPRec->bRecvReady = false;
            pSocketConn->OnRecvReady();
        }

        if ( pSocketConn->ReqRecvInd() )
        {
            bUpdated = true;
            pEPRec->bRecvInd = true;
        }
    }

    if ( (!pEPRec->bSendInd) && pSocketConn->ReqSendInd() )
    {
        if ( pEPRec->bSendReady )
        {
            pEPRec->bSendReady = false;
            pSocketConn->OnSendReady();
        }

        if ( pSocketConn->ReqSendInd() )
        {
            bUpdated = true;
            pEPRec->bSendInd = true;
        }
    }

    if ( bUpdated )
    {
        ModifyEvent(pSocketConn->_s, dwMuxId, pEPRec->bRecvInd, pEPRec->bSendInd, false);
    }
#endif
}

void ReactorEPoll::ProcessPipeEvent(DWORD dwEvents)
{
    DWORD  dwRet;
    DWORD  dwMuxIds[EPOLL_MSG_NUM];
    UINT  nSize;

    DBG(D_VERB, (L"%s:\n", FULL_FUNC_NAME));

    // may EPOLLRDHUP | EPOLLHUP | EPOLLERR | EPOLLIN
    if ( dwEvents & (EPOLLRDHUP | EPOLLHUP | EPOLLERR) )
    {
        DBG(D_INFO, (L"%s: quit\n", FULL_FUNC_NAME));
        return;
    }

    assert(dwEvents & EPOLLIN);

    nSize = sizeof(dwMuxIds);
    dwRet = PipeRecvData(_hPipeRecv, (PBYTE)dwMuxIds, nSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: PipeRecvData failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return;
    }

    if ( nSize % sizeof(DWORD) )
    {
        DBG(D_ERROR, (L"%s: invalid data size: %d\n", FULL_FUNC_NAME, nSize));
        assert(0);
        return;
    }

    nSize /= sizeof(DWORD);
    for ( UINT i = 0; i < nSize; ++i )
    {
        ProcessMuxMsg(dwMuxIds[i]);
    }
}

void ReactorEPoll::ProcessMuxEvent(DWORD dwMuxId, DWORD dwEvents)
{
    UINT  nMuxIdx;
    WORD  wMagic;
    PEPOLL_REC  pEPRec;
    SocketConnection  *pSocketConn;
    bool  bPollIn, bPollOut;
#ifdef CONFIG_EPOLL_EDGETRIGGER
    bool  bSendYield;
#else
    bool  bUpdated;
#endif
    bool  bEAgain, bRecvYield;

    nMuxIdx = WIDX_FROM_DWID(dwMuxId);
    wMagic = WMAGICNUM_FROM_DWID(dwMuxId);

    DBG(D_VERB, (L"%s: idx(%d), magic(%d), events(%#.8x)\n", FULL_FUNC_NAME, nMuxIdx, wMagic, dwEvents));

    assert(nMuxIdx < _MuxList.size());

    pEPRec = &_MuxList.at(nMuxIdx);
    if ( (!pEPRec->bValid) || (wMagic != pEPRec->wMagic) )
    {
        return;
    }

    pSocketConn = pEPRec->pSocketConn;

    if ( dwEvents & (EPOLLRDHUP | EPOLLHUP | EPOLLERR) )
    {
        DBG(D_LOG, (L"%s: hup or err: %#.8x\n", FULL_FUNC_NAME, dwEvents));
        pSocketConn->OnDisconnectedOrError();
        RemoveEvent(pSocketConn->_s);
        pEPRec->bRemoved = true;
        return;
    }

    if ( pSocketConn->_bIgnore )
    {
        pEPRec->bRecvInd = false;
        pEPRec->bSendInd = false;
#ifndef CONFIG_EPOLL_EDGETRIGGER
        ModifyEvent(pSocketConn->_s, dwMuxId, false, false, true);
#endif
        return;
    }

    bPollIn = dwEvents & EPOLLIN;
    bPollOut = dwEvents & EPOLLOUT;

    if ( pSocketConn->_bListener && bPollIn )
    {
        DBG(D_LOG, (L"%s: new connection\n", FULL_FUNC_NAME));
        // special, see issues_epoll_listen.txt
        pSocketConn->OnRecvReady(&bEAgain, &bRecvYield);
        if ( !pSocketConn->ReqRecvInd() )
        {
            pEPRec->bRecvInd = false;
            ModifyEvent(pSocketConn->_s, dwMuxId, false, false, false);
        }
        return;
    }

    if ( pSocketConn->_bAsyncConnect && bPollOut )
    {
        // process error at above
        DBG(D_LOG, (L"%s: async connected\n", FULL_FUNC_NAME));
        pSocketConn->_bAsyncConnect = false;
        pSocketConn->OnAsyncConnected();
#ifdef CONFIG_EPOLL_EDGETRIGGER
        pEPRec->bSendInd = pSocketConn->ReqSendInd();
#endif
    }

#ifdef CONFIG_EPOLL_EDGETRIGGER
    bRecvYield = false;
    // see issues_epoll_eagain4.txt
    pEPRec->bRecvReady = bPollIn;
    if ( bPollIn && pEPRec->bRecvInd )
    {
        pSocketConn->OnRecvReady(&bEAgain, &bRecvYield);
        pEPRec->bRecvInd = bEAgain || bRecvYield;
        pEPRec->bRecvReady = !bEAgain;
    }

    bSendYield = false;
    // see issues_epoll_eagain4.txt
    pEPRec->bSendReady = bPollOut;
    if ( bPollOut && pEPRec->bSendInd )
    {
        pSocketConn->OnSendReady(&bEAgain, &bSendYield);
        pEPRec->bSendInd = bEAgain || bSendYield;
        pEPRec->bSendReady = !bEAgain;
    }

    if ( bRecvYield || bSendYield )
    {
        DBG(D_VERB, (L"%s: yield\n", FULL_FUNC_NAME));
        DO_IF_DBG(++_stats.llYieldCnt);
        _YieldList.PushAtBack(dwMuxId);
    }
#else
    bUpdated = false;

    if ( bPollIn )
    {
        if ( pSocketConn->ReqRecvInd() )
        {
            pSocketConn->OnRecvReady();
        }
        else
        {
            bUpdated = true;
            pEPRec->bRecvInd = false;
            pEPRec->bRecvReady = true;
        }
    }

    if ( bPollOut )
    {
        if ( pSocketConn->ReqSendInd() )
        {
            pSocketConn->OnSendReady();
        }
        else
        {
            bUpdated = true;
            pEPRec->bSendInd = false;
            pEPRec->bSendReady = true;
        }
    }

    if ( bUpdated )
    {
        ModifyEvent(pSocketConn->_s, dwMuxId, pEPRec->bRecvInd, pEPRec->bSendInd, false);
    }
#endif
}

DWORD ReactorEPoll::MuxThread()
{
    DWORD  dwRet;
    DWORD  dwMuxId;
    int  iEvtCnt;
    struct epoll_event  epevts[EPOLL_EVT_NUM];

    DBG(D_LOG, (L"%s: enter\n", FULL_FUNC_NAME));

    while ( !_bQuit )
    {
        DO_IF_DBG(++_stats.llPollCnt);
        iEvtCnt = epoll_wait(_hEPoll, epevts, _countof(epevts), -1);

        if ( iEvtCnt == SOCKET_ERROR )
        {
            dwRet = WSAGetLastError();
            if ( dwRet == EINTR )
            {
                DBG(D_INFO, (L"%s: epoll_wait: ignore EINTR\n", FULL_FUNC_NAME));
                continue;
            }
            else
            {
                DBG(D_ERROR, (L"%s: epoll_wait failed: %d\n", FULL_FUNC_NAME, dwRet));
                assert(0);
                break;
            }
        }

        assert(iEvtCnt > 0);

        for ( int i = 0; i < iEvtCnt; ++i )
        {
            dwMuxId = epevts[i].data.u32;

            if ( dwMuxId )
            {
                ProcessMuxEvent(dwMuxId, epevts[i].events);
            }
            else
            {
                ProcessPipeEvent(epevts[i].events);
            }
        }

        while ( !_YieldList.IsEmpty() )
        {
            dwMuxId = _YieldList.PopAtFront();
            ProcessMuxMsg(dwMuxId);
        }
    }

    DBG(D_LOG, (L"%s: leave\n", FULL_FUNC_NAME));

    return ERROR_SUCCESS;
}

void *ReactorEPoll::MuxThreadProc(void *arg)
{
    ReactorEPoll  *pThis = static_cast<ReactorEPoll*>(arg);

    pThis->MuxThread();

    return NULL;
}

#endif
