/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "SignalPipe.h"
#include "SocketConnection.h"
#include "ReactorSelect.h"


#ifdef CONFIG_REACTOR_SELECT

ReactorSelect *ReactorSelect::_instance = NULL;

ReactorSelect *ReactorSelect::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new ReactorSelect();
    }
    return _instance;
}

void ReactorSelect::Destroy()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

ReactorSelect::ReactorSelect()
    : _hPipeSend(INVALID_SOCKET)
    , _hPipeRecv(INVALID_SOCKET)
    , _bMuxQuit(false)
    , _bPolling(false)
    , _nPollCnt(0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    SET_PTH_INVALID(_hMuxThread);

    pthread_mutex_init(&_hMuxLock, NULL);
}

ReactorSelect::~ReactorSelect()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(!IS_PTH_VALID(_hMuxThread));
    assert(_MuxList.empty());

    pthread_mutex_destroy(&_hMuxLock);
}

DWORD ReactorSelect::Load(UINT nMaxSelectNum)
{
    DWORD  dwRet;

    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    DO_IF_DBG(memset(&_stats, 0, sizeof(_stats)));

    dwRet = CreateSignalPipe(&_hPipeSend, &_hPipeRecv);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: create signal pipe failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    _bMuxQuit = false;
    if ( pthread_create(&_hMuxThread, NULL, MuxThreadProc, this) )
    {
        dwRet = GetLastError();
        DBG(D_ERROR, (L"%s: create mux thread failed: %d\n", FULL_FUNC_NAME, dwRet));
        SET_PTH_INVALID(_hMuxThread);
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    CloseSignalPipe(_hPipeSend);
    _hPipeSend = INVALID_SOCKET;
    CloseSignalPipe(_hPipeRecv);
    _hPipeRecv = INVALID_SOCKET;

err1:
    return dwRet;
}

void ReactorSelect::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    assert(_MuxList.empty());

    if ( IS_PTH_VALID(_hMuxThread) )
    {
        _bMuxQuit = true;
        PipeSendSignal(_hPipeSend);
        pthread_join(_hMuxThread, NULL);
        SET_PTH_INVALID(_hMuxThread);
    }

    if ( _hPipeSend != INVALID_SOCKET )
    {
        CloseSignalPipe(_hPipeSend);
        _hPipeSend = INVALID_SOCKET;
    }

    if ( _hPipeRecv != INVALID_SOCKET )
    {
        CloseSignalPipe(_hPipeRecv);
        _hPipeRecv = INVALID_SOCKET;
    }

#ifndef NDEBUG
    DBG(D_LOG, (L"\n***** select statistics data *****\n"));
    DBG(D_LOG, (L"poll count: %lld\n", _stats.llPollCnt));
    DBG(D_LOG, (L"msg call count: %lld\n", _stats.llMsgCallCnt));
    DBG(D_LOG, (L"msg send count: %lld\n", _stats.llMsgSendCnt));
    DBG(D_LOG, (L"***** select statistics data *****\n\n"));
#endif
}

DWORD ReactorSelect::Register(SocketConnection *pSocketConn)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, (int)pSocketConn->_s));

    pthread_mutex_lock(&_hMuxLock);
    if ( _MuxList.size() < REACTOR_SELECT_LIMIT )
    {
        pSocketConn->_bPollRecv = false;
        pSocketConn->_bPollSend = false;
        _MuxList.push_back(pSocketConn);
        dwRet = ERROR_SUCCESS;
    }
    else
    {
        DBG(D_WARN, (L"%s: too many sockets\n", FULL_FUNC_NAME));
        dwRet = ERROR_NOT_SUPPORTED;
    }
    pthread_mutex_unlock(&_hMuxLock);

    if ( dwRet == ERROR_SUCCESS )
    {
        PipeSendSignal(_hPipeSend);
    }

    return dwRet;
}

void ReactorSelect::Unregister(SocketConnection *pSocketConn)
{
    UINT  nCurCnt;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_lock(&_hMuxLock);
    _MuxList.remove(pSocketConn);
    pthread_mutex_unlock(&_hMuxLock);

    nCurCnt = _nPollCnt;
    PipeSendSignal(_hPipeSend);
    Sleep(10);
    while ( nCurCnt == _nPollCnt )
    {
        DBG(D_INFO, (L"%s: wait select return\n", FULL_FUNC_NAME));
        Sleep(100);
    }
}

void ReactorSelect::ResumePolling()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    DO_IF_DBG(InterlockedIncrement64(&_stats.llMsgCallCnt));

    if ( _bPolling )
    {
        DO_IF_DBG(InterlockedIncrement64(&_stats.llMsgSendCnt));
        // see misc_pipe.txt
        PipeSendSignal(_hPipeSend);
    }
}

DWORD ReactorSelect::MuxThread()
{
    DWORD  dwRet;
    int  ret;
    int  nfds;
    SOCKET  s;
    fd_set  rd_fds, wr_fds;
#ifdef _WIN32_
    fd_set  ex_fds;
#endif
    SocketConnection  *pSocketConn;
    list<SocketConnection*>::const_iterator  itr;
    list<SocketConnection*>::const_iterator  end_itr;

    DBG(D_LOG, (L"%s: enter\n", FULL_FUNC_NAME));

    while ( !_bMuxQuit )
    {
        FD_ZERO(&rd_fds);
        FD_ZERO(&wr_fds);
#ifdef _WIN32_
        FD_ZERO(&ex_fds);
#endif
        nfds = (int)_hPipeRecv;
        FD_SET(_hPipeRecv, &rd_fds);

        pthread_mutex_lock(&_hMuxLock);

        _bPolling = true;
        ++_nPollCnt;
        // ResumePolling check "_bPolling" right
        // Unregister will see "++_nPollCnt" at the first time
        MemoryBarrier();

        end_itr = _MuxList.end();
        for ( itr = _MuxList.begin(); itr != end_itr; ++itr )
        {
            pSocketConn = *itr;
            s = pSocketConn->_s;

            if ( pSocketConn->_bIgnore )
            {
                continue;
            }

            if ( pSocketConn->ReqRecvInd() )
            {
                pSocketConn->_bPollRecv = true;
                FD_SET(s, &rd_fds);
                if ( (int)s > nfds ) 
                {
                    nfds = (int)s;
                }
            }

            if ( pSocketConn->ReqSendInd() || pSocketConn->_bAsyncConnect )
            {
                pSocketConn->_bPollSend = true;
                FD_SET(s, &wr_fds);
                if ( (int)s > nfds ) 
                {
                    nfds = (int)s;
                }
            }

#ifdef _WIN32_
            if ( pSocketConn->_bAsyncConnect )
            {
                FD_SET(s, &ex_fds);
                if ( (int)s > nfds ) 
                {
                    nfds = (int)s;
                }
            }
#endif
        }

        pthread_mutex_unlock(&_hMuxLock);

        DO_IF_DBG(++_stats.llPollCnt);
#ifdef _WIN32_
        ret = select(nfds + 1, &rd_fds, &wr_fds, &ex_fds, NULL);
#else
        ret = select(nfds + 1, &rd_fds, &wr_fds, NULL, NULL);
#endif

        _bPolling = false;
        // ResumePolling will see "false" at the first time
        MemoryBarrier();

        if ( ret == SOCKET_ERROR )
        {
            dwRet = WSAGetLastError();
            DBG(D_ERROR, (L"%s: select failed: %d\n", FULL_FUNC_NAME, dwRet));
            assert(0);
            break;
        }

        assert(ret > 0);

        if ( FD_ISSET(_hPipeRecv, &rd_fds) )
        {
            PipeRecvSignal(_hPipeRecv);
            DBG(D_VERB, (L"%s: recv signal\n", FULL_FUNC_NAME));
        }

        pthread_mutex_lock(&_hMuxLock);
        end_itr = _MuxList.end();
        for ( itr = _MuxList.begin(); itr != end_itr; ++itr )
        {
            pSocketConn = *itr;
            s = pSocketConn->_s;

            if ( pSocketConn->_bIgnore )
            {
                continue;
            }

            if ( pSocketConn->_bPollSend )
            {
                pSocketConn->_bPollSend = false;
                if ( FD_ISSET(s, &wr_fds) )
                {
                    if ( pSocketConn->_bAsyncConnect )
                    {
                        pSocketConn->_bAsyncConnect = false;
#ifdef _WIN32_                    
                        // If a socket is processing a connect call (nonblocking),
                        // a socket is writeable if the connection establishment successfully completes.
                        pSocketConn->OnAsyncConnected();
#else
                        // After select(2) indicates writability,
                        // use getsockopt(2) to read the SO_ERROR option at level SOL_SOCKET to determine whether
                        // connect() completed successfully (SO_ERROR is zero) or unsuccessfully.
                        int  err;
                        socklen_t  len = sizeof(err);
                        getsockopt(s, SOL_SOCKET, SO_ERROR, &err, &len);
                        DBG(D_LOG, (L"%s: AsyncConnect: %d\n", FULL_FUNC_NAME, err));
                        err ? pSocketConn->OnDisconnectedOrError() : pSocketConn->OnAsyncConnected();
#endif
                    }
                    else
                    {
                        // If the socket is not processing a connect call,
                        // writability means a send/sendto are guaranteed to succeed.
                        pSocketConn->OnSendReady();
                    }
                }
            }

            // because under ubuntu, socket become readable also when async connect failed
            // so if we do recv first, the above code getsockopt()'s err will always be 0
            if ( pSocketConn->_bPollRecv )
            {
                pSocketConn->_bPollRecv = false;
                if ( FD_ISSET(s, &rd_fds) )
                {
                    pSocketConn->OnRecvReady();
                }
            }

#ifdef _WIN32_
            if ( pSocketConn->_bAsyncConnect )
            {
                if ( FD_ISSET(s, &ex_fds) )
                {
                    pSocketConn->_bAsyncConnect = false;
                    pSocketConn->OnDisconnectedOrError();
                }
            }
#endif
        }
        pthread_mutex_unlock(&_hMuxLock);
    }

    DBG(D_LOG, (L"%s: leave\n", FULL_FUNC_NAME));

    return ERROR_SUCCESS;
}

void *ReactorSelect::MuxThreadProc(void *arg)
{
    ReactorSelect  *pThis = static_cast<ReactorSelect*>(arg);

    pThis->MuxThread();

    return NULL;
}

#endif
