/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include <openssl/err.h>
#include "SecureCenter.h"
#include "SecureFilterAgent.h"


SecureFilterAgent::SecureFilterAgent(Connection *pConnection, bool bServer)
    : FilterProtocol(pConnection)
    , _bServer(bServer)
    , _ctx(NULL)
    , _ssl(NULL)
    , _BIORecv(NULL)
    , _BIOSend(NULL)
    , _bSecureReady(false)
    , _nRecvBufCnt(0)
    , _nSendItemCnt(0)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_init(&_hStateLock, NULL);
}

SecureFilterAgent::~SecureFilterAgent()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_destroy(&_hStateLock);
}

DWORD SecureFilterAgent::Attach(ProtocolHandler *pHandler, UINT nRecvBufCnt, UINT nSendItemCnt)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    dwRet = InitSSLEnv();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: InitSSLEnv failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    _bSecureReady = false;

    _nRecvBufCnt = nRecvBufCnt;
    _nSendItemCnt = nSendItemCnt;

    _biSendBuf.Alloc(TLS_BUF_SIZE);
    _biRecvBuf.Alloc(TLS_BUF_SIZE);

    // '0' means no send packet
    dwRet = FilterProtocol::Attach(pHandler, 1, 0);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: FilterProtocol::Attach failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    FiniSSLEnv();

err1:
    return dwRet;
}

void SecureFilterAgent::Detach()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    FilterProtocol::Detach();

    FiniSSLEnv();
}

void SecureFilterAgent::ResumeRecv(UINT nBufCnt)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nBufCnt));

    if ( _bSecureReady )
    {
        _pConnection->ResumeRecv(nBufCnt);
    }
    else
    {
        pthread_mutex_lock(&_hStateLock);
        if ( _bSecureReady )
        {
            _pConnection->ResumeRecv(nBufCnt);
        }
        else
        {
            _nRecvBufCnt += nBufCnt;
        }        
        pthread_mutex_unlock(&_hStateLock);
    }
}

void SecureFilterAgent::ResumeSend(UINT nItemCnt)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nItemCnt));

    if ( _bSecureReady )
    {
        _pConnection->ResumeSend(nItemCnt);
    }
    else
    {
        pthread_mutex_lock(&_hStateLock);
        if ( _bSecureReady )
        {
	        _pConnection->ResumeSend(nItemCnt);
        }
        else
        {
            _nSendItemCnt += nItemCnt;
        }        
        pthread_mutex_unlock(&_hStateLock);
    }
}

DWORD SecureFilterAgent::Recv(BufferItem *pBufItem)
{
    int  nSize;

    DBG(D_FUNC, (L"%s: %d @ %p\n", FULL_FUNC_NAME, pBufItem->tail_size(), pBufItem->tail()));

    // SSL_read()@OnRecvReady is called
    // so can use SSL_pending()
    nSize = SSL_pending(_ssl);
    if ( nSize == 0 )
    {
        DBG(D_WARN, (L"%s: ssl buf is empty\n", FULL_FUNC_NAME));
        assert(0);
        return EWOULDBLOCK;
    }

    while ( nSize > 0 )
    {
        if ( pBufItem->tail_size() < (UINT)nSize )
        {
            DBG(D_PERF, (L"%s: expand to %d\n", FULL_FUNC_NAME, (int)nSize));
            pBufItem->Expand((UINT)nSize);
        }

        nSize = SSL_read(_ssl, pBufItem->tail(), pBufItem->tail_size());
        if ( nSize <= 0 )
        {
            DBG(D_WARN, (L"%s: SSL_read failed\n", FULL_FUNC_NAME));
            return ERROR_ACCESS_DENIED;
        }
        pBufItem->Put((UINT)nSize);

        // SSL_set_read_ahead(_ssl, 1) is set
        // so no need to call SSL_read()
        // but no use?
        SSL_read(_ssl, NULL, 0);        
        nSize = SSL_pending(_ssl);
    }

    return ERROR_SUCCESS;
}

DWORD SecureFilterAgent::Send(BufferItem *pBufItem)
{
    int  nSize;

    DBG(D_FUNC, (L"%s: %d @ %p\n", FULL_FUNC_NAME, pBufItem->data_size(), pBufItem->data()));

    if ( _biSendBuf.data_size() > 0 )
    {
        DBG(D_WARN, (L"%s: send buf not empty\n", FULL_FUNC_NAME));
        assert(0);
        return EWOULDBLOCK;
    }

    nSize = SSL_write(_ssl, pBufItem->data(), pBufItem->data_size());
    if ( nSize <= 0 )
    {
        DBG(D_WARN, (L"%s: SSL_write failed\n", FULL_FUNC_NAME));
        return ERROR_ACCESS_DENIED;
    }
	pBufItem->Pull((UINT)nSize);

    ReadBIOSend();

    return ERROR_SUCCESS;
}

void SecureFilterAgent::OnAsyncConnected(Connection *pConnection)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    // if run before FilterProtocol::Attach()@SecureFilterAgent::Attach(),
    // client will ResumeSend(1)@ContinueHandShake,
    // SocketConnection::Attach() will reset _SendCounter.
    // if run after FilterProtocol::Attach()@SecureFilterAgent::Attach(),
    // server's ContinueHandShake@OnRecvReady() will conflict.
    // OnAsyncConnected() is sync with OnRecvReady(),
    // so call ContinueHandShake() here.
    // OnAsyncConnected() will not be called on server side
    // ContinueHandShake() will be called at OnRecvReady().
    ContinueHandShake();

    _pHandler->OnAsyncConnected(this);
}

UINT SecureFilterAgent::OnRecvReady(Connection *pConnection)
{
    DWORD  dwRet;
    int  nSize;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(pConnection == _pConnection);

    assert(_biRecvBuf.data_size() == 0);

    dwRet = _pConnection->Recv(&_biRecvBuf);
    if ( (dwRet == EAGAIN) || (dwRet == EWOULDBLOCK) )
    {
        DBG(D_VERB, (L"%s: EAGAIN\n", FULL_FUNC_NAME));
        return AGAIN;
    }

    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: recv failed: %d\n", FULL_FUNC_NAME, dwRet));
        _pHandler->OnDisconnectedOrError(this);
        return ABORT;
    }

    // write to BIO_s_mem always successfully
    nSize = BIO_write(_BIORecv, _biRecvBuf.data(), _biRecvBuf.data_size());
    assert((UINT)nSize == _biRecvBuf.data_size());
    _biRecvBuf.Reset();

    if ( !_bSecureReady )
    {
        DBG(D_INFO, (L"%s: recv %d bytes\n", FULL_FUNC_NAME, nSize));
        ContinueHandShake();
        return 0;
    }

    // data in _BIORecv is like in socket buf
    // SSL_pending return ssl data size
    // so call SSL_read to start read from bio
    SSL_read(_ssl, NULL, 0);
    if ( SSL_pending(_ssl) == 0 )
    {
        return 0;
    }

    return _pHandler->OnRecvReady(this);
}

UINT SecureFilterAgent::OnSendReady(Connection *pConnection)
{
    DWORD  dwRet;
    UINT  nCount;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(pConnection == _pConnection);

    if ( _bSecureReady && (_biSendBuf.data_size() == 0) )
    {
        nCount = _pHandler->OnSendReady(this);
        if ( nCount != 1 )
        {
            DBG(D_INFO, (L"%s: _pHandler->OnSendReady return %d\n", FULL_FUNC_NAME, nCount));
            // may ABORT (session end)
            assert((nCount != 0) && (nCount != AGAIN));
            return ABORT;
        }
    }

    assert(_biSendBuf.data_size() > 0);

    dwRet = _pConnection->Send(&_biSendBuf);
    if ( (dwRet == EAGAIN) || (dwRet == EWOULDBLOCK) )
    {
        DBG(D_VERB, (L"%s: EAGAIN\n", FULL_FUNC_NAME));
        return AGAIN;
    }

    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: send failed: %d\n", FULL_FUNC_NAME, dwRet));
        _pHandler->OnDisconnectedOrError(this);
        return ABORT;
    }

    if ( _biSendBuf.data_size() > 0 )
    {
        DBG(D_PERF, (L"%s: partial data\n", FULL_FUNC_NAME));
        return 0;
    }

    _biSendBuf.Reset();

    if ( _bSecureReady )
    {
        return 1;
    }

    ContinueHandShake();

    return (_bSecureReady ? 0 : 1);
}

DWORD SecureFilterAgent::InitSSLEnv()
{
    DWORD  dwRet;
    const SSL_METHOD  *method;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    method = _bServer ? TLSv1_2_server_method() : TLSv1_2_client_method();
    _ctx = SSL_CTX_new(method);
    if ( _ctx == NULL )
    {
        DBG(D_ERROR, (L"%s: SSL_CTX_new failed: %lu\n", FULL_FUNC_NAME, ERR_get_error()));
        goto err1;
    }

    dwRet = SecureCenter::Instance()->InstallRootCert(_ctx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: InstallRootCert failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

#ifndef CONDIF_TLS_MUTUAL
    if ( _bServer )
#endif
    {
        dwRet = SecureCenter::Instance()->InstallCertAndKey(_ctx);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: InstallCertAndKey failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto err2;
        }
    }

#ifndef CONDIF_TLS_MUTUAL
    if ( !_bServer )
#endif
    {
	    SSL_CTX_set_verify(_ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);
    }

    _BIORecv = BIO_new(BIO_s_mem());
    if ( _BIORecv == NULL )
    {
        DBG(D_ERROR, (L"%s: BIO_new for recv failed: %lu\n", FULL_FUNC_NAME, ERR_get_error()));
        goto err2;
    }

    _BIOSend = BIO_new(BIO_s_mem());
    if ( _BIOSend == NULL )
    {
        DBG(D_ERROR, (L"%s: BIO_new for send failed: %lu\n", FULL_FUNC_NAME, ERR_get_error()));
        goto err3;
    }

    _ssl = SSL_new(_ctx);
    if ( _ssl == NULL )
    {
        DBG(D_ERROR, (L"%s: SSL_new failed: %lu\n", FULL_FUNC_NAME, ERR_get_error()));
        goto err4;
    }

    SSL_set_bio(_ssl, _BIORecv, _BIOSend);

    SSL_set_read_ahead(_ssl, 1);

//    SSL_set_mode(_ssl, SSL_MODE_AUTO_RETRY);

    // prepared for the handshake init
    _bServer ? SSL_set_accept_state(_ssl) : SSL_set_connect_state(_ssl);

    return ERROR_SUCCESS;

err4:
    BIO_free(_BIOSend);
    _BIOSend = NULL;

err3:
    BIO_free(_BIORecv);
    _BIORecv = NULL;


err2:
    SSL_CTX_free(_ctx);
    _ctx = NULL;

err1:
    return ERROR_NOT_READY;
}

void SecureFilterAgent::FiniSSLEnv()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( _ssl != NULL )
    {
        SSL_shutdown(_ssl);
        SSL_free(_ssl);
        _ssl = NULL;

        // SSL_free will free bio automatically
        _BIOSend = NULL;
        _BIORecv = NULL;
    }

    if ( _BIOSend != NULL )
    {
        BIO_free(_BIOSend);
        _BIOSend = NULL;
    }

    if ( _BIORecv != NULL )
    {
        BIO_free(_BIORecv);
        _BIORecv = NULL;
    }

    if ( _ctx != NULL )
    {
        SSL_CTX_free(_ctx);
        _ctx = NULL;
    }
}

void SecureFilterAgent::ContinueHandShake()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( _biSendBuf.data_size() > 0 )
    {
        return;
    }

    DBG(D_INFO, (L"%s: SSL_do_handshake\n", FULL_FUNC_NAME));
    SSL_do_handshake(_ssl);

    if ( ReadBIOSend() > 0 )
    {
        DBG(D_INFO, (L"%s: will send %d bytes\n", FULL_FUNC_NAME, _biSendBuf.data_size()));
        _pConnection->ResumeSend(1);
        return;
    }

    // see issues_hand_1.txt
    // MUST put after SSL_do_handshake()
    // see issues_hand_3.txt
    // SHOULD put after (ReadBIOSend()>0)
    if ( SSL_is_init_finished(_ssl) )
    {
        EnterSecureState();
    }
}

void SecureFilterAgent::EnterSecureState()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_lock(&_hStateLock);

    // see ProxyFilterAgent::EnterBridgeMode
    _pConnection->ResetRecv(_nRecvBufCnt);
    _pConnection->ResetSend(_nSendItemCnt);

    // see ProxyFilterAgent::EnterBridgeMode
    _bSecureReady = true;

    pthread_mutex_unlock(&_hStateLock);
}

int SecureFilterAgent::ReadBIOSend()
{
    int  nSize;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    nSize = (int)BIO_ctrl_pending(_BIOSend);
    if ( nSize > 0 )
    {
        if ( _biSendBuf.tail_size() < (UINT)nSize )
        {
            DBG(D_PERF, (L"%s: expand to %d\n", FULL_FUNC_NAME, (int)nSize));
            _biSendBuf.Expand((UINT)nSize);
        }
    
        nSize = BIO_read(_BIOSend, _biSendBuf.tail(), _biSendBuf.tail_size());
        _biSendBuf.Put((UINT)nSize);
    }

    return nSize;
}
