/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "ProtocolHandler.h"
#include "ReactorSelect.h"
#include "ReactorEPoll.h"
#include "SocketConnection.h"


#ifndef TCP_USER_TIMEOUT
// from /usr/include/linux/tcp.h
#define TCP_USER_TIMEOUT    18  /* How long for loss retry before timeout */
#endif


SocketConnection::SocketConnection(SOCKET s, bool bListener, bool bAsyncConnect)
    : _s(s)
    , _bIgnore(false)
    , _bListener(bListener)
    , _bAsyncConnect(bAsyncConnect)
#ifdef CONFIG_REACTOR_SELECT
    , _bPollRecv(false)
    , _bPollSend(false)
#endif
#ifdef CONFIG_REACTOR_EPOLL
    , _dwMuxId(INVALID_MUX_ID)
#endif
    , _pHandler(NULL)
{
    DBG(D_INFO, (L"%s: socket(%d)\n", FULL_FUNC_NAME, (int)s));

#ifdef CONFIG_REACTOR_EPOLL
    _nYieldNum = bListener ? 1 : CONFIG_EPOLL_ET_YIELD_NUM;
#endif
}

SocketConnection::~SocketConnection()
{
    DBG(D_INFO, (L"%s: socket(%d)\n", FULL_FUNC_NAME, (int)_s));

    assert(_s != INVALID_SOCKET);
    closesocket(_s);
}

bool SocketConnection::ReqRecvInd() const
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return (!_RecvCounter.IsEmpty());
}

bool SocketConnection::ReqSendInd() const
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return (!_SendCounter.IsEmpty());
}

void SocketConnection::OnAsyncConnected()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _pHandler->OnAsyncConnected(this);
}

void SocketConnection::OnDisconnectedOrError()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _bIgnore = true;

    // ProtocolHandler's RecvQSize maybe zero
//    OnRecvReady();
    _pHandler->OnDisconnectedOrError(this);
}

void SocketConnection::OnRecvReady(bool *pbEAgain, bool *pbYield)
{
    UINT  nCount;
#ifdef CONFIG_EPOLL_EDGETRIGGER
    UINT  i;
#endif

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

#ifdef CONFIG_EPOLL_EDGETRIGGER
    assert(pbEAgain && pbYield);

    *pbEAgain = false;

    for ( i = 0; i < _nYieldNum; ++i )
    {
        nCount = _pHandler->OnRecvReady(this);

        if ( nCount == ProtocolHandler::ABORT )
        {
            _bIgnore = true;
            break;
        }

        if ( nCount == ProtocolHandler::AGAIN )
        {
            *pbEAgain = true;
            break;
        }

        if ( nCount > 0 )
        {
            assert(nCount <= _RecvCounter.GetProductNum());
            _RecvCounter.Consume(nCount);
        }

        // see issues_epoll_ret0.txt
        if ( _RecvCounter.IsEmpty() )
        {
            break;
        }
    }

    *pbYield = (i == _nYieldNum);
#else
    nCount = _pHandler->OnRecvReady(this);
    if ( nCount == ProtocolHandler::ABORT )
    {
        _bIgnore = true;
        return;
    }

    assert(nCount != ProtocolHandler::AGAIN);

    if ( nCount > 0 )
    {
        assert(nCount <= _RecvCounter.GetProductNum());
        _RecvCounter.Consume(nCount);
    }
#endif
}

void SocketConnection::OnSendReady(bool *pbEAgain, bool *pbYield)
{
    UINT  nCount;
#ifdef CONFIG_EPOLL_EDGETRIGGER
    UINT  i;
#endif

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

#ifdef CONFIG_EPOLL_EDGETRIGGER
    assert(pbEAgain && pbYield);

    *pbEAgain = false;

    for ( i = 0; i < _nYieldNum; ++i )
    {
        nCount = _pHandler->OnSendReady(this);

        if ( nCount == ProtocolHandler::ABORT )
        {
            _bIgnore = true;
            break;
        }

        if ( nCount == ProtocolHandler::AGAIN )
        {
            *pbEAgain = true;
            break;
        }

        if ( nCount > 0 )
        {
            assert(nCount <= _SendCounter.GetProductNum());
            _SendCounter.Consume(nCount);
        }

        // see issues_epoll_ret0.txt
        if ( _SendCounter.IsEmpty() )
        {
            break;
        }
    }

    *pbYield = (i == _nYieldNum);
#else
    nCount = _pHandler->OnSendReady(this);
    if ( nCount == ProtocolHandler::ABORT )
    {
        _bIgnore = true;
        return;
    }

    if ( nCount == ProtocolHandler::AGAIN )
    {
        // see issues_mux_eagain.txt
        return;
    }

    if ( nCount > 0 )
    {
        assert(nCount <= _SendCounter.GetProductNum());
        _SendCounter.Consume(nCount);
    }
#endif
}

DWORD SocketConnection::Attach(ProtocolHandler *pHandler, UINT nRecvBufCnt, UINT nSendItemCnt)
{
    DBG(D_INFO, (L"%s: recv(%d), send(%d)\n", FULL_FUNC_NAME, nRecvBufCnt, nSendItemCnt));

    assert(_s != INVALID_SOCKET);

    // SO_NOSIGPIPE not defined, use send(xxx, MSG_NOSIGNAL)
#if 0//defined(_LINUX_) || defined(_ANDROID_)
{
    DWORD  dwRet;
    int  val = 1;
    if ( setsockopt(_s, SOL_SOCKET, SO_NOSIGPIPE, (char*)&val, sizeof(val)) == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_WARN, (L"%s: setsockopt failed: %d\n", FULL_FUNC_NAME, dwRet));
    }
}
#endif

    assert(!_pHandler);
    _pHandler = pHandler;

    _bIgnore = false;
    _RecvCounter.Reset(nRecvBufCnt);
    _SendCounter.Reset(nSendItemCnt);

    // choosed by Endpoint
#if 0
    // send will block unless the socket has been placed in nonblocking mode.
    // On nonblocking stream oriented sockets, the number of bytes written can be between 1 and the requested length.
#ifdef _WIN32_
    u_long  iMode = 1;
    ioctlsocket(_s, FIONBIO, &iMode);
#else
    int  nMode = fcntl(_s, F_GETFL, 0);
    fcntl(_s, F_SETFL, (nMode | O_NONBLOCK));
#endif
#endif

#ifdef CONFIG_REACTOR_SELECT
    ReactorSelect::Instance()->Register(this);
#endif
#ifdef CONFIG_REACTOR_EPOLL
    ReactorEPoll::Instance()->Register(this, _dwMuxId);
#endif

    return ERROR_SUCCESS;
}

void SocketConnection::Detach()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

#ifdef CONFIG_REACTOR_SELECT
    ReactorSelect::Instance()->Unregister(this);
#endif
#ifdef CONFIG_REACTOR_EPOLL
    ReactorEPoll::Instance()->Unregister(_dwMuxId);
    _dwMuxId = INVALID_MUX_ID;
#endif

    assert(_pHandler);
    _pHandler = NULL;
}

DWORD SocketConnection::EnableKeepAlive(UINT nTimeout)
{
    DBG(D_INFO, (L"%s: timeout=%ds\n", FULL_FUNC_NAME, nTimeout));

#ifdef _LINUX_
    int  iVal;
    const int  iIntvl = 2;

    iVal = 1;
    setsockopt(_s, SOL_SOCKET, SO_KEEPALIVE, &iVal, sizeof(iVal));

    iVal = nTimeout / 2;
    setsockopt(_s, IPPROTO_TCP, TCP_KEEPIDLE, &iVal, sizeof(iVal));

    iVal /= iIntvl;
    setsockopt(_s, IPPROTO_TCP, TCP_KEEPINTVL, &iVal, sizeof(iVal));

    iVal = iIntvl;
    setsockopt(_s, IPPROTO_TCP, TCP_KEEPCNT, &iVal, sizeof(iVal));

    iVal = nTimeout * 1000;
    setsockopt(_s, IPPROTO_TCP, TCP_USER_TIMEOUT, &iVal, sizeof(iVal));
#else
    DWORD  dwBytes;
    struct tcp_keepalive  alive;

    alive.onoff = 1;
    alive.keepalivetime = nTimeout * 1000;
    alive.keepaliveinterval = 1 * 1000;
    WSAIoctl(_s, SIO_KEEPALIVE_VALS, &alive, sizeof(alive), NULL, 0, &dwBytes, NULL, NULL);
#endif

    return ERROR_SUCCESS;
}

void SocketConnection::ResetRecv(UINT nBufCnt)
{
    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, nBufCnt));

    _RecvCounter.Reset(0);
    ResumeRecv(nBufCnt);
}

void SocketConnection::ResetSend(UINT nItemCnt)
{
    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, nItemCnt));

    _SendCounter.Reset(0);
    ResumeSend(nItemCnt);
}

void SocketConnection::ResumeRecv(UINT nBufCnt)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nBufCnt));

    if ( _RecvCounter.Produce(nBufCnt) )
    {
#ifdef CONFIG_REACTOR_SELECT
        ReactorSelect::Instance()->ResumePolling();
#endif
#ifdef CONFIG_REACTOR_EPOLL
        ReactorEPoll::Instance()->ResumePolling(_dwMuxId);
#endif
    }
}

void SocketConnection::ResumeSend(UINT nItemCnt)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nItemCnt));

    if ( _SendCounter.Produce(nItemCnt) )
    {
#ifdef CONFIG_REACTOR_SELECT
        ReactorSelect::Instance()->ResumePolling();
#endif
#ifdef CONFIG_REACTOR_EPOLL
        ReactorEPoll::Instance()->ResumePolling(_dwMuxId);
#endif
    }
}

DWORD SocketConnection::GetLocalAddr(PSOCK_ADDR pAddr)
{
    socklen_t  len;
    struct sockaddr_in  addr;
#ifdef _LINUX_
    int  type;
#else
    DWORD  type;
#endif

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    len = sizeof(addr);
    getsockname(_s, (struct sockaddr*)&addr, &len);

    pAddr->bType = ADDR_TYPE_IPV4;
    pAddr->IPv4Addr.dwIP = addr.sin_addr.s_addr;
    pAddr->IPv4Addr.wPort = addr.sin_port;

    len = sizeof(type);
    getsockopt(_s, SOL_SOCKET, SO_TYPE, (char*)&type, &len);
    pAddr->IPv4Addr.bStream = (type == SOCK_STREAM);

    return ERROR_SUCCESS;
}

DWORD SocketConnection::GetPeerAddr(PSOCK_ADDR pAddr)
{
    struct sockaddr_in  addr;
    socklen_t  addrlen;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    addrlen = sizeof(addr);
    getpeername(_s, (struct sockaddr*)&addr, &addrlen);

    pAddr->bType = ADDR_TYPE_IPV4;
    pAddr->IPv4Addr.dwIP = addr.sin_addr.s_addr;
    pAddr->IPv4Addr.wPort = addr.sin_port;

    return ERROR_SUCCESS;
}

DWORD SocketConnection::SetRecvBufSize(DWORD  dwSize)
{
    DWORD  dwRet;
    int  size;

    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, dwSize));

    size = (int)dwSize;
    if ( setsockopt(_s, SOL_SOCKET, SO_RCVBUFFORCE, (char*)&size, sizeof(size)) == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_WARN, (L"%s: setsockopt failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    return ERROR_SUCCESS;
}

DWORD SocketConnection::SetSendBufSize(DWORD dwSize)
{
    DWORD  dwRet;
    int  size;

    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, dwSize));

    size = (int)dwSize;
    if ( setsockopt(_s, SOL_SOCKET, SO_SNDBUFFORCE, (char*)&size, sizeof(size)) == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_WARN, (L"%s: setsockopt failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    return ERROR_SUCCESS;
}

DWORD SocketConnection::PollRecv(DWORD dwMilliseconds)
{
    DWORD  dwRet;
    int  ret;
    fd_set  rd_fds;
    timeval  tv;

    DBG(D_FUNC, (L"%s: %d\n", FULL_FUNC_NAME, dwMilliseconds));

    FD_ZERO(&rd_fds);
    FD_SET(_s, &rd_fds);
    tv.tv_sec = dwMilliseconds / 1000;
    tv.tv_usec = (dwMilliseconds % 1000) * 1000;
    ret = select((int)(_s + 1), &rd_fds, NULL, NULL, &tv);
    if ( ret == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_WARN, (L"%s: select failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    if ( ret == 0 )
    {
        return WAIT_TIMEOUT;
    }

    return ERROR_SUCCESS;
}

DWORD SocketConnection::PollSend(DWORD dwMilliseconds)
{
    DWORD  dwRet;
    int  ret;
    fd_set  wr_fds;
    timeval  tv;

    DBG(D_FUNC, (L"%s: %d\n", FULL_FUNC_NAME, dwMilliseconds));

    FD_ZERO(&wr_fds);
    FD_SET(_s, &wr_fds);
    tv.tv_sec = dwMilliseconds / 1000;
    tv.tv_usec = (dwMilliseconds % 1000) * 1000;
    ret = select((int)(_s + 1), NULL, &wr_fds, NULL, &tv);
    if ( ret == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_WARN, (L"%s: select failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    if ( ret == 0 )
    {
        return WAIT_TIMEOUT;
    }

    return ERROR_SUCCESS;
}

DWORD SocketConnection::Recv(BufferItem *pBufItem)
{
    DWORD  dwRet;
    int  len;

    DBG(D_FUNC, (L"%s: %d @ %p\n", FULL_FUNC_NAME, pBufItem->tail_size(), pBufItem->tail()));

    len = recv(_s, (char*)pBufItem->tail(), pBufItem->tail_size(), 0);
    if ( len <= 0 )
    {
        dwRet = (len < 0) ? WSAGetLastError() : ERROR_NOT_CONNECTED;
        DBG(D_VERB, (L"%s: recv failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    pBufItem->Put(len);

    return ERROR_SUCCESS;
}

DWORD SocketConnection::Send(BufferItem *pBufItem)
{
    DWORD  dwRet;
    int  len;

    DBG(D_FUNC, (L"%s: %d @ %p\n", FULL_FUNC_NAME, pBufItem->data_size(), pBufItem->data()));

    len = send(_s, (char*)pBufItem->data(), pBufItem->data_size(), MSG_NOSIGNAL);
    if ( len == SOCKET_ERROR )
    {
        dwRet = WSAGetLastError();
        DBG(D_VERB, (L"%s: send failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    if ( (UINT)len < pBufItem->data_size() )
    {
//        DBG(D_PERF, (L"%s: partial: %d of %d\n", FULL_FUNC_NAME, len, pBufItem->data_size()));
    }

    pBufItem->Pull(len);

    return ERROR_SUCCESS;
}
