/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "LP2PDownloadHandler.h"


LP2PDownloadHandler::LP2PDownloadHandler(SeedTarget *pSeedTarget)
    : EpPDUHandler(LP2P_UPDOWN_DATAQSIZE, LP2P_UPDOWN_DATABSIZE*3/2,
                   LP2P_UPDOWN_CTRLQSIZE, LP2P_UPDOWN_CTRLBSIZE,
                   LP2P_UPDOWN_EVTQSIZE, LP2P_UPDOWN_EVTBSIZE)
    , _pSeedTarget(pSeedTarget)
    , _State(DH_STATE_NONE)
    , _dwSessionId(INVALID_ST_SESS_ID)
    , _nBlkNum(0)
    , _nBlkIdx(0)
    , _dwBlkId(INVALID_ST_BLK_ID)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _pLocalBitMask = pSeedTarget->GetProgressMask();

    pSeedTarget->AddRef();
}

LP2PDownloadHandler::~LP2PDownloadHandler()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_ST_BLK_ID);
    assert(_dwSessionId == INVALID_ST_SESS_ID);

    _pSeedTarget->Release();
}

DWORD LP2PDownloadHandler::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwSessionId == INVALID_ST_SESS_ID);
    dwRet = _pSeedTarget->BeginSession(false, _dwSessionId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: begin session for write failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    _nBlkNum = 0;
    _nBlkIdx = RAND_BETWEEN(0, _pSeedTarget->GetBlockCount());
    DBG(D_INFO, (L"%s: will fetch from block[#%d]\n", FULL_FUNC_NAME, _nBlkIdx));

#ifdef CONFIG_ST_BLK_COMPACT
    dwRet = _DecompAlg.Init();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init decompressor failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    DBG(D_INFO, (L"%s: alloc bytes[%d] for write buf\n", FULL_FUNC_NAME, COMPACT_CLUSTER_SIZE));
    _DecompBuf.Alloc(COMPACT_CLUSTER_SIZE);
#endif

    pConnection->SetRecvBufSize(LP2P_SOCKBUF_SIZE);

    dwRet = EpPDUHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err3;
    }

    return ERROR_SUCCESS;

err3:
#ifdef CONFIG_ST_BLK_COMPACT
    _DecompAlg.Fini();

err2:
#endif
    _pSeedTarget->EndSession(_dwSessionId);
    _dwSessionId = INVALID_ST_SESS_ID;

err1:
    return dwRet;
}

void LP2PDownloadHandler::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::Unbind();

#ifdef CONFIG_ST_BLK_COMPACT
    _DecompAlg.Fini();
#endif

    _PeerBitMask.Fini();

    if ( _dwSessionId != INVALID_ST_SESS_ID )
    {
        if ( _dwBlkId != INVALID_ST_BLK_ID )
        {
            _pSeedTarget->CloseBlock(_dwBlkId);
            _dwBlkId = INVALID_ST_BLK_ID;
        }

        _pSeedTarget->EndSession(_dwSessionId);
        _dwSessionId = INVALID_ST_SESS_ID;
    }

    _State = DH_STATE_NONE;
}

void LP2PDownloadHandler::OnAsyncConnected(Connection *pConnection)
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::OnAsyncConnected(pConnection);

    OpenSeedTargetReq();
}

void LP2PDownloadHandler::PrepareSend()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

SOCK_ADDR LP2PDownloadHandler::GetPeerAddr() const
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return _PeerAddr;
}

void LP2PDownloadHandler::SetPeerAddr(const SOCK_ADDR *pAddr)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    _PeerAddr = *pAddr;
}

void LP2PDownloadHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    assert(!bReq);

    switch ( bCode )
    {
        case C_SEEDTARGET_OPEN:
            assert(!pbiBody->data_size());
            dwRet = OnOpenSeedTargetRep(dwParam);
            break;

        case C_SEEDTARGET_GET_BITMASK:
            dwRet = OnGetSeedTargetBitmaskRep(dwParam, pbiBody);
            break;

        default:
            dwRet = ERROR_NOT_SUPPORTED;
            DBG(D_WARN, (L"%s: unknown code(%d)\n", FULL_FUNC_NAME, bCode));
            assert(0);
            break;
    }

    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: process(bCode=%d) failed: %d\n", FULL_FUNC_NAME, bCode, dwRet));
        ReqEndSession();
    }
}

void LP2PDownloadHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    bool  bLastOne, bEndOfBlk;
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert((bCode == C_SEEDTARGET_READ_BLOCK) || (bCode == C_SEEDTARGET_BLOCK_END));
    assert(!bReq);

    if ( _State == DH_STATE_ERROR )
    {
        return;
    }

    assert(_State == DH_STATE_REQ_READ);
    assert(_dwBlkId != INVALID_ST_BLK_ID);

    if ( dwParam != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: C_SEEDTARGET_READ_BLOCK failed: %d\n", FULL_FUNC_NAME, dwParam));
        goto out;
    }

    bLastOne = (bCode == C_SEEDTARGET_BLOCK_END) && (!nLeftBytes);

#ifdef CONFIG_ST_BLK_COMPACT
    dwParam = InflateBlock(pData, nDataBytes, bLastOne, bEndOfBlk);
#else
    dwParam = _pSeedTarget->WriteBlock(_dwBlkId, pData, nDataBytes, bEndOfBlk);
#endif
    if ( dwParam != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: WriteBlock failed: %d\n", FULL_FUNC_NAME, dwParam));

        if ( dwParam == ERROR_DISK_FULL )
        {
            _State = DH_STATE_ERROR;

            pbi = AllocEvent();
            if ( pbi )
            {
                EpBuildMsg(C_LP2P_ERROR_DISK_FULL, 0, 0, pbi);
                PutEvent(pbi);
            }
        }

        goto out;
    }

    assert(bEndOfBlk == bLastOne);

out:
    if ( (dwParam != ERROR_SUCCESS) || bLastOne )
    {
        _pSeedTarget->CloseBlock(_dwBlkId);
        _dwBlkId = INVALID_ST_BLK_ID;

        if ( _State == DH_STATE_REQ_READ )
        {
            _State = DH_STATE_GETMASK_READY;
            ReadSeedTargetBlockReq();
        }
    }
}

void LP2PDownloadHandler::OpenSeedTargetReq()
{
    BufferItem  *pbi;
    SEED_INFO  SeedInfo;
    PREQ_OPEN_SEEDTARGET  pOpenReq;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_State == DH_STATE_NONE);

    _pSeedTarget->GetSeedInfo(&SeedInfo);

    pbi = GetSendBuf();
    assert(pbi);

    pOpenReq = (PREQ_OPEN_SEEDTARGET)pbi->Put(sizeof(REQ_OPEN_SEEDTARGET));
    memcpy(&pOpenReq->SHA1ValSeed, &SeedInfo.sha1, sizeof(SHA1_VAL));

    BuildPDU(C_SEEDTARGET_OPEN, true, true, 0, pbi);

    _State = DH_STATE_REQ_OPEN;

    PutSendData(pbi, true);
}

DWORD LP2PDownloadHandler::OnOpenSeedTargetRep(DWORD dwResult)
{
    DBG(D_VERB, (L"%s: Open(%hs) ret=%d\n", FULL_FUNC_NAME, _pSeedTarget->GetName(), dwResult));

    assert(_State == DH_STATE_REQ_OPEN);

    if ( dwResult != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: Open(%hs) failed: %d\n", FULL_FUNC_NAME, _pSeedTarget->GetName(), dwResult));
        goto err;
    }

    DBG(D_INFO, (L"%s: Open(%hs) successfully\n", FULL_FUNC_NAME, _pSeedTarget->GetName()));

    _State = DH_STATE_OPEN_READY;

    GetSeedTargetBitmaskReq();

    return ERROR_SUCCESS;

err:
    return dwResult;
}

void LP2PDownloadHandler::GetSeedTargetBitmaskReq()
{
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_State == DH_STATE_OPEN_READY);

    pbi = GetSendBuf();
    assert(pbi);

    BuildPDU(C_SEEDTARGET_GET_BITMASK, true, true, 0, pbi);

    _State = DH_STATE_REQ_GETMASK;

    PutSendData(pbi, true);
}

DWORD LP2PDownloadHandler::OnGetSeedTargetBitmaskRep(DWORD dwResult, BufferItem *pbiBody)
{
    PRET_GET_SEEDTARGET_BITMASK  pGetSTBits;

    DBG(D_VERB, (L"%s: GetBitmask(%hs) ret=%d\n", FULL_FUNC_NAME, _pSeedTarget->GetName(), dwResult));

    assert(_State == DH_STATE_REQ_GETMASK);

    if ( dwResult != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: GetBitmask(%hs) failed: %d\n", FULL_FUNC_NAME, _pSeedTarget->GetName(), dwResult));
        goto err;
    }

    DBG(D_INFO, (L"%s: GetBitmask(%hs) successfully\n", FULL_FUNC_NAME, _pSeedTarget->GetName()));

    assert(pbiBody->data_size() > sizeof(RET_GET_SEEDTARGET_BITMASK));
    pGetSTBits = (PRET_GET_SEEDTARGET_BITMASK)pbiBody->data();
    pbiBody->Pull(sizeof(RET_GET_SEEDTARGET_BITMASK));

    if ( pGetSTBits->dwBitsCnt != _pLocalBitMask->GetBitsCount() )
    {
        dwResult = ERROR_INVALID_DATA;
        DBG(D_WARN, (L"%s: bits count not match\n", FULL_FUNC_NAME));
        goto err;
    }

    dwResult = _PeerBitMask.Init(pGetSTBits->dwBitsCnt, pbiBody->data(), pbiBody->data_size(), false);
    if ( dwResult != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init peer bitmask failed: %d\n", FULL_FUNC_NAME, dwResult));
        goto err;
    }

    _State = DH_STATE_GETMASK_READY;

    ReadSeedTargetBlockReq();

    return ERROR_SUCCESS;

err:
    assert(0);
    return dwResult;
}

void LP2PDownloadHandler::ReadSeedTargetBlockReq()
{
    DWORD  dwRet;
    BufferItem  *pbi;
    UINT  nIdx = 0, nBlkCnt;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_State == DH_STATE_GETMASK_READY);
    assert(_dwBlkId == INVALID_ST_BLK_ID);

    for ( nBlkCnt = _pSeedTarget->GetBlockCount(); _nBlkNum < nBlkCnt; )
    {
        nIdx = _nBlkIdx % nBlkCnt;
        DBG(D_VERB, (L"%s: %p: total(%d), processed(%d), now(%d->%d)\n", FULL_FUNC_NAME, this, nBlkCnt, _nBlkNum, _nBlkIdx, nIdx));

        ++_nBlkNum;
        ++_nBlkIdx;

        if ( _pLocalBitMask->IsBitSet(nIdx) || (!_PeerBitMask.IsBitSet(nIdx)) )
        {
            continue;
        }

        dwRet = _pSeedTarget->OpenBlock(_dwSessionId, nIdx, _dwBlkId);
        if ( dwRet == ERROR_SUCCESS )
        {
            break;
        }

        if ( (dwRet != ERROR_BUSY) && (dwRet != ERROR_ALREADY_EXISTS) )
        {
            DBG(D_WARN, (L"%s: OpenBlock(%d) failed: %d\n", FULL_FUNC_NAME, nIdx, dwRet));
            assert(0);
        }
    }

    if ( _dwBlkId != INVALID_ST_BLK_ID )
    {
        DBG(D_VERB, (L"%s: send read req for block[#%d]\n", FULL_FUNC_NAME, nIdx));

        pbi = GetSendBuf();
        assert(pbi);

        BuildPDU(C_SEEDTARGET_READ_BLOCK, true, true, nIdx, pbi);

        _State = DH_STATE_REQ_READ;
        PutSendData(pbi, true);
    }
    else
    {
        DBG(D_INFO, (L"%s: no more block\n", FULL_FUNC_NAME));
        ReqEndSession();
    }
}

#ifdef CONFIG_ST_BLK_COMPACT
DWORD LP2PDownloadHandler::InflateBlock(const PBYTE pData, UINT nDataBytes, bool bLast, bool &rbEndOfBlk)
{
    DWORD  dwRet;
    PBYTE  pSrc;
    UINT  nSrc, nBytes;
    bool  bAgain;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rbEndOfBlk = false;

    pSrc = pData;
    nSrc = nDataBytes;

    bAgain = true;
    while ( bAgain )
    {
        dwRet = _DecompAlg.Process(pSrc, nSrc, &_DecompBuf, bAgain, bLast);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: Decompress(%d @ %p) failed: %d\n", FULL_FUNC_NAME, nSrc, pSrc, dwRet));
            assert(0);
            break;
        }

        nBytes = _DecompBuf.data_size();
#ifdef CONFIG_ST_BLK_WR_BUF
        if ( nBytes )
#else
        // see issues_zp2p_3.txt
        if ( (nBytes == COMPACT_CLUSTER_SIZE) || (bLast && nBytes) )//|| bAgain )
#endif
        {
            // make sure test "bAgain" is unnecessary:
            // "nSrc > 0" && "_DecompBuf.tail() > 0" SHOULD NOT happen
//            assert((nBytes == COMPACT_CLUSTER_SIZE) || (bLast && nBytes));

            // block's data is finished, but endMark is on the way,
            // so bEndOfBlk will set true, but the PDU is not finished.
            // ignore the bEndOfBlk indicated by ST, see down_endmark.log
            dwRet = _pSeedTarget->WriteBlock(_dwBlkId, _DecompBuf.data(), nBytes, rbEndOfBlk);
            if ( dwRet != ERROR_SUCCESS )
            {
                DBG(D_WARN, (L"%s: WriteBlock failed: %d\n", FULL_FUNC_NAME, dwRet));
                break;
            }

            _DecompBuf.Reset();
        }
    }

    return dwRet;
}
#endif
