/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "TimerCenter.h"
#include "SeedAgentHandler.h"
#include "LP2PUploadHandler.h"
#include "LP2PSession.h"
#include "LP2PEndpoint.h"


typedef enum _LPEP_SIGNAL
{
    SIG_SAH_CONNECT = 0,

    SIG_SESSION_REQ,
    SIG_SESSION_FREE,

    LPEP_SIG_NUM
}
LPEP_SIGNAL;


LP2PEndpoint *LP2PEndpoint::_instance = NULL;

LP2PEndpoint *LP2PEndpoint::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new LP2PEndpoint();
    }
    return _instance;
}

void LP2PEndpoint::Destroy()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

LP2PEndpoint::LP2PEndpoint()
    : Endpoint()
    , _pObserver(NULL)
    , _dwTimerId(INVALID_TIMER_ID)
    , _hListener(INVALID_LISTENER_ID)
    , _pSAHandler(NULL)
    , _dwSAHConnId(INVALID_EP_CONN_ID)
    , _pHandlerList(NULL)
    , _pSessionList(NULL)
    , _pSessionEvent(NULL)
    , _STList(0)
    , _STListMate(true)
    , _dwSTMagic(0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _TrackerAddr.bType = ADDR_TYPE_IPV4;
    _TrackerAddr.IPv4Addr.bStream = true;

    pthread_rwlock_init(&_hSTValidLock, NULL);
    pthread_mutex_init(&_hUpHIdsLock, NULL);
    pthread_rwlock_init(&_hSessionLock, NULL);
}

LP2PEndpoint::~LP2PEndpoint()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_rwlock_destroy(&_hSessionLock);
    pthread_mutex_destroy(&_hUpHIdsLock);
    pthread_rwlock_destroy(&_hSTValidLock);
}

DWORD LP2PEndpoint::Load(const char *szTrackerIP, const char* szNodeName, UINT nMaxSTNum, UINT nMaxUploadNum, UINT nMaxSessionNum, LP2PEpObserver *pObserver)
{
    DWORD  dwRet;
    UINT  i, nMaxConnNum;
    LP2PUploadHandler  *pHandler;
    LP2PSession  *pSession;
    PLP2P_SEEDTARGET_REC  pSTRec;
    SOCK_ADDR  NodeAddr;

    DBG(D_LOG, (L"%s: tracker(%hs), node(%hs), st(%d), upload(%d), session(%d)\n", FULL_FUNC_NAME, szTrackerIP, szNodeName, nMaxSTNum, nMaxUploadNum, nMaxSessionNum));

    _pObserver = pObserver;

    _TrackerAddr.IPv4Addr.dwIP = inet_addr(szTrackerIP);
    _TrackerAddr.IPv4Addr.wPort = htons(SEED_TRACK_PORT);

    nMaxConnNum = nMaxUploadNum + 2;  // listener + SAH
    dwRet = Endpoint::Load(nMaxConnNum, 1, LPEP_SIG_NUM);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: load endpoint failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    _pHandlerList = new(std::nothrow) RingBuffer<LP2PUploadHandler*>(nMaxUploadNum, true);
    if ( !_pHandlerList )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc _pHandlerList failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    for ( i = 0; i < nMaxUploadNum; ++i )
    {
        pHandler = new(std::nothrow) LP2PUploadHandler();
        if ( !pHandler )
        {
            dwRet = ERROR_OUTOFMEMORY;
            DBG(D_WARN, (L"%s: alloc pHandler failed\n", FULL_FUNC_NAME));
            goto err3;
        }

        _pHandlerList->PushAtBack(pHandler);
    }

    // no need to enable internal lock
    _pSessionList = new(std::nothrow) RingBuffer<LP2PSession*>(nMaxSessionNum, false);
    if ( !_pSessionList )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc _pSessionList failed\n", FULL_FUNC_NAME));
        goto err3;
    }

    for ( i = 0; i < nMaxSessionNum; ++i )
    {
        pSession = new(std::nothrow) LP2PSession();
        if ( !pSession )
        {
            dwRet = ERROR_OUTOFMEMORY;
            DBG(D_WARN, (L"%s: alloc pSession failed\n", FULL_FUNC_NAME));
            goto err4;
        }

        _pSessionList->PushAtBack(pSession);
    }

    _pSessionEvent = new(std::nothrow) PCondEvent(nMaxSTNum);
    if ( !_pSessionEvent )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc _pSessionEvent failed\n", FULL_FUNC_NAME));
        goto err4;
    }

    _STList.resize(nMaxSTNum);
    dwRet = _STListMate.Init(nMaxSTNum);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: init VectorMate failed\n", FULL_FUNC_NAME));
        goto err5;
    }

    for ( i = 0; i < nMaxSTNum; ++i )
    {
        pSTRec = &_STList.at(i);
        DO_IF_DBG(pSTRec->bUsed = false);
        pSTRec->bValid = false;
        pSTRec->bSsValid = false;
    }

    dwRet = TimerCenter::Instance()->CreateTimer(TimerFunc, (HANDLE)this, _dwTimerId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: CreateTimer failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err6;
    }

    NodeAddr.bType = ADDR_TYPE_IPV4;
    NodeAddr.IPv4Addr.bStream = true;
    NodeAddr.IPv4Addr.dwIP = INADDR_ANY;
    NodeAddr.IPv4Addr.wPort = htons(LP2P_DOWNLOAD_PORT);
    dwRet = AddListener(&NodeAddr, false, _hListener);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: AddListener failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err7;
    }

    _pSAHandler = new(std::nothrow) SeedAgentHandler(szNodeName, &NodeAddr, nMaxSTNum);
    if ( !_pSAHandler )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc _pSAHandler failed\n", FULL_FUNC_NAME));
        goto err8;
    }

    dwRet = TimerCenter::Instance()->SetTimer(_dwTimerId, 50);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: SetTimer failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err9;
    }

    return ERROR_SUCCESS;

err9:
    delete _pSAHandler;
    _pSAHandler = NULL;

err8:
    RemoveListener(_hListener);

err7:
    TimerCenter::Instance()->DestroyTimer(_dwTimerId);
    _dwTimerId = INVALID_TIMER_ID;

err6:
    _STList.resize(0);
    _STListMate.Fini();

err5:
    delete _pSessionEvent;
    _pSessionEvent = NULL;

err4:
    while ( !_pSessionList->IsEmpty() )
    {
        delete _pSessionList->PopAtFront();
    }

    delete _pSessionList;
    _pSessionList = NULL;

err3:
    while ( !_pHandlerList->IsEmpty() )
    {
        delete _pHandlerList->PopAtFront();
    }

    delete _pHandlerList;
    _pHandlerList = NULL;

err2:
    Endpoint::Unload();

err1:
    return dwRet;
}

void LP2PEndpoint::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    if ( _dwTimerId != INVALID_TIMER_ID )
    {
        TimerCenter::Instance()->DestroyTimer(_dwTimerId);
        _dwTimerId = INVALID_TIMER_ID;
    }

    Endpoint::Unload();

    delete _pSAHandler;
    _pSAHandler = NULL;

    _STList.resize(0);
    assert(_STListMate.IsFull());
    _STListMate.Fini();

    delete _pSessionEvent;
    _pSessionEvent = NULL;

    if ( _pSessionList )
    {
        while ( !_pSessionList->IsEmpty() )
        {
            delete _pSessionList->PopAtFront();
        }

        delete _pSessionList;
        _pSessionList = NULL;
    }

    if ( _pHandlerList )
    {
        while ( !_pHandlerList->IsEmpty() )
        {
            delete _pHandlerList->PopAtFront();
        }

        delete _pHandlerList;
        _pHandlerList = NULL;
    }
}

DWORD LP2PEndpoint::AddSeedTarget(SeedTarget *pSeedTarget, DWORD dwCookie, DWORD &rdwSTId)
{
    DWORD  dwRet;
    UINT  nIdx;
    DWORD  dwMagic;
    PLP2P_SEEDTARGET_REC  pSTRec;

    DBG(D_INFO, (L"%s: %hs\n", FULL_FUNC_NAME, pSeedTarget->GetName()));

    dwRet = _STListMate.GetFreeIndex(nIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: reach the max st limit\n", FULL_FUNC_NAME));
        goto err1;
    }

    pSTRec = &_STList.at(nIdx);
    assert(!pSTRec->bUsed);
    DO_IF_DBG(pSTRec->bUsed = true);

    dwMagic = InterlockedIncrement((LONG*)&_dwSTMagic);

    pSTRec->wMagic = ADJUST_WMAGICNUM(dwMagic);
    pSTRec->pSeedTarget = pSeedTarget;
    pSTRec->dwCookie = dwCookie;
    assert(pSTRec->UpHIdList.empty());
    pSTRec->SessionReq = SESSION_REQ_NONE;
    pSTRec->bSsValid = false;
    pSTRec->pSession = NULL;
    pSTRec->dwErrCode = ERROR_SUCCESS;
    pSTRec->bValid = true;

    // see issues_connid.txt
    rdwSTId = MAKE_DWID(nIdx, pSTRec->wMagic);

    dwRet = _pSAHandler->AddSeedTarget(pSeedTarget, rdwSTId, pSTRec->dwSASTId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: AddSeedTarget to seed agent failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    // _pSAHandler->AddSeedTarget() may fail, so set SessionReq at last,
    // OnSessionFree() will only touch the valid STRec and check SessionReq
    if ( !pSeedTarget->IsFinished() )
    {
        pSTRec->SessionReq = SESSION_REQ_ACTIVE;
        _pSessionEvent->Signal(nIdx);
        RequestEpWork(SIG_SESSION_REQ);
    }

    pSeedTarget->AddRef();

    DBG(D_INFO, (L"%s: name(%hs), idx(%d)\n", FULL_FUNC_NAME, pSeedTarget->GetName(), nIdx));

    return ERROR_SUCCESS;

err2:
    pSTRec->bValid = false;
    DO_IF_DBG(pSTRec->bUsed = false);
    _STListMate.PutFreeIndex(nIdx);

err1:
    rdwSTId = INVALID_LP2P_ST_ID;

    return dwRet;
}

DWORD LP2PEndpoint::RemoveSeedTarget(DWORD dwSTId)
{
    UINT  nSTIdx;
    WORD  wMagic;
    PLP2P_SEEDTARGET_REC  pSTRec;
    list<DWORD>::const_iterator  itr;
    list<DWORD>::const_iterator  end_itr;

    nSTIdx = WIDX_FROM_DWID(dwSTId);
    wMagic = WMAGICNUM_FROM_DWID(dwSTId);

    DBG(D_INFO, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nSTIdx, wMagic));

    if ( nSTIdx >= _STList.size() )
    {
        DBG(D_WARN, (L"%s: invalid ST id\n", FULL_FUNC_NAME));
        return ERROR_INVALID_PARAMETER;
    }

    pSTRec = &_STList.at(nSTIdx);
    if ( wMagic != pSTRec->wMagic )
    {
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nSTIdx));
        return ERROR_NOT_FOUND;
    }
    assert(pSTRec->bUsed && pSTRec->bValid);

    pthread_rwlock_wrlock(&_hSTValidLock);
    // for SAH & UpH
    pSTRec->bValid = false;
    pthread_rwlock_unlock(&_hSTValidLock);

    pthread_mutex_lock(&_hUpHIdsLock);
    for ( itr = pSTRec->UpHIdList.begin(), end_itr = pSTRec->UpHIdList.end(); itr != end_itr; ++itr )
    {
        DisconnectConn(*itr);
    }
    pthread_mutex_unlock(&_hUpHIdsLock);
    while ( !pSTRec->UpHIdList.empty() )
    {
        DBG(D_INFO, (L"%s: waiting for upload(s) disconnect\n", FULL_FUNC_NAME));
        Sleep(200);
    }

    // see issues_pcond3.txt
    pthread_rwlock_wrlock(&_hSessionLock);
    // for session
    if ( pSTRec->SessionReq == SESSION_REQ_ACTIVE )
    {
        pSTRec->SessionReq = SESSION_REQ_REMOVE;
        _pSessionEvent->Signal(nSTIdx);
        RequestEpWork(SIG_SESSION_REQ);
    }
    pthread_rwlock_unlock(&_hSessionLock);
    while ( pSTRec->SessionReq != SESSION_REQ_DONE )
    {
        DBG(D_INFO, (L"%s: waiting for session[%d] unload\n", FULL_FUNC_NAME, nSTIdx));
        Sleep(200);
    }

    _pSAHandler->RemoveSeedTarget(pSTRec->dwSASTId);

    pSTRec->pSeedTarget->Release();

    // invalid the wMagic
    // OnUploadDisconnected() will check wMagic, so put it here
    pSTRec->wMagic = ~wMagic;

    DO_IF_DBG(pSTRec->bUsed = false);
    _STListMate.PutFreeIndex(nSTIdx);

    return ERROR_SUCCESS;
}

void LP2PEndpoint::OnFinished(DWORD dwCookie, DWORD dwErrCode)
{
    UINT  nSTIdx;
    WORD  wMagic;
    PLP2P_SEEDTARGET_REC  pSTRec;

    nSTIdx = WIDX_FROM_DWID(dwCookie);
    wMagic = WMAGICNUM_FROM_DWID(dwCookie);

    DBG(D_INFO, (L"%s: idx(%d), magic(%d), err(%d)\n", FULL_FUNC_NAME, nSTIdx, wMagic, dwErrCode));

    if ( nSTIdx >= _STList.size() )
    {
        DBG(D_WARN, (L"%s: invalid ST id\n", FULL_FUNC_NAME));
        assert(0);
        return;
    }

    pSTRec = &_STList.at(nSTIdx);
    // session exist, so no need lock
    assert(pSTRec->bUsed);
    // bValid may be false
//    assert(pSTRec->bValid);
    assert(wMagic == pSTRec->wMagic);

    pSTRec->dwErrCode = dwErrCode;

    _pObserver->OnSeedTargetFinished(pSTRec->dwCookie, pSTRec->dwErrCode);

    // see issues_pcond2.txt, but no need signal under lock,
    // because OnFinished() is called under session's work thread,
    // it is definite valid before session.Unload()
    _pSessionEvent->Signal(nSTIdx);
    RequestEpWork(SIG_SESSION_REQ);
}

void LP2PEndpoint::OnEnumNodes4Seed(DWORD dwCookie)
{
    UINT  nSTIdx;
    WORD  wMagic;
    PLP2P_SEEDTARGET_REC  pSTRec;

    nSTIdx = WIDX_FROM_DWID(dwCookie);
    wMagic = WMAGICNUM_FROM_DWID(dwCookie);

    DBG(D_VERB, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nSTIdx, wMagic));

    if ( nSTIdx >= _STList.size() )
    {
        DBG(D_WARN, (L"%s: invalid ST id\n", FULL_FUNC_NAME));
        assert(0);
        return;
    }

    pSTRec = &_STList.at(nSTIdx);
    // session exist, so no need lock
    assert(pSTRec->bUsed);
    // bValid may be false
//    assert(pSTRec->bValid);
    assert(wMagic == pSTRec->wMagic);

    // RemoveSeedTarget() will unload session before invalid dwSASTId
    _pSAHandler->EnumNodes4Seed(pSTRec->dwSASTId);
}

void LP2PEndpoint::OnListenerEnded(DWORD hListener)
{
    DBG(D_LOG, (L"%s: listener(%x) ended\n", FULL_FUNC_NAME, hListener));

    assert(hListener == _hListener);
    _hListener = INVALID_LISTENER_ID;
}

EpProtocolHandler *LP2PEndpoint::OnNewHandler(DWORD hListener, Connection *pConnection, DWORD dwConnId)
{
    LP2PUploadHandler  *pHandler;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    // no need lock
    if ( _pHandlerList->IsEmpty() )
    {
        DBG(D_WARN, (L"%s: reach the max conn limit\n", FULL_FUNC_NAME));
        return NULL;
    }

    // no need lock
    pHandler = _pHandlerList->PopAtFront();

    return pHandler;
}

void LP2PEndpoint::OnAsyncConnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_INFO, (L"%s: %#.8x\n", FULL_FUNC_NAME, dwConnId));

    if ( dwConnId == _dwSAHConnId )
    {
        _pObserver->OnConnected();
    }
}

void LP2PEndpoint::OnDisconnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_INFO, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));

    if ( dwConnId == _dwSAHConnId )
    {
        OnSeedAgentDisconnected();
    }
    else
    {
        OnUploadDisconnected(dwConnId, dynamic_cast<LP2PUploadHandler*>(pHandler));
    }
}

void LP2PEndpoint::OnConnEvent(DWORD dwConnId, EpProtocolHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody)
{
    DBG(D_VERB, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));

    if ( dwConnId == _dwSAHConnId )
    {
        OnSeedAgentConnEvent(pEvtHdr, pEvtBody);
    }
    else
    {
        OnUploadConnEvent(dwConnId, dynamic_cast<LP2PUploadHandler*>(pHandler), pEvtHdr, pEvtBody);
    }
}

void LP2PEndpoint::ProcessEpWork(UINT nSigIdx)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nSigIdx));

    switch ( nSigIdx )
    {
        case SIG_SAH_CONNECT:
            OnSAHConnect();
            break;

        case SIG_SESSION_REQ:
            OnSessionRequest();
            break;

        case SIG_SESSION_FREE:
            OnSessionFree();
            break;

        default:
            DBG(D_WARN, (L"%s: unknown sig(%d)\n", FULL_FUNC_NAME, nSigIdx));
            assert(0);
            break;
    }
}

void LP2PEndpoint::OnSAHConnect()
{
    DWORD  dwRet;
    SOCK_ADDR  LocalAddr;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( _dwSAHConnId != INVALID_EP_CONN_ID )
    {
        return;
    }

    _pObserver->OnConnecting();

    LocalAddr.bType = ADDR_TYPE_IPV4;
    LocalAddr.IPv4Addr.bStream = true;
    LocalAddr.IPv4Addr.dwIP = INADDR_ANY;
    LocalAddr.IPv4Addr.wPort = htons(0);
    dwRet = ConnectOut(&LocalAddr, &_TrackerAddr, _pSAHandler, _dwSAHConnId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: connect to the tacker server failed: %d\n", FULL_FUNC_NAME, dwRet));
        _pObserver->OnConnectFailed(dwRet);
    }
}

void LP2PEndpoint::OnSessionRequest()
{
    UINT  nSTIdx;
    PLP2P_SEEDTARGET_REC  pSTRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    while ( true )
    {
        nSTIdx = _pSessionEvent->Wait(0, true);

        if ( nSTIdx == INVALID_PCOND_IDX )
        {
            break;
        }

        assert(nSTIdx < _STList.size());
        pSTRec = &_STList.at(nSTIdx);
        assert(pSTRec->bUsed);

        while ( true )
        {
            if ( ProcessSession(nSTIdx, pSTRec) )
            {
                pthread_rwlock_rdlock(&_hSessionLock);
                _pSessionEvent->Reset(nSTIdx, 0);
                pSTRec->SessionReq = SESSION_REQ_DONE;
                pthread_rwlock_unlock(&_hSessionLock);
                break;
            }

            if ( _pSessionEvent->Finish(nSTIdx, 0) )
            {
                break;
            }
        }
    }
}

void LP2PEndpoint::OnSessionFree()
{
    UINT  nIdx, nSize;
    PLP2P_SEEDTARGET_REC  pSTRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_rwlock_rdlock(&_hSTValidLock);

    for ( nIdx = 0, nSize = (UINT)_STList.size(); nIdx < nSize; ++nIdx )
    {
        pSTRec = &_STList.at(nIdx);

        if ( !pSTRec->bValid )
        {
            continue;
        }

        if ( pSTRec->SessionReq != SESSION_REQ_ACTIVE )
        {
            continue;
        }

        if ( pSTRec->pSession )
        {
            continue;
        }

        if ( (!pSTRec->pSeedTarget->IsFinished()) && (pSTRec->dwErrCode == ERROR_SUCCESS) )
        {
            _pSessionEvent->Signal(nIdx);
            RequestEpWork(SIG_SESSION_REQ);
        }
    }

    pthread_rwlock_unlock(&_hSTValidLock);
}

bool LP2PEndpoint::ProcessSession(UINT nSTIdx, PLP2P_SEEDTARGET_REC pSTRec)
{
    DWORD  dwRet;
    LP2PSession  *pSession;

    DBG(D_INFO, (L"%s: idx(%d)\n", FULL_FUNC_NAME, nSTIdx));

    assert(pSTRec->SessionReq != SESSION_REQ_DONE);

    // no need lock
    if ( (pSTRec->SessionReq == SESSION_REQ_ACTIVE) && (!pSTRec->pSession) )
    {
        if ( !_pSessionList->IsEmpty() )
        {
            pSession = _pSessionList->Front();
            dwRet = pSession->Load(pSTRec->pSeedTarget, LP2P_MAX_PEER_NUM, this, MAKE_DWID(nSTIdx, pSTRec->wMagic));
            if ( dwRet == ERROR_SUCCESS )
            {
                _pSessionList->Pop();
                pSTRec->pSession = pSession;
                pSTRec->bSsValid = true;
            }
        }

        return false;
    }

    if ( (pSTRec->SessionReq == SESSION_REQ_REMOVE) || pSTRec->pSeedTarget->IsFinished() || (pSTRec->dwErrCode != ERROR_SUCCESS) )
    {
        if ( pSTRec->pSession )
        {
            pthread_rwlock_wrlock(&_hSessionLock);
            pSTRec->bSsValid = false;
            pthread_rwlock_unlock(&_hSessionLock);

            pSTRec->pSession->Unload();
            _pSessionList->PushAtBack(pSTRec->pSession);
            pSTRec->pSession = NULL;

            RequestEpWork(SIG_SESSION_FREE);
        }

        // if check SESSION_REQ_REMOVE outside, pSession may exist,
        // and signal will be reset, it is wrong.
        // it's no matter go here is for IsFinished(), now req=SESSION_REQ_REMOVE,
        return (pSTRec->SessionReq == SESSION_REQ_REMOVE);
    }

    return false;
}

void LP2PEndpoint::OnSeedAgentDisconnected()
{
    DBG(D_WARN, (L"%s: disconnected from the tracker server\n", FULL_FUNC_NAME));

    _pObserver->OnDisconnected();

    _dwSAHConnId = INVALID_EP_CONN_ID;
}

void LP2PEndpoint::OnUploadDisconnected(DWORD dwConnId, LP2PUploadHandler *pHandler)
{
    DWORD  dwSTId;
    UINT  nSTIdx;
    WORD  wMagic;
    PLP2P_SEEDTARGET_REC  pSTRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    dwSTId = pHandler->GetCookie();
    if ( dwSTId == INVALID_LP2P_ST_ID )
    {
        goto out;
    }

    nSTIdx = WIDX_FROM_DWID(dwSTId);
    wMagic = WMAGICNUM_FROM_DWID(dwSTId);

    DBG(D_VERB, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nSTIdx, wMagic));

    // no need lock
    assert(nSTIdx < _STList.size());
    pSTRec = &_STList.at(nSTIdx);

    assert(pSTRec->bUsed);
    assert(wMagic == pSTRec->wMagic);

    pthread_mutex_lock(&_hUpHIdsLock);
    pSTRec->UpHIdList.remove(dwConnId);
    pthread_mutex_unlock(&_hUpHIdsLock);

out:
    _pHandlerList->PushAtBackSafe(pHandler);
}

void LP2PEndpoint::OnSeedAgentConnEvent(PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody)
{
    UINT  nSTIdx;
    WORD  wMagic;
    PLP2P_SEEDTARGET_REC  pSTRec;

    if ( pEvtHdr->nCode == C_LP2P_NODE_REG )
    {
        DBG(D_LOG, (L"%s: C_LP2P_NODE_REG: param(%d)\n", FULL_FUNC_NAME, pEvtHdr->dwParam));
        _pObserver->OnNodeRegistered(pEvtHdr->dwParam);
        return;
    }

    nSTIdx = WIDX_FROM_DWID(pEvtHdr->dwCookie);
    wMagic = WMAGICNUM_FROM_DWID(pEvtHdr->dwCookie);

    DBG(D_VERB, (L"%s: idx(%d), code(%d), param(%d)\n", FULL_FUNC_NAME, nSTIdx, pEvtHdr->nCode, pEvtHdr->dwParam));

    pSTRec = &_STList.at(nSTIdx);
    // SA may issue event after SA.RemoveSeedTarget()
//    assert(pSTRec->bUsed);

    pthread_rwlock_rdlock(&_hSTValidLock);

    if ( (!pSTRec->bValid) || (wMagic != pSTRec->wMagic) )
    {
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nSTIdx));
        goto out;
    }
    assert(pSTRec->bUsed);

    switch ( pEvtHdr->nCode )
    {
        case C_LP2P_SEED_ADD:
            break;

        case C_LP2P_SEED_STATE_UPDATE:
            break;

        case C_LP2P_N4S_ENUM:
            assert(pEvtBody->data_size() >= sizeof(EVT_NODE4SEED_ENUM));
            pthread_rwlock_rdlock(&_hSessionLock);
            if ( pSTRec->bSsValid )
            {
                PEVT_NODE4SEED_ENUM  pEvtN4SEnum = (PEVT_NODE4SEED_ENUM)pEvtBody->data();
                assert(pSTRec->pSession);
                pSTRec->pSession->AddNodes4Seed(pEvtN4SEnum->N4SInfo, pEvtN4SEnum->nCount, &pEvtN4SEnum->LocalAddr);
            }
            pthread_rwlock_unlock(&_hSessionLock);
            break;

        default:
            DBG(D_WARN, (L"%s: unknown code(%x)\n", FULL_FUNC_NAME, pEvtHdr->nCode));
            assert(0);
            break;
    }

out:
    pthread_rwlock_unlock(&_hSTValidLock);
}

void LP2PEndpoint::OnUploadConnEvent(DWORD dwConnId, LP2PUploadHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody)
{
    DBG(D_VERB, (L"%s: connid(%#.8x), code(%d)\n", FULL_FUNC_NAME, dwConnId, pEvtHdr->nCode));

    switch ( pEvtHdr->nCode )
    {
        case C_LP2P_SEEDTARGET_OPEN:
        {
            assert(pEvtBody->data_size() == sizeof(EVT_SEEDTARGET_OPEN));
            PEVT_SEEDTARGET_OPEN  pEvtSTOpen = (PEVT_SEEDTARGET_OPEN)pEvtBody->data();
            OnOpenSeedTarget(dwConnId, pHandler, pEvtSTOpen->SHA1ValSeed);
            break;
        }

        default:
            DBG(D_WARN, (L"%s: unknown code(%x)\n", FULL_FUNC_NAME, pEvtHdr->nCode));
            assert(0);
            break;
    }
}

void LP2PEndpoint::OnOpenSeedTarget(DWORD dwConnId, LP2PUploadHandler *pHandler, const SHA1_VAL &SHA1ValSeed)
{
    UINT  nIdx, nSize;
    PLP2P_SEEDTARGET_REC  pSTRec;
    DWORD  dwResult = ERROR_NOT_FOUND;
    SeedTarget  *pSeedTarget = NULL;
    DWORD  dwCookie = INVALID_LP2P_ST_ID;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_rwlock_rdlock(&_hSTValidLock);
    for ( nIdx = 0, nSize = (UINT)_STList.size(); nIdx < nSize; ++nIdx )
    {
        pSTRec = &_STList.at(nIdx);

        if ( pSTRec->bValid && pSTRec->pSeedTarget->IsMatch(SHA1ValSeed) )
        {
            if ( pSTRec->UpHIdList.size() < pSTRec->pSeedTarget->GetMaxReadNum() )
            {
                // because using rdlock, use lock here
                pthread_mutex_lock(&_hUpHIdsLock);
                pSTRec->UpHIdList.push_back(dwConnId);
                pthread_mutex_unlock(&_hUpHIdsLock);

                dwResult = ERROR_SUCCESS;
                pSeedTarget = pSTRec->pSeedTarget;
                dwCookie = MAKE_DWID(nIdx, pSTRec->wMagic);
            }
            else
            {
                dwResult = ERROR_BUSY;
            }

            break;
        }
    }
    pthread_rwlock_unlock(&_hSTValidLock);

    // UpHIdList.push_back() under lock at above, so it is safe to use pSTRec->pSeedTarget
    // RemoveSeedTarget() will wait UpHIdList empty and then release pSeedTarget
    // OnUploadDisconnected() and OnOpenSeedTarget() is serialized
    pHandler->Attach(dwResult, pSeedTarget, dwCookie);
}

void LP2PEndpoint::TimerFunc(HANDLE hArg)
{
    LP2PEndpoint  *pThis = static_cast<LP2PEndpoint*>(hArg);

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( pThis->_dwSAHConnId == INVALID_EP_CONN_ID )
    {
        pThis->RequestEpWork(SIG_SAH_CONNECT);
    }

    TimerCenter::Instance()->SetTimer(pThis->_dwTimerId, LP2P_EP_TIMER);
}

