/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "SeedAgentHandler.h"
#include "LP2PMCastSender.h"
#include "LP2PMCastEndpoint.h"


LP2PMCastEndpoint *LP2PMCastEndpoint::_instance = NULL;

LP2PMCastEndpoint *LP2PMCastEndpoint::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new LP2PMCastEndpoint();
    }
    return _instance;
}

void LP2PMCastEndpoint::Destroy()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

LP2PMCastEndpoint::LP2PMCastEndpoint()
    : Endpoint()
    , _pObserver(NULL)
    , _pSeedTarget(NULL)
    , _dwMCastConnId(INVALID_EP_CONN_ID)
    , _dwSAHConnId(INVALID_EP_CONN_ID)
    , _dwSASTId(INVALID_LP2P_ST_ID)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

LP2PMCastEndpoint::~LP2PMCastEndpoint()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(!_pSeedTarget);
}

DWORD LP2PMCastEndpoint::Load(const char *szTrackerIP, const char* szNodeName, const char *szIntfIP, SeedTarget *pSeedTarget, LP2PMCastEpObserver *pObserver)
{
    DWORD  dwRet;
    LP2PMCastSender  *pMCastHandler;
    SeedAgentHandler  *pSAHandler;
    DWORD  dwMCastIP;
    WORD  wMCastPort;
    SOCK_ADDR  LocalAddr, RemoteAddr;

    DBG(D_LOG, (L"%s: tracker(%hs), node(%hs), intfip(%hs), seed(%hs)\n", FULL_FUNC_NAME, szTrackerIP, szNodeName, szIntfIP, pSeedTarget->GetName()));

    _pObserver = pObserver;

    pSeedTarget->AddRef();
    _pSeedTarget = pSeedTarget;

    dwRet = Endpoint::Load(2, 0, 0);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: load endpoint failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    pMCastHandler = new(std::nothrow) LP2PMCastSender(_pSeedTarget);
    if ( !pMCastHandler )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc pMCastHandler failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    dwMCastIP = LP2P_MCAST_ADDR + RAND_BETWEEN(0, LP2P_MCAST_RANGE);
    wMCastPort = LP2P_MCAST_PORT + RAND_BETWEEN(0, LP2P_MCAST_RANGE);
    DBG(D_LOG, (L"%s: mcast addr: %#.8x:%d\n", FULL_FUNC_NAME, dwMCastIP, wMCastPort));

    LocalAddr.bType = ADDR_TYPE_IPV4;
    LocalAddr.IPv4Addr.bStream = false;
    LocalAddr.IPv4Addr.dwIP = inet_addr(szIntfIP);
    LocalAddr.IPv4Addr.wPort = htons(0);

    RemoteAddr.bType = ADDR_TYPE_IPV4;
    RemoteAddr.IPv4Addr.bStream = false;
    RemoteAddr.IPv4Addr.dwIP = htonl(dwMCastIP);
    RemoteAddr.IPv4Addr.wPort = htons(wMCastPort);

    dwRet = ConnectOut(&LocalAddr, &RemoteAddr, pMCastHandler, _dwMCastConnId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ConnectOut(MCast) failed: %d\n", FULL_FUNC_NAME, dwRet));
        delete pMCastHandler;
        goto err2;
    }

    pSAHandler = new(std::nothrow) SeedAgentHandler(szNodeName, &RemoteAddr, 1);
    if ( !pSAHandler )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc pSAHandler failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    LocalAddr.bType = ADDR_TYPE_IPV4;
    LocalAddr.IPv4Addr.bStream = true;
    LocalAddr.IPv4Addr.dwIP = INADDR_ANY;
    LocalAddr.IPv4Addr.wPort = htons(0);

    RemoteAddr.bType = ADDR_TYPE_IPV4;
    RemoteAddr.IPv4Addr.bStream = true;
    RemoteAddr.IPv4Addr.dwIP = inet_addr(szTrackerIP);
    RemoteAddr.IPv4Addr.wPort = htons(SEED_TRACK_PORT);

    dwRet = ConnectOut(&LocalAddr, &RemoteAddr, pSAHandler, _dwSAHConnId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ConnectOut(SA) failed: %d\n", FULL_FUNC_NAME, dwRet));
        delete pSAHandler;
        goto err2;
    }

    dwRet = pSAHandler->AddSeedTarget(_pSeedTarget, 0, _dwSASTId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: AddSeedTarget failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    Endpoint::Unload();

err1:
    _pSeedTarget->Release();
    _pSeedTarget = NULL;

    return dwRet;
}

void LP2PMCastEndpoint::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    Endpoint::Unload();

    if ( _pSeedTarget )
    {
        _pSeedTarget->Release();
        _pSeedTarget = NULL;
    }
}

DWORD LP2PMCastEndpoint::Start()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    return QueueConnCmd(_dwMCastConnId, C_CONN_MCAST_START, NULL);
}

void LP2PMCastEndpoint::OnListenerEnded(DWORD hListener)
{
    DBG(D_LOG, (L"%s: listener(%x) ended\n", FULL_FUNC_NAME, hListener));

    assert(0);
}

EpProtocolHandler *LP2PMCastEndpoint::OnNewHandler(DWORD hListener, Connection *pConnection, DWORD dwConnId)
{
    DBG(D_LOG, (L"%s: listener(%x)\n", FULL_FUNC_NAME, hListener));

    assert(0);

    return NULL;
}

void LP2PMCastEndpoint::OnAsyncConnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_INFO, (L"%s: %#.8x\n", FULL_FUNC_NAME, dwConnId));
}

void LP2PMCastEndpoint::OnDisconnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    SeedAgentHandler  *pSAHandler;

    DBG(D_VERB, (L"%s: id(%#.8x)\n", FULL_FUNC_NAME, dwConnId));

    if ( dwConnId == _dwMCastConnId )
    {
        DBG(D_INFO, (L"%s: MCast\n", FULL_FUNC_NAME));
        dwConnId = INVALID_EP_CONN_ID;

        _pObserver->OnFinished();
    }
    else if ( dwConnId == _dwSAHConnId )
    {
        DBG(D_INFO, (L"%s: SA\n", FULL_FUNC_NAME));
        _dwSAHConnId = INVALID_EP_CONN_ID;

        if ( _dwSASTId != INVALID_LP2P_ST_ID )
        {
            pSAHandler = dynamic_cast<SeedAgentHandler*>(pHandler);
            pSAHandler->RemoveSeedTarget(_dwSASTId);
            _dwSASTId = INVALID_LP2P_ST_ID;
        }
    }

    delete pHandler;
}

void LP2PMCastEndpoint::OnConnEvent(DWORD dwConnId, EpProtocolHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody)
{
    DBG(D_VERB, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));

    if ( dwConnId == _dwSAHConnId )
    {
        DBG(D_INFO, (L"%s: sa event: code=%x, ret=%d\n", FULL_FUNC_NAME, pEvtHdr->nCode, pEvtHdr->dwParam));

        if ( pEvtHdr->dwParam != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: failed\n", FULL_FUNC_NAME));
            return;
        }

        if ( pEvtHdr->nCode == C_LP2P_SEED_ADD )
        {
            DBG(D_INFO, (L"%s: add seed OK\n", FULL_FUNC_NAME));
        }
    }
}

void LP2PMCastEndpoint::ProcessEpWork(UINT nSigIdx)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nSigIdx));

    assert(0);
}
