/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "LP2PMCastRecver.h"


LP2PMCastRecver::LP2PMCastRecver(SeedTarget *pSeedTarget)
    : EpProtocolHandler(LP2P_MCAST_RECVQSIZE, LP2P_MCAST_DATABSIZE,
                   0, 0,
                   0, 0, false)
    , _pSeedTarget(pSeedTarget)
    , _dwSessionId(INVALID_ST_SESS_ID)
    , _dwBlkId(INVALID_ST_BLK_ID)
    , _nBlkIdx(INVALID_INDEX_VALUE)
    , _dwBlkOff(INVALID_OFFSET_VALUE)
    , _bStarted(false)
    , _nKickCnt(0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pSeedTarget->AddRef();
}

LP2PMCastRecver::~LP2PMCastRecver()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_ST_BLK_ID);
    assert(_dwSessionId == INVALID_ST_SESS_ID);

    _pSeedTarget->Release();
}

DWORD LP2PMCastRecver::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwSessionId == INVALID_ST_SESS_ID);
    dwRet = _pSeedTarget->BeginSession(false, _dwSessionId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_INFO, (L"%s: begin session for write failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    pConnection->SetRecvBufSize(MCAST_R_SOCKBUF_SIZE);

    dwRet = EpProtocolHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_INFO, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    _pSeedTarget->EndSession(_dwSessionId);
    _dwSessionId = INVALID_ST_SESS_ID;

err1:
    return dwRet;
}

void LP2PMCastRecver::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpProtocolHandler::Unbind();

    if ( _dwSessionId != INVALID_ST_SESS_ID )
    {
        CloseCurBlock();

        _pSeedTarget->EndSession(_dwSessionId);
        _dwSessionId = INVALID_ST_SESS_ID;
    }

    _bStarted = false;
    _nKickCnt = 0;
}

void LP2PMCastRecver::PrepareSend()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

void LP2PMCastRecver::ProcessCmd(DWORD dwCmdMask)
{
    UINT  nLimit;

    DBG(D_VERB, (L"%s: %x\n", FULL_FUNC_NAME, dwCmdMask));

    if ( dwCmdMask & C_CONN_MCAST_TIMER )
    {
        nLimit = _bStarted ? LP2P_MCAST_ALIVE_CNT : LP2P_MCAST_WAIT_CNT;
        if ( _nKickCnt++ >= nLimit )
        {
            DBG(D_INFO, (L"%s: timeout\n", FULL_FUNC_NAME));
            ReqEndSession();
        }
    }

    EpProtocolHandler::ProcessCmd(dwCmdMask);
}

bool LP2PMCastRecver::ProcessRecvData(BufferItem *pbiData)
{
    DWORD  dwRet;
    PUDPCAST_BLK_HDR  pHdr;
    DWORD  dwSize;
    bool  bEndOfBlk;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( pbiData->data_size() < sizeof(UDPCAST_BLK_HDR) )
    {
        DBG(D_WARN, (L"%s: invalid packet\n", FULL_FUNC_NAME));
        return true;
    }

    _nKickCnt = 0;
    _bStarted = true;

    pHdr = (PUDPCAST_BLK_HDR)pbiData->data();
    pbiData->Pull(sizeof(UDPCAST_BLK_HDR));

    if ( pHdr->dwBlkIdx != (DWORD)_nBlkIdx )
    {
        CloseCurBlock();

        if ( !pHdr->dwOffset )
        {
            OpenNextBlock(pHdr->dwBlkIdx);
        }
    }

    if ( pHdr->dwOffset == INVALID_OFFSET_VALUE )
    {
        DBG(D_LOG, (L"%s: mcast finished\n", FULL_FUNC_NAME));
        ReqEndSession();
        return true;
    }

    // _dwBlkOff's valid depend on _dwBlkId's valid
    // so check _dwBlkId first
    if ( _dwBlkId == INVALID_ST_BLK_ID )
    {
        DBG(D_VERB, (L"%s: drop packet: block not opened\n", FULL_FUNC_NAME));
        return true;
    }

    if ( pHdr->dwOffset != _dwBlkOff )
    {
        DBG(D_VERB, (L"%s: drop packet: offset not match\n", FULL_FUNC_NAME));
        CloseCurBlock();
        return true;
    }

    dwSize = (DWORD)pbiData->data_size();
    dwRet = _pSeedTarget->WriteBlock(_dwBlkId, pbiData->data(), dwSize, bEndOfBlk);
    if ( (dwRet != ERROR_SUCCESS) || bEndOfBlk )
    {
        CloseCurBlock();
    }
    else
    {
        _dwBlkOff += dwSize;
    }

    return true;
}

DWORD LP2PMCastRecver::OpenNextBlock(UINT nBlkIdx)
{
    DWORD  dwRet;
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nBlkIdx));

    assert(_dwBlkId == INVALID_ST_BLK_ID);

    dwRet = _pSeedTarget->OpenBlock(_dwSessionId, nBlkIdx, _dwBlkId);
    if ( dwRet == ERROR_SUCCESS )
    {
        _nBlkIdx = nBlkIdx;
        _dwBlkOff = 0;
    }

    return dwRet;
}

void LP2PMCastRecver::CloseCurBlock()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _dwBlkId != INVALID_ST_BLK_ID )
    {
        _pSeedTarget->CloseBlock(_dwBlkId);
        _dwBlkId = INVALID_ST_BLK_ID;

        _nBlkIdx = INVALID_INDEX_VALUE;
    }
}
