/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "LP2PMCastSender.h"


LP2PMCastSender::LP2PMCastSender(SeedTarget *pSeedTarget)
    : EpProtocolHandler(0, 0,
                   LP2P_MCAST_SENDQSIZE, LP2P_MCAST_DATABSIZE,
                   0, 0, false)
    , _pSeedTarget(pSeedTarget)
    , _dwSessionId(INVALID_ST_SESS_ID)
    , _dwBlkId(INVALID_ST_BLK_ID)
    , _nBlkIdx(0)
    , _dwBlkOff(0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pSeedTarget->AddRef();
}

LP2PMCastSender::~LP2PMCastSender()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_ST_BLK_ID);
    assert(_dwSessionId == INVALID_ST_SESS_ID);

    _pSeedTarget->Release();
}

DWORD LP2PMCastSender::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwSessionId == INVALID_ST_SESS_ID);
    dwRet = _pSeedTarget->BeginSession(true, _dwSessionId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_INFO, (L"%s: begin session for read failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    dwRet = EpProtocolHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_INFO, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    _pSeedTarget->EndSession(_dwSessionId);
    _dwSessionId = INVALID_ST_SESS_ID;

err1:
    return dwRet;
}

void LP2PMCastSender::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpProtocolHandler::Unbind();

    if ( _dwSessionId != INVALID_ST_SESS_ID )
    {
        if ( _dwBlkId != INVALID_ST_BLK_ID )
        {
            _pSeedTarget->CloseBlock(_dwBlkId);
            _dwBlkId = INVALID_ST_BLK_ID;
        }

        _pSeedTarget->EndSession(_dwSessionId);
        _dwSessionId = INVALID_ST_SESS_ID;
    }

    _nBlkIdx = 0;
}

void LP2PMCastSender::ProcessCmd(DWORD dwCmdMask)
{
    DBG(D_INFO, (L"%s: %x\n", FULL_FUNC_NAME, dwCmdMask));

    if ( dwCmdMask & C_CONN_MCAST_START )
    {
        ReqSendProc(true);
    }

    EpProtocolHandler::ProcessCmd(dwCmdMask);
}

void LP2PMCastSender::PrepareSend()
{
    BufferItem  *pbi;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    while ( true )
    {
        pbi = GetSendBuf();
        if ( !pbi )
        {
            break;
        }

        if ( _dwBlkId == INVALID_ST_BLK_ID )
        {
            if ( !OpenBlock() )
            {
                SendDone(pbi);
                break;
            }
        }

        SendData(pbi);
    }
}

bool LP2PMCastSender::ProcessRecvData(BufferItem *pbiData)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return true;
}

bool LP2PMCastSender::OpenBlock()
{
    DWORD  dwRet;
    UINT  i, nCnt;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_ST_BLK_ID);

    for ( i = _nBlkIdx, nCnt = _pSeedTarget->GetBlockCount(); i < nCnt; ++i )
    {
        DBG(D_VERB, (L"%s: sending Block[%d]\n", FULL_FUNC_NAME, i));
        dwRet = _pSeedTarget->OpenBlock(_dwSessionId, i, _dwBlkId);
        if ( dwRet == ERROR_SUCCESS )
        {
            _dwBlkOff = 0;
            break;
        }

        DBG(D_WARN, (L"%s: idx[#%d] failed: %d\n", FULL_FUNC_NAME, i, dwRet));
    }

    _nBlkIdx = i;

    return (i < nCnt);
}

void LP2PMCastSender::SendDone(BufferItem *pbi)
{
    PUDPCAST_BLK_HDR  pHdr;

    DBG(D_LOG, (L"%s: mcast finished\n", FULL_FUNC_NAME));

    pHdr = (PUDPCAST_BLK_HDR)pbi->Put(sizeof(UDPCAST_BLK_HDR));
    pHdr->dwBlkIdx = INVALID_INDEX_VALUE;
    pHdr->dwOffset = INVALID_OFFSET_VALUE;

    Sleep(50);  // enough
    PutSendData(pbi, true);
    Sleep(50);  // enough

    ReqEndSession();
}

void LP2PMCastSender::SendData(BufferItem *pbi)
{
    DWORD  dwRet;
    DWORD  dwSize;
    bool  bEndOfBlk;
    PUDPCAST_BLK_HDR  pHdr;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId != INVALID_ST_BLK_ID);

    pHdr = (PUDPCAST_BLK_HDR)pbi->Put(sizeof(UDPCAST_BLK_HDR));
    pHdr->dwBlkIdx = _nBlkIdx;

    dwSize = pbi->tail_size();
    dwRet = _pSeedTarget->ReadBlock(_dwBlkId, pbi->tail(), dwSize, bEndOfBlk);
    if ( dwRet == ERROR_SUCCESS )
    {
        pbi->Put(dwSize);
        pHdr->dwOffset = _dwBlkOff;
        _dwBlkOff += dwSize;
    }
    else
    {
        DBG(D_WARN, (L"%s: ReadBlock(%d) failed: %d\n", FULL_FUNC_NAME, _nBlkIdx, dwRet));
        pHdr->dwOffset = -1;
        bEndOfBlk = true;
    }

    if ( bEndOfBlk )
    {
        _pSeedTarget->CloseBlock(_dwBlkId);
        _dwBlkId = INVALID_ST_BLK_ID;

        ++_nBlkIdx;
    }

    PutSendData(pbi, false);
}
