/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "TimerCenter.h"
#include "SeedTarget.h"
#include "LP2PDownloadHandler.h"
#include "LP2PMCastRecver.h"
#include "LP2PSession.h"


typedef enum _LPS_SIGNAL
{
    SIG_TIMER = 0,

    SIG_REQ_CONNECT,

    LPS_SIG_NUM
}
LPS_SIGNAL;


LP2PSession::LP2PSession()
    : Endpoint()
    , _pObserver(NULL)
    , _dwCookie(0)
    , _dwTimerId(INVALID_TIMER_ID)
    , _nReqN4SCnt(0)
    , _bReported(false)
    , _dwErrorCode(ERROR_SUCCESS)
    , _pSeedTarget(NULL)
    , _pBitMask(NULL)
    , _pPeerList(NULL)
    , _pHandlerList(NULL)
#ifdef CONFIG_MCAST_SUPPORT
    , _bMCastValid(false)
    , _dwMCastConnId(INVALID_EP_CONN_ID)
    , _pMCastHandler(NULL)
#endif
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_init(&_hAddrUsedLock, NULL);

    memset(&_MCastAddr, 0, sizeof(SOCK_ADDR));
}

LP2PSession::~LP2PSession()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(!_pPeerList);
    assert(!_pHandlerList);

    assert(_AddrUsedList.empty());

    assert(!_pSeedTarget);
    assert(!_pBitMask);

    pthread_mutex_destroy(&_hAddrUsedLock);
}

DWORD LP2PSession::Load(SeedTarget *pSeedTarget, UINT nMaxPeerNum, LP2PSessionObserver *pObserver, DWORD dwCookie)
{
    DWORD  dwRet;
    UINT  nDownloadNum;
    const UINT  nExTaskNum = 0;
    LP2PDownloadHandler  *pHandler;

    DBG(D_INFO, (L"%s: seed(%hs)\n", FULL_FUNC_NAME, pSeedTarget->GetName()));

    nDownloadNum = pSeedTarget->GetMaxWriteNum();

    dwRet = Endpoint::Load(nDownloadNum, nExTaskNum, LPS_SIG_NUM);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: load endpoint failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    dwRet = TimerCenter::Instance()->CreateTimer(TimerFunc, (HANDLE)this, _dwTimerId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: CreateTimer failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    _pPeerList = new(std::nothrow) RingBuffer<SOCK_ADDR>(nMaxPeerNum, false);
    if ( !_pPeerList )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_ERROR, (L"%s: alloc _pPeerList failed\n", FULL_FUNC_NAME));
        goto err3;
    }

    // "+1", see ProcessConnect()
    // no need "+1" again for mcast, because mcast only or 1+1 mode at the same time
    _pHandlerList = new(std::nothrow) RingBuffer<EpProtocolHandler*>(nDownloadNum + 1, true);
    if ( !_pHandlerList )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_ERROR, (L"%s: alloc _pHandlerList failed\n", FULL_FUNC_NAME));
        goto err4;
    }

    for ( UINT i = 0; i < nDownloadNum; ++i )
    {
        pHandler = new(std::nothrow) LP2PDownloadHandler(pSeedTarget);
        if ( !pHandler )
        {
            break;
        }
        _pHandlerList->PushAtBack(pHandler);
    }

    if ( _pHandlerList->IsEmpty() )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_ERROR, (L"%s: _pHandlerList empty\n", FULL_FUNC_NAME));
        goto err5;
    }

#ifdef CONFIG_MCAST_SUPPORT
    _pMCastHandler = new(std::nothrow) LP2PMCastRecver(pSeedTarget);
    if ( !_pMCastHandler )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_ERROR, (L"%s: alloc _pMCastHandler failed\n", FULL_FUNC_NAME));
        goto err6;
    }
#endif

    dwRet = TimerCenter::Instance()->SetTimer(_dwTimerId, 50);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: SetTimer failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err7;
    }

    _pObserver = pObserver;
    _dwCookie = dwCookie;

    pSeedTarget->AddRef();
    _pSeedTarget = pSeedTarget;
    _pBitMask = pSeedTarget->GetProgressMask();

    return ERROR_SUCCESS;

err7:
#ifdef CONFIG_MCAST_SUPPORT
    delete _pMCastHandler;
    _pMCastHandler = NULL;

err6:
#endif
    while ( !_pHandlerList->IsEmpty() )
    {
        delete _pHandlerList->PopAtFront();
    }

err5:
    delete _pHandlerList;
    _pHandlerList = NULL;

err4:
    delete _pPeerList;
    _pPeerList = NULL;

err3:
    TimerCenter::Instance()->DestroyTimer(_dwTimerId);
    _dwTimerId = INVALID_TIMER_ID;

err2:
    Endpoint::Unload();

err1:
    return dwRet;
}

void LP2PSession::Unload()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( _dwTimerId != INVALID_TIMER_ID )
    {
        TimerCenter::Instance()->DestroyTimer(_dwTimerId);
        _dwTimerId = INVALID_TIMER_ID;
    }

    Endpoint::Unload();

#ifdef CONFIG_MCAST_SUPPORT
    assert(!_bMCastValid);

    memset(&_MCastAddr, 0, sizeof(SOCK_ADDR));

    assert(_dwMCastConnId == INVALID_EP_CONN_ID);
    if ( _pMCastHandler )
    {
        delete _pMCastHandler;
        _pMCastHandler = NULL;
    }
#endif

    if ( _pHandlerList )
    {
        while ( !_pHandlerList->IsEmpty() )
        {
            delete _pHandlerList->PopAtFront();
        }

        delete _pHandlerList;
        _pHandlerList = NULL;
    }

    if ( _pPeerList )
    {
        delete _pPeerList;
        _pPeerList = NULL;
    }

    if ( _pSeedTarget )
    {
        _pBitMask = NULL;

        _pSeedTarget->Release();
        _pSeedTarget = NULL;
    }

    _bReported = false;
    _dwErrorCode = ERROR_SUCCESS;

    _nReqN4SCnt = 0;

    _pObserver = NULL;
    _dwCookie = 0;
}

void LP2PSession::AddNodes4Seed(const PNODE4SEED_INFO pN4SArray, UINT nN4SCount, const PSOCK_ADDR pLocalAddr)
{
    QWORD  qwMaskAnd, qwMaskOr, qwReversed, qwRemote;
    SOCK_ADDR  *pPeerAddr;
    UINT  i, nCount;

    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, nN4SCount));

    assert(_pBitMask);
    _pBitMask->GetCompactMask(qwMaskAnd, qwMaskOr);
    qwReversed = ~qwMaskAnd;

    for ( i = 0, nCount = 0; i < nN4SCount; ++i )
    {
        pPeerAddr = &pN4SArray[i].NodeAddr;
        if ( *pLocalAddr == *pPeerAddr )
        {
            continue;
        }

        ++nCount;

        if ( IsMCastAddr(pPeerAddr) )
        {
#ifdef CONFIG_MCAST_SUPPORT
            // mcast server send done packet,
            // mcast client OnDisconnect, then ReqEnumN4S,
            // mcast client get the same mcast addr again,
            // mcast server remove the mcast addr from the tracker.
            if ( (!_bMCastValid) && (*pPeerAddr != _MCastAddr) )
            {
                DBG(D_INFO, (L"%s: mcast\n", FULL_FUNC_NAME));
                _MCastAddr = *pPeerAddr;
                MemoryBarrier();
                _bMCastValid = true;
            }
            // add remain addrs to prevent EnumN4S@ProcessTimer during 1+1 mode
#endif
            continue;
        }

        if ( _pPeerList->IsFull() )
        {
            break;
        }

        qwRemote = pN4SArray[i].qwProgressMask;
        if ( (qwMaskAnd ^ qwRemote) && (qwReversed & qwRemote) )
        {
            _pPeerList->PushAtBack(*pPeerAddr);
        }
    }

    if ( nCount )
    {
        _nReqN4SCnt = 0;
    }
    else
    {
        DBG(D_INFO, (L"%s: no peer\n", FULL_FUNC_NAME));
    }

    RequestEpWork(SIG_REQ_CONNECT);
}

void LP2PSession::OnListenerEnded(DWORD hListener)
{
    DBG(D_INFO, (L"%s: %x\n", FULL_FUNC_NAME, hListener));

    assert(0);
}

EpProtocolHandler *LP2PSession::OnNewHandler(DWORD hListener, Connection *pConnection, DWORD dwConnId)
{
    DBG(D_INFO, (L"%s: %x\n", FULL_FUNC_NAME, hListener));

    assert(0);

    return NULL;
}

void LP2PSession::OnAsyncConnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_INFO, (L"%s: %#.8x\n", FULL_FUNC_NAME, dwConnId));
}

void LP2PSession::OnDisconnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_INFO, (L"%s: %#.8x\n", FULL_FUNC_NAME, dwConnId));

#ifdef CONFIG_MCAST_SUPPORT
    if ( dwConnId == _dwMCastConnId )
    {
        DBG(D_INFO, (L"%s: mcast\n", FULL_FUNC_NAME));
        _bMCastValid = false;
        MemoryBarrier();
        _dwMCastConnId = INVALID_EP_CONN_ID;
    }
    else
#endif
    {
        pthread_mutex_lock(&_hAddrUsedLock);
        assert(!_AddrUsedList.empty());
        _AddrUsedList.remove(dynamic_cast<LP2PDownloadHandler*>(pHandler)->GetPeerAddr());
        pthread_mutex_unlock(&_hAddrUsedLock);

        // use internal lock
        _pHandlerList->PushAtBackSafe(pHandler);
    }

    // no need to sync with Unload(),
    // the sequence is:
    // normal OnDisconnected() be called
    // all work threads quitted
    // the rest conn be called OnDisconnected(),
    // RequestEpWork() will has no effect
    // release _pEvent
    RequestEpWork(SIG_REQ_CONNECT);
}

void LP2PSession::OnConnEvent(DWORD dwConnId, EpProtocolHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody)
{
    DBG(D_INFO, (L"%s: connid(%#.8x), code(%d)\n", FULL_FUNC_NAME, dwConnId, pEvtHdr->nCode));

    switch ( pEvtHdr->nCode )
    {
        case C_LP2P_ERROR_DISK_FULL:
        {
            DBG(D_ERROR, (L"%s: disk is full\n", FULL_FUNC_NAME));
            _dwErrorCode = ERROR_DISK_FULL;
            RequestEpWork(SIG_REQ_CONNECT);
            break;
        }

        default:
            DBG(D_WARN, (L"%s: unknown code(%x)\n", FULL_FUNC_NAME, pEvtHdr->nCode));
            assert(0);
            break;
    }
}

void LP2PSession::ProcessEpWork(UINT nSigIdx)
{
    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nSigIdx));

    switch ( nSigIdx )
    {
        case SIG_TIMER:
            ProcessTimer();
            break;

        case SIG_REQ_CONNECT:
            ProcessConnect();
            break;

        default:
            DBG(D_WARN, (L"%s: unknown sig(%d)\n", FULL_FUNC_NAME, nSigIdx));
            assert(0);
            break;
    }
}

void LP2PSession::ProcessTimer()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _nReqN4SCnt >= LP2P_SESSION_RETRYNUM )
    {
        DBG(D_WARN, (L"%s: tracker offline or peer not exist\n", FULL_FUNC_NAME));
        _dwErrorCode = ERROR_NOT_READY;
        RequestEpWork(SIG_REQ_CONNECT);
        return;
    }

#ifdef CONFIG_MCAST_SUPPORT
    if ( _dwMCastConnId != INVALID_EP_CONN_ID )
    {
        QueueConnCmd(_dwMCastConnId, C_CONN_MCAST_TIMER, NULL);
        return;
    }
#endif

    if ( _pPeerList->IsEmpty() && (!_pHandlerList->IsEmpty()) && (!_pSeedTarget->IsAllBusy()) )
    {
        DBG(D_INFO, (L"%s: req enum N4S\n", FULL_FUNC_NAME));
        ++_nReqN4SCnt;
        _pObserver->OnEnumNodes4Seed(_dwCookie);
    }
}

void LP2PSession::ProcessConnect()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _pSeedTarget->IsFinished() || (_dwErrorCode != ERROR_SUCCESS) )
    {
        // see issues_99stuck.txt
        if ( !_bReported )
        {
            DBG(D_INFO, (L"%s: finished or failed\n", FULL_FUNC_NAME));
            _bReported = true;
            _pObserver->OnFinished(_dwCookie, _dwErrorCode);
        }
        return;
    }

#ifdef CONFIG_MCAST_SUPPORT
    if ( _bMCastValid )
    {
        if ( _dwMCastConnId == INVALID_EP_CONN_ID )
        {
            DBG(D_LOG, (L"%s: mcast addr: %#.8x:%d\n", FULL_FUNC_NAME, _MCastAddr.IPv4Addr.dwIP, _MCastAddr.IPv4Addr.wPort));
            ConnectIn(&_MCastAddr, _pMCastHandler, _dwMCastConnId);
        }

        return;
    }
#endif

    ConnectToPeers();
}

void LP2PSession::ConnectToPeers()
{
    DWORD  dwRet;
    DWORD  dwConnId;
    SOCK_ADDR  LocalAddr, PeerAddr;
    LP2PDownloadHandler  *pHandler;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    LocalAddr.bType = ADDR_TYPE_IPV4;
    LocalAddr.IPv4Addr.bStream = true;

    while ( (!_pPeerList->IsEmpty()) && (!_pHandlerList->IsEmpty()) )
    {
        PeerAddr = _pPeerList->PopAtFront();
        if ( find(_AddrUsedList.begin(), _AddrUsedList.end(), PeerAddr) != _AddrUsedList.end() )
        {
            DBG(D_INFO, (L"%s: skip(%#.8x/%d)\n", FULL_FUNC_NAME, PeerAddr.IPv4Addr.dwIP, ntohs(PeerAddr.IPv4Addr.wPort)));
            continue;
        }

        pHandler = dynamic_cast<LP2PDownloadHandler*>(_pHandlerList->Front());

        pthread_mutex_lock(&_hAddrUsedLock);
        _AddrUsedList.push_back(PeerAddr);
        pthread_mutex_unlock(&_hAddrUsedLock);

        pHandler->SetPeerAddr(&PeerAddr);

        LocalAddr.IPv4Addr.dwIP = INADDR_ANY;
        LocalAddr.IPv4Addr.wPort = htons(0);

        DBG(D_INFO, (L"%s: ConnectOut(%#.8x/%d)...\n", FULL_FUNC_NAME, PeerAddr.IPv4Addr.dwIP, ntohs(PeerAddr.IPv4Addr.wPort)));
        dwRet = ConnectOut(&LocalAddr, &PeerAddr, pHandler, dwConnId);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: ConnectOut(%#.8x/%d) failed: %d\n", FULL_FUNC_NAME, PeerAddr.IPv4Addr.dwIP, ntohs(PeerAddr.IPv4Addr.wPort), dwRet));
            pthread_mutex_lock(&_hAddrUsedLock);
            _AddrUsedList.pop_back();
            pthread_mutex_unlock(&_hAddrUsedLock);
            continue;
        }

        // PushAtBack()@OnDisconnected() may occur before Pop(),
        // so _pHandlerList may be full when PushAtBack(),
        // set _pHandlerList's size+1 to avoid this
        _pHandlerList->Pop();
    }
}

void LP2PSession::TimerFunc(HANDLE hArg)
{
    LP2PSession  *pThis = static_cast<LP2PSession*>(hArg);

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pThis->RequestEpWork(SIG_TIMER);

    TimerCenter::Instance()->SetTimer(pThis->_dwTimerId, LP2P_SESSION_TIMER);
}

