/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "LP2PUploadHandler.h"


LP2PUploadHandler::LP2PUploadHandler()
    : EpPDUHandler(LP2P_UPDOWN_CTRLQSIZE, LP2P_UPDOWN_CTRLBSIZE,
                   LP2P_UPDOWN_DATAQSIZE, LP2P_UPDOWN_DATABSIZE,
                   LP2P_UPDOWN_EVTQSIZE, LP2P_UPDOWN_EVTBSIZE)
    , _pSeedTarget(NULL)
    , _pBitMask(NULL)
    , _dwCookie(INVALID_COOKIE)
    , _dwSessionId(INVALID_ST_SESS_ID)
    , _dwBlkId(INVALID_ST_BLK_ID)
#ifdef CONFIG_ST_BLK_COMPACT
    , _bEndOfBlk(false)
#endif
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

LP2PUploadHandler::~LP2PUploadHandler()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_ST_BLK_ID);
    assert(_dwSessionId == INVALID_ST_SESS_ID);
    assert(!_pBitMask);
    assert(!_pSeedTarget);
}

DWORD LP2PUploadHandler::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    _dwCookie = INVALID_COOKIE;

#ifdef CONFIG_ST_BLK_COMPACT
    dwRet = _CompAlg.Init(COMPRESS_FASTEST);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init compressor failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    DBG(D_INFO, (L"%s: alloc bytes[%d] for read buf\n", FULL_FUNC_NAME, COMPACT_CLUSTER_SIZE));
    _ReadBuf.Alloc(COMPACT_CLUSTER_SIZE);
    _CompBuf.Alloc(_CompAlg.GetBound(COMPACT_CLUSTER_SIZE));
#endif

    pConnection->SetSendBufSize(LP2P_SOCKBUF_SIZE);

    dwRet = EpPDUHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
#ifdef CONFIG_ST_BLK_COMPACT
    _CompAlg.Fini();

err1:
#endif
    return dwRet;
}

void LP2PUploadHandler::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::Unbind();

#ifdef CONFIG_ST_BLK_COMPACT
    _CompAlg.Fini();
#endif

    if ( _pSeedTarget )
    {
        assert(_dwSessionId != INVALID_ST_SESS_ID);

        if ( _dwBlkId != INVALID_ST_BLK_ID )
        {
            _pSeedTarget->CloseBlock(_dwBlkId);
            _dwBlkId = INVALID_ST_BLK_ID;
        }

        _pSeedTarget->EndSession(_dwSessionId);
        _dwSessionId = INVALID_ST_SESS_ID;

        _pBitMask = NULL;

        _pSeedTarget->Release();
        _pSeedTarget = NULL;
    }
}

void LP2PUploadHandler::PrepareSend()
{
    DWORD  dwRet;
#ifndef CONFIG_ST_BLK_COMPACT
    DWORD  dwSize;
#endif
    bool  bEndOfBlk;
    BufferItem  *pbi;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    // PutSendData(pbi, bEndOfBlk) will guarantee PrepareSend() will not be called after EndOfBlk
    assert(_dwBlkId != INVALID_ST_BLK_ID);

    while ( true )
    {
        pbi = GetSendBuf();
        if ( !pbi )
        {
            break;
        }

#ifdef CONFIG_ST_BLK_COMPACT
        // see issues_zp2p.txt
        dwRet = DeflateBlock(pbi, bEndOfBlk);
#else
        dwSize = pbi->tail_size();
        dwRet = _pSeedTarget->ReadBlock(_dwBlkId, pbi->tail(), dwSize, bEndOfBlk);
        if ( dwRet == ERROR_SUCCESS )
        {
            pbi->Put(dwSize);
        }
#endif
        assert(dwRet == ERROR_SUCCESS);

        BuildPDU((bEndOfBlk ? C_SEEDTARGET_BLOCK_END : C_SEEDTARGET_READ_BLOCK), false, false, dwRet, pbi);
        PutSendData(pbi, bEndOfBlk);

        if ( (dwRet != ERROR_SUCCESS) || bEndOfBlk )
        {
            _pSeedTarget->CloseBlock(_dwBlkId);
            _dwBlkId = INVALID_ST_BLK_ID;
            break;
        }
    }
}

DWORD LP2PUploadHandler::GetCookie() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return _dwCookie;
}

void LP2PUploadHandler::Attach(DWORD dwResult, SeedTarget *pSeedTarget, DWORD dwCookie)
{
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _dwCookie = dwCookie;

    pbi = GetSendBuf();
    assert(pbi);

    if ( dwResult != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: failed: %d\n", FULL_FUNC_NAME, dwResult));
        goto out;
    }

    assert(_dwSessionId == INVALID_ST_SESS_ID);
    dwResult = pSeedTarget->BeginSession(true, _dwSessionId);
    if ( dwResult != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: BeginSession failed: %d\n", FULL_FUNC_NAME, dwResult));
        goto out;
    }

    assert(!_pSeedTarget);
    _pSeedTarget = pSeedTarget;

    assert(!_pBitMask);
    _pBitMask = pSeedTarget->GetProgressMask();

    pSeedTarget->AddRef();

    DBG(D_INFO, (L"%s: seed(%hs), cookie(%#.8x)\n", FULL_FUNC_NAME, pSeedTarget->GetName(), dwCookie));

out:
    BuildPDU(C_SEEDTARGET_OPEN, false, true, dwResult, pbi);
    PutSendData(pbi, true);
}

void LP2PUploadHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    switch ( bCode )
    {
        case C_SEEDTARGET_OPEN:
            assert(pbiBody->data_size() == sizeof(REQ_OPEN_SEEDTARGET));
            OnOpenSeedTarget((PREQ_OPEN_SEEDTARGET)pbiBody->data());
            break;

        case C_SEEDTARGET_GET_BITMASK:
            OnGetSeedTargetBitmask();
            break;

        case C_SEEDTARGET_READ_BLOCK:
            OnReadSeedTargetBlock(dwParam);
            break;

        default:
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, bCode));
            assert(0);
            break;
    }
}

void LP2PUploadHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

void LP2PUploadHandler::OnOpenSeedTarget(const PREQ_OPEN_SEEDTARGET pReqOpenST)
{
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _pSeedTarget )
    {
        DBG(D_WARN, (L"%s: already opened\n", FULL_FUNC_NAME));
        assert(0);
        return;
    }

    pbi = AllocEvent();
    if ( !pbi )
    {
        DBG(D_WARN, (L"%s: dup request\n", FULL_FUNC_NAME));
        assert(0);
        return;
    }

    memcpy(pbi->Put(sizeof(SHA1_VAL)), &pReqOpenST->SHA1ValSeed, sizeof(SHA1_VAL));
    EpBuildMsg(C_LP2P_SEEDTARGET_OPEN, 0, 0, pbi);
    PutEvent(pbi);
}

void LP2PUploadHandler::OnGetSeedTargetBitmask()
{
    UINT  nBytes;
    BufferItem  *pbi;
    PRET_GET_SEEDTARGET_BITMASK  pGetSTBits;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( !_pSeedTarget )
    {
        DBG(D_WARN, (L"%s: not opened\n", FULL_FUNC_NAME));
        assert(0);
        return;
    }

    assert(_pBitMask);
    nBytes = _pBitMask->GetBytesCount();

    pbi = GetSendBuf();
    assert(pbi);

    pGetSTBits = (PRET_GET_SEEDTARGET_BITMASK)pbi->Put(sizeof(RET_GET_SEEDTARGET_BITMASK));
    pGetSTBits->dwBitsCnt = _pBitMask->GetBitsCount();
    _pBitMask->GetFullMask(pbi->Put(nBytes), nBytes);

    BuildPDU(C_SEEDTARGET_GET_BITMASK, false, true, ERROR_SUCCESS, pbi);
    PutSendData(pbi, true);
}

void LP2PUploadHandler::OnReadSeedTargetBlock(DWORD dwBlkIdx)
{
    DWORD  dwRet;
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( !_pSeedTarget )
    {
        DBG(D_WARN, (L"%s: not opened\n", FULL_FUNC_NAME));
        assert(0);
        return;
    }

    if ( _dwBlkId != INVALID_ST_BLK_ID )
    {
        DBG(D_WARN, (L"%s: dup request\n", FULL_FUNC_NAME));
        assert(0);
        return;
    }

    assert(_dwSessionId != INVALID_ST_SESS_ID);
    dwRet = _pSeedTarget->OpenBlock(_dwSessionId, dwBlkIdx, _dwBlkId);
    if ( dwRet == ERROR_SUCCESS )
    {
#ifdef CONFIG_ST_BLK_COMPACT
        _bEndOfBlk = false;
        _CompBuf.Reset();
#endif
        ReqSendProc(true);
        return;
    }

    DBG(D_WARN, (L"%s: OpenBlock(%d) failed: %d\n", FULL_FUNC_NAME, dwBlkIdx, dwRet));
    assert(0);

    pbi = GetSendBuf();
    assert(pbi);
    BuildPDU(C_SEEDTARGET_READ_BLOCK, false, true, dwRet, pbi);
    PutSendData(pbi, true);
}

#ifdef CONFIG_ST_BLK_COMPACT
DWORD LP2PUploadHandler::DeflateBlock(BufferItem *pbi, bool &rbLast)
{
    DWORD  dwRet = ERROR_SUCCESS;
    UINT  nBytes;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rbLast = false;

    while ( pbi->tail_size() )
    {
        dwRet = DeflateCluster();
        if ( dwRet != ERROR_SUCCESS )
        {
            break;
        }

        nBytes = _CompBuf.data_size();
        if ( nBytes > pbi->tail_size() )
        {
            nBytes = pbi->tail_size();
        }
        memcpy(pbi->Put(nBytes), _CompBuf.data(), nBytes);
        _CompBuf.Pull(nBytes);

        if ( _bEndOfBlk && (!_CompBuf.data_size()) )
        {
            rbLast = true;
            break;
        }
    }

    return dwRet;
}

DWORD LP2PUploadHandler::DeflateCluster()
{
    DWORD  dwRet;
    DWORD  dwRead;
    PBYTE  pSrc;
    UINT  nSrc, nDst;
//    bool  bAgain;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _CompBuf.data_size() )
    {
        return ERROR_SUCCESS;
    }

    _CompBuf.Reset();

    dwRead = _ReadBuf.tail_size();
    dwRet = _pSeedTarget->ReadBlock(_dwBlkId, _ReadBuf.tail(), dwRead, _bEndOfBlk);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ReadBlock failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    pSrc = _ReadBuf.tail();
    nSrc = dwRead;
    nDst = _CompBuf.tail_size();
    // see issues_lz4_block.txt
//    dwRet = _CompAlg.ProcessStream(pSrc, nSrc, _CompBuf.tail(), nDst, bAgain, _bEndOfBlk);
    dwRet = _CompAlg.ProcessBlock(pSrc, nSrc, _CompBuf.tail(), nDst, false);//_bEndOfBlk);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ProcessStream failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    _CompBuf.Put(nDst);

    return ERROR_SUCCESS;
}
#endif
