/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "SeedAgentHandler.h"


typedef enum _ST_FLAG_BITS
{
    F_ST_NONE                 = 0x00,
    F_ST_ADD                  = 0x01,
    F_ST_REMOVE               = 0x02,
    F_ST_UPDATE               = 0x04,
    F_ST_ENUM                 = 0x08
}
ST_FLAG_BITS;


SeedAgentHandler::SeedAgentHandler(const char *szNodeName, const PSOCK_ADDR pNodeAddr, UINT nMaxSTNum)
    : EpPDUHandler(SEED_TRACK_NETQSIZE, SEED_TRACK_REPBSIZE,
                   SEED_TRACK_NETQSIZE, SEED_TRACK_REQBSIZE,
                   SEED_TRACK_EVTQSIZE, SEED_TRACK_EVTBSIZE)
    , _bBound(false)
    , _bBusy(false)
    , _dwNodeId(INVALID_NODE_ID)
    , _nCurST(INVALID_PCOND_IDX)
    , _STEvent(nMaxSTNum)
    , _STList(nMaxSTNum)
    , _STListMate(false)
    , _dwSTMagic(0)
{
    PSA_SEEDTARGET_REC  pSTRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_init(&_hLock, NULL);

    strncpy(_NodeInfo.name, szNodeName, MAX_NODE_NAME_LEN);
    memcpy(&_NodeInfo.addr, pNodeAddr, sizeof(SOCK_ADDR));

    for ( UINT nIdx = 0, nSize = (UINT)_STList.size(); nIdx < nSize; ++nIdx )
    {
        pSTRec = &_STList.at(nIdx);
        DO_IF_DBG(pSTRec->bUsed = false);
        pSTRec->bValid = false;
    }
    _STListMate.Init(nMaxSTNum);
}

SeedAgentHandler::~SeedAgentHandler()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    // free idx only in OnRetRemoveSeed() or DoTrackWork(), if disconnected?
#ifndef NDEBUG
    for ( UINT nIdx = 0, nSize = (UINT)_STList.size(); nIdx < nSize; ++nIdx )
    {
//        assert(!_STList.at(nIdx).bUsed);
        assert((!_STList.at(nIdx).bValid) || (!_STList.at(nIdx).pSeedTarget));
    }
#endif
//    assert(_STListMate.IsFull());
    _STListMate.Fini();

    pthread_mutex_destroy(&_hLock);
}

DWORD SeedAgentHandler::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;
    UINT  nIdx, nSize;
    PSA_SEEDTARGET_REC  pSTRec;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    dwRet = EpPDUHandler::Bind(pConnection, pEvent, nSigIdx);

    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: Bind failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    // bUsed only be clear under conn thread, no need lock
    for ( nIdx = 0, nSize = (UINT)_STList.size(); nIdx < nSize; ++nIdx )
    {
        pSTRec = &_STList.at(nIdx);
        if ( pSTRec->bValid )
        {
            _STEvent.Signal(nIdx);
        }
    }

    return ERROR_SUCCESS;
}

void SeedAgentHandler::Unbind()
{
    UINT  nIdx, nSize;
    PSA_SEEDTARGET_REC  pSTRec;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_lock(&_hLock);
    _bBound = false;
    pthread_mutex_unlock(&_hLock);

    EpPDUHandler::Unbind();

    // _dwNodeId only be set under conn thread
    _dwNodeId = INVALID_NODE_ID;

    pthread_mutex_lock(&_hLock);
    for ( nIdx = 0, nSize = (UINT)_STList.size(); nIdx < nSize; ++nIdx )
    {
        pSTRec = &_STList.at(nIdx);
        if ( pSTRec->bValid )
        {
            // keep the error state
//            pSTRec->bError = false;
            pSTRec->dwFlags |= (F_ST_ADD | F_ST_UPDATE);
            pSTRec->dwN4SId = INVALID_N4S_ID;

            // see issues_pcond_reset.txt
//           _STEvent.Reset(nIdx, 0);
        }
    }
    pthread_mutex_unlock(&_hLock);

    // _nCurST is only be set under conn thread
    _nCurST = INVALID_PCOND_IDX;

    // _bBusy is only be set under conn thread
    _bBusy = false;
}

void SeedAgentHandler::OnAsyncConnected(Connection *pConnection)
{
    SOCK_ADDR  LocalAddr;

    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::OnAsyncConnected(pConnection);

    if ( _NodeInfo.addr.IPv4Addr.dwIP == INADDR_ANY )
    {
        _pConnection->GetLocalAddr(&LocalAddr);
        _NodeInfo.addr.IPv4Addr.dwIP = LocalAddr.IPv4Addr.dwIP;
    }

    // DO NOT put in Bind(), the wrong sequence is:
    // Bind(_bBound=true) -> EnumNodes4Seed() -> PrepareSend() -> TrackNode() -> OnAsyncConnected()
    // the NodeInfo's IP is zero.
    // or move GetLocalAddr to Bind()
    pthread_mutex_lock(&_hLock);
    _bBound = true;
    pthread_mutex_unlock(&_hLock);

    ReqSendProc(true);
}

void SeedAgentHandler::PrepareSend()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_bBusy )
    {
        DoTrackWork();
    }
}

void SeedAgentHandler::ReleaseEvent(BufferItem *pbiEvt)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::ReleaseEvent(pbiEvt);

    if ( _dwNodeId == INVALID_NODE_ID )
    {
        DBG(D_WARN, (L"%s: will not do track anymore\n", FULL_FUNC_NAME));
        return;
    }

    _bBusy = false;

    DoTrackWork();
}

DWORD SeedAgentHandler::AddSeedTarget(SeedTarget *pSeedTarget, DWORD dwCookie, DWORD &rdwSTId)
{
    DWORD  dwRet;
    UINT  nSTIdx;
    DWORD  dwMagic;
    PSA_SEEDTARGET_REC  pSTRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    // _STListMate not enable internal lock
    pthread_mutex_lock(&_hLock);
    dwRet = _STListMate.GetFreeIndex(nSTIdx);
    pthread_mutex_unlock(&_hLock);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: GetFreeIndex() failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err;
    }

    pSTRec = &_STList.at(nSTIdx);
    assert(!pSTRec->bUsed);
    DO_IF_DBG(pSTRec->bUsed = true);

    dwMagic = InterlockedIncrement((LONG*)&_dwSTMagic);

    pSTRec->wMagic = ADJUST_WMAGICNUM(dwMagic);
    pSTRec->bError = false;
    pSTRec->dwFlags = F_ST_ADD;
    pSTRec->pSeedTarget = pSeedTarget;
    pSTRec->dwCookie = dwCookie;
    pSTRec->dwN4SId = INVALID_N4S_ID;
    memset(&pSTRec->SeedState, 0, sizeof(SEED_STATE));

    // OnPostUnbind() will check bUsed, but all fields are ready now
    pSTRec->bValid = true;

    rdwSTId = MAKE_DWID(nSTIdx, pSTRec->wMagic);

    pthread_mutex_lock(&_hLock);
    ReqTrackProc(nSTIdx);
    pthread_mutex_unlock(&_hLock);

    pSeedTarget->RegisterObserver(this, rdwSTId);

    pSeedTarget->AddRef();

    DBG(D_INFO, (L"%s: name(%hs), idx(%d)\n", FULL_FUNC_NAME, pSeedTarget->GetName(), nSTIdx));

    return ERROR_SUCCESS;

err:
    rdwSTId = INVALID_LP2P_ST_ID;

    return dwRet;
}

DWORD SeedAgentHandler::RemoveSeedTarget(DWORD dwSTId)
{
    UINT  nSTIdx;
    WORD  wMagic;
    PSA_SEEDTARGET_REC  pSTRec;
    SeedTarget  *pSeedTarget;

    nSTIdx = WIDX_FROM_DWID(dwSTId);
    wMagic = WMAGICNUM_FROM_DWID(dwSTId);

    DBG(D_INFO, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nSTIdx, wMagic));

    if ( nSTIdx >= _STList.size() )
    {
        DBG(D_WARN, (L"%s: invalid ST id\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_INVALID_PARAMETER;
    }

    pSTRec = &_STList.at(nSTIdx);

    if ( wMagic != pSTRec->wMagic )
    {
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nSTIdx));
        assert(0);
        return ERROR_NOT_FOUND;
    }
    assert(pSTRec->bUsed && pSTRec->bValid);

    pSeedTarget = pSTRec->pSeedTarget;
    assert(pSeedTarget);
    pSeedTarget->UnregisterObserver(this);

    // invalid the wMagic
    // OnStateUpdated() will check wMagic, so put it after UnregisterObserver()
    // see issues_magic.txt
    pSTRec->wMagic = ~wMagic;

    pthread_mutex_lock(&_hLock);
    pSTRec->dwFlags |= F_ST_REMOVE;
    pSTRec->pSeedTarget = NULL;
    ReqTrackProc(nSTIdx);
    pthread_mutex_unlock(&_hLock);

    pSeedTarget->Release();

    return ERROR_SUCCESS;
}

DWORD SeedAgentHandler::EnumNodes4Seed(DWORD dwSTId)
{
    UINT  nSTIdx;
    WORD  wMagic;
    PSA_SEEDTARGET_REC  pSTRec;

    nSTIdx = WIDX_FROM_DWID(dwSTId);
    wMagic = WMAGICNUM_FROM_DWID(dwSTId);

    DBG(D_VERB, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nSTIdx, wMagic));

    if ( nSTIdx >= _STList.size() )
    {
        DBG(D_WARN, (L"%s: invalid ST id\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_INVALID_PARAMETER;
    }

    pSTRec = &_STList.at(nSTIdx);

    if ( wMagic != pSTRec->wMagic )
    {
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, nSTIdx));
        assert(0);
        return ERROR_NOT_FOUND;
    }
    assert(pSTRec->bUsed && pSTRec->bValid);

    // no need serialization and synchronization, DoTrackWork() will check again
    if ( pSTRec->bError )
    {
        DBG(D_WARN, (L"%s: idx(%d) has error\n", FULL_FUNC_NAME, nSTIdx));
        return ERROR_ACCESS_DENIED;
    }

    pthread_mutex_lock(&_hLock);
    pSTRec->dwFlags |= F_ST_ENUM;
    ReqTrackProc(nSTIdx);
    pthread_mutex_unlock(&_hLock);

    return ERROR_SUCCESS;
}

void SeedAgentHandler::OnStateUpdated(SeedTarget *pSeedTarget, DWORD dwCookie, UINT nDegree, QWORD qwMaskAnd, QWORD qwMaskOr)
{
    UINT  nSTIdx;
    WORD  wMagic;
    PSA_SEEDTARGET_REC  pSTRec;

    nSTIdx = WIDX_FROM_DWID(dwCookie);
    wMagic = WMAGICNUM_FROM_DWID(dwCookie);

    DBG(D_VERB, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, nSTIdx, wMagic));

    assert(nSTIdx < _STList.size());

    pSTRec = &_STList.at(nSTIdx);
    assert(pSTRec->bUsed && pSTRec->bValid);

    assert(wMagic == pSTRec->wMagic);

    // no need serialization and synchronization, DoTrackWork() will check again
    if ( pSTRec->bError )
    {
        DBG(D_WARN, (L"%s: idx(%d) has error\n", FULL_FUNC_NAME, nSTIdx));
        return;
    }

    // because do UnregisterObserver() then set pSeedTarget to NULL
    assert(pSTRec->pSeedTarget == pSeedTarget);

    DBG(D_VERB, (L"%s: idx(%d), degree(%d%%), %#.16llx, %#.16llx\n", FULL_FUNC_NAME, nSTIdx, nDegree, qwMaskAnd, qwMaskOr));

    pthread_mutex_lock(&_hLock);
    pSTRec->SeedState.qwProgressMask = qwMaskOr;
    pSTRec->SeedState.bProgressDegree = nDegree;
    pSTRec->dwFlags |= F_ST_UPDATE;
    ReqTrackProc(nSTIdx);
    pthread_mutex_unlock(&_hLock);
}

void SeedAgentHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    bool  bFinish = true;

    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    assert(!bReq);

    switch ( bCode )
    {
        case C_ST_REG_NODE:
            assert(pbiBody->data_size() == sizeof(RET_REG_NODE));
            bFinish = OnRetRegisterNode(dwParam, (PRET_REG_NODE)pbiBody->data());
            break;

        case C_ST_ADD_N4S:
            assert(pbiBody->data_size() == sizeof(RET_ADD_NODE4SEED));
            bFinish = OnRetAddSeed(dwParam, (PRET_ADD_NODE4SEED)pbiBody->data());
            break;

        case C_ST_REMOVE_N4S:
            assert(!pbiBody->data_size());
            OnRetRemoveSeed(dwParam);
            break;

        case C_ST_UPDATE_N4S_STATE:
            assert(!pbiBody->data_size());
            bFinish = OnRetUpdateSeedState(dwParam);
            break;

        case C_ST_ENUM_NODES4SEED:
            bFinish = OnRetEnumNodes4Seed(dwParam, (PRET_ENUM_NODE4SEED)pbiBody->data());
            break;

        default:
            DBG(D_WARN, (L"%s: unknown code(%d)\n", FULL_FUNC_NAME, bCode));
            assert(0);
            break;
    }

    if ( bFinish )
    {
        _bBusy = false;

        DoTrackWork();
    }
}

void SeedAgentHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

bool SeedAgentHandler::OnRetRegisterNode(DWORD dwResult, PRET_REG_NODE pRetRegNode)
{
    BufferItem  *pbi;

    DBG(D_LOG, (L"%s: ret=%d, nid=%x\n", FULL_FUNC_NAME, dwResult, pRetRegNode->dwNodeId));

    assert(_dwNodeId == INVALID_NODE_ID);

    if ( dwResult == ERROR_SUCCESS )
    {
        _dwNodeId = pRetRegNode->dwNodeId;
    }

    // event is kicked by recv, NULL is impossible
    pbi = AllocEvent();
    assert(pbi);

    EpBuildMsg(C_LP2P_NODE_REG, 0, dwResult, pbi);
    PutEvent(pbi);

    return false;
}

bool SeedAgentHandler::OnRetAddSeed(DWORD dwResult, PRET_ADD_NODE4SEED pRetAddSeed)
{
    PSA_SEEDTARGET_REC  pSTRec;
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s: sidx(%d): ret=%d, sid=%x\n", FULL_FUNC_NAME, _nCurST, dwResult, pRetAddSeed->dwN4SId));

    assert(_nCurST != INVALID_PCOND_IDX);
    pSTRec = &_STList.at(_nCurST);
    assert(pSTRec->bUsed && pSTRec->bValid);

    if ( dwResult != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: sidx(%d) failed: ret=%d\n", FULL_FUNC_NAME, _nCurST, dwResult));
        pSTRec->bError = true;
    }
    else
    {
        DBG(D_LOG, (L"%s: sidx(%d) OK: sid=%x\n", FULL_FUNC_NAME, _nCurST, pRetAddSeed->dwN4SId));
        assert(pSTRec->dwN4SId == INVALID_N4S_ID);
        pSTRec->dwN4SId = pRetAddSeed->dwN4SId;
    }

    // event is kicked by recv, NULL is impossible
    pbi = AllocEvent();
    assert(pbi);

    EpBuildMsg(C_LP2P_SEED_ADD, pSTRec->dwCookie, dwResult, pbi);
    PutEvent(pbi);

    return false;
}

void SeedAgentHandler::OnRetRemoveSeed(DWORD dwResult)
{
    PSA_SEEDTARGET_REC  pSTRec;

    DBG(D_LOG, (L"%s: sidx(%d): ret=%d\n", FULL_FUNC_NAME, _nCurST, dwResult));

    assert(_nCurST != INVALID_PCOND_IDX);
    pSTRec = &_STList.at(_nCurST);
    assert(pSTRec->bUsed && pSTRec->bValid);

    _STEvent.Reset(_nCurST, 0);
    pSTRec->bValid = false;
    DO_IF_DBG(pSTRec->bUsed = false);
    // no need lock even disable internal lock
    _STListMate.PutFreeIndex(_nCurST);
    _nCurST = INVALID_PCOND_IDX;
}

bool SeedAgentHandler::OnRetUpdateSeedState(DWORD dwResult)
{
    PSA_SEEDTARGET_REC  pSTRec;
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s: sidx(%d): ret=%d\n", FULL_FUNC_NAME, _nCurST, dwResult));

    if ( dwResult == ERROR_SUCCESS )
    {
        return true;
    }

    assert(_nCurST != INVALID_PCOND_IDX);
    pSTRec = &_STList.at(_nCurST);
    assert(pSTRec->bUsed && pSTRec->bValid);

    DBG(D_WARN, (L"%s: sidx(%d) failed: ret=%d\n", FULL_FUNC_NAME, _nCurST, dwResult));
    pSTRec->bError = true;

    // event is kicked by recv, NULL is impossible
    pbi = AllocEvent();
    assert(pbi);

    EpBuildMsg(C_LP2P_SEED_STATE_UPDATE, pSTRec->dwCookie, dwResult, pbi);
    PutEvent(pbi);

    return false;
}

bool SeedAgentHandler::OnRetEnumNodes4Seed(DWORD dwResult, PRET_ENUM_NODE4SEED pRetEnumN4S)
{
    PSA_SEEDTARGET_REC  pSTRec;
    BufferItem  *pbi;
    PEVT_NODE4SEED_ENUM  pEvtN4SEnum;
    DWORD  dwSize;

    DBG(D_VERB, (L"%s: sidx(%d): ret=%d\n", FULL_FUNC_NAME, _nCurST, dwResult));

    assert(_nCurST != INVALID_PCOND_IDX);
    pSTRec = &_STList.at(_nCurST);
    assert(pSTRec->bUsed && pSTRec->bValid);

    // event is kicked by recv, NULL is impossible
    pbi = AllocEvent();
    assert(pbi);

    if ( dwResult == ERROR_SUCCESS )
    {
        DBG(D_LOG, (L"%s: sidx(%d) OK: count=%d\n", FULL_FUNC_NAME, _nCurST, pRetEnumN4S->dwCount));
        dwSize = sizeof(NODE4SEED_INFO) * pRetEnumN4S->dwCount;
        pEvtN4SEnum = (PEVT_NODE4SEED_ENUM)pbi->Put(sizeof(EVT_NODE4SEED_ENUM) - sizeof(NODE4SEED_INFO) + dwSize);
        pEvtN4SEnum->LocalAddr = _NodeInfo.addr;
        pEvtN4SEnum->nCount = (UINT)pRetEnumN4S->dwCount;
        memcpy(pEvtN4SEnum->N4SInfo, pRetEnumN4S->N4SInfo, dwSize);
    }

    EpBuildMsg(C_LP2P_N4S_ENUM, pSTRec->dwCookie, dwResult, pbi);
    PutEvent(pbi);

    return false;
}

void SeedAgentHandler::ReqTrackProc(UINT nIdx)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    // test _bBound MUST be with lock
    if ( _bBound )
    {
        _STEvent.Signal(nIdx);

        if ( !_bBusy )
        {
            // _bReqSendProc will be false once send out a packet.
            // but if ReqTrackProc() always be called during OnRetXXX() and DoTrackWork(),
            // and then schedule to PutSendData(pbi, true) set _bReqSendProc false,
            // send out RegNode PDU, then recv failed PDU,
            // then schedule to here: ReqSendProc(true) will trigger PrepareSend(),
            // then RegNode will do again.
            // to avoid this case, ProcessRecvPDU() reset _bBusy only if RegNode success.
            // add Event feature so move to ReleaseEvent()
            ReqSendProc(true);
        }
    }
}

void SeedAgentHandler::DoTrackWork()
{
    PSA_SEEDTARGET_REC  pSTRec;
    bool  bRemove;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(!_bBusy);

    if ( _dwNodeId == INVALID_NODE_ID )
    {
        // kicked by ReqSendProc@OnAsyncConnected,
        // will not consume the STSig
        TrackNode();
        return;
    }

    while ( true )
    {
        if ( _nCurST == INVALID_PCOND_IDX )
        {
            _nCurST = _STEvent.Wait(0, true);
            if ( _nCurST == INVALID_PCOND_IDX )
            {
                break;
            }
        }

        pSTRec = &_STList.at(_nCurST);
        assert(pSTRec->bUsed && pSTRec->bValid);

        if ( pSTRec->dwFlags == F_ST_NONE )
        {
            if ( _STEvent.Finish(_nCurST, 0) )
            {
                _nCurST = INVALID_PCOND_IDX;
            }
            continue;
        }

        if ( TrackSeed(pSTRec, bRemove) )
        {
            break;
        }

        _STEvent.Reset(_nCurST, 0);
        if ( bRemove )
        {
            pSTRec->bValid = false;
            DO_IF_DBG(pSTRec->bUsed = false);
            // no need lock even disable internal lock
            _STListMate.PutFreeIndex(_nCurST);
        }
        _nCurST = INVALID_PCOND_IDX;
    }
}

void SeedAgentHandler::TrackNode()
{
    BufferItem  *pbi;
    PNODE_INFO  pNodeInfo;

    DBG(D_INFO, (L"%s: %hs(%#.8x/%d)\n", FULL_FUNC_NAME, _NodeInfo.name, _NodeInfo.addr.IPv4Addr.dwIP, ntohs(_NodeInfo.addr.IPv4Addr.wPort)));

    assert(_dwNodeId == INVALID_NODE_ID);

    // send is kicked by recv or non-busy, NULL is impossible
    pbi = GetSendBuf();
    assert(pbi);

    pNodeInfo = (PNODE_INFO)pbi->Put(sizeof(NODE_INFO));
    memcpy(pNodeInfo, &_NodeInfo, sizeof(NODE_INFO));

    BuildPDU(C_ST_REG_NODE, true, true, 0, pbi);

    _bBusy = true;
    PutSendData(pbi, true);
}

bool SeedAgentHandler::TrackSeed(PSA_SEEDTARGET_REC pSTRec, bool &bRemove)
{
    DWORD  dwFlags;
    SEED_INFO  SeedInfo;
    SEED_STATE  SeedState;
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s: 0x%x\n", FULL_FUNC_NAME, pSTRec->dwFlags));

    assert(_dwNodeId != INVALID_NODE_ID);

    bRemove = false;

    pthread_mutex_lock(&_hLock);

    dwFlags = pSTRec->dwFlags;

    // TrackSeed() will only be called after recv response.
    // so it will never happed that clear F_ST_ADD flag, send C_ST_ADD_N4S,
    // set F_ST_REMOVE flag, remove it, recv N4SID...
    if ( (dwFlags & F_ST_REMOVE) && (pSTRec->dwN4SId == INVALID_N4S_ID) )
    {
        bRemove = true;
        pSTRec->bError = true;
    }
    // if dwFlags has F_ST_ADD, dwN4SId is definite invalid
    // so F_ST_ADD|F_ST_REMOVE(means pSeedTarget is NULL) and dwN4SId valid will not occur
    else if ( dwFlags & F_ST_ADD )
    {
        pSTRec->dwFlags &= (~F_ST_ADD);
        assert(pSTRec->pSeedTarget);
        pSTRec->pSeedTarget->GetSeedInfo(&SeedInfo);
    }
    else if ( dwFlags & F_ST_UPDATE )
    {
        pSTRec->dwFlags &= (~F_ST_UPDATE);
        memcpy(&SeedState, &pSTRec->SeedState, sizeof(SEED_STATE));
    }
    else if ( dwFlags & F_ST_ENUM )
    {
        // if disconnect occur after clear F_ST_ENUM flag,
        // no more enum request will send again, so LP2PSession will stuck
        // but it is no use even putting it in ReleaseEvent(),
        // because ReleaseEvent() may be called by Unbind() not by Ep.
        // so LP2PEndpoint should cooperate with SA to handle this case.
        pSTRec->dwFlags &= (~F_ST_ENUM);
    }

    pthread_mutex_unlock(&_hLock);

    if ( pSTRec->bError )
    {
        DBG(D_INFO, (L"%s: skip sidx(%d)\n", FULL_FUNC_NAME, _nCurST));
        return false;
    }

    // send is kicked by recv or non-busy, NULL is impossible
    pbi = GetSendBuf();
    assert(pbi);

    if ( dwFlags & F_ST_ADD )
    {
        assert(pSTRec->dwN4SId == INVALID_N4S_ID);
        memcpy(pbi->Put(sizeof(SEED_INFO)), &SeedInfo, sizeof(SEED_INFO));
        BuildPDU(C_ST_ADD_N4S, true, true, _dwNodeId, pbi);
    }
    else if ( dwFlags & F_ST_UPDATE )
    {
        assert(pSTRec->dwN4SId != INVALID_N4S_ID);
        memcpy(pbi->Put(sizeof(SEED_STATE)), &SeedState, sizeof(SEED_STATE));
        BuildPDU(C_ST_UPDATE_N4S_STATE, true, true, pSTRec->dwN4SId, pbi);
    }
    else if ( dwFlags & F_ST_REMOVE )
    {
        assert(pSTRec->dwN4SId != INVALID_N4S_ID);
        BuildPDU(C_ST_REMOVE_N4S, true, true, pSTRec->dwN4SId, pbi);
    }
    else if ( dwFlags & F_ST_ENUM )
    {
        assert(pSTRec->dwN4SId != INVALID_N4S_ID);
        PREQ_ENUM_NODE4SEED  pReqEnumN4S = (PREQ_ENUM_NODE4SEED)pbi->Put(sizeof(REQ_ENUM_NODE4SEED));
        pReqEnumN4S->dwStartIdx = 0;
        BuildPDU(C_ST_ENUM_NODES4SEED, true, true, pSTRec->dwN4SId, pbi);
    }

    _bBusy = true;
    PutSendData(pbi, true);

    return true;
}
