/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "SeedCenter.h"


const UINT  MAX_NODE_NUM          = 0x1000;  // 2^12
const UINT  MAX_SEED_NUM          = 0x1000;  // 2^12

#define  MAKE_NODEID(n, m)          ((((DWORD)(m)) << 12) | (((DWORD)(n)) & 0xFFF))
#define  ADJUST_NODEID_MAGICNUM(m)  (m & 0xFFFFF)
#define  NODEIDX_FROM_NODEID(x)     (((DWORD)(x)) & 0xFFF)
#define  MAGICNUM_FROM_NODEID(x)    ((((DWORD)(x)) >> 12) & 0xFFFFF)

#define  MAKE_N4SID(s, n, m)        ((((DWORD)(m)) << 24) | ((((DWORD)(n)) & 0xFFF) << 12) | (((DWORD)(s)) & 0xFFF))
#define  ADJUST_N4SID_MAGICNUM(m)   (m & 0xFF)
#define  SEEDIDX_FROM_N4SID(x)      (((DWORD)(x)) & 0xFFF)
#define  NODEIDX_FROM_N4SID(x)      ((((DWORD)(x)) >> 12) & 0xFFF)
#define  MAGICNUM_FROM_N4SID(x)     ((((DWORD)(x)) >> 24) & 0xFF)


SeedCenter *SeedCenter::_instance = NULL;

SeedCenter *SeedCenter::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new SeedCenter();
    }
    return _instance;
}

void SeedCenter::Destroy()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

SeedCenter::SeedCenter()
    : _dwNodeMagic(0)
    , _nNodeCnt(0)
    , _NodeList(0)
    , _nSeedCnt(0)
    , _SeedList(0)
    , _dwStateMagic(0)
    , _StateMatrix(0)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    pthread_rwlock_init(&_hNSLock, NULL);
    pthread_mutex_init(&_hRdLock, NULL);
}

SeedCenter::~SeedCenter()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_destroy(&_hRdLock);
    pthread_rwlock_destroy(&_hNSLock);
}

DWORD SeedCenter::Load(UINT nMaxNodeNum, UINT nMaxSeedNum)
{
    UINT  i, j;

    DBG(D_LOG, (L"%s: MaxNodeNum=%d, MaxSeedNum=%d\n", FULL_FUNC_NAME, nMaxNodeNum, nMaxSeedNum));

    if ( (nMaxNodeNum > MAX_NODE_NUM) || (nMaxSeedNum > MAX_SEED_NUM) )
    {
        DBG(D_LOG, (L"%s: not support\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_NOT_SUPPORTED;
    }

    _NodeList.resize(nMaxNodeNum);
    for ( i = 0; i < nMaxNodeNum; ++i )
    {
        memset(&_NodeList.at(i), 0, sizeof(NODE_INFO_REC));
    }

    _SeedList.resize(nMaxSeedNum);
    for ( i = 0; i < nMaxSeedNum; ++i )
    {
        memset(&_SeedList.at(i), 0, sizeof(SEED_INFO_REC));
    }

    _StateMatrix.resize(nMaxSeedNum);
    for ( i = 0; i < nMaxSeedNum; ++i )
    {
        _StateMatrix.at(i).resize(nMaxNodeNum);
        for ( j = 0; j < nMaxNodeNum; ++j )
        {
            memset(&_StateMatrix.at(i).at(j), 0, sizeof(N4S_STATE_REC));
        }
    }

    return ERROR_SUCCESS;
}

void SeedCenter::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    _nNodeCnt = 0;
    _NodeList.resize(0);

    _nSeedCnt = 0;
    _SeedList.resize(0);

    _StateMatrix.resize(0);
}

void SeedCenter::RegisterObserver(SeedTrackObserver *pObserver)
{
    UINT  i, j, nNCnt, nSCnt;
    PNODE_INFO_REC  pNodeRec;
    PSEED_INFO_REC  pSeedRec;
    PN4S_STATE_REC  pN4SRec;

    DBG(D_INFO, (L"%s: %p\n", FULL_FUNC_NAME, pObserver));

    pthread_rwlock_wrlock(&_hNSLock);

    for ( i = 0, nNCnt = 0; nNCnt < _nNodeCnt ; ++i )
    {
        pNodeRec = &_NodeList.at(i);
        if ( pNodeRec->bUsed )
        {
            ++nNCnt;
            pObserver->OnNodeRegistered(i, &pNodeRec->NodeInfo);
        }
    }

    for ( i = 0, nSCnt = 0; nSCnt < _nSeedCnt ; ++i )
    {
        pSeedRec = &_SeedList.at(i);
        if ( pSeedRec->nRefCnt )
        {
            ++nSCnt;
            pObserver->OnSeedAdded(i, &pSeedRec->SeedInfo);

            for ( j = 0, nNCnt = 0; nNCnt < pSeedRec->nRefCnt; ++j )
            {
                pN4SRec = &_StateMatrix.at(i).at(j);
                if ( pN4SRec->bUsed )
                {
                    ++nNCnt;
                    pObserver->OnNode4SeedAdded(j, i);
                    pObserver->OnNode4SeedStateUpdated(j, i, &pN4SRec->SeedState);
                }
            }
        }
    }

    _ObsList.push_back(pObserver);

    pthread_rwlock_unlock(&_hNSLock);
}

void SeedCenter::UnregisterObserver(SeedTrackObserver *pObserver)
{
    DBG(D_INFO, (L"%s: %p\n", FULL_FUNC_NAME, pObserver));

    pthread_rwlock_wrlock(&_hNSLock);
    _ObsList.remove(pObserver);
    pthread_rwlock_unlock(&_hNSLock);
}

DWORD SeedCenter::RegisterNode(const PNODE_INFO pNodeInfo, DWORD &rdwNodeId)
{
    DWORD  dwRet;
    UINT  nNodeIdx;
    DWORD  dwMagic;
    PNODE_INFO_REC  pNodeRec;

    DBG(D_INFO, (L"%s: [%hs]\n", FULL_FUNC_NAME, pNodeInfo->name));

    rdwNodeId = INVALID_NODE_ID;

    pthread_rwlock_wrlock(&_hNSLock);

    dwRet = GetFreeNode(&pNodeInfo->addr, nNodeIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: GetFreeNode failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto out;
    }

    DBG(D_INFO, (L"%s: node[%hs] @ %d\n", FULL_FUNC_NAME, pNodeInfo->name, nNodeIdx));

    pNodeRec = &_NodeList.at(nNodeIdx);

    dwMagic = _dwNodeMagic++;

    pNodeRec->bUsed = true;
    pNodeRec->dwMagic = ADJUST_NODEID_MAGICNUM(dwMagic);
    memcpy(&pNodeRec->NodeInfo, pNodeInfo, sizeof(NODE_INFO));
    pNodeRec->nSeedCnt = 0;
    pNodeRec->dwActiveSecs = GetSecondCount();

    ++_nNodeCnt;

    rdwNodeId = MAKE_NODEID(nNodeIdx, pNodeRec->dwMagic);

    for ( list<SeedTrackObserver*>::const_iterator itr = _ObsList.begin(); itr != _ObsList.end(); ++itr )
    {
        (*itr)->OnNodeRegistered(nNodeIdx, pNodeInfo);
    }

out:
    pthread_rwlock_unlock(&_hNSLock);

    return dwRet;
}

DWORD SeedCenter::UnregisterNode(DWORD dwNodeId)
{
    DWORD  dwRet = ERROR_SUCCESS;
    UINT  i, nCnt, nNodeIdx;
    DWORD  dwMagic;
    PNODE_INFO_REC  pNodeRec;
    PN4S_STATE_REC  pN4SRec;

    nNodeIdx = NODEIDX_FROM_NODEID(dwNodeId);
    dwMagic = MAGICNUM_FROM_NODEID(dwNodeId);

    DBG(D_INFO, (L"%s: node(%d), magic(%d)\n", FULL_FUNC_NAME, nNodeIdx, dwMagic));

    if ( nNodeIdx >= _NodeList.size() )
    {
        DBG(D_WARN, (L"%s: invalid node id\n", FULL_FUNC_NAME));
        return ERROR_INVALID_PARAMETER;
    }

    pNodeRec = &_NodeList.at(nNodeIdx);

    pthread_rwlock_wrlock(&_hNSLock);

    if ( (!pNodeRec->bUsed) || (dwMagic != pNodeRec->dwMagic) )
    {
        DBG(D_WARN, (L"%s: invalid node state\n", FULL_FUNC_NAME));
        dwRet = ERROR_INVALID_STATE;
        goto out;
    }

    for ( i = 0, nCnt = 0; nCnt < pNodeRec->nSeedCnt; ++i )
    {
        pN4SRec = &_StateMatrix.at(i).at(nNodeIdx);
        if ( pN4SRec->bUsed )
        {
            ++nCnt;

            ReleaseSeed(i);

            pN4SRec->bUsed = false;
        }
    }

    pNodeRec->bUsed = false;

    --_nNodeCnt;

    for ( list<SeedTrackObserver*>::const_iterator itr = _ObsList.begin(); itr != _ObsList.end(); ++itr )
    {
        (*itr)->OnNodeUnregistered(nNodeIdx);
    }

out:
    pthread_rwlock_unlock(&_hNSLock);

    return dwRet;
}

DWORD SeedCenter::AddNode4Seed(DWORD dwNodeId, const PSEED_INFO pSeedInfo, DWORD &rdwN4SId)
{
    DWORD  dwRet;
    UINT  nNodeIdx, nSeedIdx;
    DWORD  dwMagic;
    PNODE_INFO_REC  pNodeRec;
    PN4S_STATE_REC  pN4SRec;

    rdwN4SId = INVALID_N4S_ID;

    nNodeIdx = NODEIDX_FROM_NODEID(dwNodeId);
    dwMagic = MAGICNUM_FROM_NODEID(dwNodeId);

    DBG(D_INFO, (L"%s: node(%d), magic(%d), seed(%hs)\n", FULL_FUNC_NAME, nNodeIdx, dwMagic, pSeedInfo->name));

    if ( nNodeIdx >= _NodeList.size() )
    {
        DBG(D_WARN, (L"%s: invalid node id\n", FULL_FUNC_NAME));
        return ERROR_INVALID_PARAMETER;
    }

    pNodeRec = &_NodeList.at(nNodeIdx);

    pthread_rwlock_wrlock(&_hNSLock);

    if ( (!pNodeRec->bUsed) || (dwMagic != pNodeRec->dwMagic) )
    {
        DBG(D_WARN, (L"%s: invalid node state\n", FULL_FUNC_NAME));
        dwRet = ERROR_INVALID_STATE;
        goto out;
    }

    dwRet = GetFreeSeed(&pSeedInfo->sha1, nSeedIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: GetFreeSeed failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto out;
    }

    pN4SRec = &_StateMatrix.at(nSeedIdx).at(nNodeIdx);
    if ( pN4SRec->bUsed )
    {
        DBG(D_WARN, (L"%s: invalid seed state\n", FULL_FUNC_NAME));
        dwRet = ERROR_INVALID_STATE;
        goto out;
    }

    dwMagic = _dwStateMagic++;

    pN4SRec->bUsed = true;
    pN4SRec->dwMagic = ADJUST_N4SID_MAGICNUM(dwMagic);
    pN4SRec->bPreDegree = 0;
    memset(&pN4SRec->SeedState, 0, sizeof(SEED_STATE));

    PutSeed(nSeedIdx, pSeedInfo);

    ++pNodeRec->nSeedCnt;

    rdwN4SId = MAKE_N4SID(nSeedIdx, nNodeIdx, pN4SRec->dwMagic);

    for ( list<SeedTrackObserver*>::const_iterator itr = _ObsList.begin(); itr != _ObsList.end(); ++itr )
    {
        (*itr)->OnNode4SeedAdded(nNodeIdx, nSeedIdx);
    }

out:
    pthread_rwlock_unlock(&_hNSLock);

    return dwRet;
}

DWORD SeedCenter::RemoveNode4Seed(DWORD dwN4SId)
{
    DWORD  dwRet = ERROR_SUCCESS;
    UINT  nNodeIdx, nSeedIdx;
    DWORD  dwMagic;
    PNODE_INFO_REC  pNodeRec;
    PN4S_STATE_REC  pN4SRec;

    nNodeIdx = NODEIDX_FROM_N4SID(dwN4SId);
    nSeedIdx = SEEDIDX_FROM_N4SID(dwN4SId);
    dwMagic = MAGICNUM_FROM_N4SID(dwN4SId);

    DBG(D_INFO, (L"%s: node(%d), seed(%d), magic(%d)\n", FULL_FUNC_NAME, nNodeIdx, nSeedIdx, dwMagic));

    if ( (nNodeIdx >= _NodeList.size()) || (nSeedIdx >= _SeedList.size()) )
    {
        DBG(D_WARN, (L"%s: invalid node/seed id\n", FULL_FUNC_NAME));
        return ERROR_INVALID_PARAMETER;
    }

    pNodeRec = &_NodeList.at(nNodeIdx);
    pN4SRec = &_StateMatrix.at(nSeedIdx).at(nNodeIdx);

    pthread_rwlock_wrlock(&_hNSLock);

    if ( (!pN4SRec->bUsed) || (dwMagic != pN4SRec->dwMagic) )
    {
        DBG(D_WARN, (L"%s: invalid session state\n", FULL_FUNC_NAME));
        dwRet = ERROR_INVALID_STATE;
        goto out;
    }

    ReleaseSeed(nSeedIdx);

    pN4SRec->bUsed = false;

    --pNodeRec->nSeedCnt;

    for ( list<SeedTrackObserver*>::const_iterator itr = _ObsList.begin(); itr != _ObsList.end(); ++itr )
    {
        (*itr)->OnNode4SeedRemoved(nNodeIdx, nSeedIdx);
    }

out:
    pthread_rwlock_unlock(&_hNSLock);

    return dwRet;
}

DWORD SeedCenter::UpdateNode4SeedState(DWORD dwN4SId, const PSEED_STATE pSeedState)
{
    DWORD  dwRet = ERROR_SUCCESS;
    UINT  nNodeIdx, nSeedIdx;
    DWORD  dwMagic;
    PN4S_STATE_REC  pN4SRec;

    nNodeIdx = NODEIDX_FROM_N4SID(dwN4SId);
    nSeedIdx = SEEDIDX_FROM_N4SID(dwN4SId);
    dwMagic = MAGICNUM_FROM_N4SID(dwN4SId);

    DBG(D_VERB, (L"%s: node(%d), seed(%d), magic(%d)\n", FULL_FUNC_NAME, nNodeIdx, nSeedIdx, dwMagic));

    if ( (nNodeIdx >= _NodeList.size()) || (nSeedIdx >= _SeedList.size()) )
    {
        DBG(D_WARN, (L"%s: invalid node/seed id\n", FULL_FUNC_NAME));
        return ERROR_INVALID_PARAMETER;
    }

    pN4SRec = &_StateMatrix.at(nSeedIdx).at(nNodeIdx);

    pthread_rwlock_rdlock(&_hNSLock);

    if ( (!pN4SRec->bUsed) || (dwMagic != pN4SRec->dwMagic) )
    {
        DBG(D_WARN, (L"%s: invalid session state\n", FULL_FUNC_NAME));
        dwRet = ERROR_INVALID_STATE;
        goto out;
    }

    memcpy(&pN4SRec->SeedState, pSeedState, sizeof(SEED_STATE));

    if ( ((pSeedState->bProgressDegree - pN4SRec->bPreDegree) >= 1) || (pSeedState->bProgressDegree == 100) )
    {
        pN4SRec->bPreDegree = pSeedState->bProgressDegree;

        pthread_mutex_lock(&_hRdLock);
        for ( list<SeedTrackObserver*>::const_iterator itr = _ObsList.begin(); itr != _ObsList.end(); ++itr )
        {
            (*itr)->OnNode4SeedStateUpdated(nNodeIdx, nSeedIdx, pSeedState);
        }
        pthread_mutex_unlock(&_hRdLock);
    }

out:
    pthread_rwlock_unlock(&_hNSLock);

    return dwRet;
}

DWORD SeedCenter::EnumNodes4Seed(DWORD dwN4SId, UINT &nStartIdx, PNODE4SEED_INFO pN4SInfoArray, UINT &nCount)
{
    DWORD  dwRet = ERROR_SUCCESS;
    UINT  i, nCopyed, nSize;
    UINT  nNodeIdx, nSeedIdx;
    DWORD  dwMagic;
    PN4S_STATE_REC  pN4SRec;
    PSEED_INFO_REC  pSeedRec;

    nNodeIdx = NODEIDX_FROM_N4SID(dwN4SId);
    nSeedIdx = SEEDIDX_FROM_N4SID(dwN4SId);
    dwMagic = MAGICNUM_FROM_N4SID(dwN4SId);

    DBG(D_INFO, (L"%s: node(%d), seed(%d), magic(%d)\n", FULL_FUNC_NAME, nNodeIdx, nSeedIdx, dwMagic));

    if ( (nNodeIdx >= _NodeList.size()) || (nSeedIdx >= _SeedList.size()) )
    {
        DBG(D_WARN, (L"%s: invalid node/seed id\n", FULL_FUNC_NAME));
        return ERROR_INVALID_PARAMETER;
    }

    pN4SRec = &_StateMatrix.at(nSeedIdx).at(nNodeIdx);
    pSeedRec = &_SeedList.at(nNodeIdx);

    pthread_rwlock_rdlock(&_hNSLock);

    if ( (!pN4SRec->bUsed) || (dwMagic != pN4SRec->dwMagic) )
    {
        DBG(D_WARN, (L"%s: invalid session state\n", FULL_FUNC_NAME));
        dwRet = ERROR_INVALID_STATE;
        goto out;
    }

    if ( !nCount )
    {
        nCount = pSeedRec->nRefCnt;
        goto out;
    }

    for ( i = nStartIdx, nCopyed = 0, nSize = (UINT)_NodeList.size(); i < nSize; ++i )
    {
        pN4SRec = &_StateMatrix.at(nSeedIdx).at(i);
        if ( pN4SRec->bUsed )
        {
            pN4SInfoArray[nCopyed].NodeAddr = _NodeList.at(i).NodeInfo.addr;
            pN4SInfoArray[nCopyed].qwProgressMask = pN4SRec->SeedState.qwProgressMask;
            ++nCopyed;

            if ( nCopyed >= nCount )
            {
                break;
            }
        }
    }
    nStartIdx = ++i;

    nCount = nCopyed;

out:
    pthread_rwlock_unlock(&_hNSLock);

    return dwRet;
}

DWORD SeedCenter::GetFreeNode(const PSOCK_ADDR pAddr, UINT &rnIdx)
{
    UINT  i, nCnt;
    PNODE_INFO_REC  pNodeRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _nNodeCnt >= (UINT)_NodeList.size() )
    {
        DBG(D_WARN, (L"%s: full\n", FULL_FUNC_NAME));
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    rnIdx = INVALID_INDEX_VALUE;

    for ( i = 0, nCnt = 0; nCnt < _nNodeCnt; ++i )
    {
        pNodeRec = &_NodeList.at(i);

        if ( pNodeRec->bUsed )
        {
            ++nCnt;

            if ( *pAddr == pNodeRec->NodeInfo.addr )
            {
                DBG(D_WARN, (L"%s: exist\n", FULL_FUNC_NAME));
                return ERROR_ACCESS_DENIED;
            }
        }
        else if ( rnIdx == INVALID_INDEX_VALUE )
        {
            // the first unused one met during search
            rnIdx = i;
        }
    }

    if ( rnIdx == INVALID_INDEX_VALUE )
    {
        // the next one after the last used one
        rnIdx = i;
    }

    return ERROR_SUCCESS;
}

DWORD SeedCenter::GetFreeSeed(const PSHA1_VAL pSHA1, UINT &rnIdx)
{
    UINT  i, nCnt;
    PSEED_INFO_REC  pSeedRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rnIdx = INVALID_INDEX_VALUE;

    for ( i = 0, nCnt = 0; nCnt < _nSeedCnt; ++i )
    {
        pSeedRec = &_SeedList.at(i);

        if ( pSeedRec->nRefCnt > 0 )
        {
            ++nCnt;

            if ( *pSHA1 == pSeedRec->SeedInfo.sha1 )
            {
                rnIdx = i;
                return ERROR_SUCCESS;
            }
        }
        else if ( rnIdx == INVALID_INDEX_VALUE )
        {
            // the first unused one met during search
            rnIdx = i;
        }
    }

    if ( (rnIdx == INVALID_INDEX_VALUE) && (i < _SeedList.size()) )
    {
        // the next one after the last used one
        rnIdx = i;
    }

    if ( rnIdx == INVALID_INDEX_VALUE )
    {
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    return ERROR_SUCCESS;
}

void SeedCenter::PutSeed(UINT nIdx, const PSEED_INFO pSeedInfo)
{
    PSEED_INFO_REC  pSeedRec;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nIdx));

    pSeedRec = &_SeedList.at(nIdx);
    // nRefCnt is UINT, so always ">=0"
//    assert(pSeedRec->nRefCnt >= 0);

    ++pSeedRec->nRefCnt;
    if ( pSeedRec->nRefCnt == 1 )
    {
        memcpy(&pSeedRec->SeedInfo, pSeedInfo, sizeof(SEED_INFO));

        ++_nSeedCnt;

        for ( list<SeedTrackObserver*>::const_iterator itr = _ObsList.begin(); itr != _ObsList.end(); ++itr )
        {
           (*itr)->OnSeedAdded(nIdx, pSeedInfo);
        }
    }
}

void SeedCenter::ReleaseSeed(UINT nIdx)
{
    PSEED_INFO_REC  pSeedRec;

    DBG(D_VERB, (L"%s: %d\n", FULL_FUNC_NAME, nIdx));

    pSeedRec = &_SeedList.at(nIdx);
    assert(pSeedRec->nRefCnt > 0);

    --pSeedRec->nRefCnt;
    if ( !pSeedRec->nRefCnt )
    {
        --_nSeedCnt;

        for ( list<SeedTrackObserver*>::const_iterator itr = _ObsList.begin(); itr != _ObsList.end(); ++itr )
        {
           (*itr)->OnSeedRemoved(nIdx);
        }
    }
}

