/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "SHA1Hash.h"
#include "ExecutorPool.h"
#include "SeedTarget.h"


SeedTarget::SeedTarget(UINT nMaxRdNum, UINT nMaxWrNum)
    : _dwRunnId(INVALID_RUNNABLE_ID)
    , _pSeedHdr(NULL)
    , _hTargetFile(NULL)
    , _bObsAdded(false)
    , _bMaskUpdated(false)
    , _lBlkRdCnt(0)
    , _lBlkWrCnt(0)
    , _RWSList(nMaxRdNum + nMaxWrNum)
    , _nRWSDivide(nMaxRdNum)
    , _RdSListMate(true)
    , _WrSListMate(true)
    , _dwRWSMagic(0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_init(&_hStateLock, NULL);
    pthread_mutex_init(&_hObsLock, NULL);
}

SeedTarget::~SeedTarget()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_destroy(&_hObsLock);
    pthread_mutex_destroy(&_hStateLock);
}

DWORD SeedTarget::Init(const char *szSeedFile, const char *szTargetPath, bool bReadOnly, bool bAllowVerify, const char *szCacheFile)
{
    DWORD  dwRet;
    bool  bNeedVerify;
    char  szTargetFile[MAX_PATH];
    UINT  nRWSCnt;

    DBG(D_LOG, (L"%s: seed(%hs), target_path(%hs), readonly(%d), verify(%d), cache(%hs)\n", FULL_FUNC_NAME, szSeedFile, szTargetPath, bReadOnly, bAllowVerify, szCacheFile));

    dwRet = ExecutorCenter::Instance()->RegisterRunnable(this, _dwRunnId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: RegisterRunnable failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    dwRet = ReadSeedFile(szSeedFile);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ReadSeedFile(%hs) failed: %d\n", FULL_FUNC_NAME, szSeedFile, dwRet));
        goto err2;
    }

    dwRet = VerifySeedHdr();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifySeedHdr failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err3;
    }

    if ( szCacheFile )
    {
        dwRet = _baBlkProg.Init((UINT)_pSeedHdr->dwSHA1Num, szCacheFile, true);
    }
    else
    {
        dwRet = _baBlkProg.Init((UINT)_pSeedHdr->dwSHA1Num, (bReadOnly && (!bAllowVerify)), true);
    }
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init _baBlkProg failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err3;
    }

    dwRet = _baBlkBusy.Init((UINT)_pSeedHdr->dwSHA1Num, false, false);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init _baBlkBusy failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err4;
    }

    // see misc_st_init.txt
    if ( PathHasFileName(szTargetPath) )
    {
        DBG(D_INFO, (L"%s: use the target name: %hs\n", FULL_FUNC_NAME, szTargetPath));
        _snprintf(szTargetFile, _countof(szTargetFile), "%s", szTargetPath);
    }
    else
    {
        _snprintf(szTargetFile, _countof(szTargetFile), "%s%s", szTargetPath, _pSeedHdr->szFileName);
    }
    dwRet = OpenTargetFile(szTargetFile, bReadOnly, bNeedVerify);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: OpenTargetFile(%hs) failed: %d\n", FULL_FUNC_NAME, szTargetFile, dwRet));
        goto err5;
    }

    if ( bNeedVerify && bAllowVerify )
    {
        // see misc_st_init.txt
        dwRet = VerifyTargetFile((szCacheFile != NULL));
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: VerifyTargetFile(%hs) failed: %d\n", FULL_FUNC_NAME, szTargetFile, dwRet));
            goto err6;
        }
    }

#if defined(CONFIG_ST_BLK_RD_MMAP) || defined(CONFIG_ST_BLK_WR_MMAP)
    dwRet = _TargetMapping.Init(_hTargetFile, _pSeedHdr->qwFileSize, bReadOnly);
    if ( dwRet != ERROR_SUCCESS )
    {
        assert(0);
        DBG(D_WARN, (L"%s: _TargetMapping init(%lld) failed: %d\n", FULL_FUNC_NAME, _pSeedHdr->qwFileSize, dwRet));
        goto err6;
    }
#endif

    nRWSCnt = (UINT)_RWSList.size();

    dwRet = _RdSListMate.Init(_nRWSDivide);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: _RdSListMate init(%d) failed: %d\n", FULL_FUNC_NAME, _nRWSDivide, dwRet));
        goto err7;
    }

    dwRet = _WrSListMate.Init(nRWSCnt - _nRWSDivide, _nRWSDivide);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: _WrSListMate init(%d) failed: %d\n", FULL_FUNC_NAME, (nRWSCnt - _nRWSDivide), dwRet));
        goto err8;
    }

    dwRet = InitRWSEnv(szTargetFile);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: InitHashEnv failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err9;
    }

#ifdef CONFIG_ST_BLK_WR_VERIFY
    dwRet = InitHashEnv();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: InitHashEnv failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err10;
    }
#endif

    DBG(D_LOG, (L"%s: target(%hs): done: %d/%d, degree: %d%%\n", FULL_FUNC_NAME, _pSeedHdr->szFileName, _baBlkProg.GetSetsCount(), _pSeedHdr->dwSHA1Num, _baBlkProg.GetDegree()));

    return ERROR_SUCCESS;

#ifdef CONFIG_ST_BLK_WR_VERIFY
err10:
#endif
    FiniRWSEnv();

err9:
    _WrSListMate.Fini();

err8:
    _RdSListMate.Fini();

err7:
#if defined(CONFIG_ST_BLK_RD_MMAP) || defined(CONFIG_ST_BLK_WR_MMAP)
    _TargetMapping.Fini();
#endif

err6:
    fclose(_hTargetFile);
    _hTargetFile = NULL;

err5:
    _baBlkBusy.Fini();

err4:
    _baBlkProg.Fini();

err3:
    _pSeedHdr = NULL;

err2:
    ExecutorCenter::Instance()->UnregisterRunnable(_dwRunnId);
    _dwRunnId = INVALID_RUNNABLE_ID;

err1:
    return dwRet;
}

void SeedTarget::Fini()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( _dwRunnId != INVALID_RUNNABLE_ID )
    {
        ExecutorCenter::Instance()->UnregisterRunnable(_dwRunnId);
        _dwRunnId = INVALID_RUNNABLE_ID;
    }

#ifdef CONFIG_ST_BLK_WR_VERIFY
    FiniHashEnv();
#endif

    FiniRWSEnv();
    _RdSListMate.Fini();
    _WrSListMate.Fini();

#if defined(CONFIG_ST_BLK_RD_MMAP) || defined(CONFIG_ST_BLK_WR_MMAP)
    _TargetMapping.Fini();
#endif

    if ( _hTargetFile )
    {
        fclose(_hTargetFile);
        _hTargetFile = NULL;
    }

    // _baBlkProg(cache).Fini() SHOULD after fclose(_hTargetFile)
    _baBlkProg.Fini();
    _baBlkBusy.Fini();

    _lBlkRdCnt = 0;
    _lBlkWrCnt = 0;

    _pSeedHdr = NULL;
}

DWORD SeedTarget::InitRWSEnv(const char *szTargetFile)
{
    DWORD  dwRet;
    UINT  i, nRWSCnt;
    PST_RWS_REC  pRWSRec;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    nRWSCnt = (UINT)_RWSList.size();

    for ( i = 0; i < nRWSCnt; ++i )
    {
        pRWSRec = &_RWSList.at(i);

        DO_IF_DBG(pRWSRec->bUsed = false);
        pRWSRec->nBlkIdx = INVALID_INDEX_VALUE;

#if defined(CONFIG_ST_BLK_RD_FIO) || defined(CONFIG_ST_BLK_WR_FIO)
        pRWSRec->hFile = NULL;
#if defined(CONFIG_ST_BLK_RD_BUF) || defined(CONFIG_ST_BLK_WR_BUF)
        pRWSRec->pFIOBuf = NULL;
#endif
#endif
    }

#ifdef CONFIG_ST_BLK_RD_FIO
    for ( i = 0; i < _nRWSDivide; ++i )
    {
        pRWSRec = &_RWSList.at(i);

        pRWSRec->hFile = fopen(szTargetFile, "rb");
        if ( !pRWSRec->hFile )
        {
            dwRet = GetLastError();
            DBG(D_WARN, (L"%s: fopen(%hs) failed: %d\n", FULL_FUNC_NAME, szTargetFile, dwRet));
            goto err;
        }
        pRWSRec->qwCurOffset = 0;

#ifdef CONFIG_ST_BLK_RD_BUF
        if ( ST_BLK_RD_BUF_SIZE )
        {
            pRWSRec->pFIOBuf = new(std::nothrow) char[ST_BLK_RD_BUF_SIZE];
        }
        if ( pRWSRec->pFIOBuf )
        {
            DBG(D_INFO, (L"%s: buffered for fread: %d\n", FULL_FUNC_NAME, ST_BLK_RD_BUF_SIZE));
            setvbuf(pRWSRec->hFile, pRWSRec->pFIOBuf, _IOFBF, ST_BLK_RD_BUF_SIZE);
        }
        else
        {
            DBG(D_INFO, (L"%s: buffered for fread: default\n", FULL_FUNC_NAME));
        }
#else
        DBG(D_INFO, (L"%s: unbuffered for fread\n", FULL_FUNC_NAME));
        // use read(), see issues_fio_buf.txt
        setvbuf(pRWSRec->hFile, NULL, _IONBF, 0);
#endif
    }
#endif

#ifdef CONFIG_ST_BLK_WR_FIO
    for ( i = _nRWSDivide; i < nRWSCnt; ++i )
    {
        pRWSRec = &_RWSList.at(i);

        pRWSRec->hFile = fopen(szTargetFile, "r+b");
        if ( !pRWSRec->hFile )
        {
            dwRet = GetLastError();
            DBG(D_WARN, (L"%s: fopen(%hs) failed: %d\n", FULL_FUNC_NAME, szTargetFile, dwRet));
            goto err;
        }
        pRWSRec->qwCurOffset = 0;

#ifdef CONFIG_ST_BLK_WR_BUF
        if ( ST_BLK_WR_BUF_SIZE )
        {
            pRWSRec->pFIOBuf = new(std::nothrow) char[ST_BLK_WR_BUF_SIZE];
        }
        if ( pRWSRec->pFIOBuf )
        {
            DBG(D_INFO, (L"%s: buffered for fwrite: %d\n", FULL_FUNC_NAME, ST_BLK_WR_BUF_SIZE));
            setvbuf(pRWSRec->hFile, pRWSRec->pFIOBuf, _IOFBF, ST_BLK_WR_BUF_SIZE);
        }
        else
        {
            DBG(D_INFO, (L"%s: buffered for fwrite: default\n", FULL_FUNC_NAME));
        }
#else
        DBG(D_INFO, (L"%s: unbuffered for fwrite\n", FULL_FUNC_NAME));
        // use write(), see issues_fio_buf.txt
        setvbuf(pRWSRec->hFile, NULL, _IONBF, 0);
#endif
    }
#endif

    return ERROR_SUCCESS;

#if defined(CONFIG_ST_BLK_RD_FIO) || defined(CONFIG_ST_BLK_WR_FIO)
err:
    FiniRWSEnv();
#endif

    return dwRet;
}

void SeedTarget::FiniRWSEnv()
{
    UINT  i, nRWSCnt;
    PST_RWS_REC  pRWSRec;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    nRWSCnt = (UINT)_RWSList.size();

    for ( i = 0; i < nRWSCnt; ++i )
    {
        pRWSRec = &_RWSList.at(i);

#if defined(CONFIG_ST_BLK_RD_FIO) || defined(CONFIG_ST_BLK_WR_FIO)
        if ( pRWSRec->hFile )
        {
            fclose(pRWSRec->hFile);
            pRWSRec->hFile = NULL;
        }
#if defined(CONFIG_ST_BLK_RD_BUF) || defined(CONFIG_ST_BLK_WR_BUF)
        delete pRWSRec->pFIOBuf;
        pRWSRec->pFIOBuf = NULL;
#endif
#endif

        assert(pRWSRec->nBlkIdx == INVALID_INDEX_VALUE);
    }
}

#ifdef CONFIG_ST_BLK_WR_VERIFY
DWORD SeedTarget::InitHashEnv()
{
    SHA1Hash  *sha1;
    UINT  i, nRWSCnt;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    nRWSCnt = (UINT)_RWSList.size();
    for ( i = _nRWSDivide; i < nRWSCnt; ++i )
    {
        sha1 = new(std::nothrow) SHA1Hash();
        if ( !sha1 )
        {
            DBG(D_WARN, (L"%s: alloc sha1 failed\n", FULL_FUNC_NAME));
            goto err;
        }
        _RWSList.at(i).sha1 = sha1;
    }

    return ERROR_SUCCESS;

err:
    FiniHashEnv();

    return ERROR_OUTOFMEMORY;
}

void SeedTarget::FiniHashEnv()
{
    UINT  i, nRWSCnt;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    nRWSCnt = (UINT)_RWSList.size();
    for ( i = _nRWSDivide; i < nRWSCnt; ++i )
    {
        delete _RWSList.at(i).sha1;
        _RWSList.at(i).sha1 = NULL;
    }
}
#endif

DWORD SeedTarget::RegisterObserver(SeedTargetObserver *pObserver, DWORD dwCookie)
{
    PST_OBS_REC  pObsRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pObsRec = new(std::nothrow) ST_OBS_REC;
    if ( !pObsRec )
    {
        DBG(D_WARN, (L"%s: alloc ST_OBS_REC failed\n", FULL_FUNC_NAME));
        return ERROR_OUTOFMEMORY;
    }

    pObsRec->bNewOne = true;
    pObsRec->pObserver = pObserver;
    pObsRec->dwCookie = dwCookie;

    pthread_mutex_lock(&_hObsLock);
    _ObsList.push_back(pObsRec);
    pthread_mutex_unlock(&_hObsLock);

    pthread_mutex_lock(&_hStateLock);
    _bObsAdded = true;
    ReqExec();
    pthread_mutex_unlock(&_hStateLock);

    return ERROR_SUCCESS;
}

void SeedTarget::UnregisterObserver(SeedTargetObserver *pObserver)
{
    list<PST_OBS_REC>::iterator  itr;
    list<PST_OBS_REC>::const_iterator  end_itr;
    PST_OBS_REC  pObsRec = NULL;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pthread_mutex_lock(&_hObsLock);
    for ( itr = _ObsList.begin(), end_itr = _ObsList.end(); itr != end_itr; ++itr )
    {
        pObsRec = *itr;
        if ( pObsRec->pObserver == pObserver )
        {
            _ObsList.erase(itr);
            break;
        }
    }
    pthread_mutex_unlock(&_hObsLock);

    delete pObsRec;
}

bool SeedTarget::IsMatch(const SHA1_VAL &SHA1ValSeed) const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pSeedHdr);

    return (_pSeedHdr->SHA1ValSeed == SHA1ValSeed);
}

const char *SeedTarget::GetName() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pSeedHdr);

    return _pSeedHdr->szFileName;
}

void SeedTarget::GetSeedInfo(PSEED_INFO pInfo) const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pSeedHdr);

    strncpy(pInfo->name, _pSeedHdr->szFileName, MAX_SEED_NAME_LEN);
    memcpy(&pInfo->sha1, &_pSeedHdr->SHA1ValSeed, sizeof(SHA1_VAL));
}

bool SeedTarget::IsFinished() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pSeedHdr);

    return _baBlkProg.IsAllSet();
}

bool SeedTarget::IsAllBusy() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pSeedHdr);

    // no need lock, because CloseBlock() set _baBlkProg first then clear _baBlkBusy
    return (_baBlkBusy.GetSetsCount() >= (_baBlkProg.GetBitsCount() - _baBlkProg.GetSetsCount()));
}

UINT SeedTarget::GetMaxReadNum() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return _nRWSDivide;
}

UINT SeedTarget::GetMaxWriteNum() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return ((UINT)_RWSList.size() - _nRWSDivide);
}

UINT SeedTarget::GetBlockReadCount() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return (UINT)_lBlkRdCnt;
}

UINT SeedTarget::GetBlockWriteCount() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return (UINT)_lBlkWrCnt;
}

const BitArray *SeedTarget::GetProgressMask() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pSeedHdr);

    return &_baBlkProg;
}

UINT SeedTarget::GetBlockCount() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pSeedHdr);

    return (UINT)_pSeedHdr->dwSHA1Num;
}

DWORD SeedTarget::GetBlockSize() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_pSeedHdr);

    return _pSeedHdr->dwBlockSize;
}

DWORD SeedTarget::BeginSession(bool bReadOrWrite, DWORD &rdwSessionId)
{
    DWORD  dwRet;
    UINT  nRWSIdx;
    DWORD  dwMagic;
    PST_RWS_REC  pRWSRec;

    DBG(D_VERB, (L"%s: r/w(%d)\n", FULL_FUNC_NAME, bReadOrWrite));

    assert(_hTargetFile);

    dwRet = bReadOrWrite ? _RdSListMate.GetFreeIndex(nRWSIdx) : _WrSListMate.GetFreeIndex(nRWSIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_VERB, (L"%s: GetFreeIndex(%d) failed: %d\n", FULL_FUNC_NAME, bReadOrWrite, dwRet));
        goto err1;
    }

    pRWSRec = &_RWSList.at(nRWSIdx);
    assert(!pRWSRec->bUsed);

    dwMagic = InterlockedIncrement((LONG*)&_dwRWSMagic);

    pRWSRec->wMagic = ADJUST_WMAGICNUM(dwMagic);
    pRWSRec->MesaRdWr.Reset();
    pRWSRec->nBlkIdx = INVALID_INDEX_VALUE;

    DO_IF_DBG(pRWSRec->bUsed = true);

    rdwSessionId = MAKE_DWID(nRWSIdx, pRWSRec->wMagic);

    return ERROR_SUCCESS;

err1:
    rdwSessionId = INVALID_ST_SESS_ID;

    return dwRet;
}

DWORD SeedTarget::EndSession(DWORD dwSessionId)
{
    DWORD  dwRet;
    UINT  nRWSIdx;
    PST_RWS_REC  pRWSRec;

    DBG(D_VERB, (L"%s: sid(%#.8x)\n", FULL_FUNC_NAME, dwSessionId));

    dwRet = VerifySessionId(dwSessionId, nRWSIdx, pRWSRec);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifySessionId failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    if ( pRWSRec->nBlkIdx != INVALID_INDEX_VALUE )
    {
        DBG(D_WARN, (L"%s: blk(%d) not closed\n", FULL_FUNC_NAME, pRWSRec->nBlkIdx));
        assert(0);
        CloseBlock(dwSessionId);
    }

    pRWSRec->MesaRdWr.Print("ST_RW");

    DO_IF_DBG(pRWSRec->bUsed = false);
    (nRWSIdx < _nRWSDivide) ? _RdSListMate.PutFreeIndex(nRWSIdx) : _WrSListMate.PutFreeIndex(nRWSIdx);

    return ERROR_SUCCESS;
}

DWORD SeedTarget::OpenBlock(DWORD dwSessionId, UINT nBlkIdx, DWORD &rdwBlkId)
{
    DWORD  dwRet;
    UINT  nRWSIdx;
    PST_RWS_REC  pRWSRec;
    bool  bWrite, bBlkReady;

    DBG(D_VERB, (L"%s: sid(%#.8x), blk(%d)\n", FULL_FUNC_NAME, dwSessionId, nBlkIdx));

    dwRet = VerifySessionId(dwSessionId, nRWSIdx, pRWSRec);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifySessionId failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    if ( pRWSRec->nBlkIdx != INVALID_INDEX_VALUE )
    {
        dwRet = ERROR_NOT_SUPPORTED;
        DBG(D_WARN, (L"%s: not support multi open\n", FULL_FUNC_NAME));
        assert(0);
        goto err1;
    }

    bWrite = (nRWSIdx >= _nRWSDivide);

    pthread_mutex_lock(&_hStateLock);
    bBlkReady = _baBlkProg.IsBitSet(nBlkIdx);
    if ( bWrite )
    {
        if ( bBlkReady )
        {
            dwRet = ERROR_ALREADY_EXISTS;
        }
        else
        {
            dwRet = _baBlkBusy.TestAndSetBit(nBlkIdx) ? ERROR_SUCCESS : ERROR_BUSY;
        }
    }
    else
    {
        dwRet = bBlkReady ? ERROR_SUCCESS : ERROR_NOT_READY;
    }
    pthread_mutex_unlock(&_hStateLock);

    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_VERB, (L"%s: open block(%d) failed: %d\n", FULL_FUNC_NAME, nBlkIdx, dwRet));
        goto err1;
    }

    pRWSRec->nBlkIdx = nBlkIdx;

    pRWSRec->qwBlkOffset = (QWORD)_pSeedHdr->dwBlockSize * (QWORD)nBlkIdx;
    if ( (nBlkIdx + 1) == _pSeedHdr->dwSHA1Num )
    {
        pRWSRec->dwBlkSize = (DWORD)(_pSeedHdr->qwFileSize - pRWSRec->qwBlkOffset);
    }
    else
    {
        pRWSRec->dwBlkSize = _pSeedHdr->dwBlockSize;
    }
    pRWSRec->dwOffInBlk = 0;
    pRWSRec->dwLeftInBlk = pRWSRec->dwBlkSize;
    pRWSRec->pBlkBuf = NULL;

    if ( bWrite )
    {
#ifdef CONFIG_ST_BLK_WR_MMAP
        dwRet = _TargetMapping.Map(pRWSRec->qwBlkOffset, pRWSRec->dwBlkSize, !bWrite, bWrite, pRWSRec->pBlkBuf);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: mmap block(%d) failed: %d\n", FULL_FUNC_NAME, nBlkIdx, dwRet));
            assert(0);
            goto err2;
        }
#else
        // see issues_fseek.txt
        if ( pRWSRec->qwCurOffset != pRWSRec->qwBlkOffset )
        {
//            DBG(D_PERF, (L"%s: fseek(%#llx -> %#llx)\n", FULL_FUNC_NAME, pRWSRec->qwCurOffset, pRWSRec->qwBlkOffset));
#ifdef CONFIG_ST_BLK_WR_BUF
            if ( fseeko(pRWSRec->hFile, pRWSRec->qwBlkOffset, SEEK_SET) )
#else
            // see issues_fio_fd.txt
            if ( lseek(_fileno(pRWSRec->hFile), pRWSRec->qwBlkOffset, SEEK_SET) == -1 )
#endif
            {
                dwRet = GetLastError();
                DBG(D_WARN, (L"%s: fseek(%#llx) failed: %d\n", FULL_FUNC_NAME, pRWSRec->qwBlkOffset, dwRet));
                goto err2;
            };
            pRWSRec->qwCurOffset = pRWSRec->qwBlkOffset;
        }
#endif
    }
    else
    {
#ifdef CONFIG_ST_BLK_RD_MMAP
        dwRet = _TargetMapping.Map(pRWSRec->qwBlkOffset, pRWSRec->dwBlkSize, !bWrite, bWrite, pRWSRec->pBlkBuf);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: mmap block(%d) failed: %d\n", FULL_FUNC_NAME, nBlkIdx, dwRet));
            assert(0);
            goto err2;
        }
#ifdef _LINUX_
        madvise(pRWSRec->pBlkBuf, pRWSRec->dwBlkSize, MADV_SEQUENTIAL|MADV_WILLNEED);
#endif
#else
        // see issues_fseek.txt
        if ( pRWSRec->qwCurOffset != pRWSRec->qwBlkOffset )
        {
//            DBG(D_PERF, (L"%s: fseek(%#llx -> %#llx)\n", FULL_FUNC_NAME, pRWSRec->qwCurOffset, pRWSRec->qwBlkOffset));
#ifdef CONFIG_ST_BLK_RD_BUF
            if ( fseeko(pRWSRec->hFile, pRWSRec->qwBlkOffset, SEEK_SET) )
#else
            // see issues_fio_fd.txt
            if ( lseek(_fileno(pRWSRec->hFile), pRWSRec->qwBlkOffset, SEEK_SET) == -1 )
#endif
            {
                dwRet = GetLastError();
                DBG(D_WARN, (L"%s: fseek(%#llx) failed: %d\n", FULL_FUNC_NAME, pRWSRec->qwBlkOffset, dwRet));
                goto err2;
            };
            pRWSRec->qwCurOffset = pRWSRec->qwBlkOffset;
        }
#ifdef _LINUX_
        posix_fadvise(_fileno(pRWSRec->hFile), pRWSRec->qwBlkOffset, pRWSRec->dwBlkSize,
            POSIX_FADV_SEQUENTIAL|POSIX_FADV_NOREUSE);
#endif
#endif

#ifdef CONFIG_ST_BLK_RD_AHEAD
#ifdef _LINUX_
        readahead(_fileno(pRWSRec->hFile), pRWSRec->qwBlkOffset, pRWSRec->dwBlkSize);
#endif
#endif
    }

#ifdef CONFIG_ST_BLK_WR_VERIFY
    if ( bWrite )
    {
        assert(pRWSRec->sha1);
        dwRet = pRWSRec->sha1->BeginHash();
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: SHA1_BeginHash failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto err3;
        }
    }
#endif

    rdwBlkId = dwSessionId;

    return ERROR_SUCCESS;

#ifdef CONFIG_ST_BLK_WR_VERIFY
err3:
#endif
    if ( bWrite )
    {
#ifdef CONFIG_ST_BLK_WR_MMAP
        _TargetMapping.Unmap(pRWSRec->pBlkBuf, pRWSRec->dwBlkSize);
#endif
    }
    else
    {
#ifdef CONFIG_ST_BLK_RD_MMAP
        _TargetMapping.Unmap(pRWSRec->pBlkBuf, pRWSRec->dwBlkSize);
#endif
    }

err2:
    if ( bWrite )
    {
        pthread_mutex_lock(&_hStateLock);
        _baBlkBusy.ClearBit(nBlkIdx);
        pthread_mutex_unlock(&_hStateLock);
    }

err1:
    rdwBlkId = INVALID_ST_BLK_ID;

    return dwRet;
}

DWORD SeedTarget::ReadBlock(DWORD dwBlkId, PBYTE pBuf, DWORD &rdwSize, bool &rbEndOfBlk)
{
    DWORD  dwRet;
    UINT  nRWSIdx;
    PST_RWS_REC  pRWSRec;

    DBG(D_VERB, (L"%s: sid(%#.8x)\n", FULL_FUNC_NAME, dwBlkId));

    dwRet = VerifySessionId(dwBlkId, nRWSIdx, pRWSRec);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifySessionId failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    dwRet = VerifyReadWrite(true, nRWSIdx, pRWSRec);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifyReadWrite failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    assert((pRWSRec->dwOffInBlk + pRWSRec->dwLeftInBlk) <= _pSeedHdr->dwBlockSize);

    if ( rdwSize > pRWSRec->dwLeftInBlk )
    {
        rdwSize = pRWSRec->dwLeftInBlk;
    }

    pRWSRec->MesaRdWr.Start();
#ifdef CONFIG_ST_BLK_RD_MMAP
    assert(pRWSRec->pBlkBuf);
    memcpy(pBuf, &pRWSRec->pBlkBuf[pRWSRec->dwOffInBlk], rdwSize);
    dwRet = ERROR_SUCCESS;
#else
    dwRet = ReadTarget(pRWSRec, pRWSRec->qwBlkOffset + (QWORD)pRWSRec->dwOffInBlk, pBuf, rdwSize);
#endif
    if ( dwRet == ERROR_SUCCESS )
    {
        pRWSRec->MesaRdWr.Finish(rdwSize);

        pRWSRec->dwOffInBlk += rdwSize;
        pRWSRec->dwLeftInBlk -= rdwSize;
    }

    rbEndOfBlk = !pRWSRec->dwLeftInBlk;

    return dwRet;
}

DWORD SeedTarget::WriteBlock(DWORD dwBlkId, PBYTE pBuf, DWORD dwSize, bool &rbEndOfBlk)
{
    DWORD  dwRet;
    UINT  nRWSIdx;
    PST_RWS_REC  pRWSRec;

    DBG(D_VERB, (L"%s: sid(%#.8x)\n", FULL_FUNC_NAME, dwBlkId));

    dwRet = VerifySessionId(dwBlkId, nRWSIdx, pRWSRec);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifySessionId failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    dwRet = VerifyReadWrite(false, nRWSIdx, pRWSRec);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifyReadWrite failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    if ( dwSize > pRWSRec->dwLeftInBlk )
    {
        DBG(D_WARN, (L"%s: size(%d) > need(%d)\n", FULL_FUNC_NAME, dwSize, pRWSRec->dwLeftInBlk));
        assert(0);
        return ERROR_INVALID_DATA;
    }

    assert((pRWSRec->dwOffInBlk + pRWSRec->dwLeftInBlk) <= _pSeedHdr->dwBlockSize);

    pRWSRec->MesaRdWr.Start();
#ifdef CONFIG_ST_BLK_WR_MMAP
    assert(pRWSRec->pBlkBuf);
    memcpy(&pRWSRec->pBlkBuf[pRWSRec->dwOffInBlk], pBuf, dwSize);
    dwRet = ERROR_SUCCESS;
#else
    dwRet = WriteTarget(pRWSRec, pRWSRec->qwBlkOffset + (QWORD)pRWSRec->dwOffInBlk, pBuf, dwSize);
#endif
    if ( dwRet == ERROR_SUCCESS )
    {
        pRWSRec->MesaRdWr.Finish(dwSize);

#ifdef CONFIG_ST_BLK_WR_VERIFY
        assert(pRWSRec->sha1);
        dwRet = pRWSRec->sha1->DoHash(pBuf, dwSize);
#endif

        pRWSRec->dwOffInBlk += dwSize;
        pRWSRec->dwLeftInBlk -= dwSize;
    }
    else
    {
        pRWSRec->MesaRdWr.Finish(0);
    }

    rbEndOfBlk = !pRWSRec->dwLeftInBlk;

    return dwRet;
}

DWORD SeedTarget::MapBlock(DWORD dwBlkId, PBYTE &rpBuf, DWORD &rdwSize)
{
    DWORD  dwRet;
    UINT  nRWSIdx;
    PST_RWS_REC  pRWSRec;

    DBG(D_VERB, (L"%s: sid(%#.8x)\n", FULL_FUNC_NAME, dwBlkId));

    dwRet = VerifySessionId(dwBlkId, nRWSIdx, pRWSRec);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifySessionId failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    rdwSize = pRWSRec->dwBlkSize;

    if ( pRWSRec->pBlkBuf )
    {
        rpBuf = pRWSRec->pBlkBuf;
        return ERROR_SUCCESS;
    }

    rpBuf = new(std::nothrow) BYTE[rdwSize];
    if ( !rpBuf )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc rpBuf[%d] failed\n", FULL_FUNC_NAME, rdwSize));
        goto err1;
    }

#ifdef CONFIG_ST_BLK_RD_FIO
    if ( nRWSIdx < _nRWSDivide )
    {
        dwRet = ReadTarget(pRWSRec, pRWSRec->qwBlkOffset, rpBuf, rdwSize);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: ReadTarget failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto err2;
        }
    }
#endif

    pRWSRec->pBlkBuf = rpBuf;

    return ERROR_SUCCESS;

#ifdef CONFIG_ST_BLK_RD_FIO
err2:
#endif
    delete []rpBuf;
    rpBuf = NULL;

err1:
    return dwRet;
}

DWORD SeedTarget::UnmapBlock(DWORD dwBlkId, PBYTE pBuf, DWORD dwSizeUsed)
{
    DWORD  dwRet;
    UINT  nRWSIdx;
    PST_RWS_REC  pRWSRec;
    bool  bWrite;

    DBG(D_VERB, (L"%s: sid(%#.8x)\n", FULL_FUNC_NAME, dwBlkId));

    dwRet = VerifySessionId(dwBlkId, nRWSIdx, pRWSRec);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifySessionId failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    bWrite = (nRWSIdx >= _nRWSDivide);

    if ( dwSizeUsed != pRWSRec->dwLeftInBlk )
    {
        DBG(D_WARN, (L"%s: block size not match\n", FULL_FUNC_NAME));
        goto out;
    }

    if ( bWrite )
    {
#ifdef CONFIG_ST_BLK_WR_FIO
        dwRet = WriteTarget(pRWSRec, pRWSRec->qwBlkOffset, pBuf, dwSizeUsed);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: write block failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto out;
        }
#endif

#ifdef CONFIG_ST_BLK_WR_VERIFY
        assert(pRWSRec->sha1);
        dwRet = pRWSRec->sha1->DoHash(pBuf, dwSizeUsed);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: hash block failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto out;
        }
#endif
    }

    pRWSRec->dwOffInBlk += dwSizeUsed;
    pRWSRec->dwLeftInBlk -= dwSizeUsed;

out:
    assert(pBuf == pRWSRec->pBlkBuf);
    if ( bWrite )
    {
#ifdef CONFIG_ST_BLK_WR_FIO
        delete []pBuf;
        pRWSRec->pBlkBuf = NULL;
#endif
    }
    else
    {
#ifdef CONFIG_ST_BLK_RD_FIO
        delete []pBuf;
        pRWSRec->pBlkBuf = NULL;
#endif
    }

    return ERROR_SUCCESS;
}

DWORD SeedTarget::CloseBlock(DWORD dwBlkId)
{
    DWORD  dwRet;
    UINT  nRWSIdx;
    PST_RWS_REC  pRWSRec;
    bool  bWrite, bMaskUpd;
#ifdef CONFIG_ST_BLK_WR_VERIFY
    SHA1_VAL  SHA1Val;
#endif

    DBG(D_VERB, (L"%s: sid(%#.8x)\n", FULL_FUNC_NAME, dwBlkId));

    dwRet = VerifySessionId(dwBlkId, nRWSIdx, pRWSRec);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: VerifySessionId failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    DBG(D_VERB, (L"%s: sid(%#.8x), blk(%d)\n", FULL_FUNC_NAME, dwBlkId, pRWSRec->nBlkIdx));

    if ( pRWSRec->nBlkIdx == INVALID_INDEX_VALUE )
    {
        DBG(D_WARN, (L"%s: idx(%d) not open\n", FULL_FUNC_NAME, nRWSIdx));
        assert(0);
        return ERROR_INVALID_STATE;
    }

    bWrite = (nRWSIdx >= _nRWSDivide);

    if ( pRWSRec->dwLeftInBlk )
    {
        DBG(D_WARN, (L"%s: blk(%d) not finished\n", FULL_FUNC_NAME, pRWSRec->nBlkIdx));
        dwRet = ERROR_CANCELLED;
        goto out;
    }

#ifdef CONFIG_ST_BLK_WR_VERIFY
    if ( bWrite )
    {
        assert(pRWSRec->sha1);
        dwRet = pRWSRec->sha1->GetHashVal(&SHA1Val);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: SHA1_GetHashVal() failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto out;
        }

        if ( SHA1Val != _pSeedHdr->SHA1ValBlk[pRWSRec->nBlkIdx] )
        {
            DBG(D_WARN, (L"%s: checksum not match\n", FULL_FUNC_NAME));
            dwRet = ERROR_INVALID_DATA;
            goto out;
        }
    }
#endif

    InterlockedIncrement(bWrite ? &_lBlkWrCnt : &_lBlkRdCnt);

    dwRet = ERROR_SUCCESS;

out:
    if ( bWrite )
    {
#ifdef CONFIG_ST_BLK_WR_MMAP
        assert(pRWSRec->pBlkBuf);
        // do it and then SetBit
        // do it before _baBlkProg(cache).SetBit
        _TargetMapping.Unmap(pRWSRec->pBlkBuf, pRWSRec->dwBlkSize);
#else
        // call UnmapBlock() first
        assert(!pRWSRec->pBlkBuf);

        // see issues_fsync.txt
        fflush(pRWSRec->hFile);
#endif

#ifdef CONFIG_ST_BLK_WR_VERIFY
        assert(pRWSRec->sha1);
        pRWSRec->sha1->EndHash();
#endif

        pthread_mutex_lock(&_hStateLock);
        if ( dwRet == ERROR_SUCCESS )
        {
            _baBlkProg.SetBit(pRWSRec->nBlkIdx, bMaskUpd);
            if ( bMaskUpd )
            {
                _bMaskUpdated = true;
                ReqExec();
            }
        }
        _baBlkBusy.ClearBit(pRWSRec->nBlkIdx);
        pthread_mutex_unlock(&_hStateLock);
    }
    else
    {
#ifdef CONFIG_ST_BLK_RD_MMAP
        assert(pRWSRec->pBlkBuf);
        // do it and then SetBit
        // do it before _baBlkProg(cache).SetBit
        _TargetMapping.Unmap(pRWSRec->pBlkBuf, pRWSRec->dwBlkSize);
#else
        // call UnmapBlock() first
        assert(!pRWSRec->pBlkBuf);
#endif
    }

    pRWSRec->nBlkIdx = INVALID_INDEX_VALUE;

    return dwRet;
}

void SeedTarget::Execute()
{
    bool  bBlkEvt;
    UINT  nDegree;
    QWORD  qwMaskAnd, qwMaskOr;
    list<PST_OBS_REC>::const_iterator  itr;
    list<PST_OBS_REC>::const_iterator  end_itr;
    PST_OBS_REC  pObsRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( (!_bObsAdded) && (!_bMaskUpdated) )
    {
        // fast check
        return;
    }

    pthread_mutex_lock(&_hStateLock);

    _bObsAdded = false;

    bBlkEvt = _bMaskUpdated;
    _bMaskUpdated = false;

    nDegree = _baBlkProg.GetDegree();
    _baBlkProg.GetCompactMask(qwMaskAnd, qwMaskOr);

    pthread_mutex_unlock(&_hStateLock);

    pthread_mutex_lock(&_hObsLock);
    for ( itr = _ObsList.begin(), end_itr = _ObsList.end(); itr != end_itr; ++itr )
    {
        pObsRec = *itr;
        if ( bBlkEvt || pObsRec->bNewOne )
        {
            pObsRec->bNewOne = false;
            pObsRec->pObserver->OnStateUpdated(this, pObsRec->dwCookie, nDegree, qwMaskAnd, qwMaskOr);
        }
    }
    pthread_mutex_unlock(&_hObsLock);
}

DWORD SeedTarget::ReadSeedFile(const char *szSeedFile)
{
    DWORD  dwRet;
    FILE  *hFile;
    PSEED_HDR  pHdr;
    UINT  nLeft;

    DBG(D_INFO, (L"%s: %hs\n", FULL_FUNC_NAME, szSeedFile));

    hFile = fopen(szSeedFile, "rb");
    if ( !hFile )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: open seed file(%hs) failed: %d\n", FULL_FUNC_NAME, szSeedFile, dwRet));
        goto out1;
    }

    nLeft = sizeof(SEED_HDR) - sizeof(SHA1_VAL);
    _biSeedHdr.Alloc(nLeft);
    if ( fread(_biSeedHdr.Put(nLeft), 1, nLeft, hFile) != nLeft )
    {
        dwRet = ferror(hFile) ? GetLastError() : ERROR_INVALID_DATA;
        DBG(D_WARN, (L"%s: read seed file(%hs) hdr failed: %d\n", FULL_FUNC_NAME, szSeedFile, dwRet));
        goto out2;
    }

    pHdr = (PSEED_HDR)_biSeedHdr.data();
    if ( pHdr->dwMagic != SEED_MAGIC_NUM )
    {
        dwRet = ERROR_INVALID_DATA;
        DBG(D_WARN, (L"%s: invalid seed file(%hs) magic number: %x\n", FULL_FUNC_NAME, szSeedFile, pHdr->dwMagic));
        goto out2;
    }

    nLeft = sizeof(SHA1_VAL) * pHdr->dwSHA1Num;
    _biSeedHdr.Expand(nLeft);  // not touch pHdr from now
    if ( fread(_biSeedHdr.Put(nLeft), 1, nLeft, hFile) != nLeft )
    {
        dwRet = ferror(hFile) ? GetLastError() : ERROR_INVALID_DATA;
        DBG(D_WARN, (L"%s: read seed file(%hs) body failed: %d\n", FULL_FUNC_NAME, szSeedFile, dwRet));
        goto out2;
    }

    _pSeedHdr = (PSEED_HDR)_biSeedHdr.data();

    dwRet = ERROR_SUCCESS;

out2:
    fclose(hFile);

out1:
    return dwRet;
}

DWORD SeedTarget::VerifySeedHdr()
{
    DWORD  dwRet;
    SHA1Hash  sha1;
    SHA1_VAL  SHA1Val;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    dwRet = sha1.BeginHash();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_BeginHash failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto out1;
    }

    for ( DWORD i = 0; i < _pSeedHdr->dwSHA1Num; ++i )
    {
        dwRet = sha1.DoHash(_pSeedHdr->SHA1ValBlk[i].bVal, sizeof(SHA1_VAL));
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: SHA1_DoHash failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto out2;
        }
    }

    dwRet = sha1.GetHashVal(&SHA1Val);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_GetHashVal failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto out2;
    }

    if ( _pSeedHdr->SHA1ValSeed != SHA1Val )
    {
        dwRet = ERROR_INVALID_DATA;
        DBG(D_WARN, (L"%s: SHA1 value not match\n", FULL_FUNC_NAME));
        goto out2;
    }

    dwRet = ERROR_SUCCESS;

out2:
    sha1.EndHash();

out1:
    return dwRet;
}

DWORD SeedTarget::OpenTargetFile(const char *szTargetFile, bool bReadOnly, bool &rbNeedVerify)
{
    DWORD  dwRet;
    const char  *szMode;
    bool  bRegularFile;
    QWORD  qwFileSize;

    DBG(D_INFO, (L"%s: %hs\n", FULL_FUNC_NAME, szTargetFile));

    szMode = bReadOnly ? "rb" : "r+b";
    _hTargetFile = fopen(szTargetFile, szMode);
    if ( (!_hTargetFile) && (!bReadOnly) )
    {
        szMode = "w+b";
        _hTargetFile = fopen(szTargetFile, szMode);
    }
    if ( !_hTargetFile )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: open target file(%hs) with mode(%hs) failed: %d\n", FULL_FUNC_NAME, szTargetFile, szMode, dwRet));
        goto err1;
    }

    dwRet = IsRegularFile(_hTargetFile, bRegularFile);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: IsRegularFile(%hs) failed: %d\n", FULL_FUNC_NAME, szTargetFile, dwRet));
        goto err2;
    }

    if ( !bRegularFile )
    {
        DBG(D_INFO, (L"%s: target file(%hs) is not a regular file\n", FULL_FUNC_NAME, szTargetFile));
        rbNeedVerify = true;
        return ERROR_SUCCESS;
    }

    dwRet = GetFileLength(_hTargetFile, qwFileSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: GetFileLength(%hs) failed: %d\n", FULL_FUNC_NAME, szTargetFile, dwRet));
        goto err2;
    }

    if ( qwFileSize == _pSeedHdr->qwFileSize )
    {
        rbNeedVerify = true;
        return ERROR_SUCCESS;
    }

    if ( bReadOnly )
    {
        dwRet = ERROR_INVALID_DATA;
        DBG(D_INFO, (L"%s: target file(%hs) size invalid: want(%lldB), but(%lldB)\n", FULL_FUNC_NAME, szTargetFile, _pSeedHdr->qwFileSize, qwFileSize));
        goto err2;
    }

    // cache exist:
    // 1: target exist, enable verify if not trust cache
    // 2: target not exist(include length not match), and bAllowVerify=false,
    //    MUST zero the cache, and MUST do it before SetFileLength(),
    //    because if lost power occur between SetFileLength() and zero cache...
    // scripts guarantee it and support os img tune featrue
//    _baBlkProg.ReInit(false);

    DBG(D_LOG, (L"%s: SetFileLength(%hs, %lld) may take long time...\n", FULL_FUNC_NAME, szTargetFile, _pSeedHdr->qwFileSize));
    dwRet = SetFileLength(_hTargetFile, _pSeedHdr->qwFileSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: SetFileLength(%hs, %lld) failed: %d\n", FULL_FUNC_NAME, szTargetFile, _pSeedHdr->qwFileSize, dwRet));
        goto err2;
    }

    rbNeedVerify = false;

    return ERROR_SUCCESS;

err2:
    fclose(_hTargetFile);
    _hTargetFile = NULL;

err1:
    return dwRet;
}

DWORD SeedTarget::VerifyTargetFile(bool bOneOnly)
{
    DWORD  dwRet = ERROR_SUCCESS;
    bool  bValid;

    DBG(D_INFO, (L"%s: total %d blocks, may take long time...\n", FULL_FUNC_NAME, _pSeedHdr->dwSHA1Num));

    for ( UINT i = 0; i < (UINT)_pSeedHdr->dwSHA1Num; ++i )
    {
        if ( bOneOnly && (!_baBlkProg.IsBitSet(i)) )
        {
            continue;
        }

        dwRet = VerifyTargetBlock(i, bValid);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: VerifyTargetBlock(%d) failed: %d\n", FULL_FUNC_NAME, i, dwRet));
            break;
        }

        if ( (!bOneOnly) || (!bValid) )
        {
            bValid ? _baBlkProg.SetBit(i) : _baBlkProg.ClearBit(i);
        }
    }

    return dwRet;
}

DWORD SeedTarget::VerifyTargetBlock(UINT nBlkIdx, bool &rbValid)
{
    DWORD  dwRet;
    off64_t  llOffset;
    SHA1Hash  sha1;
    SHA1_VAL  SHA1Val;
    BYTE  buf[1024];
    DWORD  dwLeft, dwBytes;

    DBG(D_VERB, (L"%s: block[[#%d/#%d]\n", FULL_FUNC_NAME, nBlkIdx, _pSeedHdr->dwSHA1Num));

    llOffset = (off64_t)_pSeedHdr->dwBlockSize * (off64_t)nBlkIdx;
    if ( fseeko(_hTargetFile, llOffset, SEEK_SET) )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: fseeko(%lld) failed: %d\n", FULL_FUNC_NAME, llOffset, dwRet));
        goto out1;
    }

    dwRet = sha1.BeginHash();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_BeginHash failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto out1;
    }

    dwLeft = _pSeedHdr->dwBlockSize;
    while ( dwLeft )
    {
        dwBytes = min(sizeof(buf), (size_t)dwLeft);
        dwBytes = (DWORD)fread(buf, 1, dwBytes, _hTargetFile);
        if ( dwBytes == 0 )
        {
            if ( ferror(_hTargetFile) )
            {
                dwRet = GetLastError();
                DBG(D_WARN, (L"%s: read block(%d) failed: %d\n", FULL_FUNC_NAME, nBlkIdx, dwRet));
            }
            break;
        }

        dwRet = sha1.DoHash(buf, dwBytes);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: SHA1_DoHash failed: %d\n", FULL_FUNC_NAME, dwRet));
            break;
        }

        dwLeft -= dwBytes;
    }
    if ( dwRet != ERROR_SUCCESS )
    {
        goto out2;
    }

    dwRet = sha1.GetHashVal(&SHA1Val);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: SHA1_GetHashVal failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto out2;
    }

    rbValid = (_pSeedHdr->SHA1ValBlk[nBlkIdx] == SHA1Val);

    dwRet = ERROR_SUCCESS;

out2:
    sha1.EndHash();

out1:
    return dwRet;
}

#ifdef CONFIG_ST_BLK_RD_FIO
DWORD SeedTarget::ReadTarget(PST_RWS_REC pRWSRec, QWORD qwOffset, PBYTE pBuf, DWORD dwSize)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %d @ %#llx\n", FULL_FUNC_NAME, dwSize, qwOffset));

    // see issues_fseek.txt
//    assert(qwOffset == (QWORD)ftello(hFile));
    assert(qwOffset == pRWSRec->qwCurOffset);

#ifdef CONFIG_ST_BLK_RD_BUF
    if ( fread(pBuf, 1, dwSize, pRWSRec->hFile) != dwSize )
#else
    if ( read(_fileno(pRWSRec->hFile), pBuf, dwSize) != (ssize_t)dwSize )
#endif
    {
        // maybe eof
        dwRet = ERROR_INVALID_DATA;//GetLastError();
        DBG(D_WARN, (L"%s: fread(%d @ %#llx) failed: %d\n", FULL_FUNC_NAME, dwSize, qwOffset, dwRet));
        return dwRet;
    }

    pRWSRec->qwCurOffset += dwSize;

    return ERROR_SUCCESS;
}
#endif

#ifdef CONFIG_ST_BLK_WR_FIO
DWORD SeedTarget::WriteTarget(PST_RWS_REC pRWSRec, QWORD qwOffset, PBYTE pBuf, DWORD dwSize)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %d @ %#llx\n", FULL_FUNC_NAME, dwSize, qwOffset));

    // see issues_fseek.txt
//    assert(qwOffset == (QWORD)ftello(hFile));
    assert(qwOffset == pRWSRec->qwCurOffset);

#ifdef CONFIG_ST_BLK_WR_BUF
    if ( fwrite(pBuf, 1, dwSize, pRWSRec->hFile) != dwSize )
#else
    if ( write(_fileno(pRWSRec->hFile), pBuf, dwSize) != (ssize_t)dwSize )
#endif
    {
        // see misc_fio.txt
        // use ERROR_DISK_FULL to notify critical error
        dwRet = ERROR_DISK_FULL;//GetLastError();
        DBG(D_WARN, (L"%s: fwrite(%d @ %#llx) failed: %d\n", FULL_FUNC_NAME, dwSize, qwOffset, dwRet));
        return dwRet;
    }

    pRWSRec->qwCurOffset += dwSize;

    return ERROR_SUCCESS;
}
#endif

DWORD SeedTarget::VerifySessionId(DWORD dwSessionId, UINT &rnRWSIdx, PST_RWS_REC &rpRWSRec)
{
    DWORD  dwRet;
    WORD  wMagic;

    rnRWSIdx = WIDX_FROM_DWID(dwSessionId);
    wMagic = WMAGICNUM_FROM_DWID(dwSessionId);

    DBG(D_VERB, (L"%s: idx(%d), magic(%d)\n", FULL_FUNC_NAME, rnRWSIdx, wMagic));

    assert(_hTargetFile);

    if ( rnRWSIdx >= _RWSList.size() )
    {
        dwRet = ERROR_INVALID_PARAMETER;
        DBG(D_WARN, (L"%s: invalid idx(%d)\n", FULL_FUNC_NAME, rnRWSIdx));
        goto err;
    }

    rpRWSRec = &_RWSList.at(rnRWSIdx);
    assert(rpRWSRec->bUsed);

    if ( wMagic != rpRWSRec->wMagic )
    {
        dwRet = ERROR_NOT_FOUND;
        DBG(D_WARN, (L"%s: idx(%d) not found\n", FULL_FUNC_NAME, rnRWSIdx));
        goto err;
    }

    return ERROR_SUCCESS;

err:
    rnRWSIdx = INVALID_INDEX_VALUE;
    rpRWSRec = NULL;

    return dwRet;
}

DWORD SeedTarget::VerifyReadWrite(bool bReadOrWrite, UINT nRWSIdx, PST_RWS_REC pRWSRec)
{
    DBG(D_VERB, (L"%s: r/w(%d), idx(%d)\n", FULL_FUNC_NAME, bReadOrWrite, nRWSIdx));

    if ( bReadOrWrite )
    {
        if ( nRWSIdx >= _nRWSDivide )
        {
            DBG(D_WARN, (L"%s: idx(%d) not support read\n", FULL_FUNC_NAME, nRWSIdx));
            assert(0);
            return ERROR_ACCESS_DENIED;
        }
    }
    else
    {
        if ( nRWSIdx < _nRWSDivide )
        {
            DBG(D_WARN, (L"%s: idx(%d) not support write\n", FULL_FUNC_NAME, nRWSIdx));
            assert(0);
            return ERROR_ACCESS_DENIED;
        }
    }

    if ( pRWSRec->nBlkIdx == INVALID_INDEX_VALUE )
    {
        DBG(D_WARN, (L"%s: idx(%d) not open\n", FULL_FUNC_NAME, nRWSIdx));
        assert(0);
        return ERROR_INVALID_STATE;
    }

    if ( !pRWSRec->dwLeftInBlk )
    {
        DBG(D_WARN, (L"%s: blk(%d) was ended\n", FULL_FUNC_NAME, pRWSRec->nBlkIdx));
        assert(0);
        return ERROR_INVALID_DATA;
    }

    return ERROR_SUCCESS;
}
