/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "SeedCenter.h"
#include "SeedTrackerHandler.h"


SeedTrackerHandler::SeedTrackerHandler()
    : EpPDUHandler(SEED_TRACK_NETQSIZE, SEED_TRACK_REQBSIZE,
                   SEED_TRACK_NETQSIZE, SEED_TRACK_REPBSIZE,
                   0, 0)
    , _dwNodeId(INVALID_NODE_ID)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

SeedTrackerHandler::~SeedTrackerHandler()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

void SeedTrackerHandler::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::Unbind();

    if ( _dwNodeId != INVALID_NODE_ID )
    {
        SeedCenter::Instance()->UnregisterNode(_dwNodeId);
        _dwNodeId = INVALID_NODE_ID;
    }
}

void SeedTrackerHandler::PrepareSend()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));
}

void SeedTrackerHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DWORD  dwRet;
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    assert(bReq);

    pbi = GetSendBuf();
    if ( !pbi )
    {
        // client should send the next req only after recv the previous rep,
        // so the send buf should never be used up.
        DBG(D_WARN, (L"%s: no free send buf\n", FULL_FUNC_NAME));
        assert(0);
        ReqEndSession();
        return;
    }

    switch ( bCode )
    {
        case C_ST_REG_NODE:
            assert(pbiBody->data_size() == sizeof(NODE_INFO));
            dwRet = OnRegisterNode((PNODE_INFO)pbiBody->data(), pbi);
            break;

        case C_ST_UNREG_NODE:
            assert(!pbiBody->data_size());
            dwRet = OnUnregisterNode(dwParam, pbi);
            break;

        case C_ST_ADD_N4S:
            assert(pbiBody->data_size() == sizeof(SEED_INFO));
            dwRet = OnAddNode4Seed(dwParam, (PSEED_INFO)pbiBody->data(), pbi);
            break;

        case C_ST_REMOVE_N4S:
            assert(!pbiBody->data_size());
            dwRet = OnRemoveNode4Seed(dwParam, pbi);
            break;

        case C_ST_UPDATE_N4S_STATE:
            assert(pbiBody->data_size() == sizeof(SEED_STATE));
            dwRet = OnUpdateNode4SeedState(dwParam, (PSEED_STATE)pbiBody->data(), pbi);
            break;

        case C_ST_ENUM_NODES4SEED:
            assert(pbiBody->data_size() == sizeof(REQ_ENUM_NODE4SEED));
            dwRet = OnEnumNodes4Seed(dwParam, (PREQ_ENUM_NODE4SEED)pbiBody->data(), pbi);
            break;

        default:
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, bCode));
            assert(0);
            dwRet = ERROR_NOT_SUPPORTED;
            break;
    }

    BuildPDU(bCode, false, true, dwRet, pbi);

    PutSendData(pbi, true);
}

void SeedTrackerHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

DWORD SeedTrackerHandler::OnRegisterNode(const PNODE_INFO pNodeInfo, BufferItem *pbiSend)
{
    DWORD  dwRet;
    PRET_REG_NODE  pRetRegNode;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    dwRet = SeedCenter::Instance()->RegisterNode(pNodeInfo, _dwNodeId);
    if ( dwRet == ERROR_SUCCESS )
    {
        pRetRegNode = (PRET_REG_NODE)pbiSend->Put(sizeof(RET_REG_NODE));
        pRetRegNode->dwNodeId = _dwNodeId;
    }

    return dwRet;
}

DWORD SeedTrackerHandler::OnUnregisterNode(DWORD dwNodeId, BufferItem *pbiSend)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(dwNodeId == _dwNodeId);

    return SeedCenter::Instance()->UnregisterNode(dwNodeId);
}

DWORD SeedTrackerHandler::OnAddNode4Seed(DWORD dwNodeId, const PSEED_INFO pSeedInfo, BufferItem *pbiSend)
{
    PRET_ADD_NODE4SEED  pRetAddN4S;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(dwNodeId == _dwNodeId);

    pRetAddN4S = (PRET_ADD_NODE4SEED)pbiSend->Put(sizeof(RET_ADD_NODE4SEED));

    return SeedCenter::Instance()->AddNode4Seed(dwNodeId, pSeedInfo, pRetAddN4S->dwN4SId);
}

DWORD SeedTrackerHandler::OnRemoveNode4Seed(DWORD dwN4SId, BufferItem *pbiSend)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return SeedCenter::Instance()->RemoveNode4Seed(dwN4SId);
}

DWORD SeedTrackerHandler::OnUpdateNode4SeedState(DWORD dwN4SId, const PSEED_STATE pSeedState, BufferItem *pbiSend)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return SeedCenter::Instance()->UpdateNode4SeedState(dwN4SId, pSeedState);
}

DWORD SeedTrackerHandler::OnEnumNodes4Seed(DWORD dwN4SId, const PREQ_ENUM_NODE4SEED pReqEnumN4S, BufferItem *pbiSend)
{
    DWORD  dwRet;
    UINT  nStart, nCount;
    PRET_ENUM_NODE4SEED  pRetEnumN4S;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pRetEnumN4S = (PRET_ENUM_NODE4SEED)pbiSend->Put(sizeof(RET_ENUM_NODE4SEED) - sizeof(NODE4SEED_INFO));
    nCount = pbiSend->tail_size() / sizeof(NODE4SEED_INFO);

    nStart = (UINT)pReqEnumN4S->dwStartIdx;
    
    dwRet = SeedCenter::Instance()->EnumNodes4Seed(dwN4SId, nStart, pRetEnumN4S->N4SInfo, nCount);
    if ( dwRet == ERROR_SUCCESS )
    {
        pRetEnumN4S->dwNextIdx = nStart;
        pRetEnumN4S->dwCount = nCount;
        pbiSend->Put(sizeof(NODE4SEED_INFO) * nCount);
    }

    return dwRet;
}
