/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "DeflateCodec.h"


#ifdef _WIN32_
#ifdef CONFIG_USE_LZ4
#pragma comment(lib, "liblz4.lib")
#endif
#ifdef CONFIG_USE_ZLIB_NG
#pragma comment(lib, "zngdll.lib")
#endif
#ifdef CONFIG_USE_ZLIB
#pragma comment(lib, "zdll.lib")
#endif
#endif


DeflateAlg::DeflateAlg()
    : _bInited(false)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

DeflateAlg::~DeflateAlg()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(!_bInited);
}

#ifdef CONFIG_USE_LZ4
DWORD DeflateAlg::Init(COMPRESS_LEVEL level)
{
    DWORD  dwRet;
    size_t  ret;

    DBG(D_INFO, (L"%s_LZ4: level=%d\n", FULL_FUNC_NAME, level));

    _MesaComp.Reset();

    ret = LZ4F_createCompressionContext(&_lz4fStrm, LZ4F_VERSION);
    if ( ret != 0 )
    {
        dwRet = ERROR_NOT_SUPPORTED;
        DBG(D_ERROR, (L"%s: LZ4F_createCompressionContext failed: %hs\n", FULL_FUNC_NAME, LZ4F_getErrorName(ret)));
        goto err1;
    }

    memset(&_lz4fPrefs, 0, sizeof(_lz4fPrefs));
    _lz4fPrefs.frameInfo.blockSizeID = LZ4F_max64KB;
    _lz4fPrefs.frameInfo.blockMode = LZ4F_blockIndependent;//Linked;  // faster than Linked
    _lz4fPrefs.frameInfo.contentChecksumFlag = LZ4F_noContentChecksum;
    _lz4fPrefs.frameInfo.frameType = LZ4F_frame;
    _lz4fPrefs.frameInfo.contentSize = 0; //content size unknown
    _lz4fPrefs.frameInfo.dictID = 0; //no dictID
    _lz4fPrefs.frameInfo.blockChecksumFlag = LZ4F_noBlockChecksum;
    _lz4fPrefs.compressionLevel = level;
    // see issues_lz4_af.txt
    _lz4fPrefs.autoFlush = 1;  // call GetBound() before ProcessXXX, GetBound() return a larger value when autoFlush=1

    memset(&_lz4fOpts, 0, sizeof(_lz4fOpts));

    _bNotBegin = true;

    _bInited = true;

    return ERROR_SUCCESS;

err1:
    return dwRet;
}

void DeflateAlg::Fini()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _bInited )
    {
        _bInited = false;
        LZ4F_freeCompressionContext(_lz4fStrm);
    }

    _MesaComp.Print("LZ4F_COMP");
}

void DeflateAlg::Reset()
{
    BYTE  buf[100], *p;
    UINT  nLeft;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_bNotBegin )
    {
        p = buf;
        nLeft = _countof(buf);
        DoLZ4FEnd(p, nLeft);
    }
}

UINT DeflateAlg::GetBound(UINT nSrc)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    // see issues_lz4_bound.txt
    return (UINT)(LZ4F_HEADER_SIZE_MAX + LZ4F_compressBound(nSrc, &_lz4fPrefs));
}

DWORD DeflateAlg::ProcessBlock(const PBYTE pSrc, UINT nSrc, PBYTE pDst, UINT &rnDst, bool bReset)
{
    DWORD  dwRet;
    UINT  nLeft = rnDst;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    dwRet = DoLZ4FProcess(pSrc, nSrc, pDst, nLeft, true, bReset);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: DoLZ4FProcess failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    rnDst -= nLeft;

    return ERROR_SUCCESS;
}

DWORD DeflateAlg::ProcessStream(PBYTE &rpSrc, UINT &rnSrc, PBYTE pDst, UINT &rnDst, bool &rbAgain, bool bTrunc)
{
    DWORD  dwRet;
    UINT  nLeft;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    nLeft = rnDst;
    dwRet = DoLZ4FProcess(rpSrc, rnSrc, pDst, nLeft, false, bTrunc);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: DoLZ4FProcess failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    rpSrc += rnSrc;
    rnSrc = 0;
    rnDst -= nLeft;

    rbAgain = false;

    return ERROR_SUCCESS;
}

DWORD DeflateAlg::DoLZ4FBegin(PBYTE &rpOut, UINT &rnLeft)
{
    size_t  ret;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( rnLeft < LZ4F_HEADER_SIZE_MAX )
    {
        DBG(D_ERROR, (L"%s: need at least: %d\n", FULL_FUNC_NAME, LZ4F_HEADER_SIZE_MAX));
        assert(0);
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    ret = LZ4F_compressBegin(_lz4fStrm, rpOut, rnLeft, &_lz4fPrefs);
    if ( LZ4F_isError(ret) )
    {
        DBG(D_ERROR, (L"%s: LZ4F_compressBegin failed: %hs\n", FULL_FUNC_NAME, LZ4F_getErrorName(ret)));
        assert(0);
        return ERROR_NOT_READY;
    }

    assert(rnLeft >= ret);
    rpOut += (UINT)ret;
    rnLeft -= (UINT)ret;

    _bNotBegin = false;

    return ERROR_SUCCESS;
}

DWORD DeflateAlg::DoLZ4FEnd(PBYTE &rpOut, UINT &rnLeft)
{
    size_t  ret;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    // necessarily >= 4 (endMark), or 8 if optional frame checksum is enabled
    if ( rnLeft < 4 )
    {
        DBG(D_ERROR, (L"%s: need at least: %d\n", FULL_FUNC_NAME, 4));
        assert(0);
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    ret = LZ4F_compressEnd(_lz4fStrm, rpOut, rnLeft, &_lz4fOpts);
    if ( LZ4F_isError(ret) )
    {
        DBG(D_ERROR, (L"%s: LZ4F_compressEnd failed: %hs\n", FULL_FUNC_NAME, LZ4F_getErrorName(ret)));
        assert(0);
        return ERROR_NOT_READY;
    }

    assert(rnLeft >= ret);
    rpOut += ret;
    rnLeft -= (UINT)ret;

    _bNotBegin = true;

    return ERROR_SUCCESS;
}

DWORD DeflateAlg::DoLZ4FProcess(const PBYTE pIn, UINT nIn, PBYTE &rpOut, UINT &rnLeft, bool bAutoFlush, bool bDoEnd)
{
    DWORD  dwRet;
    size_t  ret;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    _MesaComp.Start();

    if ( _bNotBegin )
    {
        // see issues_lz4_af.txt
        _lz4fPrefs.autoFlush = (bAutoFlush ? 1 : 0);

        dwRet = DoLZ4FBegin(rpOut, rnLeft);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: DoLZ4FBegin failed: %d\n", FULL_FUNC_NAME, dwRet));
            return dwRet;
        }
    }

    ret = LZ4F_compressUpdate(_lz4fStrm, rpOut, rnLeft, pIn, nIn, &_lz4fOpts);
    if ( LZ4F_isError(ret) )
    {
        DBG(D_ERROR, (L"%s: LZ4F_compressUpdate failed: %hs\n", FULL_FUNC_NAME, LZ4F_getErrorName(ret)));
        assert(0);
        return ERROR_NOT_READY;
    }
    assert(rnLeft >= ret);
    rpOut += ret;
    rnLeft -= (UINT)ret;

    if ( bDoEnd )
    {
        dwRet = DoLZ4FEnd(rpOut, rnLeft);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_ERROR, (L"%s: DoLZ4FEnd failed: %d\n", FULL_FUNC_NAME, dwRet));
            return dwRet;
        }
    }

    _MesaComp.Finish(nIn);

    return ERROR_SUCCESS;
}
#endif

#if defined(CONFIG_USE_ZLIB_NG) || defined(CONFIG_USE_ZLIB)
DWORD DeflateAlg::Init(COMPRESS_LEVEL level)
{
    DWORD  dwRet;
    int  ret;

    DBG(D_INFO, (L"%s_Z(NG): level=%d\n", FULL_FUNC_NAME, level));

    _zlibStrm.zalloc = Z_NULL;
    _zlibStrm.zfree = Z_NULL;
    _zlibStrm.opaque = Z_NULL;
    ret = ZLIB_PREFIX(deflateInit2)(&_zlibStrm, level, Z_DEFLATED, -MAX_WBITS, MAX_MEM_LEVEL, Z_DEFAULT_STRATEGY);
    if ( ret != Z_OK )
    {
        dwRet = ERROR_NOT_SUPPORTED;
        DBG(D_ERROR, (L"%s: deflateInit2 failed: %d\n", FULL_FUNC_NAME, ret));
        goto err1;
    }

    _bInited = true;

    return ERROR_SUCCESS;

err1:
    return dwRet;
}

void DeflateAlg::Fini()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _bInited )
    {
        _bInited = false;
        ZLIB_PREFIX(deflateEnd)(&_zlibStrm);
    }
}

void DeflateAlg::Reset()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    ZLIB_PREFIX(deflateReset)(&_zlibStrm);
}

UINT DeflateAlg::GetBound(UINT nSrc)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return ZLIB_PREFIX(deflateBound)(&_zlibStrm, nSrc);
}

DWORD DeflateAlg::ProcessBlock(const PBYTE pSrc, UINT nSrc, PBYTE pDst, UINT &rnDst, bool bReset)
{
    int  ret;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    _zlibStrm.next_in = pSrc;
    _zlibStrm.avail_in = nSrc;
    _zlibStrm.next_out = pDst;
    _zlibStrm.avail_out = rnDst;
    ret = ZLIB_PREFIX(deflate)(&_zlibStrm, (bReset ? Z_FULL_FLUSH : Z_SYNC_FLUSH));
    if ( ret != Z_OK )
    {
        DBG(D_ERROR, (L"%s: deflate failed: %d\n", FULL_FUNC_NAME, ret));
        assert(0);
        return ERROR_NOT_READY;
    }

    // see misc_zlib.txt
    if ( _zlibStrm.avail_in/* || (!_zlibStrm.avail_out)*/ )
    {
        DBG(D_WARN, (L"%s: deflate need more room\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    rnDst -= _zlibStrm.avail_out;

    return ERROR_SUCCESS;
}

DWORD DeflateAlg::ProcessStream(PBYTE &rpSrc, UINT &rnSrc, PBYTE pDst, UINT &rnDst, bool &rbAgain, bool bTrunc)
{
    int  ret;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    _zlibStrm.next_in = rpSrc;
    _zlibStrm.avail_in = rnSrc;
    _zlibStrm.next_out = pDst;
    _zlibStrm.avail_out = rnDst;
    ret = ZLIB_PREFIX(deflate)(&_zlibStrm, (bTrunc ? Z_FULL_FLUSH : Z_NO_FLUSH));
    if ( (ret != Z_OK) && (ret != Z_BUF_ERROR) )
    {
        DBG(D_ERROR, (L"%s: deflate failed: %d\n", FULL_FUNC_NAME, ret));
        assert(0);
        return ERROR_NOT_READY;
    }

    rpSrc = (PBYTE)_zlibStrm.next_in;
    rnSrc = _zlibStrm.avail_in;
    rnDst -= _zlibStrm.avail_out;

    // see misc_zlib.txt
    rbAgain = (_zlibStrm.avail_out == 0);

    return ERROR_SUCCESS;
}
#endif

#ifdef CONFIG_DEFLATE_USE_ZNULL
DWORD DeflateAlg::Init(COMPRESS_LEVEL level)
{
    DBG(D_INFO, (L"%s_ZNULL: level=%d\n", FULL_FUNC_NAME, level));

    _bInited = true;

    return ERROR_SUCCESS;
}

void DeflateAlg::Fini()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _bInited = false;
}

void DeflateAlg::Reset()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));
}

UINT DeflateAlg::GetBound(UINT nSrc)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    return nSrc;
}

DWORD DeflateAlg::ProcessBlock(const PBYTE pSrc, UINT nSrc, PBYTE pDst, UINT &rnDst, bool bReset)
{

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( nSrc > rnDst )
    {
        DBG(D_WARN, (L"%s: deflate need more room\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    memcpy(pDst, pSrc, nSrc);
    rnDst = nSrc;

    return ERROR_SUCCESS;
}

DWORD DeflateAlg::ProcessStream(PBYTE &rpSrc, UINT &rnSrc, PBYTE pDst, UINT &rnDst, bool &rbAgain, bool bTrunc)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( rnSrc > rnDst )
    {
        DBG(D_WARN, (L"%s: deflate need more room\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    memcpy(pDst, rpSrc, rnSrc);
    rnDst = rnSrc;
    rpSrc += rnSrc;
    rnSrc = 0;

    rbAgain = false;

    return ERROR_SUCCESS;
}
#endif

DWORD DeflateAlg::ProcessBlock(const BufferItem *pbiSrc, BufferItem *pbiDst, bool bReset)
{
    DWORD  dwRet;
    UINT  nDst;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    nDst = pbiDst->tail_size();
    dwRet = ProcessBlock(pbiSrc->data(), pbiSrc->data_size(), pbiDst->tail(), nDst, bReset);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: Process failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    pbiDst->Put(nDst);

    return ERROR_SUCCESS;
}

DWORD DeflateAlg::ProcessStream(BufferItem *pbiSrc, BufferItem *pbiDst, bool &rbAgain, bool bTrunc)
{
    DWORD  dwRet;
    PBYTE  pSrc;
    UINT  nSrc, nLeft, nDst;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    pSrc = pbiSrc->data();
    nSrc = pbiSrc->data_size();
    nLeft = nSrc;
    nDst = pbiDst->tail_size();
    dwRet = ProcessStream(pSrc, nLeft, pbiDst->tail(), nDst, rbAgain, bTrunc);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: Process failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    pbiSrc->Pull(nSrc - nLeft);
    pbiDst->Put(nDst);

    return ERROR_SUCCESS;
}


InflateAlg::InflateAlg()
    : _bInited(false)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

}

InflateAlg::~InflateAlg()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(!_bInited);
}

#ifdef CONFIG_USE_LZ4
DWORD InflateAlg::Init()
{
    DWORD  dwRet;
    size_t  ret;

    DBG(D_INFO, (L"%s_LZ4\n", FULL_FUNC_NAME));

    _MesaDecomp.Reset();

    ret = LZ4F_createDecompressionContext(&_lz4fStrm, LZ4F_VERSION);
    if ( ret != 0 )
    {
        dwRet = ERROR_NOT_SUPPORTED;
        DBG(D_ERROR, (L"%s: LZ4F_createDecompressionContext failed: %hs\n", FULL_FUNC_NAME, LZ4F_getErrorName(ret)));
        goto err1;
    }

    memset(&_lz4fOpts, 0, sizeof(_lz4fOpts));

    _bInited = true;

    return ERROR_SUCCESS;

err1:
    return dwRet;
}

void InflateAlg::Fini()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _bInited )
    {
        _bInited = false;
        LZ4F_freeDecompressionContext(_lz4fStrm);
    }

    _MesaDecomp.Print("LZ4F_DECOMP");
}

void InflateAlg::Reset()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    LZ4F_resetDecompressionContext(_lz4fStrm);
}

DWORD InflateAlg::Process(PBYTE &rpSrc, UINT &rnSrc, PBYTE pDst, UINT &rnDst, bool &rbAgain, bool bLast)
{
    size_t  ret;
    size_t  sSrc, sDst;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    _MesaDecomp.Start();

    sSrc = rnSrc;
    sDst = rnDst;
    ret = LZ4F_decompress(_lz4fStrm, pDst, &sDst, rpSrc, &sSrc, &_lz4fOpts);
    if ( LZ4F_isError(ret) )
    {
        DBG(D_ERROR, (L"%s: LZ4F_decompress failed: %hs\n", FULL_FUNC_NAME, LZ4F_getErrorName(ret)));
        assert(0);
        return ERROR_NOT_READY;
    }

    rpSrc += sSrc;
    rnSrc -= (UINT)sSrc;

    // see issues_lz4_flush.txt
    rbAgain = (rnSrc > 0) || (bLast && (rnDst == (UINT)sDst));

    rnDst = (UINT)sDst;

    _MesaDecomp.Finish(sSrc);

    return ERROR_SUCCESS;
}
#endif

#if defined(CONFIG_USE_ZLIB_NG) || defined(CONFIG_USE_ZLIB)
DWORD InflateAlg::Init()
{
    DWORD  dwRet;
    int  ret;

    DBG(D_INFO, (L"%s_Z(NG)\n", FULL_FUNC_NAME));

    _zlibStrm.next_in = Z_NULL;
    _zlibStrm.avail_in = 0;
    _zlibStrm.zalloc = Z_NULL;
    _zlibStrm.zfree = Z_NULL;
    _zlibStrm.opaque = Z_NULL;
    ret = ZLIB_PREFIX(inflateInit2)(&_zlibStrm, -MAX_WBITS);
    if ( ret != Z_OK )
    {
        dwRet = ERROR_NOT_READY;
        DBG(D_ERROR, (L"%s: inflateInit2 failed: %d\n", FULL_FUNC_NAME, ret));
        goto err1;
    }

    _bInited = true;

    return ERROR_SUCCESS;

err1:
    return dwRet;
}

void InflateAlg::Fini()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _bInited )
    {
        _bInited = false;
        ZLIB_PREFIX(inflateEnd)(&_zlibStrm);
    }
}

void InflateAlg::Reset()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    ZLIB_PREFIX(inflateReset)(&_zlibStrm);
}

DWORD InflateAlg::Process(PBYTE &rpSrc, UINT &rnSrc, PBYTE pDst, UINT &rnDst, bool &rbAgain, bool bLast)
{
    int  ret;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    _zlibStrm.next_in = rpSrc;
    _zlibStrm.avail_in = rnSrc;
    _zlibStrm.next_out = pDst;
    _zlibStrm.avail_out = rnDst;
    // see misc_flush.txt
    ret = ZLIB_PREFIX(inflate)(&_zlibStrm, (bLast ? Z_SYNC_FLUSH : Z_NO_FLUSH));
    if ( (ret != Z_OK) && (ret != Z_BUF_ERROR) )
    {
        DBG(D_ERROR, (L"%s: inflate failed: %d\n", FULL_FUNC_NAME, ret));
        assert(0);
        return ERROR_NOT_READY;
    }

    rpSrc = (PBYTE)_zlibStrm.next_in;
    rnSrc = _zlibStrm.avail_in;
    rnDst -= _zlibStrm.avail_out;

    // see misc_zlib.txt
    rbAgain = (_zlibStrm.avail_out == 0);

    return ERROR_SUCCESS;
}
#endif

#ifdef CONFIG_DEFLATE_USE_ZNULL
DWORD InflateAlg::Init()
{
    DBG(D_INFO, (L"%s_ZNULL\n", FULL_FUNC_NAME));

    _bInited = true;

    return ERROR_SUCCESS;
}

void InflateAlg::Fini()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _bInited = false;
}

void InflateAlg::Reset()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));
}

DWORD InflateAlg::Process(PBYTE &rpSrc, UINT &rnSrc, PBYTE pDst, UINT &rnDst, bool &rbAgain, bool bLast)
{
    UINT  nBytes;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    nBytes = min(rnSrc, rnDst);
    memcpy(pDst, rpSrc, nBytes);
    rpSrc += nBytes;
    rnSrc -= nBytes;
    rnDst = nBytes;

    rbAgain = false;

    return ERROR_SUCCESS;
}
#endif

DWORD InflateAlg::Process(PBYTE &rpSrc, UINT &rnSrc, BufferItem *pbiDst, bool &rbAgain, bool bLast)
{
    DWORD  dwRet;
    UINT  nDst;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    nDst = pbiDst->tail_size();
    dwRet = Process(rpSrc, rnSrc, pbiDst->tail(), nDst, rbAgain, bLast);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: Process failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    pbiDst->Put(nDst);

    return ERROR_SUCCESS;
}

DWORD InflateAlg::Process(BufferItem *pbiSrc, BufferItem *pbiDst, bool &rbAgain, bool bLast)
{
    DWORD  dwRet;
    PBYTE  pSrc;
    UINT  nSrc, nLeft, nDst;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    pSrc = pbiSrc->data();
    nSrc = pbiSrc->data_size();
    nLeft = nSrc;
    nDst = pbiDst->tail_size();
    dwRet = Process(pSrc, nLeft, pbiDst->tail(), nDst, rbAgain, bLast);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_ERROR, (L"%s: Process failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    pbiSrc->Pull(nSrc - nLeft);
    pbiDst->Put(nDst);

    return ERROR_SUCCESS;
}
