/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "MUdfCommDef.h"
#include "MUdfUploadAgent.h"
#ifdef _WIN32_
#include "MUdfUploadAgentNt.h"
#endif
#include "MUdfDownloadAgent.h"
#include "MUdfClientEp.h"


MUdfClientEp *MUdfClientEp::_instance = NULL;

MUdfClientEp *MUdfClientEp::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new MUdfClientEp();
    }
    return _instance;
}

void MUdfClientEp::Destroy()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

MUdfClientEp::MUdfClientEp()
    : Endpoint()
    , _dwDoneRet(INVALID_MUEP_RETVAL)
    , _dwConnId(INVALID_EP_CONN_ID)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    sem_init(&_hDoneEvt, 0, 0);
}

MUdfClientEp::~MUdfClientEp()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    sem_destroy(&_hDoneEvt);
}

DWORD MUdfClientEp::Load()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    // StartUpload may occur before the previous conn free
    Endpoint::Load(2, 0, 0);

    return ERROR_SUCCESS;
}

void MUdfClientEp::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    Endpoint::Unload();
}

DWORD MUdfClientEp::StartUpload(const char *szServerIP, MUdfTarget *pMUdfTarget, MU_TARGET_TYPE mutType, const char *szRemoteMdf, const char *szRemoteUdf, bool bUpdateMode)
{
    DWORD  dwRet;
    MUdfUploadAgent  *pHandler;

    DBG(D_INFO, (L"%s: mut(%d) -> mdf(%hs)/udf(%hs) @ %hs\n", FULL_FUNC_NAME, mutType, szRemoteMdf, szRemoteUdf, szServerIP));

    pHandler = new(std::nothrow) MUdfUploadAgent(pMUdfTarget, szRemoteMdf, szRemoteUdf, bUpdateMode);
    if ( pHandler == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc MUdfUploadAgent failed\n", FULL_FUNC_NAME));
        goto err1;
    }

    dwRet = DoConnect(szServerIP, mutType, pHandler, _dwConnId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: DoConnect(%hs) failed: %d\n", FULL_FUNC_NAME, szServerIP, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    delete pHandler;

err1:
    return dwRet;
}

#ifdef _WIN32_
DWORD MUdfClientEp::StartUpload(const char *szServerIP, HANDLE hPsMUdfCtrl, HANDLE hPsMUdfFile, const char *szRemoteMdf, const char *szRemoteUdf)
{
    DWORD  dwRet;
    MUdfUploadAgentNt  *pHandler;

    DBG(D_INFO, (L"%s: -> mdf(%hs)/udf(%hs) @ %hs\n", FULL_FUNC_NAME, szRemoteMdf, szRemoteUdf, szServerIP));

    pHandler = new(std::nothrow) MUdfUploadAgentNt(hPsMUdfCtrl, hPsMUdfFile, szRemoteMdf, szRemoteUdf);
    if ( pHandler == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc MUdfUploadAgent failed\n", FULL_FUNC_NAME));
        goto err1;
    }

    dwRet = DoConnect(szServerIP, T_MUT_PS_UPLOAD, pHandler, _dwConnId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: DoConnect(%hs) failed: %d\n", FULL_FUNC_NAME, szServerIP, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    delete pHandler;

err1:
    return dwRet;
}
#endif

DWORD MUdfClientEp::StartDownload(const char *szServerIP, MUdfTarget *pMUdfTarget, MU_TARGET_TYPE mutType, const char *szRemoteMdf, const char *szRemoteUdf, const char *szLocalMdf, const char *szLocalUdf)
{
    DWORD  dwRet;
	MUdfDownloadAgent  *pHandler;

    DBG(D_INFO, (L"%s: mdf(%hs)/udf(%hs)/mut(%d) <- mdf(%hs)/udf(%hs) @ %hs\n", FULL_FUNC_NAME, szLocalMdf, szLocalUdf, mutType, szRemoteMdf, szRemoteUdf, szServerIP));

    pHandler = new(std::nothrow) MUdfDownloadAgent(pMUdfTarget, szLocalMdf, szLocalUdf, szRemoteMdf, szRemoteUdf);
    if ( pHandler == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc MUdfDownloadAgent failed\n", FULL_FUNC_NAME));
        goto err1;
    }

    dwRet = DoConnect(szServerIP, mutType, pHandler, _dwConnId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: DoConnect(%hs) failed: %d\n", FULL_FUNC_NAME, szServerIP, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    delete pHandler;

err1:
    return dwRet;
}

void MUdfClientEp::CancelUpload()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    // not serialization and synchronization, but no matter
    OnFinished(ERROR_CANCELLED);
}

DWORD MUdfClientEp::WaitForFinished(DWORD dwSeconds)
{
    timespec  tsWait;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( dwSeconds == 0 )
    {
        sem_wait(&_hDoneEvt);
    }
    else
    {
        clock_gettime(CLOCK_REALTIME, &tsWait);
        tsWait.tv_sec += dwSeconds;
        if ( sem_timedwait(&_hDoneEvt, &tsWait) == -1 )
        {
            return WAIT_TIMEOUT;
        }
    }

    // sem has no function like ResetEvent
    while ( sem_trywait(&_hDoneEvt) != -1 );

    return _dwDoneRet;
}

void MUdfClientEp::OnListenerEnded(DWORD hListener)
{
    DBG(D_LOG, (L"%s: id(%x)\n", FULL_FUNC_NAME, hListener));

    assert(0);
}

EpProtocolHandler *MUdfClientEp::OnNewHandler(DWORD hListener, Connection *pConnection, DWORD dwConnId)
{
    DBG(D_LOG, (L"%s: listener(%x), connid(%x)\n", FULL_FUNC_NAME, hListener, dwConnId));

    assert(0);

    return NULL;
}

void MUdfClientEp::OnAsyncConnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_LOG, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));
}

void MUdfClientEp::OnDisconnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_INFO, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));

    // serialization and synchronization under conn's execute context
    OnFinished(ERROR_CONNECTION_ABORTED);

    delete pHandler;
}

void MUdfClientEp::OnConnEvent(DWORD dwConnId, EpProtocolHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody)
{
    DBG(D_INFO, (L"%s: id(%#.8x), code(%d)\n", FULL_FUNC_NAME, dwConnId, pEvtHdr->nCode));

    switch ( pEvtHdr->nCode )
    {
        case C_MUDF_TRANS_END:
            // serialization and synchronization under conn's execute context
            OnFinished(pEvtHdr->dwParam);
            DisconnectConn(dwConnId);
            break;

        default:
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, pEvtHdr->nCode));
            assert(0);
            break;
    }
}

void MUdfClientEp::ProcessEpWork(UINT nSigIdx)
{
    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, nSigIdx));

    assert(0);
}

DWORD MUdfClientEp::DoConnect(const char *szServerIP, MU_TARGET_TYPE mutType, EpProtocolHandler *pHandler, DWORD &rdwConnId)
{
    DWORD  dwRet;
    WORD  wPort;
    SOCK_ADDR  laddr, raddr;

    DBG(D_INFO, (L"%s: %hs\n", FULL_FUNC_NAME, szServerIP));

    switch ( mutType )
    {
        case T_MUT_PS_UPLOAD:
            wPort = MUDFPS_UPLOAD_PORT;
            break;

        case T_MUT_OS_UPLOAD:
            wPort = MUDFOS_UPLOAD_PORT;
            break;

        case T_MUT_PS_DOWNLOAD:
            wPort = MUDFPS_DOWNLOAD_PORT;
            break;

        case T_MUT_OS_DOWNLOAD:
            wPort = MUDFOS_DOWNLOAD_PORT;
            break;

        default:
            assert(0);
            return ERROR_INVALID_PARAMETER;
    }

    laddr.bType = ADDR_TYPE_IPV4;
    laddr.IPv4Addr.bStream = 1;
    laddr.IPv4Addr.dwIP = 0;
    laddr.IPv4Addr.wPort = 0;

    raddr.bType = ADDR_TYPE_IPV4;
    raddr.IPv4Addr.bStream = 1;
    raddr.IPv4Addr.dwIP = inet_addr(szServerIP);
    raddr.IPv4Addr.wPort = htons(wPort);

    _dwDoneRet = INVALID_MUEP_RETVAL;
    dwRet = ConnectOut(&laddr, &raddr, pHandler, rdwConnId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ConnectOut(%hs) failed: %d\n", FULL_FUNC_NAME, szServerIP, dwRet));
    }

    return dwRet;
}

void MUdfClientEp::OnFinished(DWORD dwCode)
{
    DBG(D_INFO, (L"%s: pre=0x%x, cur=0x%x\n", FULL_FUNC_NAME, _dwDoneRet, dwCode));

    if ( _dwDoneRet == INVALID_MUEP_RETVAL )
    {
        _dwDoneRet = dwCode;
        sem_post(&_hDoneEvt);
    }
}
