/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "MUdfCommDef.h"
#include "MUdfDownloadAgent.h"


MUdfDownloadAgent::MUdfDownloadAgent(MUdfTarget *pMUdfTarget, const char *szLocalMdf, const char *szLocalUdf, const char *szRemoteMdf, const char *szRemoteUdf)
    : EpPDUHandler(MUDF_UPDOWN_DATAQSIZE, MUDF_UPDOWN_DATABSIZE,
                   MUDF_UPDOWN_CTRLQSIZE, MUDF_UPDOWN_DATABSIZE,
                   MUDF_UPDOWN_EVTQSIZE, MUDF_UPDOWN_EVTBSIZE)
    , _pMUdfTarget(pMUdfTarget)
    , _strLocalMdf(szLocalMdf)
    , _strLocalUdf(szLocalUdf)
    , _strRemoteMdf(szRemoteMdf)
    , _strRemoteUdf(szRemoteUdf)
    , _nBlkIdx(INVALID_MDFIDX_VAL)
    , _dwBlkId(INVALID_MUT_BLK_ID)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MUdfDownloadAgent::~MUdfDownloadAgent()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_MUT_BLK_ID);
    assert(_nBlkIdx == INVALID_MDFIDX_VAL);
}

DWORD MUdfDownloadAgent::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

#ifdef CONFIG_MUT_BLK_COMPACT
    dwRet = _DecompAlg.Init();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init decompressor failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    DBG(D_INFO, (L"%s: alloc bytes[%d] for write buf\n", FULL_FUNC_NAME, COMPACT_CLUSTER_SIZE));
    _DecompBuf.Alloc(COMPACT_CLUSTER_SIZE);
#endif

    dwRet = EpPDUHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
#ifdef CONFIG_MUT_BLK_COMPACT
    _DecompAlg.Fini();

err1:
#endif
    return dwRet;
}

void MUdfDownloadAgent::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::Unbind();

    OnFiniMUdfTarget();

#ifdef CONFIG_MUT_BLK_COMPACT
    _DecompAlg.Fini();
#endif
}

void MUdfDownloadAgent::OnAsyncConnected(Connection *pConnection)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::OnAsyncConnected(pConnection);

    dwRet = DownloadStartReq();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: DownloadStartReq failed: %d\n", FULL_FUNC_NAME, dwRet));
        OnDownloadFinished(dwRet);
    }
}

void MUdfDownloadAgent::PrepareSend()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

void MUdfDownloadAgent::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    if ( dwParam != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: error: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));
        goto out;
    }

    switch ( bCode )
    {
        case C_MUDFTARGET_DOWNLOAD:
            assert(pbiBody->data_size() == sizeof(REP_DOWNLOAD_MUDFTARGET));
            dwParam = OnInitMUdfTarget((PREP_DOWNLOAD_MUDFTARGET)pbiBody->data());
            break;

        case C_MUDFTARGET_BLOCK_START:
            assert(pbiBody->data_size() == sizeof(MUDFTARGET_BLOCK_INFO));
            dwParam = OnOpenMUdfBlock((PMUDFTARGET_BLOCK_INFO)pbiBody->data());
            break;

        case C_MUDFTARGET_FILE_END:
            dwParam = OnFiniMUdfTarget();
            break;

        default:
            dwParam = ERROR_NOT_SUPPORTED;
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, bCode));
            assert(0);
            break;
    }

out:
    if ( (bCode == C_MUDFTARGET_FILE_END) || (dwParam != ERROR_SUCCESS) )
    {
        OnDownloadFinished(dwParam);
    }
}

void MUdfDownloadAgent::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    bool  bLastOne, bEndOfBlk;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert((bCode == C_MUDFTARGET_BLOCK_DATA) || (bCode == C_MUDFTARGET_BLOCK_END));
    assert(bReq);

    if ( dwParam != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: block data error\n", FULL_FUNC_NAME));
        goto out1;
    }

    if ( _dwBlkId == INVALID_MUT_BLK_ID )
    {
        dwParam = ERROR_INVALID_STATE;
        DBG(D_WARN, (L"%s: pre open/write block failed\n", FULL_FUNC_NAME));
        goto out1;
    }

    // not use bEndOfBlk to judge whether CloseBlock & SendResponse
    bLastOne = (bCode == C_MUDFTARGET_BLOCK_END) && (nLeftBytes == 0);

#ifdef CONFIG_MUT_BLK_COMPACT
    dwParam = InflateBlock(pData, nDataBytes, bLastOne, bEndOfBlk);
#else
    dwParam = _pMUdfTarget->WriteBlock(_dwBlkId, pData, nDataBytes, bEndOfBlk);
#endif
    if ( dwParam != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: inflate or write block[%d] failed: %d\n", FULL_FUNC_NAME, _nBlkIdx, dwParam));
        goto out2;
    }

    assert(bEndOfBlk == bLastOne);

out2:
    if ( (dwParam != ERROR_SUCCESS) || bLastOne )
    {
        // WriteBlock fail, CloseBlock will fail too
        // WriteBlock OK, CloseBlock may fail
        dwParam = _pMUdfTarget->CloseBlock(_dwBlkId);
        _dwBlkId = INVALID_MUT_BLK_ID;
        _nBlkIdx = INVALID_MDFIDX_VAL;
    }

out1:
    if ( dwParam != ERROR_SUCCESS )
    {
        OnDownloadFinished(dwParam);
    }
}

DWORD MUdfDownloadAgent::DownloadStartReq()
{
    BufferItem  *pbi;
    PREQ_DOWNLOAD_MUDFTARGET  pDownloadReq;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pbi = GetSendBuf();
    assert(pbi != NULL);

    pDownloadReq = (PREQ_DOWNLOAD_MUDFTARGET)pbi->Put(sizeof(REQ_DOWNLOAD_MUDFTARGET));
    strncpy(pDownloadReq->szMdfPath, _strRemoteMdf.c_str(), MAX_MUDF_PATH_LEN);
    strncpy(pDownloadReq->szUdfPath, _strRemoteUdf.c_str(), MAX_MUDF_PATH_LEN);

    BuildPDU(C_MUDFTARGET_DOWNLOAD, true, true, 0, pbi);
    PutSendData(pbi, true);

    return ERROR_SUCCESS;
}

DWORD MUdfDownloadAgent::OnInitMUdfTarget(const PREP_DOWNLOAD_MUDFTARGET pRepOpenMUT)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s: mdf(%d,%hs), udf(%hs)\n", FULL_FUNC_NAME, pRepOpenMUT->dwMdfSize, _strLocalMdf.c_str(), _strLocalUdf.c_str()));

    dwRet = _pMUdfTarget->Init(_strLocalMdf.c_str(), _strLocalUdf.c_str(), M_MUT_OVERWRITE, pRepOpenMUT->dwMdfSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: open mudftarget failed\n", FULL_FUNC_NAME));
        goto err1;
    }

    return ERROR_SUCCESS;

err1:
    return dwRet;
}

DWORD MUdfDownloadAgent::OnFiniMUdfTarget()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( _dwBlkId != INVALID_MUT_BLK_ID )
    {
        _pMUdfTarget->CloseBlock(_dwBlkId);
        _dwBlkId = INVALID_MUT_BLK_ID;
        _nBlkIdx = INVALID_MDFIDX_VAL;
    }

    _pMUdfTarget->Fini();

    return ERROR_SUCCESS;
}

DWORD MUdfDownloadAgent::OnOpenMUdfBlock(const PMUDFTARGET_BLOCK_INFO pMUTBlkInfo)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: mdfidx=%d, udfoff=%d\n", FULL_FUNC_NAME, pMUTBlkInfo->dwMdfIdx, pMUTBlkInfo->dwUdfOff));

    if ( _dwBlkId != INVALID_MUT_BLK_ID )
    {
        DBG(D_WARN, (L"%s: previous block not closed\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_INVALID_STATE;
    }

    dwRet = _pMUdfTarget->OpenBlock(pMUTBlkInfo->dwMdfIdx, pMUTBlkInfo->dwUdfOff, false, _dwBlkId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: open block failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        goto err1;
    }

    if ( pMUTBlkInfo->dwUdfOff == INVALID_UDFOFF_VAL )
    {
        dwRet = _pMUdfTarget->CloseBlock(_dwBlkId);
        assert(dwRet == ERROR_SUCCESS);
        _dwBlkId = INVALID_MUT_BLK_ID;
    }
    else
    {
        assert(_nBlkIdx == INVALID_MDFIDX_VAL);
        _nBlkIdx = pMUTBlkInfo->dwMdfIdx;
    }

err1:
    return dwRet;
}


void MUdfDownloadAgent::OnDownloadFinished(DWORD dwErrCode)
{
    BufferItem  *pbi;

    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, dwErrCode));

    pbi = AllocEvent();
    if ( pbi != NULL )
    {
        EpBuildMsg(C_MUDF_TRANS_END, 0, dwErrCode, pbi);
        PutEvent(pbi);
    }
}

#ifdef CONFIG_MUT_BLK_COMPACT
DWORD MUdfDownloadAgent::InflateBlock(const PBYTE pData, UINT nDataBytes, bool bLast, bool &rbEndOfBlk)
{
    DWORD  dwRet;
    PBYTE  pSrc;
    UINT  nSrc, nBytes;
    bool  bAgain;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rbEndOfBlk = false;

    pSrc = pData;
    nSrc = nDataBytes;

    for ( bAgain = true; bAgain; )
    {
        dwRet = _DecompAlg.Process(pSrc, nSrc, &_DecompBuf, bAgain, bLast);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: Decompress(%d @ %p) failed: %d\n", FULL_FUNC_NAME, nSrc, pSrc, dwRet));
            assert(0);
            break;
        }

        nBytes = _DecompBuf.data_size();
        if ( nBytes > 0 )
        {
            dwRet = _pMUdfTarget->WriteBlock(_dwBlkId, _DecompBuf.data(), nBytes, rbEndOfBlk);
            if ( dwRet != ERROR_SUCCESS )
            {
                DBG(D_WARN, (L"%s: WriteBlock failed: %d\n", FULL_FUNC_NAME, dwRet));
                break;
            }

            _DecompBuf.Reset();
        }
    }

    return dwRet;
}
#endif
