/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "MUdfCommDef.h"
#include "MUdfDownloadHandler.h"


MUdfDownloadHandler::MUdfDownloadHandler(const char *szRootDir)
    : EpPDUHandler(MUDF_UPDOWN_CTRLQSIZE, MUDF_UPDOWN_CTRLBSIZE,
                   MUDF_UPDOWN_DATAQSIZE, MUDF_UPDOWN_DATABSIZE,
                   0, 0)
    , _strRootDir(szRootDir)
    , _pMUdfTarget(NULL)
    , _nBlkIdx(0)
    , _dwBlkId(INVALID_MUT_BLK_ID)
#ifdef CONFIG_MUT_BLK_COMPACT
    , _bEndOfBlk(false)
#endif
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MUdfDownloadHandler::~MUdfDownloadHandler()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_MUT_BLK_ID);
    assert(_pMUdfTarget == NULL);
}

DWORD MUdfDownloadHandler::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

#ifdef CONFIG_MUT_BLK_COMPACT
    dwRet = _CompAlg.Init(COMPRESS_FASTEST);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init compressor failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    DBG(D_INFO, (L"%s: alloc bytes[%d] for read buf\n", FULL_FUNC_NAME, COMPACT_CLUSTER_SIZE));
    _ReadBuf.Alloc(COMPACT_CLUSTER_SIZE);
    _CompBuf.Alloc(_CompAlg.GetBound(COMPACT_CLUSTER_SIZE));
#endif

    dwRet = EpPDUHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
#ifdef CONFIG_MUT_BLK_COMPACT
    _CompAlg.Fini();

err1:
#endif
    return dwRet;
}

void MUdfDownloadHandler::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::Unbind();

    if ( _pMUdfTarget != NULL )
    {
        if ( _dwBlkId != INVALID_MUT_BLK_ID )
        {
            _pMUdfTarget->CloseBlock(_dwBlkId);
            _dwBlkId = INVALID_MUT_BLK_ID;
        }

        _pMUdfTarget->Fini();

        delete _pMUdfTarget;
        _pMUdfTarget = NULL;
    }
    _nBlkIdx = 0;

#ifdef CONFIG_MUT_BLK_COMPACT
    _CompAlg.Fini();
#endif
}

void MUdfDownloadHandler::PrepareSend()
{
    DWORD  dwRet;
#ifndef CONFIG_MUT_BLK_COMPACT
    DWORD  dwSize;
#endif
    bool  bEndOfBlk;
    BufferItem  *pbi;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( _dwBlkId == INVALID_MUT_BLK_ID )
    {
        // at least one SendBuf
        dwRet = (_nBlkIdx < _pMUdfTarget->GetBlockCount()) ? DownloadBlockReq() : DownloadEndReq();
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: Download{Block/End}Req failed: %d\n", FULL_FUNC_NAME, dwRet));
            return;
        }

        if ( _dwBlkId == INVALID_MUT_BLK_ID )
        {
            return;
        }
    }

    while ( true )
    {
        pbi = GetSendBuf();
        if ( pbi == NULL )
        {
            break;
        }

#ifdef CONFIG_MUT_BLK_COMPACT
        dwRet = DeflateBlock(pbi, bEndOfBlk);
#else
        dwSize = pbi->tail_size();
        dwRet = _pMUdfTarget->ReadBlock(_dwBlkId, pbi->tail(), dwSize, bEndOfBlk);
        if ( dwRet == ERROR_SUCCESS )
        {
            pbi->Put(dwSize);
        }
#endif

        BuildPDU((bEndOfBlk ? C_MUDFTARGET_BLOCK_END : C_MUDFTARGET_BLOCK_DATA), true, false, dwRet, pbi);
        PutSendData(pbi, (dwRet == ERROR_SUCCESS) ? bEndOfBlk : true);

        if ( (dwRet != ERROR_SUCCESS) || bEndOfBlk )
        {
            _pMUdfTarget->CloseBlock(_dwBlkId);
            _dwBlkId = INVALID_MUT_BLK_ID;

            if ( dwRet == ERROR_SUCCESS )
            {
                ++_nBlkIdx;
                // schedule PrepareSend() to send block
                ReqSendProc(true);
            }

            break;
        }
    }
}

void MUdfDownloadHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    if ( dwParam != ERROR_SUCCESS )
    {
        assert(!bReq);
        DBG(D_WARN, (L"%s: error: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));
        ReqEndSession();
        return;
    }

    switch ( bCode )
    {
        case C_MUDFTARGET_DOWNLOAD:
            assert(bReq);
            assert(pbiBody->data_size() == sizeof(REQ_DOWNLOAD_MUDFTARGET));
            OnDownloadReq((PREQ_DOWNLOAD_MUDFTARGET)pbiBody->data());
            break;

        default:
            dwParam = ERROR_NOT_SUPPORTED;
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, bCode));
            assert(0);
            break;
    }

    if ( dwParam != ERROR_SUCCESS )
    {
        SendResponse(bCode, dwParam);
    }
}

void MUdfDownloadHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

void MUdfDownloadHandler::OnDownloadReq(const PREQ_DOWNLOAD_MUDFTARGET pReqOpenMUT)
{
    DWORD  dwRet;
    char  szMdfPath[MAX_PATH], szUdfPath[MAX_PATH];
    BufferItem  *pbi;
    PREP_DOWNLOAD_MUDFTARGET  pRepDownMUT;

    DBG(D_INFO, (L"%s: mdf(%hs), udf(%hs)\n", FULL_FUNC_NAME, pReqOpenMUT->szMdfPath, pReqOpenMUT->szUdfPath));

    pbi = GetSendBuf();
    assert(pbi != NULL);

    if ( _pMUdfTarget != NULL )
    {
        dwRet = ERROR_INVALID_STATE;
        DBG(D_WARN, (L"%s: previous mudftarget not fini\n", FULL_FUNC_NAME));
        goto err1;
    }

    _pMUdfTarget = new(std::nothrow) MUdfTarget();
    if ( _pMUdfTarget == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc mudftarget failed\n", FULL_FUNC_NAME));
        goto err1;
    }

    _snprintf(szMdfPath, _countof(szMdfPath), "%s%s", _strRootDir.c_str(), pReqOpenMUT->szMdfPath);
    _snprintf(szUdfPath, _countof(szUdfPath), "%s%s", _strRootDir.c_str(), pReqOpenMUT->szUdfPath);
    dwRet = _pMUdfTarget->Init(szMdfPath, szUdfPath, M_MUT_READONLY, 0);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: open mudftarget failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    pRepDownMUT = (PREP_DOWNLOAD_MUDFTARGET)pbi->Put(sizeof(REP_DOWNLOAD_MUDFTARGET));
    pRepDownMUT->dwMdfSize = _pMUdfTarget->GetMdfSize();
    DBG(D_INFO, (L"%s: mdf size: %d\n", FULL_FUNC_NAME, pRepDownMUT->dwMdfSize));

err2:
    if ( dwRet != ERROR_SUCCESS )
    {
        delete _pMUdfTarget;
        _pMUdfTarget = NULL;
    }

err1:
    BuildPDU(C_MUDFTARGET_DOWNLOAD, false, true, dwRet, pbi);
    PutSendData(pbi, true);

    if ( dwRet == ERROR_SUCCESS )
    {
        // schedule PrepareSend() to send block
        ReqSendProc(true);
    }
}

DWORD MUdfDownloadHandler::DownloadEndReq()
{
    BufferItem  *pbi;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pbi = GetSendBuf();
    assert(pbi != NULL);

    BuildPDU(C_MUDFTARGET_FILE_END, true, true, 0, pbi);
    PutSendData(pbi, true);

    return ERROR_SUCCESS;
}

DWORD MUdfDownloadHandler::DownloadBlockReq()
{
    DWORD  dwRet;
    DWORD  dwUdfOff;
    BufferItem  *pbi;
    PMUDFTARGET_BLOCK_INFO  pMUTBlkInfo;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_MUT_BLK_ID);
    assert(_pMUdfTarget != NULL);

    pbi = GetSendBuf();
    assert(pbi != NULL);

    dwRet = _pMUdfTarget->OpenBlock(_nBlkIdx, dwUdfOff, false, _dwBlkId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: OpenBlock(%d) failed: %d\n", FULL_FUNC_NAME, _nBlkIdx, dwRet));
        goto err;
    }

#ifdef CONFIG_MUT_BLK_COMPACT
   _bEndOfBlk = false;
   _CompBuf.Reset();
#endif

    pMUTBlkInfo = (PMUDFTARGET_BLOCK_INFO)pbi->Put(sizeof(MUDFTARGET_BLOCK_INFO));
    pMUTBlkInfo->dwMdfIdx = _nBlkIdx;
    pMUTBlkInfo->dwUdfOff = dwUdfOff;

err:
    BuildPDU(C_MUDFTARGET_BLOCK_START, true, true, dwRet, pbi);
    PutSendData(pbi, true);

    if ( dwRet == ERROR_SUCCESS )
    {
        if ( dwUdfOff == INVALID_UDFOFF_VAL )
        {
            _pMUdfTarget->CloseBlock(_dwBlkId);
            _dwBlkId = INVALID_MUT_BLK_ID;

            ++_nBlkIdx;
        }

        // schedule PrepareSend() to
        // send block or start next block
        ReqSendProc(true);
    }

    return dwRet;
}

void MUdfDownloadHandler::SendResponse(BYTE bCode, DWORD dwParam)
{
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    pbi = GetSendBuf();
    if ( pbi != NULL )
    {
        BuildPDU(bCode, false, true, dwParam, pbi);
        PutSendData(pbi, true);
    }
}

#ifdef CONFIG_MUT_BLK_COMPACT
DWORD MUdfDownloadHandler::DeflateBlock(BufferItem *pbi, bool &rbLast)
{
    DWORD  dwRet = ERROR_SUCCESS;
    UINT  nBytes;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rbLast = false;

    while ( pbi->tail_size() )
    {
        dwRet = DeflateCluster();
        if ( dwRet != ERROR_SUCCESS )
        {
            break;
        }

        nBytes = min(pbi->tail_size(), _CompBuf.data_size());
        memcpy(pbi->Put(nBytes), _CompBuf.data(), nBytes);
        _CompBuf.Pull(nBytes);

        if ( _bEndOfBlk && (!_CompBuf.data_size()) )
        {
            rbLast = true;
            break;
        }
    }

    return dwRet;
}

DWORD MUdfDownloadHandler::DeflateCluster()
{
    DWORD  dwRet;
    DWORD  dwRead;
    PBYTE  pSrc;
    UINT  nSrc, nDst;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _CompBuf.data_size() )
    {
        return ERROR_SUCCESS;
    }

    _CompBuf.Reset();

    dwRead = _ReadBuf.tail_size();
    dwRet = _pMUdfTarget->ReadBlock(_dwBlkId, _ReadBuf.tail(), dwRead, _bEndOfBlk);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ReadBlock failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    pSrc = _ReadBuf.tail();
    nSrc = dwRead;
    nDst = _CompBuf.tail_size();
    dwRet = _CompAlg.ProcessBlock(pSrc, nSrc, _CompBuf.tail(), nDst, false);//_bEndOfBlk);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ProcessStream failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    _CompBuf.Put(nDst);

    return ERROR_SUCCESS;
}
#endif

