/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "MUdfCommDef.h"
#include "MUdfUploadHandler.h"
#include "MUdfDownloadHandler.h"
#include "MUdfServerEp.h"


MUdfServerEp *MUdfServerEp::_instance = NULL;

MUdfServerEp *MUdfServerEp::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new MUdfServerEp();
    }
    return _instance;
}

void MUdfServerEp::Destroy()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

MUdfServerEp::MUdfServerEp()
    : Endpoint()
    , _hPsUploadListener(INVALID_LISTENER_ID)
    , _hOsUploadListener(INVALID_LISTENER_ID)
    , _hPsDownloadListener(INVALID_LISTENER_ID)
    , _hOsDownloadListener(INVALID_LISTENER_ID)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MUdfServerEp::~MUdfServerEp()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

DWORD MUdfServerEp::Load(UINT nMaxConnNum, const char *szPsUpRootDir, const char *szPsDownRootDir, const char *szOsRootDir)
{
    SOCK_ADDR  addr;

    DBG(D_LOG, (L"%s: MaxConnNum=%d, PsUpRoot=%hs, PsDownRoot=%hs, OsRoot=%hs\n", FULL_FUNC_NAME, nMaxConnNum, szPsUpRootDir, szPsDownRootDir, szOsRootDir));

    Endpoint::Load(nMaxConnNum, 3, 0);

    _strPsUpRootDir = szPsUpRootDir;
    _strPsDownRootDir = szPsDownRootDir;
    _strOsRootDir = szOsRootDir;

    addr.bType = ADDR_TYPE_IPV4;
    addr.IPv4Addr.bStream = 1;
    addr.IPv4Addr.dwIP = 0;

    addr.IPv4Addr.wPort = htons(MUDFPS_UPLOAD_PORT);
    AddListener(&addr, false, _hPsUploadListener);

    addr.IPv4Addr.wPort = htons(MUDFOS_UPLOAD_PORT);
    AddListener(&addr, false, _hOsUploadListener);

    addr.IPv4Addr.wPort = htons(MUDFPS_DOWNLOAD_PORT);
    AddListener(&addr, false, _hPsDownloadListener);

    addr.IPv4Addr.wPort = htons(MUDFOS_DOWNLOAD_PORT);
    AddListener(&addr, false, _hOsDownloadListener);

    return ERROR_SUCCESS;
}

void MUdfServerEp::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    if ( _hPsUploadListener != INVALID_LISTENER_ID )
    {
        RemoveListener(_hPsUploadListener);
        _hPsUploadListener = INVALID_LISTENER_ID;
    }

    if ( _hOsUploadListener != INVALID_LISTENER_ID )
    {
        RemoveListener(_hOsUploadListener);
        _hOsUploadListener = INVALID_LISTENER_ID;
    }

    if ( _hPsDownloadListener != INVALID_LISTENER_ID )
    {
        RemoveListener(_hPsDownloadListener);
        _hPsDownloadListener = INVALID_LISTENER_ID;
    }

    if ( _hOsDownloadListener != INVALID_LISTENER_ID )
    {
        RemoveListener(_hOsDownloadListener);
        _hOsDownloadListener = INVALID_LISTENER_ID;
    }

    Endpoint::Unload();
}

void MUdfServerEp::OnListenerEnded(DWORD hListener)
{
    DBG(D_LOG, (L"%s: id(%x)\n", FULL_FUNC_NAME, hListener));
}

EpProtocolHandler *MUdfServerEp::OnNewHandler(DWORD hListener, Connection *pConnection, DWORD dwConnId)
{
    EpPDUHandler  *pHandler;

    DBG(D_LOG, (L"%s: listener(%x), connid(%x)\n", FULL_FUNC_NAME, hListener, dwConnId));

    if ( hListener == _hPsUploadListener )
    {
        DBG(D_INFO, (L"%s: MUdfUploadHandler(Ps)\n", FULL_FUNC_NAME));
        pHandler = new(std::nothrow) MUdfUploadHandler(_strPsUpRootDir.c_str());
    }
    else if ( hListener == _hOsUploadListener )
    {
        DBG(D_INFO, (L"%s: MUdfUploadHandler(Os)\n", FULL_FUNC_NAME));
        pHandler = new(std::nothrow) MUdfUploadHandler(_strOsRootDir.c_str());
    }
    else if ( hListener == _hPsDownloadListener )
    {
        DBG(D_INFO, (L"%s: _hPsDownloadListener(Ps)\n", FULL_FUNC_NAME));
        pHandler = new(std::nothrow) MUdfDownloadHandler(_strPsDownRootDir.c_str());
    }
    else if ( hListener == _hOsDownloadListener )
    {
        DBG(D_INFO, (L"%s: _hOsDownloadListener(Os)\n", FULL_FUNC_NAME));
        pHandler = new(std::nothrow) MUdfDownloadHandler(_strOsRootDir.c_str());
    }

    else
    {
        DBG(D_ERROR, (L"%s: not match\n", FULL_FUNC_NAME));
        assert(0);
        pHandler = NULL;
    }

    return pHandler;
}

void MUdfServerEp::OnAsyncConnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_LOG, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));

    assert(0);
}

void MUdfServerEp::OnDisconnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_INFO, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));

    delete pHandler;
}

void MUdfServerEp::OnConnEvent(DWORD dwConnId, EpProtocolHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody)
{
    DBG(D_INFO, (L"%s: id(%#.8x), code(%d)\n", FULL_FUNC_NAME, dwConnId, pEvtHdr->nCode));

    assert(0);
}

void MUdfServerEp::ProcessEpWork(UINT nSigIdx)
{
    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, nSigIdx));

    assert(0);
}
