/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "MUdfCommDef.h"
#include "MUdfTarget.h"


MUdfTarget::MUdfTarget()
    : _OpenMode(M_MUT_READONLY)
    , _hMdFile(NULL)
    , _hUdFile(NULL)
    , _dwMdfSize(0)
    , _nMdfItemCnt(0)
    , _nUdfOffNext(0)
    , _pMdfItems(NULL)
    , _nDoneBlkCnt(0)
    , _qwReadBytes(0)
    , _qwWriteBytes(0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MUdfTarget::~MUdfTarget()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

DWORD MUdfTarget::Init(const char *szMdfPath, const char *szUdfPath, MUT_OPEN_MODE OpenMode, DWORD dwMdfSize)
{
    DWORD  dwRet;
    bool  bReadOnly, bReInit;
    PBYTE  pMapping;

    DBG(D_LOG, (L"%s: mdf(%hs/%d), udf(%hs), mode(%d)\n", FULL_FUNC_NAME, szMdfPath, dwMdfSize, szUdfPath, OpenMode));

    _OpenMode = OpenMode;

    bReadOnly = (OpenMode == M_MUT_READONLY);

    dwRet = OpenMdFile(szMdfPath, dwMdfSize, OpenMode, bReInit);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: OpenMdFile(%hs) failed: %d\n", FULL_FUNC_NAME, szMdfPath, dwRet));
        goto err1;
    }

    dwRet = OpenUdFile(szUdfPath, OpenMode, bReInit);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: OpenUdFile(%hs) failed: %d\n", FULL_FUNC_NAME, szUdfPath, dwRet));
        goto err2;
    }

    dwRet = _MdfMapping.Init(_hMdFile, _dwMdfSize, bReadOnly);
    if ( dwRet != ERROR_SUCCESS )
    {
        assert(0);
        DBG(D_WARN, (L"%s: _MdfMapping init(%d) failed: %d\n", FULL_FUNC_NAME, _dwMdfSize, dwRet));
        goto err3;
    }

    dwRet = _MdfMapping.Map(0, _dwMdfSize, bReadOnly, !bReadOnly, pMapping);
    if ( dwRet != ERROR_SUCCESS )
    {
        assert(0);
        DBG(D_WARN, (L"%s: _MdfMapping map(%d) failed: %d\n", FULL_FUNC_NAME, _dwMdfSize, dwRet));
        goto err4;
    }
    _pMdfItems = (DWORD*)pMapping;

    if ( bReInit )
    {
        for ( UINT i = 0; i < _nMdfItemCnt; ++i )
        {
            _pMdfItems[i] = INVALID_UDFOFF_VAL;
        }
    }

    _BlkRWRec.nBlkIdx = INVALID_MDFIDX_VAL;

    return ERROR_SUCCESS;

err4:
    _MdfMapping.Fini();

err3:
    fclose(_hUdFile);
    _hUdFile = NULL;

err2:
    fclose(_hMdFile);
    _hMdFile = NULL;

err1:
    return dwRet;
}

void MUdfTarget::Fini()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( _pMdfItems != NULL )
    {
        _MdfMapping.Unmap((PBYTE)_pMdfItems, _dwMdfSize);
        _pMdfItems = NULL;
    }

    _MdfMapping.Fini();
    
    if ( _hUdFile != NULL )
    {
        fclose(_hUdFile);
        _hUdFile = NULL;
    }

    if ( _hMdFile != NULL )
    {
        fclose(_hMdFile);
        _hMdFile = NULL;
    }

    _nDoneBlkCnt = 0;
    _qwReadBytes = 0;
    _qwWriteBytes = 0;
}

DWORD MUdfTarget::GetMdfSize() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwMdfSize != 0);

    return _dwMdfSize;
}

UINT MUdfTarget::GetBlockCount() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_nMdfItemCnt != 0);

    return _nMdfItemCnt;
}

DWORD MUdfTarget::GetBlockSize() const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    return MUDF_BLOCK_SIZE;
}

void MUdfTarget::GetStatInfo(UINT &rnDoneBlkCnt, QWORD &rqwReadBytes, QWORD &rqwWriteBytes) const
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rnDoneBlkCnt = _nDoneBlkCnt;
    rqwReadBytes = _qwReadBytes;
    rqwWriteBytes = _qwWriteBytes;
}

DWORD MUdfTarget::OpenBlock(UINT nBlkIdx, DWORD &rdwUdfOff, bool bAllocIf, DWORD &rdwBlkId)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: idx=%d, off=%d, alloc=%d\n", FULL_FUNC_NAME, nBlkIdx, rdwUdfOff, bAllocIf));

    if ( nBlkIdx >= _nMdfItemCnt )
    {
        dwRet = ERROR_INVALID_PARAMETER;
        DBG(D_WARN, (L"%s: idx=%d too big(total=%d)\n", FULL_FUNC_NAME, nBlkIdx, _nMdfItemCnt));
        assert(0);
        goto err;
    }

    if ( _BlkRWRec.nBlkIdx != INVALID_MDFIDX_VAL )
    {
        dwRet = ERROR_NOT_SUPPORTED;
        DBG(D_WARN, (L"%s: not support multi open\n", FULL_FUNC_NAME));
        assert(0);
        goto err;
    }

    if ( _OpenMode == M_MUT_READONLY )
    {
        rdwUdfOff = _pMdfItems[nBlkIdx];
    }
    else
    {
        if ( _pMdfItems[nBlkIdx] == INVALID_UDFOFF_VAL )
        {
            if ( bAllocIf )
            {
                rdwUdfOff = _nUdfOffNext++;
            }
            _pMdfItems[nBlkIdx] = rdwUdfOff;
        }
        else
        {
            if ( !bAllocIf )
            {
                _pMdfItems[nBlkIdx] = rdwUdfOff;
            }
            rdwUdfOff = _pMdfItems[nBlkIdx];
        }
    }

    if ( _pMdfItems[nBlkIdx] != INVALID_UDFOFF_VAL )
    {
        _BlkRWRec.qwBlkOffset = (QWORD)MUDF_BLOCK_SIZE * (QWORD)_pMdfItems[nBlkIdx];

        if ( _BlkRWRec.qwCurOffset != _BlkRWRec.qwBlkOffset )
        {
            if ( fseeko(_hUdFile, _BlkRWRec.qwBlkOffset, SEEK_SET) )
            {
                dwRet = GetLastError();
                DBG(D_WARN, (L"%s: fseek(%#llx) failed: %d\n", FULL_FUNC_NAME, _BlkRWRec.qwBlkOffset, dwRet));
                goto err;
            }
            _BlkRWRec.qwCurOffset = _BlkRWRec.qwBlkOffset;
        }

        _BlkRWRec.dwLeftInBlk = MUDF_BLOCK_SIZE;
    }
    else
    {
        _BlkRWRec.qwBlkOffset = 0;
        _BlkRWRec.dwLeftInBlk = 0;
    }

    _BlkRWRec.nBlkIdx = nBlkIdx;
    _BlkRWRec.dwOffInBlk = 0;

    rdwBlkId = (DWORD)(QWORD)&_BlkRWRec;

    return ERROR_SUCCESS;

err:
    rdwBlkId = INVALID_MUT_BLK_ID;

    return dwRet;
}

DWORD MUdfTarget::ReadBlock(DWORD dwBlkId, PBYTE pBuf, DWORD &rdwSize, bool &rbEndOfBlk)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _BlkRWRec.nBlkIdx == INVALID_MDFIDX_VAL )
    {
        DBG(D_WARN, (L"%s: block not opened\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_INVALID_STATE;
    }

    assert((_BlkRWRec.dwOffInBlk + _BlkRWRec.dwLeftInBlk) <= MUDF_BLOCK_SIZE);

    if ( rdwSize > _BlkRWRec.dwLeftInBlk )
    {
        rdwSize = _BlkRWRec.dwLeftInBlk;
    }

    dwRet = ReadUdFile(_BlkRWRec.qwBlkOffset + (QWORD)_BlkRWRec.dwOffInBlk, pBuf, rdwSize);
    if ( dwRet == ERROR_SUCCESS )
    {
        _BlkRWRec.dwOffInBlk += rdwSize;
        _BlkRWRec.dwLeftInBlk -= rdwSize;

        _qwReadBytes += rdwSize;
    }

    rbEndOfBlk = (_BlkRWRec.dwLeftInBlk == 0);

    return dwRet;
}

DWORD MUdfTarget::WriteBlock(DWORD dwBlkId, PBYTE pBuf, DWORD dwSize, bool &rbEndOfBlk)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _BlkRWRec.nBlkIdx == INVALID_MDFIDX_VAL )
    {
        DBG(D_WARN, (L"%s: block not opened\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_INVALID_STATE;
    }

    if ( dwSize > _BlkRWRec.dwLeftInBlk )
    {
        DBG(D_WARN, (L"%s: size(%d) > need(%d)\n", FULL_FUNC_NAME, dwSize, _BlkRWRec.dwLeftInBlk));
        assert(0);
        return ERROR_INVALID_DATA;
    }

    assert((_BlkRWRec.dwOffInBlk + _BlkRWRec.dwLeftInBlk) <= MUDF_BLOCK_SIZE);

    dwRet = WriteUdFile(_BlkRWRec.qwBlkOffset + (QWORD)_BlkRWRec.dwOffInBlk, pBuf, dwSize);
    if ( dwRet == ERROR_SUCCESS )
    {
        _BlkRWRec.dwOffInBlk += dwSize;
        _BlkRWRec.dwLeftInBlk -= dwSize;

        _qwWriteBytes += dwSize;
    }

    rbEndOfBlk = (_BlkRWRec.dwLeftInBlk == 0);

    return dwRet;
}

DWORD MUdfTarget::CloseBlock(DWORD dwBlkId)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _BlkRWRec.nBlkIdx == INVALID_MDFIDX_VAL )
    {
        DBG(D_WARN, (L"%s: block not opened\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_INVALID_STATE;
    }

    if ( _BlkRWRec.dwLeftInBlk > 0 )
    {
        DBG(D_WARN, (L"%s: blk[%d] not finished\n", FULL_FUNC_NAME, _BlkRWRec.nBlkIdx));
        dwRet = ERROR_CANCELLED;
        goto out;
    }

    if ( _OpenMode != M_MUT_READONLY )
    {
        fflush(_hUdFile);
    }

    ++_nDoneBlkCnt;

    dwRet = ERROR_SUCCESS;

out:
    _BlkRWRec.nBlkIdx = INVALID_MDFIDX_VAL;

    return dwRet;
}

DWORD MUdfTarget::OpenMdFile(const char *szMdfPath, DWORD dwMdfSize, MUT_OPEN_MODE OpenMode, bool &rbReInit)
{
    DWORD  dwRet;
    const char  *szMode;
    QWORD  qwFileSize;

    DBG(D_INFO, (L"%s: %hs(size=%d, mode=%d)\n", FULL_FUNC_NAME, szMdfPath, dwMdfSize, OpenMode));

    if ( OpenMode == M_MUT_READONLY )
    {
        szMode = "rb";
    }
    else if ( OpenMode == M_MUT_OVERWRITE)
    {
        // use "wb" file mapping will fail under windows
        szMode = "w+b";
    }
    else
    {
        szMode = "r+b";
    }

    _hMdFile = fopen(szMdfPath, szMode);
    if ( (_hMdFile == NULL) && (OpenMode == M_MUT_UPDATE) )
    {
        // use "wb" file mapping will fail under windows
        szMode = "w+b";
        _hMdFile = fopen(szMdfPath, szMode);
    }
    if ( _hMdFile == NULL )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: open mdf(%hs) with mode(%hs) failed: %d\n", FULL_FUNC_NAME, szMdfPath, szMode, dwRet));
        goto err1;
    }

    dwRet = GetFileLength(_hMdFile, qwFileSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: get mdf(%hs) length failed: %d\n", FULL_FUNC_NAME, szMdfPath, dwRet));
        goto err2;
    }

    if ( (OpenMode != M_MUT_READONLY) && ((DWORD)qwFileSize != dwMdfSize) )
    {
        qwFileSize = dwMdfSize;
        dwRet = SetFileLength(_hMdFile, qwFileSize);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: set mdf(%hs) length to %d failed: %d\n", FULL_FUNC_NAME, szMdfPath, dwMdfSize, dwRet));
            goto err2;
        }

        rbReInit = true;
    }
    else
    {
        rbReInit = false;
    }

    if ( qwFileSize == 0 )
    {
        dwRet = ERROR_INVALID_DATA;
        DBG(D_WARN, (L"%s: mdf(%hs) length is zero\n", FULL_FUNC_NAME, szMdfPath));
        goto err2;
    }

    _dwMdfSize = (DWORD)qwFileSize;
    _nMdfItemCnt = _dwMdfSize / sizeof(DWORD);

    return ERROR_SUCCESS;

err2:
    fclose(_hMdFile);
    _hMdFile = NULL;

err1:
    return dwRet;
}

DWORD MUdfTarget::OpenUdFile(const char *szUdfPath, MUT_OPEN_MODE OpenMode, bool bReInit)
{
    DWORD  dwRet;
    const char  *szMode;
    QWORD  qwFileSize;
    
    DBG(D_INFO, (L"%s: %hs(mode=%d, reinit=%d)\n", FULL_FUNC_NAME, szUdfPath, OpenMode, bReInit));

    if ( OpenMode == M_MUT_READONLY )
    {
        szMode = "rb";
    }
    else
    {
        szMode = bReInit ? "wb" : "r+b";
    }
    _hUdFile = fopen(szUdfPath, szMode);
    if ( _hUdFile == NULL )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: open udf(%hs) with mode(%hs) failed: %d\n", FULL_FUNC_NAME, szUdfPath, szMode, dwRet));
        goto err1;
    }
    _BlkRWRec.qwCurOffset = 0;

    dwRet = GetFileLength(_hUdFile, qwFileSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: get udf(%hs) length failed: %d\n", FULL_FUNC_NAME, szUdfPath, dwRet));
        goto err2;
    }

    _nUdfOffNext = (UINT)((qwFileSize + MUDF_BLOCK_SIZE - 1) / MUDF_BLOCK_SIZE);

    return ERROR_SUCCESS;

err2:
    fclose(_hUdFile);
    _hUdFile = NULL;

err1:
    return dwRet;
}

DWORD MUdfTarget::ReadUdFile(QWORD qwOffset, PBYTE pBuf, DWORD dwSize)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %d @ %#llx\n", FULL_FUNC_NAME, dwSize, qwOffset));

    assert(qwOffset == _BlkRWRec.qwCurOffset);

    if ( fread(pBuf, 1, dwSize, _hUdFile) != dwSize )
    {
        // maybe eof
        dwRet = ERROR_INVALID_DATA;//GetLastError();
        DBG(D_WARN, (L"%s: fread(%d @ %#llx) failed: %d\n", FULL_FUNC_NAME, dwSize, qwOffset, dwRet));
        return dwRet;
    }

    _BlkRWRec.qwCurOffset += dwSize;

    return ERROR_SUCCESS;
}

DWORD MUdfTarget::WriteUdFile(QWORD qwOffset, PBYTE pBuf, DWORD dwSize)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %d @ %#llx\n", FULL_FUNC_NAME, dwSize, qwOffset));

    assert(qwOffset == _BlkRWRec.qwCurOffset);

    if ( fwrite(pBuf, 1, dwSize, _hUdFile) != dwSize )
    {
        // see misc_fio.txt
        // use ERROR_DISK_FULL to notify critical error
        dwRet = ERROR_DISK_FULL;//GetLastError();
        DBG(D_WARN, (L"%s: fwrite(%d @ %#llx) failed: %d\n", FULL_FUNC_NAME, dwSize, qwOffset, dwRet));
        return dwRet;
    }

    _BlkRWRec.qwCurOffset += dwSize;

    return ERROR_SUCCESS;
}

