/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "MUdfCommDef.h"
#include "MUdfUploadAgent.h"


MUdfUploadAgent::MUdfUploadAgent(MUdfTarget *pMUdfTarget, const char *szRemoteMdf, const char *szRemoteUdf, bool bUpdateMode)
    : EpPDUHandler(MUDF_UPDOWN_CTRLQSIZE, MUDF_UPDOWN_CTRLBSIZE,
                   MUDF_UPDOWN_DATAQSIZE, MUDF_UPDOWN_DATABSIZE,
                   MUDF_UPDOWN_EVTQSIZE, MUDF_UPDOWN_EVTBSIZE)
    , _pMUdfTarget(pMUdfTarget)
    , _strRemoteMdf(szRemoteMdf)
    , _strRemoteUdf(szRemoteUdf)
    , _bUpdateMode(bUpdateMode)
    , _nBlkIdx(0)
    , _dwBlkId(INVALID_MUT_BLK_ID)
#ifdef CONFIG_MUT_BLK_COMPACT
    , _bEndOfBlk(false)
#endif
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MUdfUploadAgent::~MUdfUploadAgent()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_MUT_BLK_ID);
}

DWORD MUdfUploadAgent::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

#ifdef CONFIG_MUT_BLK_COMPACT
    dwRet = _CompAlg.Init(COMPRESS_FASTEST);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init compressor failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    DBG(D_INFO, (L"%s: alloc bytes[%d] for read buf\n", FULL_FUNC_NAME, COMPACT_CLUSTER_SIZE));
    _ReadBuf.Alloc(COMPACT_CLUSTER_SIZE);
    _CompBuf.Alloc(_CompAlg.GetBound(COMPACT_CLUSTER_SIZE));
#endif

    dwRet = EpPDUHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
#ifdef CONFIG_MUT_BLK_COMPACT
    _CompAlg.Fini();

err1:
#endif
    return dwRet;
}

void MUdfUploadAgent::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::Unbind();

    if ( _dwBlkId != INVALID_MUT_BLK_ID )
    {
        _pMUdfTarget->CloseBlock(_dwBlkId);
        _dwBlkId = INVALID_MUT_BLK_ID;
    }
    _nBlkIdx = 0;

#ifdef CONFIG_MUT_BLK_COMPACT
    _CompAlg.Fini();
#endif
}

void MUdfUploadAgent::OnAsyncConnected(Connection *pConnection)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::OnAsyncConnected(pConnection);

    dwRet = UploadStartReq();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: UploadStartReq failed: %d\n", FULL_FUNC_NAME, dwRet));
        OnUploadFinished(dwRet);
    }
}

void MUdfUploadAgent::PrepareSend()
{
    DWORD  dwRet = ERROR_SUCCESS;
#ifndef CONFIG_MUT_BLK_COMPACT
    DWORD  dwSize;
#endif
    bool  bEndOfBlk;
    BufferItem  *pbi;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( _nBlkIdx == INVALID_MDFIDX_VAL )
    {
        DBG(D_INFO, (L"%s: error state\n", FULL_FUNC_NAME));
        return;
    }

    if ( _dwBlkId == INVALID_MUT_BLK_ID )
    {
        // at least one SendBuf
        dwRet = (_nBlkIdx < _pMUdfTarget->GetBlockCount()) ? UploadBlockReq() : UploadEndReq();
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: Upload{Block/End}Req failed: %d\n", FULL_FUNC_NAME, dwRet));
            goto out;
        }

        if ( _dwBlkId == INVALID_MUT_BLK_ID )
        {
            return;
        }
    }

    while ( true )
    {
        pbi = GetSendBuf();
        if ( pbi == NULL )
        {
            break;
        }

#ifdef CONFIG_MUT_BLK_COMPACT
        dwRet = DeflateBlock(pbi, bEndOfBlk);
#else
        dwSize = pbi->tail_size();
        dwRet = _pMUdfTarget->ReadBlock(_dwBlkId, pbi->tail(), dwSize, bEndOfBlk);
        if ( dwRet == ERROR_SUCCESS )
        {
            pbi->Put(dwSize);
        }
#endif

        BuildPDU((bEndOfBlk ? C_MUDFTARGET_BLOCK_END : C_MUDFTARGET_BLOCK_DATA), true, false, dwRet, pbi);
        PutSendData(pbi, bEndOfBlk);

        if ( (dwRet != ERROR_SUCCESS) || bEndOfBlk )
        {
            _pMUdfTarget->CloseBlock(_dwBlkId);
            _dwBlkId = INVALID_MUT_BLK_ID;

            if ( dwRet == ERROR_SUCCESS )
            {
                ++_nBlkIdx;
                // schedule PrepareSend() to send block
                ReqSendProc(true);
            }

            break;
        }
    }

out:
    if ( dwRet != ERROR_SUCCESS )
    {
        OnUploadFinished(dwRet);
    }
}

void MUdfUploadAgent::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    if ( dwParam != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: error: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));
        goto out;
    }

    switch ( bCode )
    {
        case C_MUDFTARGET_UPLOAD_UPDATE:
        case C_MUDFTARGET_UPLOAD_OVERWRITE:
            // schedule PrepareSend() to send block
            ReqSendProc(true);
            break;

        case C_MUDFTARGET_FILE_END:
            DBG(D_INFO, (L"%s: finished\n", FULL_FUNC_NAME));
            break;

        default:
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, bCode));
            assert(0);
            break;
    }

out:
    if ( (bCode == C_MUDFTARGET_FILE_END) || (dwParam != ERROR_SUCCESS) )
    {
        OnUploadFinished(dwParam);
    }
}

void MUdfUploadAgent::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

DWORD MUdfUploadAgent::UploadStartReq()
{
    BufferItem  *pbi;
    PREQ_UPLOAD_MUDFTARGET  pUploadReq;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pbi = GetSendBuf();
    assert(pbi != NULL);

    pUploadReq = (PREQ_UPLOAD_MUDFTARGET)pbi->Put(sizeof(REQ_UPLOAD_MUDFTARGET));
    pUploadReq->bBlkEndAck = 0;
    pUploadReq->dwMdfSize = _pMUdfTarget->GetMdfSize();
    strncpy(pUploadReq->szMdfPath, _strRemoteMdf.c_str(), MAX_MUDF_PATH_LEN);
    strncpy(pUploadReq->szUdfPath, _strRemoteUdf.c_str(), MAX_MUDF_PATH_LEN);

    BuildPDU(_bUpdateMode ? C_MUDFTARGET_UPLOAD_UPDATE : C_MUDFTARGET_UPLOAD_OVERWRITE, true, true, 0, pbi);
    PutSendData(pbi, true);

    return ERROR_SUCCESS;
}

DWORD MUdfUploadAgent::UploadEndReq()
{
    BufferItem  *pbi;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pbi = GetSendBuf();
    assert(pbi != NULL);

    BuildPDU(C_MUDFTARGET_FILE_END, true, true, 0, pbi);
    PutSendData(pbi, true);

    return ERROR_SUCCESS;
}

DWORD MUdfUploadAgent::UploadBlockReq()
{
    DWORD  dwRet;
    DWORD  dwUdfOff;
    BufferItem  *pbi;
    PMUDFTARGET_BLOCK_INFO  pMUTBlkInfo;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_MUT_BLK_ID);
    assert(_pMUdfTarget != NULL);

    dwRet = _pMUdfTarget->OpenBlock(_nBlkIdx, dwUdfOff, false, _dwBlkId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: OpenBlock(%d) failed: %d\n", FULL_FUNC_NAME, _nBlkIdx, dwRet));
        assert(0);
        return dwRet;
    }

    if ( (!_bUpdateMode) || (dwUdfOff != INVALID_UDFOFF_VAL) )
    {
#ifdef CONFIG_MUT_BLK_COMPACT
       _bEndOfBlk = false;
       _CompBuf.Reset();
#endif

        pbi = GetSendBuf();
        assert(pbi != NULL);

        pMUTBlkInfo = (PMUDFTARGET_BLOCK_INFO)pbi->Put(sizeof(MUDFTARGET_BLOCK_INFO));
        pMUTBlkInfo->dwMdfIdx = _nBlkIdx;
        pMUTBlkInfo->dwUdfOff = dwUdfOff;

        BuildPDU(C_MUDFTARGET_BLOCK_START, true, true, 0, pbi);
        PutSendData(pbi, true);
    }

    if ( dwUdfOff == INVALID_UDFOFF_VAL )
    {
        _pMUdfTarget->CloseBlock(_dwBlkId);
        _dwBlkId = INVALID_MUT_BLK_ID;

        ++_nBlkIdx;
    }

    // schedule PrepareSend() to
    // send block or start next block
    ReqSendProc(true);

    return ERROR_SUCCESS;
}

void MUdfUploadAgent::OnUploadFinished(DWORD dwErrCode)
{
    BufferItem  *pbi;

    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, dwErrCode));

    if ( _nBlkIdx != INVALID_MDFIDX_VAL )
    {
        // mark error state
        _nBlkIdx = INVALID_MDFIDX_VAL;

        // not like MUdfUploadAgentNt, no need lock here
        pbi = AllocEvent();
        assert(pbi != NULL);

        EpBuildMsg(C_MUDF_TRANS_END, 0, dwErrCode, pbi);
        PutEvent(pbi);
    }
}

#ifdef CONFIG_MUT_BLK_COMPACT
DWORD MUdfUploadAgent::DeflateBlock(BufferItem *pbi, bool &rbLast)
{
    DWORD  dwRet = ERROR_SUCCESS;
    UINT  nBytes;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rbLast = false;

    while ( pbi->tail_size() )
    {
        dwRet = DeflateCluster();
        if ( dwRet != ERROR_SUCCESS )
        {
            break;
        }

        nBytes = min(pbi->tail_size(), _CompBuf.data_size());
        memcpy(pbi->Put(nBytes), _CompBuf.data(), nBytes);
        _CompBuf.Pull(nBytes);

        if ( _bEndOfBlk && (!_CompBuf.data_size()) )
        {
            rbLast = true;
            break;
        }
    }

    return dwRet;
}

DWORD MUdfUploadAgent::DeflateCluster()
{
    DWORD  dwRet;
    DWORD  dwRead;
    PBYTE  pSrc;
    UINT  nSrc, nDst;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _CompBuf.data_size() )
    {
        return ERROR_SUCCESS;
    }

    _CompBuf.Reset();

    dwRead = _ReadBuf.tail_size();
    dwRet = _pMUdfTarget->ReadBlock(_dwBlkId, _ReadBuf.tail(), dwRead, _bEndOfBlk);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ReadBlock failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    pSrc = _ReadBuf.tail();
    nSrc = dwRead;
    nDst = _CompBuf.tail_size();
    dwRet = _CompAlg.ProcessBlock(pSrc, nSrc, _CompBuf.tail(), nDst, false);//_bEndOfBlk);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ProcessStream failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    _CompBuf.Put(nDst);

    return ERROR_SUCCESS;
}
#endif
