/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "MUdfCommDef.h"
#include "MUdfUploadAgentNt.h"


#define IOCTL_TC_GET_NEXT_BLOCK      CTL_CODE(FILE_DEVICE_DISK, 0x803, METHOD_IN_DIRECT, FILE_READ_DATA)
#define IOCTL_TC_GET_MDF_SIZE        CTL_CODE(FILE_DEVICE_DISK, 0x805, METHOD_BUFFERED, FILE_READ_DATA)
#define IOCTL_TC_WAKE_UP_CTL_THREAD  CTL_CODE(FILE_DEVICE_DISK, 0x806, METHOD_BUFFERED, FILE_READ_DATA)

typedef struct _TC_IOCTRL_DATA_BLOCK{
	UINT32	blockIndex;
	UINT32	blockOffset;
	UINT32	bufLen;
}TC_IOCTRL_DATA_BLOCK, *PTC_IOCTRL_DATA_BLOCK;


MUdfUploadAgentNt::MUdfUploadAgentNt(HANDLE hPsMUdfCtrl, HANDLE hPsMUdfFile, const char *szRemoteMdf, const char *szRemoteUdf)
    : EpPDUHandler(MUDF_UPDOWN_CTRLQSIZE, MUDF_UPDOWN_CTRLBSIZE,
                   MUDF_UPDOWN_DATAQSIZE, MUDF_UPDOWN_DATABSIZE,
                   MUDF_UPDOWN_EVTQSIZE, MUDF_UPDOWN_EVTBSIZE, true)
    , _hPsMUdfCtrl(hPsMUdfCtrl)
    , _hPsMUdfFile(hPsMUdfFile)
    , _strRemoteMdf(szRemoteMdf)
    , _strRemoteUdf(szRemoteUdf)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    sem_init(&_hWorkEvent, 0, 0);
    SET_PTH_INVALID(_hWorkThread);
}

MUdfUploadAgentNt::~MUdfUploadAgentNt()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(!IS_PTH_VALID(_hWorkThread));
    sem_destroy(&_hWorkEvent);
}

DWORD MUdfUploadAgentNt::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    _biBlkBuf.Alloc(sizeof(TC_IOCTRL_DATA_BLOCK) + MUDF_BLOCK_SIZE);

#ifdef CONFIG_MUT_BLK_COMPACT
    dwRet = _CompAlg.Init(COMPRESS_FASTEST);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init compressor failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    DBG(D_INFO, (L"%s: compress cluster size: %d\n", FULL_FUNC_NAME, COMPACT_CLUSTER_SIZE));
    _CompBuf.Alloc(_CompAlg.GetBound(COMPACT_CLUSTER_SIZE));
#endif

    // for UploadStartReq @ OnAsyncConnected
    _CurState = S_MUDF_UPAGENT_WAITACK;
    _bQuit = false;
    if ( pthread_create(&_hWorkThread, NULL, WorkThreadProc, this) )
    {
        dwRet = GetLastError();
        DBG(D_ERROR, (L"%s: create work thread failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    dwRet = EpPDUHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err3;
    }

    return ERROR_SUCCESS;

err3:
    _bQuit = true;
    sem_post(&_hWorkEvent);
    pthread_join(_hWorkThread, NULL);
    SET_PTH_INVALID(_hWorkThread);

err2:
#ifdef CONFIG_MUT_BLK_COMPACT
    _CompAlg.Fini();

err1:
#endif
    return dwRet;
}

void MUdfUploadAgentNt::Unbind()
{
    DWORD  dwBytes;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( IS_PTH_VALID(_hWorkThread) )
    {
        _bQuit = true;
        sem_post(&_hWorkEvent);

        // if after ERROR_NO_MORE_FILES and before call IOCTL_TC_GET_NEXT_BLOCK again
        // IOCTL_TC_WAKE_UP_CTL_THREAD will no effect and IOCTL_TC_GET_NEXT_BLOCK will stuck
        DeviceIoControl(_hPsMUdfCtrl, IOCTL_TC_WAKE_UP_CTL_THREAD, NULL, 0, NULL, 0, &dwBytes, NULL);
        Sleep(500);
        // so sleep for a while and wake again
        DeviceIoControl(_hPsMUdfCtrl, IOCTL_TC_WAKE_UP_CTL_THREAD, NULL, 0, NULL, 0, &dwBytes, NULL);

        pthread_join(_hWorkThread, NULL);
        SET_PTH_INVALID(_hWorkThread);
    }


    EpPDUHandler::Unbind();

#ifdef CONFIG_MUT_BLK_COMPACT
    _CompAlg.Fini();
#endif
}

void MUdfUploadAgentNt::OnAsyncConnected(Connection *pConnection)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::OnAsyncConnected(pConnection);

    dwRet = UploadStartReq();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: UploadStartReq failed: %d\n", FULL_FUNC_NAME, dwRet));
        OnUploadFinished(dwRet);
    }
}

void MUdfUploadAgentNt::PrepareSend()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    sem_post(&_hWorkEvent);
}

void MUdfUploadAgentNt::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DWORD  dwExtra;

    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    if ( dwParam != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: error: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));
        goto out;
    }

    assert(pbiBody->data_size() == sizeof(DWORD));
    dwExtra = *(DWORD*)pbiBody->data();

    switch ( bCode )
    {
        case C_MUDFTARGET_BLOCK_END:
            assert(dwExtra == _dwBlkIdx);

        case C_MUDFTARGET_UPLOAD_UPDATE:
            DBG(D_VERB, (L"%s: {start update/block end}'s ack\n", FULL_FUNC_NAME));
            assert(_CurState == S_MUDF_UPAGENT_WAITACK);
            _CurState = S_MUDF_UPAGENT_IDLE;
            sem_post(&_hWorkEvent);            
            break;

        case C_MUDFTARGET_FILE_END:
            DBG(D_INFO, (L"%s: finished\n", FULL_FUNC_NAME));
            break;

        default:
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, bCode));
            assert(0);
            break;
    }

out:
    if ( (bCode == C_MUDFTARGET_FILE_END) || (dwParam != ERROR_SUCCESS) )
    {
        OnUploadFinished(dwParam);
    }
}

void MUdfUploadAgentNt::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

DWORD MUdfUploadAgentNt::UploadStartReq()
{
    DWORD  dwRet;
    DWORD  dwBytes, dwMdfSize;
    BufferItem  *pbi;
    PREQ_UPLOAD_MUDFTARGET  pUploadReq;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( !DeviceIoControl(_hPsMUdfFile,
                    IOCTL_TC_GET_MDF_SIZE,
                    NULL,
                    0,
                    &dwMdfSize,
                    sizeof(dwMdfSize),
                    &dwBytes,
                    NULL) )
    {
        dwRet = GetLastError();
        DBG(D_WARN, (L"%s: IOCTL_TC_GET_MDF_SIZE failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }
    DBG(D_INFO, (L"%s: mdf file size: %d\n", FULL_FUNC_NAME, dwMdfSize));

    pbi = GetSendBuf();
    assert(pbi != NULL);

    pUploadReq = (PREQ_UPLOAD_MUDFTARGET)pbi->Put(sizeof(REQ_UPLOAD_MUDFTARGET));
    pUploadReq->bBlkEndAck = 1;
    pUploadReq->dwMdfSize = dwMdfSize;
    strncpy(pUploadReq->szMdfPath, _strRemoteMdf.c_str(), MAX_MUDF_PATH_LEN);
    strncpy(pUploadReq->szUdfPath, _strRemoteUdf.c_str(), MAX_MUDF_PATH_LEN);

    BuildPDU(C_MUDFTARGET_UPLOAD_UPDATE, true, true, 0, pbi);
    PutSendData(pbi, true);

    return ERROR_SUCCESS;
}

DWORD MUdfUploadAgentNt::UploadEndReq()
{
    BufferItem  *pbi;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    pbi = GetSendBuf();
    assert(pbi != NULL);

    BuildPDU(C_MUDFTARGET_FILE_END, true, true, 0, pbi);
    PutSendData(pbi, true);

    return ERROR_SUCCESS;
}

DWORD MUdfUploadAgentNt::UploadBlockReq()
{
    DWORD  dwRet;
    BufferItem  *pbi;
    PMUDFTARGET_BLOCK_INFO  pMUTBlkInfo;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    dwRet = OpenAndReadBlock();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: OpenAndReadBlock failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    pbi = GetSendBuf();
    assert(pbi != NULL);

    pMUTBlkInfo = (PMUDFTARGET_BLOCK_INFO)pbi->Put(sizeof(MUDFTARGET_BLOCK_INFO));
    pMUTBlkInfo->dwMdfIdx = _dwBlkIdx;
    pMUTBlkInfo->dwUdfOff = _dwUdfOff;

    BuildPDU(C_MUDFTARGET_BLOCK_START, true, true, 0, pbi);
    PutSendData(pbi, true);

    return ERROR_SUCCESS;
}

DWORD MUdfUploadAgentNt::UploadBlockData(BufferItem *pbi)
{
    DWORD  dwRet;
    bool  bEndOfBlk;
#ifndef CONFIG_MUT_BLK_COMPACT
    UINT  nCopySize;
#endif
    BYTE  bCode;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

#ifdef CONFIG_MUT_BLK_COMPACT
    dwRet = DeflateBlock(pbi, bEndOfBlk);
#else
    nCopySize = min(pbi->tail_size(), _biBlkBuf.data_size());
    memcpy(pbi->Put(nCopySize), _biBlkBuf.data(), nCopySize);
    _biBlkBuf.Pull(nCopySize);
    bEndOfBlk = (_biBlkBuf.data_size() == 0);
    dwRet = ERROR_SUCCESS;
#endif

    if ( bEndOfBlk )
    {
        _CurState = S_MUDF_UPAGENT_WAITACK;
        bCode = C_MUDFTARGET_BLOCK_END;
    }
    else
    {
        bCode = C_MUDFTARGET_BLOCK_DATA;
    }
    BuildPDU(bCode, true, false, dwRet, pbi);
    PutSendData(pbi, bEndOfBlk);

    return dwRet;
}

void MUdfUploadAgentNt::OnUploadFinished(DWORD dwErrCode)
{
    BufferItem  *pbi;

    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, dwErrCode));

    // need lock, WorkThread  & ProcessRecvPDU may call it at the same time
    // UploadBlockData@WorkThread send block[N], server write error and ERROR return
    // UploadBlockData@WorkThread read block[N+1] error...
    pbi = AllocEvent(true);
    if ( pbi != NULL )
    {
        EpBuildMsg(C_MUDF_TRANS_END, 0, dwErrCode, pbi);
        PutEvent(pbi, true);
    }
}

DWORD MUdfUploadAgentNt::OpenAndReadBlock()
{
    DWORD  dwRet;
    DWORD  dwBytes;
    PTC_IOCTRL_DATA_BLOCK  pBlkInfo;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    _biBlkBuf.Reset();

    while ( true )
    {
        DBG(D_VERB, (L"%s: ack pre blk: %d\n", FULL_FUNC_NAME, _dwBlkIdx));
        if ( DeviceIoControl(_hPsMUdfFile,
                    IOCTL_TC_GET_NEXT_BLOCK, &_dwBlkIdx, sizeof(_dwBlkIdx),
                    _biBlkBuf.tail(), _biBlkBuf.tail_size(), &dwBytes, NULL) != 0 )
        {
            break;
        }

        dwRet = GetLastError();
        if ( dwRet == ERROR_NO_MORE_FILES )
        {
            // previous _dwBlkIdx is acked in DeviceIoControl call
            _dwBlkIdx = INVALID_MDFIDX_VAL;
            DBG(D_INFO, (L"%s: ERROR_NO_MORE_FILES, call again\n", FULL_FUNC_NAME));
            continue;
        }

        DBG(D_WARN, (L"%s: DeviceIoControl failed: %d\n", FULL_FUNC_NAME, dwRet));
        return dwRet;
    }

    if ( dwBytes != (sizeof(TC_IOCTRL_DATA_BLOCK) + MUDF_BLOCK_SIZE) )
    {
        DBG(D_WARN, (L"%s: invalid data size: %d\n", FULL_FUNC_NAME, dwBytes));
        assert(0);
        return ERROR_NOT_SUPPORTED;
    }

    _biBlkBuf.Put(dwBytes);

    pBlkInfo = (PTC_IOCTRL_DATA_BLOCK)_biBlkBuf.data();
    _biBlkBuf.Pull(sizeof(TC_IOCTRL_DATA_BLOCK));

    _dwBlkIdx = pBlkInfo->blockIndex;
    _dwUdfOff = pBlkInfo->blockOffset;
    assert(_dwBlkIdx != INVALID_MDFIDX_VAL);
    DBG(D_VERB, (L"%s: new blk idx=%d, off=%d\n", FULL_FUNC_NAME, _dwBlkIdx, _dwUdfOff));

#ifdef CONFIG_MUT_BLK_COMPACT
    _bEndOfBlk = false;
    _CompBuf.Reset();
#endif

    return ERROR_SUCCESS;
}

DWORD MUdfUploadAgentNt::WorkThread()
{
    DWORD  dwRet = ERROR_SUCCESS;
    BufferItem  *pbi;

    DBG(D_LOG, (L"%s: enter\n", FULL_FUNC_NAME));

    _dwBlkIdx = INVALID_MDFIDX_VAL;
    _dwUdfOff = INVALID_UDFOFF_VAL;

    // will upload blocks...
    while ( !_bQuit )
    {
        if ( _CurState == S_MUDF_UPAGENT_IDLE )
        {
            _CurState = S_MUDF_UPAGENT_START;

            DBG(D_VERB, (L"%s: will upload block...\n", FULL_FUNC_NAME));
            dwRet = UploadBlockReq();
            if ( dwRet != ERROR_SUCCESS )
            {
                DBG(D_WARN, (L"%s: UploadBlockReq failed: %d\n", FULL_FUNC_NAME, dwRet));
                break;
            }

            _CurState = S_MUDF_UPAGENT_TRANS;
        }

        if ( _CurState == S_MUDF_UPAGENT_TRANS )
        {
            pbi = GetSendBuf();
            if ( pbi == NULL )
            {
                // wait PrepareSend's signal
                sem_wait(&_hWorkEvent);
                continue;
            }

            dwRet = UploadBlockData(pbi);
            if ( dwRet != ERROR_SUCCESS )
            {
                DBG(D_WARN, (L"%s: UploadBlockData failed: %d\n", FULL_FUNC_NAME, dwRet));
                break;
            }
        }

        if ( _CurState == S_MUDF_UPAGENT_WAITACK )
        {
            // wait C_MUDFTARGET_UPLOAD_UPDATE
            // or C_MUDFTARGET_BLOCK_END's ack
            DBG(D_VERB, (L"%s: wait for ack...\n", FULL_FUNC_NAME));
            sem_wait(&_hWorkEvent);
            continue;
        }
    }

    if ( dwRet != ERROR_SUCCESS )
    {
        OnUploadFinished(dwRet);
    }

    DBG(D_LOG, (L"%s: leave\n", FULL_FUNC_NAME));

    return ERROR_SUCCESS;
}

void *MUdfUploadAgentNt::WorkThreadProc(void *arg)
{
    MUdfUploadAgentNt  *pThis = static_cast<MUdfUploadAgentNt*>(arg);

    pThis->WorkThread();

    return NULL;
}

#ifdef CONFIG_MUT_BLK_COMPACT
DWORD MUdfUploadAgentNt::DeflateBlock(BufferItem *pbi, bool &rbLast)
{
    DWORD  dwRet = ERROR_SUCCESS;
    UINT  nBytes;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rbLast = false;

    while ( pbi->tail_size() )
    {
        dwRet = DeflateCluster();
        if ( dwRet != ERROR_SUCCESS )
        {
            break;
        }

        nBytes = min(pbi->tail_size(), _CompBuf.data_size());
        memcpy(pbi->Put(nBytes), _CompBuf.data(), nBytes);
        _CompBuf.Pull(nBytes);

        if ( _bEndOfBlk && (!_CompBuf.data_size()) )
        {
            rbLast = true;
            break;
        }
    }

    return dwRet;
}

DWORD MUdfUploadAgentNt::DeflateCluster()
{
    DWORD  dwRet;
    PBYTE  pSrc;
    UINT  nSrc, nDst;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( _CompBuf.data_size() )
    {
        return ERROR_SUCCESS;
    }

    _CompBuf.Reset();

    pSrc = _biBlkBuf.data();
    nSrc = min(_biBlkBuf.data_size(), COMPACT_CLUSTER_SIZE);
    nDst = _CompBuf.tail_size();
    dwRet = _CompAlg.ProcessBlock(pSrc, nSrc, _CompBuf.tail(), nDst, false);//_bEndOfBlk);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ProcessStream failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        return dwRet;
    }

    _CompBuf.Put(nDst);

    _biBlkBuf.Pull(nSrc);
    _bEndOfBlk = (_biBlkBuf.data_size() == 0);

    return ERROR_SUCCESS;
}
#endif

