/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#include "MUdfCommDef.h"
#include "MUdfUploadHandler.h"


MUdfUploadHandler::MUdfUploadHandler(const char *szRootDir)
    : EpPDUHandler(MUDF_UPDOWN_DATAQSIZE, MUDF_UPDOWN_DATABSIZE,
                   MUDF_UPDOWN_CTRLQSIZE, MUDF_UPDOWN_CTRLBSIZE,
                   0, 0)
    , _strRootDir(szRootDir)
    , _pMUdfTarget(NULL)
    , _bBlkEndAck(false)
    , _nBlkIdx(INVALID_MDFIDX_VAL)
    , _dwBlkId(INVALID_MUT_BLK_ID)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MUdfUploadHandler::~MUdfUploadHandler()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(_dwBlkId == INVALID_MUT_BLK_ID);
    assert(_nBlkIdx == INVALID_MDFIDX_VAL);
    assert(_pMUdfTarget == NULL);
}

DWORD MUdfUploadHandler::Bind(Connection *pConnection, PCondEvent *pEvent, UINT nSigIdx)
{
    DWORD  dwRet;

    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

#ifdef CONFIG_MUT_BLK_COMPACT
    dwRet = _DecompAlg.Init();
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: init decompressor failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err1;
    }

    DBG(D_INFO, (L"%s: alloc bytes[%d] for write buf\n", FULL_FUNC_NAME, COMPACT_CLUSTER_SIZE));
    _DecompBuf.Alloc(COMPACT_CLUSTER_SIZE);
#endif

    dwRet = EpPDUHandler::Bind(pConnection, pEvent, nSigIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: bind protocol failed: %d\n", FULL_FUNC_NAME, dwRet));
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
#ifdef CONFIG_MUT_BLK_COMPACT
    _DecompAlg.Fini();

err1:
#endif
    return dwRet;
}

void MUdfUploadHandler::Unbind()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    EpPDUHandler::Unbind();

    OnFiniMUdfTarget();

#ifdef CONFIG_MUT_BLK_COMPACT
    _DecompAlg.Fini();
#endif
}

void MUdfUploadHandler::PrepareSend()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

void MUdfUploadHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DWORD  dwRet;
    bool  bSendRep = true;

    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    assert(bReq);

    if ( dwParam != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: error: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));
        assert(0);
        ReqEndSession();
        return;
    }

    switch ( bCode )
    {
        case C_MUDFTARGET_UPLOAD_UPDATE:
        case C_MUDFTARGET_UPLOAD_OVERWRITE:
            assert(pbiBody->data_size() == sizeof(REQ_UPLOAD_MUDFTARGET));
            dwRet = OnInitMUdfTarget((PREQ_UPLOAD_MUDFTARGET)pbiBody->data(), (bCode == C_MUDFTARGET_UPLOAD_OVERWRITE));
            break;

        case C_MUDFTARGET_BLOCK_START:
            bSendRep = false;
            assert(pbiBody->data_size() == sizeof(MUDFTARGET_BLOCK_INFO));
            dwRet = OnOpenMUdfBlock((PMUDFTARGET_BLOCK_INFO)pbiBody->data());
            break;

        case C_MUDFTARGET_FILE_END:
            dwRet = OnFiniMUdfTarget();
            break;

        default:
            dwRet = ERROR_NOT_SUPPORTED;
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, bCode));
            assert(0);
            break;
    }

    if ( bSendRep || (dwRet != ERROR_SUCCESS) )
    {
        SendResponse(bCode, dwRet, 0);
    }
}

void MUdfUploadHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    DWORD  dwRet;
    DWORD  dwExtra = 0;
    bool  bLastOne, bEndOfBlk;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert((bCode == C_MUDFTARGET_BLOCK_DATA) || (bCode == C_MUDFTARGET_BLOCK_END));
    assert(bReq);

    if ( dwParam != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: block data error\n", FULL_FUNC_NAME));
        ReqEndSession();
        return;
    }

    if ( _dwBlkId == INVALID_MUT_BLK_ID )
    {
        dwRet = ERROR_INVALID_STATE;
        DBG(D_WARN, (L"%s: pre open/write block failed?\n", FULL_FUNC_NAME));
        goto out1;
    }

    assert(_nBlkIdx != INVALID_MDFIDX_VAL);
    dwExtra = _nBlkIdx;

    // not use bEndOfBlk to judge whether CloseBlock & SendResponse
    bLastOne = (bCode == C_MUDFTARGET_BLOCK_END) && (nLeftBytes == 0);

#ifdef CONFIG_MUT_BLK_COMPACT
    dwRet = InflateBlock(pData, nDataBytes, bLastOne, bEndOfBlk);
#else
    dwRet = _pMUdfTarget->WriteBlock(_dwBlkId, pData, nDataBytes, bEndOfBlk);
#endif
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: inflate or write block[%d] failed: %d\n", FULL_FUNC_NAME, _nBlkIdx, dwRet));
        goto out2;
    }

    assert(bEndOfBlk == bLastOne);

out2:
    if ( (dwRet != ERROR_SUCCESS) || bLastOne )
    {
        // WriteBlock fail, CloseBlock will fail too
        // WriteBlock OK, CloseBlock may fail
        dwRet = _pMUdfTarget->CloseBlock(_dwBlkId);
        _dwBlkId = INVALID_MUT_BLK_ID;
        _nBlkIdx = INVALID_MDFIDX_VAL;
    }

out1:
    if ( (dwRet != ERROR_SUCCESS) || (bLastOne && _bBlkEndAck) )
    {
        SendResponse(bCode, dwRet, dwExtra);
    }
}

DWORD MUdfUploadHandler::OnInitMUdfTarget(const PREQ_UPLOAD_MUDFTARGET pReqOpenMUT, bool bOverwrite)
{
    DWORD  dwRet;
    char  szMdfPath[MAX_PATH], szUdfPath[MAX_PATH];

    DBG(D_INFO, (L"%s: mdf(%d,%hs), udf(%hs), ow=%d\n", FULL_FUNC_NAME, pReqOpenMUT->dwMdfSize, pReqOpenMUT->szMdfPath, pReqOpenMUT->szUdfPath, bOverwrite));

    _snprintf(szMdfPath, _countof(szMdfPath), "%s%s", _strRootDir.c_str(), pReqOpenMUT->szMdfPath);
    _snprintf(szUdfPath, _countof(szUdfPath), "%s%s", _strRootDir.c_str(), pReqOpenMUT->szUdfPath);

    if ( _pMUdfTarget != NULL )
    {
        dwRet = ERROR_INVALID_STATE;
        DBG(D_WARN, (L"%s: previous mudftarget not closed\n", FULL_FUNC_NAME));
        goto err1;
    }

    _pMUdfTarget = new(std::nothrow) MUdfTarget();
    if ( _pMUdfTarget == NULL )
    {
        dwRet = ERROR_OUTOFMEMORY;
        DBG(D_WARN, (L"%s: alloc mudftarget failed\n", FULL_FUNC_NAME));
        goto err1;
    }

    dwRet = _pMUdfTarget->Init(szMdfPath, szUdfPath, (bOverwrite ? M_MUT_OVERWRITE : M_MUT_UPDATE), pReqOpenMUT->dwMdfSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: open mudftarget failed\n", FULL_FUNC_NAME));
        goto err2;
    }

    _bBlkEndAck = (pReqOpenMUT->bBlkEndAck != 0);

    return ERROR_SUCCESS;

err2:
    delete _pMUdfTarget;
    _pMUdfTarget = NULL;

err1:
    return dwRet;
}

DWORD MUdfUploadHandler::OnFiniMUdfTarget()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    if ( _pMUdfTarget != NULL )
    {
        if ( _dwBlkId != INVALID_MUT_BLK_ID )
        {
            _pMUdfTarget->CloseBlock(_dwBlkId);
            _dwBlkId = INVALID_MUT_BLK_ID;
            _nBlkIdx = INVALID_MDFIDX_VAL;
        }

        _pMUdfTarget->Fini();

        delete _pMUdfTarget;
        _pMUdfTarget = NULL;
    }

    return ERROR_SUCCESS;
}

DWORD MUdfUploadHandler::OnOpenMUdfBlock(const PMUDFTARGET_BLOCK_INFO pMUTBlkInfo)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: mdfidx=%d, udfoff=%d\n", FULL_FUNC_NAME, pMUTBlkInfo->dwMdfIdx, pMUTBlkInfo->dwUdfOff));

    if ( _dwBlkId != INVALID_MUT_BLK_ID )
    {
        DBG(D_WARN, (L"%s: previous block not closed\n", FULL_FUNC_NAME));
        assert(0);
        return ERROR_INVALID_STATE;
    }

    dwRet = _pMUdfTarget->OpenBlock(pMUTBlkInfo->dwMdfIdx, pMUTBlkInfo->dwUdfOff, false, _dwBlkId);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: open block failed: %d\n", FULL_FUNC_NAME, dwRet));
        assert(0);
        goto err1;
    }

    if ( pMUTBlkInfo->dwUdfOff == INVALID_UDFOFF_VAL )
    {
        dwRet = _pMUdfTarget->CloseBlock(_dwBlkId);
        assert(dwRet == ERROR_SUCCESS);
        _dwBlkId = INVALID_MUT_BLK_ID;
    }
    else
    {
        assert(_nBlkIdx == INVALID_MDFIDX_VAL);
        _nBlkIdx = pMUTBlkInfo->dwMdfIdx;
    }

err1:
    return dwRet;
}

void MUdfUploadHandler::SendResponse(BYTE bCode, DWORD dwParam, DWORD dwExtra)
{
    BufferItem  *pbi;

    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d, dwExtra=%d\n", FULL_FUNC_NAME, bCode, dwParam, dwExtra));

    pbi = GetSendBuf();
    if ( pbi != NULL )
    {
        *(DWORD*)pbi->Put(sizeof(DWORD)) = dwExtra;
        BuildPDU(bCode, false, true, dwParam, pbi);
        PutSendData(pbi, true);
    }
}

#ifdef CONFIG_MUT_BLK_COMPACT
DWORD MUdfUploadHandler::InflateBlock(const PBYTE pData, UINT nDataBytes, bool bLast, bool &rbEndOfBlk)
{
    DWORD  dwRet;
    PBYTE  pSrc;
    UINT  nSrc, nBytes;
    bool  bAgain;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    rbEndOfBlk = false;

    pSrc = pData;
    nSrc = nDataBytes;

    for ( bAgain = true; bAgain; )
    {
        dwRet = _DecompAlg.Process(pSrc, nSrc, &_DecompBuf, bAgain, bLast);
        if ( dwRet != ERROR_SUCCESS )
        {
            DBG(D_WARN, (L"%s: Decompress(%d @ %p) failed: %d\n", FULL_FUNC_NAME, nSrc, pSrc, dwRet));
            assert(0);
            break;
        }

        nBytes = _DecompBuf.data_size();
        if ( nBytes > 0 )
        {
            dwRet = _pMUdfTarget->WriteBlock(_dwBlkId, _DecompBuf.data(), nBytes, rbEndOfBlk);
            if ( dwRet != ERROR_SUCCESS )
            {
                DBG(D_WARN, (L"%s: WriteBlock failed: %d\n", FULL_FUNC_NAME, dwRet));
                break;
            }

            _DecompBuf.Reset();
        }
    }

    return dwRet;
}
#endif
