/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#ifdef CONFIG_TLS_SUPPORT
#include "SecureCenter.h"
#endif
#include "SHA1Hash.h"
#include "ExecutorPool.h"
#include "TimerCenter.h"
#include "ReactorSelect.h"
#include "ReactorEPoll.h"
#include "ConnectCenter.h"
#include "SeedTarget.h"
#include "LP2PEndpoint.h"


typedef struct _ST_STAT_REC
{
    DWORD                       dwBlkMBSize;
    DWORD                       dwStartTicks;
    DWORD                       dwPreRdTicks;
    DWORD                       dwPreWrTicks;
    UINT                        nPreBlkRdCnt;
    UINT                        nPreBlkWrCnt;
}
ST_STAT_REC, *PST_STAT_REC;

typedef struct _SEEDTARGET_REC
{
#ifndef NDEBUG
    volatile bool               bUsed;
#endif

    volatile bool               bValid;

    bool                        bQuitIfDone;

    SeedTarget                * pSeedTarget;
    DWORD                       dwSTId;
    DWORD                       dwResult;

    ST_STAT_REC                 StStat;
}
SEEDTARGET_REC, *PSEEDTARGET_REC;


class MyObserver : public LP2PEpObserver
{
public:
    MyObserver();
    virtual ~MyObserver();

    virtual void OnConnectFailed(DWORD dwErrCode);
    virtual void OnConnecting();
    virtual void OnConnected();
    virtual void OnDisconnected();

    virtual void OnNodeRegistered(DWORD dwResult);

    virtual void OnSeedTargetFinished(DWORD dwCookie, DWORD dwResult);
};


static vector<SEEDTARGET_REC>   gSTList;
static VectorMate               gSTListMate(false);
static list<UINT>               gSTIdxList;

static volatile DWORD           gdwResult = ERROR_SUCCESS;
static volatile LONG            glTaskNum = 0;
static volatile LONG            glDoneNum = 0;


MyObserver::MyObserver()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MyObserver::~MyObserver()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

void MyObserver::OnConnectFailed(DWORD dwErrCode)
{
    log_wprintf(L"connect to tracker server failed: %d\n", dwErrCode);
}

void MyObserver::OnConnecting()
{
    log_wprintf(L"connecting to tracker server...\n");
}

void MyObserver::OnConnected()
{
    log_wprintf(L"connected to tracker server\n");
}

void MyObserver::OnDisconnected()
{
    log_wprintf(L"disconnected from tracker server\n");
}

void MyObserver::OnNodeRegistered(DWORD dwResult)
{
    log_wprintf(L"register node: %d\n", dwResult);
}

void MyObserver::OnSeedTargetFinished(DWORD dwCookie, DWORD dwResult)
{
    UINT  nIdx = (UINT)dwCookie;
    PSEEDTARGET_REC  pSTRec;

    log_wprintf(L"idx[%d] finished: %d\n", nIdx, dwResult);

    pSTRec = &gSTList[nIdx];
    assert(pSTRec->bValid);

    pSTRec->dwResult = dwResult;

    if ( dwResult != ERROR_SUCCESS )
    {
        gdwResult = dwResult;
    }

    if ( pSTRec->bQuitIfDone )
    {
        InterlockedIncrement(&glDoneNum);
    }
}

static void PrintCmdHelp()
{
    log_wprintf(L"a <seed_file> <target_path> <read_num> <write_num> <quit_if_done> [cache_file]\n");
    log_wprintf(L"d <seed_idx>\n");
    log_wprintf(L"l\n");
    log_wprintf(L"q\n");
}

static void CalcSpeed(SeedTarget *pSeedTarget, PST_STAT_REC pStatRec, UINT &rnDegree, DWORD &rdwRdSpeed, DWORD &rdwWrSpeed)
{
    UINT  nBlkRdCnt, nBlkWrCnt;
    DWORD  dwCurTicks, dwRdDelta, dwWrDelta;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    dwCurTicks = GetTickCount();

    nBlkRdCnt = pSeedTarget->GetBlockReadCount();
    nBlkWrCnt = pSeedTarget->GetBlockWriteCount();

    dwRdDelta = dwCurTicks - pStatRec->dwPreRdTicks;
    dwWrDelta = dwCurTicks - pStatRec->dwPreWrTicks;

    rnDegree = pSeedTarget->GetProgressMask()->GetDegree();

    rdwRdSpeed = (nBlkRdCnt - pStatRec->nPreBlkRdCnt) * pStatRec->dwBlkMBSize * 1000 / dwRdDelta;
    rdwWrSpeed = (nBlkWrCnt - pStatRec->nPreBlkWrCnt) * pStatRec->dwBlkMBSize * 1000 / dwWrDelta;

    if ( (rdwRdSpeed != 0) || ((dwCurTicks - pStatRec->dwPreRdTicks) > 5000) )
    {
        pStatRec->nPreBlkRdCnt = nBlkRdCnt;
        pStatRec->dwPreRdTicks = dwCurTicks;
    }

    if ( (rdwWrSpeed != 0) || ((dwCurTicks - pStatRec->dwPreWrTicks) > 5000) )
    {
        pStatRec->nPreBlkWrCnt = nBlkWrCnt;
        pStatRec->dwPreWrTicks = dwCurTicks;
    }
}

static void DoRemoveSeed(UINT nIdx, PSEEDTARGET_REC pSTRec)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    assert(pSTRec->bValid);
    pSTRec->bValid = false;

    assert(pSTRec->dwSTId != INVALID_LP2P_ST_ID);
    LP2PEndpoint::Instance()->RemoveSeedTarget(pSTRec->dwSTId);

    assert(pSTRec->pSeedTarget);
    pSTRec->pSeedTarget->Fini();
    pSTRec->pSeedTarget->Release();

    DO_IF_DBG(pSTRec->bUsed = false);
    gSTListMate.PutFreeIndex(nIdx);
}

DWORD AddSeed(const char *szCmd)
{
    DWORD  dwRet;
    int  nArgc;
    char  szSeedFile[MAX_PATH], szTargetPath[MAX_PATH], szCacheFile[MAX_PATH];
    UINT  nRdNum, nWrNum, nQuitIfDone;
    UINT  nIdx;
    PSEEDTARGET_REC  pSTRec;
    SeedTarget  *pSeedTarget;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    nArgc = sscanf(szCmd, "%498s %498s %u %u %u %498s", szSeedFile, szTargetPath, &nRdNum, &nWrNum, &nQuitIfDone, szCacheFile);
    if ( nArgc < 5 )
    {
        PrintCmdHelp();
        return ERROR_INVALID_PARAMETER;
    }

    if ( nQuitIfDone )
    {
        ++glTaskNum;
    }

    dwRet = gSTListMate.GetFreeIndex(nIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"get free slot for [%hs] failed: %d\n", szSeedFile, dwRet);
        goto err1;
    }

    pSTRec = &gSTList[nIdx];
    assert(!pSTRec->bUsed);
    DO_IF_DBG(pSTRec->bUsed = true);

    pSeedTarget = new(std::nothrow) SeedTarget(nRdNum, nWrNum);
    if ( !pSeedTarget )
    {
        dwRet = ERROR_OUTOFMEMORY;
        log_wprintf(L"alloc seed target for [%hs] failed: %d\n", szSeedFile, dwRet);
        goto err2;
    }

    dwRet = pSeedTarget->Init(szSeedFile, szTargetPath, (nWrNum == 0), false, ((nArgc > 5) ? szCacheFile : NULL));
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init seed target for [%hs] failed: %d\n", szSeedFile, dwRet);
        goto err3;
    }

    pSTRec->bQuitIfDone = (nQuitIfDone != 0);
    pSTRec->pSeedTarget = pSeedTarget;
    pSTRec->dwSTId = INVALID_LP2P_ST_ID;
    pSTRec->dwResult = ERROR_SUCCESS;

    pSTRec->StStat.dwBlkMBSize = pSeedTarget->GetBlockSize() >> 20;
    pSTRec->StStat.dwStartTicks = GetTickCount();
    pSTRec->StStat.dwPreRdTicks = GetTickCount();
    pSTRec->StStat.dwPreWrTicks = GetTickCount();
    pSTRec->StStat.nPreBlkRdCnt = pSeedTarget->GetBlockReadCount();
    pSTRec->StStat.nPreBlkWrCnt = pSeedTarget->GetBlockWriteCount();

    pSTRec->bValid = true;

    dwRet = LP2PEndpoint::Instance()->AddSeedTarget(pSeedTarget, nIdx, pSTRec->dwSTId);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"add seed target for [%hs] failed: %d\n", szSeedFile, dwRet);
        goto err4;
    }

    gSTIdxList.push_back(nIdx);

    log_wprintf(L"add idx[%d] for [%hs]\n", nIdx, szSeedFile);

    return ERROR_SUCCESS;

err4:
    pSeedTarget->Fini();

err3:
    pSeedTarget->Release();

err2:
    DO_IF_DBG(pSTRec->bUsed = false);
    gSTListMate.PutFreeIndex(nIdx);

err1:
    if ( nQuitIfDone )
    {
        gdwResult = dwRet;
        InterlockedIncrement(&glDoneNum);
    }

    return dwRet;
}

DWORD RemoveSeed(const char *szCmd)
{
    UINT  nIdx;
    PSEEDTARGET_REC  pSTRec;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( sscanf(szCmd, "%u", &nIdx) < 1 )
    {
        PrintCmdHelp();
        return ERROR_INVALID_PARAMETER;
    }

    if ( nIdx >= gSTList.size() )
    {
        log_wprintf(L"idx[%d] invalid\n", nIdx);
        return ERROR_INVALID_PARAMETER;
    }

    pSTRec = &gSTList[nIdx];
    if ( !pSTRec->bValid )
    {
        log_wprintf(L"idx[%d] not used\n", nIdx);
        return ERROR_INVALID_PARAMETER;
    }

    assert(!pSTRec->bQuitIfDone);

    gSTIdxList.remove(nIdx);

    DoRemoveSeed(nIdx, pSTRec);

    log_wprintf(L"removed idx[%d]\n", nIdx);

    return ERROR_SUCCESS;
}

void CleanupSeeds()
{
    UINT  nIdx;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    while ( !gSTIdxList.empty() )
    {
        nIdx = gSTIdxList.front();
        gSTIdxList.pop_front();

        DoRemoveSeed(nIdx, &gSTList[nIdx]);
    }
}

void ListSeeds()
{
    UINT  nIdx;
    SeedTarget  *pSeedTarget;
    UINT  nDegree;
    DWORD  dwRdSpeed, dwWrSpeed;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    for ( list<UINT>::const_iterator itr = gSTIdxList.begin(), end_itr = gSTIdxList.end(); itr != end_itr; ++itr )
    {
        nIdx = *itr;
        assert(nIdx < gSTList.size());

        assert(gSTList[nIdx].bValid);

        pSeedTarget = gSTList[nIdx].pSeedTarget;
        assert(pSeedTarget);

        CalcSpeed(pSeedTarget, &gSTList[nIdx].StStat, nDegree, dwRdSpeed, dwWrSpeed);

        log_wprintf(L"%d: %hs(u[%3dMB/s], d[%3dMB/s]), %3d%%\r", nIdx, pSeedTarget->GetName(), dwRdSpeed, dwWrSpeed, nDegree);
    }
}

bool ProcessCmd(const char *szCmd)
{
    bool  bQuit = false;

    DBG(D_INFO, (L"%s: %hs\n", FULL_FUNC_NAME, szCmd));

    switch ( szCmd[0] )
    {
        case 'a':
            AddSeed(szCmd + 2);
            break;

        case 'd':
            RemoveSeed(szCmd + 2);
            break;

        case 'l':
            ListSeeds();
            break;

        case 'q':
            bQuit = true;
            break;

        default:
            PrintCmdHelp();
            break;
    }

    return bQuit;
}


int main(int argc, char *argv[])
{
    DWORD  dwRet;
    MyObserver  obs;
    UINT  nMaxSeeds, nUploadNum, nSessionNum;
    char  buf[MAX_PATH];

    if ( argc < 6 )
    {
        log_wprintf(L"LiteP2PNode <tracker_ip> <node_name> <max_seeds> <upload_num> <session_num>\n");
        return ERROR_INVALID_PARAMETER;
    }

    nMaxSeeds = (UINT)atoi(argv[3]);
    nUploadNum = (UINT)atoi(argv[4]);
    nSessionNum = (UINT)atoi(argv[5]);

#if defined(_WIN32_) || defined(_WINCE_)
    WSADATA  WSAData;
    WSAStartup(MAKEWORD(2, 2), &WSAData);
#endif

    InitRand();
    log_wprintf(L"random seed: %d\n", rand());

    SHA1Hash::Startup();
#ifdef CONFIG_TLS_SUPPORT
    dwRet = SecureCenter::Instance()->Load(NULL, NULL, NULL);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"SecureCenter init failed: %d\n", dwRet);
        goto err0;
    }
#endif
    ExecutorCenter::Instance()->Load();
    TimerCenter::Instance()->Load();
    ReactorSocket::Instance()->Load();
    ConnectCenter::Instance()->Load();
    dwRet = LP2PEndpoint::Instance()->Load(argv[1], argv[2], nMaxSeeds, nUploadNum, nSessionNum, &obs);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"LP2PEndpoint init failed: %d\n", dwRet);
        goto err1;
    }

    gSTList.resize(nMaxSeeds);
    dwRet = gSTListMate.Init(nMaxSeeds);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"gSTListMate init failed: %d\n", dwRet);
        goto err2;
    }
    for ( UINT i = 0, nSize = (UINT)gSTList.size(); i < nSize; ++i )
    {
        gSTList[i].bValid = false;
    }

    InitStdIn();
    while ( true )
    {
        dwRet = ReadStdIn(buf, _countof(buf), 3000);
        if ( dwRet == WAIT_TIMEOUT )
        {
            if ( glTaskNum && (glTaskNum == glDoneNum) )
            {
                log_wprintf(L"all task(s) finished: %d\n", gdwResult);
                break;
            }

            ListSeeds();
        }
        else if ( dwRet != ERROR_SUCCESS )
        {
            log_wprintf(L"read stdin failed: %d\n", dwRet);
            gdwResult = dwRet;
            break;
        }
        else if ( ProcessCmd(buf) )
        {
            log_wprintf(L"quit\n");
            break;
        }
    }

    CleanupSeeds();

    gSTList.resize(0);
    gSTListMate.Fini();

err2:
    LP2PEndpoint::Instance()->Unload();
err1:
    ConnectCenter::Instance()->Unload();
    ReactorSocket::Instance()->Unload();
    TimerCenter::Instance()->Unload();
    ExecutorCenter::Instance()->Unload();
#ifdef CONFIG_TLS_SUPPORT
    SecureCenter::Instance()->Unload();
#endif

    LP2PEndpoint::Destroy();
    ConnectCenter::Destroy();
    ReactorSocket::Destroy();
    TimerCenter::Destroy();
    ExecutorCenter::Destroy();
#ifdef CONFIG_TLS_SUPPORT
err0:
    SecureCenter::Destroy();
#endif
    SHA1Hash::Cleanup();

#if defined(_WIN32_) || defined(_WINCE_)
    WSACleanup();
#endif

    return gdwResult;
}
