/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#include "SHA1Hash.h"
#include "ExecutorPool.h"
#include "TimerCenter.h"
#include "ReactorSelect.h"
#include "ReactorEPoll.h"
#include "ConnectCenter.h"
#include "SeedTarget.h"
#include "LP2PMCastEndpoint.h"


typedef struct _ST_STAT_REC
{
    DWORD                       dwBlkMBSize;
    DWORD                       dwStartTicks;
    DWORD                       dwPreRdTicks;
    UINT                        nPreBlkRdCnt;
}
ST_STAT_REC, *PST_STAT_REC;

class MyObserver : public LP2PMCastEpObserver
{
public:
    MyObserver();
    virtual ~MyObserver();

    virtual void OnFinished();
};


static volatile bool            gbFinished = false;

static SeedTarget             * gpSeedTarget = NULL;
static ST_STAT_REC              gSTStat;


MyObserver::MyObserver()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MyObserver::~MyObserver()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

void MyObserver::OnFinished()
{
    DBG(D_INFO, (L"%s\n", FULL_FUNC_NAME));

    // no use
//    fclose(stdin);
//    close(fileno(stdin));

    gbFinished = true;
}


static void PrintCmdHelp()
{
    log_wprintf(L"Usage:\n");
    log_wprintf(L"  s: start multicast\n");
    log_wprintf(L"  q: quit program\n");
}


static void CalcSpeed(SeedTarget *pSeedTarget, PST_STAT_REC pStatRec, UINT &rnDegree, DWORD &rdwRdSpeed)
{
    UINT  nBlkRdCnt, nBlkCnt;
    DWORD  dwCurTicks, dwRdDelta;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    dwCurTicks = GetTickCount();

    nBlkRdCnt = pSeedTarget->GetBlockReadCount();
    nBlkCnt = pSeedTarget->GetBlockCount();

    dwRdDelta = dwCurTicks - pStatRec->dwPreRdTicks;

    rnDegree = nBlkRdCnt * 100 / nBlkCnt;

    rdwRdSpeed = (nBlkRdCnt - pStatRec->nPreBlkRdCnt) * pStatRec->dwBlkMBSize * 1000 / dwRdDelta;

    if ( (rdwRdSpeed != 0) || ((dwCurTicks - pStatRec->dwPreRdTicks) > 5000) )
    {
        pStatRec->nPreBlkRdCnt = nBlkRdCnt;
        pStatRec->dwPreRdTicks = dwCurTicks;
    }
}

void PrintProgress()
{
    UINT  nDegree;
    DWORD  dwRdSpeed;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    CalcSpeed(gpSeedTarget, &gSTStat, nDegree, dwRdSpeed);
    log_wprintf(L"%3dMB/s, %3d%%\r", dwRdSpeed, nDegree);
}

bool ProcessCmd(const char *szCmd)
{
    bool  bQuit = false;

    DBG(D_INFO, (L"%s: %hs\n", FULL_FUNC_NAME, szCmd));

    switch ( szCmd[0] )
    {
        case 's':
            gSTStat.dwBlkMBSize = gpSeedTarget->GetBlockSize() >> 20;
            gSTStat.dwStartTicks = GetTickCount();
            gSTStat.dwPreRdTicks = GetTickCount();
            gSTStat.nPreBlkRdCnt = gpSeedTarget->GetBlockReadCount();
            LP2PMCastEndpoint::Instance()->Start();
            break;

        case 'q':
            bQuit = true;
            break;

        default:
            PrintCmdHelp();
            break;
    }

    return bQuit;
}

int main(int argc, char *argv[])
{
    DWORD  dwRet, dwTmp;
    MyObserver  obs;
    char  buf[256];

    if ( argc < 6 )
    {
        log_wprintf(L"MCastP2PNode <tracker_ip> <node_name> <interface_ip> <seed_file> <target_path>\n");
        return ERROR_INVALID_PARAMETER;
    }

#if defined(_WIN32_) || defined(_WINCE_)
    WSADATA  WSAData;
    WSAStartup(MAKEWORD(2, 2), &WSAData);
#endif

    InitRand();
    log_wprintf(L"random seed: %d\n", rand());

    SHA1Hash::Startup();
    ExecutorCenter::Instance()->Load();
    TimerCenter::Instance()->Load();
    ReactorSocket::Instance()->Load();
    ConnectCenter::Instance()->Load();

    gpSeedTarget = new SeedTarget(1, 0);
    dwRet = gpSeedTarget->Init(argv[4], argv[5], true, false);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init seed target failed: %d\n", dwRet);
        goto err1;
    }

    dwRet = LP2PMCastEndpoint::Instance()->Load(argv[1], argv[2], argv[3], gpSeedTarget, &obs);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"LP2PEndpoint init failed: %d\n", dwRet);
        goto err2;
    }

    InitStdIn();
    while ( true )
    {
        dwTmp = ReadStdIn(buf, _countof(buf), 2000);
        if ( dwTmp == WAIT_TIMEOUT )
        {
            if ( gbFinished )
            {
                log_wprintf(L"finished\n");
                break;
            }

            PrintProgress();
        }
        else if ( dwTmp != ERROR_SUCCESS )
        {
            log_wprintf(L"read stdin failed: %d\n", dwTmp);
            break;
        }
        else if ( ProcessCmd(buf) )
        {
            log_wprintf(L"quit\n");
            break;
        }
    }

    LP2PMCastEndpoint::Instance()->Unload();

err2:
    gpSeedTarget->Fini();

err1:
    gpSeedTarget->Release();

    ConnectCenter::Instance()->Unload();
    ReactorSocket::Instance()->Unload();
    TimerCenter::Instance()->Unload();
    ExecutorCenter::Instance()->Unload();

    LP2PMCastEndpoint::Destroy();
    ConnectCenter::Destroy();
    ReactorSocket::Destroy();
    TimerCenter::Destroy();
    ExecutorCenter::Destroy();
    SHA1Hash::Cleanup();

#if defined(_WIN32_) || defined(_WINCE_)
    WSACleanup();
#endif

    return dwRet;
}
