/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#ifdef CONFIG_TLS_SUPPORT
#include "SecureCenter.h"
#endif
#include "SHA1Hash.h"
#include "ExecutorPool.h"
#include "TimerCenter.h"
#include "ReactorSelect.h"
#include "ReactorEPoll.h"
#include "ConnectCenter.h"
#include "SeedTarget.h"
#include "LP2PEndpoint.h"


class MyObserver : public LP2PEpObserver
{
public:
    MyObserver();
    virtual ~MyObserver();

    virtual void OnConnectFailed(DWORD dwErrCode);
    virtual void OnConnecting();
    virtual void OnConnected();
    virtual void OnDisconnected();

    virtual void OnNodeRegistered(DWORD dwResult);

    virtual void OnSeedTargetFinished(DWORD dwCookie, DWORD dwResult);
};


#ifdef _WIN32_
static HANDLE                   ghParentQuit;
static HANDLE                   ghQuit;
#else
static sem_t                    ghQuit;
#endif

static DWORD                    gdwResult = ERROR_CANCELLED;

static SeedTarget             * gpSeedTarget = NULL;
static const BitArray         * gpBitMask = NULL;
static DWORD                    gdwBlkMBSize;

static DWORD                    gdwStartTicks = 0;
static DWORD                    gdwPreRdTicks = 0;
static DWORD                    gdwPreWrTicks = 0;
static UINT                     gnPreBlkRdCnt = 0;
static UINT                     gnPreBlkWrCnt = 0;


MyObserver::MyObserver()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MyObserver::~MyObserver()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

void MyObserver::OnConnectFailed(DWORD dwErrCode)
{
    log_wprintf(L"connect to tracker server failed: %d\n", dwErrCode);
}

void MyObserver::OnConnecting()
{
    log_wprintf(L"connecting to tracker server...\n");
}

void MyObserver::OnConnected()
{
    log_wprintf(L"connected to tracker server\n");
}

void MyObserver::OnDisconnected()
{
    log_wprintf(L"disconnected from tracker server\n");
}

void MyObserver::OnNodeRegistered(DWORD dwResult)
{
    log_wprintf(L"register node OK\n");
}

void MyObserver::OnSeedTargetFinished(DWORD dwCookie, DWORD dwResult)
{
    log_wprintf(L"OnSeedTargetFinished: %d\n", dwResult);

    gdwResult = dwResult;

#ifdef _WIN32_
    SetEvent(ghQuit);
#else
    sem_post(&ghQuit);
#endif
}


void PrintSpeed()
{
    UINT  nBlkRdCnt, nBlkWrCnt, nDegree;
    DWORD  dwCurTicks, dwRdDelta, dwWrDelta, dwRdSpeed, dwWrSpeed;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    dwCurTicks = GetTickCount();

    nBlkRdCnt = gpSeedTarget->GetBlockReadCount();
    nBlkWrCnt = gpSeedTarget->GetBlockWriteCount();

    dwRdDelta = dwCurTicks - gdwPreRdTicks;
    dwWrDelta = dwCurTicks - gdwPreWrTicks;

    nDegree = gpBitMask->GetDegree();

    dwRdSpeed = (nBlkRdCnt - gnPreBlkRdCnt) * gdwBlkMBSize * 1000 / dwRdDelta;
    dwWrSpeed = (nBlkWrCnt - gnPreBlkWrCnt) * gdwBlkMBSize * 1000 / dwWrDelta;
#ifdef _WIN32_
    log_wprintf(L"(u[%3dMB/s], d[%3dMB/s]), %3d%%\n", dwRdSpeed, dwWrSpeed, nDegree);
#else
    log_wprintf(L"(u[%3dMB/s], d[%3dMB/s]), %3d%%\r", dwRdSpeed, dwWrSpeed, nDegree);
#endif

    if ( dwRdSpeed != 0 )
    {
        gnPreBlkRdCnt = nBlkRdCnt;
        gdwPreRdTicks = dwCurTicks;
    }

    if ( dwWrSpeed != 0 )
    {
        gnPreBlkWrCnt = nBlkWrCnt;
        gdwPreWrTicks = dwCurTicks;
    }
}

DWORD WaitForFinished()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    gdwBlkMBSize = gpSeedTarget->GetBlockSize() >> 20;

    gdwStartTicks = GetTickCount();
    gdwPreRdTicks = GetTickCount();
    gdwPreWrTicks = GetTickCount();
    gnPreBlkRdCnt = gpSeedTarget->GetBlockReadCount();
    gnPreBlkWrCnt = gpSeedTarget->GetBlockWriteCount();

    while ( true )
    {
#ifdef _WIN32_
        HANDLE  hQuitEvts[] = {ghParentQuit, ghQuit};
        DWORD  dwWait = WaitForMultipleObjects(_countof(hQuitEvts), hQuitEvts, FALSE, 5000);
        if ( dwWait != WAIT_TIMEOUT )
#else
        timespec  tsWait;
        clock_gettime(CLOCK_REALTIME, &tsWait);
        tsWait.tv_sec += 2;
        if ( !sem_timedwait(&ghQuit, &tsWait) )
#endif
        {
            log_wprintf(L"quit with %d\n", gdwResult);
            break;
        }

        PrintSpeed();
    }

    return gdwResult;
}

int main(int argc, char *argv[])
{
    DWORD  dwRet;
    MyObserver  obs;
    UINT  nRWNum;
    DWORD  dwSTId;

    if ( argc < 6 )
    {
        log_wprintf(L"LiteP2PNode <tracker_ip> <node_name> <target_dir/[target_name]> <seed_file> <rw_num> [cache_file]\n");
        return ERROR_INVALID_PARAMETER;
    }

    nRWNum = (UINT)atoi(argv[5]);

#if defined(_WIN32_) || defined(_WINCE_)
    WSADATA  WSAData;
    WSAStartup(MAKEWORD(2, 2), &WSAData);
#endif

    InitRand();
    log_wprintf(L"random seed: %d\n", rand());

#ifdef _WIN32_
    ghParentQuit = CreateEvent(NULL, TRUE, FALSE, L"Global\\upd-process");
    ghQuit = CreateEvent(NULL, TRUE, FALSE, NULL);
#else
    sem_init(&ghQuit, 0, 0);
#endif

    SHA1Hash::Startup();
#ifdef CONFIG_TLS_SUPPORT
    dwRet = SecureCenter::Instance()->Load(NULL, NULL, NULL);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"SecureCenter init failed: %d\n", dwRet);
        goto err0;
    }
#endif
    ExecutorCenter::Instance()->Load(1);
    TimerCenter::Instance()->Load();
    ReactorSocket::Instance()->Load();
    ConnectCenter::Instance()->Load();
    LP2PEndpoint::Instance()->Load(argv[1], argv[2], 1, nRWNum, 1, &obs);

    gpSeedTarget = new SeedTarget(nRWNum, nRWNum);
    dwRet = gpSeedTarget->Init(argv[4], argv[3], false, false, ((argc > 6) ? argv[6] : NULL));
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init seed target failed: %d\n", dwRet);
        goto err1;
    }

    gpBitMask = gpSeedTarget->GetProgressMask();

    dwRet = LP2PEndpoint::Instance()->AddSeedTarget(gpSeedTarget, 0, dwSTId);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"add seed target failed: %d\n", dwRet);
        goto err2;
    }

    dwRet = WaitForFinished();

    LP2PEndpoint::Instance()->RemoveSeedTarget(dwSTId);

err2:
    gpSeedTarget->Fini();

err1:
    gpSeedTarget->Release();

    LP2PEndpoint::Instance()->Unload();
    ConnectCenter::Instance()->Unload();
    ReactorSocket::Instance()->Unload();
    TimerCenter::Instance()->Unload();
    ExecutorCenter::Instance()->Unload();
#ifdef CONFIG_TLS_SUPPORT
    SecureCenter::Instance()->Unload();
#endif

    LP2PEndpoint::Destroy();
    ConnectCenter::Destroy();
    ReactorSocket::Destroy();
    TimerCenter::Destroy();
    ExecutorCenter::Destroy();
#ifdef CONFIG_TLS_SUPPORT
err0:
    SecureCenter::Destroy();
#endif
    SHA1Hash::Cleanup();

#ifdef _WIN32_
    CloseHandle(ghQuit);
    CloseHandle(ghParentQuit);
#else
    sem_destroy(&ghQuit);
#endif

#if defined(_WIN32_) || defined(_WINCE_)
    WSACleanup();
#endif

    return dwRet;
}
