/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#ifdef CONFIG_TLS_SUPPORT
#include "SecureCenter.h"
#endif
#include "SqlWrapper.h"
#include "SHA1Hash.h"
#include "ExecutorPool.h"
#include "TimerCenter.h"
#include "ReactorSelect.h"
#include "ReactorEPoll.h"
#include "ConnectCenter.h"
#include "SeedTarget.h"
#include "LP2PEndpoint.h"


typedef struct _SEEDTARGET_REC
{
#ifndef NDEBUG
    volatile bool               bUsed;
#endif

    volatile bool               bValid;

    UINT                        nDbIdx;

    SeedTarget                * pSeedTarget;
    DWORD                       dwSTId;
}
SEEDTARGET_REC, *PSEEDTARGET_REC;


class MyObserver : public LP2PEpObserver
{
public:
    MyObserver();
    virtual ~MyObserver();

    virtual void OnConnectFailed(DWORD dwErrCode);
    virtual void OnConnecting();
    virtual void OnConnected();
    virtual void OnDisconnected();

    virtual void OnNodeRegistered(DWORD dwResult);

    virtual void OnSeedTargetFinished(DWORD dwCookie, DWORD dwResult);
};


static UINT                     gnReadNum;

static vector<SEEDTARGET_REC>   gSTList;
static VectorMate               gSTListMate(false);


MyObserver::MyObserver()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

MyObserver::~MyObserver()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

void MyObserver::OnConnectFailed(DWORD dwErrCode)
{
    log_wprintf(L"connect to tracker server failed: %d\n", dwErrCode);
}

void MyObserver::OnConnecting()
{
    log_wprintf(L"connecting to tracker server...\n");
}

void MyObserver::OnConnected()
{
    log_wprintf(L"connected to tracker server\n");
}

void MyObserver::OnDisconnected()
{
    log_wprintf(L"disconnected from tracker server\n");
}

void MyObserver::OnNodeRegistered(DWORD dwResult)
{
    log_wprintf(L"register node: %d\n", dwResult);
}

void MyObserver::OnSeedTargetFinished(DWORD dwCookie, DWORD dwResult)
{
    UINT  nIdx = (UINT)dwCookie;

    log_wprintf(L"idx[%d] finished: %d\n", nIdx, dwResult);

    assert(0);
}

static void PrintCmdHelp()
{
    log_wprintf(L"a <db_idx> <img_file> <seed_file>\n");
    log_wprintf(L"d <db_idx>\n");
    log_wprintf(L"m <db_idx> <status>\n");
    log_wprintf(L"q\n");
}

DWORD DoUpdateStatus(UINT nDbIdx, const char *szStatus)
{
    DWORD  dwRet;
    char  sql[200];

    DBG(D_INFO, (L"%s: db_idx=%d, status=%hs\n", FULL_FUNC_NAME, nDbIdx, szStatus));

    _snprintf(sql, _countof(sql), "UPDATE seed SET status='%s' WHERE id=%u", szStatus, nDbIdx);
    dwRet = SqlWrapper::Instance()->ExecSql(sql);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ExecSql(%hs) failed: %d\n", FULL_FUNC_NAME, sql, dwRet));
    }

    return dwRet;
}

static DWORD DoAddSeed(UINT nDbIdx, const char *szImgFile, const char *szSeedFile)
{
    DWORD  dwRet;
    char  szImgName[256];
    char  szTargetPath[MAX_PATH];
    UINT  nIdx;
    PSEEDTARGET_REC  pSTRec;
    SeedTarget  *pSeedTarget;

    DBG(D_INFO, (L"%s: db_idx=%d, img=%hs\n", FULL_FUNC_NAME, nDbIdx, szImgFile));

    GetFileName(szImgFile, szImgName, _countof(szImgName));
    GetFilePath(szImgFile, szTargetPath, _countof(szTargetPath));

    dwRet = gSTListMate.GetFreeIndex(nIdx);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"get free slot for [%hs] failed: %d\n", szImgName, dwRet);
        goto err1;
    }

    pSTRec = &gSTList[nIdx];
    assert(!pSTRec->bUsed);
    DO_IF_DBG(pSTRec->bUsed = true);

    pSeedTarget = new(std::nothrow) SeedTarget(gnReadNum, 0);
    if ( !pSeedTarget )
    {
        dwRet = ERROR_OUTOFMEMORY;
        log_wprintf(L"alloc seed target for [%hs] failed: %d\n", szImgName, dwRet);
        goto err2;
    }

    dwRet = pSeedTarget->Init(szSeedFile, szTargetPath, true, false);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init seed target for [%hs] failed: %d\n", szImgName, dwRet);
        goto err3;
    }

    pSTRec->nDbIdx = nDbIdx;
    pSTRec->pSeedTarget = pSeedTarget;
    pSTRec->dwSTId = INVALID_LP2P_ST_ID;

    pSTRec->bValid = true;

    dwRet = LP2PEndpoint::Instance()->AddSeedTarget(pSeedTarget, nIdx, pSTRec->dwSTId);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"add seed target for [%hs] failed: %d\n", szImgName, dwRet);
        goto err4;
    }

    log_wprintf(L"add idx[%d] for [%hs]\n", nIdx, szImgName);

    DoUpdateStatus(nDbIdx, "active");

    return ERROR_SUCCESS;

err4:
    pSeedTarget->Fini();

err3:
    pSeedTarget->Release();

err2:
    DO_IF_DBG(pSTRec->bUsed = false);
    gSTListMate.PutFreeIndex(nIdx);

err1:
    DoUpdateStatus(nDbIdx, "error");
    return dwRet;
}

static void DoRemoveSeed(UINT nIdx, PSEEDTARGET_REC pSTRec)
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    assert(pSTRec->bValid);
    pSTRec->bValid = false;

    assert(pSTRec->dwSTId != INVALID_LP2P_ST_ID);
    LP2PEndpoint::Instance()->RemoveSeedTarget(pSTRec->dwSTId);

    assert(pSTRec->pSeedTarget);
    pSTRec->pSeedTarget->Fini();
    pSTRec->pSeedTarget->Release();

    DO_IF_DBG(pSTRec->bUsed = false);
    gSTListMate.PutFreeIndex(nIdx);
}

DWORD AddSeed(const char *szCmd)
{
    UINT  nDbIdx;
    char  szImgFile[MAX_PATH], szSeedFile[MAX_PATH];

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( sscanf(szCmd, "%u %498s %498s", &nDbIdx, szImgFile, szSeedFile) < 2 )
    {
        PrintCmdHelp();
        return ERROR_INVALID_PARAMETER;
    }

    return DoAddSeed(nDbIdx, szImgFile, szSeedFile);
}

DWORD RemoveSeed(const char *szCmd)
{
    UINT  nDbIdx;
    PSEEDTARGET_REC  pSTRec;

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( sscanf(szCmd, "%u", &nDbIdx) < 1 )
    {
        PrintCmdHelp();
        return ERROR_INVALID_PARAMETER;
    }

    for ( UINT i = 0, nSize = (UINT)gSTList.size(); i < nSize; ++i )
    {
        pSTRec = &gSTList.at(i);
        if ( pSTRec->bValid && (pSTRec->nDbIdx == nDbIdx) )
        {
            DoRemoveSeed(i, pSTRec);
            log_wprintf(L"removed db_idx=%d, st_idx=%d\n", nDbIdx, i);
            break;
        }
    }

    return ERROR_SUCCESS;
}

DWORD UpdateStatus(const char *szCmd)
{
    UINT  nDbIdx;
    char  szStatus[256];

    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( sscanf(szCmd, "%u %250s", &nDbIdx, szStatus) < 2 )
    {
        PrintCmdHelp();
        return ERROR_INVALID_PARAMETER;
    }

    return DoUpdateStatus(nDbIdx, szStatus);
}

void RestoreSeeds()
{
    DWORD  dwRet;
    void  *sqlResult;
    char  **sqlRow;
    const char  *sql = "SELECT id,file_path,seed_path FROM seed WHERE status!='raw'";

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    dwRet = SqlWrapper::Instance()->BeginGetTable(sql, &sqlResult);
    if ( dwRet != ERROR_SUCCESS )
    {
        DBG(D_WARN, (L"%s: ExecSql(%hs) failed: %d\n", FULL_FUNC_NAME, sql, dwRet));
        return;
    }

    while ( true )
    {
        sqlRow = SqlWrapper::Instance()->FetchRow(sqlResult);
        if ( !sqlRow )
        {
            break;
        }

        DoAddSeed(atoi(sqlRow[0]), sqlRow[1], sqlRow[2]);
    }

    SqlWrapper::Instance()->EndGetTable(sqlResult);
}

void CleanupSeeds()
{
    PSEEDTARGET_REC  pSTRec;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    for ( UINT i = 0, nSize = (UINT)gSTList.size(); i < nSize; ++i )
    {
        pSTRec = &gSTList.at(i);
        if ( pSTRec->bValid )
        {
            DoRemoveSeed(i, pSTRec);
        }
    }
}

bool ProcessCmd(const char *szCmd)
{
    bool  bQuit = false;

    DBG(D_INFO, (L"%s: %hs\n", FULL_FUNC_NAME, szCmd));

    switch ( szCmd[0] )
    {
        case 'a':
            AddSeed(szCmd + 2);
            break;

        case 'd':
            RemoveSeed(szCmd + 2);
            break;

        case 'm':
            UpdateStatus(szCmd + 2);
            break;

        case 'q':
            bQuit = true;
            break;

        default:
            PrintCmdHelp();
            break;
    }

    return bQuit;
}


DWORD ReadDbPasswd(char *szPass, int nSize)
{
    DWORD  dwRet;
    FILE  *hFile;
    char  buf[200], ch, *p;
    const char  *szPath = "/home/tc/etc/local.ini";
    const char  *szKey = "tcp2p_pass=";

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    hFile = fopen(szPath, "r");
    if ( hFile == NULL )
    {
        dwRet = GetLastError();
        log_wprintf(L"open file (%hs) failed\n", szPath);
        goto err1;
    }

    while ( true )
    {
        if ( fgets(buf, sizeof(buf), hFile) == NULL )
        {
            // ret=0 if eof
            dwRet = ERROR_INVALID_DATA;//GetLastError();
            log_wprintf(L"read file (%hs) failed\n", szPath);
            break;
        }

        p = strstr(buf, szKey);
        if ( p == NULL )
        {
            continue;
        }

        dwRet = ERROR_SUCCESS;
        break;
    }

    if ( dwRet == ERROR_SUCCESS )
    {
        p += strlen(szKey);
        for ( int i = 0; ; ++i )
        {
            ch = p[i];
            if ( ch == 0 )
            {
                break;
            }
            if ( (ch == '\r') || (ch == '\n') )
            {
                p[i] = 0;
                break;
            }
        }
        strncpy(szPass, p, nSize);
    }

    fclose(hFile);

err1:
    return dwRet;
}

int main(int argc, char *argv[])
{
    DWORD  dwRet;
    MyObserver  obs;
    UINT  nMaxSeeds, nUploadNum;
    char  buf[300];

    if ( argc < 8 )
    {
        log_wprintf(L"TCServP2PNode <tracker_ip> <node_name> <max_seeds> <upload_num> <read_num> <db_ip> <db_port>\n");
        return ERROR_INVALID_PARAMETER;
    }

    dwRet = ReadDbPasswd(buf, _countof(buf));
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"ReadDbPasswd failed: %d\n", dwRet);
        return ERROR_NOT_READY;
    }    

    nMaxSeeds = (UINT)atoi(argv[3]);
    nUploadNum = (UINT)atoi(argv[4]);
    gnReadNum = (UINT)atoi(argv[5]);

#if defined(_WIN32_) || defined(_WINCE_)
    WSADATA  WSAData;
    WSAStartup(MAKEWORD(2, 2), &WSAData);
#endif

    InitRand();
    log_wprintf(L"random seed: %d\n", rand());

    dwRet = SqlWrapper::Instance()->Load(argv[6], atoi(argv[7]), "tcp2p", buf, "tc");
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"LP2PEndpoint init failed: %d\n", dwRet);
        goto err1;
    }
    SHA1Hash::Startup();
#ifdef CONFIG_TLS_SUPPORT
    dwRet = SecureCenter::Instance()->Load(NULL, NULL, NULL);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"SecureCenter init failed: %d\n", dwRet);
        goto err1_2;
    }
#endif
    ExecutorCenter::Instance()->Load();
    TimerCenter::Instance()->Load();
    ReactorSocket::Instance()->Load();
    ConnectCenter::Instance()->Load();
    dwRet = LP2PEndpoint::Instance()->Load(argv[1], argv[2], nMaxSeeds, nUploadNum, 0, &obs);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"LP2PEndpoint init failed: %d\n", dwRet);
        goto err2;
    }

    gSTList.resize(nMaxSeeds);
    dwRet = gSTListMate.Init(nMaxSeeds);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"gSTListMate init failed: %d\n", dwRet);
        goto err3;
    }
    for ( UINT i = 0, nSize = (UINT)gSTList.size(); i < nSize; ++i )
    {
        gSTList[i].bValid = false;
    }

    RestoreSeeds();

    while ( true )
    {
        if ( !fgets(buf, _countof(buf), stdin) )
        {
            dwRet = GetLastError();
            log_wprintf(L"read stdin failed: %d\n", dwRet);
            break;
        }
        if ( ProcessCmd(buf) )
        {
            log_wprintf(L"quit\n");
            break;
        }
    }

    CleanupSeeds();

    gSTList.resize(0);
    gSTListMate.Fini();

err3:
    LP2PEndpoint::Instance()->Unload();
err2:
    ConnectCenter::Instance()->Unload();
    ReactorSocket::Instance()->Unload();
    TimerCenter::Instance()->Unload();
    ExecutorCenter::Instance()->Unload();
#ifdef CONFIG_TLS_SUPPORT
    SecureCenter::Instance()->Unload();
#endif
    SqlWrapper::Instance()->Unload();

    LP2PEndpoint::Destroy();
    ConnectCenter::Destroy();
    ReactorSocket::Destroy();
    TimerCenter::Destroy();
    ExecutorCenter::Destroy();
#ifdef CONFIG_TLS_SUPPORT
err1_2:
    SecureCenter::Destroy();
#endif
    SHA1Hash::Cleanup();
err1:
    SqlWrapper::Destroy();

#if defined(_WIN32_) || defined(_WINCE_)
    WSACleanup();
#endif

    return ERROR_SUCCESS;
}
