/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#ifdef CONFIG_TLS_SUPPORT
#include "SecureCenter.h"
#endif
#include "ReactorSelect.h"
#include "ReactorEPoll.h"
#include "ConnectCenter.h"
#include "Endpoint.h"
#include "EpBridgeHandler.h"


class ProxyServer : protected Endpoint
{
public:
    static ProxyServer *Instance();
    static void Destroy();

private:
    static ProxyServer        * _instance;

protected:
    ProxyServer();
    virtual ~ProxyServer();

public:
    DWORD Load(UINT nMaxConnNum);
    virtual void Unload();

protected:
    virtual void OnListenerEnded(DWORD hListener);

    virtual EpProtocolHandler *OnNewHandler(DWORD hListener, Connection *pConnection, DWORD dwConnId);
    virtual void OnAsyncConnected(DWORD dwConnId, EpProtocolHandler *pHandler);
    virtual void OnDisconnected(DWORD dwConnId, EpProtocolHandler *pHandler);
    virtual void OnConnEvent(DWORD dwConnId, EpProtocolHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody);

    virtual void ProcessEpWork(UINT nSigIdx);

private:
    DWORD                       _hListener;
};


ProxyServer *ProxyServer::_instance = NULL;

ProxyServer *ProxyServer::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new ProxyServer();
    }
    return _instance;
}

void ProxyServer::Destroy()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

ProxyServer::ProxyServer()
    : Endpoint()
    , _hListener(INVALID_LISTENER_ID)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

ProxyServer::~ProxyServer()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

DWORD ProxyServer::Load(UINT nMaxConnNum)
{
    SOCK_ADDR  addr;

    DBG(D_LOG, (L"%s: MaxConnNum=%d\n", FULL_FUNC_NAME, nMaxConnNum));

    Endpoint::Load(nMaxConnNum, 0, 0);

    addr.bType = ADDR_TYPE_IPV4;
    addr.IPv4Addr.bStream = 1;
    addr.IPv4Addr.dwIP = 0;
    addr.IPv4Addr.wPort = htons(BRIDGE_SERVICE_PORT);
    AddListener(&addr, false, _hListener);

    return ERROR_SUCCESS;
}

void ProxyServer::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    if ( _hListener != INVALID_LISTENER_ID )
    {
        RemoveListener(_hListener);
        _hListener = INVALID_LISTENER_ID;
    }

    Endpoint::Unload();
}

void ProxyServer::OnListenerEnded(DWORD hListener)
{
    DBG(D_LOG, (L"%s: id(%x)\n", FULL_FUNC_NAME, hListener));
}

EpProtocolHandler *ProxyServer::OnNewHandler(DWORD hListener, Connection *pConnection, DWORD dwConnId)
{
    DBG(D_LOG, (L"%s: listener(%x), connid(%x)\n", FULL_FUNC_NAME, hListener, dwConnId));

    return new(std::nothrow) EpBridgeHandler;
}

void ProxyServer::OnAsyncConnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_LOG, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));

    assert(0);
}

void ProxyServer::OnDisconnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_INFO, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));

    delete pHandler;
}

void ProxyServer::OnConnEvent(DWORD dwConnId, EpProtocolHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody)
{
    DBG(D_INFO, (L"%s: id(%#.8x), code(%d)\n", FULL_FUNC_NAME, dwConnId, pEvtHdr->nCode));

    assert(0);
}

void ProxyServer::ProcessEpWork(UINT nSigIdx)
{
    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, nSigIdx));

    assert(0);
}


int main(int argc, char *argv[])
{
    DWORD  dwRet;
    UINT  nMaxConnNum;

    if ( argc < 2 )
    {
        log_wprintf(L"ProxyServer <max_conns> [ca] [cert] [key]\n");
        return ERROR_INVALID_PARAMETER;
    }

    nMaxConnNum = (UINT)atoi(argv[1]);

#ifdef _LINUX_
    struct rlimit  rlim;
    rlim.rlim_max = rlim.rlim_cur = nMaxConnNum * 2 + 100;
    setrlimit(RLIMIT_NOFILE, &rlim);
#endif

#if defined(_WIN32_) || defined(_WINCE_)
    WSADATA  WSAData;
    WSAStartup(MAKEWORD(2, 2), &WSAData);
#endif

#ifdef CONFIG_TLS_SUPPORT
    if ( argc >= 5 )
    {
        dwRet = SecureCenter::Instance()->Load(argv[2], argv[3], argv[4]);
    }
    else
    {
        dwRet = SecureCenter::Instance()->Load(NULL, NULL, NULL);
    }
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"SecureCenter init failed: %d\n", dwRet);
        goto err0;
    }
#endif
    ReactorSocket::Instance()->Load(nMaxConnNum * 2);
    ConnectCenter::Instance()->Load();
    dwRet = ProxyServer::Instance()->Load(nMaxConnNum);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"ProxyServer init failed: %d\n", dwRet);
        goto err1;
    }

    getchar();

    ProxyServer::Instance()->Unload();
err1:
    ConnectCenter::Instance()->Unload();
    ReactorSocket::Instance()->Unload();
#ifdef CONFIG_TLS_SUPPORT
    SecureCenter::Instance()->Unload();
#endif

    ProxyServer::Destroy();
    ConnectCenter::Destroy();
    ReactorSocket::Destroy();
#ifdef CONFIG_TLS_SUPPORT
err0:
    SecureCenter::Destroy();
#endif

#if defined(_WIN32_) || defined(_WINCE_)
    WSACleanup();
#endif

    return ERROR_SUCCESS;
}
