/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#ifdef CONFIG_TLS_SUPPORT
#include "SecureCenter.h"
#endif
#include "ReactorSelect.h"
#include "ReactorEPoll.h"
#include "ConnectCenter.h"
#include "Endpoint.h"
#include "EpPDUHandler.h"


class StaticTracker : protected Endpoint
{
public:
    static StaticTracker *Instance();
    static void Destroy();

private:
    static StaticTracker      * _instance;

protected:
    StaticTracker();
    virtual ~StaticTracker();

public:
    DWORD Load(UINT nMaxNodeNum);
    virtual void Unload();

protected:
    virtual void OnListenerEnded(DWORD hListener);

    virtual EpProtocolHandler *OnNewHandler(DWORD hListener, Connection *pConnection, DWORD dwConnId);
    virtual void OnAsyncConnected(DWORD dwConnId, EpProtocolHandler *pHandler);
    virtual void OnDisconnected(DWORD dwConnId, EpProtocolHandler *pHandler);
    virtual void OnConnEvent(DWORD dwConnId, EpProtocolHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody);

    virtual void ProcessEpWork(UINT nSigIdx);

private:
    DWORD                       _hSTListener;
};

class StaticTrackerHandler : public EpPDUHandler
{
public:
    StaticTrackerHandler();
    virtual ~StaticTrackerHandler();

    virtual void PrepareSend();

protected:
    virtual void ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody);
    virtual void ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes);

private:
    DWORD OnEnumNodes4Seed(DWORD dwN4SId, const PREQ_ENUM_NODE4SEED pReqEnumN4S, BufferItem *pbiSend);
};


static list<SOCK_ADDR>          gNodeList;


StaticTracker *StaticTracker::_instance = NULL;

StaticTracker *StaticTracker::Instance()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    if ( !_instance )
    {
        DBG(D_VERB, (L"%s: %ls\n", FULL_FUNC_NAME, L"CreateInstance"));
        _instance = new StaticTracker();
    }
    return _instance;
}

void StaticTracker::Destroy()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));

    delete _instance;
    _instance = NULL;
}

StaticTracker::StaticTracker()
    : Endpoint()
    , _hSTListener(INVALID_LISTENER_ID)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

StaticTracker::~StaticTracker()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

DWORD StaticTracker::Load(UINT nMaxNodeNum)
{
    SOCK_ADDR  addr;

    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    Endpoint::Load(nMaxNodeNum, 1, 0);

    addr.bType = ADDR_TYPE_IPV4;
    addr.IPv4Addr.bStream = 1;
    addr.IPv4Addr.dwIP = 0;
    addr.IPv4Addr.wPort = htons(STATIC_TRACK_PORT);
    AddListener(&addr, false, _hSTListener);

    return ERROR_SUCCESS;
}

void StaticTracker::Unload()
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    if ( _hSTListener != INVALID_LISTENER_ID )
    {
        RemoveListener(_hSTListener);
        _hSTListener = INVALID_LISTENER_ID;
    }

    Endpoint::Unload();
}

void StaticTracker::OnListenerEnded(DWORD hListener)
{
    DBG(D_LOG, (L"%s: %x\n", FULL_FUNC_NAME, hListener));
}

EpProtocolHandler *StaticTracker::OnNewHandler(DWORD hListener, Connection *pConnection, DWORD dwConnId)
{
    EpPDUHandler  *pHandler;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    if ( hListener == _hSTListener )
    {
        DBG(D_INFO, (L"%s: StaticTrackerHandler\n", FULL_FUNC_NAME));
        pHandler = new(std::nothrow) StaticTrackerHandler;
        return pHandler;
    }

    DBG(D_ERROR, (L"%s: not match\n", FULL_FUNC_NAME));
    assert(0);

    return NULL;
}

void StaticTracker::OnAsyncConnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_LOG, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

void StaticTracker::OnDisconnected(DWORD dwConnId, EpProtocolHandler *pHandler)
{
    DBG(D_INFO, (L"%s: id(%x)\n", FULL_FUNC_NAME, dwConnId));

    delete pHandler;
}

void StaticTracker::OnConnEvent(DWORD dwConnId, EpProtocolHandler *pHandler, PEP_MSG_HDR pEvtHdr, BufferItem *pEvtBody)
{
    DBG(D_INFO, (L"%s: id(%#.8x), code(%d)\n", FULL_FUNC_NAME, dwConnId, pEvtHdr->nCode));

    assert(0);
}

void StaticTracker::ProcessEpWork(UINT nSigIdx)
{
    DBG(D_INFO, (L"%s: %d\n", FULL_FUNC_NAME, nSigIdx));

    assert(0);
}


StaticTrackerHandler::StaticTrackerHandler()
    : EpPDUHandler(SEED_TRACK_NETQSIZE, SEED_TRACK_REQBSIZE,
                   SEED_TRACK_NETQSIZE, SEED_TRACK_REPBSIZE,
                   0, 0)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

StaticTrackerHandler::~StaticTrackerHandler()
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));
}

void StaticTrackerHandler::PrepareSend()
{
    DBG(D_FUNC, (L"%s\n", FULL_FUNC_NAME));
}

void StaticTrackerHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, BufferItem *pbiBody)
{
    DWORD  dwRet = ERROR_SUCCESS;
    BufferItem  *pbi;
    PRET_REG_NODE  pRetRegNode;
    PRET_ADD_NODE4SEED  pRetAddN4S;

    DBG(D_VERB, (L"%s: bCode=%d, dwParam=%d\n", FULL_FUNC_NAME, bCode, dwParam));

    assert(bReq);

    pbi = GetSendBuf();
    if ( pbi == NULL )
    {
        // client should send the next req only after recv the previous rep,
        // so the send buf should never be used up.
        DBG(D_WARN, (L"%s: no free send buf\n", FULL_FUNC_NAME));
        assert(0);
        ReqEndSession();
        return;
    }

    switch ( bCode )
    {
        case C_ST_REG_NODE:
            pRetRegNode = (PRET_REG_NODE)pbi->Put(sizeof(RET_REG_NODE));
            pRetRegNode->dwNodeId = GetTickCount();
            break;

        case C_ST_ADD_N4S:
            pRetAddN4S = (PRET_ADD_NODE4SEED)pbi->Put(sizeof(RET_ADD_NODE4SEED));
            pRetAddN4S->dwN4SId = GetTickCount();
            break;

        case C_ST_UNREG_NODE:
        case C_ST_REMOVE_N4S:
        case C_ST_UPDATE_N4S_STATE:
            break;

        case C_ST_ENUM_NODES4SEED:
            assert(pbiBody->data_size() == sizeof(REQ_ENUM_NODE4SEED));
            dwRet = OnEnumNodes4Seed(dwParam, (PREQ_ENUM_NODE4SEED)pbiBody->data(), pbi);
            break;

        default:
            DBG(D_WARN, (L"%s: unknown code: %d\n", FULL_FUNC_NAME, bCode));
            assert(0);
            dwRet = ERROR_NOT_SUPPORTED;
            break;
    }

    BuildPDU(bCode, false, true, dwRet, pbi);
    PutSendData(pbi, true);
}

void StaticTrackerHandler::ProcessRecvPDU(BYTE bCode, bool bReq, DWORD dwParam, const PBYTE pData, UINT nDataBytes, UINT nLeftBytes)
{
    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    assert(0);
}

DWORD StaticTrackerHandler::OnEnumNodes4Seed(DWORD dwN4SId, const PREQ_ENUM_NODE4SEED pReqEnumN4S, BufferItem *pbiSend)
{
    UINT  i, nCount;
    PRET_ENUM_NODE4SEED  pRetEnumN4S;
    list<SOCK_ADDR>::const_iterator  itr;

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    pRetEnumN4S = (PRET_ENUM_NODE4SEED)pbiSend->Put(sizeof(RET_ENUM_NODE4SEED) - sizeof(NODE4SEED_INFO));
    nCount = pbiSend->tail_size() / sizeof(NODE4SEED_INFO);

    for ( i = 0, itr = gNodeList.begin(); itr != gNodeList.end(); ++i, ++itr )
    {
        if ( i >= nCount )
        {
            break;
        }

        pRetEnumN4S->N4SInfo[i].NodeAddr = *itr;
        pRetEnumN4S->N4SInfo[i].qwProgressMask = ~0ULL;
    }
    pRetEnumN4S->dwCount = i;

    pRetEnumN4S->dwNextIdx = 0;
    pbiSend->Put(sizeof(NODE4SEED_INFO) * i);

    return ERROR_SUCCESS;
}

DWORD ReadNodesFromFile(const char *szPath)
{
    DWORD  dwRet = ERROR_SUCCESS;
    FILE  *hFile;
    char  szIP[20];
    int  nPort;
    SOCK_ADDR  addr;

    hFile = fopen(szPath, "rb");
    if ( hFile == NULL )
    {
        dwRet = GetLastError();
        log_wprintf(L"open file (%hs) failed\n", szPath);
        goto err1;
    }

    addr.bType = ADDR_TYPE_IPV4;
    addr.IPv4Addr.bStream = 1;

    while ( true )
    {
        if ( fscanf(hFile, "%16s %d", szIP, &nPort) != 2 )
        {
            break;
        }

        log_wprintf(L"add node: IP[%hs], Port[%d]\n", szIP, nPort);
        addr.IPv4Addr.dwIP = inet_addr(szIP);
        addr.IPv4Addr.wPort = htons(nPort);
        gNodeList.push_back(addr);
    }

    fclose(hFile);

err1:
    return dwRet;
}


int main(int argc, char *argv[])
{
    DWORD  dwRet;
    UINT  nMaxNodes;

    if ( argc < 3 )
    {
        log_wprintf(L"StaticTracker <max_nodes> <nodes_file>\n");
        return ERROR_INVALID_PARAMETER;
    }

    nMaxNodes = (UINT)atoi(argv[1]);

    dwRet = ReadNodesFromFile(argv[2]);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"read node list failed: %d\n", dwRet);
        goto err1;
    }

#if defined(_WIN32_) || defined(_WINCE_)
    WSADATA  WSAData;
    WSAStartup(MAKEWORD(2, 2), &WSAData);
#endif

#ifdef CONFIG_TLS_SUPPORT
    dwRet = SecureCenter::Instance()->Load(NULL, NULL, NULL);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"SecureCenter init failed: %d\n", dwRet);
        goto err1_2;
    }
#endif
    ReactorSocket::Instance()->Load();
    ConnectCenter::Instance()->Load();
    dwRet = StaticTracker::Instance()->Load(nMaxNodes);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"LP2PEndpoint init failed: %d\n", dwRet);
        goto err2;
    }

    while ( true )
    {
        SleepForSeconds(3600);
    }

    StaticTracker::Instance()->Unload();

err2:
    ConnectCenter::Instance()->Unload();
    ReactorSocket::Instance()->Unload();
#ifdef CONFIG_TLS_SUPPORT
    SecureCenter::Instance()->Unload();
#endif

    StaticTracker::Destroy();
    ConnectCenter::Destroy();
    ReactorSocket::Destroy();
#ifdef CONFIG_TLS_SUPPORT
err1_2:
    SecureCenter::Destroy();
#endif

#if defined(_WIN32_) || defined(_WINCE_)
    WSACleanup();
#endif

err1:
    return dwRet;
}
