/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <StdAfx.h>
#if defined(_WIN32_) || defined(_WINCE_)
#include <vld.h>
#endif
#include "FileMapping.h"
#include "BitArray.h"


#define MITEM_SIZE              sizeof(DWORD)
#define INVALID_MITEM           DWORD(-1)

static BitArray                 gbaCache;
static QWORD                    gqwImgSize;
static DWORD                    gdwBlkSize;
static UINT                     gnBitCnt;
static DWORD                    gdwMdfSize;


UINT CalcBitCount(QWORD qwFileSize, DWORD dwBlkSize)
{
    // allign with fujin
    return (UINT)(qwFileSize / (QWORD)dwBlkSize + 1);
//    return (UINT)((qwFileSize + dwBlkSize - 1) / (QWORD)dwBlkSize);
}

DWORD OpenOrgMUdf(const char *szMdf, FILE *&rhMdf, const char *szUdf, FILE *&rhUdf)
{
    DWORD  dwRet;
    QWORD  qwMdfSize;

    DBG(D_VERB, (L"%s: %hs\n", FULL_FUNC_NAME, szMdf));

    rhMdf = fopen(szMdf, "rb");
    if ( !rhMdf )
    {
        dwRet = GetLastError();
        log_wprintf(L"open mdf(%hs) failed: %d\n", szMdf, dwRet);
        goto err1;
    }

    dwRet = GetFileLength(rhMdf, qwMdfSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        dwRet = GetLastError();
        log_wprintf(L"get mdf(%hs) size failed: %d\n", szMdf, dwRet);
        goto err2;
    }

    if ( (DWORD)qwMdfSize < gdwMdfSize )
    {
        dwRet = ERROR_INVALID_DATA;
        log_wprintf(L"mdf(%hs) size need %d but %lld\n", szMdf, gdwMdfSize, qwMdfSize);
        goto err2;
    }

    rhUdf = fopen(szUdf, "rb");
    if ( !rhUdf )
    {
        dwRet = GetLastError();
        log_wprintf(L"open udf(%hs) failed: %d\n", szUdf, dwRet);
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    fclose(rhMdf);

err1:
    return dwRet;
}

DWORD CreateNewMUdf(const char *szMdf, FILE *&rhMdf, const char *szUdf, FILE *&rhUdf)
{
    DWORD  dwRet;

    DBG(D_VERB, (L"%s: %hs\n", FULL_FUNC_NAME, szMdf));

    // MAP_SHARED was requested and PROT_WRITE is set,
    // fd must be open in read/write (O_RDWR) mode.
    rhMdf = fopen(szMdf, "w+b");
    if ( !rhMdf )
    {
        dwRet = GetLastError();
        log_wprintf(L"create mdf(%hs) failed: %d\n", szMdf, dwRet);
        goto err1;
    }

    dwRet = SetFileLength(rhMdf, gdwMdfSize);
    if ( dwRet != ERROR_SUCCESS )
    {
        dwRet = GetLastError();
        log_wprintf(L"set mdf(%hs) size to %d failed: %d\n", szMdf, gdwMdfSize, dwRet);
        goto err2;
    }

    rhUdf = fopen(szUdf, "wb");
    if ( !rhUdf )
    {
        dwRet = GetLastError();
        log_wprintf(L"create udf(%hs) failed: %d\n", szUdf, dwRet);
        goto err2;
    }

    return ERROR_SUCCESS;

err2:
    fclose(rhMdf);

err1:
    return dwRet;
}

DWORD CopyBlock(FILE *hOrgUdf, DWORD dwOrgIdx, FILE *hNewUdf, DWORD dwNewIdx)
{
    DWORD  dwRet = ERROR_SUCCESS;
    QWORD  qwOrgOff, qwNewOff;
    size_t  nBytes, nLeft;
    BYTE  buf[512];

    DBG(D_VERB, (L"%s\n", FULL_FUNC_NAME));

    qwOrgOff = (QWORD)gdwBlkSize * (QWORD)dwOrgIdx;
    qwNewOff = (QWORD)gdwBlkSize * (QWORD)dwNewIdx;

    if ( fseeko(hOrgUdf, qwOrgOff, SEEK_SET) )
    {
        dwRet = GetLastError();
        log_wprintf(L"seek org udf to %lld failed: %d\n", qwOrgOff, dwRet);
        return dwRet;
    }

    if ( fseeko(hNewUdf, qwNewOff, SEEK_SET) )
    {
        dwRet = GetLastError();
        log_wprintf(L"seek new udf to %lld failed: %d\n", qwNewOff, dwRet);
        return dwRet;
    }

    nLeft = gdwBlkSize;
    while ( nLeft > 0 )
    {
        nBytes = min(sizeof(buf), nLeft);

        nBytes = fread(buf, 1, nBytes, hOrgUdf);
        if ( nBytes == 0 )
        {
            if ( ferror(hOrgUdf) )
            {
                dwRet = GetLastError();
                log_wprintf(L"read from org udf failed: %d\n", dwRet);
            }
            break;
        }

        if ( fwrite(buf, 1, nBytes, hNewUdf) != nBytes )
        {
            // see misc_fio.txt
            dwRet = GetLastError();
            log_wprintf(L"write to new udf failed: %d\n", dwRet);
            break;
        }

        nLeft -= nBytes;
    }

    return dwRet;
}

int main(int argc, char *argv[])
{
    DWORD  dwRet;
    FILE  *hOrgMdf, *hOrgUdf, *hNewMdf, *hNewUdf;
    FileMapping  OrgMdfMap, NewMdfMap;
    BYTE  *pMapping;
    DWORD  *pOrgMItem, *pNewMItem;
    DWORD  dwOrgIdx, dwNewIdx;
    UINT  nOldPer, nCurPer;

    if ( argc < 8 )
    {
        log_wprintf(L"TuneMUdf.exe <mdf> <udf> <mdf_new> <udf_new> <cache_file> <img_size> <blk_size>\n");
        return ERROR_INVALID_PARAMETER;
    }

    gqwImgSize = _atoi64(argv[6]);
    gdwBlkSize = atoi(argv[7]);
    gnBitCnt = CalcBitCount(gqwImgSize, gdwBlkSize);
    gdwMdfSize = gnBitCnt * MITEM_SIZE;
    log_wprintf(L"img: %lldMB, blk: %dKB, bits: %d, mdf: %d\n", gqwImgSize>>20, gdwBlkSize>>10, gnBitCnt, gdwMdfSize);

    dwRet = OpenOrgMUdf(argv[1], hOrgMdf, argv[2], hOrgUdf);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"open org mdf(%hs) or udf(%hs) failed: %d\n", argv[1], argv[2], dwRet);
        goto err1;
    }

    dwRet = CreateNewMUdf(argv[3], hNewMdf, argv[4], hNewUdf);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"create new mdf(%hs) or udf(%hs) failed: %d\n", argv[3], argv[4], dwRet);
        goto err2;
    }

    dwRet = OrgMdfMap.Init(hOrgMdf, gdwMdfSize, true);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init org mdf(%hs) map failed: %d\n", argv[1], dwRet);
        goto err3;
    }

    dwRet = OrgMdfMap.Map(0, gdwMdfSize, true, false, pMapping);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"map org mdf(%hs) failed: %d\n", argv[1], dwRet);
        goto err4;
    }
    pOrgMItem = (DWORD*)pMapping;

    dwRet = NewMdfMap.Init(hNewMdf, gdwMdfSize, false);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init new mdf(%hs) map failed: %d\n", argv[3], dwRet);
        goto err5;
    }

    dwRet = NewMdfMap.Map(0, gdwMdfSize, false, true, pMapping);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"map new mdf(%hs) failed: %d\n", argv[3], dwRet);
        goto err6;
    }
    pNewMItem = (DWORD*)pMapping;
    memset(pNewMItem, 0xFF, gdwMdfSize);

    dwRet = gbaCache.Init(gnBitCnt, argv[5], false);
    if ( dwRet != ERROR_SUCCESS )
    {
        log_wprintf(L"init cache file(%hs) failed: %d\n", argv[5], dwRet);
        goto err7;
    }

    dwNewIdx = 0;
    nOldPer = 0;
    for ( UINT i = 0; i < gnBitCnt; ++i )
    {
        dwOrgIdx = pOrgMItem[i];

        if ( dwOrgIdx == INVALID_MITEM )
        {
            continue;
        }

        if ( gbaCache.IsBitSet(i) )
        {
            continue;
        }

        pNewMItem[i] = dwNewIdx;
        dwRet = CopyBlock(hOrgUdf, dwOrgIdx, hNewUdf, dwNewIdx);
        if ( dwRet != ERROR_SUCCESS )
        {
            log_wprintf(L"copy block[%d] failed: %d\n", i, dwRet);
            break;
        }
        ++dwNewIdx;

        nCurPer = (i + 1) * 100 / gnBitCnt;
        if ( nCurPer != nOldPer )
        {
            log_wprintf(L"progress: %d%%\r", nCurPer);
            nOldPer = nCurPer;
        }
    }

    gbaCache.Fini();

err7:
    NewMdfMap.Unmap((PBYTE)pNewMItem, gdwMdfSize);

err6:
    NewMdfMap.Fini();

err5:
    OrgMdfMap.Unmap((PBYTE)pOrgMItem, gdwMdfSize);

err4:
    OrgMdfMap.Fini();

err3:
    fclose(hNewMdf);
    fclose(hNewUdf);

err2:
    fclose(hOrgMdf);
    fclose(hOrgUdf);

err1:
    return dwRet;
}
