/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file

**/
#define _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <tchar.h>
#include <windows.h>
#include <locale.h>
#include <string.h>
#include <errno.h>

#include "TcConfig.h"
#include "TcDebugLib.h"
#include "TcGetoptLib.h"
#include "TcNtLib.h"

#define RESOLUTION_CONF "\\Device\\Harddisk0\\Partition1\\resolution.config"

//
// version string
//
#define MAJOR_VERSION    0x1
#define MINOR_VERSION    0x1
#define MICRO_VERSION    0x0
#define VERSION_STRING   L"1.1.0"

//
// define program exit value
//
static enum RetVal {
	RET_SUCCESS = 0,            // success
	RET_ILLEGAL_ARGUMENTS = 1,  // illegal argument
	RET_RT = 2                  // others
};

//
// argument flags struct
//
struct ResoMenuMap {
	int flag_is_attached;
	int flag_width;
	DWORD width;
	int flag_height;
	DWORD height;
	int flag_displayindex;
	DWORD displayindex;
	int flag_version;
	int flag_print;
	int flag_tc;
	int flag_tc_save;
	int flag_test;
	int flag_help;
};

//
// argument flags variable
//
static struct ResoMenuMap g_menumap = { 0 };


/**
get screen width and height from config file

@param width
@param height
@return: 1 on success, 0 on failure
**/
static int getWidthHeight(DWORD *width, DWORD *height) {
	int ret;
	int t;
	unsigned char *p;
	unsigned char buf[256];

	DEBUG((D_INFO, L"%hs:\n", __FUNCTION__));

	ret = 0;
	memset(buf, 0, sizeof(buf));

	if (width == NULL || height == NULL) {
		return ret;
	}

	t = TcOneTimeNtReadFile(buf, sizeof(buf), RESOLUTION_CONF);
	DEBUG((D_INFO, L"%hs: TcOneTimeNtReadFile return: 0x%X\n", __FUNCTION__, ret));
	DEBUG((D_INFO, L"%hs: strlen(buf): 0x%X\n", __FUNCTION__, strlen(buf)));

	if (t <= 0) {
		DEBUG((D_ERROR, L"%hs: fail to load config file: 0x%X\n", __FUNCTION__, ret));
		return ret;
	}

	DEBUG((D_INFO, L"%hs: %hs\n", __FUNCTION__, buf));

	for (p = buf; *p != '\0'; p++) {
		if (*p == 'x' || *p == 'X') {
			*p = '\0';
			break;
		}
	}
	p++;

	if ((atoi(buf) > 0) && (atoi(p) > 0)) {
		*width = (DWORD)atoi(buf);
		*height = (DWORD)atoi(p);

		DEBUG((D_INFO, L"%hs: width = %d\n", __FUNCTION__, *width));
		DEBUG((D_INFO, L"%hs: height = %d\n", __FUNCTION__, *height));
		ret = 1;
	}

	return ret;
}


/**
  Get screen resolution
  @param pdwWidth   pointer to screen width
  @param pdwHeight  pointer to screen height
  @return           TRUE on success, FALSE on failure
**/
BOOL TcGetScreenResolution(LPDWORD pdwWidth, LPDWORD pdwHeight)
{
	BOOL bResult = FALSE;
	DWORD dwResult;

	DEBUG((D_INFO, L"%hs:\n", __FUNCTION__));

	if (!pdwWidth || !pdwHeight)
	{
		DEBUG((D_ERROR, L"illegal arguments\n"));
		goto exit;
	}

	//bResult = FALSE;
	dwResult = GetSystemMetrics(SM_CXSCREEN);
	if (dwResult == 0)
	{
		DEBUG((D_ERROR, L"Fail to get current screen resolution width: 0x%X\n", GetLastError()));
		goto exit;
	}
	DEBUG((D_INFO, L"Get current screen resolution width: %d\n", dwResult));
	*pdwWidth = dwResult;

	dwResult = GetSystemMetrics(SM_CYSCREEN);
	if (dwResult == 0)
	{
		DEBUG((D_ERROR, L"Fail to get current screen resolution height: 0x%X\n", GetLastError()));
		goto exit;
	}
	DEBUG((D_INFO, L"Get current screen resolution height: %d\n", dwResult));
	*pdwHeight = dwResult;

	bResult = TRUE;

exit:
	return bResult;
}


/**
Get current display settings
@param pdwWidth   pointer to screen width
@param pdwHeight  pointer to screen height
@return           TRUE on success, FALSE on failure
**/
BOOL TcGetCurrentSettings(LPDWORD pdwWidth, LPDWORD pdwHeight)
{
	BOOL bResult;
	DWORD dwResult;
	DEVMODE devMode;

	DEBUG((D_INFO, L"%hs:\n", __FUNCTION__));

	bResult = FALSE;

	if (!pdwWidth || !pdwHeight)
	{
		DEBUG((D_ERROR, L"illegal arguments\n"));
		goto exit;
	}

	ZeroMemory(&devMode, sizeof(devMode));
	devMode.dmSize = sizeof(devMode);
	if (EnumDisplaySettings(NULL,                  // current display device
		ENUM_CURRENT_SETTINGS, // get current settings
		&devMode))
	{
		*pdwWidth = devMode.dmPelsWidth;
		*pdwHeight = devMode.dmPelsHeight;
	}
	else
	{
		DEBUG((D_ERROR, L"fail to get current display settings\n"));
		goto exit;
	}

	DEBUG((D_INFO, L"Get current display width: %d\n", *pdwWidth));
	DEBUG((D_INFO, L"Get current display height: %d\n", *pdwHeight));

	bResult = TRUE;

exit:
	return bResult;
}


/**
 check if specified display device is attached

@param dwWidth
@param dwHeight
@param dwDisplayIndex
@return on success, return 0
otherwise, exit with a non-zero value
**/
DWORD TcIsDisplayAttached(LPDWORD pdwWidth, LPDWORD pdwHeight, DWORD dwDisplayIndex)
{
	DEVMODE devMode;
	DISPLAY_DEVICE displayDevice;
	DWORD dwHeight;
	DWORD dwWidth;
	BOOL bRet;

	if (dwDisplayIndex < 0) {
		fwprintf(stderr, L"illegal arguments.\n");
		return RET_ILLEGAL_ARGUMENTS;
	}

	displayDevice.cb = sizeof(DISPLAY_DEVICE);
	bRet = EnumDisplayDevices(NULL, dwDisplayIndex, &displayDevice, 0);
	DEBUG((D_INFO, L"EnumDisplayDevices return: 0x%X\n", bRet));

	if (bRet) {
		DEBUG((D_INFO, L"Device name: %s\n", displayDevice.DeviceName));
		DEBUG((D_INFO, L"Device ID: %s\n", displayDevice.DeviceID));
		DEBUG((D_INFO, L"Device StateFlags: 0x%X\n", displayDevice.StateFlags));

		if ((displayDevice.StateFlags & DISPLAY_DEVICE_ATTACHED_TO_DESKTOP) > 0) {
			DEBUG((D_INFO, L"Device was attached to desktop\n"));
		}
		else {
			DEBUG((D_ERROR, L"Display was not attached to PC\n"));
			return RET_RT;
		}
	}
	else {
		DEBUG((D_ERROR, L"EnumDisplayDevices()\n"));
		return RET_RT;
	}

	//
	// get current settings
	//
	ZeroMemory(&devMode, sizeof(devMode));
	devMode.dmSize = sizeof(DEVMODE);
	devMode.dmDriverExtra = 0;
	wcsncpy(devMode.dmDeviceName, displayDevice.DeviceName, CCHDEVICENAME);
	//devMode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT;
	devMode.dmFields = 0;

	if (EnumDisplaySettings(NULL,  // current display device
		ENUM_CURRENT_SETTINGS,     // get current settings
		&devMode))
	{
		dwWidth = devMode.dmPelsWidth;
		dwHeight = devMode.dmPelsHeight;
	}
	else
	{
		DEBUG((D_ERROR, L"fail to get current display settings\n"));
		return RET_RT;
	}
	DEBUG((D_INFO, L"Get current display width: %d\n", dwWidth));
	DEBUG((D_INFO, L"Get current display height: %d\n", dwHeight));

	if (pdwWidth != NULL)
	{
		*pdwWidth = dwWidth;
	}
	if (pdwHeight != NULL)
	{
		*pdwHeight = dwHeight;
	}

	return RET_SUCCESS;
}


/**
change resolution for specified display device

@param dwWidth               screen resolution width
@param dwHeight              screen resolution height
@param dwDisplayIndex        display index
@param dwFlags               can be set to CDS_TEST to test if changing display settings could succeed,
							 or CDS_UPDATEREGISTRY | CDS_GLOBAL to change display settings immediately
@return on success, return 0
otherwise, exit with a non-zero value
**/
DWORD TcChangeDisplayResolution(DWORD dwWidth, DWORD dwHeight, DWORD dwDisplayIndex, DWORD dwFlags)
{
	DEVMODE devMode;
	DISPLAY_DEVICE displayDevice;
	BOOL bRet;
	LONG lr;
	DWORD dwRet;

	if (dwWidth <= 0 || dwHeight <= 0 || dwDisplayIndex < 0 || dwFlags < 0) {
		fwprintf(stderr, L"illegal arguments.\n");
		return RET_ILLEGAL_ARGUMENTS;
	}

	displayDevice.cb = sizeof(DISPLAY_DEVICE);
	bRet = EnumDisplayDevices(NULL, dwDisplayIndex, &displayDevice, 0);
	DEBUG((D_INFO, L"EnumDisplayDevices return: 0x%X\n", bRet));

	if (bRet) {
		DEBUG((D_INFO, L"Device name: %s\n", displayDevice.DeviceName));
		DEBUG((D_INFO, L"Device ID: %s\n", displayDevice.DeviceID));
		DEBUG((D_INFO, L"Device StateFlags: 0x%X\n", displayDevice.StateFlags));

		if ((displayDevice.StateFlags & DISPLAY_DEVICE_ATTACHED_TO_DESKTOP) > 0) {
			DEBUG((D_INFO, L"Device was attached to desktop\n"));
		}
		else {
			DEBUG((D_ERROR, L"Display was not attached to PC\n"));
			return RET_RT;
		}
	}
	else {
		DEBUG((D_ERROR, L"EnumDisplayDevices()\n"));
		return RET_RT;
	}

	devMode.dmSize = sizeof(DEVMODE);
	devMode.dmDriverExtra = 0;
	wcsncpy(devMode.dmDeviceName, displayDevice.DeviceName, CCHDEVICENAME);
	devMode.dmPelsWidth = dwWidth;
	devMode.dmPelsHeight = dwHeight;
	devMode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT;
	//dwFlags = CDS_UPDATEREGISTRY | CDS_GLOBAL;
	DEBUG((D_INFO, L"Device name: %ls (bufferlen = %d)\n", devMode.dmDeviceName, CCHDEVICENAME));

	lr = ChangeDisplaySettings(&devMode, dwFlags);
	DEBUG((D_INFO, L"ChangeDisplaySettings return: 0x%X\n", lr));
	if (lr == DISP_CHANGE_SUCCESSFUL) {
		dwRet = RET_SUCCESS;
		DEBUG((D_INFO, L"Change resolution successfully\n"));
	}
	else
	{
		dwRet = RET_RT;
		if (lr == DISP_CHANGE_BADMODE) {
			DEBUG((D_INFO, L"Mode not supported\n"));
		}
		else if (lr == DISP_CHANGE_RESTART) {
			DEBUG((D_INFO, L"The computer needs to restart before the change takes effect\n"));
		}
		else
		{
			DEBUG((D_INFO, L"Code = %ld\n", lr));
		}
	}
	return dwRet;
}


/**
save config to tc storeage device
if both dwWidth and dwHeight is 0, then save current display settings
@param
@param
@return 0 on success
**/
DWORD TcSaveDisplaySettings(DWORD dwWidth, DWORD dwHeight)
{
	DWORD dwRet;
	DWORD dwTemp;
	DWORD dwLocalWidth;
	DWORD dwLocalHeight;
	char buf[MAX_PATH];

	DEBUG((D_INFO, L"%hs:\n", __FUNCTION__));

	dwRet = 1;

	if (dwWidth == 0 && dwHeight == 0)
	{
		DEBUG((D_INFO, L"will try to get current display settings\n"));
		if (!TcGetCurrentSettings(&dwLocalWidth, &dwLocalHeight))
		{
			DEBUG((D_ERROR, L"TcGetCurrentSettings: fail to get current settings\n"));
			goto exit;
		}
	}
	else
	{
		dwLocalWidth = dwWidth;
		dwLocalHeight = dwHeight;
	}

	DEBUG((D_INFO, L"%hs: dwLocalWidth = %d\n", __FUNCTION__, dwLocalWidth));
	DEBUG((D_INFO, L"%hs: dwLocalHeight = %d\n", __FUNCTION__, dwLocalHeight));

	dwTemp = TcChangeDisplayResolution(dwLocalWidth, dwLocalHeight, 0, CDS_UPDATEREGISTRY | CDS_GLOBAL);
	if (dwTemp != 0)
	{
		DEBUG((D_ERROR, L"TcChangeDisplayResolution: fail to set the settings\n"));
		goto exit;
	}

	ZeroMemory(buf, sizeof(buf));
	sprintf_s(buf, sizeof(buf), "%dx%d", dwLocalWidth, dwLocalHeight);
	dwTemp = TcOneTimeNtWriteFile(buf, strlen(buf), RESOLUTION_CONF);
	if (dwTemp != 0)
	{
		DEBUG((D_ERROR, L"TcOneTimeNtWriteFile: fail to write buf\n"));
		goto exit;
	}

	dwRet = 0;

exit:
	return dwRet;
}



static void show_help(LPCWSTR lpszProgname)
{
	fwprintf(stderr, L"DESCRIPTION\n");
	fwprintf(stderr, L"  Change screen resolutions\n");
	fwprintf(stderr, L"\n");
	fwprintf(stderr, L"SYNOPSIS\n");
	fwprintf(stderr, L"  %s [OPTIONS]\n", lpszProgname);
	fwprintf(stderr, L"\n");
	fwprintf(stderr, L"OPTIONS\n");
	fwprintf(stderr, L"       --is-attached                   check if display is attached, default display index is 0\n");
	fwprintf(stderr, L"  -d,  --displayindex <DISPLAY_INDEX>  set display index, default 0\n");
	fwprintf(stderr, L"       --help                          show this help page\n");
	fwprintf(stderr, L"  -h,  --height <HEIGHT>               screen height\n");
	fwprintf(stderr, L"       --print                         print the current display resolution\n");
	fwprintf(stderr, L"       --tc                            tc specific screen resolution\n");
	fwprintf(stderr, L"       --tc-save                       save tc specific screen resolution\n");
	fwprintf(stderr, L"       --test                          test if specific screen resolution could be set, default 1024x768\n");
	fwprintf(stderr, L"  -v,  --version                       print version string\n");
	fwprintf(stderr, L"  -w,  --width <WIDTH>                 screen width\n");
	fwprintf(stderr, L"\n");
	fwprintf(stderr, L"EXAMPLE\n");
	fwprintf(stderr, L"  %ls -w 800 -h 600\n", lpszProgname);
	fwprintf(stderr, L"  %ls -w 1024 -h 768 -d 0\n", lpszProgname);
	fwprintf(stderr, L"  %ls -w 1024 -h 768 -d 0 --test\n", lpszProgname);
	fwprintf(stderr, L"  %ls --is-attached --displayindex 0\n", lpszProgname);
}

static void show_usage(LPCWSTR lpszProgname)
{
	fwprintf(stderr, L"%s: illegal arguments!\n", lpszProgname);
	fwprintf(stderr, L"\n");
	fwprintf(stderr, L"Usage: try \"%s --help\" for mor information\n", lpszProgname);
	fwprintf(stderr, L"\n");
}

/**
parse arguments and return

@return
0 on success on a single option is set without any of ( -width & -height )
1 on success and both --width & --height is specified
-1 on error
**/
static int parse_arguments(int argc, WCHAR* argv[]) {
	int k;
	int c;
	int option_index;
	WCHAR short_options[] = L"d:h:tw:v";
	struct option long_options[] =
	{
	  {L"is-attached", no_argument, &g_menumap.flag_is_attached, 1},
	  {L"displayindex", required_argument, NULL, L'd' },
	  {L"height", required_argument, NULL, L'h'},
	  {L"help", no_argument, &g_menumap.flag_help, 1},
	  {L"print", no_argument, &g_menumap.flag_print, 1 },
	  {L"tc", no_argument, &g_menumap.flag_tc, 1 },
	  {L"tc-save", no_argument, &g_menumap.flag_tc_save, 1 },
	  {L"test", no_argument, &g_menumap.flag_test, 1},
	  {L"version", no_argument, NULL, L'v'},
	  {L"width", required_argument, NULL, L'w'},
	  {NULL, 0, NULL, 0}
	};

	while (1)
	{
		option_index = 0;
		c = getopt_long(argc, argv, short_options, long_options, &option_index);

		if (c == -1)
			break;

		switch (c)
		{
		case 0:
			if (long_options[option_index].flag != NULL)
			{
				DEBUG((D_INFO, L"--%ls\n", long_options[option_index].name));
			}
			break;

		case L'd':
			DEBUG((D_INFO, L"option -d, %ls\n", optarg));
			g_menumap.flag_displayindex = 1;
			g_menumap.displayindex = (DWORD)_wtoi(optarg);
			break;

		case L'h':
			DEBUG((D_INFO, L"option -h, %ls\n", optarg));
			g_menumap.flag_height = 1;
			g_menumap.height = _wtoi(optarg);
			break;

		case L'v':
			DEBUG((D_INFO, L"option -v\n"));
			g_menumap.flag_version = 1;
			break;

		case L'w':
			DEBUG((D_INFO, L"option -w, %ls\n", optarg));
			g_menumap.flag_width = 1;
			g_menumap.width = _wtoi(optarg);
			break;

		case L'?':
			DEBUG((D_ERROR, L"unknown option\n"));
			return -1;
			break;

		default:
			DEBUG((D_ERROR, L"You are not supposed to reach here\n"));
			return -1;
			break;
		}
	}

	k = 0;
	if (g_menumap.flag_is_attached) {
		k++;
	}
	//if (g_menumap.flag_displayindex) {
	   // k++;
	//}
	//if (g_menumap.flag_height) {
	//  k++;
	//}
	if (g_menumap.flag_help) {
		k++;
	}
	//if (g_menumap.flag_width) {
	//  k++;
	//}
	if (g_menumap.flag_version) {
		k++;
	}
	if (g_menumap.flag_print) {
		k++;
	}
	if (g_menumap.flag_tc) {
		k++;
	}
	if (g_menumap.flag_tc_save) {
		k++;
	}
	if (g_menumap.flag_test) {
		k++;
	}

	//
	// no single option
	//
	if (k == 0) {
		if (!g_menumap.flag_width && !g_menumap.flag_height)
		{
			//
			// no arguments
			//
			return 0;
		}
		else if (g_menumap.flag_width && g_menumap.flag_height)
		{
			//
			// --width <WIDTH> --height <HEIGHT> [ --displayindex <INDEX> ]
			//
			return 1;
		}
		else
		{
			//
			// --width <WIDTH> [ --displayindex <INDEX> ]
			// --height <HEIGHT> [ --displayindex <INDEX> ]
			//
			return -1;
		}
	}
	//
	// --version
	//
	if (g_menumap.flag_version && k > 1) {
		//DEBUG((D_ERROR, L"illegal arguments\n"));
		return -1;
	}
	//
	// --help
	//
	if (g_menumap.flag_help && k > 1) {
		//DEBUG((D_ERROR, L"illegal arguments\n"));
		return -1;
	}
	//
	// --print
	//
	if (g_menumap.flag_print && k > 1) {
		//DEBUG((D_ERROR, L"illegal arguments\n"));
		return -1;
	}
	//
	// --tc
	//
	if (g_menumap.flag_tc && k > 1) {
		//DEBUG((D_ERROR, L"illegal arguments\n"));
		return -1;
	}
	//
	// --tc-save --width <WIDTH> --height <HEIGHT>
	// --tc-save --width <WIDTH> --height <HEIGHT> [ --displayindex <INDEX> ]
	//
	if (g_menumap.flag_tc_save) {
		if (k > 1)
		{
			//DEBUG((D_ERROR, L"illegal arguments\n"));
			return -1;
		}
		if (!g_menumap.flag_width && !g_menumap.flag_height) {
			return 0;
		}
		else if (g_menumap.flag_width && g_menumap.flag_height) {
			return 1;
		}
		else
		{
			//DEBUG((D_ERROR, L"illegal arguments\n"));
			return -1;
		}
	}
	//
	// --test
	// --test --width <WIDTH> --height <HEIGHT>
	// --test --width <WIDTH> --height <HEIGHT> [ --displayindex <INDEX> ]
	//
	if (g_menumap.flag_test) {
		if (k > 1) {
			//DEBUG((D_ERROR, L"illegal arguments\n"));
			return -1;
		}
		if (!g_menumap.flag_width && !g_menumap.flag_height) {
			DEBUG((D_INFO, L"test with default value: --width 1024 --height 768\n"));
			return 0;
		}
		else if (g_menumap.flag_width && g_menumap.flag_height) {
			return 1;
		}
		else 
		{
			//DEBUG((D_ERROR, L"illegal arguments\n"));
			return -1;
		}
	}

	return 0;
}

// on success, exit with zero
// otherwise, exit with a non-zero value
int __cdecl wmain(int argc, WCHAR* argv[])
{
	int nParseArgs;
	DWORD dwRet;
	DWORD dwFlags;

	DEBUG_INIT(TO_STD_CONSOLE, NULL);
	_wsetlocale(LC_ALL, L"");

	//
	// parse arguments
	//
	nParseArgs = parse_arguments(argc, argv);
	if (nParseArgs < 0) {
		show_usage((LPCWSTR)argv[0]);
		dwRet = RET_ILLEGAL_ARGUMENTS;
		goto exit0;
	}
	else if (nParseArgs == 0) {
		if (g_menumap.flag_help) {
			show_help((LPCWSTR)argv[0]);
			dwRet = RET_SUCCESS;
			goto exit0;
		}

		else if (g_menumap.flag_version) {
			fwprintf(stdout, L"version: %ls\n", VERSION_STRING);
			dwRet = RET_SUCCESS;
			goto exit0;
		}

		else if (g_menumap.flag_is_attached) {
			dwRet = TcIsDisplayAttached(&g_menumap.width, &g_menumap.height, g_menumap.displayindex);
			goto exit1;
		}

		else if (g_menumap.flag_tc) {
			if (getWidthHeight(&g_menumap.width, &g_menumap.height)) {
				DEBUG((D_INFO, L"load config successfully\n"));
			}
			else {
				DEBUG((D_ERROR, L"fail to load config\nexit!!\n"));
				dwRet = -1;
				goto exit1;
			}
		}

		else if (g_menumap.flag_tc_save) {
			//
			// get and set current settings
			//
			dwRet = TcSaveDisplaySettings(0, 0);
			goto exit1;
		}

		else if (g_menumap.flag_test)
		{
			g_menumap.width = 1024;
			g_menumap.height = 768;
		}

		else if (g_menumap.flag_print) {
			dwRet = TcGetScreenResolution(&g_menumap.width, &g_menumap.height) ? 0 : 1;
			goto exit1;
		}

		else {
			//
			// no options, print help
			//
			show_help((LPCWSTR)argv[0]);
			dwRet = RET_SUCCESS;
			goto exit0;
		}
	}
	else
	{
		// nParseArgs > 0
		// tc width & height
		// 
		if (g_menumap.flag_tc_save) {
			dwRet = TcSaveDisplaySettings(g_menumap.width, g_menumap.height);
			goto exit1;
		}
		else if (g_menumap.flag_test)
		{
			;
		}
		else // (g_menumap.flag_width && g_menumap.flag_height)
		{
			;
		}
	}

	//
	// default index: 0
	//
	if (g_menumap.flag_displayindex == 0) {
		g_menumap.displayindex = 0;
	}

	DEBUG((D_INFO, L"dwDisplayIndex = %lu\n", g_menumap.displayindex));
	DEBUG((D_INFO, L"dwWidth = %lu\n", g_menumap.width));
	DEBUG((D_INFO, L"dwHeight = %lu\n", g_menumap.height));

	if (g_menumap.flag_test)
	{
		dwFlags = CDS_TEST;
	}
	else
	{
		dwFlags = CDS_UPDATEREGISTRY | CDS_GLOBAL;
	}

	dwRet = TcChangeDisplayResolution(g_menumap.width, g_menumap.height, g_menumap.displayindex, dwFlags);

exit1:
	DEBUG((D_INFO, L"dwRet = %lu\n", dwRet));

exit0:
	DEBUG_FINISH();
	return dwRet;
}
