/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file
  debug implementation
**/
#include <wchar.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#if defined(_WIN32)
#include <windows.h> // windows.h provides OutputDebugString
#else
#include <linux/limits.h> // PATH_MAX
#endif
#include "TcDebugLib.h"

typedef struct _DEBUG_FILE_POINTER
{
  int   printtype;
  int   debugtype;
  FILE  *debug;
  FILE  *print;
} DEBUG_FILE_POINTER;

static DEBUG_FILE_POINTER gFp = {0};

////////////////////////////////////////////////////////////////////////////////

static int mywcstombs(char **szstr, wchar_t *wszstr)
{
  int ret;
  int cnt;

  ret = -1;
  cnt = wcstombs(NULL, wszstr, 0);
  *szstr = (char *) calloc(cnt + 1, sizeof(char));
  if (*szstr == NULL)
  {
    return ret;
  }

  ret = wcstombs(*szstr, wszstr, cnt + 1);

  return ret;
}

static int mymbstowcs(wchar_t **wszstr, char *szstr)
{
  int ret;
  int cnt;

  ret = -1;
  cnt = mbstowcs(NULL, szstr, 0);
  *wszstr = (wchar_t *) calloc(cnt + 1, sizeof(wchar_t));
  if (*wszstr == NULL)
  {
    return ret;
  }

  ret = mbstowcs(*wszstr, szstr, cnt + 1);

  return ret;
}


/**
  Set value
  @param streamType
  @return none
**/
static
void
SetDebugType (
  OUTPUT_STREAM_ENUM streamType
  )
{
  gFp.debugtype = streamType;
}

/**
  Get value
  @param streamType
  @return value of the instance
**/
static
OUTPUT_STREAM_ENUM
GetDebugType (
  void
  )
{
  return gFp.debugtype;
}

/**
  Set value
  @param streamType
  @return none
**/
static
void
SetPrintType (
  OUTPUT_STREAM_ENUM streamType
  )
{
  gFp.printtype = streamType;
}


/**
  Get value
  @param streamType
  @return value of the instance
**/
static
OUTPUT_STREAM_ENUM
GetPrintType (
  void
  )
{
  return gFp.printtype;
}

/**
  Set value
  @param streamType
  @return none
**/
static
void
SetDebugFp (
  FILE *fp
  )
{
  gFp.debug = fp;
}

/**
  Get value
  @param streamType
  @return value of the instance
**/
static
FILE*
GetDebugFp (
  void
  )
{
  return gFp.debug;
}

/**
  Set value
  @param streamType
  @return none
**/
static
void
SetPrintFp (
  FILE *fp
  )
{
  gFp.print = fp;
}

/**
  Get value
  @param streamType
  @return value of the instance
**/
static
FILE*
GetPrintFp (
  void
  )
{
  return gFp.print;
}

/**
  Debug initialization
  
  @param streamEnumType  one value of TO_STD_CONSOLE, TO_FILE and TO_CONSOLE
  @param filename        when streamEnumType is TO_FILE, filename specify a file name;
                         when streamEnumType has other value, this parameter is ignored.
  
  @return 0 on success
**/

int
TC_STDCALL
TcDebugInit (
  OUTPUT_STREAM_ENUM streamEnumType,
  wchar_t *filename
  )
{
  int    ret;
  FILE   *fp;
#if !defined(_WIN32)
  char   filenamea[PATH_MAX];
#endif
  
  ret = 0;
  
  switch (streamEnumType)
  {
  case TO_STD_CONSOLE:
    SetDebugType(TO_STD_CONSOLE);
    SetDebugFp(stderr);
    break;

  case TO_FILE:
#if defined(_WIN32)
    _wfopen_s(&fp, filename, L"w+");
#else
    memset(filenamea, 0, sizeof(filenamea));
    wcstombs(filenamea, filename, sizeof(filenamea) - 1);
    fp = fopen(filenamea, "w+");
#endif
    if (fp == NULL)
    {
      ret = 1;
    }
    else
    {
      SetDebugType(TO_FILE);
      SetDebugFp(fp);
	  fclose(fp);
    }
    break;

  case TO_CONSOLE:
    SetDebugType(TO_CONSOLE);
    break;

  default:
    ret = 2;
    break;
  }
  
  return ret;
}

/**
  Print initialization
  
  @param streamEnumType  one value of TO_STD_CONSOLE, TO_FILE and TO_CONSOLE
  @param filename        when streamEnumType is TO_FILE, filename specify a file name;
                         when streamEnumType has other value, this parameter is ignored.
  
  @return 0 on success
**/
int
TC_STDCALL
TcPrintInit (
  OUTPUT_STREAM_ENUM streamEnumType,
  wchar_t *filename
  )
{
  int    ret;
  int    tmp;
  FILE   *fp;
#if !defined(_WIN32)
  char   filenamea[PATH_MAX];
#endif
  
  ret = 0;
  
  switch (streamEnumType)
  {
  case TO_STD_CONSOLE:
    SetPrintType(TO_STD_CONSOLE);
    SetPrintFp(stdout);
    break;

  case TO_FILE:
#if defined(_WIN32)
    _wfopen_s(&fp, filename, L"w+");
#else
    memset(filenamea, 0, sizeof(filenamea));
    wcstombs(filenamea, filename, sizeof(filenamea) - 1);
    fp = fopen(filenamea, "w+");
#endif
    if (fp == NULL)
    {
      ret = 1;
    }
    else
    {
      SetPrintType(TO_FILE);
      SetPrintFp(fp);
	  fclose(fp);
    }
    break;

  case TO_CONSOLE:
    SetPrintType(TO_CONSOLE);
    break;

  default:
    ret = 2;
    break;
  }

  return ret;
}

/**
  close the open deubg handle
  
  @return none
**/
int
TC_STDCALL
TcDebugFinish (
  void
  )
{
  OUTPUT_STREAM_ENUM debugType;
  FILE               *debugFp;
  
  debugType = GetDebugType();
  debugFp   = GetDebugFp();

  switch (debugType)
  {
  case TO_STD_CONSOLE:
    break;
    
  case TO_FILE:
    if (debugFp != NULL)
    {
      fclose(debugFp);
      SetDebugFp(NULL);
    }
    break;
  
  case TO_CONSOLE:
    break;
    
  default:
    break;
  }
  
  return 0;
}

/**
  close the open print handle
  
  @return none
**/
int
TC_STDCALL
TcPrintFinish (
  void
  )
{
  OUTPUT_STREAM_ENUM printType;
  FILE               *printFp;
  
  printType = GetPrintType();
  printFp   = GetPrintFp();

  switch (printType)
  {
  case TO_STD_CONSOLE:
    break;
    
  case TO_FILE:
    if (printFp != NULL)
    {
      fclose(printFp);
      SetPrintFp(NULL);
    }
    break;
  
  case TO_CONSOLE:
    break;
    
  default:
    break;
  }
  
  return 0;
}


/**
  output a format string to debug console
  
  @param fmt     string include specifiers
  @param ...
  @return number of characters written, or negative value if errors occurr
**/
int
TC_STDCALL
TcDebugOutput (
  const wchar_t *fmt,
  ...
  )
{
  #define LINE_LIMIT_NUMBER 1024
  va_list   argptr;
  wchar_t   wbuf[LINE_LIMIT_NUMBER];
  int       ret;
  
  va_start(argptr, fmt);
  switch (gFp.debugtype)
  {
  case TO_STD_CONSOLE:
  case TO_FILE:
    ret = vfwprintf(GetDebugFp(), fmt, argptr);
    fflush(GetDebugFp());
    break;
    
  case TO_CONSOLE:
    ret = vswprintf(wbuf, sizeof(wbuf) / sizeof(wbuf[0]), fmt, argptr);
#if defined(_WIN32)
    OutputDebugStringW(wbuf); 
#endif
    break;
  
  default:
    break;
  }
  va_end(argptr);
  
  return ret;
}


/**
  output a format string to print console
  
  @param fmt     string include specifiers
  @param ...
  @return number of characters written, or negative value if errors occurr
**/
int
TC_STDCALL
TcPrintOutput (
  const wchar_t *fmt,
  ...
  )
{
  #define LINE_LIMIT_NUMBER 1024
  va_list   argptr;
  wchar_t   wbuf[LINE_LIMIT_NUMBER];
  int       ret;
  
  va_start(argptr, fmt);
  switch (gFp.printtype)
  {
  case TO_STD_CONSOLE:
  case TO_FILE:
    ret = vfwprintf(GetPrintFp(), fmt, argptr);
    fflush(GetPrintFp());
    break;
    
  case TO_CONSOLE:
    ret = vswprintf(wbuf, sizeof(wbuf) / sizeof(wbuf[0]), fmt, argptr);
#if defined(_WIN32)
    OutputDebugStringW(wbuf);
#endif
    break;
  
  default:
    break;
  }
  va_end(argptr);
  
  return ret;
}
