/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file

**/
#define _CRT_SECURE_NO_WARNINGS
#include <tchar.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "TcConfig.h"
#include "TcDebugLib.h"
#include "TcNtLib.h"
#include "TcKeyValueLib.h"
#include "TcGetoptLib.h"

#define VERSION_STRING  _TEXT("0.2")
#define TC_NETFILE      "\\Device\\Harddisk0\\Partition1\\tcnet.conf"
#define TC_DELIMITER    "="

#define NETINFO_MAX     7
#define NETKEY_SZ       64
#define NETVAL_SZ       64
#define TMPBUFSZ        1024

#define KEY_BOOTPROTO   "BOOTPROTO"
#define KEY_HWADDR      "HWADDR"
#define KEY_IPADDR      "IPADDR"
#define KEY_NETMASK     "NETMASK"
#define KEY_GATEWAY     "GATEWAY"
#define KEY_DNS_PRIMARY "DNS"
#define KEY_DNS_BACKUP  "DNS2"

typedef struct keyval_struc
{
  char key[NETKEY_SZ];
  char val[NETVAL_SZ];
}TCNETINFO [NETINFO_MAX];
TCNETINFO g_netinfo = {0};

static void init_tcnetinfo() 
{
  strcpy(g_netinfo[0].key, KEY_BOOTPROTO);
  strcpy(g_netinfo[1].key, KEY_HWADDR);
  strcpy(g_netinfo[2].key, KEY_IPADDR);
  strcpy(g_netinfo[3].key, KEY_NETMASK);
  strcpy(g_netinfo[4].key, KEY_GATEWAY);  
  strcpy(g_netinfo[5].key, KEY_DNS_PRIMARY);
  strcpy(g_netinfo[6].key, KEY_DNS_BACKUP);
}

static const char* set_tcnetinfo(const char *key, const char *val) 
{
  int j;
  for (j = 0; j < NETINFO_MAX; j++)
  {
    if (!strcmp(g_netinfo[j].key, key))
    {
      DEBUG((D_INFO, L"set key-value pair [%hs, %hs]\n", key, val));
      memset(g_netinfo[j].val, 0, sizeof(g_netinfo[j].val));
      strncpy(g_netinfo[j].val, val, sizeof(g_netinfo[j].val) - 1);
      return g_netinfo[j].val;
    }
  }
  return NULL;
}

static const char* get_tcnetinfo(const char *key)
{
  int j;
  for (j = 0; j < NETINFO_MAX; j++)
  {
    if (!strcmp(g_netinfo[j].key, key))
    {
      return g_netinfo[j].val;
    }
  }
  return NULL;
}

// on success, return 0
static int print_tcnetinfo(const char *key)
{
  int ret = 0;
  const char *val = get_tcnetinfo(key);
  if (val && *(char *)val){
    PRINT(L"%hs\n", val);
  }
  return ret;
}

// on success, return index
// otherwise, return -1
static int get_key_index(const char *key)
{
  int ret;
  int j;
  
  ret = -1;
  for (j = 0; j < NETINFO_MAX; j++)
  {
    if (!strcmp(g_netinfo[j].key, key))
    {
      ret = j;
      break;
    }
  }
  return ret;
}

// on success, return 0
static print_tcnetinfo_row(int j)
{
  if (j < 0 || j > NETINFO_MAX || j == NETINFO_MAX)
  {
    return 5;
  }
  
  if (strcmp(g_netinfo[j].val, ""))
  {
    PRINT(L"%hs = %hs\n", g_netinfo[j].key, g_netinfo[j].val);
  }
  return 0;
}

static int print_tcnetinfo_all()
{
  int j;
  for (j = 0; j < NETINFO_MAX; j++)
  {
    print_tcnetinfo_row(j);
  }
  return 0;
}

static int get_empty_tcnetinfo_num()
{
  int j;
  int ret;
  
  ret = 0;
  for (j = 0; j < NETINFO_MAX; j++)
  {
    if (!strcmp(g_netinfo[j].val, ""))
    {
      ret++;
    }
  }
  return ret;
}

// return the available number of tcnet entries
static int load_tcnetinfo(const char *filename, const char *delimiter)
{
  int j;
  int n;
  int ret;
  int bytes;
  char buf[TMPBUFSZ];
  struct config_data *db;

  ret = 0;

  // check arguments
  if (!filename || !delimiter)
  {
    DEBUG((D_ERROR, L"get_key_val(): illegal arguments\n"));
    return -1;
  }
  
  if (('\0' == *filename) || ('\0' == *delimiter))
  {
    DEBUG((D_ERROR, L"get_key_val(): impropper arguments\n"));
    return -1;
  }
  
  // load file content to buffer
  memset(buf, 0, sizeof(buf));
  bytes = TcOneTimeNtReadFile(buf, sizeof(buf), filename);
  
  if (bytes <= 0)
    {
    DEBUG((D_ERROR, L"get_key_value(): bytes = %d\n", bytes));
    return -1;
  }
  else
  {
    DEBUG((D_INFO, L"get_key_value(): bytes = %d\n", bytes));
  }
  
  n = sget_key_val(buf, bytes, delimiter, &db);
  
  DEBUG((D_INFO, L"get_key_val(): got %d items\n", n));
  
  init_tcnetinfo();
  
  for (j = 0; j < n && db != NULL; j++, db = db->next)
  {
    if (db->key == NULL)
    {
      DEBUG((D_INFO, L"get_key_val(): empty items, j = %d\n", j));
      continue;
    }
    
    if (!strcmp(KEY_BOOTPROTO, db->key))
    {
      set_tcnetinfo(KEY_BOOTPROTO, db->value);
    }
        else if(!strcmp(KEY_HWADDR, db->key))
    {
      set_tcnetinfo(KEY_HWADDR, db->value);
    }
    else if(!strcmp(KEY_IPADDR, db->key))
    {
      set_tcnetinfo(KEY_IPADDR, db->value);
    }
    else if(!strcmp(KEY_NETMASK, db->key))
    {
      set_tcnetinfo(KEY_NETMASK, db->value);
    }
    else if(!strcmp(KEY_GATEWAY, db->key))
    {
      set_tcnetinfo(KEY_GATEWAY, db->value);
    }
    else if(!strcmp(KEY_DNS_PRIMARY, db->key))
    {
      set_tcnetinfo(KEY_DNS_PRIMARY, db->value);
    }
    else if(!strcmp(KEY_DNS_BACKUP, db->key))
    {
      set_tcnetinfo(KEY_DNS_BACKUP, db->value);
    }
  }
  
  return NETINFO_MAX - get_empty_tcnetinfo_num();
}

static void print_tcnetinfo_for_internal_test(int num)
{
  if (num < 0 || num > 6)
  {
    num = 6;
  }
  if (num > 0)
    PRINT(L"%hs = %hs\n", KEY_BOOTPROTO, "static");
  if (num > 1)
    PRINT(L"%hs = %hs\n", KEY_HWADDR, "74-D4-35-05-26-58");
  if (num > 2)
    PRINT(L"%hs = %hs\n", KEY_IPADDR, "192.168.0.100");
  if (num > 3)
    PRINT(L"%hs = %hs\n", KEY_NETMASK, "255.255.255.0");
  if (num > 4)
    PRINT(L"%hs = %hs\n", KEY_GATEWAY, "192.168.0.2");
  if (num > 5)
    PRINT(L"%hs = %hs\n", KEY_DNS_PRIMARY, "8.8.8.8");
  if (num > 6)
    PRINT(L"%hs = %hs\n", KEY_DNS_BACKUP, "8.8.4.4");
}

static void initlocale()
{
  _tsetlocale(LC_ALL, _TEXT(""));
}

static void usage(TCHAR prog[])
{
  PRINT(_TEXT("Usage: %s [OPTION]\n"), prog);
  PRINT(_TEXT("\n"));
  PRINT(_TEXT("[OPTION] can be one of the following\n"));
  PRINT(_TEXT("     --all              print all desirable network configurations for tc\n"));
  PRINT(_TEXT("-d,  --dns              print all doman name server configurations\n"));
  PRINT(_TEXT("     --dns-primary      print primary DNS\n"));
  PRINT(_TEXT("     --dns-alternative  print DNS alternative\n"));
  PRINT(_TEXT("-g,  --default-gateway  print ip address for default gateway\n"));
  PRINT(_TEXT("-h,  --help             print this help\n"));
  PRINT(_TEXT("-i,  --ipv4             print ipv4 address and mask\n"));
  PRINT(_TEXT("     --ipv4-address     print only ipv4 address\n"));
  PRINT(_TEXT("     --mac              print only physical address\n"));
  PRINT(_TEXT("-m,  --netmask          print only netmask\n"));
  PRINT(_TEXT("-v,  --version          print version of the application\n"));
  //PRINT(_TEXT("     --internal-test    emulator the output of option`--all' for testing\n"));
}

int _tmain(int argc, _TCHAR *argv[])
{
  int ret = 0;
  int flag_all = 0;
  int flag_dns = 0;
  int flag_dns_primary = 0;
  int flag_dns_alternative = 0;
  int flag_default_gateway = 0;
  int flag_help = 0;
  int flag_ipv4 = 0;
  int flag_ipv4_address = 0;
  int flag_mac = 0;
  int flag_netmask = 0;
  int flag_version = 0;
  int flag_internal_test = 0;
  
  int c;
  int option_index;  
  TCHAR short_options[] = _TEXT("dghimv");
  struct option long_options[] =
  {
    {_TEXT("all"), no_argument, &flag_all, 1},
    {_TEXT("dns"), no_argument, 0, _TEXT('d')},
    {_TEXT("dns-primary"), no_argument, &flag_dns_primary, 1},
    {_TEXT("dns-alternative"), no_argument, &flag_dns_alternative, 1},
    {_TEXT("default-gateway"), no_argument, 0, _TEXT('g')},
    {_TEXT("help"),  no_argument, 0, _TEXT('h')},
    {_TEXT("ipv4"),  no_argument, 0, _TEXT('i')},
    {_TEXT("ipv4-address"),  no_argument, &flag_ipv4_address, 1},
    {_TEXT("mac"), no_argument, &flag_mac, 1},
    {_TEXT("netmask"),  no_argument, 0, _TEXT('m')},
    {_TEXT("version"), no_argument, 0, _TEXT('v')},
    {_TEXT("internal-test"), no_argument, &flag_internal_test, 1},
    {NULL, 0, NULL, 0}
  };

  DEBUG_INIT(TO_STD_CONSOLE, NULL);
  PRINT_INIT(TO_STD_CONSOLE, NULL);

  initlocale();

  if (argc < 2)
  {
    usage(argv[0]);
    ret = 1;
    goto Exit;
  }

  while(1)
  {
    option_index = 0;
    c = getopt_long(argc, argv, short_options, long_options, &option_index);
    
    if (c == -1)
      break;

    switch (c)
        {
    case 0:
      if (long_options[option_index].flag != NULL)
      {
        DEBUG((D_INFO, _TEXT("--%s\n"), long_options[option_index].name));
      }
      break;

    case _TEXT('d'):
      DEBUG((D_INFO, _TEXT("option -d\n")));
      flag_dns = 1;
      break;

    case _TEXT('g'):
      DEBUG((D_INFO, _TEXT("option -g\n")));
      flag_default_gateway = 1;
      break;

        case _TEXT('h'):
      DEBUG((D_INFO, _TEXT("option -h\n")));
      flag_help = 1;
      break;

        case _TEXT('i'):
      DEBUG((D_INFO, _TEXT("option -i\n")));
      flag_ipv4 = 1;
      break;

        case _TEXT('m'):
      DEBUG((D_INFO, _TEXT("option -m\n")));
      flag_netmask = 1;
      break;

    case _TEXT('v'):
      DEBUG((D_INFO, _TEXT("option -v\n")));
      flag_version = 1;
      break;

    case _TEXT('?'):
      ret = 2;
      goto Exit;
      break;

    default:
      ret = 3;
      goto Exit;
        }
  }

  if (flag_help == 1){
    DEBUG((D_INFO, _TEXT("--help\n")));
    usage(argv[0]);
    ret = 0;
    goto Exit;
  }
  else if (flag_version == 1){
    DEBUG((D_INFO, _TEXT("--version\n")));
    PRINT(_TEXT("%s\n"), VERSION_STRING);
    ret = 0;
    goto Exit;
  }
  else if (flag_internal_test == 1){
    DEBUG((D_INFO, _TEXT("--internal-test\n")));
    print_tcnetinfo_for_internal_test(rand() % (1 + NETINFO_MAX));
    ret = 0;
    goto Exit;
  }

  // // load configurations
  // // at least the first 4 entries are essential
  // if (load_tcnetinfo(TC_NETFILE, TC_DELIMITER) < 4)
  // {
    // ERR(_TEXT("insufficient entryies\n"));
    // ret = 4;
    // goto Exit;
  // }
  
  //  
  // check the result, just proceed the load process
  //
  if (load_tcnetinfo(TC_NETFILE, TC_DELIMITER) <= 0)
  {
    DEBUG((D_ERROR, _TEXT("fail to load config file\n")));
    ret = 1;
    goto Exit;
  }
  if (flag_all == 1){
    DEBUG((D_INFO, _TEXT("--all\n")));
    print_tcnetinfo_all();
    ret = 0;
    goto Exit;
  }
  else if (flag_dns == 1){
    DEBUG((D_INFO, _TEXT("--dns\n")));
    print_tcnetinfo_row(get_key_index(KEY_DNS_PRIMARY));
    print_tcnetinfo_row(get_key_index(KEY_DNS_BACKUP));
    ret = 0;
    goto Exit;
  }
  else if (flag_dns_primary == 1){
    DEBUG((D_INFO, _TEXT("--dns-primary\n")));
    ret = print_tcnetinfo(KEY_DNS_PRIMARY);
    goto Exit;
  }
  else if (flag_dns_alternative == 1){
    DEBUG((D_INFO, _TEXT("--dns-alternative\n")));
    ret = print_tcnetinfo(KEY_DNS_BACKUP);
    goto Exit;
  }
  else if (flag_default_gateway == 1){
    DEBUG((D_INFO, _TEXT("--default-gateway\n")));
    ret = print_tcnetinfo(KEY_GATEWAY);
    goto Exit;
  }
  else if (flag_ipv4 == 1){
    DEBUG((D_INFO, _TEXT("--ipv4\n")));
    print_tcnetinfo_row(get_key_index(KEY_IPADDR));
    print_tcnetinfo_row(get_key_index(KEY_NETMASK));
    ret = 0;
    goto Exit;
  }
  else if (flag_ipv4_address == 1){
    DEBUG((D_INFO, _TEXT("--ipv4-address\n")));
    ret = print_tcnetinfo(KEY_IPADDR);
    goto Exit;
  }
  else if (flag_mac == 1){
    DEBUG((D_INFO, _TEXT("--mac\n")));
    ret = print_tcnetinfo(KEY_HWADDR);
    goto Exit;
  }
  else if (flag_netmask == 1){
    DEBUG((D_INFO, _TEXT("--netmask\n")));
    ret = print_tcnetinfo(KEY_NETMASK);
    goto Exit;
  }
  else{
    DEBUG((D_INFO, _TEXT("no options were found\n")));
    ret = 6;
    goto Exit;
  }

Exit:
  DEBUG_FINISH();
  PRINT_FINISH();
  
  return ret;
}
