/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file
  Get the username for the logged on TC user and print it to stdout.

  RETURN
  On success, return 0;
  otherwise, return a non-zero value.
**/
#define _CRT_SECURE_NO_WARNINGS

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "TcConfig.h"
#include "TcDebugLib.h"
#include "TcNtLib.h"
#include "TcKeyValueLib.h"

#define TMPBUFSZ          1024
#define TC_USERNAME_KEY   "username"
#define TC_KEY_DELIMITER  "="
#define TC_PASSFILE       "\\Device\\Harddisk0\\Partition1\\.tc.passwd"

/**
  get key value from a given file and store it to dest.
  
  @param dest
  @param filename
  @param keystr
  @param dest delimiter
  @return on success, return a pointer pointing to the result;
          otherwise, return NULL.
**/
static char* get_key_value(char **dest, const char *filename, const char *keystr, const char *delimiter)
{
  int j;
  int n;
  int ret;
  int bytes;
  char buf[TMPBUFSZ];
  struct config_data *db;

  ret = -1;

  //
  // check arguments
  //
  if (!dest || !filename || !keystr || !delimiter)
  {
    DEBUG((D_ERROR, L"get_key_val(): illegal arguments\n"));
    return NULL;
  }
  
  *dest = NULL;
  
  if (('\0' == *filename) || ('\0' == *keystr) || ('\0' == *delimiter))
  {
    DEBUG((D_ERROR, L"get_key_val(): impropper arguments\n"));
    return *dest;
  }
  
  // load file content to buffer
  memset(buf, 0, sizeof(buf));
  bytes = TcOneTimeNtReadFile(buf, sizeof(buf), filename);
  
  if (bytes <= 0)
    {
    DEBUG((D_ERROR, L"get_key_value(): bytes = %d\n", bytes));
    return *dest;
  }
  else
  {
    DEBUG((D_INFO, L"get_key_value(): bytes = %d\n", bytes));
  }
  
  n = sget_key_val(buf, bytes, delimiter, &db);
  
  DEBUG((D_INFO, L"get_key_val(): got %d items\n", n));
  
  for (j = 0; j < n && db != NULL; j++, db = db->next)
  {
    if (db->key == NULL)
    {
      DEBUG((D_INFO, L"get_key_val(): empty items, j = %d\n", j));
      continue;
    }
    
    if (!strcmp(keystr, db->key))
    {
      DEBUG((D_INFO, L"get_key_val(): find key: %hs\n", db->key));
      if (db->value == NULL)
      {
        DEBUG((D_ERROR, L"get_key_val(): found empty value for key \"%hs\"\n", db->key));
        return *dest;
      }
      else
      {
        DEBUG((D_INFO, L"get_key_val(): find key value for key [%hs, %hs]\n", db->key, db->value));
        *dest = (char *)malloc(strlen(db->value) + 1);
        if (*dest == NULL)
        {
          DEBUG((D_ERROR, L"get_key_val(): failed to alloc memory\n"));
          DEBUG((D_ERROR, L"get_key_val(): strlen = %d\n", strlen(db->value)));
          return *dest;
        }
        else
        {
          strcpy(*dest, db->value);
        }
      }
      break;
    }
  }
  
  if (*dest == NULL)
  {
    DEBUG((D_ERROR, L"get_key_val(): cannot find (key,value) pair with key = \"%hs\" \n", keystr));
  }
  
  return *dest;
}

int
__cdecl
main (
  int argc,
  char *argv[]
  )
{
  int ret;
  char *res;
  
  DEBUG_INIT(TO_STD_CONSOLE, NULL);
  ret = 1;
  get_key_value(&res, TC_PASSFILE, TC_USERNAME_KEY, TC_KEY_DELIMITER);
  
  // print to stdout
    if (res != NULL)
  {
    printf("%s", res);

    // clean up
    free(res);
    res = NULL;
    
    ret = 0;
  }
  else
  {
    DEBUG((D_ERROR, L"main(): get_key_val() return with NULL\n"));
  }

  DEBUG_FINISH();
  return ret;
}
