/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
#include <stdlib.h>
#include <malloc.h>
#include <stdio.h>

#include "TcGetoptLib.h"

#define OPT_ORDERING_REQUIRE_ORDER   0
#define OPT_ORDERING_PERMUTE         1
#define OPT_ORDERING_RETURN_IN_ORDER 2

int g_parseopt_ind = 1;
int g_parseopt_err = 1;
int g_parseopt_opt = '?';


static struct parseopt_data_struc_wide
{
  int     next_index;
  int     error_msg_flag;
  int     optsymbol;
  int     init_data_ready;
  int     arrangement;
  int     nonopt_head;
  int     nonopt_tail;
  wchar_t *optval;
  wchar_t *next_symbol;
};
wchar_t *g_parseopt_arg_wide;

/**
* wrapper for internal algorithm implementation
*/
static
int
parseopt_wrapper_wide(
            wchar_t *const                       *listw,
            int                                  listcnt,
            const wchar_t                        *shortopts,
            const struct parse_option_struc_wide *longopts,
            int                                  *longopt_index,
            int                                  longonly_mode
            );


/**
*  -a
*  -a arg
*/
int PARSEOPT_STDCALL
parseopt_wide
(
 int                                  listcnt,
 wchar_t *const                       *listw,
 const wchar_t                        *shortopts
 ) PARSEOPT_THROW
{
  return parseopt_wrapper_wide(
    listw,
    listcnt,
    shortopts,
    NULL,
    NULL,
    0);
}

/**
*  -a
*  -a arg
* --long_option
* --long_option xxx
* --long_option=xxx
*/
int PARSEOPT_STDCALL
parseopt_long_wide
(
 int                                  listcnt,
 wchar_t *const                       *listw,
 const wchar_t                        *shortopts,
 const struct parse_option_struc_wide *longopts,
 int                                  *longopt_index
 ) PARSEOPT_THROW
{
  return parseopt_wrapper_wide(
    listw,
    listcnt,
    shortopts,
    longopts,
    longopt_index,
    0);
}

/**
*  -long_option
*  -long_option xxx
*  -long_option=xxx
* --long_option
* --long_option xxx
* --long_option=xxx
*/
int PARSEOPT_STDCALL
parseopt_long_only_wide
(
 int                                  listcnt,
 wchar_t *const                       *listw,
 const wchar_t                        *shortopts,
 const struct parse_option_struc_wide *longopts,
 int                                  *longopt_index
 ) PARSEOPT_THROW
{
  return parseopt_wrapper_wide(
    listw,
    listcnt,
    shortopts,
    longopts,
    longopt_index,
    1);
}


/**
* swap the i-th and j-th elements for array 'a'
*/
static void swap_elements(void *a[], int i, int j)
{
  void *middleman;

  middleman = a[i];
  a[i]      = a[j];
  a[j]      = middleman;
}

/**
* exchange the elements before index middle, the front elements,
* to tail of the array
* keep the front elements in original order,
* and keep rest elements in original order as well
*/
static void rearrange(void *data[], int first, int middle, int last)
{
  int j;
  int gap;

#define MINI(x,y) ((x) < (y) ? (x) : (y))
  gap = MINI(middle - first, last - middle);
  while (gap > 0)
  {
    if (gap == middle - first)
    {
      last -= gap;
      for (j = 0; j < gap; j++)
      {
        swap_elements(data, first + j, last + j);
      }
    }
    else
    {
      for (j = 0; j < gap; j++)
      {
        swap_elements(data, first + j, middle + j);
      }
      first += gap;
    }
    gap = MINI(middle - first, last - middle);
  }
}

/**
* initialize data structure
*/
static const wchar_t *init_parseopt_data_wide(const wchar_t *shortopts, struct parseopt_data_struc_wide *q)
{
  q->next_symbol = NULL;
  q->nonopt_head = q->next_index;
  q->nonopt_tail = q->next_index;

  if (L'-' == shortopts[0])
  {
    shortopts++;
    q->arrangement = OPT_ORDERING_RETURN_IN_ORDER;
  }
  else if (L'+' == shortopts[0])
  {
    shortopts++;
    q->arrangement = OPT_ORDERING_REQUIRE_ORDER;
  }
  else
  {
    q->arrangement = OPT_ORDERING_PERMUTE;
  }

  return shortopts;
}


static int
parseopt_impl_wide
(
 wchar_t *const                       *listw,
 int                                  listcnt,
 const wchar_t                        *shortopts,
 const struct parse_option_struc_wide *longopts,
 int                                  *longind,
 int                                  onlylong_mode,
 struct parseopt_data_struc_wide      *pd
  )
{
  //
  // define a flag,
  // to indicate whether error messages could be printed.
  // 0 indicate to print error messages,
  // 1 indicates not to print error messages.
  //
  // note: this flag work with the short shortopts to determine the error message capability
  // see below
  //
  int prnerr;

  int strictly_matched;
  int longindex_found;
  int temp_longopt_index;
  size_t namelength;
  wchar_t *nametail;
  struct temp_optionlist
  {
    const struct parse_option_struc_wide *option;
    struct temp_optionlist *successor;
  } *possible_oplist;
  const struct parse_option_struc_wide *q;
  const struct parse_option_struc_wide *found_longnode;

  wchar_t wch;
  wchar_t *pwch;

  pd->optval = NULL;

  //
  // if no arguments, just return as if scaning was done
  //
  if (listcnt < 1)
  {
    return -1;
  }

  //
  // initialize the data structure
  //
  if (pd->next_index == 0 || !pd->init_data_ready)
  {
    if (pd->next_index == 0)
    {
      pd->next_index = 1;
    }
    shortopts = init_parseopt_data_wide(shortopts, pd);
    //
    // static structure was initialized to all '0' by default
    // 0 indicates the data stucture is uninitialized
    // 1 indicates the data stucture is initialized
    //
    pd->init_data_ready = 1;
  }

  //
  // skip the possible leading '+' or '-' characters of short option string
  //
  else if (shortopts[0] == L'-' || shortopts[0] == L'+')
  {
    shortopts++;
  }

  //
  // if leading character (optionally followed by possible leading '+' or '-' sign) 
  // of short option string is ':',  then error messages will not be printed
  // 
  // otherwiese, the error message capability will depend on value of prnerr set previously
  //
  prnerr = pd->error_msg_flag;
  if (shortopts[0] == L':')
  {
    prnerr = 0;
  }

  //
  // if it is the first time to scan for the short option string
  // or if the short option string is scaned completely
  //
  if (pd->next_symbol == NULL || *pd->next_symbol == L'\0')
  {
    if (pd->nonopt_tail > pd->next_index)
    {
      pd->nonopt_tail = pd->next_index;
    }

    if (pd->nonopt_head > pd->next_index)
    {
      pd->nonopt_head = pd->next_index;
    }

    //
    // premute all the argument list, including option and non-option arguments
    //
    if (pd->arrangement == OPT_ORDERING_PERMUTE)
    {
      //
      // if arugment list is not well-formed
      // place all non-option arguments to tail of the argument list
      //
      if (pd->nonopt_head != pd->nonopt_tail && pd->nonopt_tail != pd->next_index)
      {
        rearrange((void **)listw, pd->nonopt_head, pd->nonopt_tail, pd->next_index);
        pd->nonopt_head += pd->next_index - pd->nonopt_tail;
        pd->nonopt_tail = pd->next_index;
      }

      //
      // otherwise, check if there exists no non-option argument,
      // let the first nonopt index point to next_index
      //
      else if (pd->nonopt_tail != pd->next_index) // pd->nonopt_head == pd->nonopt_tail
      {
        // pd->nonopt_head == pd->nonopt_tail = pd->next_index;
        pd->nonopt_head = pd->next_index;
      }

      //
      // scan the argument list
      // if the argument is a non-option argument,
      // or if the argument is exactly a '-' sign,
      // just move the next_index forward by 1.
      //
      while (pd->next_index < listcnt && (listw[pd->next_index][0] != L'-' || listw[pd->next_index][1] == L'\0'))
      {
        pd->next_index++;
      }

      //
      // keep the last nonopt index to the value of next_index
      //
      pd->nonopt_tail = pd->next_index;
    }

    //
    // if the argument is exactly "--",
    // then treate all remaining arguments as non-option arguments
    //
    if (pd->next_index != listcnt && !wcscmp(listw[pd->next_index], L"--"))
    {
      pd->next_index++;
      if (pd->nonopt_head != pd->nonopt_tail && pd->nonopt_tail != pd->next_index)
      {
        rearrange((void **)listw, pd->nonopt_head, pd->nonopt_tail, pd->next_index);
        pd->nonopt_head += pd->next_index - pd->nonopt_tail;
        pd->nonopt_tail = pd->next_index;
      }
      else if (pd->nonopt_head == pd->nonopt_tail)
      {
        pd->nonopt_head = pd->next_index;
      }

      pd->nonopt_tail = listcnt;
      pd->next_index = listcnt;
    }

    //
    // if no more aguments, just return    
    //
    if (pd->next_index == listcnt)
    {
      // if there exist non-option argument(s),
      // next_index point to the first non-option argument
      if (pd->nonopt_head != pd->nonopt_tail)
      {
        pd->next_index = pd->nonopt_head;
      }
      return -1;
    }

    //
    // if listw[i] is a non-option,
    // or if listw[i] is '-',
    //
    if ((listw[pd->next_index][0] != L'-' || listw[pd->next_index][1] == L'\0'))
    {
      if (pd->arrangement == OPT_ORDERING_REQUIRE_ORDER)
      {
        return -1;
      }

      pd->optval = listw[pd->next_index];
      pd->next_index++;
      return 1;
    }

    //
    // for short option argument, for example, -f, skip 1 character
    // for long option arugment, for example, --file, skip 2 characters
    //
    pd->next_symbol = (listw[pd->next_index] + 1 + (longopts != NULL && listw[pd->next_index][1] == L'-'));
  }

  //
  // if longopts is provided,
  //
  //
  // if longopts is provided,
  if (longopts != NULL && 
    ( // --xxx, long option
    listw[pd->next_index][1] == L'-' || 
    // specified only long options are supported
    (  onlylong_mode && 
    ( // is not exactly "--"
    // "abcdefg...", 'b' can be found in shortopts
    listw[pd->next_index][2] || !wcschr(shortopts, listw[pd->next_index][1])
    )
    )
    )
    )
  {
    strictly_matched = 0;
    longindex_found = -1;
    possible_oplist = NULL;
    found_longnode = NULL;

    //
    // continue scanning the input argument list
    // -f, then name is "f"
    // --file, then name is "file"
    // --file=, then name is "file"
    //
    for (nametail = pd->next_symbol; *nametail && *nametail != L'='; nametail++)
    {
      ;
    }
    namelength = (size_t)(nametail - pd->next_symbol);

    //
    // traverse the long option struct array
    // to find which option elements are matched
    // if mutiple elements are matched, then add them to the ambigous linked-list
    // if there is an strictly_matched match, break the loop
    //
    for (q = longopts, temp_longopt_index = 0; q->name; q++, temp_longopt_index++)
    {
      //
      // if the long option name matches
      // namelength is the xxx length from input argument (--xxx)
      //
      if (!wcsncmp(q->name, pd->next_symbol, namelength))
      {
        //
        // user input option length matches with the program internal option name length
        //
        if (namelength == (size_t)wcslen(q->name))
        {
          //
          // this is exactly the match option
          // end loop
          //
          found_longnode = q;
          longindex_found = temp_longopt_index;
          strictly_matched = 1;
          break;
        }
        //
        // user input long option name length is less than that of a name in the option name list
        // found_longnode point to the 1st matched one w/ possible ambiguity
        //
        else if (found_longnode == NULL)
        {
          found_longnode = q;
          longindex_found = temp_longopt_index;
        }
        //
        // this user input option name might be just a prefix of
        // at least two names in the long option name list
        //
        // why onlylong_mode?
        // the struct long option has 4 arguments, if any of the arugment is different from the 1st one matched,
        // insert it to ambigous linked-list front
        //
        else if (onlylong_mode || found_longnode->has_arg != q->has_arg || found_longnode->flag != q->flag || found_longnode->val != q->val)
        {
          // to crate a linked-list node and insert it to the front of the ambigous linked-list
          struct temp_optionlist *a_list_node = (struct temp_optionlist*)alloca(sizeof(*a_list_node));
          a_list_node->option = q;
          a_list_node->successor = possible_oplist;
          possible_oplist = a_list_node;
        }
      }
    }

    //
    // if the ambigous linked-list exists
    // and there is no one fully matched
    // return '?'
    //
    if (possible_oplist != NULL && !strictly_matched)
    {
      if (prnerr)
      {
        struct temp_optionlist first_longnode_found;
        first_longnode_found.option = found_longnode;
        first_longnode_found.successor = possible_oplist;
        possible_oplist = &first_longnode_found;
        fwprintf(stderr, L"%s: possible answers for option '%s':", listw[0], listw[pd->next_index]);
        do
        {
          // alloca() will allocate temporary space on caller stack frame and freed when caller return
          fwprintf(stderr, L" '--%s'", possible_oplist->option->name);
          possible_oplist = possible_oplist->successor;
        } while (possible_oplist != NULL);
        fwprintf(stderr, L"\n");
      }
      // skip the nextchar pointer by the current scanned name (user input) length
      pd->next_symbol += wcslen(pd->next_symbol);
      pd->next_index++;
      pd->optsymbol = 0;
      return L'?';
    }

    //
    // if at least one match, and in following two cases the ambig could be NULL or not NULL.
    // 1) w/o ambiguity, strictly_matched == 1
    // 2) w/ ambiguity, strictly_matched == 0
    //
    if (found_longnode != NULL)
    {
      temp_longopt_index = longindex_found;
      pd->next_index++;

      // if nametail is a euqal sign '='
      if (*nametail)
      {
        // if required or optional augument
        // for example, --file=FILE
        // optarg point to F
        if (found_longnode->has_arg)
        {
          pd->optval = nametail + 1;
        }
        else // no argument
        {
          if (prnerr)
          {
            // long options, --file==FILE
            if (listw[pd->next_index - 1][1] == L'-')
            {
              fwprintf(stderr, L"%s: option '--%s' does not parse any arguments\n", listw[0], found_longnode->name);
            }
            else // only long options, -file=FILE
            {
              fwprintf(stderr, L"%s: option '%c%s' does not parse any arguments\n", listw[0], listw[pd->next_index - 1][0], found_longnode->name);
            }
          }
          pd->next_symbol += wcslen(pd->next_symbol);
          pd->optsymbol = found_longnode->val;
          return L'?';
        }
      }
      else if (found_longnode->has_arg == 1)
      {
        if (pd->next_index < listcnt)
        {
          pd->optval = listw[pd->next_index++];
        }
        else
        {
          if (prnerr)
          {
            fwprintf(stderr, L"%s: missing argument for option '--%s'\n", listw[0], found_longnode->name);
          }
          pd->next_symbol += wcslen(pd->next_symbol);
          pd->optsymbol = found_longnode->val;
          return shortopts[0] == L':' ? L':' : L'?';
        }
      }

      pd->next_symbol += wcslen(pd->next_symbol);
      if (longind != NULL)
      {
        *longind = temp_longopt_index;
      }

      if (found_longnode->flag)
      {
        *(found_longnode->flag) = found_longnode->val;
        return 0;
      }
      return found_longnode->val;
    }

    if (!onlylong_mode || listw[pd->next_index][1] == L'-' || wcschr(shortopts, *pd->next_symbol) == NULL)
    {
      if (prnerr)
      {
        if (listw[pd->next_index][1] == L'-')
        {
          fwprintf(stderr, L"%s: unknown option '--%s'\n", listw[0], pd->next_symbol);
        }
        else
        {
          fwprintf(stderr, L"%s: unknown option '%c%s'\n", listw[0], listw[pd->next_index][0], pd->next_symbol);
        }
      }
      pd->next_symbol = (wchar_t *)L"";
      pd->next_index++;
      pd->optsymbol = 0;
      return L'?';
    }
  }

  wch = *pd->next_symbol++;
  pwch = (wchar_t*)wcschr(shortopts, wch);

  if (*pd->next_symbol == L'\0')
  {
    ++pd->next_index;
  }

  if (pwch == NULL || wch == L':' || wch == L';')
  {
    if (prnerr)
    {
      fwprintf(stderr, L"%s: unknown option -- '%c'\n", listw[0], wch);
    }

    pd->optsymbol = wch;
    return L'?';
  }

  if (pwch[1] == L':')
  {
    if (pwch[2] == L':')
    {
      if (*pd->next_symbol != L'\0')
      {
        pd->optval = pd->next_symbol;
        pd->next_index++;
      }
      else
      {
        pd->optval = NULL;
      }
      pd->next_symbol = NULL;
    }
    else
    {
      if (*pd->next_symbol != L'\0')
      {
        pd->optval = pd->next_symbol;
        pd->next_index++;
      }
      else if (pd->next_index == listcnt)
      {
        if (prnerr)
        {
          fwprintf(stderr, L"%s: missing argument for option -- '%c'\n", listw[0], wch);
        }
        pd->optsymbol = wch;
        if (shortopts[0] == L':')
        {
          wch = L':';
        }
        else
        {
          wch = L'?';
        }
      }
      else
      {
        pd->optval = listw[pd->next_index++];
      }
      pd->next_symbol = NULL;
    }
  }
  return wch;
}

/**
* wrapper for internal algorithm implementation
*/
static
int
parseopt_wrapper_wide(
            wchar_t *const                       *listw,
            int                                  listcnt,
            const wchar_t                        *shortopts,
            const struct parse_option_struc_wide *longopts,
            int                                  *longopt_index,
            int                                  longonly_mode
            )
{
  static struct parseopt_data_struc_wide parseopt_data_wide = {0};
  int ret;

  parseopt_data_wide.next_index     = g_parseopt_ind;
  parseopt_data_wide.error_msg_flag = g_parseopt_err;

  ret = parseopt_impl_wide(
    listw,
    listcnt,
    shortopts,
    longopts,
    longopt_index,
    longonly_mode,
    &parseopt_data_wide);

  g_parseopt_ind      = parseopt_data_wide.next_index;
  g_parseopt_arg_wide = parseopt_data_wide.optval;
  g_parseopt_opt      = parseopt_data_wide.optsymbol;

  return ret;
}