/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file

**/
#define _CRT_SECURE_NO_WARNINGS
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <wchar.h>
#include <string.h>
#include <assert.h>

#include "TcKeyValueLib.h"

// if src is NULL or is an empty string
// just return itself;
// otherwise, trim it from its head
char*
TC_STDCALL
trim_head(char *src)
{
    char *p = src;

    if (p == NULL || *p == '\0')
    {
        return p;
    }

	size_t len = strlen(p);
    while(*p && isspace(*p))
    {
        ++p;
        --len;
    }
    return memmove(src, p, len + 1);
}

// if src is NULL or is an empty string
// just return itself;
// otherwise, trim it from its tail
char*
TC_STDCALL
trim_tail(char *src)
{
    char *p = src;

    if (p == NULL || *p == '\0')
    {
        return p;
    }

	size_t len = strlen(p);
    while(isspace(p[len - 1]))
    {
        len--;
        p[len] = 0;
    }
    return memmove(src, p, len + 1);
}

// if src is NULL or is an empty string
// just return itself;
// otherwise, trim it from both its tail and its tail
char*
TC_STDCALL
trim(char *src)
{
    char *p = src;
    
    if (p == NULL || *p == '\0')
    {
        return p;
    }

	size_t len = strlen(p);
    while(isspace(p[len - 1]))
    {
        len--;
        p[len] = 0;
    }
    while(*p && isspace(*p))
    {
        ++p;
        --len;
    }

    return memmove(src, p, len + 1);
}

#ifdef _WIN32
static char* strsep(char **stringp, const char *delim)
{
    char *start;
    char *p;

    start = *stringp;
    p = start ? strpbrk(start, delim) : NULL;

    if (!p) {
        *stringp = NULL;
    }
    else
    {
        *p = 0;
        *stringp = p + 1;
    }
    
    return start;
}
#endif //_WIN32

// divide str with first delim and store left part to key, right part to value
// the key will be trimed, and a key contains only spaces is illegal 
// the value will also be trimed, and a value consists of only spaces is allowed and will be trimed to an empty string
// e.g., let '=' be the delim, legal input:
// key = value
// key = 
// illegal input:
//     = 
//     = value
// on success, return 0
// otherwise, return -1
int
TC_STDCALL
get_a_key_val(char *key, int keysz, char *value, int valuesz, const char *str, const char *delim)
{
    char *p;
    char *q;
    char *r;
    size_t plen;
    size_t qlen;

    if (!key || !value || !str || !delim)
    {
        return -1;
    }

    q = (char *) calloc(strlen(str) + 1, sizeof(char));
    assert(q != NULL);
	// if not find delim
	if (q == NULL)
	{
		return -1;
	}
    // search and replace delimeters with '\0'
    r = q;
    strcpy(q, str);
    p = strsep(&q, delim);
    
    
    // if find delim but the left side is all spaces
    trim(p);
    if (*p == '\0')
    {
		free(r);
        return -1;
    }

    // trim the value
    trim(q);

    // buffer size matters
    plen = strlen(p);
    qlen = strlen(q);
    if (plen + qlen > (size_t)valuesz)
    {
		free(r);
        return -1;
    }

    // store it into list
    strcpy(key, p);
    strcpy(value, q);
    
    if (r != NULL)
    {
        free(r);
        r = NULL;
    }
    
    return 0;
}

// load configs from file, save legal key-val pairs,
// and return the number legal options
// if pconf is not NULL, the all key-val pairs would be saved there.
// if find key-val pairs, return the number of legal key-val pairs, > 0
// if not find key-val pairs, return = 0
// if errors, return < 0
//
int
TC_STDCALL
get_key_val(const char *filename, const char *delim, struct config_data **pconf)
{
    int n;
    int t;
    size_t len;
    char *p;
    char buf[CONF_LINEBUF_SZ];
    char key[CONF_KEYBUF_SZ];
    char val[CONF_VALBUF_SZ];
    FILE *fp;
    struct config_data *phead, *ptail, *ptmp;

    fp = fopen(filename, "r");
    if (fp == NULL)
    {
        return -1;
    }

    if (pconf != NULL)
    {
        phead = NULL;
    }
    n = 0;
    while (NULL != fgets(buf, sizeof(buf), fp))
    {
        p = buf;
        len = strlen(p);

        // remove CR-LF, if any
        while( (len > 0) && (p[len - 1] == '\r' || p[len - 1] == '\n') )
        {
            len--;
            p[len] = '\0';
        }

        // get key and val from the string
        t = get_a_key_val(key, sizeof(key), val, sizeof(val), p, delim);
        if (0 == t)
        {
            n++;

            // add a memeber to the linked list
            if (pconf != NULL)
            {
                ptmp = (struct config_data *) calloc(1, sizeof(struct config_data));
                assert(ptmp != NULL);

				if (ptmp == NULL)
				{
					break;
				}
                strcpy(ptmp->key, key);
                strcpy(ptmp->value, val);
                ptmp->next = NULL;

                if(phead == NULL)
                {
                    phead = ptmp;
                    ptail = phead;
                }
                else
                {
                    ptail->next = ptmp;
                    ptail = ptail->next;
                }
            }
        }
    }

    if (fp != NULL)
    {
        fclose(fp);
        fp = NULL;
    }

    if (pconf != NULL)
    {
        *pconf = phead;
    }

    return n;
}

// load configs from a string, save legal key-val pairs,
// and return the number legal options
// Note: the string may contain one or mutiple CR-LF('\r''\n') or LF('\n').
// And make sure each key-val pair in a line.
// if pconf is not NULL, the all key-val pairs would be saved there.
// if find key-val pairs, return the number of legal key-val pairs, > 0
// if not find key-val pairs, return = 0
// if errors, return < 0
//
int
TC_STDCALL
sget_key_val(const char *input, int inputsz, const char *delim, struct config_data **pconf)
{
    int n;
    int t;
    int len;
    int done;
    int donesz;
    char line[CONF_LINEBUF_SZ];
    const char *p, *q, *r;
    char key[CONF_KEYBUF_SZ];
    char val[CONF_VALBUF_SZ];
    struct config_data *phead, *ptail, *ptmp;
    
    const char TEMINATOR = '\n';
    
    if (input == NULL || inputsz <= 0 || delim == NULL)
    {
        return -1;
    }
    
    if (pconf != NULL)
    {
        phead = NULL;
    }
    
    n = 0;
    done = 0;
    donesz = 0;
    r = input;
    while (!done)
    {
        q = r;
        p = r;
        
        while (donesz < inputsz && *q != TEMINATOR)
        {
            q++;
            donesz++;
        }
        
        // for cases the line being scanned is without any '\n'
        if (donesz == inputsz)
        {
            // the input string is scanned fully over
            q--;
            done = 1;
        }
        else if ( *q == TEMINATOR )
        {
            // for case at the last line 
            if (donesz + 1 == inputsz)
            {
                done = 1;
            }
            else
            {
                r = q + 1;
            }
        }
        
        len = (int) (q - p + 1);
        memmove(line, p, len);
        line[len] = '\0';

        // remove CR-LF, if any
        while( (len > 0) && (line[len - 1] == '\r' || line[len - 1] == '\n') )
        {
            len--;
            line[len] = '\0';
        }

        // get key and val from the string
        t = get_a_key_val(key, sizeof(key), val, sizeof(val), line, delim);
        if (0 == t)
        {
            n++;

            // add a memeber to the linked list
            if (pconf != NULL)
            {
                ptmp = (struct config_data *) calloc(1, sizeof(struct config_data));
                assert(ptmp != NULL);
				if (ptmp == NULL)
				{
					break;
				}
                strcpy(ptmp->key, key);
                strcpy(ptmp->value, val);
                ptmp->next = NULL;

                if(phead == NULL)
                {
                    phead = ptmp;
                    ptail = phead;
                }
                else
                {
                    ptail->next = ptmp;
                    ptail = ptail->next;
                }
            }
        }
    }

    if (pconf != NULL)
    {
        *pconf = phead;
    }

    return n;
}


void
TC_STDCALL
free_key_val(struct config_data *pconf)
{
    struct config_data *p, *q;

    p = q = pconf;
    while (q != NULL)
    {
        p = q -> next;
        free(q);
        q = p;
    }
}

void
TC_STDCALL
print_key_val(struct config_data *pconf)
{
    int j;
    struct config_data *p;

    j = 0;
    for (p = pconf; p != NULL; p = p->next)
    {
        wprintf(L"%02d\n", j + 1);
        wprintf(L"    \"%hs\"\n", p->key);
        wprintf(L"    %d\n", strlen(p->key));
        wprintf(L"    \"%hs\"\n", p->value);
        wprintf(L"    %d\n", strlen(p->value));
        j++;
    }
}

#ifdef __cplusplus
extern "C"
#endif
static
int
TC_STDCALL
test_get_a_key_val(const char *text, const char *delim)
{
    int t;
    char key[CONF_KEYBUF_SZ];
    char value[CONF_VALBUF_SZ];

    t = get_a_key_val(key, sizeof(key), value, sizeof(value), text, delim);
    if (t != 0)
    {
        return -1;
    }

    wprintf(L"text:\n");
    wprintf(L"    \"%hs\"\n", text);
    wprintf(L"    %d\n", strlen(text));
    wprintf(L"key:\n");
    wprintf(L"    \"%hs\"\n", key);
    wprintf(L"    %d\n", strlen(key));
    wprintf(L"value:\n");
    wprintf(L"    \"%hs\"\n", value);
    wprintf(L"    %d\n", strlen(value));
    wprintf(L"\n");

    return 0;
}

// a simple test for get option interface
#if defined(__cplusplus)
extern "C"
#endif
static
void
TC_STDCALL
test_get_key_val_from_file(const char *filename, const char *delim)
{
    int n;
    struct config_data *pconf;

    pconf = NULL;
    n = get_key_val(filename, delim, &pconf);
    if (n <= 0)
    {
        return;
    }

    wprintf(L"n = %d\n", n);
    print_key_val(pconf);

    if (pconf != NULL)
    {
        free_key_val(pconf);
        pconf = NULL;
    }
}

#if defined(__cplusplus)
extern "C"
#endif
static
void
TC_STDCALL
test_get_key_val_sample_file(const char *filename, const char *delim)
{
    FILE *fp;

    if (NULL == (fp = fopen(filename, "w")))
    {
        fwprintf(stderr, L"fail to open file \"%hs\"\n", filename);
        exit(-1);
    }

    fwprintf(fp, L"\n");
    fwprintf(fp, L"\r\n");
    fwprintf(fp, L"key1\n");
    fwprintf(fp, L" key2\n");
    fwprintf(fp, L" key3 \n");
    fwprintf(fp, L" key 4 \n");
    fwprintf(fp, L"%hs\n", delim);
    fwprintf(fp, L" %hs\n", delim);
    fwprintf(fp, L"%hs \n", delim);
    fwprintf(fp, L" %hs \n", delim);
    fwprintf(fp, L"key5 %hs \n", delim);
    fwprintf(fp, L" key5 %hs \n", delim);
    fwprintf(fp, L" key 5 %hs \n", delim);
    fwprintf(fp, L" %hsvalue6\n", delim);
    fwprintf(fp, L" %hs value6\n", delim);
    fwprintf(fp, L" %hs value6 \n", delim);
    fwprintf(fp, L" %hs value 6 \n", delim);
    fwprintf(fp, L"key7%hsvalue7\n", delim);
    fwprintf(fp, L" key8%hsvalue8\n", delim);
    fwprintf(fp, L" key9 %hsvalue9\n", delim);
    fwprintf(fp, L" key 10 %hsvalue10\n", delim);
    fwprintf(fp, L" key 11 %hs value11\n", delim);
    fwprintf(fp, L" key 12 %hs value12 \n", delim);
    fwprintf(fp, L" key 13 %hs value 13 \n", delim);
    fwprintf(fp, L" key 14 %hs value 14 \r\n", delim);
    fwprintf(fp, L" key 15 %hs value 15 ", delim);
    fwprintf(fp, L" key 16 %hs value 16 ", delim);
    fflush(fp);

    if (fp != NULL)
    {
        fclose(fp);
        fp = NULL;
    }
}

void
TC_STDCALL
test_get_key_val()
{
    const char buf[] = " adks al - = sl  ";
    const char *filename = "get_key_val_sample_file.conf";
    const char *delim = "=";

    test_get_a_key_val(buf, "=");

    test_get_key_val_sample_file(filename, delim);
    test_get_key_val_from_file(filename, delim);
}
