/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file

**/
#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <tchar.h>
#include <windows.h>
#include <Wincrypt.h>
#include <locale.h>

#include "TcConfig.h"
#include "TcDebugLib.h"
#include "TcGetoptLib.h"

//
// version string
//
#define MAJOR_VERSION 0x1
#define MINOR_VERSION 0x0
#define MICRO_VERSION 0x0

#define VERSION_NUMBER (SHORT ( ((MAJOR_VERSION & 0xF) << 8) | ((MINOR_VERSION & 0xF) << 4) | ((MICRO_VERSION & 0xF) << 0) ) )
static CHAR VERSION_STRING[64];

static DWORD TcStdInMd5sum(PUINT8);
static DWORD TcFileMd5sum(PUINT8, LPCTSTR);
static CONST PCHAR InitVersionString();
static BOOL IsPatternFileMd5(LPCSTR digest, LPCTSTR szFile);
static BOOL CheckFilesMd5(LPCTSTR);

void Usage(LPCTSTR progname)
{
  _tprintf(_T("\n"));
  _tprintf(_T("md5sum - compute and check MD5 message digest\n"));
  _tprintf(_T("\n"));
  _tprintf(_T("usage:\n"));
  _tprintf(_T("%s <FILE>...\n"), progname);
  _tprintf(_T("%s < <FILE>\n"), progname);
  _tprintf(_T("%s [OPTION] [FILE]\n"), progname);
  _tprintf(_T("  -c,--check <FILE>    read MD5 sums from the FILEs and check them\n"));
  _tprintf(_T("  -f,--file <FILE>     compute MD5 message digest\n"));
  _tprintf(_T("  -h,--help            print this help\n"));
  _tprintf(_T("  -v,--version         output version information and exit\n"));
}

/**
* entry
* argv[1]: a file name
*/
int _tmain(int argc, TCHAR* argv[])
{
  DWORD dwRet = 1;
  UCHAR buf[32 + 1] = { 0 };
  INT flag_check = 0;
  INT flag_file = 0;
  INT flag_help = 0;
  INT flag_version = 0;
  INT tempSum;
  LPTSTR file_optarg = NULL;
  LPTSTR check_optarg = NULL;

  int c;
  int option_index;
  TCHAR short_options[] = _T("c:f:hv");
  struct option long_options[] =
  {
    {_T("check"), required_argument, NULL, _T('c') },
    {_T("file"), required_argument, NULL, _T('f') },
    {_T("help"), no_argument, NULL, _T('h')},
    {_T("version"), no_argument, NULL, _T('v')},
    {NULL, 0, NULL, 0}
  };

  DEBUG_INIT(TO_STD_CONSOLE, NULL);

  _tsetlocale(LC_ALL, _T(""));
  InitVersionString();

  while (1)
  {
    option_index = 0;
    c = getopt_long(argc, argv, short_options, long_options, &option_index);

    if (c == -1)
      break;

    switch (c)
    {
    case 0:
      if (long_options[option_index].flag != NULL)
      {
        DEBUG((D_INFO, _T("--%s\n"), long_options[option_index].name));
      }
      break;

    case _T('c'):
      DEBUG((D_INFO, _T("option -c\n")));
      flag_check = 1;
      check_optarg = optarg;
      break;

    case _T('f'):
      DEBUG((D_INFO, _T("option -f\n")));
      flag_file = 1;
      file_optarg = optarg;
      break;

    case _T('h'):
      DEBUG((D_INFO, _T("option -h\n")));
      flag_help = 1;
      break;

    case _T('v'):
      DEBUG((D_INFO, _T("option -v\n")));
      flag_version = 1;
      break;

    case _T('?'):
      DEBUG((D_ERROR, _T("unknown option\n")));
      dwRet = 2;
	  goto Exit;
      break;

    default:
      DEBUG((D_ERROR, _T("You are not supposed to reach here\n")));
      dwRet = 3;
	  goto Exit;
      break;
    }
  }

  //
  // check arguments
  //
  tempSum = flag_check + flag_file + flag_help + flag_version;
  if (tempSum > 1)
  {
    DEBUG((D_ERROR, _T("fail to parse arguments.\n")));
    DEBUG((D_ERROR, _T("try %s -h to see supported commands\n"), argv[0]));
    dwRet = 4;
	goto Exit;
  }

  //
  // address for no option cases
  // md5sum [FILE]...
  // md5sum < [FILE]
  //
  if (tempSum == 0)
  {
    int i;

    // md5sum < [FILE]
    if (argc == 1)
    {
      dwRet = TcStdInMd5sum(buf);
      if (dwRet == 0)
      {
        _tprintf(_T("%hs"), buf);
        goto Exit;
      }
      else {
        DEBUG((D_ERROR, _T("compute md5 digest for file\n")));
        dwRet = 1;
		goto Exit;
      }
    }
    else {
      // md5sum [FILE]...
      for (i = 1; i < argc; i++)
      {
        dwRet = TcFileMd5sum(buf, argv[i]);

        if (dwRet == 0)
        {
          _tprintf(_T("%hs"), buf);
          _tprintf(_T("  "));
          _tprintf(_T("%s"), argv[i]);
          _tprintf(_T("\n"));
        }
        else
        {
          DEBUG((D_ERROR, _T("compute md5 digest for file %s\n"), argv[i]));
          dwRet = 1;
		  goto Exit;
        }
      }
      goto Exit;
    }
  }

  ZeroMemory(buf, sizeof(buf));

  //
  // tempSum == 1
  //
  if (flag_check == 1)
  {
    return CheckFilesMd5(check_optarg) == TRUE ? 0 : 1;
  }

  if (flag_file == 1)
  {
    dwRet = TcFileMd5sum(buf, file_optarg);

    if (dwRet == 0)
    {
      _tprintf(_T("%hs"), buf);
      _tprintf(_T("  "));
      _tprintf(_T("%s"), file_optarg);
      _tprintf(_T("\n"));
    }
    else
    {
      DEBUG((D_ERROR, _T("compute md5 digest for file %s\n"), file_optarg));
      dwRet = 1;
	  goto Exit;
    }
  }

  if (flag_help == 1)
  {
    Usage(argv[0]);
    dwRet = 0;
	goto Exit;
  }

  if (flag_version == 1)
  {
    _tprintf(_T("%hs\n"), VERSION_STRING);
    dwRet = 0;
	goto Exit;
  }

Exit:
  DEBUG_FINISH();
  return dwRet;
}

static CONST PCHAR InitVersionString()
{
  sprintf_s(VERSION_STRING, sizeof(VERSION_STRING)/sizeof(VERSION_STRING[0]), "%d.%d.%d", MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION);
  return VERSION_STRING;
}

/*
 * check with file
 * return TRUE if the given digest is the computed digest
 * return FALSE, otherwise.
 */
static BOOL IsPatternFileMd5(LPCSTR digest, LPCTSTR szFile)
{
  UCHAR buf[32 + 1];
  DWORD dwTemp;

  dwTemp = TcFileMd5sum(buf, szFile);
  if (dwTemp == 0)
  {
    if (!strcmp(digest, (LPCSTR)buf))
    {
      return TRUE;
    }
  }
  return FALSE;
}

/**
*  compare a serial of files with given digest
*  filenames and digests are read from a given text file
*/
static BOOL CheckFilesMd5(LPCTSTR file)
{
  DWORD dwTotal = 0, dwPass = 0, dwFail = 0;
  CHAR szBuf[512];
  CHAR szDigest[32 + 1];
  CHAR szFile[512];
  FILE *fpMd5DigestFile;
  PWCHAR wszFile;
  PCHAR md5DigestFile;
  LPTSTR scanFile;

#ifdef _UNICODE
  size_t requiredSize;
  requiredSize = wcstombs(NULL, file, 0);
  md5DigestFile = (PCHAR)calloc(requiredSize + 1, sizeof(CHAR));
  if (md5DigestFile == NULL)
  {
	  return -1;
  }
  wcstombs(md5DigestFile, file, requiredSize + 1);
#else
  md5DigestFile = _strdup(file);
#endif

  if (0 != fopen_s(&fpMd5DigestFile, md5DigestFile, "r"))
  {
    DEBUG((D_ERROR, _T("fail to open file %hs\n"), md5DigestFile));
    return FALSE;
  }
  if (!fpMd5DigestFile)
  {
	  goto error;
  }
  while (NULL != fgets(szBuf, sizeof(szBuf), fpMd5DigestFile))
  {
    PCHAR p;
    DWORD dwLenTemp = (DWORD)strlen(szBuf);

    if (dwLenTemp > 1 && szBuf[dwLenTemp - 1] == '\n')
    {
      szBuf[dwLenTemp - 1] = 0;
      dwLenTemp--;
    }
    if (dwLenTemp > 1 && szBuf[dwLenTemp - 1] == '\r')
    {
      szBuf[dwLenTemp - 1] = 0;
      dwLenTemp--;
    }
    if (dwLenTemp > 1 && szBuf[dwLenTemp - 1] == ' ')
    {
      szBuf[dwLenTemp - 1] = 0;
      dwLenTemp--;
    }
    if (dwLenTemp > 1 && szBuf[dwLenTemp - 1] == '\t')
    {
      szBuf[dwLenTemp - 1] = 0;
      dwLenTemp--;
    }

    p = szBuf;
    while (*p == ' ' || *p == '\t')
    {
      p++;
    }

    memset(szDigest, 0, sizeof(szDigest));
    strncpy(szDigest, p, sizeof(szDigest) - 1);
    p += sizeof(szDigest) - 1;

    while (*p == ' ' || *p == '\t')
    {
      p++;
    }
    memset(szFile, 0, sizeof(szFile));
    strncpy(szFile, p, sizeof(szFile) - 1);

    DEBUG((D_INFO, _T("Read digest: %hs, read file name: %hs\n"), szDigest, szFile));

#ifdef _UNICODE
    // DWORD requiredSize;
    requiredSize = mbstowcs(NULL, szFile, 0);
    wszFile = (PWCHAR)calloc(requiredSize + 1, sizeof(WCHAR));
	if (!wszFile)
	{
		goto error;
	}
    mbstowcs(wszFile, szFile, requiredSize + 1);
    scanFile = wszFile;
#else
    scanFile = _strdup(file);
#endif
    //
    // compute and compare the md5sum digest
    //
    if (IsPatternFileMd5(szDigest, scanFile))
    {
      _tprintf(_T("%s: OK\n"), scanFile);
      dwPass++;
    }
    else
    {
      _tprintf(_T("%s: FAILED\n"), scanFile);
      dwFail++;
    }

    if (scanFile)
    {
      free(scanFile);
    }
  }
error:
  if (fpMd5DigestFile)
  {
    fclose(fpMd5DigestFile);
  }

  if (md5DigestFile)
  {
    free(md5DigestFile);
  }

  dwTotal = dwPass + dwFail;

  if (dwFail > 0)
  {
    DEBUG((D_WARNING, _T("%d of %d computed checksums did NOT match"), dwFail, dwTotal));
  }

  return (dwFail == 0);
}


/**
* compute md5sum of a file
* dest: a pointer to start of a buf for the result
* filename: filename
*/
static DWORD TcFileMd5sum(PUINT8 dest, LPCTSTR filename)
{
  //CONST DWORD BUFSIZE = 1024;
  //CONST DWORD MD5LEN = 16;
#define BUFSIZE 1024
#define MD5LEN 16
  LPSTR HASH_DIGITS = "0123456789abcdef";

  HANDLE hFile;
  HCRYPTPROV hCryptProv;
  HCRYPTHASH hCryptHash;
  BYTE rFileBuf[BUFSIZE];
  BYTE szHashBuf[MD5LEN];
  PCHAR szHashDigits;
  BOOL bTmpRet;
  DWORD i;
  DWORD dwRet;
  DWORD dwRdBufSz;
  DWORD dwHashBufSz;

  hFile = NULL;
  szHashDigits = HASH_DIGITS;
  hCryptProv = 0;
  hCryptHash = 0;
  dwRdBufSz = 0;
  dwHashBufSz = 0;
  bTmpRet = FALSE;
  dwRet = 0;

  hFile = CreateFile(
    filename,
    GENERIC_READ,
    FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
    NULL,
    OPEN_EXISTING,
    FILE_FLAG_SEQUENTIAL_SCAN,
    NULL);

  if (INVALID_HANDLE_VALUE == hFile)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("failed to call CreateFile: %s\n"), filename));
    DEBUG((D_ERROR, _T("errorcode: 0x%x\n"), dwRet));
    return dwRet;
  }

  bTmpRet = CryptAcquireContext(&hCryptProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT);

  if (!bTmpRet)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("fail to get handle to CryptAcquireContext, errorcode: 0x%x\n"), dwRet));
    CloseHandle(hFile);
    return dwRet;
  }

  bTmpRet = CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hCryptHash);

  if (!bTmpRet)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("fail to call CryptCreateHash, errorcode: 0x%x\n"), dwRet));
    CloseHandle(hFile);
    CryptReleaseContext(hCryptProv, 0);
    return dwRet;
  }

  while (bTmpRet = ReadFile(hFile, rFileBuf, BUFSIZE, &dwRdBufSz, NULL))
  {
    if (0 == dwRdBufSz)
    {
      break;
    }

    if (!CryptHashData(hCryptHash, rFileBuf, dwRdBufSz, 0))
    {
      dwRet = GetLastError();
      DEBUG((D_ERROR, _T("fail to call CryptHashData, errorcode: 0x%x\n"), dwRet));
      CryptReleaseContext(hCryptProv, 0);
      CryptDestroyHash(hCryptHash);
      CloseHandle(hFile);
      return dwRet;
    }
  }

  if (!bTmpRet)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("ReadFile fail.\n")));
    DEBUG((D_ERROR, _T("errorcode: 0x%x\n"), dwRet));
    CryptReleaseContext(hCryptProv, 0);
    CryptDestroyHash(hCryptHash);
    CloseHandle(hFile);
    return dwRet;
  }

  dwHashBufSz = MD5LEN;
  bTmpRet = CryptGetHashParam(hCryptHash, HP_HASHVAL, szHashBuf, &dwHashBufSz, 0);

  if (bTmpRet)
  {
    for (i = 0; i < dwHashBufSz; i++)
    {
      sprintf_s((char *)dest + 2 * i, 3, "%c%c", szHashDigits[szHashBuf[i] >> 4], szHashDigits[szHashBuf[i] & 0xf]);
    }
  }
  else
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("fail to call CryptGetHashParam, errorcode: 0x%x\n"), dwRet));
  }

  CryptDestroyHash(hCryptHash);
  CryptReleaseContext(hCryptProv, 0);
  CloseHandle(hFile);

  return dwRet;
#undef BUFSIZE
#undef MD5LEN
}


/**
* compute the md5sum of a buf
* dest: a buf for the result
* data: a pointer to start of a bunch of data
* nBytes: number of bytes for the data bulk
*/
static DWORD TcBufMd5sum(PUINT8 dest, LPBYTE data, DWORD nBytes)
{
  //CONST DWORD BUFSIZE = 1024;
  //CONST DWORD MD5LEN = 16;
#define BUFSIZE 1024
#define MD5LEN 16
  LPSTR HASH_DIGITS = "0123456789abcdef";

  DWORD i;
  DWORD dwRet;
  BOOL bTmpRet;
  HCRYPTPROV hCryptProv;
  HCRYPTHASH hCryptHash;
  BYTE szHashBuf[MD5LEN];
  DWORD dwHashBufSz;
  PCHAR szHashDigits;

  szHashDigits = HASH_DIGITS;
  hCryptProv = 0;
  hCryptHash = 0;
  dwHashBufSz = 0;
  dwRet = 0;
  bTmpRet = FALSE;

  bTmpRet = CryptAcquireContext(&hCryptProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT);
  if (!bTmpRet)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("failed to call CryptAcquireContext, retval: 0x%x\n"), dwRet));
    return dwRet;
  }

  bTmpRet = CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hCryptHash);
  if (!bTmpRet)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("failed to call CryptCreateHash retval: %d\n"), dwRet));
    CryptReleaseContext(hCryptProv, 0);
    return dwRet;
  }

  bTmpRet = CryptHashData(hCryptHash, data, nBytes, 0);
  if (!bTmpRet)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("failed to call CryptHashData retval: 0x%x\n"), dwRet));
    CryptReleaseContext(hCryptProv, 0);
    CryptDestroyHash(hCryptHash);
    return dwRet;
  }

  dwHashBufSz = MD5LEN;
  bTmpRet = CryptGetHashParam(hCryptHash, HP_HASHVAL, szHashBuf, &dwHashBufSz, 0);
  if (bTmpRet)
  {
    for (i = 0; i < dwHashBufSz; i++)
    {
      sprintf_s((char *)dest + 2 * i, 3, "%c%c", szHashDigits[szHashBuf[i] >> 4], szHashDigits[szHashBuf[i] & 0xf]);
    }
  }
  else
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("failed to call CryptGetHashParam, retval: 0x%x\n"), dwRet));
  }

  CryptDestroyHash(hCryptHash);
  CryptReleaseContext(hCryptProv, 0);

  return dwRet;
#undef BUFSIZE
#undef MD5LEN

}


/**
* compute md5sum of a file
* dest: a pointer to start of a buf for the result
*/
static DWORD TcStdInMd5sum(PUINT8 dest)
{
  //CONST DWORD BUFSIZE = 1024;
  //CONST DWORD MD5LEN = 16;
#define BUFSIZE 1024
#define MD5LEN 16
  LPSTR HASH_DIGITS = "0123456789abcdef";

  HANDLE hFile;
  HCRYPTPROV hCryptProv;
  HCRYPTHASH hCryptHash;
  BYTE rFileBuf[BUFSIZE];
  BYTE szHashBuf[MD5LEN];
  PCHAR szHashDigits;
  BOOL bTmpRet;
  DWORD i;
  DWORD dwRet;
  DWORD dwRdBufSz;
  DWORD dwHashBufSz;

  hFile = NULL;
  szHashDigits = HASH_DIGITS;
  hCryptProv = 0;
  hCryptHash = 0;
  dwRdBufSz = 0;
  dwHashBufSz = 0;
  bTmpRet = FALSE;
  dwRet = 0;

  hFile = GetStdHandle(STD_INPUT_HANDLE);
  if (hFile == INVALID_HANDLE_VALUE)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("fail to get std handle, exitcode=0x%x"), dwRet));
    return dwRet;
  }

  bTmpRet = CryptAcquireContext(&hCryptProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT);

  if (!bTmpRet)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("fail to get handle to CryptAcquireContext, errorcode: 0x%x\n"), dwRet));
    CloseHandle(hFile);
    return dwRet;
  }

  bTmpRet = CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hCryptHash);

  if (!bTmpRet)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("fail to call CryptCreateHash, errorcode: 0x%x\n"), dwRet));
    CloseHandle(hFile);
    CryptReleaseContext(hCryptProv, 0);
    return dwRet;
  }

  while (bTmpRet = ReadFile(hFile, rFileBuf, BUFSIZE - 2, &dwRdBufSz, NULL))
  {
    if (0 == dwRdBufSz)
    {
      break;
    }

    if (!CryptHashData(hCryptHash, rFileBuf, dwRdBufSz, 0))
    {
      dwRet = GetLastError();
      DEBUG((D_ERROR, _T("fail to call CryptHashData, errorcode: 0x%x\n"), dwRet));
      CryptReleaseContext(hCryptProv, 0);
      CryptDestroyHash(hCryptHash);
      CloseHandle(hFile);
      return dwRet;
    }
  }

  if (!bTmpRet)
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("ReadFile fail.\n")));
    DEBUG((D_ERROR, _T("errorcode: 0x%x\n"), dwRet));
    CryptReleaseContext(hCryptProv, 0);
    CryptDestroyHash(hCryptHash);
    CloseHandle(hFile);
    return dwRet;
  }

  dwHashBufSz = MD5LEN;
  bTmpRet = CryptGetHashParam(hCryptHash, HP_HASHVAL, szHashBuf, &dwHashBufSz, 0);

  if (bTmpRet)
  {
    for (i = 0; i < dwHashBufSz; i++)
    {
      sprintf_s((char *)dest + 2 * i, 3, "%c%c", szHashDigits[szHashBuf[i] >> 4], szHashDigits[szHashBuf[i] & 0xf]);
    }
  }
  else
  {
    dwRet = GetLastError();
    DEBUG((D_ERROR, _T("fail to call CryptGetHashParam, errorcode: 0x%x\n"), dwRet));
  }

  CryptDestroyHash(hCryptHash);
  CryptReleaseContext(hCryptProv, 0);
  CloseHandle(hFile);

  return dwRet;

#undef BUFSIZE
#undef MD5LEN
}
