/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file
launch a process into any session under the local system account
**/
#include <Windows.h>
#include <WtsApi32.h>
#include <UserEnv.h>
#include <TlHelp32.h>
#include "TcConfig.h"
#include "TcDebugLib.h"
#include "TcWin32Ctrl.h"
#pragma comment(lib,"AdvApi32.lib")
#pragma comment(lib,"WtsApi32.lib")
#pragma comment(lib,"UserEnv.lib")

//
// Custom message for service to launch an app into session 1
//
#define SERVICE_CONTROL_TC_MESSAGE_ID1 0x0085

/**
Launch the process into a different session

@param szExeFilename   The executable file path to be loaded and run
@param szParamList     parameter list for the executable, multiple parameters separated with spaces
@return TRUE on success, FALSE on failure
**/
BOOL
TC_WIN32_API
LaunchAppIntoDifferentSession(
  LPCWSTR szExeFileName,
  LPCWSTR szParamList,
  LPDWORD pdwResult
)
{
  BOOL                  bTemp;
  BOOL                  bResult;
  DWORD                 dwTemp;
  DWORD                 dwSessionNumber;
  DWORD                 dwWinlogonSessionNumber;
  DWORD                 dwWinlogonProcId;
  DWORD                 dwCreationFlags;
  HANDLE                hUsrToken;
  HANDLE                hUsrToken2;
  HANDLE                hProcToken;
  HANDLE                hProc;
  HANDLE                hSnapshot;
  PROCESS_INFORMATION   theProcInfo;
  STARTUPINFO           theStartupInfo;
  TOKEN_PRIVILEGES      theTokenPrivileges;
  LUID                  theLuid;
  LPVOID                pTheEnv;
  PROCESSENTRY32        theProcEntry;

  bResult = FALSE;
  dwWinlogonProcId = 0;
  //
  // Log the client on to the local computer
  //
  dwSessionNumber = WTSGetActiveConsoleSessionId();
  DEBUG((D_INFO, L"%hs: WTSGetActiveConsoleSessionId returned: 0x%X\n", __FUNCTION__, dwSessionNumber));

  //
  // Find the winlogon process
  //
  hSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  if (hSnapshot == INVALID_HANDLE_VALUE)
  {
    DEBUG((D_ERROR, L"%hs: CreateToolhelp32Snapshot error: 0x%X\n", __FUNCTION__, GetLastError()));
    goto exit1;
  }

  theProcEntry.dwSize = sizeof(PROCESSENTRY32);
  if (!Process32First(hSnapshot, &theProcEntry))
  {
    DEBUG((D_ERROR, L"%hs: Process32First error: 0x%X\n", __FUNCTION__, GetLastError()));
    goto exit2;
  }

  do
  {
    if (_wcsicmp(theProcEntry.szExeFile, L"winlogon.exe") == 0)
    {
      //
      // found a winlogon process
      // make sure it's running in the console session
      //
      dwWinlogonSessionNumber = 0;
      if (ProcessIdToSessionId(theProcEntry.th32ProcessID, &dwWinlogonSessionNumber)
        && dwWinlogonSessionNumber == dwSessionNumber)
      {
        dwWinlogonProcId = theProcEntry.th32ProcessID;
        DEBUG((D_INFO, L"%hs: winLogonPid = %d\n", __FUNCTION__, dwWinlogonProcId));
        break;
      }
    }
  } while (Process32Next(hSnapshot, &theProcEntry));

  //
  // query user token
  //
  if (!WTSQueryUserToken(dwSessionNumber, &hUsrToken))
  {
    DEBUG((D_ERROR, L"%hs: WTSQueryUserToken error: 0x%X\n", __FUNCTION__, GetLastError()));
    goto exit2;
  }

  ZeroMemory(&theProcInfo, sizeof(theProcInfo));
  ZeroMemory(&theStartupInfo, sizeof(STARTUPINFO));

  theStartupInfo.cb = sizeof(STARTUPINFO);
  theStartupInfo.lpDesktop = L"winsta0\\default";
  dwCreationFlags = NORMAL_PRIORITY_CLASS | CREATE_NEW_CONSOLE;

  //
  // approach one:
  // To Hide the process Window
  //
  theStartupInfo.dwFlags = STARTF_USESHOWWINDOW;
  theStartupInfo.wShowWindow = SW_HIDE;

  //
  // Or
  // approach two:
  // to hide process window
  // CREATE_NEW_CONSOLE: The new process has a new console, instead of inheriting its parent's console (the default)
  // CREATE_NO_WINDOW: The process is a console application that is being run without a console window
  //
  //dwCreationFlags &= ~CREATE_NEW_CONSOLE;
  //dwCreationFlags |= CREATE_NO_WINDOW;

  hProc = OpenProcess(MAXIMUM_ALLOWED, FALSE, dwWinlogonProcId);
  if (hProc == NULL)
  {
    DEBUG((D_ERROR, L"%hs: OpenProcess error: 0x%X\n", __FUNCTION__, GetLastError()));
    goto exit3;
  }

  DEBUG((D_INFO, L"%hs: try to call OpenProcessToken\n", __FUNCTION__));
  bTemp = OpenProcessToken(
    hProc,
    TOKEN_ADJUST_SESSIONID | TOKEN_READ | TOKEN_WRITE | TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY | TOKEN_DUPLICATE | TOKEN_ASSIGN_PRIMARY,
    &hProcToken);

  if (!bTemp)
  {
    DEBUG((D_ERROR, L"%hs: OpenProcessToken error: 0x%X\n", __FUNCTION__, GetLastError()));
    goto exit4;
  }

  DEBUG((D_INFO, L"%hs: try to call LookupPrivilegeValue\n", __FUNCTION__));
  bTemp = LookupPrivilegeValue(NULL, SE_DEBUG_NAME, &theLuid);
  if (!bTemp)
  {
    DEBUG((D_ERROR, L"%hs: LookupPrivilegeValue error: 0x%X\n", __FUNCTION__, GetLastError()));
    goto exit5;
  }

  //
  // set token privileges
  //
  theTokenPrivileges.Privileges[0].Luid = theLuid;
  theTokenPrivileges.PrivilegeCount = 1;
  theTokenPrivileges.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

  //
  // To duplicate a user token
  //
  bTemp = DuplicateTokenEx(hProcToken, MAXIMUM_ALLOWED, NULL, SecurityIdentification, TokenPrimary, &hUsrToken2);
  if (!bTemp)
  {
    DEBUG((D_ERROR, L"%hs: DuplicateTokenEx error: 0x%X\n", __FUNCTION__, GetLastError()));
    goto exit5;
  }

  //
  // SetTokenInformation
  //
  DEBUG((D_INFO, L"%hs: SetTokenInformation return value will be ignored\n", __FUNCTION__));
  bTemp = SetTokenInformation(hUsrToken2, TokenSessionId, (void*)&dwSessionNumber, sizeof(DWORD));
  if (!bTemp)
  {
    DEBUG((D_WARNING, L"%hs: SetTokenInformation error: 0x%X\n", __FUNCTION__, GetLastError()));
    //goto exit6;
  }

  //
  // Adjust Token privilege
  //
  dwTemp = -1;
  bTemp = AdjustTokenPrivileges(hUsrToken2, FALSE, &theTokenPrivileges, sizeof(TOKEN_PRIVILEGES), (PTOKEN_PRIVILEGES)NULL, NULL);
  if (!bTemp)
  {
    dwTemp = GetLastError();
    DEBUG((D_ERROR, L"%hs: AdjustTokenPrivileges error: 0x%X\n", __FUNCTION__, dwTemp));
    goto exit6;
  }

  DEBUG((D_INFO, L"%hs: AdjustTokenPrivileges info: return TRUE\n", __FUNCTION__));
  if (dwTemp == ERROR_NOT_ALL_ASSIGNED)
  {
    DEBUG((D_WARNING, L"%hs: AdjustTokenPrivileges error: ERROR_NOT_ALL_ASSIGNED\n", __FUNCTION__));
  }

  //
  // create environment block
  //
  bTemp = CreateEnvironmentBlock(&pTheEnv, hUsrToken2, TRUE);
  if (bTemp)
  {
    dwCreationFlags |= CREATE_UNICODE_ENVIRONMENT;
  }
  else
  {
    DEBUG((D_WARNING, L"%hs: CreateEnvironmentBlock error: 0x%X\n", __FUNCTION__, GetLastError()));
    pTheEnv = NULL;
  }


  //
  // Launch a process in the client logon session and begin impersonation of client
  //
  bTemp = CreateProcessAsUser(
    hUsrToken2,          // client's access token
    szExeFileName,       // file to execut
    (LPWSTR)szParamList, // command line
    NULL,                // pointer to process SECURITY_ATTRIBUTES
    NULL,                // pointer to thread SECURITY_ATTRIBUTES
    FALSE,               // handles are not inheritable
    dwCreationFlags,     // creation flags
    pTheEnv,             // pointer to new environment block 
    NULL,                // name of current directory 
    &theStartupInfo,     // pointer to STARTUPINFO structure
    &theProcInfo         // receives information about new process
  );

  while (TRUE)
  {
    if (GetExitCodeProcess(theProcInfo.hProcess, &dwTemp))
    {
      if (dwTemp == STILL_ACTIVE)
      {
        Sleep(300);
      }
      else
      {
        break;
      }
    }
  }

  DEBUG((D_INFO, L"%hs: CreateProcessAsUser exit value: 0x%X\n", __FUNCTION__, dwTemp));
  if (pdwResult != NULL)
  {
    *pdwResult = dwTemp;
  }

  if (!bTemp)
  {
    DEBUG((D_ERROR, L"%hs: CreateProcessAsUser error: 0x%X\n", __FUNCTION__, GetLastError()));
    goto exit7;
  }

  bResult = TRUE;

  //
  // clean up resources
  //
  if (theProcInfo.hThread != NULL)
  {
    CloseHandle(theProcInfo.hThread);
    theProcInfo.hThread = NULL;
  }
  if (theProcInfo.hProcess != NULL)
  {
    CloseHandle(theProcInfo.hProcess);
    theProcInfo.hProcess = NULL;
  }

exit7:
  if (pTheEnv != NULL)
  {
    DestroyEnvironmentBlock(pTheEnv);
    pTheEnv = NULL;
  }

exit6:
  if (hUsrToken2 != NULL)
  {
    CloseHandle(hUsrToken2);
    hUsrToken2 = NULL;
  }

exit5:
  if (hProcToken != NULL)
  {
    CloseHandle(hProcToken);
    hProcToken = NULL;
  }

exit4:
  if (hProc != NULL)
  {
    CloseHandle(hProc);
    hProc = NULL;
  }

exit3:
  if (hUsrToken != NULL)
  {
    CloseHandle(hUsrToken);
    hUsrToken = NULL;
  }

exit2:
  if (hSnapshot != INVALID_HANDLE_VALUE)
  {
    CloseHandle(hSnapshot);
    hSnapshot = NULL;
  }

exit1:
  return bResult;
}


/**
  Determine the session ID of the currently logged-on user
**/
static
DWORD
GetCurrentSessionId(
  void
)
{
  DWORD             i;
  DWORD             n_sessions;
  DWORD             SessionId;
  BOOL              bRet;
  WTS_SESSION_INFO  *pSessionInfo;

  n_sessions = 0;
  bRet = WTSEnumerateSessions(WTS_CURRENT_SERVER, 0, 1, &pSessionInfo, &n_sessions);
  if (!bRet)
  {
    return 0;
  }

  SessionId = 0;

  for (i = 0; i < n_sessions; ++i)
  {
    if (pSessionInfo[i].State == WTSActive)
    {
      SessionId = pSessionInfo[i].SessionId;
      break;
    }
  }

  WTSFreeMemory(pSessionInfo);
  return SessionId;
}

/**
  Launch process as the currently logged-on user
**/
BOOL
TC_WIN32_API
TcLaunchProcess(
  LPCWSTR exeName,
  LPCWSTR paramList
)
{
  DWORD                SessionId;
  HANDLE               hToken;
  BOOL                 bRet;
  BOOL                 bResult;
  void                 *environment;
  STARTUPINFO          theStartupInfo = { sizeof(theStartupInfo) };
  PROCESS_INFORMATION  theProcInfo = { 0 };
  DWORD                dwCreationFlags;

  bResult = FALSE;

  SessionId = GetCurrentSessionId();
  DEBUG((D_INFO, L"%hs: GetCurrentSessionId returned: 0x%X\n", __FUNCTION__, SessionId));

  if (SessionId == 0)    // no-one logged in
  {
    DEBUG((D_ERROR, L"%hs: no one logged in\n", __FUNCTION__));
    goto exit1;
  }

  bRet = WTSQueryUserToken(SessionId, &hToken);
  if (!bRet)
  {
    DEBUG((D_ERROR, L"%hs:WTSQueryUserToken\n", __FUNCTION__));
    goto exit1;
  }

  environment = NULL;
  bRet = CreateEnvironmentBlock(&environment, hToken, TRUE);
  if (!bRet)
  {
    DEBUG((D_ERROR, L"%hs: CreateEnvironmentBlock\n", __FUNCTION__));
    goto exit2;
  }

  //
  // Do NOT want to inherit handles here
  //
  ZeroMemory(&theProcInfo, sizeof(theProcInfo));
  ZeroMemory(&theStartupInfo, sizeof(theStartupInfo));
  theStartupInfo.cb = sizeof(theStartupInfo);
  theStartupInfo.lpDesktop = L"winsta0\\default";
  dwCreationFlags = NORMAL_PRIORITY_CLASS | CREATE_UNICODE_ENVIRONMENT;

  bRet = CreateProcessAsUser(hToken, exeName, (LPWSTR)paramList, NULL, NULL, FALSE, dwCreationFlags, environment, NULL, &theStartupInfo, &theProcInfo);
  DEBUG((D_INFO, L"%hs: CreateProcessAsUser returned: 0x%X\n", __FUNCTION__, bRet));
  if (!bRet)
  {
    DEBUG((D_ERROR, L"%hs: CreateProcessAsUser\n", __FUNCTION__));
    goto exit3;
  }

  bResult = TRUE;

  CloseHandle(theProcInfo.hThread);
  CloseHandle(theProcInfo.hProcess);

exit3:
  DestroyEnvironmentBlock(environment);

exit2:
  CloseHandle(hToken);

exit1:
  return bResult;
}