/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file

Build Environment 1
Windows 7 SP1 x64
WinDDK 7600.16385.1
Microsoft Visual Studio 2008 SP1

Build Environment 2
Windows 10 x64
WinDDK 7600.16385.1
Microsoft Visual Studio 2015 update 3
**/

#include "TcConfig.h"
#include "TcDebugLib.h"
#include "TcCtl.h"
#include "TcWin32Ctrl.h"
#include "TcNtLib.h"
#include <process.h>
#include <WtsApi32.h>

//
// if undefined, debug to file instead
//
//#define SVC_DEBUG_TO_CONSOLE

#define SERVICE_NAME             L"TcSvcMonitor"
#define IFSETUP_CMD_STR          L"C:\\ifsetup2.bat"

#ifdef SHUTDOWN_CMD_STR
#undef SHUTDOWN_CMD_STR
#define SHUTDOWN_CMD_STR         L"shutdown -f -s -t 0"
#endif

#ifdef REBOOT_CMD_STR
#undef REBOOT_CMD_STR
#define REBOOT_CMD_STR           L"shutdown -f -r -t 0"
#endif

//
// command to change screen resolution
//
#define TC_CHGRES_APPNAME_STR    L"C:\\TcChgres.exe"
#define TC_CHGRES_TEST_DISPLAY   L"C:\\TcChgres.exe --test"
#define TC_USBCTRL_APPNAME_STR   L"C:\\TcUsbCtrl.bat"

#define TC_LOCAL_MSG_STOP_SVC    10
#define TC_SERVER_MSG_SHUTDOWN   TCDMSG_SHUTDOWN   // 11
#define TC_SERVER_MSG_REBOOT     TCDMSG_REBOOT     // 12
//#define TC_LOCAL_MSG_SHUTDOWN  13 /* moved to tcctl.h */

#define SERVICE_ARGV0            SERVICE_NAME
#define SERVICE_ARGV1            L"0"
#define SERVICE_ARGV2            L"eth0"
#define SERVICE_ARGV3            L"12345"

#define TCSVC_THREAD_NUM         3
LPWSTR                  gLpwszArgv[] = { SERVICE_ARGV0, SERVICE_ARGV1, SERVICE_ARGV2, SERVICE_ARGV3 };
SERVICE_STATUS          gServiceStatus;
SERVICE_STATUS_HANDLE   gHandleServiceStatus;
HANDLE                  ghChgresEvent;
HANDLE                  ghUsbCtrlEvent;
HANDLE                  ghStopEvent;
HANDLE                  ghThreads[TCSVC_THREAD_NUM];
BOOL                    gbThreadIsAlive[TCSVC_THREAD_NUM] = { 0 };
BOOL                    gbIsUserLoggedOn = FALSE;
BOOL                    gbIsDesktopReady = FALSE;

//
// 0 indicates no command
// value above 0 indicates there exists command to handle
//
int gbCmdToSetResolution;
int gbCmdToSetUsbStorageAccess;

//
// value of screen Width & Height
//
int gnScreenWidth;
int gnScreenHeight;
WCHAR gwszCmdToSetResolution[MAX_PATH];

int gnUsbAccessValue;
WCHAR gwszCmdToSetUsbStorageAccess[MAX_PATH];

/**
Free
**/
VOID
WINAPI
FreeArgv(
	TH_PARMS *pparmsMsgLoop
);

BOOL
WINAPI
MallocArgv(
	DWORD dwArgc,
	LPWSTR *lpwszArgv,
	TH_PARMS *pparmsMsgLoop
);

/**
execute an outside executable
**/
DWORD
WINAPI
SystemWrapper(
	LPVOID pparms
)
{
	DWORD dwRet;

	dwRet = 0;

	DEBUG((D_INFO, L"%hs:\n", __FUNCTION__));
	DEBUG((D_INFO, L"%hs: run command: %ls\n", __FUNCTION__, (LPCTSTR)pparms));

	dwRet = _wsystem((LPCWSTR)pparms);
	//dwRet = WEXITSTATUS(dwRet);

	DEBUG((D_INFO, L"%hs: \"%ls\" exited with %d\n", __FUNCTION__, (LPCTSTR)pparms, dwRet));
	return dwRet;
}


/**
  Report service status to Service Control Manager
**/
VOID
ReportServiceStatus(
	DWORD dwCurrentState,
	DWORD dwWin32ExitCode,
	DWORD dwWaitHint
)
{
	static DWORD dwCheckPoint = 1;

	gServiceStatus.dwCurrentState = dwCurrentState;
	gServiceStatus.dwWin32ExitCode = dwWin32ExitCode;
	gServiceStatus.dwWaitHint = dwWaitHint;

	if (dwCurrentState == SERVICE_START_PENDING)
	{
		gServiceStatus.dwControlsAccepted = 0;
	}
	else
	{
		gServiceStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP |
			SERVICE_ACCEPT_PRESHUTDOWN |
			SERVICE_ACCEPT_SESSIONCHANGE;
	}

	if ((dwCurrentState == SERVICE_RUNNING) ||
		(dwCurrentState == SERVICE_STOPPED))
	{
		gServiceStatus.dwCheckPoint = 0;
	}
	else
	{
		gServiceStatus.dwCheckPoint = dwCheckPoint++;
	}

	SetServiceStatus(gHandleServiceStatus, &gServiceStatus);
}

/**
  Service Control Handler function
**/
DWORD
WINAPI
ServiceControlHandlerExtend(
	DWORD dwCtrl,
	DWORD dwEventType,
	LPVOID lpEventData,
	LPVOID lpContext
)
{
	DEBUG((D_INFO, L"%hs:\n", __FUNCTION__));
	DEBUG((D_INFO, L"%hs: dwCtrl = 0x%X\n", __FUNCTION__, dwCtrl));

	switch (dwCtrl)
	{
	case SERVICE_CONTROL_STOP:
		ReportServiceStatus(SERVICE_STOP_PENDING, NO_ERROR, 0);
		g_exit = TC_LOCAL_MSG_STOP_SVC;
		if (!SetEvent(ghStopEvent))
		{
			DEBUG((D_WARNING, L"SERVICE_CONTROL_STOP: fail to set ghStopEvent\n"));
		}

		DEBUG((D_INFO, L"SERVICE_CONTROL_STOP: g_exit = %d\n", g_exit));
		return 0;

	case SERVICE_CONTROL_INTERROGATE:
		DEBUG((D_INFO, L"SERVICE_CONTROL_INTERROGATE: g_exit = %d\n", g_exit));
		break;

	case SERVICE_CONTROL_PRESHUTDOWN:
		ReportServiceStatus(SERVICE_STOP_PENDING, NO_ERROR, 0);
		g_exit = TC_LOCAL_MSG_SHUTDOWN;

		DEBUG((D_INFO, L"SERVICE_CONTROL_PRESHUTDOWN: g_exit = %d\n", g_exit));
		return 0;

	case SERVICE_CONTROL_SESSIONCHANGE:
		DEBUG((D_INFO, L"SERVICE_CONTROL_SESSIONCHANGE: dwEventType = 0x%X\n", dwEventType));
		switch (dwEventType)
		{
		case WTS_SESSION_LOGON:
			DEBUG((D_INFO, L"%hs: WTS_SESSION_LOGON\n", __FUNCTION__));
			break;

		case WTS_SESSION_UNLOCK:
			DEBUG((D_INFO, L"%hs: WTS_SESSION_UNLOCK\n", __FUNCTION__));
			break;

		default:
			DEBUG((D_INFO, L"SERVICE_CONTROL_SESSIONCHANGE: unaddressed[0x%X]\n", dwEventType));
			break;
		}
		break;

	default:
		DEBUG((D_INFO, L"%hs: g_exit = %d\n", __FUNCTION__, g_exit));
		break;
	}

	ReportServiceStatus(gServiceStatus.dwCurrentState, NO_ERROR, 0);
	return 0;
}


/**
  Check if any user logged on to System

  @param none
  @return TRUE if there is a user logged on to system
**/
BOOL
TC_WIN32_API
TcCheckUserLogon(
	VOID
)
{
	BOOL  bRet;
	DWORD dwIndex;
	DWORD dwCnt;
	DWORD dwTemp;
	PWTS_SESSION_INFO pWtsSessionInfo;

	pWtsSessionInfo = NULL;
	bRet = WTSEnumerateSessions(
		WTS_CURRENT_SERVER_HANDLE, // the RD Session Host server on which your application is running
		0,                         // reserved
		1,                         // API version
		&pWtsSessionInfo,          // pointer to WTS_SESSION_INFO structure instance
		&dwCnt
	);

	if (!bRet)
	{
		DEBUG((D_ERROR, L"%hs: WTSEnumerateSessions error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto exit;
	}

	bRet = FALSE;
	dwTemp = -1;
	DEBUG((D_INFO, L"%hs: WTSEnumerateSessions returned: dwCnt = 0x%X\n", __FUNCTION__, dwCnt));
	for (dwIndex = 0; dwIndex < dwCnt; dwIndex++)
	{
		DEBUG((D_INFO, L"%hs: WTSEnumerateSessions 0x%X: SessionId       = 0x%X\n", __FUNCTION__, dwIndex, pWtsSessionInfo[dwIndex].SessionId));
		DEBUG((D_INFO, L"%hs: WTSEnumerateSessions 0x%X: pWinStationName = %ls\n", __FUNCTION__, dwIndex, pWtsSessionInfo[dwIndex].pWinStationName));
		DEBUG((D_INFO, L"%hs: WTSEnumerateSessions 0x%X: State           = 0x%X\n", __FUNCTION__, dwIndex, pWtsSessionInfo[dwIndex].State));

		if (pWtsSessionInfo[dwIndex].State == WTSActive)
		{
			bRet = TRUE;
			dwTemp = dwIndex;
			gbIsUserLoggedOn = TRUE;
		}
	}

	if (dwTemp < dwCnt)
	{
		DEBUG((D_INFO, L"%hs: Active session Index: 0x%X\n", __FUNCTION__, dwTemp));
		DEBUG((D_INFO, L"%hs: Active session Id   : 0x%X\n", __FUNCTION__, pWtsSessionInfo[dwTemp].SessionId));
		DEBUG((D_INFO, L"%hs: Active session name : %ls\n", __FUNCTION__, pWtsSessionInfo[dwTemp].pWinStationName));
		DEBUG((D_INFO, L"%hs: Active session state: 0x%X\n", __FUNCTION__, pWtsSessionInfo[dwTemp].State));
	}

exit:
	if (pWtsSessionInfo != NULL)
	{
		WTSFreeMemory(pWtsSessionInfo);
		pWtsSessionInfo = NULL;
	}

	return bRet;
}

//
// return the config value of usb storage permission
// return 1 by default (no config file or enalbed USB storage)
// return 0 if config file exists and expects USB storage disabled
//
static DWORD GetUsbAccessValue()
{
	DWORD ret;
	INT t;
	char buf[1024];

	ret = 1;
	ZeroMemory(buf, _countof(buf));
	t = TcOneTimeNtReadFile(buf, sizeof(buf), USBSTORAGE_CONF);
	DEBUG((D_INFO, L"%hs: TcOneTimeNtReadFile return: 0x%X\n", __FUNCTION__, t));
	DEBUG((D_INFO, L"%hs: strlen(buf): 0x%X\n", __FUNCTION__, strlen(buf)));

	if (t <= 0) {
		DEBUG((D_ERROR, L"%hs: fail to load config file: 0x%X\n", __FUNCTION__, t));
		return ret;
	}

	DEBUG((D_INFO, L"%hs: %hs\n", __FUNCTION__, buf));
	ret = atoi(buf);
	return ret;
}

//
// set the config value of usb storage permission
// return 1 on success
// return 0 on failure
//
static DWORD SetUsbAccessValue(int usb_mode)
{
	DWORD ret;
	INT t;
	char buf[1024];

	ret = 0;
	ZeroMemory(buf, _countof(buf));

	sprintf(buf, "%d", usb_mode);
	DEBUG((D_INFO, L"%hs: %hs\n", __FUNCTION__, buf));
	DEBUG((D_INFO, L"%hs: strlen(buf): 0x%X\n", __FUNCTION__, strlen(buf)));

	t = TcOneTimeNtWriteFile(buf, sizeof(buf), USBSTORAGE_CONF);
	DEBUG((D_INFO, L"%hs: TcOneTimeNtWriteFile return: 0x%X\n", __FUNCTION__, t));

	if (t != 0) {
		DEBUG((D_ERROR, L"%hs: fail to save config file: 0x%X\n", __FUNCTION__, t));
		return ret;
	}

	ret = 1;

	return ret;
}


/**
  The worker thread to change screen resolution
**/
unsigned
WINAPI
TcCheckUserLogonWorkerThread(
	LPVOID *pParm
)
{
	unsigned ret = 0;
	DWORD dwStatus;
	DWORD dwResult;
	BOOL bResolutionFlag;
	BOOL bUsbStorageFlag;

	bResolutionFlag = FALSE;
	bUsbStorageFlag = FALSE;

	//
	// loop until desktop is ready
	//
	while (!gbIsDesktopReady)
	{
		dwStatus = WaitForSingleObject(ghStopEvent, 0);
		if (dwStatus == WAIT_OBJECT_0)
		{
			//
			// reset the event to non-signaled state
			//
			if (!ResetEvent(ghStopEvent))
			{
				DEBUG((D_WARNING, L"%hs: fail to reset ghStopEvent\n", __FUNCTION__));
			}
			goto exit;
		}

		if (gbIsUserLoggedOn || TcCheckUserLogon())
		{
			//
			// query if display is attached
			//
			if (LaunchAppIntoDifferentSession(TC_CHGRES_APPNAME_STR, TC_CHGRES_TEST_DISPLAY, &dwResult))
			{
				DEBUG((D_INFO, L"%hs: LaunchAppIntoDifferentSession exit value: 0x%X\n", __FUNCTION__, dwResult));
			}
			else
			{
				DEBUG((D_ERROR, L"%hs: LaunchAppIntoDifferentSession fail\n", __FUNCTION__));
			}

			//
			// 0 indicates display is attached
			//
			if (dwResult == 0)
			{
				//
				// set the desktop status to ready
				//
				gbIsDesktopReady = TRUE;

				//
				// prepare ingredients
				//
				ZeroMemory(gwszCmdToSetResolution, _countof(gwszCmdToSetResolution));
				swprintf(gwszCmdToSetResolution, _countof(gwszCmdToSetResolution), L"%ls --tc", TC_CHGRES_APPNAME_STR);
				DEBUG((D_INFO, L"%hs: %ls\n", __FUNCTION__, gwszCmdToSetResolution));

				//
				// send a message to trigger the resolution changes task
				// and exit the loop
				//
				if (SetEvent(ghChgresEvent))
				{
					DEBUG((D_INFO, L"%hs: set change res event\n", __FUNCTION__));
					bResolutionFlag = TRUE;
				}
				else
				{
					DEBUG((D_WARNING, L"%hs: fail to set change res event\n", __FUNCTION__));
				}


				//
				// prepare ingredients
				//
				ZeroMemory(gwszCmdToSetUsbStorageAccess, _countof(gwszCmdToSetUsbStorageAccess));
				swprintf(gwszCmdToSetUsbStorageAccess, _countof(gwszCmdToSetUsbStorageAccess), L"%ls --enable %d", TC_USBCTRL_APPNAME_STR, GetUsbAccessValue());
				DEBUG((D_INFO, L"%hs: %ls\n", __FUNCTION__, gwszCmdToSetUsbStorageAccess));
				//
				// send a message to trigger the usb storage control task
				// and exit the loop
				//
				if (SetEvent(ghUsbCtrlEvent))
				{
					DEBUG((D_INFO, L"%hs: set usb ctrl event\n", __FUNCTION__));
					bUsbStorageFlag = TRUE;
				}
				else
				{
					DEBUG((D_WARNING, L"%hs: fail to set usb ctrl event\n", __FUNCTION__));
				}

				//
				// run again
				//
				Sleep(2 * 1000);
				DEBUG((D_INFO, L"%hs: set event for usb ctrl again\n", __FUNCTION__));
				if (SetEvent(ghUsbCtrlEvent))
				{
					DEBUG((D_INFO, L"%hs: set usb ctrl event\n", __FUNCTION__));
					bUsbStorageFlag = TRUE;
				}
				else
				{
					DEBUG((D_WARNING, L"%hs: fail to set usb ctrl event\n", __FUNCTION__));
				}

				//
				// exit the thread
				//
				if (bResolutionFlag && bUsbStorageFlag)
				{
					break;
				}
			}
			else
			{
				DEBUG((D_INFO, L"%hs: Seems display was not attached!\n", __FUNCTION__));
			}
		}
		else
		{
			DEBUG((D_INFO, L"%hs: no user was logged on!\n", __FUNCTION__));
		}

		Sleep(1000);
	}

exit:
	_endthreadex(0);

	return ret;
}


/**
  The worker thread to change screen resolution
**/
unsigned
WINAPI
TcChangeResolutionWorkerThread(
	LPVOID pParms
)
{
	UINT ret = 0;
	DWORD dwStatus;
	DWORD dwResult;
	dwResult = 0;
	while (TRUE)
	{
		dwStatus = WaitForSingleObject(ghStopEvent, 0);
		if (dwStatus == WAIT_OBJECT_0)
		{
			//
			// reset the event to non-signaled state
			//
			if (!ResetEvent(ghStopEvent))
			{
				DEBUG((D_WARNING, L"%hs: fail to reset ghStopEvent\n", __FUNCTION__));
			}
			goto exit;
		}

		dwStatus = WaitForSingleObject(ghChgresEvent, 0);
		if (dwStatus == WAIT_OBJECT_0)
		{
			DEBUG((D_INFO, L"%hs: receive event gHChgresEvent\n", __FUNCTION__));
			if (!gbIsDesktopReady)
			{
				DEBUG((D_WARNING, L"%hs: nobody logon to the system, gHChgresEvent would be discarded\n", __FUNCTION__));
			}
			else
			{
				//
				// no guarantee that the screen resolution is changed successfully
				//
				if (LaunchAppIntoDifferentSession(TC_CHGRES_APPNAME_STR, gwszCmdToSetResolution, &dwResult))
				{
					DEBUG((D_INFO, L"%hs: LaunchAppIntoDifferentSession exit value: 0x%X\n", __FUNCTION__, dwResult));
				}
				else
				{
					DEBUG((D_ERROR, L"%hs: LaunchAppIntoDifferentSession error: 0x%X\n", __FUNCTION__, dwResult));
				}
			}

			//
			// reset the event to non-signaled state
			//
			if (!ResetEvent(ghChgresEvent))
			{
				DEBUG((D_WARNING, L"%hs: fail to reset ghChgresEvent\n", __FUNCTION__));
			}
		}

		dwStatus = WaitForSingleObject(ghUsbCtrlEvent, 0);
		if (dwStatus == WAIT_OBJECT_0)
		{
			DEBUG((D_INFO, L"%hs: receive event ghUsbCtrlEvent\n", __FUNCTION__));
			if (!gbIsUserLoggedOn)
			{
				DEBUG((D_WARNING, L"%hs: nobody logon to the system, ghUsbCtrlEvent would be discarded\n", __FUNCTION__));
			}
			else
			{
				//
				// no guarantee that the usb stroage is correctly set
				//
				if (LaunchAppIntoDifferentSession(TC_USBCTRL_APPNAME_STR, gwszCmdToSetUsbStorageAccess, &dwResult))
				{
					DEBUG((D_INFO, L"%hs: LaunchAppIntoDifferentSession exit value: 0x%X\n", __FUNCTION__, dwResult));
				}
				else
				{
					DEBUG((D_ERROR, L"%hs: LaunchAppIntoDifferentSession error: 0x%X\n", __FUNCTION__, dwResult));
				}
			}

			//
			// reset the event to non-signaled state
			//
			if (!ResetEvent(ghUsbCtrlEvent))
			{
				DEBUG((D_WARNING, L"%hs: fail to reset ghUsbCtrlEvent\n", __FUNCTION__));
			}
		}

		Sleep(1000);
	}

exit:
	_endthreadex(dwResult);

	return ret;
}

/**
  The worker thread to server
**/
unsigned
WINAPI
TcClientListenerThread(
	TH_PARMS *pparmsMsgLoop
)
{
	unsigned ret;

	ret = ServiceWorkerThread(pparmsMsgLoop);
	_endthreadex(0);

	return ret;
}


/**
  Service initialization
**/
VOID
WINAPI
ServiceInitialization(
	DWORD dwArgc,
	LPWSTR *lpwszArgv
)
{
	TH_PARMS parmsMsgLoop;
	DWORD    dwResult;
	DWORD    dwIndex;

	//
	// report status "Running" to SCM
	//
	ReportServiceStatus(SERVICE_RUNNING, NO_ERROR, 0);

	//
	// initialize handles
	//
	for (dwIndex = 0; dwIndex < TCSVC_THREAD_NUM; dwIndex++)
	{
		ghThreads[dwIndex] = NULL;
	}

	//
	// pass arguments to the struct
	//
	if (!MallocArgv(dwArgc, lpwszArgv, &parmsMsgLoop))
	{
		DEBUG((D_ERROR, L"%hs: mallocArgv\n", __FUNCTION__));
		goto exit1;
	}

	ghChgresEvent = CreateEvent(
		NULL,               // default security attributes
		TRUE,               // manual-reset event
		FALSE,              // initial state is nonsignaled
		L"ChgresEvent"      // object name
	);
	if (ghChgresEvent == NULL)
	{
		DEBUG((D_ERROR, L"%hs: CreateEvent for ghChgresEvent: 0x%X\n", __FUNCTION__, GetLastError()));
		goto exit2;
	}

	ghUsbCtrlEvent = CreateEvent(
		NULL,               // default security attributes
		TRUE,               // manual-reset event
		FALSE,              // initial state is nonsignaled
		L"UsbCtrlEvent"     // object name
	);
	if (ghUsbCtrlEvent == NULL)
	{
		DEBUG((D_ERROR, L"%hs: CreateEvent for ghUsbCtrlEvent: 0x%X\n", __FUNCTION__, GetLastError()));
		goto exit3;
	}

	ghStopEvent = CreateEvent(
		NULL,               // default security attributes
		TRUE,               // manual-reset event
		FALSE,              // initial state is nonsignaled
		L"StopEvent"        // object name
	);
	if (ghStopEvent == NULL)
	{
		DEBUG((D_ERROR, L"%hs: CreateEvent for ghStopEvent: 0x%X\n", __FUNCTION__, GetLastError()));
		goto exit4;
	}

	//
	// Perform service task here...
	//

	//
	// firstly,
	// set up network configuration
	//
	dwResult = SystemWrapper(IFSETUP_CMD_STR);
	DEBUG((D_INFO, L"%hs: SystemWrapper returned: %d\n", __FUNCTION__, dwResult));

	//
	// secondly,
	// create thread to check if any user logged on
	//
	ghThreads[0] = (HANDLE)_beginthreadex(NULL, 0, &TcCheckUserLogonWorkerThread, NULL, 0, NULL);
	if (ghThreads[0] == NULL)
	{
		DEBUG((D_ERROR, L"%hs: TcCheckUserLogonWorkerThread error: 0x%X\n", __FUNCTION__, errno));
		goto exit5;
	}

	//
	// create thread to handle change resolution event
	//
	ghThreads[1] = (HANDLE)_beginthreadex(NULL, 0, &TcChangeResolutionWorkerThread, NULL, 0, NULL);
	if (ghThreads[1] == NULL)
	{
		DEBUG((D_ERROR, L"%hs: TcChangeResolutionWorkerThread error: 0x%X\n", __FUNCTION__, errno));
		goto exit6;
	}

	//
	// create thread to connect to TC server daemon
	//
	ghThreads[2] = (HANDLE)_beginthreadex(NULL, 0, &TcClientListenerThread, &parmsMsgLoop, 0, NULL);
	if (ghThreads[2] == NULL)
	{
		DEBUG((D_ERROR, L"%hs: TcMessageLoopThread error: 0x%X\n", __FUNCTION__, errno));
		goto exit7;
	}

	//
	// global variables indicate thread status: alive or dead
	//
	gbThreadIsAlive[0] = TRUE;
	gbThreadIsAlive[1] = TRUE;
	gbThreadIsAlive[2] = TRUE;

	//
	// loop until all threads return
	//
	while (gbThreadIsAlive[0] || gbThreadIsAlive[1] || gbThreadIsAlive[2])
	{
		//
		// msg thread checking
		//
		if (gbThreadIsAlive[0])
		{
			dwResult = WaitForSingleObject(ghThreads[0], 0);
			if (WAIT_OBJECT_0 == dwResult)
			{
				gbThreadIsAlive[0] = FALSE;
			}
		}

		if (gbThreadIsAlive[1])
		{
			dwResult = WaitForSingleObject(ghThreads[1], 0);
			if (WAIT_OBJECT_0 == dwResult)
			{
				gbThreadIsAlive[1] = FALSE;
			}
		}

		if (gbThreadIsAlive[2])
		{
			dwResult = WaitForSingleObject(ghThreads[2], 0);
			if (WAIT_OBJECT_0 == dwResult)
			{
				gbThreadIsAlive[2] = FALSE;
				//
				// inform other threads to terminate lives
				//
				if (!SetEvent(ghStopEvent))
				{
					DEBUG((D_WARNING, L"%hs: fail to set ghStopEvent\n", __FUNCTION__));
				}
			}
		}

		//
		// when receive set resolution commands from server
		//
		if (gbCmdToSetResolution > 0)
		{
			ZeroMemory(gwszCmdToSetResolution, _countof(gwszCmdToSetResolution));
			swprintf(gwszCmdToSetResolution, _countof(gwszCmdToSetResolution), L"%ls --tc-save --width %d --height %d", TC_CHGRES_APPNAME_STR, gnScreenWidth, gnScreenHeight);
			DEBUG((D_INFO, L"%hs: command line: %ls\n", __FUNCTION__, gwszCmdToSetResolution));

			if (!SetEvent(ghChgresEvent))
			{
				DEBUG((D_WARNING, L"%hs: fail to set ghChgresEvent\n", __FUNCTION__));
			}
			gbCmdToSetResolution = 0;
		}

		//
		// when receive usb-storage-control commands from server
		//
		if (gbCmdToSetUsbStorageAccess > 0)
		{
			ZeroMemory(gwszCmdToSetUsbStorageAccess, _countof(gwszCmdToSetUsbStorageAccess));
			swprintf(gwszCmdToSetUsbStorageAccess, _countof(gwszCmdToSetUsbStorageAccess), L"%ls --enable %d", TC_USBCTRL_APPNAME_STR, gnUsbAccessValue);
			DEBUG((D_INFO, L"%hs: command line: %ls\n", __FUNCTION__, gwszCmdToSetUsbStorageAccess));
			//
			// save the settings
			//
			if (SetUsbAccessValue(gnUsbAccessValue) == 1)
			{
				DEBUG((D_INFO, L"%hs: command line: save the usb-control config successfully\n", __FUNCTION__));
			}
			else
			{
				DEBUG((D_ERROR, L"%hs: command line: fail to save the usb-control config\n", __FUNCTION__));
			}

			if (!SetEvent(ghUsbCtrlEvent))
			{
				DEBUG((D_WARNING, L"%hs: fail to set ghUsbCtrlEvent\n", __FUNCTION__));
			}
			gbCmdToSetUsbStorageAccess = 0;
		}

		//
		// wait for a moment
		//
		Sleep(1000 * 2);
	}

	//
	// thirdly,
	// check reason that message loop exits,
	// and take some actions later...
	//
	DEBUG((D_INFO, L"%hs: ServiceWorkerThread returned, g_exit = %d\n", __FUNCTION__, g_exit));

	//
	// remote shutdown command
	//
	if (g_exit == TC_SERVER_MSG_SHUTDOWN)
	{
		//
		// when shutdown or reboot, no need to clean up resources
		//
		dwResult = SystemWrapper(SHUTDOWN_CMD_STR);
		DEBUG((D_INFO, L"%hs: SystemWrapper returned: %d\n", __FUNCTION__, dwResult));
	}

	//
	// remote reboot command
	//
	if (g_exit == TC_SERVER_MSG_REBOOT)
	{
		//
		// when shutdown or reboot, no need to clean up resources :)
		// 
		dwResult = SystemWrapper(REBOOT_CMD_STR);
		DEBUG((D_INFO, L"%hs: SystemWrapper returned: %d\n", __FUNCTION__, dwResult));
	}

	//
	// clean up
	//
	if (ghThreads[2] != NULL)
	{
		CloseHandle(ghThreads[2]);
		ghThreads[2] = NULL;
	}

exit7:
	if (ghThreads[1] != NULL)
	{
		CloseHandle(ghThreads[1]);
		ghThreads[1] = NULL;
	}

exit6:
	if (ghThreads[0] != NULL)
	{
		CloseHandle(ghThreads[0]);
		ghThreads[0] = NULL;
	}

exit5:
	if (ghStopEvent != NULL)
	{
		CloseHandle(ghStopEvent);
		ghStopEvent = NULL;
	}

exit4:
	if (ghUsbCtrlEvent != NULL)
	{
		CloseHandle(ghUsbCtrlEvent);
		ghUsbCtrlEvent = NULL;
	}

exit3:
	if (ghChgresEvent != NULL)
	{
		CloseHandle(ghChgresEvent);
		ghChgresEvent = NULL;
	}

exit2:
	FreeArgv(&parmsMsgLoop);

exit1:
	//
	// Finally,
	// report status "Stopped" to SCM
	//
	ReportServiceStatus(SERVICE_STOPPED, NO_ERROR, 0);
}


/**
  Service entry point function
**/
VOID
WINAPI
ServiceMain(
	DWORD dwArgc,
	LPWSTR *lpwszArgv
)
{
	DWORD   dwNewArgc;
	LPWSTR  *lpwszNewArgv;
	DWORD   i;

	dwNewArgc = dwArgc;
	lpwszNewArgv = lpwszArgv;

	DEBUG((D_INFO, L"%hs: %d\n", __FUNCTION__, dwArgc));

	//
	// if no cmdline arguments,
	// then use default ones
	//
	if (dwNewArgc < 2)
	{
		dwNewArgc = 4;
		lpwszNewArgv = gLpwszArgv;
	}

	//
	// print arguments
	//
	for (i = 0; i < dwNewArgc; i++)
	{
		DEBUG((D_INFO, L"%hs: %ls\n", __FUNCTION__, lpwszNewArgv[i]));
	}

	gHandleServiceStatus = RegisterServiceCtrlHandlerEx(
		SERVICE_NAME,
		ServiceControlHandlerExtend,
		NULL
	);

	if (!gHandleServiceStatus)
	{
		DEBUG((D_ERROR, L"%hs: RegisterServiceCtrlHandler failed with %d\n", __FUNCTION__, GetLastError()));
		return;
	}

	gServiceStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	gServiceStatus.dwServiceSpecificExitCode = 0;

	ReportServiceStatus(SERVICE_START_PENDING, NO_ERROR, 3000);

	//
	// invoke service initialization function
	//
	ServiceInitialization(dwNewArgc, lpwszNewArgv);
}

//
// Entry for service process
//
void
__cdecl
wmain(
	int argc,
	wchar_t *argv[]
)
{
	SERVICE_TABLE_ENTRY DispatchTable[] =
	{
	  { SERVICE_NAME, (LPSERVICE_MAIN_FUNCTION)ServiceMain },
	  { NULL, NULL }
	};

#ifdef SVC_DEBUG_TO_CONSOLE
	DEBUG_INIT(TO_CONSOLE, NULL);
#else
	DEBUG_INIT(TO_FILE, L"C:\\TcSvcMonitor_Debug.txt");
#endif
	PRINT_INIT(TO_FILE, L"C:\\tcsvcmonitor.txt");

	DEBUG((D_INFO, L"%hs: Entry of service\n", __FUNCTION__));
	PRINT(L"%hs: Entry of service\n", __FUNCTION__);

	//
	// this function will never return unless service is stopped
	//
	if (!StartServiceCtrlDispatcher(DispatchTable))
	{
		DEBUG((D_ERROR, L"%hs: StartServiceCtrlDispatcher failed with %d\n", __FUNCTION__, GetLastError()));
	}

	DEBUG((D_INFO, L"%hs: Exit of service\n", __FUNCTION__));
	PRINT(L"%hs: Exit of service\n", __FUNCTION__);

	DEBUG_FINISH();
	PRINT_FINISH();
}




/**
Free
**/
VOID
WINAPI
FreeArgv(
	TH_PARMS *pparmsMsgLoop
)
{
	DWORD dwIndex;
	DWORD dwArgc;
	char **argv;

	if (pparmsMsgLoop == NULL)
	{
		return;
	}

	dwArgc = pparmsMsgLoop->argc;
	argv = pparmsMsgLoop->argv;
	if (argv != NULL)
	{
		for (dwIndex = 0; dwIndex < dwArgc; dwIndex++)
		{
			if (argv[dwIndex] != NULL)
			{
				free(argv[dwIndex]);
				argv[dwIndex] = NULL;
			}
		}
		free(argv);
		argv = NULL;
	}
	pparmsMsgLoop->argc = 0;
	pparmsMsgLoop->argv = argv;
}

BOOL
WINAPI
MallocArgv(
	DWORD dwArgc,
	LPWSTR *lpwszArgv,
	TH_PARMS *pparmsMsgLoop
)
{
	DWORD dwIndex;
	DWORD dwLen;
	char **argv;
	BOOL bResult;

	bResult = FALSE;
	argv = NULL;
	if (lpwszArgv == NULL || pparmsMsgLoop == NULL)
	{
		DEBUG((D_ERROR, L"%hs: illgal arguments\n", __FUNCTION__));
		goto exit1;
	}

	//
	// convert wide string to single bytes strings
	//
	argv = (char **)calloc(dwArgc, sizeof(char*));
	if (argv == NULL)
	{
		DEBUG((D_ERROR, L"%hs: fail to alloc memory\n", __FUNCTION__));
		goto exit1;
	}
	for (dwIndex = 0; dwIndex < dwArgc; dwIndex++)
	{
		argv[dwIndex] = NULL;
	}
	for (dwIndex = 0; dwIndex < dwArgc; dwIndex++)
	{
		dwLen = (DWORD)wcstombs(NULL, lpwszArgv[dwIndex], 0);
		argv[dwIndex] = (char *)calloc(dwLen + 1, sizeof(char));
		if (argv[dwIndex] == NULL)
		{
			DEBUG((D_ERROR, L"%hs: fail to alloc memory\n", __FUNCTION__));
			goto exit2;
		}
		wcstombs(argv[dwIndex], lpwszArgv[dwIndex], dwLen + 1);
	}

	pparmsMsgLoop->argc = dwArgc;
	pparmsMsgLoop->argv = argv;

	bResult = TRUE;
	return bResult;

exit2:
	if (pparmsMsgLoop != NULL) {
		FreeArgv(pparmsMsgLoop);
	}
	free(argv);
exit1:
	return bResult;
}
