/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file
  launch a process into any session under the local system account
**/
#include <Windows.h>
#include <WinSvc.h>
#include <time.h>
#include <WtsApi32.h>
#include <UserEnv.h>
#include <TlHelp32.h>
#include "TcConfig.h"
#include "TcDebugLib.h"
#include "TcWin32Ctrl.h"
#pragma comment(lib,"AdvApi32.lib")
#pragma comment(lib,"WtsApi32.lib")
#pragma comment(lib,"UserEnv.lib")

//
// Custom message for service to launch an app into session 1
//
#define SERVICE_CONTROL_TC_MESSAGE_ID1 0x0085

/**
  Launch the process into a different session

  @param szExeFilename   The executable file path to be loaded and run
  @param szParamList     parameter list for the executable, multiple parameters separated with spaces
  @return TRUE on success, FALSE on failure
**/
BOOL
TC_WIN32_API
LaunchAppIntoDifferentSession(
	LPCWSTR szExeFileName,
	LPCWSTR szParamList,
	LPDWORD pdwResult
  )
{
	BOOL                  bTemp;
	BOOL                  bResult;
	DWORD                 dwTemp;
	DWORD                 dwSessionId;
	DWORD                 dwWinLogonSessionId;
	DWORD                 dwWinLogonPid;
	DWORD                 dwCreationFlags;
	HANDLE                hUserToken;
	HANDLE                hUserTokenDup;
	HANDLE                hPToken;
	HANDLE                hProcess;
	HANDLE                hSnap;
	PROCESS_INFORMATION   pi;
	STARTUPINFO           si;
	TOKEN_PRIVILEGES      tp;
	LUID                  luid;
	LPVOID                pEnv;
	PROCESSENTRY32        procEntry;

	bResult = FALSE;
	dwWinLogonPid = 0;
	//
	// Log the client on to the local computer
	//
	dwSessionId = WTSGetActiveConsoleSessionId();
	DEBUG((D_INFO, L"%hs: WTSGetActiveConsoleSessionId returned: 0x%X\n", __FUNCTION__, dwSessionId));
	
	//
	// Find the winlogon process
	//
	hSnap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
	if (hSnap == INVALID_HANDLE_VALUE)
	{
		DEBUG((D_ERROR, L"%hs: CreateToolhelp32Snapshot error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto LaunchAppIntoDifferentSession_Exit_1;
	}

	procEntry.dwSize = sizeof(PROCESSENTRY32);
	if (!Process32First(hSnap, &procEntry))
	{
		DEBUG((D_ERROR, L"%hs: Process32First error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto LaunchAppIntoDifferentSession_Exit_2;
	}

	do
	{
		if (_wcsicmp(procEntry.szExeFile, L"winlogon.exe") == 0)
		{
			//
			// found a winlogon process
			// make sure it's running in the console session
			//
			dwWinLogonSessionId = 0;
			if ( ProcessIdToSessionId(procEntry.th32ProcessID, &dwWinLogonSessionId) 
				&& dwWinLogonSessionId == dwSessionId )
			{
				dwWinLogonPid = procEntry.th32ProcessID;
				DEBUG((D_INFO, L"%hs: winLogonPid = %d\n", __FUNCTION__, dwWinLogonPid));
				break;
			}
		}
	} while (Process32Next(hSnap, &procEntry));

	//
	// query user token
	//
	if (!WTSQueryUserToken(dwSessionId, &hUserToken))
	{
		DEBUG((D_ERROR, L"%hs: WTSQueryUserToken error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto LaunchAppIntoDifferentSession_Exit_2;
	}
  
	ZeroMemory(&pi, sizeof(pi));
	ZeroMemory(&si, sizeof(STARTUPINFO));
  
	si.cb           = sizeof(STARTUPINFO);
	si.lpDesktop    = L"winsta0\\default";
	dwCreationFlags = NORMAL_PRIORITY_CLASS | CREATE_NEW_CONSOLE;

	//
	// approach one:
	// To Hide the process Window
	//
	si.dwFlags      = STARTF_USESHOWWINDOW;
	si.wShowWindow  = SW_HIDE;

	//
	// Or
	// approach two:
	// to hide process window
	// CREATE_NEW_CONSOLE: The new process has a new console, instead of inheriting its parent's console (the default)
	// CREATE_NO_WINDOW: The process is a console application that is being run without a console window
	//
	//dwCreationFlags &= ~CREATE_NEW_CONSOLE;
	//dwCreationFlags |= CREATE_NO_WINDOW;

	hProcess = OpenProcess(MAXIMUM_ALLOWED, FALSE, dwWinLogonPid);
	if (hProcess == NULL)
	{
		DEBUG((D_ERROR, L"%hs: OpenProcess error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto LaunchAppIntoDifferentSession_Exit_3;
	}

	if (!OpenProcessToken(
		hProcess,
		TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY | TOKEN_DUPLICATE | TOKEN_ASSIGN_PRIMARY | TOKEN_ADJUST_SESSIONID | TOKEN_READ | TOKEN_WRITE,
		&hPToken))
	{
		DEBUG((D_ERROR, L"%hs: OpenProcessToken error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto LaunchAppIntoDifferentSession_Exit_4;
	}
	
	if (!LookupPrivilegeValue(NULL, SE_DEBUG_NAME, &luid))
	{
		DEBUG((D_ERROR, L"%hs: LookupPrivilegeValue error: 0x%X\n", __FUNCTION__, GetLastError())); 
		goto LaunchAppIntoDifferentSession_Exit_5;
	}
	
	tp.PrivilegeCount           = 1;
	tp.Privileges[0].Luid       = luid;
	tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

	if (!DuplicateTokenEx(hPToken, MAXIMUM_ALLOWED, NULL, SecurityIdentification, TokenPrimary, &hUserTokenDup))
	{
		DEBUG((D_ERROR, L"%hs: DuplicateTokenEx error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto LaunchAppIntoDifferentSession_Exit_5;
	}

	DEBUG((D_INFO, L"%hs: \n", __FUNCTION__));

	//
	// Adjust Token privilege
	//
	if (!SetTokenInformation(hUserTokenDup, TokenSessionId, (void*)&dwSessionId, sizeof(DWORD)))
	{
		DEBUG((D_WARNING, L"%hs: SetTokenInformation error: 0x%X\n", __FUNCTION__, GetLastError()));
		//goto LaunchAppIntoDifferentSession_Exit_6;
	}

	dwTemp = -1;
	if (!AdjustTokenPrivileges(hUserTokenDup, FALSE, &tp, sizeof(TOKEN_PRIVILEGES), (PTOKEN_PRIVILEGES)NULL, NULL))
	{
		dwTemp = GetLastError();
		DEBUG((D_ERROR, L"%hs: AdjustTokenPrivileges error: 0x%X\n", __FUNCTION__, dwTemp));
		goto LaunchAppIntoDifferentSession_Exit_6;
	}
	if (dwTemp == ERROR_NOT_ALL_ASSIGNED)
	{
		DEBUG((D_WARNING, L"%hs: AdjustTokenPrivileges error: ERROR_NOT_ALL_ASSIGNED\n", __FUNCTION__));
	}

	if (CreateEnvironmentBlock(&pEnv, hUserTokenDup, TRUE))
	{
		dwCreationFlags |= CREATE_UNICODE_ENVIRONMENT;
	}
	else
	{
		DEBUG((D_WARNING, L"%hs: CreateEnvironmentBlock error: 0x%X\n", __FUNCTION__, GetLastError()));
		pEnv = NULL;
	}

	//
	// Launch the process in the client's logon session.
	// and begin impersonation of client
	//
	bTemp = CreateProcessAsUser(
		hUserTokenDup,     // client's access token
		szExeFileName,     // file to execut
		(LPWSTR)szParamList,       // command line
		NULL,              // pointer to process SECURITY_ATTRIBUTES
		NULL,              // pointer to thread SECURITY_ATTRIBUTES
		FALSE,             // handles are not inheritable
		dwCreationFlags,   // creation flags
		pEnv,              // pointer to new environment block 
		NULL,              // name of current directory 
		&si,               // pointer to STARTUPINFO structure
		&pi                // receives information about new process
	);

	while (TRUE)
	{
		if (GetExitCodeProcess(pi.hProcess, &dwTemp))
		{
			if (dwTemp == STILL_ACTIVE)
			{
				Sleep(100);
			}
			else
			{
				break;
			}
		}
	}

	DEBUG((D_INFO, L"%hs: CreateProcessAsUser exit value: 0x%X\n", __FUNCTION__, dwTemp));
	if (pdwResult != NULL)
	{
		*pdwResult = dwTemp;
	}

	if (!bTemp)
	{
		DEBUG((D_ERROR, L"%hs: CreateProcessAsUser error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto LaunchAppIntoDifferentSession_Exit_7;
	}
	
  bResult = TRUE;
  
	//
	// clean up resources
	//
	if (pi.hThread != NULL)
	{
		CloseHandle(pi.hThread);
		pi.hThread = NULL;
	}
	if (pi.hProcess != NULL)
	{
		CloseHandle(pi.hProcess);
		pi.hProcess = NULL;
	}

LaunchAppIntoDifferentSession_Exit_7:
	if (pEnv != NULL)
	{
		DestroyEnvironmentBlock(pEnv);
		pEnv = NULL;
	}

LaunchAppIntoDifferentSession_Exit_6:
	if (hUserTokenDup != NULL)
	{
		CloseHandle(hUserTokenDup);
		hUserTokenDup = NULL;
	}

LaunchAppIntoDifferentSession_Exit_5:
	if (hPToken != NULL)
	{
		CloseHandle(hPToken);
		hPToken = NULL;
	}

LaunchAppIntoDifferentSession_Exit_4:
	if (hProcess != NULL)
	{
		CloseHandle(hProcess);
		hProcess = NULL;
	}
  
LaunchAppIntoDifferentSession_Exit_3:
	if (hUserToken != NULL)
	{
		CloseHandle(hUserToken);
		hUserToken = NULL;
	}

LaunchAppIntoDifferentSession_Exit_2:
	if (hSnap != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hSnap);
		hSnap = NULL;
	}

LaunchAppIntoDifferentSession_Exit_1:
	return bResult;
}