/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/
/** @file

OpenService error code 1053

visual studio 2008, no issue
visual studio 2015, occurs on virtual box and TC unactivated clients
**/

#include "TcConfig.h"
#include "TcDebugLib.h"
#include "TcCtl.h"
#include "TcWin32Ctrl.h"
#include <process.h>
#include <WtsApi32.h>

//
// if undefined, debug to file instead
//
#define SVC_DEBUG_TO_CONSOLE

#define SERVICE_NAME             L"TestService"
//
// command to change screen resolution
//
#define TC_CHGRES_APPNAME_STR    L"C:\\TcChgres.exe"
//#define TC_CHGRES_ARGLIST_STR    L"C:\\TcChgres.exe --tc"
#define TC_CHGRES_ARGLIST_STR    L"C:\\TcChgres.exe --width 1024 --height 768"
//#define TC_CHGRES_ARGLIST_STR    L"C:\\TcChgres.exe --width 1600 --height 1200"
#define TC_CHK_DISPLAY_ATTACHED  L"C:\\TcChgres.exe --test"

#define TC_LOCAL_MSG_STOP_SVC    10
#define TC_SERVER_MSG_SHUTDOWN   TCDMSG_SHUTDOWN   // 11
#define TC_SERVER_MSG_REBOOT     TCDMSG_REBOOT     // 12
//#define TC_LOCAL_MSG_SHUTDOWN  13 /* moved to tcctl.h */

#define SERVICE_ARGV0            SERVICE_NAME
#define SERVICE_ARGV1            L"0"
#define SERVICE_ARGV2            L"eth0"
#define SERVICE_ARGV3            L"12345"

#define TCSVC_THREAD_NUM           2
LPWSTR                  gLpwszArgv[] = { SERVICE_ARGV0, SERVICE_ARGV1, SERVICE_ARGV2, SERVICE_ARGV3 };
SERVICE_STATUS          gServiceStatus;
SERVICE_STATUS_HANDLE   gHandleServiceStatus;
HANDLE                  ghChgresEvent;
HANDLE                  ghStopEvent;
HANDLE                  ghThreads[TCSVC_THREAD_NUM];
BOOL                    gbThreadIsAlive[TCSVC_THREAD_NUM] = { 0 };
UCHAR                   g_exit = 0;
BOOL                    gbIsSomeUserLogon = FALSE;

/**
Report service status to Service Control Manager
**/
VOID
ReportServiceStatus(
	DWORD dwCurrentState,
	DWORD dwWin32ExitCode,
	DWORD dwWaitHint
)
{
	static DWORD dwCheckPoint = 1;

	gServiceStatus.dwCurrentState = dwCurrentState;
	gServiceStatus.dwWin32ExitCode = dwWin32ExitCode;
	gServiceStatus.dwWaitHint = dwWaitHint;

	if (dwCurrentState == SERVICE_START_PENDING)
	{
		gServiceStatus.dwControlsAccepted = 0;
	}
	else
	{
		gServiceStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP |
			SERVICE_ACCEPT_PRESHUTDOWN |
			SERVICE_ACCEPT_SESSIONCHANGE;
	}

	if ((dwCurrentState == SERVICE_RUNNING) ||
		(dwCurrentState == SERVICE_STOPPED))
	{
		gServiceStatus.dwCheckPoint = 0;
	}
	else
	{
		gServiceStatus.dwCheckPoint = dwCheckPoint++;
	}

	SetServiceStatus(gHandleServiceStatus, &gServiceStatus);
}

/**
Service Control Handler function
**/
DWORD
WINAPI
ServiceControlHandlerExtend(
	DWORD dwCtrl,
	DWORD dwEventType,
	LPVOID lpEventData,
	LPVOID lpContext
)
{
	DEBUG((D_INFO, L"%hs:\n", __FUNCTION__));
	DEBUG((D_INFO, L"%hs: dwCtrl = 0x%X\n", __FUNCTION__, dwCtrl));

	switch (dwCtrl)
	{
	case SERVICE_CONTROL_STOP:
		ReportServiceStatus(SERVICE_STOP_PENDING, NO_ERROR, 0);
		g_exit = TC_LOCAL_MSG_STOP_SVC;
		if (!SetEvent(ghStopEvent))
		{
			DEBUG((D_WARNING, L"SERVICE_CONTROL_STOP: fail to set ghStopEvent\n"));
		}

		DEBUG((D_INFO, L"SERVICE_CONTROL_STOP: g_exit = %d\n", g_exit));
		return 0;

	case SERVICE_CONTROL_INTERROGATE:
		DEBUG((D_INFO, L"SERVICE_CONTROL_INTERROGATE: g_exit = %d\n", g_exit));
		break;

	case SERVICE_CONTROL_PRESHUTDOWN:
		ReportServiceStatus(SERVICE_STOP_PENDING, NO_ERROR, 0);
		g_exit = TC_LOCAL_MSG_SHUTDOWN;

		DEBUG((D_INFO, L"SERVICE_CONTROL_PRESHUTDOWN: g_exit = %d\n", g_exit));
		return 0;

	case SERVICE_CONTROL_SESSIONCHANGE:
		DEBUG((D_INFO, L"SERVICE_CONTROL_SESSIONCHANGE: dwEventType = 0x%X\n", dwEventType));
		switch (dwEventType)
		{
		case WTS_SESSION_LOGON:
			DEBUG((D_INFO, L"%hs: WTS_SESSION_LOGON\n", __FUNCTION__));
			break;

		case WTS_SESSION_UNLOCK:
			DEBUG((D_INFO, L"%hs: WTS_SESSION_UNLOCK\n", __FUNCTION__));
			break;

		default:
			DEBUG((D_INFO, L"SERVICE_CONTROL_SESSIONCHANGE: unaddressed[0x%X]\n", dwEventType));
			break;
		}
		break;

	default:
		DEBUG((D_INFO, L"%hs: g_exit = %d\n", __FUNCTION__, g_exit));
		break;
	}

	ReportServiceStatus(gServiceStatus.dwCurrentState, NO_ERROR, 0);
	return 0;
}


/**
  Check if any user logged on to System

  @param none
  @return TRUE if there is a user logged on to system
**/
BOOL
TC_WIN32_API
TcCheckUserLogon(
	VOID
)
{
	BOOL  bRet;
	DWORD dwIndex;
	DWORD dwCnt;
	DWORD dwTemp;
	PWTS_SESSION_INFO pWtsSessionInfo;

	pWtsSessionInfo = NULL;
	bRet = WTSEnumerateSessions(
		WTS_CURRENT_SERVER_HANDLE, // the RD Session Host server on which your application is running
		0,                         // reserved
		1,                         // API version
		&pWtsSessionInfo,          // pointer to WTS_SESSION_INFO structure instance
		&dwCnt
	);

	if (!bRet)
	{
		DEBUG((D_ERROR, L"%hs: WTSEnumerateSessions error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto exit;
	}

	bRet = FALSE;
	dwTemp = -1;
	DEBUG((D_INFO, L"%hs: WTSEnumerateSessions returned: dwCnt = 0x%X\n", __FUNCTION__, dwCnt));
	for (dwIndex = 0; dwIndex < dwCnt; dwIndex++)
	{
		DEBUG((D_INFO, L"%hs: WTSEnumerateSessions 0x%X: SessionId       = 0x%X\n", __FUNCTION__, dwIndex, pWtsSessionInfo[dwIndex].SessionId));
		DEBUG((D_INFO, L"%hs: WTSEnumerateSessions 0x%X: pWinStationName = %ls\n", __FUNCTION__, dwIndex, pWtsSessionInfo[dwIndex].pWinStationName));
		DEBUG((D_INFO, L"%hs: WTSEnumerateSessions 0x%X: State           = 0x%X\n", __FUNCTION__, dwIndex, pWtsSessionInfo[dwIndex].State));

		if (pWtsSessionInfo[dwIndex].State == WTSActive)
		{
			bRet = TRUE;
			dwTemp = dwIndex;
		}
	}

	if (dwTemp < dwCnt && dwTemp >= 0)
	{
		DEBUG((D_INFO, L"%hs: Active session Index: 0x%X\n", __FUNCTION__, dwTemp));
		DEBUG((D_INFO, L"%hs: Active session Id   : 0x%X\n", __FUNCTION__, pWtsSessionInfo[dwTemp].SessionId));
		DEBUG((D_INFO, L"%hs: Active session name : 0x%ls\n", __FUNCTION__, pWtsSessionInfo[dwTemp].pWinStationName));
		DEBUG((D_INFO, L"%hs: Active session state: 0x%X\n", __FUNCTION__, pWtsSessionInfo[dwTemp].State));
	}

exit:
	if (pWtsSessionInfo != NULL)
	{
		WTSFreeMemory(pWtsSessionInfo);
		pWtsSessionInfo = NULL;
	}

	return bRet;
}



/**
  The worker thread to change screen resolution
**/
unsigned
WINAPI
TcCheckUserLogonWorkerThread(
	LPVOID *pParm
)
{
	unsigned ret = 0;
	DWORD dwStatus;
	DWORD dwResult;

	while (!gbIsSomeUserLogon)
	{
		dwStatus = WaitForSingleObject(ghStopEvent, 0);
		if (dwStatus == WAIT_OBJECT_0)
		{
			//
			// reset the event to non-signaled state
			//
			if (!ResetEvent(ghStopEvent))
			{
				DEBUG((D_WARNING, L"%hs: fail to reset ghStopEvent\n", __FUNCTION__));
			}
			goto exit;
		}

		if (TcCheckUserLogon())
		{
			//
			// query if display is attached
			//
			if (LaunchAppIntoDifferentSession(TC_CHGRES_APPNAME_STR, TC_CHK_DISPLAY_ATTACHED, &dwResult))
			{
				DEBUG((D_INFO, L"%hs: LaunchAppIntoDifferentSession exit value: 0x%X\n", __FUNCTION__, dwResult));
			}
			else
			{
				DEBUG((D_ERROR, L"%hs: LaunchAppIntoDifferentSession fail\n", __FUNCTION__));
			}

			//
			// 0 indicates display is attached
			//
			if (dwResult == 0)
			{
				//
				// set the logon status
				//
				gbIsSomeUserLogon = TRUE;

				//
				// send a message to trigger the resolution changes task
				// and exit the loop
				//
				if (SetEvent(ghChgresEvent))
				{
					DEBUG((D_INFO, L"%hs: set change res event\n", __FUNCTION__));
					break;
				}
				else
				{
					DEBUG((D_WARNING, L"%hs: fail to set change res event\n", __FUNCTION__));
				}
			}
			else
			{
				DEBUG((D_INFO, L"%hs: Seems display was not attached!\n", __FUNCTION__));
			}
		}
		else
		{
			DEBUG((D_INFO, L"%hs: no user was logged on!\n", __FUNCTION__));
		}

		Sleep(1000);
	}

exit:
	_endthreadex(0);

	return ret;
}



/**
  The worker thread to change screen resolution
**/
unsigned
WINAPI
TcChangeResolutionWorkerThread(
	LPVOID pParms
)
{
	UINT ret = 0;
	DWORD dwStatus;
	DWORD dwResult;
	dwResult = 0;
	while (TRUE)
	{
		dwStatus = WaitForSingleObject(ghStopEvent, 0);
		if (dwStatus == WAIT_OBJECT_0)
		{
			//
			// reset the event to non-signaled state
			//
			if (!ResetEvent(ghStopEvent))
			{
				DEBUG((D_WARNING, L"%hs: fail to reset ghStopEvent\n", __FUNCTION__));
			}
			goto exit;
		}

		dwStatus = WaitForSingleObject(ghChgresEvent, 0);
		if (dwStatus == WAIT_OBJECT_0)
		{
			DEBUG((D_INFO, L"%hs: receive event gHChgresEvent\n", __FUNCTION__));
			if (!gbIsSomeUserLogon)
			{
				DEBUG((D_WARNING, L"%hs: nobody logon to the system, gHChgresEvent would be discarded\n", __FUNCTION__));
			}
			else
			{
				//
				// no guarantee that the screen resolution is changed successfully
				//
				if (LaunchAppIntoDifferentSession(TC_CHGRES_APPNAME_STR, TC_CHGRES_ARGLIST_STR, &dwResult))
				{
					DEBUG((D_INFO, L"%hs: LaunchAppIntoDifferentSession exit value: 0x%X\n", __FUNCTION__, dwResult));
				}
				else
				{
					DEBUG((D_ERROR, L"%hs: LaunchAppIntoDifferentSession error: 0x%X\n", __FUNCTION__, dwResult));
				}
			}

			//
			// reset the event to non-signaled state
			//
			if (!ResetEvent(ghChgresEvent))
			{
				DEBUG((D_WARNING, L"%hs: fail to reset ghChgresEvent\n", __FUNCTION__));
			}
		}

		Sleep(1000);
	}

exit:
	_endthreadex(dwResult);

	return ret;
}


/**
Service initialization
**/
VOID
WINAPI
ServiceInitialization(
	DWORD dwArgc,
	LPWSTR *lpwszArgv
)
{
	DWORD    dwResult;
	DWORD    dwIndex;
	BOOL     bLoopTag;

	//
	// report status "Running" to SCM
	//
	ReportServiceStatus(SERVICE_RUNNING, NO_ERROR, 0);


	for (dwIndex = 0; dwIndex < TCSVC_THREAD_NUM; dwIndex++)
	{
		ghThreads[dwIndex] = NULL;
	}

	ghChgresEvent = CreateEvent(
		NULL,               // default security attributes
		TRUE,               // manual-reset event
		FALSE,              // initial state is nonsignaled
		L"ChgresEvent"      // object name
	);
	if (ghChgresEvent == NULL)
	{
		DEBUG((D_ERROR, L"%hs: CreateEvent for ghChgresEvent: 0x%X\n", __FUNCTION__, GetLastError()));
	}

	ghStopEvent = CreateEvent(
		NULL,               // default security attributes
		TRUE,               // manual-reset event
		FALSE,              // initial state is nonsignaled
		L"StopEvent"        // object name
	);
	if (ghStopEvent == NULL)
	{
		DEBUG((D_ERROR, L"%hs: CreateEvent for ghStopEvent: 0x%X\n", __FUNCTION__, GetLastError()));
	}

	ghThreads[0] = (HANDLE)_beginthreadex(NULL, 0, &TcCheckUserLogonWorkerThread, NULL, 0, NULL);
	if (ghThreads[0] == NULL)
	{
		DEBUG((D_ERROR, L"%hs: TcCheckUserLogonWorkerThread error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto exit0;
	}

	ghThreads[1] = (HANDLE)_beginthreadex(NULL, 0, &TcChangeResolutionWorkerThread, NULL, 0, NULL);
	if (ghThreads[1] == NULL)
	{
		DEBUG((D_ERROR, L"%hs: TcChangeResolutionWorkerThread error: 0x%X\n", __FUNCTION__, GetLastError()));
		goto exit1;
	}

	gbThreadIsAlive[0] = TRUE;
	gbThreadIsAlive[1] = TRUE;
	while (TRUE)
	{
		dwResult = WaitForSingleObject(ghThreads[0], 0);
		if (WAIT_OBJECT_0 == dwResult)
		{
			gbThreadIsAlive[0] = FALSE;
		}

		dwResult = WaitForSingleObject(ghThreads[1], 0);
		if (WAIT_OBJECT_0 == dwResult)
		{
			gbThreadIsAlive[1] = FALSE;
		}

		if (!gbThreadIsAlive[0] && !gbThreadIsAlive[1])
		{
			break;
		}

		//
		// wait for a moment
		//
		//DEBUG((D_INFO, L"%hs: Wait for 1 second\n", __FUNCTION__));
		Sleep(1000 * 1);
	}

	//
	// thirdly,
	// check reason that message loop exits,
	// and take some actions later...
	//
	DEBUG((D_INFO, L"%hs: ServiceWorkerThread returned, g_exit = %d\n", __FUNCTION__, g_exit));

exit2:
	if (ghThreads[1] != NULL)
	{
		CloseHandle(ghThreads[1]);
		ghThreads[1] = NULL;
	}

exit1:
	if (ghThreads[0] != NULL)
	{
		CloseHandle(ghThreads[0]);
		ghThreads[0] = NULL;
	}

exit0:
	if (ghStopEvent != NULL)
	{
		CloseHandle(ghStopEvent);
		ghStopEvent = NULL;
	}

	if (ghChgresEvent != NULL)
	{
		CloseHandle(ghChgresEvent);
		ghChgresEvent = NULL;
	}

	ReportServiceStatus(SERVICE_STOPPED, NO_ERROR, 0);
}


/**
Service entry point function
**/
VOID
WINAPI
ServiceMain(
	DWORD dwArgc,
	LPWSTR *lpwszArgv
)
{
	DWORD   dwNewArgc;
	LPWSTR  *lpwszNewArgv;
	DWORD   i;

	dwNewArgc = dwArgc;
	lpwszNewArgv = lpwszArgv;

	DEBUG((D_INFO, L"%hs: %d\n", __FUNCTION__, dwArgc));

	//
	// if no cmdline arguments,
	// then use default ones
	//
	if (dwNewArgc < 2)
	{
		dwNewArgc = 4;
		lpwszNewArgv = gLpwszArgv;
	}

	//
	// print arguments
	//
	for (i = 0; i < dwNewArgc; i++)
	{
		DEBUG((D_INFO, L"%hs: %ls\n", __FUNCTION__, lpwszNewArgv[i]));
	}

	gHandleServiceStatus = RegisterServiceCtrlHandlerEx(
		SERVICE_NAME,
		ServiceControlHandlerExtend,
		NULL
	);

	if (!gHandleServiceStatus)
	{
		DEBUG((D_ERROR, L"RegisterServiceCtrlHandler failed with %d\n", GetLastError()));
		return;
	}

	gServiceStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	gServiceStatus.dwServiceSpecificExitCode = 0;

	ReportServiceStatus(SERVICE_START_PENDING, NO_ERROR, 3000);

	//
	// invoke service initialization function
	//
	ServiceInitialization(dwNewArgc, lpwszNewArgv);
}

//
// Entry for service process
//
void
__cdecl
wmain(
	int argc,
	wchar_t *argv[]
)
{
	SERVICE_TABLE_ENTRY DispatchTable[] =
	{
		{ SERVICE_NAME, (LPSERVICE_MAIN_FUNCTION)ServiceMain },
		{ NULL, NULL }
	};

#ifdef SVC_DEBUG_TO_CONSOLE
	DEBUG_INIT(TO_CONSOLE, NULL);
#else
	DEBUG_INIT(TO_FILE, L"C:\\TestService_Debug.txt");
#endif
	PRINT_INIT(TO_FILE, L"C:\\TestService.txt");

	DEBUG((D_INFO, L"Entry of service\n"));
	PRINT(L"Entry of service\n");

	//
	// this function will never return unless service is stopped
	//
	if (!StartServiceCtrlDispatcher(DispatchTable))
	{
		DEBUG((D_ERROR, L"StartServiceCtrlDispatcher failed with %d\n", GetLastError()));
	}

	DEBUG((D_INFO, L"Exit of service\n"));
	PRINT(L"Exit of service\n");

	DEBUG_FINISH();
	PRINT_FINISH();
}


