/******************************************************************************
Copyright 2012 - 2020 Intel Corporation

For licensing information, see the file 'LICENSE' in the root folder of
this software module.
******************************************************************************/


#include <stdio.h>
#include "conf.h"
#include "tcd.h"
#include "bmap.h"


int init_bulkmap(BULKMAP *bmp)
{
	int k;

	if (bmp == NULL)
	{
		fprintf(stderr, "init_bulkmap: NULL bmp\n");
		return -1;
	}

	bmp->bits = 0;
	for (k=0; k<MAPSZ; k++)
		bmp->map[k] = 0;

	return 0;
}


int set_bulkmap(BULKMAP *bmp, uint32 slot)
{
	DBG("set_bulkmap: %ld\n", slot);

	if (bmp == NULL)
	{
		fprintf(stderr, "set_bulkmap: NULL bmp\n");
		return -1;
	}

	if (slot >= TOTALSZ)
	{
		fprintf(stderr, "set_bulkmap: slot %d too big\n", slot);
		return -1;
	}

	/*
	bmp->map[slot >> 6] |= (uint64)1 << (slot & 0x000000000000003fULL);
	bmp->bits ++;
	*/

	if (! (bmp->map[slot >> 6] & ((uint64)1 << (slot & 0x000000000000003fULL))) )
	{
		bmp->map[slot >> 6] |= (uint64)1 << (slot & 0x000000000000003fULL);
		bmp->bits ++;
	}

	return 0;
}


int clear_bulkmap(BULKMAP *bmp, uint32 slot)
{
	DBG("clear_bulkmap: %ld\n", slot);

	if (bmp == NULL)
	{
		fprintf(stderr, "clear_bulkmap: NULL bmp\n");
		return -1;
	}

	if (slot >= TOTALSZ)
	{
		fprintf(stderr, "clear_bulkmap: slot %d too big\n", slot);
		return -1;
	}

	/*
	bmp->map[slot >> 6] &= ~((uint64)1 << (slot & 0x000000000000003fULL));
	bmp->bits --;
	*/

	if (bmp->map[slot >> 6] & ((uint64)1 << (slot & 0x000000000000003fULL)));
	{
		bmp->map[slot >> 6] &= ~((uint64)1 << (slot & 0x000000000000003fULL));
		bmp->bits --;
	}

	return 0;
}


int isset_bulkmap(BULKMAP *bmp, uint32 slot)
{
	DBG("isset_bulkmap: %ld\n", slot);

	if (bmp == NULL)
	{
		fprintf(stderr, "isset_bulkmap: NULL bmp\n");
		return -1;
	}

	if (slot >= TOTALSZ)
	{
		fprintf(stderr, "isset_bulkmap: slot %d too big\n", slot);
		return -1;
	}

	if ( (bmp->map[slot >> 6] & ((uint64)1 << (slot & 0x000000000000003fULL))) )
		return 1;
	else
		return 0;
}

int isclear_bulkmap(BULKMAP *bmp, uint32 slot)
{
	DBG("isset_bulkmap: %ld\n", slot);

	if (bmp == NULL)
	{
		fprintf(stderr, "isset_bulkmap: NULL bmp\n");
		return -1;
	}

	if (slot >= TOTALSZ)
	{
		fprintf(stderr, "isset_bulkmap: slot %d too big\n", slot);
		return -1;
	}

	if ( !(bmp->map[slot >> 6] & ((uint64)1 << (slot & 0x000000000000003fULL))) )
		return 1;
	else
		return 0;
}




// bmp <- bmp .and. bmp0
int and_bulkmap(BULKMAP *bmp, BULKMAP *bmp0)
{
	int k;

	DBG("and_bulkmap: \n");

	if (bmp == NULL || bmp0 == NULL)
	{
		fprintf(stderr, "and_bulkmap: NULL bmp\n");
		return -1;
	}


	bmp->bits = 0;
	for (k=0; k<MAPSZ; k++)
		bmp->map[k] &= bmp0->map[k];

	return 0;
}


// bmp <- bmp .or. bmp0
int or_bulkmap(BULKMAP *bmp, BULKMAP *bmp0)
{
	int k;

	DBG("and_bulkmap: \n");

	if (bmp == NULL || bmp0 == NULL)
	{
		fprintf(stderr, "and_bulkmap: NULL bmp\n");
		return -1;
	}


	bmp->bits = 0;
	for (k=0; k<MAPSZ; k++)
		bmp->map[k] |= bmp0->map[k];

	return 0;
}



// there are two types of full-bulkmap: 
// 1, a bulkmap with TOTALSZ of slots
// 2, the last bulkmap, w/o empty slots.  
int isfull_bulkmap(BULKMAP *bmp, uint32 hasnext)
{
	if (bmp == NULL)
	{
		fprintf(stderr, "isfull_bulkmap: NULL bmp\n");
		return -1;
	}

	// a small issue here: we should check while hasnext=0, whether all bits ahead of hasnext is 1. 
	return (bmp->bits == TOTALSZ || hasnext==0);
}

int print_bulkmap(BULKMAP *bmp)
{
	int k, j;

	if (bmp == NULL)
	{
		fprintf(stderr, "print_bulkmap: NULL bmp\n");
		return -1;
	}

	printf("bulkmap (total bits = %d): \n", bmp->bits);

	printf("   ");
	for (k=0; k<64; k++)
		if (k%10 == 0)
			printf("%d", k/10);
		else
			printf(" ");
	printf("\n");

	printf("   ");
	for (k=0; k<64; k++)
		printf("%d", k%10);
	printf("\n");

	printf("   ");
	for (k=0; k<64; k++)
		printf("=");
	printf("\n");

	for (k=0; k<MAPSZ; k++)
	{
		printf("%2d ", k);
		j = 0;
		uint64 d = 1;
		do 
		{
			if (bmp->map[k] & d)
				printf("*");
			else
				printf(".");
			d <<= 1;
		}while (++j<64);
		printf("\n");
	}
	printf("\n");

	return 0;
}

/*
main()
{
	BULKMAP bmp, bmp0;

	init_bulkmap(&bmp);
	// print_bulkmap(&bmp);

	set_bulkmap(&bmp, 0);
	set_bulkmap(&bmp, 1);
	set_bulkmap(&bmp, 2);
	print_bulkmap(&bmp);


	init_bulkmap(&bmp0);
	set_bulkmap(&bmp0, 0);
	set_bulkmap(&bmp0, 1);
	set_bulkmap(&bmp0, 3);
	print_bulkmap(&bmp0);

	and_bulkmap(&bmp, &bmp0);
	print_bulkmap(&bmp);

	return; 

	set_bulkmap(&bmp, 31);
	print_bulkmap(&bmp);

	set_bulkmap(&bmp, 32);
	print_bulkmap(&bmp);

	set_bulkmap(&bmp, 33);
	print_bulkmap(&bmp);

	set_bulkmap(&bmp, 1023);
	print_bulkmap(&bmp);

	set_bulkmap(&bmp, 1024);
	print_bulkmap(&bmp);


	set_bulkmap(&bmp, 4094);
	print_bulkmap(&bmp);
	set_bulkmap(&bmp, 4095);
	print_bulkmap(&bmp);
	set_bulkmap(&bmp, 4096);
	print_bulkmap(&bmp);


	clear_bulkmap(&bmp, 0);
	print_bulkmap(&bmp);
	clear_bulkmap(&bmp, 1);
	print_bulkmap(&bmp);
	clear_bulkmap(&bmp, 2);
	print_bulkmap(&bmp);

	clear_bulkmap(&bmp, 31);
	print_bulkmap(&bmp);

	clear_bulkmap(&bmp, 32);
	print_bulkmap(&bmp);

	clear_bulkmap(&bmp, 33);
	print_bulkmap(&bmp);

	clear_bulkmap(&bmp, 1023);
	print_bulkmap(&bmp);
	clear_bulkmap(&bmp, 1024);
	print_bulkmap(&bmp);

	clear_bulkmap(&bmp, 4094);
	print_bulkmap(&bmp);
	clear_bulkmap(&bmp, 4095);
	print_bulkmap(&bmp);
	clear_bulkmap(&bmp, 4096);
	print_bulkmap(&bmp);
	
}


*/


