/******************************************************************************
Copyright 2012 - 2020 Intel Corporation

For licensing information, see the file 'LICENSE' in the root folder of
this software module.
******************************************************************************/

#include <assert.h>

#define CFG_FILE_PARSER_ANSI
#include "cfgsmalllib.h"

#include <stdio.h>

#define CFG_FILETYPE                 FILE*
#define cfg_fopenwrite(filename,fp)  ((*(fp) = fopen((filename),"wb")) != NULL)
#define cfg_fopenread(filename,fp)   ((*(fp) = fopen((filename),"rb")) != NULL)
#define cfg_fclose(fp)               (fclose(*(fp)) == 0)
#define cfg_rename(source,dest)      (rename((source), (dest)) == 0)
#define cfg_remove(filename)         (remove(filename) == 0)
#define cfg_fgets(buffer,size,fp)    (fgets((buffer),(size),*(fp)) != NULL)
#define cfg_fputs(buffer,fp)         (fputs((buffer),*(fp)) >= 0)


#define CFG_FILEPOS_TYPE              fpos_t
#define cfg_fgetpos(fp,pos)           (fgetpos(*(fp), (pos)) == 0)
#define cfg_fsetpos(fp,pos)           (fsetpos(*(fp), (pos)) == 0)

#define cfg_float2str(str,val)        sprintf((str),"%f",(val))
#define cfg_str2float(str)            (float)strtod((str),NULL)

#if (!defined(_UNICODE) && !defined(UNICODE)) || defined(CFG_FILE_PARSER_ANSI)
  #include <ctype.h>
  #include <string.h>
  #include <stdlib.h>
  #define __T(x)    x
  #define _tremove  remove
  #define _trename  rename
  #define _totupper toupper
  #define _tcscat   strcat
  #define _tcschr   strchr
  #define _tfgets   fgets
  #define _tfputs   fputs  
  #define _tcslen   strlen
  #define _tcscmp   strcmp
  #define _tcscpy   strcpy
  #define _tcsicmp  stricmp
  #define _tcsncmp  strncmp
  #define _tcsnicmp cfg_strnicmp
  #define _tcsrchr  strrchr
  #define _tcstol   strtol
  #define _tcstod   strtod
  #define _stprintf sprintf
  #define _tfopen   fopen
#endif

#if defined(_MSC_VER)
  #pragma warning(disable: 4996)
#endif

#if !defined CFG_LINE_BREAK
  #define CFG_LINE_BREAK    __T("\n")
#endif

#if !defined _countof
  #define _countof(array) (sizeof(array) / sizeof(array[0]))
#endif

typedef enum {
  CFG_QUOTE_NONE,
  // add escape
  CFG_QUOTE_ENQUOTE,
  // remove escape
  CFG_QUOTE_DEQUOTE,
} cfg_enum_quote_option;


/**
 * compare two array with at most first n characters
 */
static int cfg_strnicmp(const TCHAR *left, const TCHAR *right, size_t n)
{
  int ch1, ch2;

  while (n > 0 && (*left || *right)) {
    ch1 = *left++;
    ch2 = *right++;
    if ('A' <= ch1 && ch1 <= 'Z') {
      ch1 += ('a' - 'A');
    }
    if ('A' <= ch2 && ch2 <= 'Z') {
      ch2 += ('a' - 'A');
    }
    if (ch1 != ch2) {
      return ch1 - ch2;
    }
    n--;
  }
  return 0;
}

/**
 * skip leading white spaces
 * @return a pointer to '\0' or the first non-white-spaces char
 */
static TCHAR *skip_leading_whitespaces(const TCHAR *head)
{
  assert(head != NULL);
  //
  // ' ' and value less than ' ' are all considered as white spaces
  //
  while (*head != '\0' && *head <= ' ') {
    head++;
  }
  return (TCHAR *)head;
}

/**
 * skip white spaces
 * @return a pointer to '\0' char or last white spaces char counting backward in the string.
 * <==
 */
static TCHAR *skip_trailing_whitespaces_internal(const TCHAR *head, const TCHAR *tail)
{
  assert(head != NULL);
  assert(tail != NULL);
  while (tail > head && *(tail-1) <= ' ') {
    tail--;
  }
  return (TCHAR *)tail;
}

/**
 * truncate trailing white spaces in the string
 */
static TCHAR *skip_trailing_whitespaces(TCHAR *head)
{
  TCHAR *p;

  p = skip_trailing_whitespaces_internal(head, _tcschr(head, '\0'));
  assert(p != NULL);
  *p = '\0';
  return head;
}

/**
 * save a copy of the src to dst
 * the quotation indicator have 3 values:
 * CFG_QUOTE_NONE: raw copy
 * CFG_QUOTE_ENQUOTE: add escaping handlling and leading/trailing double quotations
 * CFG_QUOTE_DEQUOTE: remove internal escaped chars (\ and ")
 */
static TCHAR *quote_strncpy(TCHAR *dst, const TCHAR *src, size_t dsize, cfg_enum_quote_option quotation)
{
  size_t s;
  size_t d;

  assert(dsize > 0);
  assert(dst <= src || dst >= src + dsize);

  //
  // leading quote, trailing quote and one terminal char takes 3 bytes
  //
  if (quotation == CFG_QUOTE_ENQUOTE && dsize < sizeof('"') + sizeof('"') + sizeof('\0')) {
    quotation = CFG_QUOTE_NONE;
  }

  switch (quotation) {
    //
    // raw strcpy
    //
    case CFG_QUOTE_NONE:
      //
      // first at most copy (dsize - 1) valid chars
      //
      for (d = 0; d < dsize - 1 && src[d] != '\0'; d++) {
        dst[d] = src[d];
      }
      //
      // followed by a string terminator
      //
      dst[d] = '\0';
      break;

    //
    // add leading ", trailing " and possible \ for special char escaping
    //
    case CFG_QUOTE_ENQUOTE:
      //
      // add leading quote
      //
      d = 0;
      dst[d] = '"';

      //
      // dst should count from index 1 to buffer size minux 2 (2 bytes to be occupied by trailing quote and string teminator)
      //
      for (s = 0, d = 1; src[s] != '\0' && d < dsize - sizeof('"') - sizeof('\0'); s++, d++) {
        //
        // when met a quote in the src text, added '\\' before it
        //
        if (src[s] == '"') {
          //
          // no space to store the escape character, the quote and the terminator
          //
          if (d >= dsize - sizeof('\\') - sizeof('"') - sizeof('\0')) {
            break;
          } else {
            dst[d] = '\\';
            d++;
          }
        }
        dst[d] = src[s];
      }
      //
      // add trailing quote and string teminator
      //
      dst[d] = '"';
      d++;
      dst[d] = '\0';
      break;

    //
    // when met \"(or ""), skip leading \ (or ")
    //
    case CFG_QUOTE_DEQUOTE:
      for (s = 0, d = 0; src[s] != '\0' && d < dsize - sizeof('\0'); s++, d++) {
        if ((src[s] == '"' || src[s] == '\\') && src[s + 1] == '"') {
          s++;
        }
        dst[d] = src[s];
      }
      dst[d] = '\0';
      break;

    default:
      assert(0);
  }
  return dst;
}

/**
 * remove possible trailing commment
 * remove possible trailing white spaces
 * remove possible double quotations surrounding the text
 * @result the result contain two part: a text and a qutation value,
 *         the returned quotation values could be CFG_QUOTE_NONE: indicates the input have NO surrounding double qutations
 *         the returned quotation values could be CFG_QUOTE_DEQUOTE: indicates the input have surrounding double qutations and have been removed, one might need to take more actions to address possible escaping characters within the retured text
 * @example
 *   input: "123 456"   # comment
 *   ouput: 123 456
 */
static TCHAR *remove_comment_trailingwhitespaces_surroundingquotations(TCHAR *head, cfg_enum_quote_option *quotation)
{
  int inside_string;
  TCHAR *p;

  assert(head != NULL);
  assert(quotation != NULL);

  //
  // Remove a trailing comment
  //
  inside_string = 0;
  for (p = head; *p != '\0'; p++) {
    //
    // when met ';' or '#' outside of a string, stop scanning
    //
    if (*p == ';' || *p == '#' ) {
      if (!inside_string) {
        break;
      }
    }

    if (*p == '"') {
      if (*(p + 1) == '"') {
        //
        // skip both quotation ""
        //
        p++;
      } else {
        //
        // single quote, to toggle inside_string
        //
        inside_string = !inside_string;
      }
    } else if (*p == '\\' && *(p + 1) == '"') {
      //
      // skip escaped quotation \"
      //
      p++;
    }
  }

  //
  // scan is over, truncate trailing comments
  //
  assert(*p == '\0' || *p == ';' || *p == '#');
  *p = '\0';

  //
  // truncate trailing white spaces
  //
  skip_trailing_whitespaces(head);

  //
  // Remove quotations surrounding a text
  //
  *quotation = CFG_QUOTE_NONE;
  p = _tcschr(head, '\0');
  if (p != NULL) {
    p--;
    if (*head == '"' && *p == '"') {
      head++;
      *p = '\0';
      //
      // the final text might contain escape chars like \" (or "")
      // so marked the result to "DEQUOTE" which indicates it might need more manipulation to make it clean
      //
      *quotation = CFG_QUOTE_DEQUOTE;
    }
  }
  return head;
}

/**
 * get a section by section index
 * or given a section, get a key by key index; 
 * or not given a section, get a key before the possible first section by key index
 * or get a key value by given section name and key name
 * @return 0 on fail, 1 on success
 */
static int get_section_or_key_or_value(CFG_FILETYPE *fp, const TCHAR *section, const TCHAR *key, int section_index, int key_index, TCHAR *buf, int buf_size)
{
  int len;
  int index;
  TCHAR *p;
  TCHAR *q;
  TCHAR local_buf[CFG_BUFSZ];
  cfg_enum_quote_option quotation;

  assert(fp != NULL);

  //
  // section processing:
  // scan the file 1 line at a time until a section is matched or end-of-file encountered.
  // a section is matched means: either an input non-null section was find within the file,
  //                             or the input section is null but we have section_index sections skipped before the current section
  //
  len = (section != NULL) ? (int)_tcslen(section) : 0;
  if (len > 0 || section_index >= 0) {
    index = -1;
    while (1) {
      if (!cfg_fgets(local_buf, CFG_BUFSZ, fp)) {
        //
        // end of file
        //
        return 0;
      }

      //
      // check leading '['
      //
      p = skip_leading_whitespaces(local_buf);
      if (*p != '[') {
        continue;
      }

      //
      // check trailing ']'
      //
      if ((q = _tcschr(p, ']')) == NULL) {
        continue;
      }

      //
      // now it is a legal section
      // plus index by 1
      //
      index++;
      //
      // if index match, just return the section name
      //
      if (index == section_index) {
        break;
      }

      //
      // get section names within '[' and ']' and check if they match
      // the match here have two possible cases:
      //   1. the input string is NULL or empty, in this case the section is empty
      //   2. the input string is not NULL/empty and the is a real string match case
      //
      if ((int)(q - p - 1) == len && _tcsnicmp(section, p + 1, len) == 0) {
        break;
      }
    }

    //
    // the while loop above has only two exit: the two breaks in the if-statement.
    // Now the scan did not reach end-of-file and we have 3 possible results
    // the section matches,
    // or the index matches,
    // or both match,
    // if neither matches it will reach end-of-file in the while loop above and return.

    //
    // case section_index >= 0:
    // address index match/mismatch cases and stop all actions and return
    //
    if (section_index >= 0) {
      //
      // index matches:
      // save the section name scanned to buffer and return
      //
      if (index == section_index) {
        assert(q != NULL);
        assert(*q == ']');
        *q = '\0';
        quote_strncpy(buf, p + 1, buf_size, CFG_QUOTE_NONE);
        return 1;
      }
      //
      // index does not match:
      // file scanning is over but unable to find so many sections
      // note even if section is NULL the empty section might match NULL section ([] match NULL), and find a way coming out of while loop
      //
      return 0;
    }
  }

  //
  // case len > 0:
  // i.e. section is NOT NULL and section matches
  // search the entry uplon leaving this section's area.
  //
  assert(key != NULL || key_index >= 0);
  len = (key != NULL) ? (int)_tcslen(key) : 0;
  index = -1;
  while (1) {
    if (!cfg_fgets(local_buf, CFG_BUFSZ, fp)) {
      return 0;
    }

    p = skip_leading_whitespaces(local_buf);
    if (*p == '[') {
      return 0;
    } else if (*p == ';' || *p == '#') {
      //
      // this is a comment
      //
      continue;
    }

    //
    // find equal sign
    //
    q = _tcschr(p, '=');
    if (q == NULL) {
      q = _tcschr(p, ':');
    }
    if (q == NULL) {
      //
      // unable to find the equal sign and colon
      //
      continue;
    }

    //
    // this is a key value pair
    //
    index++;

    //
    // check if index matches
    //
    if (index == key_index) {
      break;
    }

    //
    // check if key matches
    //
    if ( (int)(skip_trailing_whitespaces_internal(p, q) - p) == len && _tcsnicmp(p, key, len) == 0) {
      break;
    }
  }

  //
  // case key_index >= 0:
  // address index match/mismatch cases and stop all actions and return
  //
  if (key_index >= 0) {
      //
      // index matches:
      // save the key name scanned to buffer and return
      //
    if (index == key_index) {
      assert(q != NULL);
      assert(*q == '=' || *q == ':');
      *q = '\0';
      skip_trailing_whitespaces(p);
      quote_strncpy(buf, p, buf_size, CFG_QUOTE_NONE);
      return 1;
    }
    //
    // index does not match:
    // section/file scanning is over but unable to find so many keys
    // note even if key is NULL the empty key might match NULL section ([] match NULL), and find a way coming out of while loop
    //
    return 0;
  }

  //
  // case len > 0:
  // i.e. key is NOT NULL and key matches
  // save a copy of the value of the key.
  // here only cover at most buf_size bytes
  //
  assert(q != NULL);
  assert(*q == '=' || *q == ':');
  //
  // skip possible leading whate spaces
  //
  p = skip_leading_whitespaces(q + 1);
  //
  // skip possible trailing comments
  // skip possible trailing white spaces
  // remove possible double quotations surrounding the value
  //
  p = remove_comment_trailingwhitespaces_surroundingquotations(p, &quotation);
  //
  // save a copy of the value text and skip possible escaping chars (\ or ")
  //
  quote_strncpy(buf, p, buf_size, quotation);

  return 1;
}


/**
 * @param section       section name
 * @param key           key name
 * @param default_value default key value when searching failes
 * @param buf           the result should put here
 * @param buf_size      the buffer size
 * @param filename      the config file name
 * @return              the number of characters write to the buffer (not including the terminator)
 */
int cfg_getstring(const TCHAR *section, const TCHAR *key, const TCHAR *default_value, TCHAR *buf, int buf_size, const TCHAR *filename)
{
  int r;
  CFG_FILETYPE fp;

  //
  // key cannot be NULL
  //
  if (buf == NULL || buf_size <= 0 || key == NULL) {
    return 0;
  }

  //
  // fail to open file for reading
  //
  if (!cfg_fopenread(filename, &fp)) {
    goto exit2;
  }

  //
  // fail to find the key
  //
  if (get_section_or_key_or_value(&fp, section, key, -1, -1, buf, buf_size))
  {
    goto exit1;
  }

  //
  // close the file
  //
  r = cfg_fclose(&fp);
  assert (r != 0);

exit2:
  quote_strncpy(buf, default_value, buf_size, CFG_QUOTE_NONE);
exit1:
  return _tcslen(buf);
}

/**
 * @param section         section name
 * @param key             key name
 * @param default_value   default value when searching fails
 * @param filename        confi file name
 * @return                the value for the key
 */
long cfg_getlong(const TCHAR *section, const TCHAR *key, long default_value, const TCHAR *filename)
{
  int nchars;
  TCHAR buf[128];

  nchars = cfg_getstring(section, key, __T(""), buf, _countof(buf), filename);
  if (nchars == 0) {
    //
    // key not found, just return default value
    //
    return default_value;
  } else if (nchars >= 2 && (buf[1] == 'x' || buf[1] == 'x')) {
    //
    // hexadecimal
    //
    return _tcstol(buf, NULL, 16);
  } else {
    //
    // decimal
    //
    return _tcstol(buf, NULL, 10);
  }
}


/**
 * @param section         section name
 * @param key             key name
 * @param default_value   default value when searching fails
 * @param filename        config file name
 * @return                the value for the key
 */
float cfg_getfloat(const TCHAR *section, const TCHAR *key, float default_value, const TCHAR *filename)
{
  int nchars;
  TCHAR buf[128];

  nchars = cfg_getstring(section, key, __T(""), buf, _countof(buf), filename);
  if (nchars == 0) {
    return default_value;
  } else {
    return cfg_str2float(buf);
  }
}


/**
 * @param section         section name
 * @param key             key name
 * @param default_value   default value when searching fails
 * @param filename        config file name
 * @return                the value for the key converted to boolean values: 0 or 1
 * string with leading 1, T, t, Y, y are all considered to be true value: 1
 * string with leading 0, F, f, N, n are all considered to be false value: 0
 * if neither matches, just return the given default_value
 */
int cfg_getbool(const TCHAR *section, const TCHAR *key, int default_value, const TCHAR *filename)
{
  int ret;
  TCHAR buf[2];

  //
  // this is tricky,
  // we do not check if we found the key
  // and set default string to empty
  //
  ret = default_value;
  cfg_getstring(section, key, __T(""), buf, _countof(buf), filename);
  if (buf[0] == '1' || buf[0] == 'T' || buf[0] == 't' || buf[0] == 'Y' || buf[0] == 'y') {
    ret = 1;
  }
  if (buf[0] == '0' || buf[0] == 'F' || buf[0] == 'f' || buf[0] == 'N' || buf[0] == 'n') {
    ret = 0;
  }
  return ret;
}

/**
 * @param index      section index zero-based
 * @param buf        the buffer
 * @param buf_size   buffer chars capacity
 * @param filename   config file name
 * @return           number of characters copied
 */
int cfg_getsection(int index, TCHAR *buf, int buf_size, const TCHAR *filename)
{
  int r;
  CFG_FILETYPE fp;

  if (buf == NULL || buf_size <= 0 || index < 0) {
    return 0;
  }

  if (!cfg_fopenread(filename, &fp)) {
    goto exit2;
  }
  if (get_section_or_key_or_value(&fp, NULL, NULL, index, -1, buf, buf_size)) {
    goto exit1;
  }

  //
  // close file
  //
  r = cfg_fclose(&fp);
  assert (r != 0);

exit2:
  *buf = '\0';
exit1:
  return _tcslen(buf);
}

/**
 * @param section    section name to scan, or NULL
 * @param index      key index zero-based
 * @param buf        the buffer
 * @param buf_size   buffer chars capacity
 * @param filename   config file name
 * @return           number of characters copied
 */
int  cfg_getkey(const TCHAR *section, int index, TCHAR *buf, int buf_size, const TCHAR *filename)
{
  int r;
  CFG_FILETYPE fp;

  if (buf == NULL || buf_size <= 0 || index < 0) {
    return 0;
  }
  if (!cfg_fopenread(filename, &fp)) {
    goto exit2;
  }
  if (get_section_or_key_or_value(&fp, section, NULL, -1, index, buf, buf_size)) {
    goto exit1;
  }
  //
  // close file
  //
  r = cfg_fclose(&fp);
  assert (r != 0);
 
exit2:
    *buf = '\0';
exit1:
  return _tcslen(buf);
}


