/******************************************************************************
Copyright 2012 - 2020 Intel Corporation

For licensing information, see the file 'LICENSE' in the root folder of
this software module.
******************************************************************************/

#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#include "conf.h"
#include "tcd.h"
#include "cltdb.h"
#include "tcdmsg.h"
#include "tm.h"

// CLT clts[MAXCLTS];
// extern SCB scb; 
// int clts_count=0;
// extern int svr_cbulk, svr_remain_slots;


// initialize the clts data structure in server control block structure
void init_clts(SCB *scbp)
{
	CLT *cp;
	int k;

	for (k=0; k<MAXCLTS; k++)
	{
		cp = &(scbp->clts[k]);

		cp->used = 0;
		cp->srcipadd = 0;
		cp->srcport = 0;
		cp->cbulk = 0;
		cp->cbulkfull = CLT_CBULK_EMPTY;
		init_bulkmap(&(cp->cbulkmap));

		cp->machine_status = CLT_MACHINE_STATUS_UNKNOWN;
		cp->runacked = 0;
	}

	scbp->clts_count = 0;
	scbp->clts_normal_count = 0;

#ifdef USE_MYSQL
	db_clear_deploy_client();
	// added on May 11
	db_clear_user_status();
#endif

	return;	
}

// return the # of clients in the registration table. 
int count_clts(SCB *scbp)
{
	return scbp->clts_count;
}

// return # of normal clients in the registration table.
int normal_count_clts(SCB *scbp)
{
	return scbp->clts_normal_count;
}


// find the clts entry with (ip, port) pair
CLT *find_clts(SCB *scbp, uint32 ip, uint16 port)
{
	CLT *cp;
	int k=-1;

	while (++k < MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;

		if (cp->srcipadd == ip && cp->srcport == port)
			return (CLT *)cp;
	}

	return (CLT *)NULL;	
}


// insert the clts entry with (ip, port) pair, and return NULL if it's full already
CLT *insert_clts(SCB *scbp, uint32 ip, uint16 port, unsigned char *hwa, uint32 machine_status)
{
	CLT *cp;
	int k=-1;
	
	cp = find_clts(scbp, ip, port);
	if (cp != NULL)
	{
#ifdef STANDALONE_MODE
		fprintf(stderr, "insert_clts: already have\n");
#endif
		return cp;
	}

	while (++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
		{
			cp->used = 1;
			cp->srcipadd = ip;
			cp->srcport = port;
			memcpy(&(cp->hwa[0]), (char *)hwa, HWA_SZ);

			init_bulkmap(&(cp->cbulkmap));
			cp->cbulk = 0;
			cp->cbulkfull = 0;
			cp->machine_status = machine_status;

			cp->runacked = 0;

			scbp->clts_count++;

			if (machine_status == CLT_MACHINE_STATUS_NORMAL)
				scbp->clts_normal_count++;

#ifdef USE_MYSQL
			db_insert_deploy_client(ip, hwa, machine_status);
#endif

			return cp;
		}
	}

#ifdef STANDALONE_MODE
	fprintf(stderr, "insert_clts: full\n");
#endif
	return NULL;	
}

CLT *delete_clts(SCB *scbp, uint32 ip, uint16 port)
{
	CLT *cp;
	int k=-1;
	
	cp = find_clts(scbp, ip, port);
	if (cp == NULL)
		return NULL;
	else
	{
		cp->used = 0;
		cp->srcipadd = 0;
		cp->srcport = 0;
		cp->cbulk = 0;
		cp->cbulkfull = 0;
		init_bulkmap(&(cp->cbulkmap));


		if (cp->machine_status == CLT_MACHINE_STATUS_NORMAL)
			scbp->clts_normal_count--;

		cp->machine_status = CLT_MACHINE_STATUS_UNKNOWN;
		cp->runacked = 0;

		scbp->clts_count--;



#ifdef USE_MYSQL
		db_delete_deploy_client(ip);
#endif

		return cp;
	}
}


// drop client #
void drop_clts(SCB *scbp, int idx)
{
	CLT *cp;
	int k=-1;
	uint32 ip;
	
	if (idx < 0 || idx >= MAXCLTS)
	{
#ifdef STANDALONE_MODE
		fprintf(stderr, "drop_clts: illegal idx %d, should be in [0, %d)\n", idx, MAXCLTS);
#endif
		return;
	}


	cp = &(scbp->clts[idx]);

	cp->used = 0;
	// to reserve the ip address...
	ip = cp->srcipadd;
	cp->srcipadd = 0;
	cp->srcport = 0;
	cp->cbulk = 0;
	cp->cbulkfull = 0;
	init_bulkmap(&(cp->cbulkmap));

	if (cp->machine_status == CLT_MACHINE_STATUS_NORMAL)
		scbp->clts_normal_count--;


	cp->machine_status = CLT_MACHINE_STATUS_UNKNOWN;

	cp->runacked = 0;

	scbp->clts_count--;

#ifdef USE_MYSQL
	//db_delete_deploy_client(ip);
	db_update_deploy_client(ip, "interrupted");
#endif

	return;

}



// drop client #
int drop_clts2(SCB *scbp, CLT *clt)
{
	CLT *cp;
	int k=-1;
	uint32 ip;
	
	if (clt == NULL)
		return -1;

	cp = clt;

	cp->used = 0;
	// to reserve the ip address...
	ip = cp->srcipadd;
	cp->srcipadd = 0;
	cp->srcport = 0;
	cp->cbulk = 0;
	cp->cbulkfull = 0;
	init_bulkmap(&(cp->cbulkmap));

	if (cp->machine_status == CLT_MACHINE_STATUS_NORMAL)
		scbp->clts_normal_count--;


	cp->machine_status = CLT_MACHINE_STATUS_UNKNOWN;

	cp->runacked = 0;

	scbp->clts_count--;

#ifdef USE_MYSQL
	//db_delete_deploy_client(ip);
	db_update_deploy_client(ip, "interrupted");
#endif

	return 0;

}



// to clear the clts table
void clear_clts(SCB *scbp)
{
	init_clts(scbp);

	return;
}


void print_clts_item(SCB *scbp, CLT *cp)
{
	struct in_addr in;

	if (cp != NULL)
	{
		in.s_addr = cp->srcipadd;
#ifdef STANDALONE_MODE
		printf("clt [#%d] --- %s:%d [%02X:%02X:%02X:%02X:%02X:%02X] current bulk#=%d mach_status=%s\n", (int)(cp - &(scbp->clts[0])), inet_ntoa(in), cp->srcport, \
			cp->hwa[0], cp->hwa[1], cp->hwa[2], cp->hwa[3], cp->hwa[4], cp->hwa[5], cp->cbulk, CLT_MACHINE_STATUS_STR(cp->machine_status));
#endif
		// print_bulkmap(&(cp->cbulkmap));
	}
	else
	{
#ifdef STANDALONE_MODE
		printf("empty clts item\n");
#endif
	}

	return;
}

void print_clts(SCB *scbp)
{
	CLT *cp;
	int k=-1, count=0;

#ifdef STANDALONE_MODE
	printf("========== total clients: ==========\n");
	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;
		print_clts_item(scbp, cp);
		count ++;
	}
	printf("=============[%2d(%2d)]=============\n", scbp->clts_count, scbp->clts_normal_count);
	printf("\n");
#endif
	return;	
}



void log_clts(SCB *scbp)
{
	CLT *cp;
	int k=-1;

	log_printf("clt cbulk# {");
	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;

		log_printf("%d, ", cp->cbulk);
	}
	log_printf("}\n\n");
	return;	
}



// newly added for clear clts counter in 2nd file transfer .
void clear_cbulk_clts(SCB *scbp)
{
	CLT *cp;
	int k=-1;

	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;
		cp->cbulk = 0;
		cp->cbulkfull = CLT_CBULK_EMPTY;
		init_bulkmap(&(cp->cbulkmap));
	}
	return;	
}


// clear cbulkfull bit of all entries.
void clear_cbulkfull_status_clts(SCB *scbp)
{
	CLT *cp;
	int k=-1;

	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;

		cp->cbulkfull = CLT_CBULK_EMPTY;
	}
	return;	
}

// put all client ip address into a memory block
uint32 put_clts(SCB *scbp, uint32 *list)
{
	CLT *cp;
	int k=-1, count=0;
	uint32 *l = list;

	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;
		// print_clts_item(scbp, cp);
		*l = cp->srcipadd;
		l++;
		count ++;
	}

	return count;	
}


// put all client (ipaddr, cbulk#) pair into a memory block
uint32 put_clts_ipcbulk(SCB *scbp, uint32 *ipcbulklist)
{
	CLT *cp;
	int k=-1, count=0;
	uint32 *l = ipcbulklist;

	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;
		// print_clts_item(scbp, cp);
		*l = cp->srcipadd;
		l++;
		*l = cp->cbulk;
		l++;

		count ++;
	}

	return count;	
}

// check if all clients have received packet acknowledge and each ack is full-sloted.
int all_client_rcvd_and_full_clts(SCB *scbp)
{
	CLT *cp;
	int k=-1;
	int allfull = 1;

	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;

		// kick off broken machines, e.g., no hard drive, or hard drive not mounted. 
		if (cp->machine_status != CLT_MACHINE_STATUS_NORMAL)
			continue; 

		if (cp->cbulkfull != CLT_CBULK_FULL)
		{	
			allfull = 0;
			break;
		}
	}
	return allfull;
}

// check if all clients have received packet acknowledgement.
int all_client_rcvd_pktack_clts(SCB *scbp)
{
	CLT *cp;
	int k=-1;
	int all_pktacked = 1;

	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;

		// kick off broken machines, e.g., no hard drive, or hard drive not mounted. 
		if (cp->machine_status != CLT_MACHINE_STATUS_NORMAL)
			continue; 

		if (cp->cbulkfull == CLT_CBULK_EMPTY)
		{	
			all_pktacked = 0;
			break;
		}
	}
	return all_pktacked;
}

// check if no client has received packet acknowledgement.
int no_client_rcvd_pktack_clts(SCB *scbp)
{
	CLT *cp;
	int k=-1;
	int no_pktacked = 1;

	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;

		// kick off broken machines, e.g., no hard drive, or hard drive not mounted. 
		if (cp->machine_status != CLT_MACHINE_STATUS_NORMAL)
			continue; 


		if (cp->cbulkfull != CLT_CBULK_EMPTY)
		{	
			no_pktacked = 0;
			break;
		}
	}
	return no_pktacked;
}


// check if some client was dropped, and if so, we will kick it off.
int check_offline_client(SCB *scbp)
{
	CLT *cp;
	int k=-1; 
	int min_k = -1;	// index to record the offset with smallest client->cbulk
	uint32 min_cbulk = 0;

	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		if (cp->used == 0)
			continue;

		// kick off broken machines, e.g., no hard drive, or hard drive not mounted. 
		if (cp->machine_status != CLT_MACHINE_STATUS_NORMAL)
			continue; 

		if (min_k == -1)
		{
			min_cbulk = cp->cbulk;
			min_k = k;
		}
		else if (cp->cbulk < min_cbulk)
		{
			min_cbulk = cp->cbulk;
			min_k = k;
		}
	}

	return min_k;
}

void check_and_drop_non_runack_clients(SCB *scbp)
{
	CLT *cp;
	int k=-1; 
//	int min_k = -1;	// index to record the offset with smallest client->cbulk
//	uint32 min_cbulk = 0;
	uint32 ip;

	while(++k<MAXCLTS)
	{
		cp = &(scbp->clts[k]);
		// for empty entry, just neglect. 
		if (cp->used == 0)
			continue;

		// normal client who has received runacked message
		if (cp->runacked == 1)
			continue; 

		// drop the client which was not acked. 

		cp->used = 0;
		// to reserve the ip address...
		ip = cp->srcipadd;
		cp->srcipadd = 0;
		cp->srcport = 0;
		cp->cbulk = 0;
		cp->cbulkfull = 0;
		init_bulkmap(&(cp->cbulkmap));

		if (cp->machine_status == CLT_MACHINE_STATUS_NORMAL)
			scbp->clts_normal_count--;

		cp->machine_status = CLT_MACHINE_STATUS_UNKNOWN;

		cp->runacked = 0;

		scbp->clts_count--;

#ifdef USE_MYSQL
		//db_delete_deploy_client(ip);
		db_update_deploy_client(ip, "interrupted");
#endif
	
	}

	return;
	
}



/*
main()
{
	init_clts();
	insert_clts(0x11111111, 0x11);
	insert_clts(0x22222222, 0x22);
	insert_clts(0x33333333, 0x33);
	insert_clts(0x44444444, 0x44);
	print_clts();

	CLT *cp;
	cp = find_clts(0x33333333, 0x33);
	print_clts_item(cp);
	cp = find_clts(0x44444444, 0x44);
	print_clts_item(cp);

	if ((cp = delete_clts(0x33333333, 0x33))==NULL)
		printf("delete 0x33333333, 0x33 failed\n");
	else
		printf("delete 0x33333333, 0x33 successfully\n");

	if ((cp = delete_clts(0x33333333, 0x33))==NULL)
		printf("delete 0x33333333, 0x33 failed\n");
	else
		printf("delete 0x33333333, 0x33 successfully\n");

	print_clts();

}
*/


























// initialize the clts data structure in server control block structure
void init_actusrs(SCB *scbp)
{
	ACTUSR *ap;
	int k;

	if (scbp == NULL)
	{
		fprintf(stderr, "init_actusrs: NULL scbp\n");
		return;
	}

	for (k=0; k<MAXACTUSRS; k++)
	{
		ap = &(scbp->actusrs[k]);
		ap->used = 0;

		ap->srcipadd = 0;
		ap->srcport = 0;
		ap->hwa[0] = ap->hwa[1] = ap->hwa[2] = ap->hwa[3] = ap->hwa[4] = ap->hwa[5] = 0; 
		strcpy(ap->uname, "\0");
		strcpy(ap->osname, "\0");

	}
	scbp->actusrs_count = 0;

#ifdef USE_MYSQL
//	db_clear_deploy_client();
#endif

	return;	
}

// return the # of clients in the registration table. 
int count_actusrs(SCB *scbp)
{
	if (scbp == NULL)
	{
		fprintf(stderr, "count_actusrs: NULL scbp\n");
		return -1;
	}

	return scbp->actusrs_count;
}

// find the clts entry with (ip, port) pair
ACTUSR *find_actusrs(SCB *scbp, uint32 ip/*, uint16 port*/)
{
	ACTUSR *ap;
	int k=-1;

	if (scbp == NULL)
	{
		fprintf(stderr, "find_actusrs: NULL scbp\n");
		return NULL;
	}

	while (++k < MAXACTUSRS)
	{
		ap = &(scbp->actusrs[k]);
		if (ap->used == 0)
			continue;

		if (ap->srcipadd == ip /*&& ap->srcport == port*/)
			return (ACTUSR *)ap;
	}

	return (ACTUSR *)NULL;	
}


// insert the clts entry with (ip, port) pair, and return NULL if it's full already
ACTUSR *insert_actusrs(SCB *scbp, uint32 ip, /*uint16 port, */unsigned char *hwa, char *uname, char *osname)
{
	ACTUSR *ap;
	int k=-1;

	if (scbp == NULL)
	{
		fprintf(stderr, "insert_actusrs: NULL scbp\n");
		return NULL;
	}
	
	ap = find_actusrs(scbp, ip/*, port*/);
	if (ap != NULL)
	{
		fprintf(stderr, "insert_actusrs: already have\n");
		return ap;
	}

	while (++k < MAXACTUSRS)
	{
		ap = &(scbp->actusrs[k]);

		if (ap->used == 0)
		{
			ap->used = 1;

			ap->srcipadd = ip;
			// ap->srcport = port;
			memcpy(&(ap->hwa[0]), (char *)hwa, HWA_SZ);

			strcpy(ap->uname, uname);
			strcpy(ap->osname, osname);

			// update the timestamp for this user ...
			ap->ts = getts();

			scbp->actusrs_count++;
#ifdef USE_MYSQL
//			db_insert_deploy_client(ip, hwa);
#endif

			return ap;
		}
	}

	fprintf(stderr, "insert_actusrs: full\n");
	return NULL;	
}


ACTUSR *touch_actusers(SCB *scbp, uint32 ip/*, uint16 port*/)
{
	ACTUSR *ap;
	int k=-1;

	if (scbp == NULL)
	{
		fprintf(stderr, "touch_actusrs: NULL scbp\n");
		return NULL;
	}
	
	ap = find_actusrs(scbp, ip/*, port*/);
	if (ap == NULL)
		return NULL;
	else
	{
		// update the timestamp for this user ...
		ap->ts = getts();

#ifdef USE_MYSQL
//		db_delete_deploy_client(ip);
#endif

		return ap;
	}
}


#ifdef POST_REPORT_HWINFO
// used for periodically update message process from clients. 
// if the entry has already been there, just touch it by updating the time stamp
// if the entry cannot be found, just insert it. 
ACTUSR *insert_or_touch_actusrs(SCB *scbp, uint32 ip, /*uint16 port, */unsigned char *hwa, char *uname, char *osname, int preboot, int disksz, int memsz, const char *strcpu)
{
	ACTUSR *ap;
	int k=-1;

	if (scbp == NULL)
	{
		fprintf(stderr, "insert_or_touch_actusrs: NULL scbp\n");
		return NULL;
	}

	ap = find_actusrs(scbp, ip/*, port*/);
	if (ap != NULL)
	{
		printf("insert_or_touch_actusrs: already have, touch it\n");

		// update the timestamp for this user ...
		ap->ts = getts();
		if (preboot)
		{
			// preboot stage, need to refresh

			// ap->srcipadd = ip;
			// ap->srcport = port;
			memcpy(&(ap->hwa[0]), (char *)hwa, HWA_SZ);

			strcpy(ap->uname, uname);
			strcpy(ap->osname, osname);

			strcpy(ap->strcpu, strcpu);
            ap->disksz = disksz;
            ap->memsz = memsz;
		}


		return ap;
	}

	// not there, insert it ...
	printf("insert_or_touch_actusrs: not there, insert it uname [%s] osname [%s]\n", uname, osname);
	k = -1;
	while (++k < MAXACTUSRS)
	{
		ap = &(scbp->actusrs[k]);

		if (ap->used == 0)
		{
			ap->used = 1;

			ap->srcipadd = ip;
			// ap->srcport = port;
			memcpy(&(ap->hwa[0]), (char *)hwa, HWA_SZ);

			strcpy(ap->uname, uname);
			strcpy(ap->osname, osname);

			strcpy(ap->strcpu, strcpu);
            ap->disksz = disksz;
            ap->memsz = memsz;

			// update the timestamp for this user ...
			ap->ts = getts();

			scbp->actusrs_count++;
#ifdef USE_MYSQL
//			db_insert_deploy_client(ip, hwa);
#endif

			return ap;
		}
	}

	fprintf(stderr, "insert_or_touch_actusrs: full\n");

	return NULL;	
	
}
#else //POST_REPORT_HWINFO

// used for periodically update message process from clients. 
// if the entry has already been there, just touch it by updating the time stamp
// if the entry cannot be found, just insert it. 
ACTUSR *insert_or_touch_actusrs(SCB *scbp, uint32 ip, /*uint16 port, */unsigned char *hwa, char *uname, char *osname, int preboot)
{
	ACTUSR *ap;
	int k=-1;

	if (scbp == NULL)
	{
		fprintf(stderr, "insert_or_touch_actusrs: NULL scbp\n");
		return NULL;
	}

	ap = find_actusrs(scbp, ip/*, port*/);
	if (ap != NULL)
	{
		printf("insert_or_touch_actusrs: already have, touch it\n");

		// update the timestamp for this user ...
		ap->ts = getts();
		if (preboot)
		{
			// preboot stage, need to refresh

			// ap->srcipadd = ip;
			// ap->srcport = port;
			memcpy(&(ap->hwa[0]), (char *)hwa, HWA_SZ);

			strcpy(ap->uname, uname);
			strcpy(ap->osname, osname);
		}


		return ap;
	}

	// not there, insert it ...
	printf("insert_or_touch_actusrs: not there, insert it uname [%s] osname [%s]\n", uname, osname);
	k = -1;
	while (++k < MAXACTUSRS)
	{
		ap = &(scbp->actusrs[k]);

		if (ap->used == 0)
		{
			ap->used = 1;

			ap->srcipadd = ip;
			// ap->srcport = port;
			memcpy(&(ap->hwa[0]), (char *)hwa, HWA_SZ);

			strcpy(ap->uname, uname);
			strcpy(ap->osname, osname);

			// update the timestamp for this user ...
			ap->ts = getts();

			scbp->actusrs_count++;
#ifdef USE_MYSQL
//			db_insert_deploy_client(ip, hwa);
#endif

			return ap;
		}
	}

	fprintf(stderr, "insert_or_touch_actusrs: full\n");

	return NULL;	
	
}
#endif //POST_REPORT_HWINFO

ACTUSR *delete_actusrs(SCB *scbp, uint32 ip/*, uint16 port*/)
{
	ACTUSR *ap;
	int k=-1;

	if (scbp == NULL)
	{
		fprintf(stderr, "delete_actusrs: NULL scbp\n");
		return NULL;
	}
	
	ap = find_actusrs(scbp, ip/*, port*/);
	if (ap == NULL)
		return NULL;
	else
	{
		ap->used = 0;

		ap->srcipadd = 0;
		// ap->srcport = 0;

		ap->hwa[0] = ap->hwa[1] = ap->hwa[2] = ap->hwa[3] = ap->hwa[4] = ap->hwa[5] = 0; 
		strcpy(ap->uname, "\0");
		strcpy(ap->osname, "\0");

		scbp->actusrs_count--;

#ifdef USE_MYSQL
//		db_delete_deploy_client(ip);
#endif

		return ap;
	}
}


// drop client #
void drop_actusrs(SCB *scbp, int idx)
{
	ACTUSR *ap;
	int k=-1;
	uint32 ip;

	if (scbp == NULL)
	{
		fprintf(stderr, "drop_actusrs: NULL scbp\n");
		return;
	}
	
	if (idx < 0 || idx >= MAXACTUSRS)
	{
		fprintf(stderr, "drop_actusrs: illegal idx %d, should be in [0, %d)\n", idx, MAXACTUSRS);
		return;
	}

	ap = &(scbp->actusrs[idx]);
	ap->used = 0;

	// to reserve the ip address...
	ip = ap->srcipadd;
	ap->srcipadd = 0;
	// ap->srcport = 0;

	ap->hwa[0] = ap->hwa[1] = ap->hwa[2] = ap->hwa[3] = ap->hwa[4] = ap->hwa[5] = 0; 
	strcpy(ap->uname, "\0");
	strcpy(ap->osname, "\0");

	scbp->actusrs_count--;

#ifdef USE_MYSQL
	// db_update_deploy_client(ip, "interrupted");
#endif

	return;

}


// to clear the clts table
void clear_actusrs(SCB *scbp)
{
	if (scbp == NULL)
	{
		fprintf(stderr, "clear_actusrs: NULL scbp\n");
		return;
	}

	init_actusrs(scbp);
	return;	
}


void print_actusrs_item(SCB *scbp, ACTUSR *ap)
{
	struct in_addr in;

	if (ap != NULL)
	{
		in.s_addr = ap->srcipadd;
		struct timeval ts = getts();

		printf("actusr [#%d] --- %s [%02X:%02X:%02X:%02X:%02X:%02X] user: [%s] os: [%s] ts: [%lld]\n", \
			(int)(ap - &(scbp->actusrs[0])), inet_ntoa(in), \
			ap->hwa[0], ap->hwa[1], ap->hwa[2], ap->hwa[3], ap->hwa[4], ap->hwa[5], ap->uname, ap->osname, \
			diff_tv(ap->ts, ts));

/*
		printf("actusr [#%d] --- %s:%d [%02X:%02X:%02X:%02X:%02X:%02X] user: [%s] os: [%s] ts: [%lld]\n", \
			(int)(ap - &(scbp->actusrs[0])), inet_ntoa(in), ap->srcport, \
			ap->hwa[0], ap->hwa[1], ap->hwa[2], ap->hwa[3], ap->hwa[4], ap->hwa[5], ap->uname, ap->osname, \
			diff_tv(ap->ts, ts));
*/

/*
		printf("current ts: ");
		printts(ts); 
		printf(" --- actusr ts: ");
		printts(ap->ts);
*/
// diff_tv(ap->ts, ts)

		// print_bulkmap(&(cp->cbulkmap));
	}
	else
		printf("empty actusrs item\n");

	return;
}

void print_actusrs(SCB *scbp)
{
	ACTUSR *ap;
	int k=-1, count=0;

	if (scbp == NULL)
	{
		fprintf(stderr, "print_actusrs: NULL scbp\n");
		return;
	}

	printf("========== total active users: ==========\n");
	while(++k<MAXACTUSRS)
	{
		ap = &(scbp->actusrs[k]);
		if (ap->used == 0)
			continue;
		print_actusrs_item(scbp, ap);
		count ++;
	}
	printf("================[%2d]================\n", count);
	printf("\n");
	return;	
}


// scan the actusr table. if some entries were not updated for a certain period of time (DEFAULT_SVR_ACTUSR_TIMEOUT, for 5 seconds), delete it. 
void handle_actusers_timeout(SCB *scbp)
{
	ACTUSR *ap;
	int k=-1;

	uint64 diff;
	struct timeval ts = getts();
	int retval;

//	diff = diff_tv(scbp->init_state_tv, ts);
//	if (diff > DEFAULT_SVR_HELLO_TIMEOUT)


	if (scbp == NULL)
	{
		fprintf(stderr, "handle_actusers_timeout: NULL scbp\n");
		return;
	}

	while(++k<MAXACTUSRS)
	{
		ap = &(scbp->actusrs[k]);
		if (ap->used == 0)
			continue;
//		print_actusrs_item(scbp, ap);

		diff = diff_tv(ap->ts, ts);		
		if (diff > DEFAULT_SVR_ACTUSR_TIMEOUT)
		{
			printf("=====>>>>> handle_actusers_timeout: delete %d actusr\n", k);
			ap->used = 0;

			// to reserve the ip address...
			ap->srcipadd = 0;
			// ap->srcport = 0;

			ap->hwa[0] = ap->hwa[1] = ap->hwa[2] = ap->hwa[3] = ap->hwa[4] = ap->hwa[5] = 0; 
			strcpy(ap->uname, "\0");
			strcpy(ap->osname, "\0");

			scbp->actusrs_count--;
		}
	}
	return;	
	
}


