/******************************************************************************
Copyright 2012 - 2020 Intel Corporation

For licensing information, see the file 'LICENSE' in the root folder of
this software module.
******************************************************************************/

#include "conf.h"

#include <time.h>
// #include "predefine.h"
#include "db.h"
#include "tcd.h"
// #include "debug.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef USE_MYSQL

#include <mysql/mysql.h>	//mysql header file
#include "tcdmsg.h"

static MYSQL g_db_connection; 		//used in db connection
static int g_db_connected=0;		//db connection status
static int g_need_clean_status=1;
static char *g_db_username=NULL,*g_db_password=NULL;

int start_db(char *name, char *pwd);
int update_pcnt_info(char *imagename, int pcnt);


static int safe_mysql_query(MYSQL *sql_conn,const char *query)
{
	int mysql_error_code,errno;
	int retry;
	
	retry=0;
	while(1)
	{
		mysql_error_code = mysql_query(sql_conn,query);
		errno=mysql_errno(sql_conn);		
		if(retry || errno!=2006)  //server has gone away
		   break;		
		
		DEBUG("sql error %d: %s\n", errno,mysql_error(sql_conn));
		DEBUG("try to restart db...\n");
		g_db_connected=0;
		g_need_clean_status=0;
		if(start_db(g_db_username,g_db_password)!=0){
			g_need_clean_status=1;
			break;
		}
		g_need_clean_status=1;
		retry=1;	
	}
	return mysql_error_code;
}

int update_user_status(int user_id, int image_id, int status_id)
{  
  char sql[MAX_STRING_LEN] = {0};

  //DEBUG("update_user_status\n");
  sprintf(sql,"update userstatus set isOnline='%d' where userId='%d' and osImageId='%d'", status_id, user_id, image_id);
  //DEBUG("SQL:%s\n", sql);
  
  return safe_mysql_query(&g_db_connection,sql);
}

void reset_all_status(int status_id, int removeAll)
{  
  char sql[MAX_STRING_LEN] = {0};
  
  if(!g_need_clean_status)
  	return;	

  //DEBUG("reset all status\n");
  if(removeAll)
  	sprintf(sql,"delete from userstatus");
  else
  	sprintf(sql,"update userstatus set isOnline='%d'", status_id);

  safe_mysql_query(&g_db_connection,sql);
  g_need_clean_status=0;  
}

void set_conn_timeout(unsigned int timeout)
{
	char sql[MAX_STRING_LEN*2];
	
	memset(sql,0,sizeof(sql));

	snprintf(sql,sizeof(sql),"set wait_timeout=%d",timeout);	
	if(mysql_real_query(&g_db_connection,sql,strlen(sql)))
		ERR("select error: %s\n", mysql_error(&g_db_connection));

	snprintf(sql,sizeof(sql),"set interactive_timeout=%d",timeout);	
	if(mysql_real_query(&g_db_connection,sql,strlen(sql)))
		ERR("select error: %s\n", mysql_error(&g_db_connection));
}

void save_authen_info(char *name,char *pwd)
{
	char *tmpname,*tmppwd;
	size_t namelen,pwdlen;
	
	if(!name || !pwd)
		return;
	
	tmpname = g_db_username;
	tmppwd = g_db_password;		
	
	namelen = strlen(name) + 1;
	pwdlen = strlen(pwd) + 1;

	g_db_username = (char *)calloc(namelen, sizeof(char));
	g_db_password = (char *)calloc(pwdlen, sizeof(char));

	strcpy(g_db_username, name);
	strcpy(g_db_password, pwd);
}

/** 
 * There are some functions used to connect db and search infomation in db
**/
//start the database connection
int start_db(char *name, char *pwd)
{	
	if(g_db_connected) 
	  return 0;

	//DEBUG("init mysql\n");
	mysql_init(&g_db_connection);

	if(!mysql_real_connect(&g_db_connection,"127.0.0.1",name,pwd,"tc",0,NULL,0)){
		DEBUG("db: name = %s , pwd = %s\n",name,pwd);
		DEBUG("Database connection failure: %d, %s\n",mysql_errno(&g_db_connection),mysql_error(&g_db_connection));
		return -1;
	}else{
		DEBUG("DB connection established\n");
	}	
	
	set_conn_timeout(30*24*3600);
	save_authen_info(name,pwd);
	g_db_connected = 1;        
	//reset_all_status();	
	
	return 0;
}

//close the database connection
void end_db()
{
	if(g_db_connected){ 
		mysql_close(&g_db_connection);
    }
		
	g_db_connected = 0;
	
	if(g_db_username){
		free(g_db_username);
		g_db_username = NULL;
    }
		
	if(g_db_password){
		free(g_db_password);	
		g_db_password = NULL;
    }
}

//get all users' status from db
int get_all_user_status(char *statuslist)
{	
	MYSQL_RES *res_ptr;
	MYSQL_ROW mysqlrow;
	int res, i;
	char qs[MAX_STRING_LEN *2] = {0};

  	//DEBUG("db: get_all_user_status\n");
  	if(statuslist == NULL) {
		ERR("status list pointer is NULL\n");
		return -1;
  	}

	sprintf(qs, "SELECT A.userId, B.username, A.tcId, A.mac, A.osimageid, A.osimagename FROM userstatus A "
				"left join users B on B.id=A.userId "
				"WHERE isOnline!='0'");

	res = safe_mysql_query(&g_db_connection,qs);
	statuslist[0] = 0;
	if(res) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;
	} else {
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if(mysql_num_rows(res_ptr)>0) {
				i = 0;
        			
		    		while((mysqlrow=mysql_fetch_row(res_ptr))) {
					if(i > 0)
          					strcat(statuslist, ";");
					strcat(statuslist, mysqlrow[0]);
					strcat(statuslist, ",");
					strcat(statuslist, mysqlrow[1]);
					strcat(statuslist, ",");
					strcat(statuslist, mysqlrow[2]);
					strcat(statuslist, ",");
					strcat(statuslist, mysqlrow[3]);
					strcat(statuslist, ",");
					strcat(statuslist, mysqlrow[4]);
					strcat(statuslist, ",");
					strcat(statuslist, mysqlrow[5]);
					i++;
			  	}
			} else {
				//DEBUG("%s\n",qs);
        			ERR("get_all_user_status: sql returns no result\n");
        			goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				ERR("get_all_user_status retrive error: %s\n",mysql_error(&g_db_connection));
        			goto sql_error;
			}

      			mysql_free_result(res_ptr);
			
		} else {
			ERR("failed to get all user status list\n");
			goto sql_error;
   	 	}

	}

  	return 0;	

sql_error:
  	DEBUG("SQL:%s\n", qs);
  	mysql_free_result(res_ptr);
	return -1;
}


int get_username_by_id(uint32_t userid, char *username) 
{
  char qs[MAX_STRING_LEN];
  int  res;
  MYSQL_RES *res_ptr;
  MYSQL_ROW mysqlrow;
  int ret = -1;

  //DEBUG("db: get_username_by_id \n");
  
  sprintf(qs,"select userName from users WHERE id='%u'", userid);

  res = safe_mysql_query(&g_db_connection,qs); 
  if(res){
	ERR("select error: %s\n", mysql_error(&g_db_connection));
	return ret;
		
 }else{
	res_ptr = mysql_store_result(&g_db_connection);
    	if(res_ptr){
		if((mysql_num_rows(res_ptr))>0){
			mysqlrow=mysql_fetch_row(res_ptr);
			sscanf(mysqlrow[0], "%s", username);
			//DEBUG("username : %s\n",username);
			ret = 0;
		} else {
        		DEBUG("get_username_by_id: sql returns no result\n");
        		goto sql_error;
		}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);

		} 
		
	}

  return ret;
  
sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return ret;
  
}

int get_uid_gid_by_name(char *username, int *userid, int *groupid) 
{
	char qs[MAX_STRING_LEN];
	int  res;
	MYSQL_RES *res_ptr;
	MYSQL_ROW mysqlrow;
	//int userid = -1;

	//DEBUG("db: get_userid_by_name \n");

	if(!username || !userid || !groupid) {
		ERR("ERROR get_uid_gid_by_name: Paras NULL\n");
		return -1;
	}

	sprintf(qs,"select id, groupId from users WHERE username = '%s'", username);

	res = safe_mysql_query(&g_db_connection,qs); 
	if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -2;		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    		if(res_ptr){
			if((mysql_num_rows(res_ptr))>0){
				mysqlrow=mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d",userid);
				sscanf(mysqlrow[1],"%d",groupid);
				//DEBUG("userid : %u\n",*userid);
			} else {
		        	DEBUG("get_uid_gid_by_name: sql returns no result\n");
		        	goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}
			mysql_free_result(res_ptr);
		}		
	}

  	return 0;
  
sql_error:
	DEBUG("SQL:%s\n", qs);
  	mysql_free_result(res_ptr);
	return -1;  
}


int get_user_status(char *username, int osimage_id) 
{
  char qs[MAX_STRING_LEN*2];
  int  res;
  MYSQL_RES *res_ptr;
  MYSQL_ROW mysqlrow;
  int isOnline = 0;

  //DEBUG("db: get_user_status \n");
  
  sprintf(qs,"select isOnline from userstatus, users WHERE userstatus.userid=users.id and users.username = '%s' and osimageid=%d", username, osimage_id);

  res = safe_mysql_query(&g_db_connection,qs); 
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return isOnline;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr))>0){
				mysqlrow=mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d",&isOnline);
				DEBUG("usersignin : %d\n",isOnline);
			} else {
        DEBUG("get_user_status: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);

		} 
		
	}

  return isOnline;
  
sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return isOnline;
  
}


int db_delete_from_user_status_by_ip (const char* ipv4){
	int ret; 
	MYSQL_RES *res_ptr;
	char sqls[MAX_STRING_LEN] = {0};

	sprintf(sqls, "SELECT * FROM userstatus WHERE ip = \'%s\' ", ipv4);
	ret = safe_mysql_query(&g_db_connection, sqls);
	if (ret){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		ret = -1;
	}
	else{
		res_ptr = mysql_store_result (&g_db_connection);
		if (res_ptr){
			if((mysql_num_rows(res_ptr))>0){
				sprintf(sqls, "DELETE FROM userstatus WHERE ip = \'%s\' ", ipv4);
				ret = safe_mysql_query(&g_db_connection, sqls);
				if (ret){
					ERR ("delete error: %s\n", mysql_error (&g_db_connection));
					ret = -1;
				}
				else{
					DEBUG ("delete entry with ip(%s) from userstatus\n", ipv4);
					ret = 1;
				}
			}
			mysql_free_result (res_ptr);
		}
		else{
			DEBUG ("No entry with ip(%s) was found in userstatus.\n", ipv4);
			ret = 0;
		}
	}
	return ret;
}


int db_delete_from_deploy_client_by_ip (const char* ipv4){
	int ret; 
	MYSQL_RES *res_ptr;
	char sqls[MAX_STRING_LEN] = {0};

	sprintf(sqls, "SELECT * FROM deploy_client WHERE ip_addr = \'%s\' ", ipv4);
	ret = safe_mysql_query(&g_db_connection, sqls);
	if (ret){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		ret = -1;
	}
	else{
		res_ptr = mysql_store_result (&g_db_connection);
		if (res_ptr){
			if((mysql_num_rows(res_ptr))>0){
				sprintf(sqls, "DELETE FROM deploy_client WHERE ip_addr = \'%s\' ", ipv4);
				ret = safe_mysql_query(&g_db_connection, sqls);
				if (ret){
					ERR ("delete error from deploy_client: %s\n", mysql_error (&g_db_connection));
					ret = -1;
				}
				else{
					DEBUG ("delete entry with ip_addr(%s) from deploy_client\n", ipv4);
					ret = 1;
				}
			}
			mysql_free_result (res_ptr);
		}
		else{
			DEBUG ("No entry with ip_addr(%s) was found in deploy_client.\n", ipv4);
			ret = 0;
		}
	}
	return ret;
}


int get_user_status_by_id(uint32_t user_id, int osimage_id) 
{
  char qs[MAX_STRING_LEN*2];
  int  res;
  MYSQL_RES *res_ptr;
  MYSQL_ROW mysqlrow;
  int isOnline = 0;

  //DEBUG("db: get_user_status \n");
  
  sprintf(qs,"select isOnline from userstatus, users WHERE userstatus.userid=users.id and users.id = '%d' and osimageid=%d", user_id, osimage_id);

  res = safe_mysql_query(&g_db_connection,qs); 
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return isOnline;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr))>0){
				mysqlrow=mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d",&isOnline);
				DEBUG("usersignin : %d\n",isOnline);
			} else {
        DEBUG("get_user_status: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);

		} 
		
	}

  return isOnline;
  
sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return isOnline;
  
}


int get_user_status_by_name(char *username) 
{
  char qs[MAX_STRING_LEN*2];
  int  res;
  MYSQL_RES *res_ptr;
	MYSQL_ROW mysqlrow;
	int isOnline = 0;

  //DEBUG("db: get_user_status \n");
  
  sprintf(qs,"select isOnline from userstatus, users WHERE userstatus.userid=users.id and users.username = '%s'", username);

  res = safe_mysql_query(&g_db_connection,qs);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return isOnline;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr))>0){
				mysqlrow=mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d",&isOnline);
				DEBUG("usersignin : %d\n",isOnline);
			} else {
        DEBUG("get_user_status_by_name\n\rsql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);

		} 
		
	}

  return isOnline;
  
sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return isOnline;
  
}


int insert_user_status(int userId, int tcId, int osimage_id, char *imagename, char *mac, int imageaccess, int online)
{	
	char qs[MAX_STRING_LEN*3];	
	//char subqs[MAX_STRING_LEN*2];
	char signtime[25];
	time_t timep;
	struct tm *p;

	//DEBUG("db: insert_user_status\n");

  //sprintf(subqs,"select id from users where username = '%s'", username);
  
	
	time(&timep);
	p = localtime(&timep);
	sprintf(signtime,"%d-%d-%d %d:%d:%d",(1900+p->tm_year),(1+p->tm_mon),p->tm_mday,p->tm_hour,p->tm_min,p->tm_sec);
	sprintf(qs,"insert into userstatus(userid, tcId,isOnline, osImageId,osimagename,ip,mac,osRunmodeId,lastSignInTime)" 
	            //"values((%s),'%s','%d','%d',%d,'%s','000.000','%s',%d,'%s')", 
    	        //subqs,
    	        "values(%d,'%d','%d',%d,'%s','000.000','%s',%d,'%s')", 
    	        userId,
    	        tcId,
    	        online,
    	        osimage_id,
    	        imagename,
    	        mac,
    	        imageaccess,
    	        signtime);
		

	//DEBUG("update sql:%s\n", qs);
	
	return safe_mysql_query(&g_db_connection,qs);		
}

int delete_user_status(int user_id, int osimage_id)
{
	char qs[MAX_STRING_LEN*2]={0};	

  //DEBUG("db: delete_user_status\n");
  
  sprintf(qs, "delete from userstatus where userid =%d and osimageid =%d",
              user_id,
              osimage_id);
  
	//DEBUG("update sql:%s\n",qs);

	return safe_mysql_query(&g_db_connection,qs);	
}

//get the image name if os image matched
int is_osimage_matched(int imageid, char *imagepath, char *imagename)
{
	char qs[MAX_STRING_LEN*2];
  	int  res;
  	MYSQL_RES *res_ptr;
	MYSQL_ROW mysqlrow;
	int matched = -1;

  	//DEBUG("db: is_osimage_matched \n");
  
  	sprintf(qs,"select imageName from osimages WHERE id='%d' and path = '%s'", imageid, imagepath);

  	res = safe_mysql_query(&g_db_connection,qs);
  	if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return matched;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    		if(res_ptr){
			if((mysql_num_rows(res_ptr))>0){
				sscanf(mysqlrow[0],"%s", imagename);
				matched = 0;
			} else {
        			DEBUG("is_osimage_matched: sql returns no result\n");
       			goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
		} 
		
	}

  return matched;
  
sql_error:
  	DEBUG("SQL:%s\n", qs);
  	mysql_free_result(res_ptr);
	return matched;
}

//to update image path after merge the diff into the old image
int update_image_status(char* rawfile, char* newfile, int needupdate)
{
	char qs[MAX_STRING_LEN*2]={0};	
	char updatetime[25];
	time_t timep;
	struct tm *p;

  //DEBUG("db: update_image_status\n");
  
  if(newfile != NULL && needupdate == 0) {
  	time(&timep);
		p=gmtime(&timep);
		sprintf(updatetime,"%d-%d-%d %d:%d:%d",(1900+p->tm_year),(1+p->tm_mon),p->tm_mday,p->tm_hour,p->tm_min,p->tm_sec);
		
		sprintf(qs,"update osimages set path='%s',lastUpdateTime='%s',needUpdate='0' WHERE path = '%s'",newfile,updatetime,rawfile);
	}	else if (newfile == NULL && needupdate){
		sprintf(qs,"update osimages set needUpdate='%d' WHERE path = '%s'", needupdate, rawfile);
  } else {
    return 0;
  }
  
	//DEBUG("update sql:%s",qs);

	return safe_mysql_query(&g_db_connection,qs);	
}

int update_server_status(int server_id,int status)
{
	char qs[MAX_STRING_LEN*2];	
	char signtime[25];
	time_t timep;
	struct tm *p;

	switch(status) {
	case SERVER_STATUS_ONLINE:
		time(&timep);
		p=localtime(&timep);
		sprintf(signtime,"%d-%d-%d %d:%d:%d",(1900+p->tm_year),(1+p->tm_mon),p->tm_mday,p->tm_hour,p->tm_min,p->tm_sec);
		sprintf(qs,"update servers set status='%d',lastPowerOnTime='%s' WHERE id = '%d'",status,signtime,server_id);
	  break;

	case SERVER_STATUS_OVERLOAD:
	  sprintf(qs,"update servers, serverconfigs set servers.status='%d', serverconfigs.configvalue='%d'" 
	                     "WHERE servers.id=serverconfigs.serverid and serverconfigs.configname = 'maxClientNum'"
	                     "and servers.id=%d",
	                     status,
	                     MAX_CLIENTS,
	                     server_id);
	  break;
	  
	case SERVER_STATUS_NORMAL:
		sprintf(qs,"update servers set status='%d' WHERE id = '%d'",SERVER_STATUS_ONLINE,server_id);
		break;
		
	case SERVER_STATUS_OFFLINE:
	  sprintf(qs,"update servers set status='%d' WHERE id = '%d'",SERVER_STATUS_OFFLINE,server_id);
	  break;
	  
	}

	//DEBUG("update sql:%s\n",qs);

	return safe_mysql_query(&g_db_connection,qs);	
}

//get client and ospath in db
int get_user_oslist(char *username, char *pwd, char *oslist, int srv_major, int srv_minor)
{	
	MYSQL_RES *res_ptr;
//	MYSQL_RES *res_ptr1;
	MYSQL_ROW mysqlrow;
	int res;
	char qs[MAX_STRING_LEN*2] = {0};
	char ns[MAX_STRING_LEN] = {0};
	char ps[MAX_STRING_LEN] = {0};
  int groupid;

  //DEBUG("db: get_user_oslist\n");  
  
  mysql_escape_string(ns,username,strlen(username));
	mysql_escape_string(ps,pwd,strlen(pwd));

	sprintf(qs,"SELECT groupid FROM users WHERE userName = '%s' AND pwd = '%s' and isApproved='1' limit 1",ns,ps);

	res=safe_mysql_query(&g_db_connection,qs);
	if(res){
		ERR("select error %d: %s\n",res,mysql_error(&g_db_connection));
		return -2;
		
	}else{
		res_ptr=mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr))>0){

					while((mysqlrow=mysql_fetch_row(res_ptr))){
  					sscanf(mysqlrow[0],"%d",&groupid);
  					//MSG("group id : %d\n",groupid);					
				  }
			} else {
			  DEBUG("get_user_oslist1: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);

		} else {
      goto sql_error;
		}
 
	}

  oslist[0]=0;
  /*
  //will cause problem such as ,INSTR(",1,22,3",2) returns true
  
  sprintf(qs,"SELECT imagename FROM osimages WHERE major = %d and minor=%d and (INSTR(groupread, %d)" 
                                                "or INSTR(groupdiffwrite,%d)"
                                                "or INSTR(grouprawwrite, %d)"
                                                "or INSTR(groupinstall, %d))",		
                                                major, minor,
                                                groupid,
                                                groupid,
                                                groupid,
                                                groupid);
*/
	 sprintf(qs,"SELECT imagename FROM osimages WHERE major = %d and minor=%d and (groupread like '%%,%d,%%' " 
                                                "or groupdiffwrite like '%%,%d,%%' "
                                                "or grouprawwrite like '%%,%d,%%' "
                                                "or groupinstall like '%%,%d,%%' )",
                                                srv_major, srv_minor,
                                                groupid,
                                                groupid,
                                                groupid,
                                                groupid);
	res = safe_mysql_query(&g_db_connection,qs);
	if(res) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -2;
	} else {
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if(mysql_num_rows(res_ptr)>0) {

        mysqlrow = mysql_fetch_row(res_ptr);
   			strcat(oslist, mysqlrow[0]);
   			
		    while((mysqlrow=mysql_fetch_row(res_ptr))) {
          strcat(oslist, ";");
					strcat(oslist, mysqlrow[0]);
			  }	

			} else {
	//DEBUG("%s\n",qs);
        ERR("get_user_oslist2: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				ERR("Retrive error: %s\n",mysql_error(&g_db_connection));
        goto sql_error;
			}

      mysql_free_result(res_ptr);
			
		} else {
		  ERR("failed to find oslist for user %s\n", username);
		  goto sql_error;
    }

	}

  return 0;	

sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return -1;
}

int get_image_id_path_by_name(char *name, int *image_id, char *path)
{
  char qs[MAX_STRING_LEN*2];
  int  res;
  MYSQL_RES *res_ptr;
	MYSQL_ROW mysqlrow;

  //DEBUG("db: get_image_id_path_by_name\n");

  sprintf(qs,"select id, path from osimages WHERE imagename = '%s'", name);

  res = safe_mysql_query(&g_db_connection,qs);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -2;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr))>0){
				mysqlrow=mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d",image_id);
				strcpy(path, mysqlrow[1]);
				//sscanf(mysqlrow[0],"%s",path);
				//DEBUG("osimage path %s\n",path);
			} else {
        DEBUG("get_image_id_path_by_name: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);

		} 
		
	}

  return 0;
  
sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return -1;
}

int get_osimageid_by_name(char *name) 
{
  char qs[MAX_STRING_LEN*2];
  int  res;
  int  osimageid;
  MYSQL_RES *res_ptr;
	MYSQL_ROW mysqlrow;

  //DEBUG("db: get_ospath_by_name\n");

  sprintf(qs,"select id from osimages WHERE imagename = '%s'", name);

  res = safe_mysql_query(&g_db_connection,qs);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -2;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr))>0){
				mysqlrow=mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d", &osimageid);
				//DEBUG("osimage id is %d\n",osimageid);
			} else {
        DEBUG("get_osimageid_by_name: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);

		} 
		
	}

  return osimageid;
  
sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return -1;
}

int get_osimagename_by_id(int image_id, char *osimage_name) 
{
  char qs[MAX_STRING_LEN*2];
  int  res;
  MYSQL_RES *res_ptr;
	MYSQL_ROW mysqlrow;

  //DEBUG("db: get_ospath_by_name\n");

  sprintf(qs,"select imagename from osimages WHERE id = '%d'", image_id);

  res = safe_mysql_query(&g_db_connection,qs);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -2;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr))>0){
				mysqlrow=mysql_fetch_row(res_ptr);
				strcpy(osimage_name, mysqlrow[0]);
				//DEBUG("osimage id is %d\n",osimageid);
			} else {
		        DEBUG("get_osimagename_by_id: sql returns no result\n");
		        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);

		} 
		
	}

  return 0;
  
sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return -1;
}


int get_imageaccess(char *osname, char *username, IMAGE_ACCESS *imageaccess) 
{
  char qs[MAX_STRING_LEN*2];
  int  res;
  int  groupid;
  char tmpstr[12] = {0};
  MYSQL_RES *res_ptr;
	MYSQL_ROW mysqlrow;

  //DEBUG("db: get_imageaccess\n");

  sprintf(qs,"select groupid from users WHERE username = '%s'", username);

  res = safe_mysql_query(&g_db_connection,qs);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -2;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr))>0){
				mysqlrow=mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d", &groupid);
				//DEBUG("groupid is %d\n", groupid);
			} else {
        ERR("get_imageaccess1: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);

		} 
		
	}

  sprintf(tmpstr, "%d", groupid);
	sprintf(qs,"SELECT groupRead,groupDiffWrite,groupInstall,groupRawWrite FROM osimages WHERE imageName = '%s' limit 1",osname);
	res = safe_mysql_query(&g_db_connection,qs);
	if(res) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -2;
	} else {
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if(mysql_num_rows(res_ptr)>0) {
		
				while((mysqlrow=mysql_fetch_row(res_ptr))) {
					if(mysqlrow[3] && strstr(mysqlrow[3],tmpstr))	//if in groupInstall
						*imageaccess=IMAGE_RAW_READWRITE;
					else if(mysqlrow[2] && strstr(mysqlrow[2],tmpstr))
						*imageaccess=IMAGE_INSTALL_RWRITE;
					else if(mysqlrow[1] && strstr(mysqlrow[1],tmpstr))
						*imageaccess=IMAGE_UDF_READWRITE;
					else if(mysqlrow[0] && strstr(mysqlrow[0],tmpstr))
						*imageaccess=IMAGE_READONLY;
					else
						*imageaccess=IMAGE_NO_RIGHTS;		//no rights to access the image
        }	

        //DEBUG("imageaccess = %d\n", (int) *imageaccess);        
			} else {
        		ERR("get_imageaccess2: sql returns no result\n");
        		goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				ERR("Retrive error: %s\n",mysql_error(&g_db_connection));
        		goto sql_error;
			}

      mysql_free_result(res_ptr);
			
		} 

	}

  return 0;
  
sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return -1;
}

int get_imageaccess_by_gid(int imageid, int groupid, IMAGE_ACCESS *imageaccess) 
{
  char qs[MAX_STRING_LEN*2];
  int  res;
  char tmpstr[12] ={0};
  MYSQL_RES *res_ptr;
  MYSQL_ROW mysqlrow;

  //DEBUG("db: get_imageaccess_by_gid\n");


  sprintf(tmpstr, "%d", groupid);
	sprintf(qs,"SELECT groupRead,groupDiffWrite,groupInstall,groupRawWrite FROM osimages WHERE id = '%d' limit 1",imageid);
	res = safe_mysql_query(&g_db_connection,qs);
	if(res) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -2;
	} else {
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if(mysql_num_rows(res_ptr)>0) {
		
				while((mysqlrow=mysql_fetch_row(res_ptr))) {
					if(mysqlrow[3] && strstr(mysqlrow[3],tmpstr))	//if in groupInstall
						*imageaccess=IMAGE_RAW_READWRITE;
					else if(mysqlrow[2] && strstr(mysqlrow[2],tmpstr))
						*imageaccess=IMAGE_INSTALL_RWRITE;
					else if(mysqlrow[1] && strstr(mysqlrow[1],tmpstr))
						*imageaccess=IMAGE_UDF_READWRITE;
					else if(mysqlrow[0] && strstr(mysqlrow[0],tmpstr))
						*imageaccess=IMAGE_READONLY;
					else
						*imageaccess=IMAGE_NO_RIGHTS;		//no rights to access the image
        }	

        //DEBUG("imageaccess = %d\n", (int) *imageaccess);        
			} else {
        		ERR("get_imageaccess_by_gid: sql returns no result\n");
        		goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				ERR("get_imageaccess_by_gid retrieve error: %s\n",mysql_error(&g_db_connection));
        		goto sql_error;
			}

      mysql_free_result(res_ptr);
			
		} 

	}

  return 0;
  
sql_error:
  DEBUG("SQL:%s\n", qs);
  mysql_free_result(res_ptr);
	return -1;
}

char *get_default_server_config(int serverId,char *configName)
{	
	char qs[MAX_STRING_LEN];
	int  res;
	MYSQL_RES *res_ptr;
	MYSQL_ROW mysqlrow;
	char *configValue;

	//DEBUG("db: get_default_server_config \n");

	snprintf(qs,MAX_STRING_LEN,"select configValue from serverconfigs where serverId=%d and configName='%s' limit 1",serverId,configName);
	res = safe_mysql_query(&g_db_connection,qs); 
	if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return NULL;		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr))>0){
				mysqlrow=mysql_fetch_row(res_ptr);
				if(!(configValue=(char *)calloc(strlen(mysqlrow[0])+1,sizeof(char))))
				{
					ERR("get_default_server_config error:could not allocate space for configValue!");
					return NULL;
				}
				strcpy(configValue,mysqlrow[0]);				
				//DEBUG("configValue : %s\n",configValue);
			} else {
			        DEBUG("configValue: sql returns no result\n");
			        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}
			mysql_free_result(res_ptr);
		} 		
	}

	return configValue;

sql_error:
	DEBUG("SQL:%s\n", qs);
	mysql_free_result(res_ptr);
	return NULL;
}

char *get_udf_dir()
{
	return get_default_server_config(TC_SERVER_ID, "userDataPath");
}

char *get_image_dir()
{
	return get_default_server_config(TC_SERVER_ID, "osImagePath");
}

//////////////////////////////////////////////////////////////////////
// added by wming for status update. 
//////////////////////////////////////////////////////////////////////

int db_update_pcnt_info(char *imagename, int pcnt, float speed)
{
	char qs[MAX_STRING_LEN*3];	
	//char subqs[MAX_STRING_LEN*2];

	//DEBUG("db: insert_user_status\n");

	// sprintf(qs, "update progress from deploy_image where imagename = '%s'", imagename);
	sprintf(qs, "update deploy_image set progress='%d', speed='%f' where image_path = '%s'", pcnt, speed, imagename);

	//DEBUG("update sql:%s\n", qs);
	
	return safe_mysql_query(&g_db_connection,qs);		
}



// insert the client to the deploy_client table
int db_insert_deploy_client(uint32 ipadd, char *mac, uint32 machine_status)
{
	char qs[MAX_STRING_LEN*3];

	if (machine_status == CLT_MACHINE_STATUS_UNMOUNT)
		sprintf(qs,"insert into deploy_client (ip_addr, mac_addr, error) values(\'%s\', \'%02X-%02X-%02X-%02X-%02X-%02X\', \'disk\')", \
			inet_ntoa(*(struct in_addr *)&ipadd), \
			(unsigned char)mac[0], (unsigned char)mac[1], (unsigned char)mac[2], (unsigned char)mac[3], (unsigned char)mac[4], (unsigned char)mac[5]);
	else
		sprintf(qs,"insert into deploy_client (ip_addr, mac_addr) values(\'%s\', \'%02X-%02X-%02X-%02X-%02X-%02X\')", \
			inet_ntoa(*(struct in_addr *)&ipadd), \
			(unsigned char)mac[0], (unsigned char)mac[1], (unsigned char)mac[2], (unsigned char)mac[3], (unsigned char)mac[4], (unsigned char)mac[5]);

#ifdef STANDALONE_MODE
	printf("db_insert_deploy_client: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);
}


// delete the client from the deploy_client table
int db_delete_deploy_client(uint32 ipadd)
{
	char qs[MAX_STRING_LEN*3];	
	sprintf(qs,"delete from deploy_client where ip_addr = \'%s\'", inet_ntoa(*(struct in_addr *)&ipadd));
#ifdef STANDALONE_MODE
	printf("db_delete_deploy_client: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);	
}


// update the deploy_client table entry
int db_update_deploy_client(uint32 ipadd, char *status)
{
	char qs[MAX_STRING_LEN*3];

	sprintf(qs,"update deploy_client set status=\'%s\' where ip_addr=\'%s\'", status, inet_ntoa(*(struct in_addr *)&ipadd));
	
#ifdef STANDALONE_MODE
	printf("db_update_deploy_client: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);	
}

// update the deploy_client table entry
int db_update_deploy_client_with_machine_status(uint32 ipadd, char *status, uint32 machine_status)
{
	char qs[MAX_STRING_LEN*3];

	if (machine_status == CLT_MACHINE_STATUS_UNMOUNT)
		sprintf(qs,"update deploy_client set status=\'%s\', error='disk' where ip_addr=\'%s\'", status, inet_ntoa(*(struct in_addr *)&ipadd));
	else if (machine_status == CLT_MACHINE_STATUS_INIT_ILLEGAL_SIZE)
		sprintf(qs,"update deploy_client set status=\'%s\', error='reserved_size' where ip_addr=\'%s\'", status, inet_ntoa(*(struct in_addr *)&ipadd));
	else if (machine_status == CLT_MACHINE_STATUS_INIT_HD_FAILURE)
		sprintf(qs,"update deploy_client set status=\'%s\', error='initialization' where ip_addr=\'%s\'", status, inet_ntoa(*(struct in_addr *)&ipadd));
	else
		sprintf(qs,"update deploy_client set status=\'%s\', error='NULL' where ip_addr=\'%s\'", status, inet_ntoa(*(struct in_addr *)&ipadd));
	
#ifdef STANDALONE_MODE
	printf("db_update_deploy_client: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);	
}


// making the deploy_client entries to "waiting"
int db_update_deploy_client_all_waiting()
{
	char qs[MAX_STRING_LEN*3];
	sprintf(qs,"update deploy_client set status=\'%s\'", "waiting");
	
#ifdef STANDALONE_MODE
	printf("db_update_deploy_client_all_waiting: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);
}

// making the deploy_client entries to "preparing"
int db_update_deploy_client_all_preparing()
{
	char qs[MAX_STRING_LEN*3];
	sprintf(qs,"update deploy_client set status=\'%s\'", "preparing");
	
#ifdef STANDALONE_MODE
	printf("db_update_deploy_client_all_waiting: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);
}


// clear all clients from deploy_client table
int db_clear_deploy_client()
{
	char qs[MAX_STRING_LEN*3];	
	sprintf(qs,"delete from deploy_client");
#ifdef STANDALONE_MODE
	printf("db_clear_deploy_client: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);		
}

// count the number of clients labeled with given status 'status'
int db_count_deploy_client_with_status(const char *status)
{
	int ret;
	char qs[MAX_STRING_LEN*3];	
	MYSQL_RES *res_ptr;

	sprintf(qs, "SELECT id FROM deploy_client WHERE status=\'%s\'", status);
  	ret = safe_mysql_query(&g_db_connection,qs);

  	if (ret)
	{
		perror("db_count_deploy_client_status: safe_mysql_query");
		return -1;
	}
	else
	{
		res_ptr = mysql_store_result(&g_db_connection);
	    	if(res_ptr)
		{
			return mysql_num_rows(res_ptr);
		}
		else
		{
			perror("db_count_deploy_client_status: mysql_store_result");
			return -1;
		}
	}
}

// insert deploy_history table
int db_insert_deploy_history(char *evt_type, char *evt_text)
{
	char qs[MAX_STRING_LEN*3];
	sprintf(qs,"insert into deploy_history(event_type, event_text) values(\'%s\', \'%s\');", evt_type, evt_text);
#ifdef STANDALONE_MODE
	printf("db_insert_deploy_history: qs=<%s>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);		
}


// check whether this client has been registered. If yes, return 0, else return -1
int db_chkreg_client(char *mac)
{
	char qs[MAX_STRING_LEN*3];
	MYSQL_RES *res_ptr;
	
	sprintf(qs,"select * from machines where mac = \'%02X-%02X-%02X-%02X-%02X-%02X\';", \
		(unsigned char)mac[0], (unsigned char)mac[1], (unsigned char)mac[2], (unsigned char)mac[3], (unsigned char)mac[4], (unsigned char)mac[5]);

#ifdef STANDALONE_MODE
	printf("db_chkreg_client: qs=<'%s'>\n", qs);
#endif

	if (safe_mysql_query(&g_db_connection,qs))
	{
		perror("db_chkreg_client: safe_mysql_query");
		return -1;	
	}

	res_ptr = mysql_store_result(&g_db_connection);
    	if(res_ptr)
	{
		if((mysql_num_rows(res_ptr))>0)
		{
#ifdef STANDALONE_MODE
			printf("db_chkreg_client: find  rows\n");
#endif
			return 0;
		}
		else 
		{
#ifdef STANDALONE_MODE
			printf("db_chkreg_client: no rows\n");
#endif
			return -1;
		}
	}
	else
	{
		perror("db_chkreg_client: mysql_store_result");
		return -1;
	}

}


// check whether this client has been registered
int db_reg_client(char *mac, char *hn, TCDMSG *msg)
{
	char qs[MAX_STRING_LEN*3];
	MYSQL_RES *res_ptr;

	// check whether the machine name is duplicated ... 
	sprintf(qs,"select * from machines where machineName = \'%s\';", hn);
#ifdef STANDALONE_MODE
	printf("db_reg_client: qs=<'%s'>\n", qs);
#endif

	if (safe_mysql_query(&g_db_connection,qs))
	{
		perror("db_reg_client: safe_mysql_query");
		return -1;	
	}

	res_ptr = mysql_store_result(&g_db_connection);
    	if(res_ptr)
	{
		if((mysql_num_rows(res_ptr))>0)
		{
			// machine name duplicated.
#ifdef STANDALONE_MODE
			printf("db_reg_client: find  rows\n");
#endif
			return -2;
		}
		else 
		{
			// machine name not duplicated, so can insert to database
#ifdef STANDALONE_MODE
			printf("db_reg_client: no rows\n");
#endif

			//sprintf(qs,"insert into machines (mac, machineName, memo) values (\'%02X-%02X-%02X-%02X-%02X-%02X\', \'%s\', \'%s\');", \
				(unsigned char)mac[0], (unsigned char)mac[1], (unsigned char)mac[2], (unsigned char)mac[3], (unsigned char)mac[4], (unsigned char)mac[5], \
				hn, hn);
			sprintf(qs,"insert into machines (mac, machineName, memo, cpu_model, memory, disk) values (\'%02X-%02X-%02X-%02X-%02X-%02X\', \'%s\', \'%s\', \'%s\', \'%u\', \'%u\');", 
				(unsigned char)mac[0], (unsigned char)mac[1], (unsigned char)mac[2], (unsigned char)mac[3], (unsigned char)mac[4], (unsigned char)mac[5], 
				hn, hn,
                                msg->u.rg.cpumodelname,
                                msg->u.rg.totalram,
                                msg->u.rg.diskcapacity);
#ifdef STANDALONE_MODE
			printf("db_chkreg_client: qs=<%s>\n", qs);
#endif
			return safe_mysql_query(&g_db_connection,qs);
		}
	}
	else
	{
		perror("db_reg_client: mysql_store_result");
		return -1;
	}

}

// update this client if registered
int db_update_client(char *mac, char *hn, TCDMSG *msg)
{
	char qs[MAX_STRING_LEN*3];

	// update this entry with new values
	sprintf(qs, "update machines set cpu_model=\'%s\', memory=\'%u\', disk=\'%u\' where mac=\'%02X-%02X-%02X-%02X-%02X-%02X\';", 
		msg->u.cur.cpumodelname,
		msg->u.cur.totalram,
		msg->u.cur.diskcapacity,
		(unsigned char)mac[0], (unsigned char)mac[1], (unsigned char)mac[2],
		(unsigned char)mac[3], (unsigned char)mac[4], (unsigned char)mac[5]
	);
#ifdef STANDALONE_MODE
	printf("db_update_client: qs=<%s>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);

}


// to update the user status table: altering all entries whose status from deploying to waiting. 
int db_update_user_status_all_waiting()
{
	char qs[MAX_STRING_LEN*3];
	sprintf(qs,"update userstatus set client_status=\'%s\', userId=\'\', osImageId=\'\' where client_status='\%s\'", "waiting", "deploying");
	
#ifdef STANDALONE_MODE
	printf("db_update_user_status_all_waiting: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);
}

// update userstatus set client_status='unknown' where Unix_timestamp() - heartbeat > 3

// to check if there are any clients who were not active for a certain period of time. If so, change the status to unknown
int db_update_user_status_check_inactive_clients()
{
	char qs[MAX_STRING_LEN*3];
//	sprintf(qs,"update userstatus set client_status=\'unknown\' where Unix_timestamp() - heartbeat > 15 and client_status <> 'intializing_disk' and client_status <> 'updating_firmware' and client_status <> 'deploying' and client_status <> 'interrupted' and client_status <> 'waiting'");

	sprintf(qs,"update userstatus set client_status=\'unknown\' where Unix_timestamp() - heartbeat > 15");
	
#ifdef STANDALONE_MODE
	printf("db_update_user_status_check_inactive: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);
}

#ifdef POST_REPORT_HWINFO
// check the entry with mac address, if exist, update the userid/osid, if not, insert a new entry
int db_update_user_status(char *mac, uint32 ipadd, char *uname, char *osid, uint32 act, int disksz, int memsz, const char *strcpu)
{
    int j;
    int t;
    char str_hwaddr[18];
    char str_ipaddr[16];
    //char str_currtm[20];
	char qs[MAX_STRING_LEN*3];
	MYSQL_RES *res_ptr, *res;
	char *status_str;
	char uid[128];
	MYSQL_ROW mysqlrow;

    char disksz_gib[32];

    // if memsz is 0
    // think of the coming entry is invalid
    // and do not update/insert it into the table
    if (memsz == 0)
    {
        return -1;
    }

    // convert mac data to a string
    for (j = 0; j < 6; j++)
    {
        sprintf(&str_hwaddr[3 * j], "%02X", (unsigned char)mac[j]);

        if (j < 5)
            sprintf(&str_hwaddr[3 * j + 2], "-");
    }

    // convert ip data to a string
    strncpy(str_ipaddr, inet_ntoa(*(struct in_addr *)&ipadd), sizeof(str_ipaddr));
    str_ipaddr[sizeof(str_ipaddr) - 1] = '\0';

    // get current time and save it to a string
    // Do Not use str_currtm
    // Use Unix_timestamp() from MySql instead
    //clktmstr(str_currtm, sizeof(str_currtm), "%F %T");

    // convert disk size from MiB to GiB
    sprintf(disksz_gib, "%.2f", disksz / 1024.0);

    /*tcd_syslog("mac[%s], ip[%s], uname[%s], osid[%s], status[%d], disksz[%d], memsz[%d], cpu[%s]\n",
               str_hwaddr, str_ipaddr, uname, osid, act, disksz, memsz, strcpu);*/

/*
#define ACT_POWERON		1	// nothing
#define ACT_AUTHENTICATED	2	// userid
#define ACT_SYNCING		3	// userid and osid
#define ACT_BOOTING		4	// userid and osid
#define ACT_RUNNING		5	// nothing
#define ACT_SHUTTING_DOWN	6	// nothing
#define ACT_UPLOADING		7	// userid and osid

#define ACT_INITIALIZING_DISK	8	// nothing
#define ACT_UPDATING_FIRMWARE	90	// nothing
#define ACT_UPDATING_FIRMWARE_CMD_SENT	91	// nothing
#define ACT_UPDATING_FIRMWARE_SUCCESS	92	// nothing
#define ACT_UPDATING_FIRMWARE_FAILURE	93	// nothing
#define ACT_DEPLOYING		10	// nothing
#define ACT_INTERRUPTED		11	// nothing
#define ACT_WAITING		12	// nothing

#define ACT_HEARTBEAT		17	// nothing
*/

	switch(act)
	{
		case ACT_POWERON:
			status_str = "power_on";
			break;
		case ACT_AUTHENTICATED:
			status_str = "authenticated";
			break;
		case ACT_SYNCING:
			status_str = "syncing";
			break;
		case ACT_BOOTING:
			status_str = "booting";
			break;
		case ACT_RUNNING:
			status_str = "running";
			break;
		case ACT_SHUTTING_DOWN:
			// status_str = "shutting_down";
			status_str = "power_off";
			break;
		case ACT_UPLOADING:
			status_str = "uploading";
			break;

		case ACT_INITIALIZING_DISK:
			status_str = "initializing_disk";
			break;
		case ACT_UPDATING_FIRMWARE_CMD_SENT:
			status_str = "updating_firmware_cmd_sent";
			break;
		case ACT_UPDATING_FIRMWARE:
			status_str = "updating_firmware";
			break;
		case ACT_UPDATING_FIRMWARE_SUCCESS:
			status_str = "updating_firmware_done";
			break;
		case ACT_UPDATING_FIRMWARE_FAILURE:
			status_str = "updating_firmware_failed";
			break;
		case ACT_DEPLOYING:
			status_str = "deploying";
			break;
		case ACT_INTERRUPTED:
			status_str = "interrupted";
			break;
		case ACT_WAITING:
			status_str = "waiting";
			break;

		case ACT_HEARTBEAT:
			// status_str = "waiting";
#ifdef STANDALONE_MODE
			printf("heartbeat...\n");
#endif
			//return 0;
			break;

		default: 
			status_str = "unknown";
			break;
	}


	if (act == ACT_AUTHENTICATED || act == ACT_SYNCING || act == ACT_BOOTING || act == ACT_UPLOADING || act == ACT_RUNNING) // need userid
	{
		sprintf(qs, "select id from users where userName=\'%s\'", uname);
		t = safe_mysql_query(&g_db_connection, qs);
		if(t)
		{
			perror("get user status");
			tcd_syslog("get user status\n");
			return -1;
		}
		else
		{
			res_ptr = mysql_store_result(&g_db_connection);
			if(res_ptr) 
			{
				if((mysql_num_rows(res_ptr))>0)
				{
					mysqlrow = mysql_fetch_row(res_ptr);
					strcpy(uid, mysqlrow[0]);
			        mysql_free_result(res_ptr);
				}
				else
				{
#ifdef STANDALONE_MODE
					fprintf(stderr, "failed to get id for username %s", uname);
#endif
					tcd_syslog("failed to get id for username %s\n", uname);
			        mysql_free_result(res_ptr);
					return -1;
				}
			}
			else
			{
#ifdef STANDALONE_MODE
				fprintf(stderr, "failed to store result for username %s", uname);
#endif
				tcd_syslog("failed to store result for username %s\n", uname);
				return -1;
			}
		}
	}
	// uid ...

	// check whether the machine MAC is there ... 
	sprintf(qs,"select * from userstatus where mac = \'%s\';", str_hwaddr);
#ifdef STANDALONE_MODE
	printf("db_update_user_status: qs=<'%s'>\n", qs);
#endif
	if (safe_mysql_query(&g_db_connection, qs))
	{
		perror("db_update_user_status: safe_mysql_query");
		tcd_syslog("db_update_user_status: safe_mysql_query\n");
		return -1;
	}
	res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr)
	{
		if((mysql_num_rows(res_ptr))>0)
		{
            //tcd_syslog("mac exists\n");
			// MAC exists, we should update

			if (act == ACT_AUTHENTICATED)	// only uid
			{
				sprintf(qs,"update userstatus set userId=\'%s\', osImageId=\'\', client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp(), disk_size=\'%s\', memory_size=\'%d\', cpu_model=\'%s\' where mac=\'%s\';", \
					uid, status_str, str_ipaddr, /*str_currtm,*/ disksz_gib, memsz, strcpu, str_hwaddr); 
			}
			else if (act == ACT_SYNCING || act == ACT_BOOTING || act == ACT_UPLOADING)	// both uid and osid
			{
				sprintf(qs,"update userstatus set userId=\'%s\', osImageId=\'%s\', client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp(), disk_size=\'%s\', memory_size=\'%d\', cpu_model=\'%s\' where mac=\'%s\';", uid, osid, status_str, str_ipaddr, /*str_currtm,*/ disksz_gib, memsz, strcpu, str_hwaddr); 
			}

			else if (act == ACT_POWERON) // nothing needed but need to clear user id and os image id. 
			{
				sprintf(qs,"update userstatus set userId=\'\', osImageId=\'\', client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp(), disk_size=\'%s\', memory_size=\'%d\', cpu_model=\'%s\' where mac=\'%s\';", \
					status_str, str_ipaddr, /*str_currtm,*/ disksz_gib, memsz, strcpu, str_hwaddr); 
			}
			else if (act == ACT_SHUTTING_DOWN || act == ACT_INITIALIZING_DISK || act == ACT_UPDATING_FIRMWARE || act == ACT_UPDATING_FIRMWARE_CMD_SENT || act == ACT_UPDATING_FIRMWARE_SUCCESS || act == ACT_UPDATING_FIRMWARE_FAILURE )	// USER ID AND OS IMAGE SET TO Null
			{
				sprintf(qs,"update userstatus set userId=\'\', osImageId=\'\', client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp(), disk_size=\'%s\', memory_size=\'%d\', cpu_model=\'%s\' where mac=\'%s\';", \
					status_str, str_ipaddr, /*str_currtm,*/ disksz_gib, memsz, strcpu, str_hwaddr); 
			}
			else if (act == ACT_RUNNING)	// all
			{
				sprintf(qs,"update userstatus set userId=\'%s\', osImageId=\'%s\', client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp(), disk_size=\'%s\', memory_size=\'%d\', cpu_model=\'%s\' where mac=\'%s\';", uid, osid, status_str, str_ipaddr, /*str_currtm,*/ disksz_gib, memsz, strcpu, str_hwaddr); 
			}
			else	// nothing needed and keep user id and os image id
			{
				sprintf(qs,"update userstatus set client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp(), disk_size=\'%s\', memory_size=\'%d\', cpu_model=\'%s\' where mac=\'%s\';", \
					status_str, str_ipaddr, /*str_currtm,*/ disksz_gib, memsz, strcpu, str_hwaddr); 
			}


#ifdef STANDALONE_MODE
			printf("db_update_user_status: qs=<%s>\n", qs);
#endif
			mysql_free_result(res_ptr);
			return safe_mysql_query(&g_db_connection,qs);
		}
		else
		{
			// MAC not exist, we should insert
            //tcd_syslog("mac does not exists\n");

			if (act == ACT_AUTHENTICATED)	// only uid
			{
				sprintf(qs,"insert into userstatus (mac, userId, osImageId, client_status, ip, heartbeat, disk_size, memory_size, cpu_model) values (\'%s\', \'%s\', \'\', \'%s\', \'%s\', Unix_timestamp(), \'%s\', \'%d\', \'%s\');", \
					str_hwaddr, \
					uid, status_str, str_ipaddr/*, str_currtm*/, disksz_gib, memsz, strcpu);
			}
			else if (act == ACT_SYNCING || act == ACT_BOOTING || act == ACT_UPLOADING)	// both uid and osid
			{
				sprintf(qs,"insert into userstatus (mac, userId, osImageId, client_status, ip, heartbeat, disk_size, memory_size, cpu_model) values (\'%s\', \'%s\', \'%s\', \'%s\', \'%s\', Unix_timestamp(), \'%s\', \'%d\', \'%s\');", \
					str_hwaddr, \
					uid, osid, status_str, str_ipaddr/*, str_currtm*/, disksz_gib, memsz, strcpu);
			}

			else if (act == ACT_POWERON)	// nothing needed but need to clear user id and os image id
			{
				sprintf(qs,"insert into userstatus (mac, userId, osImageId, client_status, ip, heartbeat, disk_size, memory_size, cpu_model) values (\'%s\', \'\', \'\', \'%s\', \'%s\', Unix_timestamp(), \'%s\', \'%d\', \'%s\');", \
					str_hwaddr, \
					status_str, str_ipaddr/*, str_currtm*/, disksz_gib, memsz, strcpu);
			}
			else if (act == ACT_RUNNING)	// all
			{
				sprintf(qs,"insert into userstatus (mac, userId, osImageId, client_status, ip, heartbeat, disk_size, memory_size, cpu_model) values (\'%s\', \'%s\', \'%s\', \'%s\', \'%s\', Unix_timestamp(), \'%s\', \'%d\', \'%s\');", \
					str_hwaddr, \
					uid, osid, status_str, str_ipaddr/*, str_currtm*/, disksz_gib, memsz, strcpu);
			}

			else	// nothing needed and keep user id and os image id
			{
				sprintf(qs,"insert into userstatus (mac, client_status, ip, heartbeat, disk_size, memory_size, cpu_model) values (\'%s\', \'%s\', \'%s\', Unix_timestamp(), \'%s\', \'%d\', \'%s\');", \
					str_hwaddr, \
					status_str, str_ipaddr /*,str_currtm*/, disksz_gib, memsz, strcpu);
			}


#ifdef STANDALONE_MODE
			printf("db_update_user_status: qs=<%s>\n", qs);
#endif
			mysql_free_result(res_ptr);
			return safe_mysql_query(&g_db_connection,qs);
		}
	}
	else
	{
		perror("db_update_user_status: mysql_store_result");	
		return -1;
	}
}
#else // POST_REPORT_HWINFO

// check the entry with mac address, if exist, update the userid/osid, if not, insert a new entry
int db_update_user_status(char *mac, uint32 ipadd, char *uname, char *osid, uint32 act)
{
    int j;
    int t;
    char str_hwaddr[18];
    char str_ipaddr[16];
    //char str_currtm[20];
	char qs[MAX_STRING_LEN*3];
	MYSQL_RES *res_ptr, *res;
	char *status_str;
	char uid[128];
	MYSQL_ROW mysqlrow;

    // convert mac data to a string
    for (j = 0; j < 6; j++)
    {
        sprintf(&str_hwaddr[3 * j], "%02X", (unsigned char)mac[j]);

        if (j < 5)
            sprintf(&str_hwaddr[3 * j + 2], "-");
    }

    // convert ip data to a string
    strncpy(str_ipaddr, inet_ntoa(*(struct in_addr *)&ipadd), sizeof(str_ipaddr));
    str_ipaddr[sizeof(str_ipaddr) - 1] = '\0';

    // get current time and save it to a string
    // Do Not use str_currtm
    // Use Unix_timestamp() from MySql instead
    //clktmstr(str_currtm, sizeof(str_currtm), "%F %T");

/*
#define ACT_POWERON		1	// nothing
#define ACT_AUTHENTICATED	2	// userid
#define ACT_SYNCING		3	// userid and osid
#define ACT_BOOTING		4	// userid and osid
#define ACT_RUNNING		5	// nothing
#define ACT_SHUTTING_DOWN	6	// nothing
#define ACT_UPLOADING		7	// userid and osid

#define ACT_INITIALIZING_DISK	8	// nothing
#define ACT_UPDATING_FIRMWARE	90	// nothing
#define ACT_UPDATING_FIRMWARE_CMD_SENT	91	// nothing
#define ACT_UPDATING_FIRMWARE_SUCCESS	92	// nothing
#define ACT_UPDATING_FIRMWARE_FAILURE	93	// nothing
#define ACT_DEPLOYING		10	// nothing
#define ACT_INTERRUPTED		11	// nothing
#define ACT_WAITING		12	// nothing

#define ACT_HEARTBEAT		17	// nothing
*/

	switch(act)
	{
		case ACT_POWERON:
			status_str = "power_on";
			break;
		case ACT_AUTHENTICATED:
			status_str = "authenticated";
			break;
		case ACT_SYNCING:
			status_str = "syncing";
			break;
		case ACT_BOOTING:
			status_str = "booting";
			break;
		case ACT_RUNNING:
			status_str = "running";
			break;
		case ACT_SHUTTING_DOWN:
			// status_str = "shutting_down";
			status_str = "power_off";
			break;
		case ACT_UPLOADING:
			status_str = "uploading";
			break;

		case ACT_INITIALIZING_DISK:
			status_str = "initializing_disk";
			break;
		case ACT_UPDATING_FIRMWARE_CMD_SENT:
			status_str = "updating_firmware_cmd_sent";
			break;
		case ACT_UPDATING_FIRMWARE:
			status_str = "updating_firmware";
			break;
		case ACT_UPDATING_FIRMWARE_SUCCESS:
			status_str = "updating_firmware_done";
			break;
		case ACT_UPDATING_FIRMWARE_FAILURE:
			status_str = "updating_firmware_failed";
			break;
		case ACT_DEPLOYING:
			status_str = "deploying";
			break;
		case ACT_INTERRUPTED:
			status_str = "interrupted";
			break;
		case ACT_WAITING:
			status_str = "waiting";
			break;

		case ACT_HEARTBEAT:
			// status_str = "waiting";
#ifdef STANDALONE_MODE
			printf("heartbeat...\n");
#endif
			//return 0;
			break;

		default: 
			status_str = "unknown";
			break;
	}


	if (act == ACT_AUTHENTICATED || act == ACT_SYNCING || act == ACT_BOOTING || act == ACT_UPLOADING) // need userid
	{
		sprintf(qs, "select id from users where userName=\'%s\'", uname);
		t = safe_mysql_query(&g_db_connection, qs);
		if(t)
		{
			perror("get user status");
			return -1;
		}
		else
		{
			res_ptr = mysql_store_result(&g_db_connection);
			if(res_ptr) 
			{
				if((mysql_num_rows(res_ptr))>0)
				{
					mysqlrow = mysql_fetch_row(res_ptr);
					strcpy(uid, mysqlrow[0]);
				}
				else
				{
#ifdef STANDALONE_MODE
					fprintf(stderr, "failed to get id for username %s", uname);
#endif
					return -1;
				}
			}
			else
			{
#ifdef STANDALONE_MODE
				fprintf(stderr, "failed to store result for username %s", uname);
#endif
				return -1;
			}
		}
	}
	// uid ...

	// check whether the machine MAC is there ... 
	sprintf(qs,"select * from userstatus where mac = \'%s\';", str_hwaddr);
#ifdef STANDALONE_MODE
	printf("db_update_user_status: qs=<'%s'>\n", qs);
#endif
	if (safe_mysql_query(&g_db_connection, qs))
	{
		perror("db_update_user_status: safe_mysql_query");
		return -1;
	}
	res_ptr = mysql_store_result(&g_db_connection);
    	if(res_ptr)
	{
		if((mysql_num_rows(res_ptr))>0)
		{
			// MAC exists, we should update

			if (act == ACT_AUTHENTICATED)	// only uid
			{
				sprintf(qs,"update userstatus set userId=\'%s\', osImageId=\'\', client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp() where mac=\'%s\';", \
					uid, status_str, str_ipaddr, /*str_currtm,*/ str_hwaddr); 
			}
			else if (act == ACT_SYNCING || act == ACT_BOOTING || act == ACT_UPLOADING)	// both uid and osid
			{
				sprintf(qs,"update userstatus set userId=\'%s\', osImageId=\'%s\', client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp() where mac=\'%s\';", uid, osid, status_str, str_ipaddr, /*str_currtm,*/ str_hwaddr); 
			}

			else if (act == ACT_POWERON) // nothing needed but need to clear user id and os image id. 
			{
				sprintf(qs,"update userstatus set userId=\'\', osImageId=\'\', client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp() where mac=\'%s\';", \
					status_str, str_ipaddr, /*str_currtm,*/ str_hwaddr); 
			}
			else if (act == ACT_SHUTTING_DOWN || act == ACT_INITIALIZING_DISK || act == ACT_UPDATING_FIRMWARE || act == ACT_UPDATING_FIRMWARE_CMD_SENT || act == ACT_UPDATING_FIRMWARE_SUCCESS || act == ACT_UPDATING_FIRMWARE_FAILURE )	// USER ID AND OS IMAGE SET TO Null
			{
				sprintf(qs,"update userstatus set userId=\'\', osImageId=\'\', client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp() where mac=\'%s\';", \
					status_str, str_ipaddr, /*str_currtm,*/ str_hwaddr); 
			}
			else	// nothing needed and keep user id and os image id
			{
				sprintf(qs,"update userstatus set client_status=\'%s\', ip=\'%s\', heartbeat=Unix_timestamp() where mac=\'%s\';", \
					status_str, str_ipaddr, /*str_currtm,*/ str_hwaddr); 
			}


#ifdef STANDALONE_MODE
			printf("db_update_user_status: qs=<%s>\n", qs);
#endif
			return safe_mysql_query(&g_db_connection,qs);
		}
		else
		{
			// MAC not exist, we should insert

			if (act == ACT_AUTHENTICATED)	// only uid
			{
				sprintf(qs,"insert into userstatus (mac, userId, osImageId, client_status, ip, heartbeat) values (\'%s\', \'%s\', \'\', \'%s\', \'%s\', Unix_timestamp());", \
					str_hwaddr, \
					uid, status_str, str_ipaddr/*, str_currtm*/);
			}
			else if (act == ACT_SYNCING || act == ACT_BOOTING || act == ACT_UPLOADING)	// both uid and osid
			{
				sprintf(qs,"insert into userstatus (mac, userId, osImageId, client_status, ip, heartbeat) values (\'%s\', \'%s\', \'%s\', \'%s\', \'%s\', Unix_timestamp());", \
					str_hwaddr, \
					uid, osid, status_str, str_ipaddr/*, str_currtm*/);
			}

			else if (act == ACT_POWERON)	// nothing needed but need to clear user id and os image id
			{
				sprintf(qs,"insert into userstatus (mac, userId, osImageId, client_status, ip, heartbeat) values (\'%s\', \'\', \'\', \'%s\', \'%s\', Unix_timestamp());", \
					str_hwaddr, \
					status_str, str_ipaddr/*, str_currtm*/);
			}
			else if (act == ACT_RUNNING)	// both uid and osid
			{
				sprintf(qs,"insert into userstatus (mac, userId, osImageId, client_status, ip, heartbeat) values (\'%s\', \'%s\', \'%s\', \'%s\', \'%s\', Unix_timestamp());", \
					str_hwaddr, \
					uid, osid, status_str, str_ipaddr/*, str_currtm*/);
			}

			else	// nothing needed and keep user id and os image id
			{
				sprintf(qs,"insert into userstatus (mac, client_status, ip, heartbeat) values (\'%s\', \'%s\', \'%s\', Unix_timestamp());", \
					str_hwaddr, \
					status_str, str_ipaddr /*,str_currtm*/);
			}


#ifdef STANDALONE_MODE
			printf("db_update_user_status: qs=<%s>\n", qs);
#endif
			return safe_mysql_query(&g_db_connection,qs);
		}
	}
	else
	{
		perror("db_update_user_status: mysql_store_result");	
		return -1;
	}
}
#endif // POST_REPORT_HWINFO


// clear user status table
int db_clear_user_status()
{
	char qs[MAX_STRING_LEN*3];	
	sprintf(qs,"delete from userstatus");
#ifdef STANDALONE_MODE
	printf("db_clear_user_status: qs=<'%s'>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);		
}

//DELETE FROM userstatus WHERE NOW() - heartbeat < 10 ;
int db_refresh_user_status()
{
    char qs[MAX_STRING_LEN*3];
    sprintf(qs, "DELETE FROM userstatus WHERE NOW() - heartbeat > \'%d\' ;", REFRESH_STATUS_EXPIRE_INTERVAL);
#ifdef STANDALONE_MODE
    printf("db_refresh_user_status: qs=<'%s'>\n", qs);
#endif
    return safe_mysql_query(&g_db_connection, qs);
}

// update a filed for an existing userstatus entry
int db_update_existing_user_status_by_ip(const char* ipv4, uint32 act)
{
	char qs[MAX_STRING_LEN*3];
	MYSQL_RES *res_ptr;
	char *status_str;

    tcd_syslog("db_update_existing_user_status_by_ip(): ip[%s], status[%d]\n", ipv4, act);

	switch(act)
	{
		case ACT_POWERON:
			status_str = "power_on";
			break;
		case ACT_AUTHENTICATED:
			status_str = "authenticated";
			break;
		case ACT_SYNCING:
			status_str = "syncing";
			break;
		case ACT_BOOTING:
			status_str = "booting";
			break;
		case ACT_RUNNING:
			status_str = "running";
			break;
		case ACT_SHUTTING_DOWN:
			// status_str = "shutting_down";
			status_str = "power_off";
			break;
		case ACT_UPLOADING:
			status_str = "uploading";
			break;

		case ACT_INITIALIZING_DISK:
			status_str = "initializing_disk";
			break;
		case ACT_UPDATING_FIRMWARE_CMD_SENT:
			status_str = "updating_firmware_cmd_sent";
			break;
		case ACT_UPDATING_FIRMWARE:
			status_str = "updating_firmware";
			break;
		case ACT_UPDATING_FIRMWARE_SUCCESS:
			status_str = "updating_firmware_done";
			break;
		case ACT_UPDATING_FIRMWARE_FAILURE:
			status_str = "updating_firmware_failed";
			break;
		case ACT_DEPLOYING:
			status_str = "deploying";
			break;
		case ACT_INTERRUPTED:
			status_str = "interrupted";
			break;
		case ACT_WAITING:
			status_str = "waiting";
			break;

		case ACT_HEARTBEAT:
			// status_str = "waiting";
#ifdef STANDALONE_MODE
			printf("heartbeat...\n");
#endif
			//return 0;
			break;

		default: 
			status_str = "unknown";
			break;
	}

	// check whether the ip is there ... 
	sprintf(qs,"select * from userstatus where ip = \'%s\';", ipv4);
#ifdef STANDALONE_MODE
	printf("db_update_existing_user_status_by_ip: qs=<'%s'>\n", qs);
#endif
	if (safe_mysql_query(&g_db_connection, qs))
	{
		perror("db_update_existing_user_status_by_ip: safe_mysql_query");
		tcd_syslog("db_update_existing_user_status_by_ip: safe_mysql_query\n");
		return -1;
	}

	res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr)
	{
		if((mysql_num_rows(res_ptr))>0)
		{
            tcd_syslog("db_update_existing_user_status_by_ip: ip exists\n");
			// IP exists, we should update

			sprintf(qs,"update userstatus set client_status=\'%s\' where ip=\'%s\';", status_str, ipv4);


#ifdef STANDALONE_MODE
			printf("db_update_existing_user_status_by_ip: qs=<%s>\n", qs);
#endif
			return safe_mysql_query(&g_db_connection,qs);
		}
		else
		{
			// IP does not exist, do nothing
            tcd_syslog("db_update_existing_user_status_by_ip: ip does exists\n");
            return -1;
		}
	}
	else
	{
		perror("db_update_existing_user_status_by_ip: mysql_store_result");	
		return -1;
	}
}

// update a filed for all existing userstatus entry
int db_update_all_existing_user_status(uint32 act)
{
	char qs[MAX_STRING_LEN*3];
	MYSQL_RES *res_ptr;
	char *status_str;

    tcd_syslog("db_update_all_existing_user_status(): status[%d]\n", act);

	switch(act)
	{
		case ACT_POWERON:
			status_str = "power_on";
			break;
		case ACT_AUTHENTICATED:
			status_str = "authenticated";
			break;
		case ACT_SYNCING:
			status_str = "syncing";
			break;
		case ACT_BOOTING:
			status_str = "booting";
			break;
		case ACT_RUNNING:
			status_str = "running";
			break;
		case ACT_SHUTTING_DOWN:
			// status_str = "shutting_down";
			status_str = "power_off";
			break;
		case ACT_UPLOADING:
			status_str = "uploading";
			break;

		case ACT_INITIALIZING_DISK:
			status_str = "initializing_disk";
			break;
		case ACT_UPDATING_FIRMWARE_CMD_SENT:
			status_str = "updating_firmware_cmd_sent";
			break;
		case ACT_UPDATING_FIRMWARE:
			status_str = "updating_firmware";
			break;
		case ACT_UPDATING_FIRMWARE_SUCCESS:
			status_str = "updating_firmware_done";
			break;
		case ACT_UPDATING_FIRMWARE_FAILURE:
			status_str = "updating_firmware_failed";
			break;
		case ACT_DEPLOYING:
			status_str = "deploying";
			break;
		case ACT_INTERRUPTED:
			status_str = "interrupted";
			break;
		case ACT_WAITING:
			status_str = "waiting";
			break;

		case ACT_HEARTBEAT:
			// status_str = "waiting";
#ifdef STANDALONE_MODE
			printf("heartbeat...\n");
#endif
			//return 0;
			break;

		default: 
			status_str = "unknown";
			break;
	}

    sprintf(qs,"update userstatus set client_status=\'%s\';", status_str);
#ifdef STANDALONE_MODE
	printf("db_update_all_existing_user_status: qs=<%s>\n", qs);
#endif
	return safe_mysql_query(&g_db_connection,qs);
}


// update heart-beat time for the entry
// this function shoud Not be called
//int db_update_hb_by_mac(const char *str_hwaddr, const char *str_new_hbtime)
//{
//	char qs[MAX_STRING_LEN*3];
//    sprintf(qs, "UPDATE userstatus SET heartbeat=\'%s\' WHERE mac=\'%s\';", str_new_hbtime, str_hwaddr);
//#ifdef STANDALONE_MODE
//    printf("db_update_hb_by_mac: qs=<'%s'>\n", qs);
//#endif
//    return safe_mysql_query(&g_db_connection, qs);
//}


// delete the entry from table
// this function shoud Not be called
//int db_delete_user_status_by_mac(const char *str_hwaddr)
//{
//	char qs[MAX_STRING_LEN*3];
//    sprintf(qs, "DELETE FROM userstatus WHERE mac = \'%s\';", str_hwaddr);
//#ifdef STANDALONE_MODE
//    printf("db_delete_user_status_by_mac: qs=<'%s'>\n", qs);
//#endif
//    return safe_mysql_query(&g_db_connection, qs);
//}


// if the entry expires, delete the entry from table
// this function shoud Not be called
//DELETE FROM userstatus WHERE NOW() - heartbeat < 10 ;
//int db_refresh_user_status()
//{
//    struct timespec ts = {0};
//
//    char str_hwaddr[32];
//    char str_old_hbtime[32];
//    char str_new_hbtime[32];
//    
//	char qs[MAX_STRING_LEN*3];
//	MYSQL_RES *res_ptr, *res;
//	char *status_str;
//	MYSQL_ROW mysqlrow;
//
//    //memset(str_hwaddr, 0, sizeof(str_hwaddr));
//    //memset(str_old_hbtime, 0, sizeof(str_old_hbtime));
//
//    sprintf(qs, "SELECT mac, heartbeat FROM userstatus ;");
//    res = safe_mysql_query(&g_db_connection, qs);
//    if(res)
//    {
//        perror("get user status");
//        return -1;
//    }
//    else
//    {
//        res_ptr = mysql_store_result(&g_db_connection);
//        if(res_ptr) 
//        {
//            while (mysqlrow = mysql_fetch_row(res_ptr))
//            {
//                strcpy(str_hwaddr, mysqlrow[0]);
//                strcpy(str_old_hbtime, mysqlrow[1]);
//
//                // get current time
//                // and compare to the time string from db table
//                // if the interval exceeds the specified value, then remove the entry
//                clock_gettime(CLOCK_REALTIME, &ts);
//#define HB_REFRESH_TM 10
//                if (str2timet(str_old_hbtime, sizeof(str_old_hbtime)) + HB_REFRESH_TM < ts.tv_sec)
//                {
//                    db_delete_user_status_by_mac(str_hwaddr);
//                }
//            }
//        }
//        else
//        {
//#ifdef STANDALONE_MODE
//            printf("mysql_store_result: qs=<%s>\n", qs);
//            printf("mysql_store_result: fail to \n");
//#endif
//            return -1;
//        }
//    }
//    return 0;
//}



//////////////////////////////////////////////////////////////////////
// end of wming's status update
//////////////////////////////////////////////////////////////////////


#endif // USE_MYSQL



