/******************************************************************************
Copyright 2012 - 2020 Intel Corporation

For licensing information, see the file 'LICENSE' in the root folder of
this software module.
******************************************************************************/

#ifndef __DB_H__
#define __DB_H__

// get from Fujin's pre_define.h

//#ifndef __PREDEFINE_H__
//#define __PREDEFINE_H__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <stdint.h>

#include "tcd.h"

//Whether to verify the license file on server
#define VERIFY_LICENSE			1

// move the definition of USE_MYSQL to conf.h ...
// Whether to use mysql database to store server/client setttings
// #define USE_MYSQL 				1


//Whether to collect and print the performance data
#define SYS_PERF				0
//Whether to run the server in daemon mode 
#define RUN_IN_DAEMON_MODE		1
//Whether we will use giant mtu
#define GIANT_MTU_ENABLED		0
//Whether we will use a force timer to clean powered off client status
#define USE_POWERED_OFF_TIMER	0

#define TC_SECTOR_SIZE			512
#define SECTOR_SIZE_INDEX  		9


#if GIANT_MTU_ENABLED
#define TC_MAX_SECTORS         	8
#else
#define TC_MAX_SECTORS         	2
#endif

#define MAX_ETH_NAME_LEN		32
#define DFT_APP_NAME			"tc-server"
#define SINGLE_INST_LOCK_FORMAT	"/var/tmp/%s-%s-%d-%d.lock"
#define TC_MDF_FORMAT			"%s-%s.mdf"
#define TC_UDF_FORMAT			"%s-%s.udf"
#define	VAL_INTACT				((uint32_t)(-1))

#define	MAJOR(x)				((x) >> 24 & 0xFF)
#define	MINOR(x)				((x) & 0xffffff)
#define	MAKE_DEV(x, y)			((x) << 24 | (y))

#define MAX_STRING_LEN  		128
#define MAX_SI_FILE_LEN			80
#define UDF_PAGE_SIZE    		4096
// 4096 = 2 ^ 12
#define UDF_PAGE_INDEX    		12
#define DFT_UDF_DIR				"/tc/userdata"
#define DFT_IMAGE_DIR			"/tc/diskimage"
#define MAX_CLIENTS     		120
#define PKT_RCV_BUF_SIZE		(TC_MAX_SECTORS * TC_SECTOR_SIZE * 4)	//(1<<16)


#define MIN_HB_COUNT    		1
#define CHECK_TIMER_INTERVAL  	60
// 2 mins
#define HB_CHECK_TIMER  		CHECK_TIMER_INTERVAL*2
//if we hadn't received client's HB for longer than HB_LOST_TIMEOUT seconds,
//then we'll destory the client. -1 means never destory the client in this case
#define HB_LOST_TIMEOUT  		-1 //HB_CHECK_TIMER*2 
// 2 mins
#define RESET_CONF_INTERVAL   	HB_CHECK_TIMER
#define RW_BUF_CNT      		0x1000 
#define TC_SERVER_ID   			4
// 1 * HB_CHECK_TIMER
#define BOOT_TIMEOUT			0



typedef enum {
  POWERED_OFF = 0x0,
  AUTHENTICATED = 0x01,
  RUNNING = 0x02,  
  SYNCING = 0x03,
  SHUTDOWN_CMD_RECVED = 0x04,
  INVALID = 0x05
} CLIENT_STATUS;

typedef enum {
  SERVER_STATUS_OFFLINE = 0x0,
  SERVER_STATUS_ONLINE = 0x1,
  SERVER_STATUS_OVERLOAD = 0x2,
  SERVER_STATUS_NORMAL = 0x3
} SERVER_STATUS;

typedef enum {
	IMAGE_NO_RIGHTS=0x0,
	IMAGE_READONLY=0x1,
	IMAGE_UDF_READWRITE=0x2,	
	IMAGE_RAW_READWRITE=0x3,
	IMAGE_INSTALL_RWRITE=0x4	
}IMAGE_ACCESS;

/**
 * Variables associated with a server.
 **/
typedef struct {
	char 		*cur_dir;
	char		eth_name[32];		/**< nic used for license check >  **/
	char 		single_inst_file[MAX_SI_FILE_LEN];
	int 		socket;	     		/**< The socket of this server. */
	long 		major;   			/**< major number we're exporting this server at */
	long		minor;				/**< minor number we're exporting this server at */	
	int     	eth_mtu;
	int			conf_reply_secs;
	uint8_t	  	eth_mac[6];
	uint64_t	runTime;
} SERVER_CONFIG;


// #endif //__PREDEFINE_H__


// #ifndef __DEBUG_H__
// #define __DEBUG_H__

// #include "predefine.h"


#define LOG_SRV_NAME	"tc-server"
// Set ENABLE_DEBUG to 1 if you want a debugging version of tc-server 
// Otherwise set ENABLE_DEBUG to 0 
#define ENABLE_DEBUG 		1
//Whether to print the read/write detail infos
#define RW_DEBUG			0


#if RUN_IN_DAEMON_MODE
#define MSG(fmt,arg...) do {\
							syslog(LOG_INFO,fmt,##arg);\
							}while(0)
#define TC_OUTPUT(fmt,arg...) do {\
							syslog(LOG_ERR,fmt,##arg);\
							printf(fmt,##arg);\
							}while(0)

#define ERR(fmt,arg...) do {\
							syslog(LOG_ERR,fmt,##arg);\
							printf(fmt,##arg);\
							}while(0)

#define PERR(msg) 		do {\
							syslog(LOG_ERR,msg);\
							printf(msg);\
							}while(0)

#if ENABLE_DEBUG
#define DEBUG(fmt,arg...)     do {\
								syslog(LOG_INFO,fmt,##arg);\
							  }while(0)
                               
                                
#else
#define DEBUG(fmt,arg...)
#endif

#else
#define MSG(fmt,arg...)         do {\
                                  printf(fmt,##arg);\
                                } while(0)

#define TC_OUTPUT(fmt,arg...)   do {\
                                  printf(fmt,##arg);\
                                } while(0)
                        
#define ERR(fmt,arg...)         do {\
                                  printf(fmt,##arg);\
                                } while(0) 

#define PERR(msg)        		do {\
                                  perror(msg);\
                                } while(0)


#if ENABLE_DEBUG
#define DEBUG(fmt,arg...)      do {\
                                printf(fmt,##arg);\
                              } while(0) 
                             
                              
#else
#define DEBUG(fmt,arg...)
#endif                                
                                
#endif

#if RW_DEBUG
#define TC_RW_DEBUG(fmt,arg...)  do {\
                                    printf(fmt, ##arg);\
                                  } while(0)
#else
#define TC_RW_DEBUG(fmt,arg...)
#endif

void logging_init(void); 

// #endif	// __DEBUG_H__



//start the database connection
int database_start(char *name, char *pwd);

//close the database connection
void database_end();

int update_user_status(int user_id, int image_id, int status_id);
int get_all_user_status(char *statuslist);

int get_uid_gid_by_name(char *username, int *, int*) ;

int get_user_status(char *username, int osimageid);
int get_user_status_by_name(char *username) ;

void reset_all_status(int, int);

int update_server_status(int server_id,int status);

int insert_user_status(int userId, int tcId, int imageid, char *imagename, char *mac, int imageaccess, int online);

int db_refresh_user_status();

int update_image_status(char* rawfile, char* newfile, int needupdate);

int update_server_status(int server_id,int status);

int get_user_oslist(char *username, char *password, char *oslist, int srv_major, int srv_minor);

int get_image_id_path_by_name(char *name, int *image_id, char *path);

int get_osimageid_by_name(char *name);

int get_osimagename_by_id(int image_id, char *osimage_name);

//int get_imageaccess(char *osname, char* username, IMAGE_ACCESS *imageaccess);
//int get_imageaccess_by_gid(int imageid, int groupid, IMAGE_ACCESS *imageaccess) ;

int delete_user_status(int user_id, int osimage_id);

char *get_default_server_config(int serverId,char *configName);

char *get_udf_dir();

char *get_image_dir();

// added by wming
// int db_update_pcnt_info(char *imagename, int pcnt);
int db_update_pcnt_info(char *imagename, int pcnt, float speed);

int db_insert_deploy_client(uint32 ipadd, char *mac, uint32 machine_status);
int db_delete_deploy_client(uint32 ipadd);

int db_update_deploy_client(uint32 ipadd, char *status);
int db_update_deploy_client_with_machine_status(uint32 ipadd, char *status, uint32 machine_status);
int db_update_deploy_client_all_waiting();

int db_clear_deploy_client();
int db_insert_deploy_history(char *evt_type, char *evt_text);

int db_count_deploy_client_with_status(const char *status);


int db_delete_from_user_status_by_ip (const char* ipv4);
int db_delete_from_deploy_client_by_ip (const char* ipv4);

int db_update_all_existing_user_status(uint32 act);
int db_update_existing_user_status_by_ip(const char* ipv4, uint32 act);

#endif
