/******************************************************************************
Copyright 2012 - 2020 Intel Corporation

For licensing information, see the file 'LICENSE' in the root folder of
this software module.
******************************************************************************/


/* TC Deployment Client (the Receiver) */

#include "conf.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <termios.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "tcd.h"
#include "cltdb.h"
#include "tcdmsg.h"
#include "term.h"

#include "bmap.h"
#include "fl.h"

#include "tm.h"
#include "vos_mem.h"

#define PR(x)	printf("=====>> checkpoint [%d]\n", (x))

// extern int base, coeff, intpktdelay;

char *cifn, *localip;
unsigned char *localmac;

unsigned char *ifn2mac(char *ifn);

struct in_addr ina;

// create a socket to receive unicast messages
int g_clt_sock = -1;
struct sockaddr_in g_clt_sock_addr = {0};


// to check whether the hard disk was successfully mounted.
int checkmnt()
{
	int retval;

	retval = system("mount | grep -q \"" RSVDDEV "\" && mount | grep -q \"" MNTDEV "\"");

	return WEXITSTATUS (retval);
}


void clear_ccb(CCB *ccbp)
{
	ccbp->clt_id = 999;
	ccbp->clt_cbulk = -1;
	ccbp->clt_totalbytes = 0;

	ccbp->clt_hasnofile = 0;

	init_bulkmap(&(ccbp->clt_bulkmap));

	ccbp->pkacktv = getts();

	return;
}

void init_ccb(CCB *ccbp)
{
	ccbp->tcdclt_debugger = 0;

	printf("client: initial state to [init]\n");
	ccbp->clt_state = CLT_STATE_INIT;

	// default log file, defined in conf.h
	ccbp->clt_logfile = TCDLOGFN;

	ccbp->pktack_delay_base = 0; 
	ccbp->pktack_delay_coeff = 0;
	ccbp->pktack_delay_intpkt = 0;

	clear_ccb(ccbp);

	ccbp->clt_machine_status = checkmnt();

	return;
}


uint64 diff;

char *flw_name_idver=NULL, flw_dir[512];

// defined in ifn2ipadd.c
extern char *ifn2ip_str(char *ifn);
extern char *ifn2mac_str(char *ifn);

// help for the client
void clt_help()
{
	printf("\n");
	printf("==============================================\n");
	printf("tcd client help: \n\n");
	printf("h: print the help information\n");
	printf("d: turn on/off the debugger\n");
	printf("c: check the current status\n");

	printf("b: reboot\n");
	printf("0: poweroff\n");
	printf("1: restart tcdclt application\n");

	printf("v: print the build and version info\n");
	printf("q: quit\n");
	printf("==============================================\n");
	printf("\n");
}

void print_clt_status(CCB *ccbp)
{
	printf("==============================================\n");
	printf("current bulk#: %d\n", ccbp->clt_cbulk);
	printf("base: %d coeff: %d intpkt: %d\n", ccbp->pktack_delay_base, ccbp->pktack_delay_coeff, ccbp->pktack_delay_intpkt);
	printf("==============================================\n");
}

void print_clt_instant_speed(CCB *ccbp, char *fn)
{
	uint64 diff;
	double bulk_pcnt, bps;

	ccbp->lastcurrtv = ccbp->currtv; 
	ccbp->currtv = getts();

	diff = (uint64)diff_tv(ccbp->lastcurrtv, ccbp->currtv);
	bps = (double)(TOTALSZ * SLOT_BYTES * SAMPLE_FREQ * 8)/(double)diff * 1000000;

	fflush(stdout);
	if (bps >= 1000000)
		printf("\rInstant Receiving Speed [%s]: %8.2fMbps", fn, bps/1000000);
	else if (bps >= 1000)
		printf("\rInstant Receiving Speed [%s]: %8.2fKbps", fn, bps/1000);
	else
		printf("\rInstant Receiving Speed [%s]: %8.2fbps", fn, bps);

	bulk_pcnt = (double)ccbp->clt_cbulk * (double)TOTALSZ / (double)(ccbp->clt_total_slots) * 100.0;
	printf("   %8.2f%% ", bulk_pcnt);

	return;
}


void print_clt_summary_speed(CCB *ccbp, char *fn)
{
	uint64 diff;

	ccbp->endtv = getts();					
	diff = (uint64)diff_tv(ccbp->starttv, ccbp->endtv);
	double bps = (double)ccbp->clt_totalbytes*8*1000000/(double)diff;

	printf("\n\nSummary for [%s]: \ntotal bytes: %lld total seconds: %8.2f \n", fn, ccbp->clt_totalbytes, (double)diff/1000000.0);

	if (bps >= 1000000)
		printf("receiving speed: %8.2fMbps\n\n", bps/1000000);
	else if (bps >= 1000)
		printf("receiving speed: %8.2fKbps\n\n", bps/1000);
	else
		printf("receiving speed: %8.2fbps\n\n", bps);

	return;

}



/*

Here list the brief description of the state machine at client side: 

there are four states: init, ready, run and stop, same as the four states at server side. 

1, init: the very beginning state of the client. At this state, the client can accept the broadcasted hello packet
   from the server, and acknowledged with the registration request message. 
a) accept registration acknowledgement message from the server, go to state 'ready' and take notes accordingly. 
b) when pressing 'q' key, will go to state 'stop' 

2, ready: the client is ready to run. In this state the client is waiting for the runrequest message only
a) when getting the runreq message from server, it will send runack message back to server, go to state 'run' 
   and start to send/receive data packets. 
b) when pressing 'q' key, it will go to state 'stop' to terminate the data sending/receiving. 

3, run: the running state of the client, where the massive inbounding and outbounding packet processing happens here. 
a) when pressing 'q' key, will go to state 'stop' 
b) when getting the last packet of the file, will go to state 'stop'

4, stop: All packet transfer was done. 
a) to print the summary of this transfer and quit.  

*/

// message processing in client part
int proc_tcdmsg_clt(int socket, char *pkt, struct sockaddr_in singlesock, struct sockaddr_in groupsock, CCB *ccbp)
{
	TCDMSG *msg = (TCDMSG *)pkt;
	char *msgbuf;
	int msgbuf_len;
	CLT *cp;
	BULKMAP *bmp;
	double bps, bulk_pcnt;
	int k;
	uint32 localipadd;
	uint32 maxcbulk=0;
	char system_buf[1024];
    char tmpimagename[FNAME_SZ] = {0};


	switch(msg->type)
	{
		case TCDMSG_REGACK:
			if (ccbp->clt_state != CLT_STATE_INIT)
			{
				fprintf(stderr, "proc_tcdmsg_clt: receive REGACK in state [%s], should be [CLT_STATE_INIT]\n", CLT_STATE_STR(ccbp->clt_state));
				break;
			}

			SET_CLT_STATE(ccbp, CLT_STATE_READY);

			break;

		case TCDMSG_RUNREQ:
			if (ccbp->clt_state != CLT_STATE_READY)
			{
				fprintf(stderr, "proc_tcdmsg_clt: receive RUNREQ in state [%s], should be [CLT_STATE_READY]\n", CLT_STATE_STR(ccbp->clt_state));
				break;
			}

			if (ccbp->clt_machine_status != 0)
			{
				printf("do not mount local hard disk, so cannot deploy\n");
				// break;

				// newly added to fix the ctrl-c issue
				SET_CLT_STATE(ccbp, CLT_STATE_STOP);
				runack_send(socket, singlesock, 0);

				return -1;
			}

			// to identify local client according to client list from this message. 

			clear_ccb(ccbp);


			printf("=======================>>>>> runreq message identifying\n");
			printf("client count: %d\n", msg->u.ur.clt_count);

			if (msg->u.ur.clt_count > MAXCLTS)
			{
				printf("fatal error: illegal clt_count in server: %d\n", msg->u.ur.clt_count);
				// while(1);
			}

			for (k=0; k<msg->u.ur.clt_count; k++)
			{
				printf("%d --- 0x%x\n", k, msg->u.ur.clt_list[k]);
			}
			printf("=======================\n");

			localipadd = ifn2ip(cifn);
			k = 0;
			while(k < msg->u.ur.clt_count)
			{
				if ((uint32)(msg->u.ur.clt_list[k]) == (uint32)localipadd)
				{
					printf("it's #%d client\n", k);
					print_num(k);
					ccbp->clt_id = k;
					break;
				}
				k++;
			}
			// printf("local ip: %s --- 0x%x\n", localip, ifn2ip(cifn));

			// debug before runreq: 
			printf("DEBUG: ----------> total slot: 0x%x\n", msg->u.ur.u32tslots);

/*
#ifdef MACHINE_X64
printf("DEBUG: ----------> total slot: 0x%llx\n", NTOHUL(msg->u.ur.tslots));
//print_buf(msg, 128);
#else
printf("DEBUG: ----------> total slot: 0x%llx\n", msg->u.ur.tslots);
#endif
*/

			DBG("run request\n");

			// last file in the file list was transferred, so just quit. 
			if (msg->u.ur.hasnofile == 1)
			{
				SET_CLT_STATE(ccbp, CLT_STATE_STOP);
				notify_send(socket, singlesock, NF_LAST_FILE, 0);

				return -1;
			}

			printf("runreq: filename: <%s> tslots: %d [0x%x]\n", msg->u.ur.fname, msg->u.ur.u32tslots, msg->u.ur.u32tslots);
			ccbp->clt_total_slots = msg->u.ur.u32tslots; 

			printf("delay base: %d delay coeff: %d delay intpkt: %d mapsz: %d\n", \
				msg->u.ur.pktack_delay_base, msg->u.ur.pktack_delay_coeff, msg->u.ur.pktack_delay_intpkt, msg->u.ur.dyna_mapsz);
			ccbp->pktack_delay_base = msg->u.ur.pktack_delay_base; 
			ccbp->pktack_delay_coeff = msg->u.ur.pktack_delay_coeff; 
			ccbp->pktack_delay_intpkt = msg->u.ur.pktack_delay_intpkt; 
			ccbp->dyna_mapsz = msg->u.ur.dyna_mapsz;


			// if (flw_name_idver == NULL)
			// 	flw_name_idver = (char *)malloc(1024);
			flw_name_idver = NULL;
			flw_name_idver = (char *)malloc(512);
			if (flw_name_idver == NULL)
			{
				/* code */
				return -1;
			}
			memset(flw_name_idver, 0, sizeof(flw_name_idver));
			strncpy(flw_name_idver, flw_dir, sizeof(flw_dir));
			// here in msg->u.ur.fname there is a string consisting <name>_<osid>_<ver>
			// so we need to extract the fname with fl_name()
			if (strlen(flw_name_idver) > 256)
			{
				/* code */
				return -1;
			}
			flw_name_idver = strcat(flw_name_idver, msg->u.ur.fname);
			
			
			fl_fname_ex(tmpimagename, sizeof(tmpimagename), flw_name_idver);
			printf("path flw_name: <%s>\n", tmpimagename);



			// pre deploy trigger at the beginning of each file transfer --- by Weiyu
			sprintf(system_buf, "/usr/sbin/tc/pre_deploy_trigger.sh %d 1 %s", fl_osid(flw_name_idver), inet_ntoa(ina));
			printf("system_buf: [%s]\n", system_buf);
			system(system_buf);


			// create the image file ...
			flw_fd = fl_creat(tmpimagename);
			if (flw_fd == -1)
			{
				fprintf(stderr, "create <%s> failed\n", tmpimagename);
				notify_send(socket, singlesock, NF_CREATE_FAILURE, 0);
				return -1;
			}
			else
			{
				printf("create <%s> successfully\n", tmpimagename);
			}

			ccbp->lastcurrtv = ccbp->currtv = ccbp->starttv = getts(); 

			// to set the parameters for packet receiving ...
			ccbp->clt_cbulk = 0;

			SET_CLT_STATE(ccbp, CLT_STATE_RUN);
			runack_send(socket, singlesock, 1);

			break;


		case TCDMSG_DATA:
			// data can only be transferred in state RUN
			if (ccbp->clt_state != CLT_STATE_RUN)
			{
				//fprintf(stderr, "proc_tcdmsg_clt: receive DATA in state [%s], should be [CLT_STATE_RUN]\n", CLT_STATE_STR(ccbp->clt_state));
				break;
			}

            fl_fname_ex(tmpimagename, sizeof(tmpimagename), flw_name_idver);
			// the new message belongs to current bulk ...
			if (msg->u.d.bulk == ccbp->clt_cbulk)
			{
				// belongs to this bulk, but not the last slot
				set_bulkmap(&(ccbp->clt_bulkmap), msg->u.d.slot);

				// printf("tcdclt: before fl_write_slot: data: bulk=%ld, slot=%ld\n", msg->u.d.bulk, msg->u.d.slot);
				if (fl_write_slot(flw_fd, (msg->u.d.bulk*TOTALSZ + msg->u.d.slot), msg->u.d.data, msg->u.d.bytes) == -1)
				{
					fprintf(stderr, "fl_write_slot: fatal error, disk full\n");
					notify_send(socket, singlesock, NF_WRITE_FAILURE, 0);
					// fatal error, system will hang.
					printf ("System is halt.");
					exit(-2);
					// return -1;
				}

				ccbp->clt_totalbytes += msg->u.d.bytes;

				// last slot of this bulk ...
				if ((msg->u.d.slot % TOTALSZ == (TOTALSZ-1)) || (msg->u.d.hasnext == 0))
				{
					#ifdef _DEBUG_
					print_bulkmap(&(ccbp->clt_bulkmap));
					#endif

					// to send back the acknowledgement for this bulk
					if (isfull_bulkmap(&(ccbp->clt_bulkmap), msg->u.d.hasnext))
					{
						// struct timeval tv = getts();
						// printf("full/last bulk #: %d [%ld:%ld]\n", msg->u.d.bulk, tv.tv_sec, tv.tv_usec);

						// the workaround to avoid the congestion of packet acknowledgement sending
						// when getting the broadcasting data packet. 
						log_printf("pktack_send: bulk# [%d], cbulk: [%d]\n", ccbp->clt_cbulk, ccbp->clt_cbulk);



						// vos_random_usleep();
						vos_usleep(ccbp->pktack_delay_base + ccbp->clt_id*ccbp->pktack_delay_coeff/*RANDOM_DELAY_COEFF*/);
						pktack_send(socket, singlesock, ccbp->clt_cbulk, 1, &(ccbp->clt_bulkmap));
						ccbp->pkacktv = getts();

#ifdef PRINT_INSTANT_SPEED
						if (ccbp->clt_cbulk % SAMPLE_FREQ == SAMPLE_FREQ-1)
							print_clt_instant_speed(ccbp, tmpimagename);
#endif

						ccbp->clt_cbulk ++;
						init_bulkmap(&(ccbp->clt_bulkmap));
					}
					else
					{
						printf("loss slots in bulk #%d\n", msg->u.d.bulk);
						print_bulkmap(&(ccbp->clt_bulkmap));
#ifdef _DEBUG_
#endif
						// workaround for avoid the packet acknowledgement congestion
						// log_printf("pktack_send: bulk# [%d], cbulk: [%d]\n", ccbp->clt_cbulk, ccbp->clt_cbulk);
						// vos_random_usleep();

						vos_usleep(ccbp->pktack_delay_base + ccbp->clt_id * ccbp->pktack_delay_coeff/*RANDOM_DELAY_COEFF*/);
						pktack_send(socket, singlesock, ccbp->clt_cbulk, 0, &(ccbp->clt_bulkmap));
						ccbp->pkacktv = getts(); 
					}
				}

				// the last packet of the whole file
				if (msg->u.d.hasnext == 0)
				{
					printf("last packet of the whole file, exit\n");

					fl_close(flw_fd);





					// to create the version file after the entire file is received.
					// create the version file ...
					char vname[1024];
					int v_fd;

					// sprintf(vname, "%s%d.ver", flw_dir, fl_osid(msg->u.ur.fname));
					sprintf(vname, "%s%d.ver", flw_dir, fl_osid(flw_name_idver));
					v_fd = fl_creat(vname);
					if (v_fd == -1)
					{
						fprintf(stderr, "create <%s> failed\n", vname);
						notify_send(socket, singlesock, NF_CREATE_FAILURE, 0);
						return -1;
					}
					else
					{
						printf("create <%s> successfully\n", vname);
					}

					// sprintf(vname, "%d", fl_ver(msg->u.ur.fname));
					sprintf(vname, "%d", fl_ver(flw_name_idver));
					if (write(v_fd, vname, strlen(vname)) < 0)
					{
						perror("verfile_write");
						return -1;
					}

					fl_close(v_fd);



					// char system_buf[1024];

					printf("=====>>>>> invoke weiyu\'s script: id=%d image=[%s]\n", fl_osid(flw_name_idver), tmpimagename);
					// May 11: updated according to Weiyu's latest deploy_trigger script


					//printf("client: hello received, server[%s:%d]\n", inet_ntoa(ina), dstport);

					// add the 4th parameter for server IP address
					sprintf(system_buf, "/usr/sbin/tc/deploy_trigger.sh %d %s 1 %s", \
						fl_osid(flw_name_idver), tmpimagename, inet_ntoa(ina));
					printf("system_buf: [%s]\n", system_buf);
					system(system_buf);

					print_clt_summary_speed(ccbp, tmpimagename);
					notify_send(socket, singlesock, NF_RECV_DONE, 0);

// wming
					SET_CLT_STATE(ccbp, CLT_STATE_READY);

				}

			}
#if 1
			else
				// printf("bulk# overflow or underflow: current bulk [%d] --- incoming bulk [%d-%d]\n", ccbp->clt_cbulk, msg->u.d.bulk, msg->u.d.slot);
				log_printf("bulk# mismatch: incoming bulk [%d-%d] --- cbulk [%d]\n", msg->u.d.bulk, msg->u.d.slot, ccbp->clt_cbulk);
#endif

			break;

		case TCDMSG_SHUTDOWN:
			printf("=======================>>>>> shutdown message\n");
			notify_send(socket, singlesock, NF_SHUTDOWN, 0);
			system(TCDCLT_SCRIPT_POWEROFF);
			return -1;
			break;

		case TCDMSG_SHUTDOWN_IP:
			printf("=======================>>>>> shutdown ip message\n");
			printf("IP address in message: %s\n", msg->u.sd_ip.ipv4);
			printf("local IP address: %s\n", localip);
			if (!strcmp (msg->u.sd_ip.ipv4, localip)){
				notify_send(socket, singlesock, NF_SHUTDOWN_IP, 0);
				system(TCDCLT_SCRIPT_POWEROFF);
			}
			//return -1;
			break;

		case TCDMSG_REBOOT:
			printf("=======================>>>>> reboot message\n");
			notify_send(socket, singlesock, NF_REBOOT, 0);
			system(TCDCLT_SCRIPT_REBOOT);
			return -1;
			break;

		case TCDMSG_REBOOT_IP:
			printf("=======================>>>>> reboot ip message\n");
			printf("IP address in message: %s\n", msg->u.rb_ip.ipv4);
			printf("local IP address: %s\n", localip);
			if (!strcmp (msg->u.rb_ip.ipv4, localip)){
				notify_send(socket, singlesock, NF_REBOOT_IP, 0);
				system(TCDCLT_SCRIPT_REBOOT);
			}
			//return -1;
			break;

		case TCDMSG_RESTARTAPP:
			printf("=======================>>>>> restart application message\n");
			notify_send(socket, singlesock, NF_RESTARTAPP, 0);
			return -1;
			break;


		case TCDMSG_IDENTIFY:
			printf("=======================>>>>> identify message\n");


			//printf("name: %s hasnofile: %d tslots: 0x%llx\n", msg->u.id.fname, msg->u.id.hasnofile, msg->u.id.tslots);
/*
			printf("client count: %d\n", msg->u.ur.clt_count);
			for (k=0; k<msg->u.ur.clt_count; k++)
			{
				printf("%d --- 0x%x\n", k, msg->u.ur.clt_list[k]);
			}
			printf("=======================\n");

			localipadd = ifn2ip(cifn);
			k = 0;
			while(k < msg->u.ur.clt_count)
			{
				if ((uint32)(msg->u.ur.clt_list[k]) == (uint32)localipadd)
				{
					printf("it's #%d client\n", k);
					print_num(k);
					ccbp->clt_id = k;
					break;
				}
				k++;
			}
			// printf("local ip: %s --- 0x%x\n", localip, ifn2ip(cifn));
*/

			print_num(ccbp->clt_id);
			//printf("local ip: %s --- 0x%x\n", localip, ifn2ip(cifn));
			break;


		case TCDMSG_USERCMD:
			printf("=======================>>>>> pktack user-defined command\n");
			printf("client count: %d\n", msg->u.id.clt_count);
			for (k=0; k<msg->u.id.clt_count; k++)
			{
				printf("%d --- 0x%x cbulk#: %d\n", k, msg->u.pac.clt_ipcbulk_list[k*2], msg->u.pac.clt_ipcbulk_list[k*2+1]);
			}
			printf("=======================\n");

			localipadd = ifn2ip(cifn);
			k = 0;
			while(k < msg->u.id.clt_count)
			{
				if ((uint32)(msg->u.id.clt_list[k*2]) == (uint32)localipadd)
				{
					printf("it's #%d client\n", k);
					// print_num(k);
					printf("------------------->>>> cbulk: %d acked full: %d\n", ccbp->clt_cbulk, ccbp->clt_cbulk-1);
					pktack_full_send(socket, singlesock, ccbp->clt_cbulk-1, 1);
					break;
				}
				k++;
			}
			printf("local ip: %s --- 0x%x\n", localip, ifn2ip(cifn));
			break;

		case TCDMSG_FORMAT:
			printf("=======================>>>>> format the local harddisk\n");
			notify_send(socket, singlesock, NF_FORMAT, 0);
			// wming
			system(TCDCLT_SCRIPT_PATH"/autopart.sh /dev/sda /dev/sda1");
			notify_send(socket, singlesock, NF_FORMAT, 1);
			return -1;
			break;

		case TCDMSG_INITIALIZE:
			printf("=======================>>>>> initialize the local harddisk\n");
			notify_send(socket, singlesock, NF_INITIALIZE, 0);
			// wming
			//system(TCDCLT_SCRIPT_PATH"/inithd.sh /dev/sda");
			sprintf (system_buf, TCDCLT_SCRIPT_PATH"/inithd.sh /dev/sda %u", msg->u.id.sys_part_capacity);
			k = system (system_buf);
			notify_send(socket, singlesock, NF_INITIALIZE, 1 + WEXITSTATUS(k));

			// obsolete
			// newly added to solve the high deploy failure rate issue found before the Oct 1 holiday 
			//printf("=======================>>>>> restart the system after initializing the system\n");
			//system(TCDCLT_SCRIPT_REBOOT);

			// masked: will reboot, unmask: will not reboot
			// return -1;
			break;

		case TCDMSG_UPDATEFIRMWARE:
			printf("=======================>>>>> firmware update for local harddisk\n");
			notify_send(socket, singlesock, NF_UPDATEFIRMWARE, 0);
			// wming
			k = system(TCDCLT_SCRIPT_PATH"/updatefirmware.sh /dev/sda " RSVDDEV);
			notify_send(socket, singlesock, NF_UPDATEFIRMWARE, 1 + WEXITSTATUS(k));
			return -1;
			break;

		case TCDMSG_CGSVRIP:
			notify_send(socket, singlesock, NF_CGSVRIP, 0);
			char cgsbuf[256];
			uint32 ip = msg->u.cgs.srvip;
			printf("=======================>>>>> change server IP cache in local client [%s]\n", inet_ntoa(*(struct in_addr *)&ip));
			sprintf(cgsbuf, "/usr/sbin/tc/servip_trigger.sh %s", inet_ntoa(*(struct in_addr *)&ip));
			system(cgsbuf);
			notify_send(socket, singlesock, NF_CGSVRIP, 1);
			// return -1;
			break;


		case TCDMSG_PKTACK_CHECK:
			//printf("=======================>>>>> pktack check message\n");
			//printf("client count: %d\n", msg->u.pac.clt_count);
			for (k=0; k<msg->u.pac.clt_count; k++)
			{
				//printf("%d-0x%x cbk#[%d]  ", k, msg->u.pac.clt_ipcbulk_list[k*2], msg->u.pac.clt_ipcbulk_list[k*2+1]);
				//if (k % 3 == 2)
				//	printf("\n");
				maxcbulk = ((msg->u.pac.clt_ipcbulk_list[k*2+1])>maxcbulk)?(msg->u.pac.clt_ipcbulk_list[k*2+1]):maxcbulk;
			}

			localipadd = ifn2ip(cifn);

			k = -1;
			while(++k < msg->u.pac.clt_count)
			{
				if ((uint32)(msg->u.pac.clt_ipcbulk_list[k*2]) == (uint32)localipadd)
				{
					// find the corresponding cbulk# from the server
					printf("it's #%d client\n", k);

					if (ccbp->clt_cbulk > msg->u.pac.clt_ipcbulk_list[k*2+1])
					{
						printf("---------->>>> cbulk mismatch: client [%d] and server [%d]\n", \
							ccbp->clt_cbulk, /*localipadd, */msg->u.pac.clt_ipcbulk_list[k*2+1]/*, msg->u.pac.clt_ipcbulk_list[k*2]*/);
						pktack_full_send(socket, singlesock, ccbp->clt_cbulk-1, 1);
					}
					break;
				}
			}
			// printf("local ip: %s --- 0x%x\n", localip, ifn2ip(cifn));

			break;



		case TCDMSG_PKTINQ:
			if (ccbp->clt_state != CLT_STATE_RUN)
			{
				fprintf(stderr, "proc_tcdmsg_clt: receive PKTINQ in state [%s], should be [CLT_STATE_RUN]\n", CLT_STATE_STR(ccbp->clt_state));
				break;
			}

			DBG("packet inquiry: bulk=%ld, slot=%ld\n", msg->u.pi.bulk, msg->u.pi.slot);

			break;

		default:
			// printf("unknown tcdmsg type @ client: [%d]\n", msg->type);
			break;
	}

	return 0;
}

#if 0
int proc_tcdmsg_clt_hello(int socket, char *pkt, struct sockaddr_in singlesock, struct sockaddr_in groupsock, CCB *ccbp)
{
	TCDMSG *msg = (TCDMSG *)pkt;
	char *msgbuf;
	int msgbuf_len;
	CLT *cp;



	switch(msg->type)
	{
		case TCDMSG_HELLO:
			if (ccbp->clt_state == CLT_STATE_INIT)
			{
				printf("client: hello received, server[%x:%d]\n", msg->u.h.srvip, msg->u.h.srvport);
				
				regreq_send(socket, singlesock, localmac, ccbp->clt_machine_status);
			}
			else
			{
				fprintf(stderr, "proc_tcdmsg_clt: receive HELLO in state [%s], should be [CLT_STATE_INIT]\n", CLT_STATE_STR(ccbp->clt_state));
			}
			break;

		default:
			printf("illegal tcdmsg type @ client: [%d]\n", msg->type);
			break;
	}

	return 0;
}
#endif


// client command line processing
int clt_cmdline_proc(char ch, CCB *ccbp)
{
	switch(ch)
	{
		case 'h':
			clt_help();
			break;

		case 'q':
/*
			printf("client: change state to stop\n");
			SET_CLT_STATE(ccbp, CLT_STATE_STOP);
			return -1;
*/
			printf("not support in this version ...\n");
			break;

		case 'd':
			if (ccbp->tcdclt_debugger)
			{
				printf("turn off the debugger\n");
				ccbp->tcdclt_debugger = 0;
			}
			else
			{
				printf("turn on the debugger\n");
				ccbp->tcdclt_debugger = 1;
			}
			break;

		case 'c':
			print_clt_status(ccbp);
			break;

		case 'b':
			printf("reboot the machine\n");
			system(TCDCLT_SCRIPT_REBOOT);
			break;
		case '0':
			printf("power off the client\n");
			system(TCDCLT_SCRIPT_POWEROFF);
			break;
		case '1':
/*			
			printf("restart the tcdclt application\n");
			printf("client: change state to stop\n");
			SET_CLT_STATE(ccbp, CLT_STATE_STOP);
			return -1;
*/
			printf("not support in this version ...\n");

			break;


		case 'v':
			print_buildinfo();
			break;

		default: 
			DBG("clt_cmdline_proc: unknown char\n");
			break;
	}

	return 0;
}

BULKMAP fullbmap = {TOTALSZ, 0xffffffffffffffffULL, }; 


void proc_tcdclt_timeout(CCB *ccbp, int sock, struct sockaddr_in singlesock)
{
	uint64 diff;
	struct timeval ts;

	// only effective in RUN state
	if (ccbp->clt_state != CLT_STATE_RUN)
		return;
	
#ifdef _DEBUG_
	printf("in proc_tcdclt_timeout ...\n");
#endif

	ts = getts();

	diff = diff_tv(ccbp->pkacktv, ts);
	if (diff > DEFAULT_CLT_PKTACK_LOSS_TIMEOUT)
	{
		// packet acknowledgement loss ...
		// printf("packet acknowledgement for cbulk [%d] loss ... \n", ccbp->clt_cbulk-1);
		// pktack_send(sock, singlesock, ccbp->clt_cbulk-1, 1, &fullbmap/*&(ccbp->clt_bulkmap)*/);
		// pktack_full_send(sock, singlesock, ccbp->clt_cbulk-1, 1);
	}

	return;
}


/* client FSM main routine: 
   cskt: the control socket. It's used to talk with server's control plane. 
   dskt: the data socket. It's used to receive server's data plane packets. 
   servadd: server's control plane information
 */
int clt_fsm_main(int sock, struct sockaddr_in singlesock, struct sockaddr_in groupsock, CCB *ccbp)
{
	fd_set rdfds;
	struct timeval tv;
	struct sockaddr_in sain;
	int sain_len;
	int res;
	char ch;
	char buf[2048];
	// struct termios new_settings, stored_settings;

	printf("enter clt_fsm_main() loop ...\n");


	// to enable the raw terminal mode so we can use hotkey ...
	term_raw();


	// the while deadloop(), which can only be out if 
	//	clt_cmdline_proc() returns -1, or 
	// 	proc_tcdmsg_clt() returns -1. 
	while(1)
	{
		// add network socket and stdin to the FDSET
		FD_ZERO(&rdfds);
		FD_SET(0, &rdfds);
		FD_SET(sock, &rdfds);
		FD_SET(g_clt_sock, &rdfds);
		// and set the select timeout to forever
		tv.tv_sec = 0;
		tv.tv_usec = 0; //DEFAULT_CLT_PKTACK_LOSS_TIMEOUT;

		res = select((sock > g_clt_sock ? sock : g_clt_sock) + 1, &rdfds, NULL, NULL, &tv);
		// DBG("res=%d\n", res);
		
		if (res > 0)
		{
			memset (buf, 0, sizeof (buf));
			// stdin events
			if (FD_ISSET(0, &rdfds))
			{
				ch = getchar();
				// DBG("0x%x [%c]\n", ch, ch);
				if (clt_cmdline_proc(ch, ccbp) == -1)
					break; // goto _OUTCLT
			}

			// socket events handling
			else if (FD_ISSET(sock, &rdfds))
			{
				// receiving data packets until the last packet ...
				DBG("start receiving messages ...\n");

				// change here for making it work in uclibc platform
				// if (recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr *)&sain, &sain_len) != -1)
				if (recv(sock, buf, sizeof(buf), 0) != -1)
				{
					//DBG("0x%x : 0x%x : 0x%x\n", sain.sin_family, sain.sin_addr.s_addr, sain.sin_port);
					// printf("Receive from [%s:%d]: \n", inet_ntoa(sockaddr.sin_addr),  htons(sain.sin_port));
					if (proc_tcdmsg_clt(sock, buf, singlesock, groupsock, ccbp) == -1)
						break; 
				}
				else
					printf("clt_fsm_main: recvfrom failed!\n");
			}
			else if (FD_ISSET (g_clt_sock, &rdfds))
			{
				DBG("start receiving reboot/shutdown messages ...\n");
				if (recv(g_clt_sock, buf, sizeof(buf), 0) != -1)
				{
					if (proc_tcdmsg_clt(sock, buf, singlesock, groupsock, ccbp) == -1)
						break; 
				}
				else
					printf("clt_fsm_main: recv from g_clt_sock failed!\n");
			}
		}
		else if (res == 0)
		{
			// printf("clt_fsm_main: select timeout\n");
			proc_tcdclt_timeout(ccbp, sock, singlesock);
		}
		else
		{
			perror("clt_fsm_main: select()");
		}


	}

	// to restore the terminal mode
	term_restore();	

	return 0;
}

void tcdclt_usage(char *argv0)
{
	printf("\nUsage: ");

	printf("\n\t%s <port> \n\t\t--- from default outbounding interface eth0 and with current working dir\n", argv0);
	printf("\n\t\t    example: %s 12345", argv0);
	printf("\n\t%s <port> <dir>\n\t\t--- use eth0, save to <dir>\n", argv0);
	printf("\n\t\t    example: %s 12345 /home/data", argv0);
	printf("\n\t%s <interf> <port> \n\t\t--- from outbounding interface <interf> and with current working dir\n", argv0);
	printf("\n\t\t    example: %s eth1 12345", argv0);
	printf("\n\t%s <interf> <port> <dir>\n\t\t--- from <interf>, save to <dir>\n", argv0);
	printf("\n\t\t    example: %s eth1 12345 /home/data", argv0);

	printf("\nArguments: ");
	printf("\n\t<interf>:\tinbounding interface like eth0, eth1, lo");
	printf("\n\t<port>:\tserver's listening UDP port");
	printf("\n\t<dir>:\tfile folder to save the file\n\n");

	return;
}

// format: tcdclt <ip> <port>, ip and port are the target server's IP address and port#
int main(int argc, char **argv)
{

	// the entity of the client control block
	// the definition of CCB is in tcdmsg.h
	CCB ccb; 


	// data channel socket, UDP
	int sock; 
	struct sockaddr_in servadd;
	char buf[2048];

	int dstport;
	uint32 dstip;

	struct sockaddr_in localsock;
	struct ip_mreq group;

	char *msgbuf;
	int msgbuf_len;

	// parameter handling
	if (argc != 2 && argc != 3 && argc != 4)
	{
		tcdclt_usage(argv[0]);

		exit(0);
	}

	if (argc == 2)	// %s <port>
	{
		cifn = "eth0";
		dstport = atoi(argv[1]);
		strcpy(flw_dir, "./");
	}
	else if (argc == 3)
	{
		if (!strncmp(argv[1], "eth", 3))	//%s <interf> <port>
		{
			cifn = argv[1];
			dstport = atoi(argv[2]);
			strcpy(flw_dir, "./");
		}
		else	// %s <port> <dir>
		{
			cifn = "eth0";
			dstport = atoi(argv[1]);
			memset(flw_dir, 0, sizeof(flw_dir));
			strncpy(flw_dir, argv[2], sizeof(flw_dir)-2);
			
		}
	}
	else if (argc == 4) // %s <interf> <port> <dir>
	{
		cifn = argv[1];
		dstport = atoi(argv[2]);
		memset(flw_dir, 0, sizeof(flw_dir));
		strncpy(flw_dir, argv[3], sizeof(flw_dir)-2);
	}

	if (flw_dir[strlen(flw_dir)-1] != '/')
	{
		int x = strlen(flw_dir);
		if (x < sizeof(flw_dir)-2)
		{
			/* code */
			flw_dir[x] = '/';
			flw_dir[x+1] = 0;
		}
			
	}

	// get the local ipadd of the interface
	if ((localip = ifn2ip_str(cifn)) == NULL)
	{
		perror("ifn2ip_str");
		exit(-1);
	}
	printf("==========>>>>>>> local ip: %s\n", localip);

	// get the local mac address of the interface

	if ((localmac = (unsigned char *)ifn2mac(cifn)) == NULL)
	{
		perror("ifn2mac");
		exit(-1);
	}
	printf("==========>>>>>>>>>> mac address: [%02X:%02X:%02X:%02X:%02X:%02X]\n", localmac[0], localmac[1], localmac[2], localmac[3], localmac[4], localmac[5]);

	printf("\nMulticast Deployment Client\n");
	if (flw_name_idver != NULL)
		printf("Interface [%s] Network [%s:%d] File [%s]\n", cifn, localip, dstport, flw_name_idver);
	else
		printf("Interface [%s] Network [%s:%d] \n", cifn, localip, dstport);
		

	// printf("outbounding via [%s] with local ip [%s]\n", cifn, localip);
	
	// create a thread to receive unicast messages
	g_clt_sock_addr.sin_family = AF_INET;
	g_clt_sock_addr.sin_port = htons (CLT_PORT);
	g_clt_sock_addr.sin_addr.s_addr = inet_addr (localip);

	if ((g_clt_sock = socket (AF_INET, SOCK_DGRAM, 0)) < 0){
		perror ("socket(), g_clt_sock");
		_exit (-1);
	}

	if (bind (g_clt_sock, (struct sockaddr *)&g_clt_sock_addr, sizeof (g_clt_sock_addr)) < 0){
		perror ("bind(), g_clt_sock");
		close (g_clt_sock);
		_exit (-1);
	}

	// to create the data socket. the hello packet will also be from this socket. 
	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		perror("socket");
		exit(-1);
	}
	DBG("tcdclt: socket successfully ...\n");

	int reuse = 1;
	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse, sizeof(reuse)) < 0)
	{
		perror("setting SO_REUSEADDR");
		close(sock);
		exit(-1);
	}
	DBG("tcdclt: setsockopt SO_REUSEADDR successfully\n");

	int nSendBuf=SND_BUF_SZ; // something like 64*1024, defined in conf.h
	if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (const char *)&nSendBuf, sizeof(int)) < 0)
	{
		perror("setting socket buffer");
		close(sock);
		exit(-1);
	}

	int nRecvBuf=RCV_BUF_SZ; // something like 64*1024, defined in conf.h
	if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (const char *)&nRecvBuf, sizeof(int)) < 0)
	{
		perror("setting socket buffer");
		close(sock);
		exit(-1);
	}
	DBG("tcdsrv: set socket buffer to 32K successfully ...\n");

	memset((char *)&localsock, 0, sizeof(localsock));
	localsock.sin_family = AF_INET; 
	localsock.sin_port = htons(GROUP_PORT);
	localsock.sin_addr.s_addr = INADDR_ANY;

	if (bind(sock, (struct sockaddr *)&localsock, sizeof(localsock)))
	{
		perror("bind");
		close(sock);
		exit(-1);
	}
	DBG("tcdclt: bind successfully\n");


	// add the local interface to multicast group
	group.imr_multiaddr.s_addr = inet_addr(GROUP_IP);
	group.imr_interface.s_addr = inet_addr(localip);
	if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&group, sizeof(group))<0)
	{
		perror("adding multicast group error");
		close(sock);
		exit(-1);
	}
	DBG("dskt: add to multicast group successfully\n");




	// wait for hello packet ...
	DBG("wait for the hello packets ...\n");

	struct sockaddr_in sain;
	int sain_len;
	if (recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr *)&sain, &sain_len) == -1)
	{
		perror("recvfrom dskt");
		close(sock);
		exit(-1);
	}

	//DBG("0x%x : 0x%x : 0x%x\n", sain.sin_family, sain.sin_addr.s_addr, sain.sin_port);
	//DBG("Receive from [%x:%d]: \n", htonl(sain.sin_addr.s_addr), htons(sain.sin_port));

	TCDMSG *msg = (TCDMSG *)buf;
	if (msg->type != TCDMSG_HELLO)
	{
		printf("illegal tcdmsg type @ client: [%d]\n", msg->type);
		perror("msg type");
		close(sock);
		exit(-1);
	}

	dstip = msg->u.h.srvip;
	dstport = msg->u.h.srvport;
// defined as global variable
//	struct in_addr ina;
	ina.s_addr = dstip;
	printf("client: hello received, server[%s:%d]\n", inet_ntoa(ina), dstport);


	// to create the groupsock and singlesock, used for multicast and unicast packet sending. 
	// setup the multicast group socket
	struct sockaddr_in groupsock;
	memset((char *)&groupsock, 0, sizeof(groupsock));
	groupsock.sin_family = AF_INET;
	groupsock.sin_addr.s_addr = inet_addr(GROUP_IP);
	groupsock.sin_port = htons(GROUP_PORT);

	struct sockaddr_in singlesock;
	memset((char *)&singlesock, 0, sizeof(singlesock));
	singlesock.sin_family = AF_INET; 
	singlesock.sin_addr.s_addr = dstip;
	singlesock.sin_port = htons(dstport);

	// initialize the memory ...
	if (vos_meminit() == -1)
	{
		fprintf(stderr, "vos_meminit: failed\n");
		close(sock);
		exit(-1);
	}

	// the initialize the client control block
	init_ccb(&ccb);

	// init the logfile
	log_init(ccb.clt_logfile);

	// register to the server according to the hello packet address ...
	regreq_send(sock, singlesock, localmac, ccb.clt_machine_status);


	// to start the client FSM
	clt_fsm_main(sock, singlesock, groupsock, &ccb);

	// to send the de-register request
	deregreq_send(sock, singlesock);

	SET_CLT_STATE((&ccb), CLT_STATE_STOP);

	// to close the control socket	
	DBG("tcdclt: close the socket\n");
	close(sock);
	close (g_clt_sock);

	// close the logfile
	log_close();

	exit(0);
}


