/******************************************************************************
Copyright 2012 - 2020 Intel Corporation

For licensing information, see the file 'LICENSE' in the root folder of
this software module.
******************************************************************************/


/* TC Deployment Controller (the Controller) */

#include "conf.h"
#include "tcd.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <termios.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>

#include <getopt.h>

#include "tcd.h"
#include "cltdb.h"
#include "tcdmsg.h"
#include "term.h"

#include "bmap.h"
#include "fl.h"

#include "tm.h"
#include "vos_mem.h"

#define countof(arr) (sizeof(arr)/sizeof(arr[0]))

static const char* argv0;
char *flw_name=NULL, flw_dir[256];

// defined in ifn2ipadd.c
extern char *ifn2ip_str(char *ifn);
extern char *ifn2mac_str(char *ifn);
void tcdctl_usage();

struct cmdline_option {
  const char *long_name;
  char short_name;
  enum {
    OPT_VAL,
    OPT_BOOLEAN,
    OPT_FUNCALL
  } type;
  const void *data;
  int argtype;
};

struct cmdline_option_flag_value
{
  int flag;    // this flag is set when a corresponding option is scanned.
  char *arg;   // required_argument
  int enabled; // no_argument
};

static struct cmdline_option option_data[] =
{
  {"start-service", 0, OPT_BOOLEAN, "startservice", -1}, //ignore
  {"stop-service", 0, OPT_BOOLEAN, "stopservice", -1},
  {"reboot", 'r', OPT_BOOLEAN, "reboot", -1},
  {"poweroff", 'p', OPT_BOOLEAN, "poweroff", -1},
  {"initialize", 0, OPT_BOOLEAN, "initialize", -1},
  {"update-firmware", 0, OPT_BOOLEAN, "updatefirmware", -1},
  {"change-server", 0, OPT_BOOLEAN, "changeserver", -1},
  {"remove-image", 0, OPT_BOOLEAN, "removeimage", -1},
  {"boot", 'b', OPT_BOOLEAN, "boot", -1},
  {"all", 'a', OPT_BOOLEAN, "all", -1},
  {"ip", 'i', OPT_VAL, "ip", -1},
  {"ip-file", 0, OPT_VAL, "ipfile", -1},
  {"user-partsz", 0, OPT_VAL, "userpartsz", -1},
  {"user-name", 'u', OPT_VAL, "username", -1},
  {"image-id", 0, OPT_VAL, "imageid", -1},
  {"image-name", 0, OPT_VAL, "imagename", -1},
  {"image-org-name", 0, OPT_VAL, "imageorgname", -1},
  {"user-token", 0, OPT_VAL, "usertoken", -1},
  {"server-ip", 0, OPT_VAL, "serverip", -1},
  {"set-resolution", 0, OPT_BOOLEAN, "set-resolution", -1},
  {"width", 0, OPT_VAL, "width", -1},
  {"height", 0, OPT_VAL, "height", -1},
  {"enable-usb-storage", 0, OPT_VAL, "enable-usb-storage", -1},
  {"help", 'h', OPT_BOOLEAN, "help", -1}
};
static struct option long_options[2 * countof(option_data) + 1];
static char short_options[128];
static unsigned char optmap[128 - 32]; 
static struct cmdline_option_flag_value option_data_val[countof(option_data)];
#define IDXOPT(x) get_cmdline_index(x)
#define STROPT(i) option_data[i].long_name

static void init_switches()
{
  char *p = short_options;
  int i, o = 0;
  
  for (i = 0; i < countof(option_data); i++)
  {
    struct cmdline_option *cmdopt = &option_data[i];
    struct option *longopt;
    
    if (!cmdopt->long_name)
    {
      continue;
    }
    
    longopt = &long_options[o++];
    longopt->name = cmdopt->long_name;
    longopt->val = i;
    
    if (cmdopt->short_name)
    {
      *p++ = cmdopt->short_name;
      optmap[cmdopt->short_name - 32] = longopt - long_options;
    }
    
    switch(cmdopt->type)
    {
      case OPT_VAL:
        longopt->has_arg = required_argument;
        if (cmdopt->short_name)
        {
          *p++ = ':';
        }
        break;

      case OPT_BOOLEAN:
        longopt->has_arg = no_argument;
        break;
        
      default:
        if (cmdopt->argtype == -1){
          DBG("unknown option\n");
          exit(-1);
        }
        longopt->has_arg = cmdopt->argtype;
        if (cmdopt->short_name)
        {
          if (longopt->has_arg == required_argument)
          {
            *p++ = ':';
          }
        }
        break;      
    }
  }
  *p = '\0';
  
  int tmp = countof(long_options);
  DBG("o : %d\n", o);
  DBG("tmp : %d\n", tmp);
  DBG("sizeof(countof(long_options)) : %d\n", sizeof(countof(long_options)));
  if (o >= tmp)
  {
    DBG("exceeds max allowed number of arguments.\n");
    exit(-1);
  }
}

//
// find string in array,
// return index
// if not found, return -1
//
static int get_cmdline_index(const char *longoptstr)
{
  int i;
  for (i = 0; i < countof(option_data); i++)
  {
    if (!strcasecmp(option_data[i].long_name, longoptstr))
    {
      return i;
    }
  }
  DBG("cannot find option: %s\n", longoptstr);
  tcdctl_usage();
  exit(-1);
}

static FILE* ipfilefp;
//
// open ip-file
//
static FILE *get_ip_file_fp(const char *ipfile)
{
  ipfilefp = fopen(ipfile, "r");
  return ipfilefp;
}

//
// load ip-file and store it to memory
//
static int get_filelines(const char *file)
{
  int c;
  int i = -1;
  FILE *fp = fopen(file, "r");
  if (fp == NULL)
  {
    return i;
  }
  
  i = 0;
  c = getc(fp);
  while (c != EOF)
  {
    if (c == '\n')
    {
      i++;
    }
    c = getc(fp);
  }

  fclose(fp);
  return i;
}

//
// load ip-file to memory
// return 0 if end of file or error
// return 1 if success
// return 2 if it is not an ip
//
static int get_next_ip_from_file(char *dest, int destlen)
{
  char localbuf[1024];

  if (destlen < 16)
  {
    return 0;
  }

  memset(dest, 0, destlen);
  memset(localbuf, 0, sizeof(localbuf));
  if (NULL != fgets(localbuf, sizeof(localbuf), ipfilefp))
  {
    int i;
    strncpy(dest, localbuf, 16);
    DBG("localbuf = %s\n", localbuf);
    DBG("dest = %s\n", dest);
    dest[15] = '\0';

    for (i = 0; i < destlen; i++)
    {
      if (dest[i] == '\n' || dest[i] == '\r')
      {
        dest[i] = '\0';
        break;
      }
    }


    if (1 == is_valid_ipv4(dest))
    {
      return 1;
    }
    else
    {
      return 2;
    }
  }
  else
  {
    return 0;
  }
}

void ctl_help()
{
  printf("=======================================\n");
  printf("h: print this list\n");
  printf("s: start the TCDSRV\n");
  printf("l: print all clients connected to the TCDSRV\n");
  printf("d: shutdown the TCDSRV\n");
  printf("i: identify each clients\n");
  printf("b: reboot all clients\n");
  printf("0: poweroff all clients\n");
  printf("1: restart all ciients' tcdclt application\n");
  printf("f: format hard disk of ciients\n");
  printf("v: print the build and version info\n");
  printf("u: some user-defined commands, just for debugging\n");
  printf("q: quit\n");
  printf("=======================================\n");

  return;
}






#ifdef USE_MYSQL
void tcdctl_usage()
{
  //
  // define macro used only in this function at first
  // undefine it at end of function
  //
#define prf(fmt,...) printf("%s "fmt, argv0, ##__VA_ARGS__)
#define prd(fmt,...) printf("    "fmt, ##__VA_ARGS__)
#define prn()        printf("\n")
//#define prn()

  printf("\n");
  printf("Usage:\n");

#if 0
  prf("--start-service ...\n");
  prd("start the TCD service daemon\n");
  prn();

  prf("--stop-service ...\n");
  prd("stop the TCD service daemon\n");
  prn();
#endif

  prf("--reboot { --all | --ip <ipaddr> | --ip-file <file> }\n");
  prd("send reboot command to a client with IP address <ipaddr>\n");
  prd("send reboot command to client(s) with IP address read from <file>\n");
  prd("send reboot command to client(s) in the multicast group\n");
  prn();

  prf("--poweroff { --all | --ip <ipaddr> | --ip-file <file> }\n");
  prd("send poweroff command to a client with IP address <ipaddr>\n");
  prd("send poweroff command to client(s) with IP address read from <file>\n");
  prd("send poweroff command to client(s) in the multicast group\n");
  prn();

  prf("--initialize --user-partsz <userPartSizeInMiB> { --all | --ip <ipaddr> | --ip-file <file> }\n");
  prd("send initialize command to a client with IP address <ipaddr>\n");
  prd("send initialize command to client(s) with IP address read from <file>\n");
  prd("send initialize command to client(s) in the multicast group\n");
  prd("userPartSizeInMiB: desirable size of user partition (Unit: MiB)\n");
  prn();

  prf("--update-firmware { --all | --ip <ipaddr> | --ip-file <file> }\n");
  prd("send update-firmware command to a client with IP address <ipaddr>\n");
  prd("send update-firmware command to client(s) with IP address read from <file>\n");
  prd("send update-firmware command to client(s) in the multicast group\n");
  prn();

  prf("--change-server --server-ip <ipaddr> { --all | --ip-file <file> }\n");
  prd("send change-server command to client(s)\n");
  prn();

  prf("--remove-image --image-org-name <imageOrgName> --image-id <imageId> { --all | --ip-file <file> }\n");
  prd("send remove-image command to client(s)\n");
  prn();

  prf("--boot --image-id <image_id> --image-org-name <image_org_name> --image-name <image_name> --user-token <user_token> --user-name <user_name> --ip <ipaddr>\n");
  prd("send boot command to a client in the multicast group\n");
  prn();

  prf("--set-resolution --width <screenHidth> --height <screenHeight> { --ip <ipaddr> | --ip-file <file> }\n");
  prd("send set-resolution command to certain TC client systems\n");
  prn();

  prf("--enable-usb-storage { 0 | 1 } { --ip <ipaddr> | --ip-file <file> }\n");
  prd("send enable/disable usb-storage command to certain TC client systems\n");
  prn();
#if 0
  prf("--start <file1> <file2> ...\n");
  prd("start the deployment in current working dir\n");
  prn();

  prf("--stop ...\n");
  prd("stop the deployment\n");
  prn();

  prf("--restartapp ...\n");
  prd("restart deployment application on clients\n");
  prn();

  prf("--identify ...\n");
  prd("identify each client by printing client#\n");
  prn();

  prf("--format ...\n");
  prd("format all clients\n");
  prn();
#endif

#ifdef prf
#undef prf
#endif
#ifdef prd
#undef prd
#endif
#ifdef prn
#undef prn
#endif
  return;
}

//
// parse argument
// on success, return 1;
// on failure, return 0;
// currently do not support arguments w/o perfix description
//
static int parse_arguments(int argc, char *argv[])
{
  int ret = 0;
  int c;
  int opt_index;
  
  init_switches();
  
  
  //
  // parse arguments
  //
  while (1)
  {
    int val;
    struct cmdline_option *cmdopt;
    struct cmdline_option_flag_value *cmdoptval;
    
    opt_index = -1;
    c = getopt_long(argc, argv, short_options, long_options, &opt_index);

    if (-1 == c)
    {
      break;
    }

    //
    // a short option is matched, if the OPTINDEX is not changed
    //
    if (opt_index == -1)
    {
      if (c == '?')
      {
        tcdctl_usage();
        exit(-1);
      }
      opt_index = optmap[c - 32];
    }
    
    val = long_options[opt_index].val;
    cmdopt = &option_data[val];
    cmdoptval = &option_data_val[val];
    
    //
    // this option has been recognized
    //
    cmdoptval->flag = 1;
    
    switch (cmdopt->type)
    {
      case OPT_VAL:
        if (optarg)
        {
          cmdoptval->arg = optarg;
        }
        DBG("option: --%s, optarg = %s\n", long_options[opt_index].name, optarg);
        break;

      case OPT_BOOLEAN:
        cmdoptval->enabled = 1;
        DBG("option: --%s\n", long_options[opt_index].name);
        break;
        
      case OPT_FUNCALL:
        {
          void (*func) (void) = (void (*) (void)) cmdopt->data;
          func ();
        }
        DBG("option: --%s\n", long_options[opt_index].name);
        return 0;
        break;
        
      default:
        DBG("program should not reach here\n");
        return 0;
        break;
    }
  }
  return 1;
}

int main(int argc, char **argv)
{
  int sock;
  struct sockaddr_in servadd;
  char *localip;

  int dstport;
  uint32 dstip;
  char *dstip_str=NULL;
  struct sockaddr_in localsock;
  int ch, res;
  fd_set rdfds;
  struct timeval tv;

#define MAXDEPLOYFILE  16
  char *fn[MAXDEPLOYFILE]; 
  int fncount = 0, k;

  argv0 = (const char *)argv[0];

  // parameter handling
  if (argc < 2)
  {
    tcdctl_usage();
    exit(0);
  }
  else if (argc > MAXDEPLOYFILE+2)
  {
    printf("to many files, maximal to %d\n", MAXDEPLOYFILE);
    exit(0);
  }

  dstip_str = "127.0.0.1";
  dstip = inet_addr(dstip_str);
  dstport = 12345; // atoi(argv[2]);

  if (!parse_arguments(argc, argv))
  {
    DBG("fail to parse arguments\n");
    exit(-1);
  }

  for (k=2; k<argc; k++)
  {
    fn[k-2] = argv[k];
  }
  fncount = argc - 2;
  // fn and fncount could be filename only, or filename:osid:ver string. 

  // to print the file list ...
  // for (k=0; k<fncount; k++)
  // {
    // printf("file: %d --- %s\n", k, fn[k]);
  // }

  strcpy(flw_dir, "./");

  //printf("flw_dir=<%s>\n", flw_dir);
  if (flw_dir[strlen(flw_dir)-1] != '/')
  {
    int x = strlen(flw_dir);
    flw_dir[x] = '/';
    flw_dir[x+1] = 0;  
  }
  //printf("flw_dir=<%s>\n", flw_dir);

  printf("\nMulticast Deployment Controller\n");
  // to create the data socket. the hello packet will also be from this socket. 
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    perror("socket");
    exit(-1);
  }
  DBG("tcdclt: socket successfully ...\n");

  struct sockaddr_in singlesock;
  bzero(&singlesock, sizeof(singlesock));
  singlesock.sin_family = AF_INET;
  singlesock.sin_addr.s_addr = inet_addr(dstip_str);
  singlesock.sin_port = htons(dstport);

  // initialize the memory ...
  if (vos_meminit() == -1)
  {
    fprintf(stderr, "vos_meminit: failed\n");
    close(sock);
    exit(-1);
  }

  //
  // --ip-file
  // address a serial of commands here
  //
  int j;
  j = IDXOPT("ip-file");
  if (option_data_val[j].flag && option_data_val[j].arg != NULL)
  {
    if (get_filelines(option_data_val[j].arg) <= 0)
    {
      DBG("file %s might be empty!\n", option_data_val[j].arg);
      goto exit_of_main;
    }
    if (NULL == get_ip_file_fp(option_data_val[j].arg))
    {
      DBG("fail to open file %s\n", option_data_val[j].arg);
      goto exit_of_main;
    }
  }
  
  //
  // reboot
  //
  if (option_data_val[IDXOPT("reboot")].flag && option_data_val[IDXOPT("reboot")].enabled)
  {
    if (option_data_val[IDXOPT("ip")].flag && option_data_val[IDXOPT("ip")].arg != NULL)
    {
      if (1 == is_valid_ipv4 (option_data_val[IDXOPT("ip")].arg))
      {
        printf("reboot client with IP: %s\n", option_data_val[IDXOPT("ip")].arg);
        cmdreboot_ip_send(sock, singlesock, option_data_val[IDXOPT("ip")].arg);
      }
    }
    else if (option_data_val[IDXOPT("ip-file")].flag && option_data_val[IDXOPT("ip-file")].arg != NULL)
    {
      //
      char buf[16];
      int r;
      
      while (r = get_next_ip_from_file(buf, sizeof(buf)))
      {
        if (r == 1)
        {
          printf("reboot client with IP: %s\n", buf);
          cmdreboot_ip_send(sock, singlesock, buf);
        }
        else if (r == 2)
        {
          continue;
        }
      }
      fclose(ipfilefp);
    }
    else if (option_data_val[IDXOPT("all")].flag && option_data_val[IDXOPT("all")].enabled)
    {
      printf("reboot all clients\n");
      cmdreboot_send(sock, singlesock);
    }
    else
    {
      printf("insufficient arguments for %s command\n", "reboot");
      goto exit_of_main;
    }
  }
  
  //
  // poweroff
  //
  if (option_data_val[IDXOPT("poweroff")].flag && option_data_val[IDXOPT("poweroff")].enabled)
  {
    if (option_data_val[IDXOPT("ip")].flag && option_data_val[IDXOPT("ip")].arg != NULL)
    {
      if (1 == is_valid_ipv4 (option_data_val[IDXOPT("ip")].arg))
      {
        printf("poweroff client with IP: %s\n", option_data_val[IDXOPT("ip")].arg);
        cmdshutdown_ip_send(sock, singlesock, option_data_val[IDXOPT("ip")].arg);
      }
    }
    else if (option_data_val[IDXOPT("ip-file")].flag && option_data_val[IDXOPT("ip-file")].arg != NULL)
    {
      //
      char buf[16];
      int r;
      
      while (r = get_next_ip_from_file(buf, sizeof(buf)))
      {
        if (r == 1)
        {
          printf("poweroff client with IP: %s\n", buf);
          cmdshutdown_ip_send(sock, singlesock, buf);
        }
        else if (r == 2)
        {
          continue;
        }
      }
      fclose(ipfilefp);
    }
    else if (option_data_val[IDXOPT("all")].flag && option_data_val[IDXOPT("all")].enabled)
    {
      printf("poweroff all clients\n");
      cmdshutdown_send(sock, singlesock);
    }
    else
    {
      printf("insufficient arguments for %s command\n", "poweroff");
      goto exit_of_main;
    }
  }

  //
  // initialize
  //
  if (option_data_val[IDXOPT("initialize")].flag && option_data_val[IDXOPT("initialize")].enabled)
  {
    int user_partsz = 8192;
    if (option_data_val[IDXOPT("user-partsz")].flag && option_data_val[IDXOPT("user-partsz")].arg != NULL)
    {
      user_partsz = atoi(option_data_val[IDXOPT("user-partsz")].arg);
    }
    
    if (option_data_val[IDXOPT("ip")].flag && option_data_val[IDXOPT("ip")].arg != NULL)
    {
      if (1 == is_valid_ipv4 (option_data_val[IDXOPT("ip")].arg))
      {
        printf("initialize client with IP: %s, user_partsz: %d\n", option_data_val[IDXOPT("ip")].arg, user_partsz);
        cmdinitialize_ip_send(sock, singlesock, user_partsz, option_data_val[IDXOPT("ip")].arg);
      }
    }
    else if (option_data_val[IDXOPT("ip-file")].flag && option_data_val[IDXOPT("ip-file")].arg != NULL)
    {
      //
      char buf[16];
      int r;
      
      while (r = get_next_ip_from_file(buf, sizeof(buf)))
      {
        if (r == 1)
        {
          printf("initialize client with IP: %s, user_partsz: %d\n", buf, user_partsz);
          cmdinitialize_ip_send(sock, singlesock, user_partsz, buf);
        }
        else if (r == 2)
        {
          continue;
        }
      }
      fclose(ipfilefp);
    }
    else if (option_data_val[IDXOPT("all")].flag && option_data_val[IDXOPT("all")].enabled)
    {
      printf("initialize all clients, user_partsz: %d\n", user_partsz);
      cmdinitialize_send(sock, singlesock, user_partsz);
    }
    else
    {
      printf("insufficient arguments for %s command\n", "initialize");
      goto exit_of_main;
    }
  }
  
  //
  // update-firmware
  //
  if (option_data_val[IDXOPT("update-firmware")].flag && option_data_val[IDXOPT("update-firmware")].enabled)
  {
    if (option_data_val[IDXOPT("ip")].flag && option_data_val[IDXOPT("ip")].arg != NULL)
    {
      if (1 == is_valid_ipv4 (option_data_val[IDXOPT("ip")].arg))
      {
        printf("update-firmware client with IP: %s\n", option_data_val[IDXOPT("ip")].arg);
        cmdupdatefirmware_ip_send(sock, singlesock, option_data_val[IDXOPT("ip")].arg);
      }
    }
    else if (option_data_val[IDXOPT("ip-file")].flag && option_data_val[IDXOPT("ip-file")].arg != NULL)
    {
      //
      char buf[16];
      int r;
      
      while (r = get_next_ip_from_file(buf, sizeof(buf)))
      {
        if (r == 1)
        {
          printf("update-firmware client with IP: %s\n", buf);
          cmdupdatefirmware_ip_send(sock, singlesock, buf);
        }
        else if (r == 2)
        {
          continue;
        }
      }
      fclose(ipfilefp);
    }
    else if (option_data_val[IDXOPT("all")].flag && option_data_val[IDXOPT("all")].enabled)
    {
      printf("update-firmware all clients\n");
      cmdupdatefirmware_send(sock, singlesock);
    }
    else
    {
      printf("insufficient arguments for %s command\n", "update-firmware");
      goto exit_of_main;
    }
  }
  
  //
  // change-server
  //
  if (option_data_val[IDXOPT("change-server")].flag && option_data_val[IDXOPT("change-server")].enabled)
  {
    if (! (option_data_val[IDXOPT("server-ip")].flag && option_data_val[IDXOPT("server-ip")].arg != NULL))
    {
      printf("insufficient arguments for %s command\n", "change-server");
      goto exit_of_main;
    }
    
    if (option_data_val[IDXOPT("ip-file")].flag && option_data_val[IDXOPT("ip-file")].arg != NULL)
    {
      //
      char buf[16];
      int r;
      
      while (r = get_next_ip_from_file(buf, sizeof(buf)))
      {
        if (r == 1)
        {
          printf("send change-server command to client with IP: %s\n", buf);
          cmdcgsvrip_ip_send(sock, singlesock, inet_addr(option_data_val[IDXOPT("server-ip")].arg),buf);
        }
        else if (r == 2)
        {
          continue;
        }
      }      
      fclose(ipfilefp);
    }
    else if (option_data_val[IDXOPT("all")].flag && option_data_val[IDXOPT("all")].enabled)
    {
      printf("send change-server command to all clients\n");
      cmdcgsvrip_send(sock, singlesock, inet_addr(option_data_val[IDXOPT("server-ip")].arg));
    }
    else
    {
      printf("insufficient arguments for %s command\n", "change-server");
      goto exit_of_main;
    }
  }  
  
  
  //
  // remove-image
  //
  if (option_data_val[IDXOPT("remove-image")].flag && option_data_val[IDXOPT("remove-image")].enabled)
  {
    
    if ( !(option_data_val[IDXOPT("image-org-name")].flag && option_data_val[IDXOPT("image-org-name")].arg != NULL)
      || !(option_data_val[IDXOPT("image-id")].flag && option_data_val[IDXOPT("image-id")].arg != NULL) )
    {
      printf("insufficient arguments for %s command\n", "remove-image");
      goto exit_of_main;
    }
    
    if (option_data_val[IDXOPT("ip-file")].flag && option_data_val[IDXOPT("ip-file")].arg != NULL)
    {
      //
      char buf[16];
      int r;
      
      while (r = get_next_ip_from_file(buf, sizeof(buf)))
      {
        if (r == 1)
        {
          printf("send remove-image command to client with IP: %s\n", buf);
          cmdrmimage_ip_send(sock, singlesock, option_data_val[IDXOPT("image-org-name")].arg, atoi(option_data_val[IDXOPT("image-id")].arg), buf);
        }
        else if (r == 2)
        {
          continue;
        }
      }        

      fclose(ipfilefp);
    }
    else if (option_data_val[IDXOPT("all")].flag && option_data_val[IDXOPT("all")].enabled)
    {
      printf("send remove-image command to all clients\n");
      cmdrmimage_send(sock, singlesock, option_data_val[IDXOPT("image-org-name")].arg, atoi(option_data_val[IDXOPT("image-id")].arg) );
    }
    else
    {
      printf("insufficient arguments for %s command\n", "remove-image");
      goto exit_of_main;
    }
  }
 
  //
  // set-resolution
  //
  if (option_data_val[IDXOPT("set-resolution")].flag && option_data_val[IDXOPT("set-resolution")].enabled)
  {
    int client_scr_width;
    int client_scr_height;
    if (! (option_data_val[IDXOPT("width")].flag && option_data_val[IDXOPT("width")].arg != NULL))
    {
      printf("insufficient arguments for %s command\n", "set-resolution");
      goto exit_of_main;
    }
    if (! (option_data_val[IDXOPT("height")].flag && option_data_val[IDXOPT("height")].arg != NULL))
    {
      printf("insufficient arguments for %s command\n", "set-resolution");
      goto exit_of_main;
    }

    client_scr_width = atoi(option_data_val[IDXOPT("width")].arg);
    client_scr_height = atoi(option_data_val[IDXOPT("height")].arg);
 
    if (option_data_val[IDXOPT("ip")].flag && option_data_val[IDXOPT("ip")].arg != NULL)
    {
      if (1 == is_valid_ipv4 (option_data_val[IDXOPT("ip")].arg))
      {
        printf("set-resolution client with IP: %s\n", option_data_val[IDXOPT("ip")].arg);
        cmdresol_ip_send(sock, singlesock, option_data_val[IDXOPT("ip")].arg, client_scr_width, client_scr_height);
      }
    }
    else if (option_data_val[IDXOPT("ip-file")].flag && option_data_val[IDXOPT("ip-file")].arg != NULL)
    {
      //
      char buf[16];
      int r;
      
      while (r = get_next_ip_from_file(buf, sizeof(buf)))
      {
        if (r == 1)
        {
          printf("set-resolution client with IP: %s\n", buf);
          cmdresol_ip_send(sock, singlesock, buf, client_scr_width, client_scr_height);
        }
        else if (r == 2)
        {
          continue;
        }
      }
      fclose(ipfilefp);
    }
    /*else if (option_data_val[IDXOPT("all")].flag && option_data_val[IDXOPT("all")].enabled)
    {
      printf("reboot all clients\n");
      cmdresol_send(sock, singlesock);
    }*/
    else
    {
      printf("insufficient arguments for %s command\n", "set-resolution");
      goto exit_of_main;
    }
  }
  
  //
  // enable/disable access to usb-storage devices
  //
  if (option_data_val[IDXOPT("enable-usb-storage")].flag && option_data_val[IDXOPT("enable-usb-storage")].arg != NULL)
  {
    int enable_usb_storage;

    enable_usb_storage = atoi(option_data_val[IDXOPT("enable-usb-storage")].arg);
 
    if (option_data_val[IDXOPT("ip")].flag && option_data_val[IDXOPT("ip")].arg != NULL)
    {
      if (1 == is_valid_ipv4 (option_data_val[IDXOPT("ip")].arg))
      {
        printf("set usb storage mode [%d],  client with IP: %s\n", enable_usb_storage, option_data_val[IDXOPT("ip")].arg);
        cmdusbctrl_ip_send(sock, singlesock, option_data_val[IDXOPT("ip")].arg, enable_usb_storage);
      }
    }
    else if (option_data_val[IDXOPT("ip-file")].flag && option_data_val[IDXOPT("ip-file")].arg != NULL)
    {
      //
      char buf[16];
      int r;
      
      while (r = get_next_ip_from_file(buf, sizeof(buf)))
      {
        if (r == 1)
        {
          printf("set usb storage mode [%d], client with IP: %s\n", enable_usb_storage, buf);
          cmdusbctrl_ip_send(sock, singlesock, buf, enable_usb_storage);
        }
        else if (r == 2)
        {
          continue;
        }
      }
      fclose(ipfilefp);
    }
    /*else if (option_data_val[IDXOPT("all")].flag && option_data_val[IDXOPT("all")].enabled)
    {
      printf("reboot all clients\n");
      cmdresol_send(sock, singlesock);
    }*/
    else
    {
      printf("insufficient arguments for %s command\n", "set-resolution");
      goto exit_of_main;
    }
  }
  
 

  //
  // boot
  // to-do
  //  
/*  
 if (!strcmp(argv[1], "--boot"))
  {
    if (argc == 8)
    {
      printf("boot client image_id[%s], image_org_name[%s], image_name[%s], user_token[%s], user_name[%s], client_ip[%s]\n", argv[2], argv[3], argv[4], argv[5], argv[6], argv[7]);
      if (1 == is_valid_ipv4 (argv[7]))
      {
        cmdboot_ip_send(sock, singlesock, atoi(argv[2]), argv[3], argv[4], argv[5], argv[6], inet_addr(argv[7]));
      }
      else
      {
        printf("invalid IPv4: %s\n", argv[7]);
      }
    }
    else
    {
      printf("illegal syntax\n");
    }
  }
  else
  {
    printf("unknown parameter: %s\n", argv[1]);
  }
*/


  //
  // help
  //
  if (option_data_val[IDXOPT("help")].flag && option_data_val[IDXOPT("help")].enabled)
  {
    tcdctl_usage();
  }
  // to close the control socket
exit_of_main:
  DBG("tcdclt: close the socket\n");
  close(sock);

  exit(0);
}


#else

void tcdctl_usage()
{
  printf("\nUsage: ");

  printf("\n\t%s <ip> <port> [filelist]\n\t\t--- use eth0, current working dir\n", argv0);

  printf("\nArguments: ");
  printf("\n\t<ip>:\tip address of TCD server");
  printf("\n\t<port>:\tserver's listening UDP port");
  printf("\n\t<filelist>:\tfiles to be send\n\n");

  return;
}

// interactive mode ... 
// format: tcdctl <ip> <port> <filename1> <filename2> ..., 
// <ip> and <port> are the target server's IP and listening port#
int main(int argc, char **argv)
{
  int sock; 
  struct sockaddr_in servadd;
  char *cifn, *localip;

  int dstport;
  uint32 dstip;
  char *dstip_str=NULL;
  struct sockaddr_in localsock;
  int ch, res;
  fd_set rdfds;
  struct timeval tv;


  // parameter handling
  if (argc < 3)
  {
    tcdctl_usage();

    exit(0);
  }

  cifn = "eth0";
  dstip_str = argv[1];
  dstip = inet_addr(dstip_str);
  dstport = atoi(argv[2]);

  

  strcpy(flw_dir, "./");

  printf("flw_dir=<%s>\n", flw_dir);
  if (flw_dir[strlen(flw_dir)-1] != '/')
  {
    int x = strlen(flw_dir);
    flw_dir[x] = '/';
    flw_dir[x+1] = 0;  
  }
  printf("flw_dir=<%s>\n", flw_dir);

  // get the local ipadd of the interface
  if ((localip = ifn2ip_str(cifn)) == NULL)
  {
    perror("ifn2ip_str");  
    exit(-1);
  }

  printf("\nMulticast Deployment Controller\n");
  printf("Interface [%s] LocalIP: [%s] Network [0x%x:%d] \n", cifn, localip, dstip, dstport);

    
  // flw_name: the file to be sent ...

  // to create the data socket. the hello packet will also be from this socket. 
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    perror("socket");
    exit(-1);
  }
  DBG("tcdclt: socket successfully ...\n");


  struct sockaddr_in singlesock;
  bzero(&singlesock, sizeof(singlesock));
  singlesock.sin_family = AF_INET;
  singlesock.sin_addr.s_addr = inet_addr(dstip_str);
  singlesock.sin_port = htons(dstport);

  // initialize the memory ...
  if (vos_meminit() == -1)
  {
    fprintf(stderr, "vos_meminit: failed\n");
    close(sock);
    exit(-1);
  }


//  char *fn[] = {"../ce11.img", "../ce12.img", "../ce13.img", "../ce14.img", "../ce15.img"};
  // char *fn[] = {"../ce15.img", "../ce14.img", "../ce13.img", "../ce12.img", "../ce11.img"};
//  char *fn[] = {"../ce1.img", "../ce2.img", "../ce.img"}; 
//  char *fn[] = {"../tree_151218.tar.gz", "../ce1.img", "../buildroot-2011.11.tar.bz2", "../msock_160121.tar.gz", "../ha_151106.tar.gz"};
//  char *fn[] = {"../msock_160121.tar.gz", "../ha_151106.tar.gz"};
//  char *fn[] = {"../ce1.img", "../msock_160121.tar.gz", "../ha_151106.tar.gz", "../ce2.img", "../ce4.img", "../ce6.img", "../ce8.img", "../ce.img"}; 

//  char *fn[] = {"../ce4.img", "../ce8.img"}; 
//  char *fn[] = {"../ce1.img", "../ce2.img"}; 
//  char *fn[] = {"../ce1.img", "../ce2.img", "../ce4.img", "../ce8.img"}; 
//  char *fn[] = {"../ce4.img", "../ce2.img", "../ce1.img"}; 
//  char *fn[] = {"../diskimage/ce1.img", "../diskimage/ce2.img", "../diskimage/ce4.img", "../diskimage/ce.img"}; 
  char *fn[] = {"../ce4.img", "../ce2.img"}; 


// used in 3F lab...
//  char *fn[] = {"/tc/diskimage/ubuntu-10.img", "/tc/diskimage/win7_en.img"}; 

  // to enable the raw terminal mode so we can use hotkey ...
  term_raw();


  while(1)
  {
    // add network socket and stdin to the FDSET
    FD_ZERO(&rdfds);
    FD_SET(0, &rdfds);
    // and set the select timeout to forever
    tv.tv_sec = 0;
    tv.tv_usec = 0;

    res = select(0+1, &rdfds, NULL, NULL, &tv);
    // DBG("res=%d\n", res);
  
    if (res > 0)
    {
      // stdin events
      if (FD_ISSET(0, &rdfds))
      {
        ch = getchar();

        switch(ch)
        {
          case 's':
            printf("start the deployment ...\n");
            cmdstart_send(sock, singlesock, fn, sizeof(fn)/sizeof(char *));
            break;


          case 'l':
            printf("list ...\n");
            cmdlist_send(sock, singlesock);
            break;
          case 'd':
            printf("quit the TCD server ...\n");
            cmdquit_send(sock, singlesock);
            break;
          case 'b':
            printf("reboot all clients ...\n");
            cmdreboot_send(sock, singlesock);
            break;
          case '1':
            printf("restart clients tcdclt application ...\n");
            cmdrestartapp_send(sock, singlesock);
            break;

          case '0':
            printf("shutdown all clients ...\n");
            cmdshutdown_send(sock, singlesock);
            break;
          case 'f':
            printf("format hard disk of clients ...\n");
            cmdformat_send(sock, singlesock);
            break;


          case 'i':
            printf("identify each client ...\n");
            cmdidentify_send(sock, singlesock);
            break;

          case 'u':
            printf("user-defined command ...\n");
            cmduser_send(sock, singlesock); 
            break;

          case 'h':
            ctl_help();
            break;
          case 'q':
            printf("quit the tcd controller ...\n");
            goto _TCDCTL_END_LOOP;
            break;

          // command "version": to print the build info.
          case 'v':
            print_buildinfo();
            break;


          default:
            printf("unknown command [%c : 0x%x]\n", ch, ch);
            break;
        }


      }

    }
  }

_TCDCTL_END_LOOP:

  // to restore the terminal mode
  term_restore();  

  // to close the control socket  
  DBG("tcdclt: close the socket\n");
  close(sock);

  exit(0);
}


#endif

