#/bin/sh
# if run successfully, exit with 0;
# otherwise, exit with a non-zero value.

SCRIPT_PATH=$0
SCRIPT_NAME=${SCRIPT_PATH##*/}
# define the disk to be initialized
DEV_NODE='/dev/sda'

# define MOUNTING POINT
DISKDIR='/mnt/disk'
USBDIR='/mnt/usb'
RSVDDIR='/mnt/rsvd'
INITHD_LOG_FILE="/tmp/${SCRIPT_NAME%.sh}.log"
INITRD_CPIO_FILE="$RSVDDIR/boot/initrd.cpio"
INITRD_IMAGE_FILE="$RSVDDIR/boot/initrd.img"
FLAG_DISABLE_VD=1


######################### Micros def (begin) ###################################
# define error code
# exit normally
ERR_NORMAL=0
# exit if illegal parameter
ERR_PARAMETER=1
# exit if given parameter `userPartSize` is improper
ERR_SIZE=2
# exit when user interrupted
ERR_INTERRUPT=3
# exit if no hard drive
ERR_NOHD=4
# exit if umount failed
ERR_UMOUNT_SYSPART=5
ERR_UMOUNT_DSKPART=6
# exit if clean MBR failed
ERR_CLEAN_MBR=7
# exit if tag MBR failed
ERR_TAG_MBR=8
# exit if install MBR hook failed
ERR_HOOK_MBR=9
# exit if fdisk failed
ERR_FDISK=10
# exit if mount failed
ERR_MOUNT_SYSPART=11
ERR_MOUNT_DSKPART=12
# exit if format failed
ERR_FORMAT_SYSPART=13
ERR_FORMAT_DSKPART=14
ERR_FORMAT_USRPART=15
# exit if part not exist
ERR_SYSPART_NOTEXIST=16
ERR_DSKPART_NOTEXIST=17
ERR_USRPART_NOTEXIST=18
# exit if install booting files failed
ERR_RM_BLFILES=19
ERR_CP_BLFILES=20
ERR_GEN_INITRD=21
ERR_COMP_INITRD=22
ERR_INITRD_MISSING=23
ERR_INSTALL_BL=24
# exit if gen VD
ERR_GEN_VD=25
########################## Micros def (end) ####################################


######################## Function def (begin) ##################################
# helpful infomartion
usage()
{
    cat >&7 <<EOT
SYNOPSIS
    ${SCRIPT_PATH} [userPartSize]

DESCRIPTION
    userPartSize, size of the user partition (unit: MiB).
    If \`userPartSize' is not specified, then the script would use 8192 instead.
    if \`userPartSize' is zero, then the user partition would not be created.
    Otherwise, the script tries to create a user partition with given size.

EXAMPLE
    "${SCRIPT_PATH}" creates a 8 GiB partition as the user partition
    "${SCRIPT_PATH} 0" creates no user partitions
    "${SCRIPT_PATH} 10240" creates a 10 GiB partition as the user partition
EOT
}

tcmsg_init()
{
    echo -n "" > $INITHD_LOG_FILE
    exec 7>&1 8>&2
    exec >> $INITHD_LOG_FILE 2>&1
}

# no color
tcprn()
{
    tccount=$(($tccount + 1))
    tccount_ex=`echo "$tccount" | awk '{printf "%02d", $1}'`
    # print on the screen
    #echo -e "$@" >&7
    echo "[$tccount_ex] $@" >&7
    # print to log file
    echo -e "\n[$tccount_ex] $@"
}

# colorful
# print "[Info] " before every piece of information
tcinfo()
{
    # print on the screen
    echo -e "\033[32m[INFO]\033[0m $@" >&7
    # print to log file
    echo "[INFO] $@"
}

tcerr()
{
    # print on the screen
    echo -e "\033[31m[ERROR]\033[0m $@" >&8
    # print to log file
    echo "[ERROR] $@" >&2
}

tcwarn()
{
    # print on the screen
    echo -e "\033[35m[WARNING]\033[0m $@" >&8
    # print to log file
    echo "[WARNING] $@" >&2
}

tcsavelogas(){
    logdevnode=${DEV_NODE}1
    logdestmp=$DISKDIR
    tick=10
    tcinfo "try saving another copy of logs to $logdestmp"
    while [ ! -b $logdevnode -a $tick -gt 0 ] ; do
        tcinfo "save logs, time left: $tick seconds"
        tick=$(($tick - 1))
        sleep 1
    done

    if [ ! -b $logdevnode ] ; then
        tcwarn "$logdevnode : no such file"
        tcwarn "fail to save another copy of logs to $logdestmp"
        return
    fi

    mount | grep -q $logdevnode
    if [ $? -ne 0 ] ; then
        mount $logdevnode $logdestmp
        t=$?
        sleep 1
        if [ $t -ne 0 ] ; then
            tcwarn "fail to mount $logdevnode to $logdestmp"
            tcwarn "fail to save another copy of logs to $logdestmp"
            return
        fi
        mount_op=YES
    fi

    mount | grep -q $logdevnode
    if [ $? -eq 0 ] ; then
        newdestmp=`df $logdevnode | sed -n '2p' | awk '{print $NF}'`
        tcinfo "mounting point for $logdevnode : $newdestmp"

        tcinfo "mount_op = $mount_op"
        tcinfo "saving a copy of file $INITHD_LOG_FILE to $newdestmp successfully"
        cp $INITHD_LOG_FILE $newdestmp

        if [ "x$mount_op" == "xYES" ] ; then
            tcinfo "try umounting $newdestmp ..."
            umount $newdestmp
            if [ $? -ne 0 ] ; then
                tcwarn "fail to umount $newdestmp"
            fi
        fi
    else
        tcwarn "fail to find mounting point for $logdevnode"
        tcwarn "fail to save another copy of logs to $logdestmp"
        return
    fi
}

tcexit()
{
    tcinfo "save logs to file $INITHD_LOG_FILE"
    tcsavelogas

    exec 1>&7 2>&8
    exec 7>&- 8>&-
    exit $1
}

# if parameter is a number return 0, otherwise return a non-zero value
is_a_number()
{
    if [ -z $1 ] ; then
        return 1
    elif echo $1 | egrep -q "^[0-9]+$" ; then
        return 0
    else
        return 2
    fi
}

int_handler()
{
    tcerr "received TERM or INT signal"
    tcerr "task is terminated"
    sleep 1
    tcexit $ERR_INTERRUPT
}
######################### Function def (end) ###################################

######################## Set log file (begin) ##################################
tcmsg_init
tccount=0
tcinfo "set log to file $INITHD_LOG_FILE"
curr_time=`date "+%F %T"`
tcinfo "logging time: ${curr_time}"
######################### Set log file (end) ###################################



############################# Program entry ####################################
# script entry:
trap int_handler INT
trap int_handler TERM

# address script parameters
if [ $# -lt 1 ] ; then
    PART_2_SIZE_MIB=8192
    if [ "x$FLAG_DISABLE_VD" == "x1" ] ; then
        D_FLAG=1
        VD_FLAG=0
        VDSZ=0
    else
        D_FLAG=0
        VD_FLAG=1
        VDSZ=8192
    fi
    tcinfo "size for user partition (unit: MiB): ${PART_2_SIZE_MIB}"
else
    if [ -b "$DEV_NODE" ] ; then
        tcinfo "device block node: ${DEV_NODE}"
    else
        tcerr  "$DEV_NODE: no such block node"
        tcexit $ERR_NOHD
    fi

    para_tmp=$1
    is_a_number $para_tmp
    if [ $? -eq 0 ] && [ $para_tmp -ge 0 ] ; then
        PART_2_SIZE_MIB=$para_tmp
        if [ $PART_2_SIZE_MIB -gt 0 ] ; then
            # create second partition
            if [ "x$FLAG_DISABLE_VD" == "x1" ] ; then
                D_FLAG=1
                VD_FLAG=0
                VDSZ=$PART_2_SIZE_MIB
            else
                D_FLAG=0
                VD_FLAG=1
                VDSZ=$PART_2_SIZE_MIB
            fi
        else
            # do not create second partition
            D_FLAG=0
            VD_FLAG=0
            VDSZ=$PART_2_SIZE_MIB
        fi
    else
        tcerr "$para_tmp: illegal parameter"
        usage
        tcexit $ERR_PARAMETER
    fi
    tcinfo "specified size for user partition (unit: MiB): ${PART_2_SIZE_MIB}"
fi

# sleep before umount, login UI will not prompt err msg
sleep 2
sh /tcdscript/pre-inithd.sh
ret=$?
if [ $ret -ne 0 ] ; then
    tcerr "fail to umount partitions"
    tcexit $ret
fi

# disk name, such as "sda"
DEV_NAME=${DEV_NODE##*/}
tcinfo "device name: $DEV_NAME"

DEV_SECTORS_MAX_STR='2^32'
DEV_SECTORS_MAX=`echo "$DEV_SECTORS_MAX_STR" | bc`
tcinfo "address only 2^32 (i.e., $DEV_SECTORS_MAX) sectors"

# clear disk table
tcprn "clear disk partition table "
tcinfo "clear disk partition table ...."
dd if=/dev/zero of=$DEV_NODE bs=1 count=66 seek=446
if [ ! $? -eq 0 ] ; then
    tcerr "clear disk partition table"
    tcexit $ERR_CLEAN_MBR
fi

tcprn "clear disk signature ..."
dd if=/dev/zero of=$DEV_NODE bs=1 count=4 seek=440
if [ ! $? -eq 0 ] ; then
    tcerr "clear disk signature"
    tcexit $ERR_CLEAN_MBR
fi

# disk constants
BYTES_PER_SECTOR=`fdisk -u -l $DEV_NODE | sed -n 4p | awk '{print $9}'`
SECTORS_PER_TRACK=`fdisk -l $DEV_NODE | sed -n 3p | awk '{print $3}'`
HEADS_PER_CYLINDER=`fdisk -l $DEV_NODE | sed -n 3p | awk '{print $1}'`

tcinfo "sector size: $BYTES_PER_SECTOR"
tcinfo "sectors per track: $SECTORS_PER_TRACK"
tcinfo "heads per cylinder: $HEADS_PER_CYLINDER"

# disk size (unit: sectors)
FILE_SECTORS="/sys/block/$DEV_NAME/size"
if [ -f $FILE_SECTORS ] ; then
    DEV_SECTORS=`cat $FILE_SECTORS`
else
    DEV_SECTORS=`fdisk -u -l $DEV_NODE | sed -n 3p | awk '{print $8}'`
fi

# do NOT use $DEV_SECTORS directly, use $DEV_SECTORS_SEC instead
tcinfo "device sectors: $DEV_SECTORS"

DEV_MIBS=`echo "${DEV_SECTORS} * ${BYTES_PER_SECTOR} / 2^20" | bc`
tcinfo "device capacity (unit: MiB): $DEV_MIBS"

# do NOT use $DEV_CYLINDERS directly, use DEV_CYLINDERS_SEC instead
DEV_CYLINDERS=`fdisk -l $DEV_NODE | sed -n 3p | awk '{print $5}'`
tcinfo "device cylinders: $DEV_CYLINDERS"

# aligns partition beginning sector in 2048 sector units
DEV_ALIGNMENT='2048'



################## TC Partitioning Specification (Begin) #######################
## partitioning description
# disk capacity: use at most 2^32 sectors couting from head of the device
# partition size:
#   1) assign fixed size for reserved system partition firstly;
#   2) assign specified size for user partition secondly;
#   3) assign the remaining size for system partition finally;
# partition head: all partitions begin at the sector with which it compilies
#   the 2048 sectors alignment mechanism
# partition tail: all partitions end at the end of a cylinder, this is
#   (cylinder alignment)
 
# primary partition
# partition No.: 4
# description: reserved system partition for TC firmware
# partition (volume) label: RESERVED
# filesystem: ext3
# partition id: 0x12
# partition end at (sector): if disk capacity is not greater than 2TiB,
#   this partition ends at the last sector of the last cylinder of the disk;
#   otherwise, this partition ends at the last sector of the cylinder conforming
#   to 2 TiB limits.
# partition size (unit: MiB): 480

# primary partition
# partition No.: 2
# description: store user proprietary data
# partition (volume) label: USER
# filesystem: NTFS
# partition id: 0x7
# partition end at: the last sector of the cylinder which is previously adjacent
#   to the cylinder covering the first sector of the reserved system partition
#   (cylinder alignment)
# partition size (unit: MiB): user specified
# partition begin at: the 1st 2048-alignment sector meeting the user patition
#   size user specified, counting backward, i.e., from the last partition sector
#   to disk head

# primary partition
# partition No.: 1
# description: store system data
# partition (volume) label: SYSTEM
# filesystem: ext3
# partition id: 0x12
# partition begin at (sector): 2048 
# partition size (unit: MiB): the remaining disk resource after allocating for
#   reserved system partition and user partition 
# partition end at (sector): the last sector of the cylinder which is previously
#   adjacent to the cylinder covering the first sector of the user partition
#   (cylinder alignment)


################### TC Partitioning Specification (end) ########################



################## TC Partitions Secure Check (begin) ##########################
# if the capacity of disk sectors exceeds 2^32, only use the front 2^32 sectors
# of the disk
t=`echo "${DEV_SECTORS} > ${DEV_SECTORS_MAX}" | bc`
if [ $t -eq 1 ] ; then
    DEV_SECTORS_SEC=${DEV_SECTORS_MAX}
    tcwarn "the number of disk sectors is greater than 2^32"
    tcinfo "adjust disk secure sectors to: ${DEV_SECTORS_MAX}"
else
    DEV_SECTORS_SEC=$DEV_SECTORS
fi
tcinfo "disk secure sectors: ${DEV_SECTORS_SEC}"

# calculate the secure last cylinder
SECTORS_PER_CYLINDER=`echo "${SECTORS_PER_TRACK} * ${HEADS_PER_CYLINDER}" | bc`
DEV_CYLINDERS_SEC=`echo "${DEV_SECTORS_SEC} / ${SECTORS_PER_CYLINDER}" | bc`
t=`echo "$DEV_CYLINDERS_SEC > ${DEV_CYLINDERS}" | bc`
if [ $? -eq 1 ] ; then
    DEV_CYLINDERS_SEC=${DEV_CYLINDERS}
fi
tcinfo "disk secure cylinders: ${DEV_CYLINDERS_SEC}"

################### TC Partitions Secure Check (end) ###########################



###################### Partition 4 Calc (begin) ################################
# 1. ends at the last sector of the last cylinder or the last secure cylinder
# 2. starts at the greatest 2048-sector-alignment sector position meeting the
#    partition size requirement (a fixed value) couting from disk head to disk
# 3. partition size: 480 MiB

DEV_SECTORS_SEC_REMAINING=$DEV_SECTORS_SEC
# if the desirable parition size (480, unit: MiB) of reserved system partition
# exceeds the disk capacity, exit with error messages
PART_4_SIZE_MIB='480'
part_4_size_sectors=`echo "${PART_4_SIZE_MIB}*2^20 / ${BYTES_PER_SECTOR}" | bc`
t=`echo "$part_4_size_sectors > $DEV_SECTORS_SEC_REMAINING" | bc`
if [ $t -eq 1 ] ; then
    tcerr "insufficient storage for reserved system partition (partition 4)"
    tcinfo "storage(unit: MiB): ${DEV_MIBS}"
    tcinfo "reserved system partition storage(unit: MiB): ${PART_4_SIZE_MIB}"
    tcexit $ERR_SIZE
fi

# align partition end sector to end of last secure cylinder of the disk
part_4_end_sector=`echo "${DEV_SECTORS_SEC} / ${SECTORS_PER_CYLINDER} * \
    ${SECTORS_PER_CYLINDER}" | bc`
part_4_end_sector_alignment=$part_4_end_sector

part_4_begin_sector=`echo "$part_4_end_sector - ${part_4_size_sectors}" | \
    bc`
part_4_begin_sector_alignment=`echo "${part_4_begin_sector} / \
    ${DEV_ALIGNMENT} * ${DEV_ALIGNMENT}" | bc`

# check the begin sector
t=`echo "$part_4_begin_sector_alignment <= $DEV_ALIGNMENT" | bc`
if [ $t -eq 1 ] ; then
    tcerr "insufficient storage for reserved system partition (partition 4)"
    tcinfo "storage(unit: MiB): ${DEV_MIBS}"
    tcinfo "reserved system partition storage(unit: MiB): ${PART_4_SIZE_MIB}"
    tcinfo "partition 4 starts at (unit: sector):$part_4_begin_sector_alignment"
    tcexit $ERR_SIZE
fi

part_4_size_sectors_alignment=`echo "${part_4_end_sector_alignment} - \
    ${part_4_begin_sector_alignment}" | bc`
part_4_size_mib_alignment=`echo "${part_4_size_sectors_alignment} * \
    ${BYTES_PER_SECTOR} / 2^20" | bc`
#part_4_end_sector_exact=$part_4_end_sector_alignment
####################### Partition 4 Calc (end) #################################


###################### Partition 2 Calc (begin) ################################
# 1. ends at the last sector of the cylinder previously adjacent to the cylinder
#    covering the first sector of partition 4
# 2. starts at the greatest 2048-sector-alignment position meeting the
#    partition size requirement (user specified) couting from disk head to disk
#    tail
# 3. size: user specified
# 
if [ $D_FLAG -eq 0 ] ; then
   PART_2_SIZE_MIB=0
fi
DEV_SECTORS_SEC_REMAINING=$part_4_begin_sector_alignment
# calculate the end sector
part_2_size_sectors=`echo "$PART_2_SIZE_MIB * 2^20 / ${BYTES_PER_SECTOR}" | bc`
t=`echo "$part_2_size_sectors > $DEV_SECTORS_SEC_REMAINING" | bc`
if [ $t -eq 1 ] ; then
    tcerr "insufficient storage for reserved system partition (partition 2)"
    tcinfo "storage(unit: MiB): ${DEV_MIBS}"
    tcinfo "user partition storage(unit: MiB): ${PART_2_SIZE_MIB}"
    tcexit $ERR_SIZE
fi

part_2_end_sector=`echo "$part_4_begin_sector_alignment - 0" | bc`
part_2_end_sector_alignment=`echo "$part_2_end_sector / \
    ${SECTORS_PER_CYLINDER} * $SECTORS_PER_CYLINDER" | bc`

# calculate the start sector
part_2_begin_sector=`echo "$part_2_end_sector_alignment - \
    $part_2_size_sectors" | bc`
part_2_begin_sector_alignment=`echo "$part_2_begin_sector / \
    $DEV_ALIGNMENT * $DEV_ALIGNMENT" | bc`

# check the begin sector
t=`echo "$part_2_begin_sector_alignment <= $DEV_ALIGNMENT" | bc`
if [ $? -eq 1 ] ; then
    tcerr "insufficient storage for reserved system partition (partition 2)"
    tcinfo "storage(unit: MiB): ${DEV_MIBS}"
    tcinfo "user partition storage(unit: MiB): ${PART_2_SIZE_MIB}"
    tcinfo "partition 2 starts at (unit: sector):$part_2_begin_sector_alignment"
    tcexit $ERR_SIZE
fi

part_2_size_sectors_alignment=`echo "$part_2_end_sector_alignment - \
    $part_2_begin_sector_alignment" | bc`
part_2_size_mib_alignment=`echo "$part_2_size_sectors_alignment * \
    $BYTES_PER_SECTOR / 2^20" | bc`
####################### Partition 2 Calc (end) #################################




###################### Partition 1 Calc (begin) ################################
# 1. starts at 2048 sector
# 2. ends at the last sector of the cylinder previously adjacent to the cylinder
#    covering the first sector of partition 2
DEV_SECTORS_SEC_REMAINING=$part_2_begin_sector_alignment
# calculate the end sector
part_1_end_sector=`echo "$part_2_begin_sector_alignment - 0" | bc`
# cylinder alignment 
part_1_end_sector_alignment=`echo "($part_1_end_sector + 0) / \
    ${SECTORS_PER_CYLINDER} * $SECTORS_PER_CYLINDER - 0" | bc`

part_1_begin_sector_alignment=$DEV_ALIGNMENT

part_1_size_sectors_alignment=`echo "$part_1_end_sector_alignment - \
    $part_1_begin_sector_alignment + 0" | bc`
part_1_size_mib_alignment=`echo "$part_1_size_sectors_alignment * \
    $BYTES_PER_SECTOR / 2^20" | bc`

# check
t=`echo "$part_1_begin_sector_alignment >= $part_1_end_sector_alignment" | bc`
if [ $t -eq 1 ] ; then
    tcerr "insufficient storage for system partition (partition 1)"
    tcinfo "storage(unit: MiB): ${DEV_MIBS}"
    tcinfo "user partition storage(unit: MiB): ${PART_1_SIZE_MIB}"
    tcexit $ERR_SIZE
fi

t=`echo "$part_1_size_sectors_alignment <= 0" | bc`
if [ $t -eq 1 ] ; then
    tcerr "insufficient storage for system partition (partition 1)"
    tcinfo "storage(unit: MiB): ${DEV_MIBS}"
    tcinfo "user partition storage(unit: MiB): ${PART_1_SIZE_MIB}"
    tcexit $ERR_SIZE
fi

####################### Partition 1 Calc (end) #################################

# log 
tcinfo "part 1"
tcinfo "begin                  : $part_1_begin_sector_alignment"
tcinfo "end                    : $part_1_end_sector_alignment"
tcinfo "storage (unit: sectors): $part_1_size_sectors_alignment"
tcinfo "storage (unit: MiB)    : $part_1_size_mib_alignment"
if ! [ $D_FLAG -eq 0 ] ; then
    tcinfo "part 2"
    tcinfo "begin                  : $part_2_begin_sector_alignment"
    tcinfo "end                    : $part_2_end_sector_alignment"
    tcinfo "storage (unit: sectors): $part_2_size_sectors_alignment"
    tcinfo "storage (unit: MiB)    : $part_2_size_mib_alignment"
fi
tcinfo "part 4"
tcinfo "begin                  : $part_4_begin_sector_alignment"
tcinfo "end                    : $part_4_end_sector_alignment"
tcinfo "storage (unit: sectors): $part_4_size_sectors_alignment"
tcinfo "storage (unit: MiB)    : $part_4_size_mib_alignment"


###################### umount partitions (begin) ###############################
if df | grep -q "$RSVDDIR" ; then
    echo "umount $RSVDDIR ..."
    umount $RSVDDIR
    if [ ! $? -eq 0 ] ; then
        tcerr "umount $RSVDDIR"
        tcexit $ERR_UMOUNT_SYSPART
    fi
fi

if df | grep -q "$DISKDIR" ; then
    echo "umount $DISKDIR..."
    umount $DISKDIR
    if [ ! $? -eq 0 ] ; then
        tcerr "umount $DISKDIR"
        tcexit $ERR_UMOUNT_DSKPART
    fi
fi
####################### umount partitions (end) ################################



####################### write tc loader (begin) ################################
tcprn "new --- tagging the hard disk, updating MBR (+63 sectors) ..."
echo "TC" > /tmp/hb.sig
dd if=/tmp/hb.sig of=$DEV_NODE bs=512 seek=63
if [ ! $? -eq 0 ] ; then
    tcerr "tagging the disk"
    tcexit $ERR_TAG_MBR
fi

dd if=/usr/sbin/tc/MBRCode.bin of=$DEV_NODE bs=512 seek=64
if [ ! $? -eq 0 ] ; then
    tcerr "update mbr"
    tcexit $ERR_HOOK_MBR
fi

dd if=/usr/sbin/tc/BootLoad.bin of=$DEV_NODE bs=512 seek=65
if [ ! $? -eq 0 ] ; then
    tcerr "update boot loader"
    tcexit $ERR_HOOK_MBR
fi

######################## write tc loader (end) #################################

#################### write disk identifier (begin) #############################
#
# write a disk identifier
#
tcprn "initialize disk identifier"
seconds=`date +%s`
magicnumber=`echo "obase=16; $seconds" | bc`
magiclen=${#magicnumber}

tcinfo "magicnumber = 0x$magicnumber"
tcinfo "magiclen = $magiclen"

startoffset=$(($magiclen - 2))
byte0=${magicnumber:$startoffset:2}
startoffset=$(($magiclen - 4))
byte1=${magicnumber:$startoffset:2}
startoffset=$(($magiclen - 6))
byte2=${magicnumber:$startoffset:2}
startoffset=$(($magiclen - 8))
byte3=${magicnumber:$startoffset:2}

echo byte0 = $byte0
echo byte1 = $byte1
echo byte2 = $byte2
echo byte3 = $byte3

echo -ne "\x$byte0\x$byte1\x$byte2\x$byte3" | dd of=$DEV_NODE bs=1 count=4 seek=440
##################### write disk identifier (end) ##############################



######################## Disk partition (begin) ################################
sleep 1

tcprn "creating partitions..."

if [ $D_FLAG -eq 0 ] ; then
fdisk -u $DEV_NODE <<EOT
n
p
1
$part_1_begin_sector_alignment
$(($part_1_end_sector_alignment - 1))
n
p
4
$part_4_begin_sector_alignment
$(($part_4_end_sector_alignment - 1))
t
1
12
t
4
12
a
4
w
EOT
else
fdisk -u $DEV_NODE <<EOT
n
p
1
$part_1_begin_sector_alignment
$(($part_1_end_sector_alignment - 1))
n
p
2
$part_2_begin_sector_alignment
$(($part_2_end_sector_alignment - 1))
n
p
4
$part_4_begin_sector_alignment
$(($part_4_end_sector_alignment - 1))
t
1
12
t
2
7
t
4
12
a
4
w
EOT
fi

if [ ! $? -eq 0 ] ; then
    tcerr "creating partitions"
    tcexit $ERR_FDISK
fi

sync
######################### Disk partition (end) #################################



######################## Make filesystem (begin) ###############################
# it seems `sleep 1' is not enough
# sleep 1
# at most delay 30 seconds for detecting dev node
NODE_CHK_DELAY=30

tcprn "creating ext3 filesystem for partition 1..."
part_1_node=${DEV_NODE}1
PART_1_LABEL="SYSTEM"

tick=$NODE_CHK_DELAY
while [ ! -b $part_1_node -a $tick -gt 0 ] ; do
    tcinfo "checking dev node: $part_1_node, time left: $tick"
    tick=$(($tick - 1))
    sleep 1
done
if [ -b $part_1_node ] ; then
    mkfs.ext3 -F -L $PART_1_LABEL -b 4096 $part_1_node
    if [ ! $? -eq 0 ] ; then
        tcerr "creating ext3 filesystem for partition 1"
        tcexit $ERR_FORMAT_DSKPART
    fi
else
    tcerr "fail to find dev node $part_1_node"
    tcexit $ERR_DSKPART_NOTEXIST
fi

sleep 1
if ! [ $D_FLAG -eq 0 ] ; then
    tcprn "creating ntfs filesystem for partition 2..."
    part_2_node=${DEV_NODE}2
    PART_2_LABEL="USER"
    mkfs.ntfs --fast --force --label $PART_2_LABEL $part_2_node
    if [ ! $? -eq 0 ] ; then
        tcerr "creating ntfs filesystem for partition 2"
        tcexit $ERR_FORMAT_USRPART
    fi
fi

sleep 1
tcprn "creating ext3 filesystem for partition 4..."
part_4_node=${DEV_NODE}4
PART_4_LABEL="RESERVED"
mkfs.ext3 -F -L $PART_4_LABEL -b 4096 $part_4_node
if [ ! $? -eq 0 ] ; then
    tcerr "creating ext3 filesystem for partition 4"
    tcexit $ERR_FORMAT_SYSPART
fi

######################### Make filesystem (end) ################################



####################### install firmware (begin) ###############################
sleep 1

tcprn "mount the target partition ..."
#mount `findfs LABEL="RESERVED"` $RSVDDIR
mount /dev/sda4 $RSVDDIR
if [ ! $? -eq 0 ] ; then
    tcerr "mounting reserved partition"
    tcexit $ERR_MOUNT_SYSPART
fi

rm -rf $RSVDDIR/boot
if [ ! $? -eq 0 ] ; then
    tcerr "removing old firmware from reserved partition"
    tcexit $ERR_RM_BLFILES
fi

tcprn "install system firmware, this may take several minutes..."
cp -a /usr/share/boot $RSVDDIR
if [ ! $? -eq 0 ] ; then
    tcerr "copying new firmware"
    tcexit $ERR_CP_BLFILES
fi

tcprn "making firmwares..."
cd /
cat /usr/sbin/tc/file.lst | cpio -H newc -o > $INITRD_CPIO_FILE
if [ ! $? -eq 0 ] ; then
    tcerr "fail to generate .cpio file $INITRD_CPIO_FILE"
    tcexit $ERR_GEN_INITRD
fi

tcprn "compressing firmwares..."
gzip -c -9 -n $INITRD_CPIO_FILE > $INITRD_IMAGE_FILE
if [ ! $? -eq 0 ] ; then
    tcerr "fail to compress file $INITRD_CPIO_FILE"
    tcexit $ERR_COMP_INITRD
fi

tcprn "checking firmwares..."
if [ ! -f $INITRD_IMAGE_FILE ] ; then
    tcerr "$INITRD_IMAGE_FILE : no such file"
    tcexit $ERR_INITRD_MISSING
fi

sync

tcprn "umount partition ..."
umount $RSVDDIR
if [ ! $? -eq 0 ] ; then
    tcerr "umount reserved partition"
    tcexit $ERR_UMOUNT_SYSPART
fi
######################## install firmware (end) ################################



######################## install grub (begin) ##################################
#tcprn "display the mounted partitions..."
#df

sleep 1

tcprn "install grub ..."
#grub --batch <<EOT
grub --batch <<EOT | grep 'Running.*succeeded'
root (hd0,3)
setup (hd0)
quit
EOT

if [ ! $? -eq 0 ] ; then
    tcerr "failed to install grub."
    tcexit $ERR_INSTALL_BL
fi

######################### install grub (end) ###################################

################## create virtual hard drive (begin) ###########################
mount | grep -q $part_1_node
if [ ! $? -eq 0 ] ; then
    mount $part_1_node $DISKDIR
    if [ ! $? -eq 0 ] ; then
        tcerr "fail to mount $part_1_node"
        tcexit $ERR_MOUNT_DSKPART
    fi
fi

if [ $VD_FLAG -eq 1 ] ; then
    tcprn "create virtual device..."
    t=`sh /tcdscript/_initvd.sh $VDSZ`
    if [ ! $t -eq 0 ] ; then
        tcerr "fail to create virtual device: $t"
        tcexit $ERR_GEN_VD
    else
        tcinfo "create virtual device successfully"
    fi
fi

umount $part_1_node
if [ ! $? -eq 0 ] ; then
    tcerr "fail to umount $part_1_node"
    tcexit $ERR_UMOUNT_DSKPART
fi
################### create virtual hard drive (end) ############################

sh /tcdscript/post-inithd.sh
ret=$?
if [ $ret -ne 0 ] ; then
    tcerr "fail to mount partitions"
    tcexit $ret
fi
# sleep after mount, login UI will not prompt err msg
sleep 2

tcprn "initialized disk successfully"

# notification
FILE_DISKRDY=/tmp/f-diskrdy
echo -n $ERR_NORMAL > $FILE_DISKRDY

tcexit $ERR_NORMAL

