#!/bin/sh
#
# update firmware and reboot automatically
# @return 0, on success
# @return 1, on no updatable firmware are found
# @return n (n>1), on other circumstances
#
source /usr/sbin/tc/global_def.sh

# define the disk to be initialized
DEV_NODE='/dev/sda'

# define MOUNTING POINT
DISKDIR='/mnt/disk'
USBDIR='/mnt/usb'
RSVDDIR='/mnt/rsvd'
UPFW_LOG_FILE="/var/log/updatefirmware.log"
tccount=0

# define IPv4 address for TC Server
IP_FILE="${RSVDDIR}/tc-serv.addr"
if [ ! -f $IP_FILE ] ; then
    IP_ADDR='192.168.0.1'
else
    IP_ADDR=`cat $IP_FILE`
fi

tcmsg_init()
{
    echo -n "" > $UPFW_LOG_FILE
    exec 7>&1 8>&2
    exec >> $UPFW_LOG_FILE 2>&1
}

# no color
tcprn()
{
    tccount=$(($tccount + 1))
    tccount_ex=`echo "$tccount" | awk '{printf "%02d", $1}'`
    # print on the screen
    #echo -e "$@" >&7
    echo "[$tccount_ex] $@" >&7
    # print to log file
    echo -e "\n[$tccount_ex] $@"
}

# print "[Info] " before every piece of information
tcinfo(){
    echo -e "\033[32m[INFO]\033[0m $@" >&7
    echo "[INFO] $@"
}

tcerr(){
    echo -e "\033[31m[ERROR]\033[0m $@" >&8
    echo "[ERROR] $@"
}

tcwarn(){
    echo -e "\033[35m[WARNING]\033[0m $@" >&8
    echo "[WARNING] $@"
}

tcexit(){
    if [ "x$1" == "x" -o "x$1" == "x0" ] ; then
        statusCode=${ACT_UPDATING_FIRMWARE_SUCCESS}
    else
        statusCode=${ACT_UPDATING_FIRMWARE_FAILURE}
    fi

    exitCode=$1

    echo "${statusCode}: :" > $STATUS_CONF_FILE

    tcinfo "save logs to file $UPFW_LOG_FILE"
    exec 1>&7 2>&8
    exec 7>&- 8>&-
    exit $exitCode
}

report_client_status(){
    echo "$1 : :" > $STATUS_CONF_FILE
}


######################## Set log file (begin) ##################################
tcmsg_init
tcinfo "set log to file $UPFW_LOG_FILE"
######################### Set log file (end) ###################################


tcinfo "server ip: $IP_ADDR"

################################### Micros def (begin) ##########################
# define error code exit normally
ERR_NORMAL=0
# exit if no available fw
ERR_NO_FW=1
# exit if illegal parameter
ERR_PARAMETER=2
# exit if given parameter `firstPartSize` is improper
ERR_SIZE=3
# exit when a command fail
ERR_RUNTIME=4
# exit when user interrupted
ERR_INTERRUPT=5
# exit if no hard drive
ERR_NOHD=6
# pre-upfw.sh exited with a non-zero value
ERR_PRE_SCRIPT=7
# no reserved partition
ERR_NO_FS_RSVD=8
# umount diskdev
ERR_UMOUNT_DISKDIR=9
# umount reserved dir
ERR_UMOUNT_RSVDDIR=10
# mount reserved dir
ERR_MOUNT_RSVDDIR=11
# no /tmp
ERR_NO_TMPFS=12
# no md5sum was provided
ERR_FW_ILLEGAL_MD5SUM=13
# fail to unzip fw
ERR_FW_UNZIP=14
# no kernel in fw
ERR_FW_NO_KERNEL=15
# fail to umount rsvd after install fw files
ERR_UMOUNT_RSVDDIR_2=16
# fail to update mbr
ERR_INST_MBR=17
# fail to update boot loader
ERR_INST_BOOTLOADER=18
# fail to update grub
ERR_INST_GRUB=19
# post-upfw.sh exited with a non-zero value
ERR_POST_SCRIPT=20

#################################### Micros def (end) ###########################

# tell tc server the client is about to udpate its firmware
report_client_status ${ACT_UPDATING_FIRMWARE}

#if [ $# -lt 2 ] ; then
#    tcerr "illegal parameter"
#    echo "Usage: updatefirmware.sh <device> <partition>"
#    exit $ERR_PARAMETER
#fi
if [ $# -ne 0 ] ; then
    tcerr "illegal parameters"
    tcexit $ERR_PARAMETER
fi

if [ ! -b "$DEV_NODE" ] ; then
    tcerr "$DEV_NODE: no such block node"
    tcexit $ERR_NOHD
fi

# sleep before umount, login UI will not prompt err msg
sleep 2
sh /tcdscript/pre-updatefirmware.sh
retval=$?
if ! [ $retval -eq 0 ] ; then
    tcerr "failed to run prerequsite script"
    tcexit $ERR_PRE_SCRIPT
fi

tcprn "search partition with label \"RESERVED\"..."
rsvdpart=`findfs LABEL="RESERVED"`
if [ -z "$rsvdpart" ] ; then
    tcerr "no partition found with label \"RESERVED\""
    tcexit $ERR_NO_FS_RSVD
else
    if [ ! -b "$rsvdpart" ] ; then
        tcerr "$rsvdpart: no such partition"
        tcexit $ERR_NO_FS_RSVD
    else
        tcinfo "$rsvdpart is with label \"RESERVED\""
    fi
fi

if df | grep -q "$DISKDIR" ; then
    tcinfo "umount the mountpoint $DISKDIR ..."
    umount $DISKDIR
    if [ $? -ne 0 ] ; then
        tcerr "fail to umount $DISKDIR"
        tcexit $ERR_UMOUNT_DISKDIR
    fi
fi

if df | grep -q "$RSVDDIR" ; then
    tcinfo "umount the mountpoint $RSVDDIR ..."
    umount $RSVDDIR
    if [ $? -ne 0 ] ; then
        tcerr "fail to umount $RSVDDIR"
        tcexit $ERR_UMOUNT_RSVDDIR
    fi
fi

tcprn "mount $rsvdpart to the mountpoint $RSVDDIR ..."
mount $rsvdpart $RSVDDIR
if [ $? -ne 0 ] ; then
    tcerr "fail to mount partition with label \"RESERVED\""
    tcexit $ERR_MOUNT_RSVDDIR
fi
tcinfo "mount successfully"

TC_KERNEL="$RSVDDIR/boot/bzImage"
TC_INITRD="$RSVDDIR/boot/initrd.img"
FW_URL="http://$IP_ADDR/firmwares/rootfs.cpio.gz"
FW_MD5_URL="http://$IP_ADDR/firmwares/rootfs.cpio.gz.md5"

# check if /tmp tmpfs exists
if ! df | grep -q "^tmpfs" ; then
    tcerr "tmpfs: mount point was not found"
    tcexit $ERR_NO_TMPFS
else
    LOCAL_DIR=`df | grep "^tmpfs" | awk '{print $NF}'`
fi

FW=$LOCAL_DIR/${FW_URL##*/}
FW_MD5=$LOCAL_DIR/${FW_MD5_URL##*/}

if [ ! -f "$TC_KERNEL" ] ; then
    tcwarn "$TC_KERNEL : no such file"
fi

if [ ! -f "$TC_INITRD" ] ; then
    tcwarn "$TC_INITRD : no such file"
    TC_INITRD_MD5=
else
    TC_INITRD_MD5=`md5sum $TC_INITRD | awk '{print $1}'`
fi

# "update firmware..."
rm -f $FW_MD5
wget -c --tries=10 --timeout=3 --retry-connrefused --output-document $FW_MD5 $FW_MD5_URL
ret_dw_fw_md5=$?
if [ $ret_dw_fw_md5 -ne 0 ] ; then
    tcwarn "fail to download $FW_MD5_URL"
fi

rm -f $FW
wget -c --tries=10 --timeout=3 --retry-connrefused --output-document $FW $FW_URL
ret_dw_fw=$?
if [ $ret_dw_fw -ne 0 ] ; then
    tcerr "fail to download from $FW_URL"
    # exit if firmware was not found
    tcerr "no firmware was found on server: $IP_ADDR"
    tcexit $ERR_NO_FW
fi

# verify md5 checksum
if [ $ret_dw_fw_md5 -eq 0 ] ; then
    tcprn "check md5 checksum..."
    ORI_MD5SUM=`cat $FW_MD5`
    LOCAL_MD5SUM=`md5sum $FW | awk '{print $1}'`

    if [ -n "$ORI_MD5SUM" ] && [ "$ORI_MD5SUM" == "$LOCAL_MD5SUM" ] ; then
        tcinfo "veryfy md5sum successfully"
    elif [ -z "$ORI_MD5SUM" ] ; then
        tcwarn "no valid md5sum were found in file $FW_MD5_URL"
    else
        tcerr "fail to verify md5sum, the client will NOT be updated"
        tcexit $ERR_FW_ILLEGAL_MD5SUM
    fi
else
    tcwarn "ignore veryfying md5 checksum due to file missing: $FW_MD5_URL"
fi

# unziped and update firmware
tcprn "update firmware..."
# update initrd
rm -f $TC_INITRD
mv $FW $TC_INITRD

cd $LOCAL_DIR
rm -rf folder
mkdir -p folder
cd folder
gzip -dc $TC_INITRD | cpio -idm
if [ ! $? -eq 0 ] ; then
    tcerr "failed to unzip the firmware"
    tcexit $ERR_FW_UNZIP
fi

rm -f $TC_KERNEL
mv usr/share/boot/bzImage $TC_KERNEL
if [ ! $? -eq 0 ] ; then
    tcerr "failed to find kernel file from firmware"
    tcexit $ERR_FW_NO_KERNEL
fi

sync
tcinfo "upate firmware successfully"

if df | grep -q "$RSVDDIR" ; then
    tcinfo "umount the mountpoint \"$RSVDDIR\"..."
    umount $RSVDDIR
    if [ $? -ne 0 ] ; then
        tcerr "fail to umount $RSVDDIR"
        tcexit $ERR_UMOUNT_RSVDDIR_2
    fi
fi

#echo "clear disk signature ..."
#dd if=/dev/zero of=$DEV_NODE bs=1 count=4 seek=440           1>/dev/null 2>&1

tcprn "new --- tagging the hard disk..."
tcinfo "update MBR (+63 sectors)..."
echo "TC" > /tmp/hb.sig
dd if=/tmp/hb.sig of=$DEV_NODE bs=512 seek=63
dd if=$LOCAL_DIR/folder/usr/sbin/tc/MBRCode.bin of=$DEV_NODE bs=512 seek=64
if [ ! $? -eq 0 ] ; then
    tcerr "failed to update mbrcode.bin"
    tcexit $ERR_INST_MBR
fi

dd if=$LOCAL_DIR/folder/usr/sbin/tc/BootLoad.bin of=$DEV_NODE bs=512 seek=65
if [ ! $? -eq 0 ] ; then
    tcerr "failed to update bootload.bin"
    tcexit $ERR_INST_BOOTLOADER
fi

grub --batch <<EOT | grep 'Running.*succeeded'
root (hd0,3)
setup (hd0)
quit
EOT

if [ ! $? -eq 0 ] ; then
    tcerr "failed to install grub."
    tcexit $ERR_INST_GRUB
fi

sh /tcdscript/post-updatefirmware.sh
retval=$?
if ! [ $retval -eq 0 ] ; then
    tcerr "failed to run post script"
    tcexit $ERR_POST_SCRIPT
fi
# sleep after mount, login UI will not prompt err msg
sleep 2

tcinfo "updatefirmware.sh successfully ..."

# tell the server that firmware was successfully updated.
# report_client_status ${ACT_UPDATING_FIRMWARE_SUCCESS}
report_client_status ${ACT_UPDATING_FIRMWARE_SUCCESS}

# reboot the client
sleep $HB_INTERVAL
/usr/sbin/tc/tc-shutdown.sh --reboot

tcexit $ERR_NORMAL

