#!/bin/sh
# get-client-entry.sh 
# 
# query if a client has already been registered into the TC system
# and get machine name and auto login accout into conf file accordingly.
# account into file $AUTO_LOGIN_FILE
# machine name into file $MACHINE_NAME_FILE 
# Ret Value:
# 0 if operation success
# value greater than 0 on other conditions.
# 2 if unregistered, and
#   if auto registration is open, output the desired client name to $TC_NEW_CLIENT_NAME_FILE;
#   otherwise delete the file $TC_NEW_CLIENT_NAME_FILE
#
ERR_SUCCESS=0
ERR_QC_UNREGISTERED=2
ERR_OTHER=11
ERR_PARA=12
ERR_CURL=13
ERR_HWADDR=14

SCRIPT_NAME=$0
. /usr/sbin/tc/global_def.sh

SCRIPT_GET_HWADDR=$TCDIR/get-hwaddr.sh
NIF=$DEFAULT_IFACE
if [ -f $IFACE_FILE ] ; then
  NIF=`cat $IFACE_FILE`
fi
SEPARATOR="-"

if [ $# -eq 0 ] ; then
    serverip=${SERVER}
    if ! [ -b ${DISKRSVD} ] ; then
        echo "${DISKRSVD}: no such block node"
        exit $ERR_OTHER
    fi

    mount | grep -q -E "${DISKRSVD}.*${RSVDDIR}"
    if [ $? -eq 0 ] ; then
        mounted=YES
        mount_action=NO
    else
        mount_action=YES

        TRY_MAX=5
        try_count=0
        until [ $try_count -ge $TRY_MAX ] || mount | grep -q -E "${DISKRSVD}.*${RSVDDIR}" ; do
            mount ${DISKRSVD} ${RSVDDIR}
            try_count=$(($try_count + 1))
            sleep 1
        done

        mount | grep -q -E "${DISKRSVD}.*${RSVDDIR}"
        if [ $? -eq 0 ] ; then
            mounted=YES
        else
            mounted=NO
        fi
    fi

    if [ "x${mounted}" == "xYES" -a -f ${SERVADDR_FILE} ] ; then
        serverip=`cat ${SERVADDR_FILE}`
    fi

    if [ "x${mounted}" == "xYES" -a "x${mount_action}" == "xYES" ] ; then
        umount ${RSVDDIR}
    fi
elif [ $# -eq 1 ] ; then
    serverip=$1
else
    cat <<EOT
SYNOPSIS
    $SCRIPT_NAME [ServerIP]

DESCRIPTION
    Query if a client is in DB resided on server <ServerIP> and save its machine name and autlogin
    account into file $MACHINE_NAME_FILE and file $AUTO_LOGIN_FILE

EXAMPLE
    $SCRIPT_NAME 192.168.0.1
EOT
    exit $ERR_PARA
fi

sh $SCRIPT_GET_HWADDR $NIF $SEPARATOR >/dev/null 2>&1
if [ $? -ne 0 ] ; then
    exit $ERR_HWADDR
fi

HWADDR=`sh $SCRIPT_GET_HWADDR $NIF $SEPARATOR`
URL_QC=$HTTP_PROT://$DOMAINNAME/tc/rest/client.php
JSON_QC_TMP=/tmp/query-client.json
URL_DST=

if [ -f $JSON_QC_TMP ] ; then
    rm -f $JSON_QC_TMP
fi

curl $CURL_OPTS -s -X POST --retry 3 --retry-delay 2 --output $JSON_QC_TMP --data "{\"mac\":\"${HWADDR}\"}" --header "Content-Type: application/json" $URL_QC

if [ $? -eq 0 ] ; then
    if [ -f $JSON_QC_TMP ] ; then
        t=`jq -r '.success' $JSON_QC_TMP`
        if ! [ $? -eq 0 ] ; then
            exit $ERR_OTHER
        fi

        if [ "x$t" == "xtrue" ] ; then
            # for registered client
            # save machine name and auto login name to local file
            t=`jq -r '.result.name' $JSON_QC_TMP`
            if [ $? -eq 0 ] ; then
                echo "$t" > $MACHINE_NAME_FILE
            else
                exit $ERR_OTHER
            fi

            #
            # screen resolution
            #
            t1=`jq -r '.result.resolution' $JSON_QC_TMP`
            if [ $? -eq 0 ] ; then
                echo "$t1" > $TC_RESOLUTION_FILE
            else
                echo "error: screen resolution"
                exit $ERR_OTHER
            fi

            #
            # USB storage device control
            #
            t1=`jq -r '.result.usb_storage' $JSON_QC_TMP`
            if [ $? -eq 0 ] ; then
                echo "$t1" > $TC_USBSTORAGE_FILE
            else
                echo "error: usb storage"
                exit $ERR_OTHER
            fi

            #
            # network bandwith settings
            #
            t1=`jq -r '.result.download_KBS' $JSON_QC_TMP`
            if [ $? -ne 0 ] ; then
                exit $ERR_OTHER
            fi
            t2=`jq -r '.result.upload_KBS' $JSON_QC_TMP`
            if [ $? -ne 0 ] ; then
                exit $ERR_OTHER
            fi
            $TCDIR/tc-netbw-trigger.sh $t1 $t2

            t1=`jq -r '.result.autoboot_user' $JSON_QC_TMP`
            if [ $? -ne 0 ] ; then
                exit $ERR_OTHER
            fi
            t2=`jq -r '.result.auto_login_delay' $JSON_QC_TMP`
            if [ $? -ne 0 ] ; then
                exit $ERR_OTHER
            fi
            t3=`jq -r '.result.autoboot_ciid' $JSON_QC_TMP`
            if [ $? -ne 0 ] ; then
                exit $ERR_OTHER
            fi
            t4=`jq -r '.result.autoboot_ciid_global' $JSON_QC_TMP`
            if [ $? -ne 0 ] ; then
                exit $ERR_OTHER
            fi
            t5=`jq -r '.result.autoboot_image' $JSON_QC_TMP`
            if [ $? -ne 0 ] ; then
                exit $ERR_OTHER
            fi
            t6=`jq -r '.result.autoboot_image_global' $JSON_QC_TMP`
            if [ $? -ne 0 ] ; then
                exit $ERR_OTHER
            fi
            echo "$t1:$t2:$t3:$t4:$t5:$t6" > $AUTO_LOGIN_FILE

            #
            # set flag: registered
            #
            gIsClientRegistered=true
        elif [ "x$t" == "xfalse" ] ; then
            #
            # set flag: unregistered
            #
            gIsClientRegistered=false
        else
            # for non-available registeration service
            exit $ERR_OTHER
        fi
    else
        exit $ERR_OTHER
    fi
else
    exit $ERR_CURL
fi

#
# if client is registered, exit
# otherwise, check if auto naming is enabled and to get auto name from server
#
if [ "x$gIsClientRegistered" == "xtrue" ] ; then
  exit $ERR_SUCCESS
fi

#
# if unregistered, check if auto registration is enabled
#
URL_SETTINGS=$HTTP_PROT://$DOMAINNAME/tc/rest/system.php/settings
JSON_SETTINGS=/tmp/client-settings.json
if [ -f $JSON_SETTINGS ] ; then
    rm -f $JSON_SETTINGS
fi
curl $CURL_OPTS -s -X GET --retry 3 --retry-delay 2 --output $JSON_SETTINGS $URL_SETTINGS
if [ $? -eq 0 ] ; then
  if [ -f $JSON_SETTINGS ] ; then
    t=`jq -r '.result.client_naming' $JSON_SETTINGS`
    if ! [ $? -eq 0 ] ; then
       exit $ERR_OTHER
    fi

    if [ "x$t" == "xtrue" -o "x$t" == "x1" ] ; then
      #
      # client auto naming is enabled
      #
      gClientAutoNaming=true
    else
      #
      # client auto naming is not enabled
      #
      gClientAutoNaming=false
    fi
  else
    exit $ERR_OTHER
  fi
else
  exit $ERR_CURL
fi

#
# if auto naming is not enabled, exit with code indicating client is unregistered
# otherwise, try to get the auto name from server and write to a temporary file
#
if [ "x$gClientAutoNaming" == "xfalse" ] ; then
  exit $ERR_QC_UNREGISTERED
fi

if [ -f $TC_NEW_CLIENT_NAME_FILE ] ; then
  exit $ERR_QC_UNREGISTERED
fi

URL_CLIENT_NAMING=$HTTP_PROT://$DOMAINNAME/tc/rest/system.php/auto_client_naming
JSON_CLIENT_NAMING=/tmp/client-naming.json
if [ -f $JSON_CLIENT_NAMING ] ; then
    rm -f $JSON_CLIENT_NAMING
fi
curl $CURL_OPTS -s -X GET --retry 3 --retry-delay 2 --output $JSON_CLIENT_NAMING $URL_CLIENT_NAMING
if [ $? -eq 0 ] ; then
  if [ -f $JSON_CLIENT_NAMING ] ; then
    t=`jq -r '.success' $JSON_CLIENT_NAMING`
    if ! [ $? -eq 0 ] ; then
      exit $ERR_OTHER
    fi

    #
    # seems we get an available client name
    #
    if [ "x$t" == "xtrue" ] ; then
      t=`jq -r '.result' $JSON_CLIENT_NAMING`
      if [ $? -eq 0 ] ; then
        #
        # write the name to a file
        #
        echo "$t" > $TC_NEW_CLIENT_NAME_FILE 
      else
        exit $ERR_OTHER
      fi
    #
    # client naming service might not be available
    #
    elif [ "x$t" == "xfalse" ] ; then
      t=`jq -r '.result.error' $JSON_CLIENT_NAMING`
      #
      # the option is off on server
      #
      if [ "x$t" == "xe_operation_denied" ] ; then
        exit $ERR_OTHER
      fi

      #
      # cannot calculate a new name,
      # for example, for the next name after 99 if the digit width is 2
      #
      if [ "x$t" == "xe_client_naming_max" ] ; then
        exit $ERR_OTHER
      fi

      #
      # invalid regex rule or other unexpected errors
      #
      if [ "x$t" == "xe_client_naming" ] ; then
        exit $ERR_OTHER
      fi
    #
    # error handling
    #
    else
      exit $ERR_OTHER
    fi

  else
    exit $ERR_OTHER
  fi
else
  exit $ERR_CURL
fi

#
# exit with err code: unregistered
#
exit $ERR_QC_UNREGISTERED

