#!/bin/sh
# reg-client.sh
# 
# reg a client into the TC system.
#
# Ret Value:
# 0 on registration success
# 3 on duplicate MAC address
# 4 on duplicate client name
# 5 illegal characters in client name
# value greater than 5 on other conditions.
ERR_REG_SUCCESS=0
ERR_SERVICE_OFF=1
ERR_REG_DUPLICATE_HWADDR=3
ERR_REG_DUPLICATE_NAME=4
ERR_REG_ILLEGAL_NAME=5
ERR_OTHER=11
ERR_PARA=12
ERR_CURL=13
ERR_HWADDR=14

SCRIPT_NAME=$0
. /usr/sbin/tc/global_def.sh

SCRIPT_GET_HWADDR=$TCDIR/get-hwaddr.sh
NIF=$DEFAULT_IFACE
if [ -f $IFACE_FILE ] ; then
  NIF=`cat $IFACE_FILE`
fi
SEPARATOR="-"

if [ $# -eq 1 ] ; then
    clt_name=$1
    if [ -f ${SERVADDR_FILE} ] ; then
        serverip=`cat ${SERVADDR_FILE}`
    else
        serverip=${SERVER}
    fi
elif [ $# -eq 2 ] ; then
    clt_name=$1
    serverip=$2
else
    cat <<EOT
SYNOPSIS
    $SCRIPT_NAME <CLT_NAME> [ServerIP]

DESCRIPTION
    Register a client to DB resided on server <ServerIP>

EXAMPLE
    $SCRIPT_NAME client001 192.168.0.1
EOT
    exit $ERR_PARA
fi

# check if reg-service is ON
sh /usr/sbin/tc/query-reg-service.sh $serverip
if [ $? -ne 0 ] ; then
    echo "registration server is OFF"
    exit $ERR_SERVICE_OFF
fi

# cpu information
if ! [ -f $GET_CPU_SCRIPT ] ; then
    tcerr "$GET_CPU_SCRIPT : no such file"
    exit $ERR_OTHER
fi
str_cpu=`sh $GET_CPU_SCRIPT`
retval=$?
if ! [ $retval -eq 0 ] ; then
    tcerr "$GET_CPU_SCRIPT : run"
    exit $ERR_OTHER
fi


# client memory
#clt_memo_kib=`cat /proc/meminfo | grep ^MemTotal | awk '{print $2}'`
#clt_memo=$(($clt_memo_kib / 1024))
if ! [ -f $GET_MEMORY_SCRIPT ] ; then
    tcerr "$GET_MEMORY_SCRIPT : no such file"
    exit $ERR_OTHER
fi
str_memory=`sh $GET_MEMORY_SCRIPT`
retval=$?
if ! [ $retval -eq 0 ] ; then
    tcerr "$GET_MEMORY_SCRIPT : run"
    exit $ERR_OTHER
fi

# disk information
if ! [ -f $GET_DISK_SIZE_SCRIPT ] ; then
    tcerr "$GET_DISK_SIZE_SCRIPT : no such file"
    exit $ERR_OTHER
fi
str_disk=`sh $GET_DISK_SIZE_SCRIPT`
retval=$?
if ! [ $retval -eq 0 ] ; then
    tcerr "$GET_DISK_SIZE_SCRIPT : run"
    exit $ERR_OTHER
fi
# convert to GiB
# save 2 digit after the dot
str_disk=`echo "scale=2; $str_disk / 1024" | bc -l`
retval=$?
if ! [ $retval -eq 0 ] ; then
    tcerr "fail to run the bc ultility"
    exit $ERR_OTHER
fi



# hwaddr
sh $SCRIPT_GET_HWADDR $NIF $SEPARATOR >/dev/null 2>&1
if [ $? -ne 0 ] ; then
    exit $ERR_HWADDR
fi

# memo
str_memo='self registration'

HWADDR=`sh $SCRIPT_GET_HWADDR $NIF $SEPARATOR`
URL_REG=$HTTP_PROT://$DOMAINNAME/tc/rest/machine.php
JSON_REG_TMP=/tmp/reg-client.json
URL_DST=

if [ -f $JSON_REG_TMP ] ; then
    rm -f $JSON_REG_TMP
fi

# firmware information
str_fwver=`cat /usr/sbin/tc/VERSION`
retval=$?
if ! [ $retval -eq 0 ] ; then
    tcerr "fail to get firmware version information"
    exit $ERR_OTHER
fi

# JSON format
# {
#     "mac":"00-00-00-00-00-00",
#     "cpu_model":"Intel(R) Xeon(R) CPU E5-2650 0 @ 2.00GHz",
#     "memory_size":"32844",
#     "disk_size":"58.00",
#     "name":"Dell_XPS_01",
#     "firmware_version":"1902.8928",
#     "memo":"a memorandam"
# }
curl $CURL_OPTS -s -X POST --retry 3 --retry-delay 2 --output $JSON_REG_TMP \
 --header "Content-Type: application/json" $URL_REG \
--data "{\"mac\":\"${HWADDR}\",\"cpu_model\":\"${str_cpu}\",\"memory_size\":\"${str_memory}\",\"disk_size\":\"${str_disk}\",\"name\":\"${clt_name}\",\"firmware_version\":\"${str_fwver}\",\"memo\":\"${str_memo}\"}"

if [ $? -eq 0 ] ; then
    if [ -f $JSON_REG_TMP ] ; then
        t=`jq -r '.success' $JSON_REG_TMP`
        if ! [ $? -eq 0 ] ; then
            exit $ERR_OTHER
        fi

        if [ "x$t" == "xtrue" ] ; then
            # success registration
            exit $ERR_REG_SUCCESS
        elif [ "x$t" == "xfalse" ] ; then
            # registeration failure
            r=`jq -r '.result.error' $JSON_REG_TMP`
            if ! [ $? -eq 0 ] ; then
                exit $ERR_OTHER
            fi
            echo "reg err: $r"
            if [ "x$r" == "xe_duplicate_value" ] ; then
                extra=`jq -r '.result.extra' $JSON_REG_TMP`
                echo "extra=$extra"
                echo $extra | grep $HWADDR
                if [ $? -eq 0 ] ; then
                    exit $ERR_REG_DUPLICATE_HWADDR
                else
                    exit $ERR_REG_DUPLICATE_NAME
                fi
            elif [ "x$r" == "xe_fail_register_user" ] ; then
                exit $ERR_REG_DUPLICATE_NAME
            elif [ "x$r" == "xe_bad_client_name" ] ; then
                exit $ERR_REG_ILLEGAL_NAME
            else
                exit $ERR_OTHER
            fi
        else
            # for non-available registeration service
            exit $ERR_OTHER
        fi
    else
        exit $ERR_OTHER
    fi
else
    exit $ERR_CURL
fi

