#!/bin/sh
# test-smb.sh
# 1) test if the smb service is working and 
#    try mount it onto filesystem
# 2) if the service has already been loaded,
#    unloaded it or just keep it as-is
#
SCRIPT_PATH=$0
SCRIPT_NAME=${SCRIPT_PATH##*/}
SERVER_IP=192.168.0.1
CLIENT_IP=192.168.0.100
CLIENT_MASK=255.255.255.0
NET_INTERFACE=eth0
SMB_AUTH=/etc/samba/smb-tc-auth.conf
SMB_SERVICE=//$SERVER_IP/tcimage
SMB_MOUNT_DIR=/mnt/tmp
TEST_RW_FILE=$SMB_MOUNT_DIR/helloworld.txt
TEST_LOG_FILE=/var/log/test-smb.log
###############################################################################
ERR_SUCCESS=0
ERR_FAILURE=1
###############################################################################

usage()
{
    cat <<EOT
SYNOPSIS
    $SCRIPT_PATH [ServerIP [ClientIP [ClientMask]]]

DESCRIPTION
    Test if the samba service resided in TC server with ServerIP is working.

EXAMPLE
    $SCRIPT_PATH 192.168.0.1
    $SCRIPT_PATH 192.168.0.1 192.168.0.100
    $SCRIPT_PATH 192.168.0.1 192.168.0.100 255.255.255.0
EOT
}
###############################################################################

tcmsg_init()
{
    currtime=$(date "+%F %T")
    echo -n "" > $TEST_LOG_FILE
    echo -n "$currtime" >> $TEST_LOG_FILE
    echo -e "\033[32m[INFO]\033[0m $currtime"

    exec 7>&1 8>&2
    exec >> $TEST_LOG_FILE 2>&1
}

# colorful
# print "[INFO] " before every piece of information
tcinfo()
{
    # print on the screen
    echo -e "\033[32m[INFO]\033[0m $@" >&7
    # print to log file
    echo "[INFO] $@"
}

# colorful
# print "[ERROR] " before every piece of information
tcerr()
{
    # print on the screen
    echo -e "\033[31m[ERROR]\033[0m $@" >&8
    # print to log file
    echo "[ERROR] $@" >&2
}

# colorful
# print "[WARNING] " before every piece of information
tcwarn()
{
    # print on the screen
    echo -e "\033[35m[WARNING]\033[0m $@" >&8
    # print to log file
    echo "[WARNING] $@" >&2
}

# no color
tcstep()
{
    gstep=$(($gstep + 1))
    gstepex=`echo "$gstep" | awk '{printf "%02d", $1}'`
    # print on the screen
    #echo -e "$@" >&7
    echo "[$gstepex] $@" >&7
    # print to log file
    echo -e "\n[$gstepex] $@"
}

tcmsg_finish()
{
    tcinfo "save logs to file $TEST_LOG_FILE"

    exec 1>&7 2>&8
    exec 7>&- 8>&-
}

tcexit()
{
    tcmsg_finish
    exit $1
}

# no color
tcprompt()
{
    gstep=$(($gstep + 1))
    gstepex=`echo "$gstep" | awk '{printf "%02d", $1}'`
    # print on the screen
    #echo -e "$@" >&7
    echo -e -n "[$gstepex] $@?[Y/n]" >&7
    # print to log file
    echo -e -n "[$gstepex] $@?[Y/n]"
    read -s -n 1 c
    echo -e -n "\n" >&7
    echo -e -n "\n"
    if [ "x$c" = "x" -o "x$c" = "xy" -o "x$c" = "xY" ] ; then
        return 0
    else 
        return 1
    fi
}

###############################################################################
## -h
## -help
## --help
case $1 in
    -h|-help|--help)
        usage
        exit $ERR_SUCCESS
        ;;
    *)
        ;;
esac

###############################################################################
gstep=0
tcmsg_init
tcstep "set log files to $TEST_LOG_FILE"

###############################################################################
tcstep "parse arguments"

if [ $# -ge 1 ] ; then
    server_ip=$1
fi

if [ $# -ge 2 ] ; then
    client_ip=$2
fi

if [ $# -ge 3 ] ; then
    client_mask=$3
fi

if [ "x$server_ip" == "x" ] ; then
    server_ip=$SERVER_IP
fi

if [ "x$client_ip" == "x" ] ; then
    client_ip=$CLIENT_IP
fi

if [ "x$client_mask" == "x" ] ; then
    client_mask=$CLIENT_MASK
fi

nic=$NET_INTERFACE

tcinfo "assume server_ip             $server_ip"
tcinfo "assume client_ip             $client_ip"
tcinfo "assume client_mask           $client_mask"
tcinfo "assume network interface     $nic"
###############################################################################

tcstep "check if $nic exist"
grep -q $nic /proc/net/dev
if [ $? -eq 0 ] ; then
    tcinfo "yes"
else
    tcerr "no"
    tcexit $ERR_FAILURE
fi

tcstep "bring up network interface $nic"
ifconfig $nic up
if [ $? -eq 0 ] ; then
    tcinfo "yes"
else
    tcerr "no"
    tcexit $ERR_FAILURE
fi

tcstep "check if $nic is alive"
ifconfig | grep -q "^$nic"
if [ $? -eq 0 ] ; then
    tcinfo "yes"
else
    tcerr "no"
    tcexit $ERR_FAILURE
fi

tcstep "try to set client ip with dhcp client"
udhcpc -t 5 -n -i $nic
if [ $? -eq 0 ] ; then
    tcinfo "yes"
else
    tcwarn "no"
fi

tcstep "detect client ip"
ifconfig $nic | grep -q 'inet addr:'
if [ $? -eq 0 ] ; then
    tcinfo "yes"
else
    tcwarn "no"
fi

tcstep "try to fetch client ip"
client_ip2=`ifconfig $nic | grep 'inet addr:' | sed 's/:/ /' |  awk '{print $3}'`
if [ $? -eq 0 ] ; then
    tcinfo "client ip: $client_ip2"
else
    tcwarn "no"
fi

if [ "x$client_ip2" == "x" ] ; then
    tcinfo "try set the static ip address"

    ifconfig $nic $client_ip netmask $client_mask
    if [ $? -eq 0 ] ; then
        tcinfo "client ip: $client_ip"
    else
        tcerr "no"
        tcexit $ERR_FAILURE
    fi
fi

tcstep "test the connection to $server_ip"
ping -c 2 -I $nic $server_ip
if [ $? -eq 0 ] ; then
    tcinfo "yes"
else
    tcerr "no"
    tcexit $ERR_FAILURE
fi

tcstep "test if smbtree works properly"
smbtree -A $SMB_AUTH
if [ $? -eq 0 ] ; then
    tcinfo "yes"
else
    tcerr "no"
    tcexit $ERR_FAILURE
fi


tcstep "test if smbclient work properly and list remote stuff"
smbclient $SMB_SERVICE -A $SMB_AUTH -c "ls"
if [ $? -eq 0 ] ; then
    tcinfo "yes"
else
    tcerr "no"
    tcexit $ERR_FAILURE
fi


tcstep "test if samba service had been loaded"
if df | grep -q "$SMB_SERVICE" ; then
    tcinfo "Service $SMB_SERVICE is alreadly loaded"

    tcprompt "umount $SMB_SERVICE"
    if [ $? -eq 0 ] ; then
        tmpmt=`mount | grep "$SMB_SERVICE" | awk '{print $3}'`
        umount -f $tmpmt
        if [ $? -eq 0 ] ; then
            tcinfo "yes"
            tcinfo "$SMB_SERVICE is \e[32munloaded\e[0m successfully"
            tcexit $ERR_SUCCESS
        else
            tcerr "no"
            tcerr "fail to unload $SMB_SERVICE"
            tcexit $ERR_FAILURE
        fi
    else
        tcinfo "samba service had been loaded and will be kept \e[32mas-is\e[0m"
        tcexit $ERR_SUCCESS
    fi
else
    tcinfo "no"
fi

tcstep "try load samba service onto local filesystem"
mount.cifs $SMB_SERVICE $SMB_MOUNT_DIR -o credentials=$SMB_AUTH
if [ $? -eq 0 ] ; then
    tcinfo "yes"
else
    tcerr "no"
    tcexit $ERR_FAILURE
fi

tcstep "test if samba service is loaded successfully"
if df | grep -q $SMB_SERVICE ; then
    tcinfo "yes"
else
    tcerr "no"
    tcexit $ERR_FAILURE
fi

tcstep "Test write permission"
echo "Hello, world!" >$TEST_RW_FILE
if [ $? -eq 0 ] ; then
    rdonly=no
    tcinfo "yes"
else
    rdonly=yes
    tcwarn "no"
fi

if [ "x$rdonly" == "xyes" ] ; then
    tcinfo "samba service is \e[32mloaded read-only\e[0m"
else
    tcinfo "samba service is \e[32mloaded\e[0m"
fi
tcexit $ERR_SUCCESS

###############################################################################

