/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <getopt.h>
#include "common.h"
#include "debug.h"

#define REQUIRED_NUMS	9

int find_sep(char *authStr, char sep)
{
	size_t aLen = 0;
	int i;

	aLen = strlen(authStr);
	for(i = 0; i < aLen; i++)
	{
		if(authStr[i] == sep)
		{			
			break;
		}
	}
	if(i == aLen)
	{
		ERR("No seprator '%c' found in specified string!\n", sep);
		return -1;
	}
	return i;
}

int parseOsInfo(char* authStr, char *osname, char* curver)
{
	char sep = INFO_SEP_CHAR;
	int idx = -1;
	size_t aLen = 0;

	idx = find_sep(authStr, sep);
	if(idx < 0)
		return -1;
	
	aLen = strlen(authStr);	
    if(idx == 0) {
		ERR("Error: no os name specified!\n");
		return -1;
	}
	else if(idx == (aLen - 1)) {
		ERR("Error: no current version specified!\n");
		return -1;
	}

	if(idx >= MAX_OS_NAME_LEN) {
	  	ERR("Error: os name exceeds the max len( %d )\n", MAX_OS_NAME_LEN);
		return -1;
    }

	if((aLen - idx) >= MAX_VER_LEN) {
	  	ERR("Error: os version exceeds the max len( %d )\n", MAX_VER_LEN);
		return -1;
    }
	
	snprintf(osname, idx+1, "%s", authStr);
	snprintf(curver, aLen - idx, "%s", &authStr[idx+1]);
	
	return 0;
}


int parseAuth(char* authStr, char *username, char* password)
{
	char sep = INFO_SEP_CHAR;
	int idx = -1;
	size_t aLen = 0;
	
	idx = find_sep(authStr, sep);
	if(idx < 0)
		return -1;

	aLen = strlen(authStr);		
	if(idx == 0) {
		ERR("Error: no username specified!\n");
		return -1;
	}
	else if(idx == (aLen - 1)) {
		ERR("Error: no password specified!\n");
		return -1;
	}

	if(idx >= MAX_U_NAME_LEN) {
	  	ERR("Error: user name exceeds the max len( %d )\n", MAX_U_NAME_LEN);
		return -1;
    }

	if((aLen - idx) >= MAX_PWD_LEN) {
	  	ERR("Error: password exceeds the max len( %d )\n", MAX_PWD_LEN);
		return -1;
    }
	
	snprintf(username, idx+1, "%s", authStr);
	snprintf(password, aLen - idx, "%s", &authStr[idx+1]);
	
	return 0;
}

#if (defined SDL_BUILD_ENABLE) && (defined ENABLE_TC_MUTUAL_TLS)

int 
setup_cert_params(char **pp_cert_full_file_path,
	char *cert_dir,
	char *cert_file
	) {
	int ret;
	char *tmp_file = NULL;
	size_t str_len;
	
	if(!cert_dir || 
		!pp_cert_full_file_path || 
		!cert_file)
		return -1;

	*pp_cert_full_file_path = NULL;
	str_len = strlen(cert_dir) + strlen(cert_file) + 4;
	tmp_file = (char *)calloc(str_len, sizeof(char));
	if(!tmp_file) {
		ERR("Error: failed to alloc mem for ca file\n");
		ret = -1
		goto invalid;
	}

	snprintf(tmp_file, str_len, "%s\%s", cert_dir, cert_file);
	*pp_cert_full_file_path = tmp_file;
	ret = 0;

invalid:
	if(ret != 0) {
		if(tmp_file)
			free(tmp_file);
	}
	return ret;
}
#endif

SERVER_CONFIG* parse_cmdline(int argc, char *argv[]) {
	int ret = -1;
	int c,i, sum;
	char valid[REQUIRED_NUMS] = {0};
	SERVER_CONFIG *s_config = NULL;
	char *authStr = NULL;
	char *osInfo = NULL;
	char *tmp_path = NULL;
	struct option long_options[] = {
		{"ip", required_argument, NULL, 'i'},
		{"port", required_argument, NULL, 'p'},
		{"authenticate", required_argument, NULL, 'a'},		
		{"sync", required_argument, NULL, 's'},
		{"disconnect", no_argument, NULL, 'd'},
		{"upload_check", required_argument, NULL, 'c'},
		{"upload", required_argument, NULL, 'u'},
		{"nic", required_argument, NULL, 'n'},
		{"filename", required_argument, NULL, 'f'},
		{"tokenfile", required_argument, NULL, 't'},
		{"sync_data", required_argument, NULL, 'q'},
		{"upload_data_check", required_argument, NULL, 'v'},
		{"upload_data", required_argument, NULL, 'g'},
		{"list_user_id", no_argument, NULL, 'l'},
#if (defined SDL_BUILD_ENABLE) && (defined ENABLE_TC_MUTUAL_TLS)
		{"cert_dir", required_argument, NULL, 'k'},
#endif
		{0, 0, 0, 0}
	};	

	s_config = (SERVER_CONFIG*)calloc(1, sizeof(SERVER_CONFIG));
	if(!s_config){
		ERR("Failed to allocate mem for config cls\n");
		return s_config;
	}
	//memset(s_config,0x0, sizeof(SERVER_CONFIG));

#if (defined SDL_BUILD_ENABLE) && (defined ENABLE_TC_MUTUAL_TLS)
	while((c = getopt_long(argc, argv, "a:i:p:s:dc:u:n:f:t:q:v:g:lk:", 
#else
	while((c = getopt_long(argc, argv, "a:i:p:s:dc:u:n:f:t:q:v:g:l", 
#endif
		long_options, &i)) >= 0) {
		switch (c) {
		case 'i':
			strcpy(s_config->ip, optarg);
			break;
		case 'p':
			s_config->port = strtol(optarg, NULL, 0);
			break;
		case 'a':
			s_config->status = AUTHENTICATION;
			if(authStr)
				free(authStr);
			authStr = strdup(optarg);
			valid[0] = 1;
			break;
		case 's':
			s_config->status = SYNC;
			if(osInfo)
				free(osInfo);
			osInfo = strdup(optarg);
			valid[1] = 1;
			break;
		case 'd':
			s_config->status = DISCONNECT;
			valid[2] = 1;
			break;
		case 'c':
			s_config->status = UPLOAD_CHECK;
			if(osInfo)
				free(osInfo);
			osInfo = strdup(optarg);
			valid[3] = 1;
			break;
		case 'u':
			s_config->status = UPLOAD;
			if(osInfo)
				free(osInfo);
			osInfo = strdup(optarg);
			valid[4] = 1;
			break;
		case 'n':
			strcpy(s_config->nicName, optarg);
			break;
   		case 'f':
			strcpy(s_config->filename, optarg);
			break;
   		case 't':
			strcpy(s_config->tokenfile, optarg);
			break;
		case 'v':
			s_config->status = UPLOAD_CHECK_DATA;
			if(osInfo)
				free(osInfo);
			osInfo = strdup(optarg);
			valid[5] = 1;
			break;
		case 'g':
			s_config->status = UPLOAD_DATA;
			if(osInfo)
				free(osInfo);
			osInfo = strdup(optarg);
			valid[6] = 1;
			break;
		case 'q':
			s_config->status = SYNC_DATA;
			if(osInfo)
				free(osInfo);
			osInfo = strdup(optarg);
			valid[7] = 1;
			break;
		case 'l':
			s_config->status = GET_USER_ID;
			valid[8] = 1;
			break;
#if (defined SDL_BUILD_ENABLE) && (defined ENABLE_TC_MUTUAL_TLS)
		case 'k':
			size_t str_len;

			str_len = strlen(optarg) + 2;
			if(tmp_path)
				free(tmp_path);
			tmp_path = (char *)calloc(str_len, sizeof(char));
			if(!tmp_path) {
				ERR("Failed to allocate mem for cert dir\n");
				ret = -1;
				goto invalid;
			}
			strncpy(tmp_path, optarg, str_len);
			break;
#endif
		case '?' : 
			if ( optopt == 'i' || optopt == 'p' || optopt == 'n' || 
			  	optopt == 'a'|| optopt == 's'|| optopt == 'c'|| optopt == 'u' ||
			  	optopt == 'v' ||optopt == 'g' ||optopt == 'q')  
                ERR("Error: option -%c requires an argument\n", optopt);  
            //else if ( isprint(optopt) )  
             //   ERR("Error: unknown option '-%c'\n", optopt);  
            else  
                ERR("Error: unknown option character '0x%x'\n", optopt);

			ret = -1;
			goto invalid;
		default:
			ERR("Unsupported command '%c'!\n", c);
			ret = -1;
			goto invalid;
		}
	}
	
	for(i = 0, sum=0; i < REQUIRED_NUMS; i++)
	{
		sum += valid[i];
	}
	
	if(sum == 0)
	{
		ERR("Error: no communicate mode [-a|-s|-d|-c|-u|-q|-v|-g|-l] is set!\n");
		ret = -1;
		goto invalid;
	}
	else if(sum > 1)
	{
		ERR("Error: only one communicate mode [-a|-s|-d|-c|-u|-q|-v|-g|-l] "
			"can be set at a time!\n");
		ret = -1;
		goto invalid;
	}

	if(s_config->ip[0] == 0)
	{
		DEBUG("Ip is not set, will use \"%s\" by default\n", SRV_IP);
		strcpy(s_config->ip, SRV_IP);
	}
	
	if(!s_config->port)
	{
		DEBUG("Port is not set, will use \"%d\" by default\n", SRV_PORT);
		s_config->port = SRV_PORT;
	}
	
	if(s_config->nicName[0] == 0)
	{
		//DEBUG("Nic is not set, will use \"%s\" by default\n", NIC_NAME);
		strcpy(s_config->nicName, NIC_NAME);
	}

#if (defined SDL_BUILD_ENABLE) && (defined ENABLE_TC_MUTUAL_TLS)

	if(!tmp_path) {
		ERR("Failed to get cert dir\n");
		ret = -1;
		goto invalid;
	}

	ret = setup_cert_params(&s_config->ca_file,
					tmp_path,
					DFT_CA_FILE);
	if(ret != 0) {
		ERR("Failed to set up ca file path: %s\n");
		goto invalid;
	}
	
	ret = setup_cert_params(&s_config->client_cert_file,
			tmp_path,
			DFT_CERT_FILE);
	if(ret != 0) {
		ERR("Failed to set up cert file path: %s\n", tmp_path);
		goto invalid;
	}

	ret = setup_cert_params(&s_config->key_file,
			tmp_path,
			DFT_KEY_FILE);
	if(ret != 0) {
		ERR("Failed to set up key file path: %s\n", tmp_path);
		goto invalid;
	}
	
#endif

	if(authStr != NULL) {
		if(parseAuth(authStr, s_config->username, s_config->password) < 0) {
			ret = -1;
			goto invalid;
		}

		//DEBUG("username is %s, password is %s\n", s_config->username, s_config->password);
	}

	if(osInfo != NULL) {
		if(parseOsInfo(osInfo, s_config->osname, s_config->curVer) < 0) {
			ret = -1;
			goto invalid;
		}

		DEBUG("os_id is %s, current version is %s\n", s_config->osname, s_config->curVer);
	}	

    if (s_config->tokenfile[0] == 0)
    {
		ERR("Error: please set the token file!\n");
		ret = -1;
		goto invalid;
    }

	ret = 0;
invalid:
	if(tmp_path)
		free(tmp_path);
	if(authStr)
		free(authStr);
	if(osInfo)
		free(osInfo);

	if((ret != 0) && 
		s_config) {
	
#if (defined SDL_BUILD_ENABLE) && (defined ENABLE_TC_MUTUAL_TLS)
		if(s_config->ca_file)
			free(s_config->ca_file);

		if(s_config->cert_file)
			free(s_config->cert_file);

		if(s_config->key_file)
			free(s_config->key_file);
#endif
	
		free(s_config);
		s_config = NULL;
	}
	
	return s_config;
}
