/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "protocol.h"
#include "session.h"
#include "debug.h"

tc_hd_t tc_hd[2];
extern int max_pkt_size;

void str2hex(const char *str, int len, __uint8_t *hex)
{
    __uint8_t  h1, h2;
    int i;
    for ( i = 0; i < len/2; ++i )
    {
        h1 = *str++ - 'A';
        h2 = *str++ - 'A';
        hex[i] = (h1 << 4) | h2;
    }
}

void hex2str(__uint8_t *hex, int len, char *str)
{
    int i;
    for ( i = 0; i < len; ++i )
    {
        *str++ = (hex[i] >> 4) + 'A';
        *str++ = (hex[i] & 0x0F) + 'A';
    }
}

int allc_pkt_list(tc_hd_t *hd)
{
	uchar     dst[6];
  
  	hd->snd_pkt = (tc_pkt_t *)malloc(max_pkt_size);
  	if(hd->snd_pkt == NULL) {
		ERR("not enough memory\r\n");
		return -1;
  	}
  
  	hd->rcv_pkt = (tc_pkt_t *)malloc(max_pkt_size);
  	if(hd->rcv_pkt == NULL) {
		ERR("not enough memory\r\n");
		free(hd->snd_pkt);
		hd->snd_pkt = NULL;
		return -1;
  	}

  	if(init_tc_hdr(&(hd->snd_pkt->hdr)) != 0) {
		free(hd->rcv_pkt);
		free(hd->snd_pkt);
		hd->rcv_pkt = NULL;
		hd->snd_pkt = NULL;
		return -1;
  	}
	
  	return 0;
}

void free_pkt_list(tc_hd_t *hd)
{
	if(!hd)
		return;
	
	if(hd->rcv_pkt != NULL) {
		free(hd->rcv_pkt);
		hd->rcv_pkt = NULL;
	}
	if(hd->snd_pkt != NULL) {
		free(hd->snd_pkt);
		hd->snd_pkt = NULL;
	}
}


int send_conf_cmd(tc_hd_t *hd, int cmd, char*confData, uchar prtRcvData) 
{
  int ret;
  uint snd_len;
  uint rcv_len;
  int conf_len;
  tc_pkt_t *snd_pkt = hd->snd_pkt;
  tc_pkt_t *rcv_pkt = hd->rcv_pkt;
  
  if(!confData)
  {
  	ERR("Conf data should not be NULL!\n");
  	return -1;
  }

  if(cmd == AUTH_CONF)
  {
      memset(snd_pkt->hdr.tag, 0, 4);
  }
  else
  {
    FILE *t;
    char buf[8];
    t = fopen(hd->s_config->tokenfile,"rb");
    if (t)
    {
      fread(buf, sizeof(buf), 1, t);
      str2hex(buf, sizeof(buf), snd_pkt->hdr.tag);
      fclose(t);
    }
  }

  conf_len = strlen(confData);
   
  snd_pkt->hdr.cmd = TC_CMD_CONFIG;
  snd_pkt->tc_cmd.vercmd = 0xf0 & (cmd << 4);
  memset(snd_pkt->tc_cmd.data, 0x0, conf_len + 1);
  strcpy((char *)snd_pkt->tc_cmd.data, confData);  
  snd_pkt->tc_cmd.len = conf_len;

  snd_len = sizeof(tc_pkt_t) + snd_pkt->tc_cmd.len;
  rcv_len = max_pkt_size;

  ret = tc_cmd(snd_pkt, snd_len, rcv_pkt, &rcv_len);
  if(ret != 0) {
    //DEBUG("tc_cmd error \r\n");
    return ret;
  }

  if(cmd == AUTH_CONF)
  {
    FILE *t;
    char buf[8];
    t = fopen(hd->s_config->tokenfile,"wb");
    if (t)
    {
      hex2str(rcv_pkt->hdr.tag, 4, buf);
      fwrite(buf, sizeof(buf), 1, t);
      fclose(t);
    }
  }
  else if((cmd ==RESET_CONF) && (rcv_pkt->hdr.error==0))
  {
    remove(hd->s_config->tokenfile);
  }

  if(prtRcvData) {
        FILE *f;
        char *data;
        size_t len;

        data = (char *)(rcv_pkt->tc_cmd.data);
        len = strlen(data);

        if ( len > 0 )
        {
            data[len] = '\n';
            len++;

            PRT("%s\r\n", data);
	    	if(hd->s_config->filename[0] != 0) {
            	f = fopen(hd->s_config->filename,"wb");
            	if (f)
            	{
                	fwrite(rcv_pkt->tc_cmd.data, len, 1, f);
                	fclose(f);
            	}
	  		}
        }
  }
  
  return 0;
  
}

void logError(int FullErrorCode)
{
	char errCode;
	char leftTimes = 0;

	errCode = (char)FullErrorCode;
	switch(errCode) {
	case TC_ERR_ONLINE:
		ERR("Error: User with selected OS has been online already\n");
		break;
	case TC_ERR_PWD_UNAME:
    	ERR("Error: Incorrect username or password\r\n");
		leftTimes = (char)(FullErrorCode >> 8);
		if(leftTimes > 0) {
			ERR("%d times left\r\n", leftTimes);
		}
		break;
	case TC_ERR_CONN_CLOSED:
		ERR("Error: Connection has been closed by server\n");
		break;
	case TC_ERR_RECV:
		ERR("Error: Receive failure :%s\r\n", strerror(errno));
		break;
	case TC_ERR_SEND:
		ERR("Error: Send failure :%s\r\n", strerror(errno));
		break;
	case TC_ERR_NOT_LOGINED:
		ERR("Error: You did not login or session timeout, re-login please\n");
		break;
	case TC_ERR_CONFIG_ERR:
		ERR("Error: Incorrect config cmd string sent\n");
		break;
	case TC_ERR_BAD_VERSION:
		ERR("Error: Os image version on client differs from the one on server\n");
		break;
	case TC_ERR_PERMISSION:
		ERR("Error: User permission error\n");
		break;
	case TC_ERR_IMAGE_LOCKED:
		ERR("Error: Image was being locked\n");
		break;
	case TC_ERR_IMAGE_NOT_READY:
		ERR("Error: Image was not ready\n");
		break;
	case TC_ERR_USER_BLOCKED:
		ERR("Error: User was locked\n");
		break;
	case TC_ERR_MAXIUM_FAILURE:
		ERR("Error: maximum login failure times reached, please try again later\n");
		break;
	default:
		ERR("Error: Server set the response error bit: 0x%x\n", errCode);
		break;
	}
}

int auth_user()
{
  int len;
  int ret;
  SERVER_CONFIG *s_config;
  
  s_config = tc_hd[0].s_config;
  
  ret = do_auth(&tc_hd[0], s_config->username, s_config->password);
  if(ret != 0) {
  	logError(ret);
    return ret;
  }
  
  return 0;
}

int get_user_id()
{
	int len;
	int ret;
	FILE *fd;
	char buf[8] = {0};
	SERVER_CONFIG *s_config;

	s_config = tc_hd[0].s_config;

	ret = send_conf_cmd(&tc_hd[0], GET_USER_ID_CONF, " ", 1);
	if(ret != 0) {
		logError(ret);
		return ret;
	}
  
 	return 0;
}


int get_oslist(tc_hd_t *hd, char *username, char *password)
{
  int           ret;
  char			auth_data[256] = {0};

  sprintf((char *)auth_data, "username=%s%spassword=%s", 
  					username, INFO_SEP_STR, password);	
  ret = send_conf_cmd(hd, AUTH_CONF, auth_data, 1);
  
  if(ret != 0) {
    //DEBUG("tc_cmd error \r\n");
    return ret;
  }

  //DEBUG("oslist\n");  
  
  return ret;
}

int do_auth(tc_hd_t *hd, char *username, char *password)
{
	int ret;
  ret = get_oslist(hd, username, password);
  if(ret != 0) { 
    //DEBUG("failed to get oslist \r\n");
    goto error;
  }

  return 0;
  
error:
  return ret;
}

int generate_os_info(SERVER_CONFIG *s_config, char *osinfo)
{
	if(!osinfo)
		return -1;

	snprintf(osinfo, MAX_STRING_LEN, "%s%s%s", 
				s_config->osname, INFO_SEP_STR, s_config->curVer);

	return 0;
}

int select_osimage(tc_hd_t *hd) 
{
  int ret;
  uint snd_len;
  uint rcv_len;
  char osinfo[MAX_STRING_LEN] = {0};  
  tc_pkt_t *snd_pkt = hd->snd_pkt;
  tc_pkt_t *rcv_pkt = hd->rcv_pkt;

  generate_os_info(hd->s_config, osinfo);
   
  snd_pkt->hdr.cmd = TC_CMD_CONFIG;
  snd_pkt->tc_cmd.vercmd = 0xf0 & (SET_OS_CONF << 4);
  memset(snd_pkt->tc_cmd.data, 0x0, MAX_STRING_LEN);
  strcpy((char *)snd_pkt->tc_cmd.data, osinfo);

  snd_len = sizeof(tc_pkt_t) + MAX_STRING_LEN;
  rcv_len = max_pkt_size;

  ret = tc_cmd(snd_pkt, snd_len, rcv_pkt, &rcv_len);
  if(ret != 0) {
    //ERR("tc_cmd in select_osimage error \r\n");
    return ret;
  }  
  return 0;
  
}

int sync_os(tc_hd_t *hd)
{
	int ret;
	char osinfo[MAX_STRING_LEN] = {0};

	if(!hd) {
		ERR("Sync_os: hd is null\n");
		return -1;
	}

	generate_os_info(hd->s_config, osinfo);
	
	DEBUG("\r\nSyncing %s ...\n", osinfo);
  
  	ret = select_osimage(hd);
  
  	return ret;
}

int do_sync_data()
{
	int ret =-1;
	tc_hd_t *hd = &tc_hd[0];
	char osname[MAX_OS_NAME_LEN + 1] = {0};
  	char curver[MAX_VER_LEN + 1] = {0};
	char osinfo[MAX_STRING_LEN] = {0};
	SERVER_CONFIG *s_config;
		
	//if((ret = sync_os(hd)) != 0)
	//	goto error;	

  	s_config = tc_hd[0].s_config;

	generate_os_info(s_config, osinfo);
	
	ret = send_conf_cmd(hd, SYNC_DATA_CONF, osinfo, 1);
	if(ret != 0)
		goto error;

	return 0;
	
error:	
	logError(ret);
	ERR("Fail to sync %s\r\n", osinfo);
	return ret;
  	
}

int do_sync()
{
	int ret =-1;
	tc_hd_t *hd = &tc_hd[0];
	char osname[MAX_OS_NAME_LEN + 1] = {0};
  	char curver[MAX_VER_LEN + 1] = {0};
	char osinfo[MAX_STRING_LEN] = {0};
	SERVER_CONFIG *s_config;
		
	//if((ret = sync_os(hd)) != 0)
	//	goto error;	

  	s_config = tc_hd[0].s_config;

	generate_os_info(s_config, osinfo);
	
	ret = send_conf_cmd(hd, SYNC_CONF, osinfo, 1);
	if(ret != 0)
		goto error;

	return 0;
	
error:	
	logError(ret);
	ERR("Fail to sync %s\r\n", osinfo);
	return ret;
  	
}

int do_upload_check_data()
{
	int ret;
	char osinfo[MAX_STRING_LEN] = {0};
	
	tc_hd_t *hd = &tc_hd[0];
	generate_os_info(hd->s_config, osinfo);
	
	ret = send_conf_cmd(hd, UPLOAD_CHECK_DATA_CONF, osinfo, 0);
	if(ret != 0)
		goto error;

	return 0;
	
error:	
	logError(ret);
	ERR("Failed to check uploading conditions \r\n");
	return ret;
}

int do_upload_check()
{
	int ret;
	char osinfo[MAX_STRING_LEN] = {0};
	
	tc_hd_t *hd = &tc_hd[0];
	generate_os_info(hd->s_config, osinfo);
	
	ret = send_conf_cmd(hd, UPLOAD_CHECK_CONF, osinfo, 0);
	if(ret != 0)
		goto error;

	return 0;
	
error:	
	logError(ret);
	ERR("Failed to check uploading conditions \r\n");
	return ret;
}


int do_upload_data()
{
	int ret;
	char osinfo[MAX_STRING_LEN] = {0};
	
	tc_hd_t *hd = &tc_hd[0];
	generate_os_info(hd->s_config, osinfo);
	
	ret = send_conf_cmd(hd, UPLOAD_DATA_CONF, osinfo, 0);
	if(ret != 0)
		goto error;

	return 0;
	
error:	
	logError(ret);
	ERR("Failed to upload \r\n");
	return ret;
}


int do_upload()
{
	int ret;
	char osinfo[MAX_STRING_LEN] = {0};
	
	tc_hd_t *hd = &tc_hd[0];
	generate_os_info(hd->s_config, osinfo);
	
	ret = send_conf_cmd(hd, UPLOAD_CONF, osinfo, 0);
	if(ret != 0)
		goto error;

	return 0;
	
error:	
	logError(ret);
	ERR("Failed to upload \r\n");
	return ret;
}

int disconnect()
{
	int ret;
	char osinfo[MAX_STRING_LEN] = {0};
	
	tc_hd_t *hd = &tc_hd[0];
	generate_os_info(hd->s_config, osinfo);
	
	ret = send_conf_cmd(hd, RESET_CONF, osinfo, 0);
	if(ret != 0)
		goto error;

	return 0;
	
error:	
	logError(ret);
	ERR("Failed to disconnect \r\n");
	return ret;
}


