/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <netinet/tcp.h>
#include <netinet/in.h>		/* sockaddr_in, htons, in_addr */
#include <netdb.h>		/* hostent, gethostby*, getservby* */
#include <stdio.h>
#include <fcntl.h>
#include <syslog.h>
#include <stdlib.h>
#include <sys/mount.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <linux/ioctl.h>
#include <errno.h>
#include "session.h"
#include "debug.h"

void dump_buf(unsigned char *buf, int sz)
{
	int i,j;
	for(i = 0; i < sz/16; i++) {
		DEBUG("0x%04x: ", i*16);
		for(j = 0; j < 16; j++) {
	 		DEBUG("%02x ", buf[i*16+j]);
		}
		DEBUG("\r\n");
	}

	DEBUG("0x%04x: ", i*16);
	for(j = 0; j < sz%16; j++) {
		DEBUG("%02x ", buf[i*16+j]);
	}
	DEBUG("\r\n");
}

void setmysockopt(int sock) {
	int size = 1;
#ifdef	IPPROTO_TCP
	if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &size, sizeof(int)) < 0)
		 ERR("no sockopt TCP_NODELAY: %s\n", strerror(errno));
#endif
	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &size, sizeof(int)) < 0)
		 ERR("no sockopt SO_REUSEADDR: %s\n", strerror(errno));
}

#ifdef SDL_BUILD_ENABLE

int setup_client_ssl(SERVER_CONFIG *s_config,
	int sock)
{
	int ret;
	
	if(!s_config)
		return -1;
	
	s_config->p_ssl = SSL_new(s_config->p_tc_ssl_ctx);
	if(s_config->p_ssl == NULL) {
		ERR("setup_client_ssl: failed to new SSL\n");
		ret = -1;
		goto setup_client_ssl_cleanup;
	}

	SSL_set_fd(s_config->p_ssl, sock);
	 
	if(SSL_connect(s_config->p_ssl) == -1) {
		ERR("setup_client_ssl: failed to connect fd\n");
		ret = -1;
		goto setup_client_ssl_cleanup;
	}
	
	//printf("Client with %s encryption\n", SSL_get_cipher(s_config->p_ssl));

	ret = 0;

setup_client_ssl_cleanup:
	if(ret != 0) {
		if(s_config->p_ssl) {
			SSL_shutdown(s_config->p_ssl);
			SSL_free(s_config->p_ssl);
			s_config->p_ssl = NULL;
		}
	}
	
	return ret;
}

int init_ssl_work(SERVER_CONFIG *p_s_config)
{	
	int ret;
	SSL_CTX *p_tc_ssl_ctx = NULL;
	const SSL_METHOD *p_tc_ssl_method;

	if(!p_s_config 
#ifdef ENABLE_TC_MUTUAL_TLS	
		|| !p_s_config->ca_file ||
		!p_s_config->client_cert_file || 
		!p_s_config->client_key_file
#endif
	) {
		ERR("Error: ssl params are NULL\n");
		return -1;
	}

	p_s_config->p_tc_ssl_ctx = NULL;

	SSL_library_init();
	OpenSSL_add_all_algorithms();
	SSL_load_error_strings();
	
	p_tc_ssl_method = TLSv1_2_client_method();	
	p_tc_ssl_ctx = SSL_CTX_new(p_tc_ssl_method);
	if (p_tc_ssl_ctx == NULL) {
		ERR("Failed to create ssl ctx\n");
		ret = -1;
		goto init_ssl_cleanup;
	}

#ifdef ENABLE_TC_MUTUAL_TLS	
	SSL_CTX_set_verify(p_tc_ssl_ctx, SSL_VERIFY_PEER | 
			SSL_VERIFY_FAIL_IF_NO_PEER_CERT, 0);
	
	if (SSL_CTX_load_verify_locations(p_tc_ssl_ctx, 
			p_s_config->ca_file, 
			0) != 1) {
		ERR("Failed to load CA file: %s\n", p_s_config->ca_file);
		ret = -2;
		goto init_ssl_cleanup;
	}
	
	if (SSL_CTX_set_default_verify_paths(p_tc_ssl_ctx) != 1) {
	 	ERR("Call to SSL_CTX_set_default_verify_paths failed");
		ret = -2;
		goto init_ssl_cleanup;
	}
	
	if (SSL_CTX_use_certificate_file(p_tc_ssl_ctx, 
			p_s_config->client_cert_file, 
			SSL_FILETYPE_PEM) <= 0) {
		ERR("Failed to load cert file: %s\n", p_s_config->client_cert_file);
		ret = -3;
		goto init_ssl_cleanup;
	}
	
	if (SSL_CTX_use_PrivateKey_file(p_tc_ssl_ctx, 
			p_s_config->client_key_file, 
			SSL_FILETYPE_PEM) <= 0) {
		ERR("Failed to load key file: %s\n", p_s_config->client_key_file);
		ret = -4;
		goto init_ssl_cleanup;;
	}
	
	if (!SSL_CTX_check_private_key(p_tc_ssl_ctx)) {
		ERR("Failed to check private key\n");
		ret = -1;
		goto init_ssl_cleanup;
	}
	 
	SSL_CTX_set_mode(p_tc_ssl_ctx, SSL_MODE_AUTO_RETRY);

#endif
	p_s_config->p_tc_ssl_ctx  = p_tc_ssl_ctx;
	
	ret = 0;
	
init_ssl_cleanup:

	if(ret != 0) {	
		if(p_tc_ssl_ctx)
			SSL_CTX_free(p_tc_ssl_ctx);
	}
	
	return ret;
}

#endif

int opennet(SERVER_CONFIG *s_config) {
	int sock;
	struct sockaddr_in xaddrin;
	int xaddrinlen = sizeof(xaddrin);
	struct hostent *hostn;
	int af;


	DEBUG("Connecting to %s:%u\n", s_config->ip, (unsigned int)s_config->port);
	hostn = gethostbyname(s_config->ip);
	if (!hostn)
	{
		ERR("Gethostname failed: %s\n", strerror(errno));
		return -1;
	}

	af = AF_INET;
	if ((sock = socket(af, SOCK_STREAM, IPPROTO_TCP)) < 0)
	{
		ERR("Socket failed: %s\n", strerror(errno));
		return -1;
	}

	s_config->socket = sock;

	xaddrin.sin_family = af;
	xaddrin.sin_port = htons(s_config->port);
	xaddrin.sin_addr.s_addr = *((int *) hostn->h_addr);
	if ((connect(sock, (struct sockaddr *) &xaddrin, xaddrinlen) < 0))
	{
		ERR("Connect: %s\n", strerror(errno));
		close(sock);
		s_config->socket = -1;
		return -1;
	}

	setmysockopt(sock);
	return 0;
}


void usage()
{	
	printf("tc sync client version %s\n", "1.0.0");
	printf("Usage: \ntc-sync-client -i host -p port -f filename -t tokenfile" 
#ifdef SDL_BUILD_ENABLE
#ifdef ENABLE_TC_MUTUAL_TLS
		" -k certDir"
#endif
#endif
		"\n" \
		"\t< [<-a|--authenticate> username:password] | \n"\
		"\t[-q|--sync_data os_id:current_version] | \n"\
		"\t[-v|--upload_check_data os_id:current_version] | \n"\
		"\t[-g|--upload_data os_name:current_version] | \n" \
		"\t[-s|--sync os_id:current_version] | \n"\
		"\t[-c|--upload_check os_id:current_version] | \n"\
		"\t[-l|--get_user_id] | \n"\
		"\t[-u|--upload os_name:current_version] | [-d|--disconnect]>\n");
#ifdef SDL_BUILD_ENABLE
#ifdef ENABLE_TC_MUTUAL_TLS
	   "Notes: \"-k\" assign the directory where file \"" DFT_CA_FILE "\" & \"" \
	   DFT_CERT_FILE "\" & \"" DFT_KEY_FILE "\" resides"
#endif
#endif
	
	return;	
}

int main(int argc, char *argv[]) {
	int ret;
	SERVER_CONFIG *s_config;

	s_config = parse_cmdline(argc, argv);
	
	if(!s_config) {		
		usage();
		return -1;
	}

#ifdef SDL_BUILD_ENABLE
	ret = init_ssl_work(s_config);
	if (ret != 0) {
		ERR("Failed to init ssl\n");
		ret = TC_ERR_SSL_FAILURE;
	  	goto init_invalid;
	}
#endif
	
	ret = opennet(s_config);
	if(ret != 0) {		
		ERR("opennet failed.\n");
        ret = TC_ERR_CONN_CLOSED;
		goto init_invalid;
	}

#ifdef SDL_BUILD_ENABLE
	ret = setup_client_ssl(s_config, s_config->socket);
	if (ret != 0) {
		ERR("Failed to init ssl\n");
		ret = TC_ERR_SSL_FAILURE;
		goto init_invalid;
	}
	MSG("TLS enabled\n");
#endif
		
	if((ret = allc_pkt_list(&tc_hd[0])) != 0) {
		goto init_invalid;
	}

	tc_hd[0].s_config = s_config;

	switch(s_config->status) {
		case AUTHENTICATION:	
			ret = auth_user();
			break;
		case SYNC_DATA:
			ret = do_sync_data();
			break;
		case SYNC:
			ret = do_sync();
			break;
		case UPLOAD_CHECK_DATA:
			ret = do_upload_check_data();
			break;
		case UPLOAD_CHECK:
			ret = do_upload_check();
			break;
		case UPLOAD_DATA:
			ret = do_upload_data();
			break;
		case UPLOAD:
			ret = do_upload();
			break;
		case DISCONNECT:
			ret = disconnect();
			break;
		case GET_USER_ID:	
			ret = get_user_id();
			break;
		default:
			usage();
			ret = -1;
			goto init_invalid;
	}	

init_invalid:
	if(s_config) {
		
#ifdef SDL_BUILD_ENABLE
#ifdef ENABLE_TC_MUTUAL_TLS
		if(s_config->ca_file)
			free(s_config->ca_file);

		if(s_config->cert_file)
			free(s_config->cert_file);

		if(s_config->key_file)
			free(s_config->key_file);
#endif

		if(s_config->p_ssl) {
			SSL_shutdown(s_config->p_ssl);
			SSL_free(s_config->p_ssl);
			s_config->p_ssl = NULL;
		}

		if(s_config->p_tc_ssl_ctx)
			SSL_CTX_free(s_config->p_tc_ssl_ctx);
#endif
		if(s_config->socket > 0) {
			close(s_config->socket);
			s_config->socket = -1;
		}
		free(s_config);
	}

	free_pkt_list(&tc_hd[0]);

	return ret;
}



