/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include "predefine.h"
#include "tcprotocol.h"
#include "client.h"
#include "db.h"
#include "debug.h"
#include "tctimer.h"

extern Client g_running_clients[];
extern fd_set g_mset;

int mark_client_status(Client* client, int status_id)
{
	if(!client)
		return -1;

	client->status = status_id;
	update_user_status(client->username, client->image_id, status_id);
	return 0;
}

int mark_client_invalid(Client* client)
{
	return mark_client_status(client, INVALID);
}

int mark_client_running(Client* client)
{
	return mark_client_status(client, RUNNING);
}


int 
get_client(
	Client client_array[], 
	unsigned char *addr, 
	char *username
	)
{
	int i = -1, aval_idx =-1;

	if(addr != NULL)
	{		
		for(i = 0; i < MAX_CLIENTS; i++) {
			//if the client was from the  same IP, we may treat it as client reboot
			if(IN_USE(client_array+i)) {
			   if(memcmp(addr, client_array[i].addr, 4) == 0 &&
			   		(client_array[i].sock == -1)) {
			   		if(username != NULL) {
						if(strcmp(client_array[i].username, 
							username) == 0) {
							DEBUG("Reuse exists client(%s) %d\n", 
								username,
								i);
							break;
						}
			   		} else {
				   		DEBUG("Reuse exists client %d\n", i);
						break;
			   		}
			   	}
			}else if( aval_idx < 0){
				//DEBUG("aval_idx is %d, client->usedIdx is %d",i, client_array[i].usedIdx);
				aval_idx = i;
			}
		}
	}else {
		for(i = 0; i < MAX_CLIENTS; i++)
		{
			if(NOT_IN_USE(client_array+i))
			  break;
		}
	}

	if( i == MAX_CLIENTS )
		i = aval_idx;

	if(i == -1) {
		ERR("no client slot available, max slot is: %d\n", MAX_CLIENTS);
		return -1;
	}

FOUND:

#ifdef SDL_BUILD_ENABLE
	if(client_array[i].p_ssl) {
		SSL_shutdown(client_array[i].p_ssl);
		SSL_free(client_array[i].p_ssl);
		client_array[i].p_ssl = 0;
	}
#endif
	if(client_array[i].sock > 0) {
		close(client_array[i].sock);
		client_array[i].sock = -1;
	}
	client_array[i].usedIdx = i;
	remove_timer(i);
	return i;
}

void client_array_init(Client *client_array, int item_cnt)
{
	int i;
	
	if(!client_array || item_cnt < 0)
		return;

	for(i = 0; i < item_cnt; i++)
		free_client(client_array+i);
}

int setup_client(Client *client, tc_pkt_t *p)
{
  int   online;
  int   ret;
  char  strmac[64] = {0};
  int   osimageid;
  char   osimageidStr[64] = {0};
  //uint8_t olderror;
  char * tmpStr;
  int isABUser = 0;
  size_t imagename_len;

  //olderror = p->hdr.error;
  p->hdr.error = TC_ERR_DATABASE;

  if(client->image_name){
  	free(client->image_name);
	client->image_name = NULL;
  }
  
  imagename_len = strlen(p->tc_cmd.data);
  client->image_name = (char *)calloc(imagename_len + 1, sizeof(char));
  if(!client->image_name) {
  	 ERR("Failed to alloc memory for image name\n");
	 return -1;
  }
  
  strncpy(client->image_name, p->tc_cmd.data, imagename_len);
  //client->image_name[MAX_STRING_LEN - 1] = '\0';
  
  //just to avoid os name string mixed with os version, such as "win7.img;4"
  if((tmpStr = strchr(client->image_name, INFO_SEP_CHAR)) != NULL) {
  	DEBUG("Mixed os name and os version string: %s\n", client->image_name);
  	client->image_name[tmpStr - client->image_name] = 0;
	DEBUG("Client os name is %s after fixed\n", client->image_name);
  }
  
  osimageid = get_osimageid_by_name(client->image_name);
  if(osimageid < 0) {
    ERR("Failed to find osimageid by name %s\n", client->image_name);
    return -1;
  }
  
  client->image_id = osimageid;

  isABUser = is_autoboot_user(client->username);  
  online = get_user_status(client->username, osimageid);
  if(online && !isABUser){	  	
    DEBUG("user is online already\n");
	p->hdr.error = TC_ERR_ONLINE;
    return -1;	  	
  }

  snprintf(osimageidStr, 64, "%d", osimageid);

  ret = get_imageaccess(osimageidStr, 
  					client->username, &client->imageaccess);
  if(ret < 0) {
    DEBUG("fail to get imageaccess\n");
    return -1;
  }  

  ret = is_image_seed_ready(osimageid);
   if(ret == 0) {
    DEBUG("Image seed(%s) was not ready\n", client->image_name);
	p->hdr.error = TC_ERR_IMAGE_NOT_READY;
    return -1;
  }  

  snprintf(strmac,64, "%d-%d-%d-%d", client->addr[0],client->addr[1],
						client->addr[2], client->addr[3]);

  ret = insert_user_status(client->username, client->image_id, strmac, 
  					client->imageaccess, BOOTING);
  if(ret != 0){
  	if(!isABUser || (ret != ERR_DUP_ENTRY_IN_DB)){
	    DEBUG("failed to set online status\n");
	    return -1;
  	}
  }

  p->hdr.error = 0;
  //p->hdr.error = olderror;
  
  return 0;
}

void free_client(Client *client)
{
	if(!client)
		return;
	if (NOT_IN_USE(client)) {
		goto exit_free;
	}

#ifdef SDL_BUILD_ENABLE
	if(client->p_ssl) {
		SSL_shutdown(client->p_ssl);
		SSL_free(client->p_ssl);
	}
#endif

	if(client->sock > 0) {
		FD_CLR(client->sock, &g_mset);
		close(client->sock);
	}

exit_free:
	if(client->username)
		free(client->username);

	if(client->image_name)
		free(client->image_name);
	memset(client, 0x0, sizeof(Client));
	client->usedIdx = -1;
	client->sock = -1;
}

