/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "predefine.h"
#include "tcprotocol.h"
#include "client.h"
#include "db.h"
#include "debug.h"
#include "process.h"
#include "conf_cmd.h"

extern Client g_running_clients[];

int _parse_info(char *infoStr, 
	char **ppdata1, int data1_offset, int max_data1_len,
	char **ppdata2, int data2_offset, int max_data2_len)
{
	int ret = -1;
	char *pos, *tmpStr;
	size_t dataLen;

	if(!ppdata1 || !ppdata2) {
		ERR("_parseinfo error: invalid parameter\n");
		return ret;
	}

	*ppdata1 = NULL;
	*ppdata2 = NULL;
	
	pos = strtok(infoStr, INFO_SEP_STR);
	if (pos == NULL) {
		ERR("_parseinfo error: failed to call strtok1\n");
		goto parse_exit;
	}

	tmpStr = (char *)(pos + data1_offset);
	dataLen = strlen(tmpStr);

	if(max_data1_len > 0 && dataLen > max_data1_len) {
		ERR("_parseinfo error: uname(%s;%lu) is too long\n",
			tmpStr,
			dataLen);
		goto parse_exit;
	}
	ret = escape_parameter(ppdata1, tmpStr);
	if(ret != 0) {
		ERR("_parseinfo error: failed to escape params1\n");
		goto parse_exit;
	}
	/*
	
	*ppdata1 = malloc(dataLen + 1);
	
	if(!(*ppdata1)) {
		ERR("_parseinfo error: failed to alloc1\n");
		goto parse_exit;
	}

	strncpy(*ppdata1, tmpStr, dataLen);
	(*ppdata1)[dataLen] = '\0';
	*/

	pos = strtok(NULL, INFO_SEP_STR);
	if (pos == NULL) {
		ERR("_parseinfo error: failed to call strtok2\n");
		goto parse_exit;
	}

	tmpStr = (char *)(pos + data2_offset);
	dataLen = strlen(tmpStr);

	if(max_data2_len > 0 && dataLen > max_data2_len) {
		ERR("_parseinfo error: pwd(%lu) is too long\n",
			dataLen);
		goto parse_exit;
	}
	
	ret = escape_parameter(ppdata2, tmpStr);
	if(ret != 0) {
		ERR("_parseinfo error: failed to escape params2\n");
		goto parse_exit;
	}
	/*
	dataLen = strlen(tmpStr);
	*ppdata2 = malloc(dataLen + 1);
	if(!(*ppdata2)) {
		ERR("_parseinfo error: failed to alloc2\n");
		goto parse_exit;
	}
	
	strncpy(*ppdata2, tmpStr, dataLen);
	(*ppdata2)[dataLen] = '\0';
	*/

	ret = 0;

parse_exit:

	if(ret != 0) {
		if(ppdata1 && *ppdata1)
			free(*ppdata1);

		if(ppdata2 && *ppdata2)
			free(*ppdata2);
	}
	
	return ret;
}

int _parse_user_info(char *userinfo, char **pUsername, char **pPassword)
{
	return _parse_info(userinfo, pUsername, 9, MAX_UNAME_LEN,
				pPassword, 9, MAX_PWD_LEN);
}

int _parse_os_info(char *osinfo, char **pOsid, char **pCurver)
{
	return _parse_info(osinfo, pOsid, 0, MAX_STRING_LEN,
				pCurver, 0, MAX_STRING_LEN);
}

int set_os_conf(Client *client, tc_pkt_t *p)
{
	int ret;

	if(NOT_IN_USE(client) || !client->username || client->username[0] == 0){
		ERR("client is not connected\n");
		p->hdr.error = TC_ERR_NOT_LOGINED;
		goto error;
	}

	ret = setup_client(client, p);
	if(ret != 0) {
		ERR("setup client %s failed\n", client->username);
		goto error;
	} else {
		MSG("client is setup for %s\n", client->username);		
	}
	p->tc_cmd.len = 0;
	return 0;
	
error:	
	p->hdr.ver_flags |= TC_FL_ERROR;
	return -1;	

}

int auth_conf(Client *client, tc_pkt_t *p)
{
	int ret = -1;
	int loginLeftTimes = 0;
	char *oslist = NULL;
	char *pUsername = NULL;
	char *pPassword = NULL;
	SERVER_CONFIG *s_config = NULL;

	p->hdr.error = TC_ERR_PWD_UNAME;
	p->hdr.cmd = 0;

	s_config = get_server_config();
	if(!s_config)
		goto auth_exit;
	
	/*oslist = (char *)malloc(MAX_BUF_SIZE);
	if(!oslist){
		ERR("failed to allocate mem with size %d for oslist in auth_conf\n", 
			MAX_BUF_SIZE);
		goto auth_exit;
	}*/

	ret = _parse_user_info((char *)p->tc_cmd.data, &pUsername, &pPassword);
	if(ret != 0){
		ERR("failed to parse user info %s\n", p->tc_cmd.data);
		goto auth_exit;
	}

	ret = get_user_oslist(pUsername, pPassword, 
			&oslist, &loginLeftTimes, s_config->login_failure_interval);
	if(ret != 0) {
		DEBUG("failed to get oslist for %s:%d\n", p->tc_cmd.data, ret);
		if(ret == -2)
			p->hdr.error = TC_ERR_NO_IMAGE;
		else if(ret == -3)
			p->hdr.error = TC_ERR_USER_BLOCKED;
		else if(ret == -4)
			p->hdr.cmd = (__uint8_t)loginLeftTimes;
		else if(ret == -5)
			p->hdr.error = TC_ERR_MAXIUM_FAILURE;
		
		goto auth_exit;
	} else {
		if(!oslist) {
			ret = -1;
			DEBUG("Error: oslist is NULL\n");
			goto auth_exit;
		}
		//DEBUG("os list for %s: %s\n", pUsername, oslist);
		p->tc_cmd.len = strlen(oslist) + 1;

		// No need to set the last char to 0 since the src is shorter than dest
		strncpy((char *)p->tc_cmd.data, oslist, p->tc_cmd.len);
		
		//strncpy(client->username, pUsername, MAX_STRING_LEN);
		//client->username[MAX_STRING_LEN - 1] = '\0';
		if(client->username) {
			free(client->username);
		}
		client->username = pUsername;
		
		DEBUG("Auth OK, client was setup for %s(tag: %d)\n", client->username, client->usedIdx + 1);
		client->boot_timeout = BOOT_TIMEOUT;

		SET_TAG(p->hdr.tag, client->usedIdx);
		//MSG("Auth hdr:tag is %d-%d-%d-%d\n",p->hdr.tag[0],p->hdr.tag[1],p->hdr.tag[2],p->hdr.tag[3]);
	}

	p->hdr.error = 0;
	ret = 0;
    
auth_exit:
	if(oslist)
		free(oslist);
	if(pPassword)
		free(pPassword);

	if(ret != 0) {
		p->hdr.ver_flags |= TC_FL_ERROR;
		strncpy(p->tc_cmd.data, "", 2);
		p->tc_cmd.data[1] = '\0';
		p->tc_cmd.len = strlen(p->tc_cmd.data) + 1;

		if(pUsername) {
			if(client->username == pUsername)
				client->username = NULL;
			
			free(pUsername);
		}
	}
	return ret;
}

int get_user_id_conf(Client *client, tc_pkt_t *p)
{		
	uint8_t old_error;
	int ret;
	char userIdStr[MAX_STRING_LEN] = {0};	
	int userId;

	old_error = p->hdr.error;
	p->hdr.error = TC_ERR_CONFIG_ERR;

	if(NOT_IN_USE(client) || !client->username || client->username[0] == 0){
		ERR("client is not connected\n");
		p->hdr.error = TC_ERR_NOT_LOGINED;
		goto error;
	}
	
	userId = get_user_id(client->username);

	if(userId < 0) {
		ERR("Failed to find userId by id %s\n", client->username);
		p->hdr.error = TC_ERR_DATABASE;
		goto error;
	}

	//DEBUG("user id is: %d\n", userId);

	snprintf(userIdStr, MAX_STRING_LEN, "%d", userId);

	p->tc_cmd.len = strlen(userIdStr) + 1;
	strncpy(p->tc_cmd.data, userIdStr, p->tc_cmd.len);
	//p->tc_cmd.data[p->tc_cmd.len - 1] = '\0';
	
	p->hdr.error = old_error;
	
	return 0;

error:	
	p->hdr.ver_flags |= TC_FL_ERROR;
	return -1;
}

int update_hb(Client *client, tc_pkt_t *p)
{	
	if(NOT_IN_USE(client) || !client->username || client->username[0] == 0){
		ERR("client is not connected\n");
		if(p)
			p->hdr.error = TC_ERR_NOT_LOGINED;
		goto error;
	}

	client->hb_count++;
	//DEBUG("HB count for user %s is 0x%x\n", client->username, client->hb_count);

	return 0;

error:
	if(p) {
		p->hdr.ver_flags |= TC_FL_ERROR;
	}
	return -1;  
}

int reset_conf_handler(int idx, char need_free_client)
{
	Client *client;

	if(idx < 0) {
		ERR("Error: idx(%d) <0 in reset_conf_handler\n", idx);
		return -1;
	}

	client = g_running_clients+idx;
	if(NOT_IN_USE(client) || !client->username || client->username[0] == 0){
		ERR("Error: client is not connected in reset_conf_handler\n");
		goto error;
	}	
	if(!is_autoboot_user(client->username))
		delete_user_status(client->username, client->image_id);

	insert_user_login_log(client->userid, client->username, "logout", 0);

	MSG("user %s end access %d\n", client->username, client->image_id);

	if(need_free_client)
		free_client(client);
	return 0;

error:	
	return -1;
}

void reset_conf(Client *client, tc_pkt_t *p)
{  
  if(NOT_IN_USE(client) || !client->username || client->username[0] == 0){
		ERR("client is not connected\n");
		p->hdr.error = TC_ERR_NOT_LOGINED;
		goto error;
  }
  client->status = SHUTTING_DOWN;
  //set_timer(client->usedIdx, RESET_CONF_INTERVAL, reset_conf_handler, 
  //			client->usedIdx, 0);
  //DEBUG("Timer set %d\n", client->usedIdx);
  //if(!is_autoboot_user(client->username))
  reset_conf_handler(client->usedIdx, FALSE);
error:  
  return;  
}

int _check_upload_conditions(Client *client, tc_pkt_t *p)
{  
	int ret, fnc_ret;
	long clientOsVer, curVerOnServer, verDiff;
	char *pOsId = NULL;
	char *pClientOsVerStr = NULL;
	char curVerOnServerStr[MAX_STRING_LEN] = {0};
	IMAGE_ACCESS imgacc;

	ret = TC_ERR_CONFIG_ERR;

	if(NOT_IN_USE(client) || !client->username || client->username[0] == 0){
		ERR("client is not connected\n");
		ret = TC_ERR_NOT_LOGINED;
		goto check_upload_cond_exit;
	}

  	fnc_ret = _parse_os_info((char *)p->tc_cmd.data, &pOsId, &pClientOsVerStr);
	if((fnc_ret != 0) || !pClientOsVerStr){
		ERR("Failed to parse os info %s\n", p->tc_cmd.data);
		goto check_upload_cond_exit;
	}

	if(str_to_long(pClientOsVerStr, &clientOsVer) == -1) {
		ERR("Failed to format client os version\n"); 
		goto check_upload_cond_exit;			
	}
	if(clientOsVer < 0)
	{
		ERR("Invalid client os version: %ld\n", clientOsVer);
		goto check_upload_cond_exit;
	}

	fnc_ret = is_image_locked(pOsId);
	if(fnc_ret != 0){
		ERR("Error: the image %s was being locked!\n", pOsId);
		ret = TC_ERR_IMAGE_LOCKED;
		goto check_upload_cond_exit;
	}
	
	fnc_ret = get_image_cur_ver(0, pOsId, curVerOnServerStr);
	if(fnc_ret != 0){
		ERR("Failed to get image version info of %s\n", pOsId);
		ret = TC_ERR_GET_VERSION;
		goto check_upload_cond_exit;
	} 

	if(str_to_long(curVerOnServerStr, &curVerOnServer) == -1) {
		ERR("Failed to format current version on server\n"); 
		ret = TC_ERR_GET_VERSION;
		goto check_upload_cond_exit;
	}
	if(curVerOnServer < 0) {
		ERR("Invalid os version of %s on server: %ld\n", pOsId, curVerOnServer); 
		ret = TC_ERR_GET_VERSION;
		goto check_upload_cond_exit;
	}

	verDiff = curVerOnServer - clientOsVer;

	if(verDiff != 0) {
		ERR("Error: version on server(%ld) differ from client(%ld)\n",curVerOnServer, clientOsVer);
		ret = TC_ERR_BAD_VERSION;
		goto check_upload_cond_exit;
	}

	fnc_ret = get_imageaccess(pOsId, client->username, &imgacc);
	if(fnc_ret != 0){
		ERR("Failed to get image access for user %s:%s\n", client->username, pOsId);
		ret = TC_ERR_DATABASE;
		goto check_upload_cond_exit;
	}

	if((imgacc != IMAGE_INSTALL_RWRITE) && (imgacc != IMAGE_DIFF_READWRITE)) {
		DEBUG("Permission denied for user %s:%s\n", client->username, pOsId);
		ret = TC_ERR_PERMISSION;
		goto check_upload_cond_exit;
	}	
	
	ret = 0;

check_upload_cond_exit:

	if(pOsId)
		free(pOsId);

	if(pClientOsVerStr)
		free(pClientOsVerStr);
	
	return ret;   
}

int _check_upload_data_conditions(Client *client, tc_pkt_t *p)
{  
	int ret, fnc_ret;
	long clientOsVer, curVerOnServer, verDiff;
	char *pOsId = NULL;
	char *pClientOsVerStr = NULL;
	char curVerOnServerStr[MAX_STRING_LEN] = {0};
	int userId;

	ret = TC_ERR_CONFIG_ERR;

	if(NOT_IN_USE(client) || !client->username || client->username[0] == 0){
		ERR("client is not connected\n");
		ret = TC_ERR_NOT_LOGINED;
		goto check_upload_data_cond_exit;
	}

  	fnc_ret = _parse_os_info((char *)p->tc_cmd.data, &pOsId, &pClientOsVerStr);
	if((fnc_ret != 0) || !pClientOsVerStr){
		ERR("Failed to parse os info %s\n", p->tc_cmd.data);
		goto check_upload_data_cond_exit;
	}

	if(str_to_long(pClientOsVerStr, &clientOsVer) == -1) {
		ERR("Failed to format client os version\n"); 
		goto check_upload_data_cond_exit;			
	}
	if(clientOsVer < 0)
	{
		ERR("Invalid client os version: %ld\n", clientOsVer);
		goto check_upload_data_cond_exit;
	}

	userId = get_user_id(client->username);
	
	if(userId < 0) {
		ERR("Failed to find userId by id %s\n", client->username);
		ret = TC_ERR_DATABASE;
		goto check_upload_data_cond_exit;
	}

	fnc_ret = is_data_image_locked(pOsId, userId);
	if(fnc_ret != 0){
		ERR("Error: the data image %s-%d was being locked!\n", 
			pOsId,
			userId);
		ret = TC_ERR_IMAGE_LOCKED;
		goto check_upload_data_cond_exit;
	}

	fnc_ret = get_data_image_cur_ver(userId, pOsId, curVerOnServerStr);
	if(fnc_ret != 0){
		ERR("Failed to get image version info of %s\n", pOsId);
		ret = TC_ERR_GET_VERSION;
		goto check_upload_data_cond_exit;
	} 

	if(str_to_long(curVerOnServerStr, &curVerOnServer) == -1) {
		ERR("Failed to format current version on server\n"); 
		ret = TC_ERR_GET_VERSION;
		goto check_upload_data_cond_exit;
	}
	if(curVerOnServer < 0) {
		ERR("Invalid os version of %s on server: %ld\n", 
			pOsId,
			curVerOnServer); 
		ret = TC_ERR_GET_VERSION;
		goto check_upload_data_cond_exit;
	}

	verDiff = curVerOnServer - clientOsVer;

	if(verDiff != 0) {
		ERR("Error: version on server(%ld) differ from client(%ld)\n", 
			curVerOnServer, 
			clientOsVer);
		ret = TC_ERR_BAD_VERSION;
		goto check_upload_data_cond_exit;
	}
	
	ret = 0;

check_upload_data_cond_exit:

	if(pOsId)
		free(pOsId);

	if(pClientOsVerStr)
		free(pClientOsVerStr);

	return ret;
}

int upload_check_conf(
		Client *client, 
		tc_pkt_t *p, 
		P_CHECK_UPLOAD_CONDITIONS p_check_upload_conditions)
{  
	int ret;
	char *outputStr = "";

	ret = p_check_upload_conditions(client, p);
	if(ret != 0) {
		p->hdr.error = ret;
		goto error;
	}

	p->tc_cmd.len = strlen(outputStr) + 1;
	strncpy(p->tc_cmd.data, outputStr, p->tc_cmd.len);
	//p->tc_cmd.data[p->tc_cmd.len - 1] = '\0';
	
	return 0;

error:	
	p->hdr.ver_flags |= TC_FL_ERROR;
	return -1;   
}

int upload_conf(Client *client, tc_pkt_t *p)
{  
	int ret;
	char *pOsId = NULL;
	char *pClientOsVerStr = NULL;
	char *outputStr = "";
	char clientCmdDataStr[DB_MAX_STRING_LEN] = {0};
	IMAGE_ACCESS imgacc;

	snprintf(clientCmdDataStr, 
		DB_MAX_STRING_LEN,
		"%s", 
		(char *)p->tc_cmd.data);
	
	ret = _check_upload_conditions(client, p);
	if(ret != 0) {
		p->hdr.error = ret;
		goto upload_conf_exit;
	}

	ret = _parse_os_info(clientCmdDataStr, &pOsId, &pClientOsVerStr);
	if(ret != 0){
		ERR("Failed to parse os info %s\n", p->tc_cmd.data);
		p->hdr.error = TC_ERR_CONFIG_ERR;
		goto upload_conf_exit;
	}

	ret = get_imageaccess(pOsId, client->username, &imgacc);
	if(ret != 0){
		ERR("Failed to get image access for user %s:%s\n", 
			client->username,
			pOsId);
		p->hdr.error = TC_ERR_DATABASE;
		goto upload_conf_exit;
	}

	if(imgacc == IMAGE_INSTALL_RWRITE) {
		ret = update_image_status_by_id(pOsId, 1);
		if(ret != 0){
			DEBUG("failed to update image status: %s\n", pOsId);
			p->hdr.error = TC_ERR_DATABASE;
			goto upload_conf_exit;
		}
	}

	p->tc_cmd.len = strlen(outputStr) + 1;
	strncpy(p->tc_cmd.data, outputStr, p->tc_cmd.len);
	
	ret = 0;

upload_conf_exit:	

	if(ret != 0) {
		p->hdr.ver_flags |= TC_FL_ERROR;
	}

	if(pOsId)
		free(pOsId);

	if(pClientOsVerStr)
		free(pClientOsVerStr);
	
	return -1;   
}

int upload_data_conf(Client *client, tc_pkt_t *p)
{  
	int ret;
	char *outputStr = "";
	
	ret = _check_upload_data_conditions(client, p);
	if(ret != 0) {
		p->hdr.error = ret;
		goto upload_data_conf_exit;
	}

	p->tc_cmd.len = strlen(outputStr) + 1;
	strncpy(p->tc_cmd.data, outputStr, p->tc_cmd.len);	
	
	ret = 0;

upload_data_conf_exit:
	
	if(ret != 0) {
		p->hdr.ver_flags |= TC_FL_ERROR;
	}
	return -1;   
}

int get_update_list_by_ver(long clientOsVer, 
	long curVerOnServer, 
	int userId,
	char *osId,
	char *outlist
	)
{
	int ret = -1;
	long verDiff;
	char *maxVerKeptStr = NULL;	
	long maxVerKept;
	
	verDiff = curVerOnServer - clientOsVer;

	//Get max version kept num
	maxVerKeptStr = get_default_server_config(TC_SERVER_ID, "maxVerKept");
	if(maxVerKeptStr == NULL)
		maxVerKept = DFT_MAX_VER_KEPT;
	else{
		if(str_to_long(maxVerKeptStr, &maxVerKept) == -1) {
			ERR("Failed to format max versions kept num\n"); 
			ret = -1;
			goto error;			
		}
		if(maxVerKept < 0)
		{
			ERR("Invalid max version kept: %ld\n", maxVerKept); 
			ret = -1;
			goto error;
		}
	}

GetUpdateListMain:	
	//we just send "curVerOnServer" only to tell client to sync the image directly under following cases:
	//1. if the client os version was 0, which means there is no image on client side
	//2. if the version diff between client and server was larger than max version kept on server	
	//3(removed). if the version on client was bigger/newer than that on server
	if((clientOsVer == 0) ||(verDiff > maxVerKept))// || verDiff < 0))	
		snprintf(outlist, MAX_STRING_LEN, "%ld%s%s", curVerOnServer, 
						INFO_SEP_STR, INFO_SEP_STR);
	else {
		ret = get_image_update_list(userId, osId, clientOsVer, 
					curVerOnServer, outlist);
		// ret == 1 means some ver lists were missing/deleted in db
		// so we need to download the whole image instead
		if(ret == 1) {
			verDiff = maxVerKept + 1;
			goto GetUpdateListMain;
		} else if(ret != 0) {
			ERR("Failed to get image udpate list\n");
			ret = -1;
			goto error;
		}
	}
	
	ret = 0;

error:
	if(maxVerKeptStr)
		free(maxVerKeptStr);

	return ret;
}

int get_data_update_list_by_ver(long clientOsVer, 
	long curVerOnServer,
	int userId,
	char *osId,
	char *outlist
	)
{
	int ret = -1;
	long verDiff;
	char *maxVerKeptStr = NULL;	 
	long maxVerKept;
	
	verDiff = curVerOnServer - clientOsVer;

	//Get max version kept num
	maxVerKeptStr = get_default_server_config(TC_SERVER_ID, "maxDataVerKept");
	if(maxVerKeptStr == NULL)
		maxVerKept = DFT_MAX_DATA_VER_KEPT;
	else{
		if(str_to_long(maxVerKeptStr, &maxVerKept) == -1) {
			ERR("Failed to format max versions kept num\n"); 
			ret = -1;
			goto error;			
		}
		if(maxVerKept < 0)
		{
			ERR("Invalid max version kept: %ld\n", maxVerKept); 
			ret = -1;
			goto error;
		}
	}

GetDataUpdateListMain:	
	//we just send "curVerOnServer" only to tell client to sync the image directly under following cases:
	//1. if the client os version was 0, which means there is no image on client side
	//2. if the version diff between client and server was larger than max version kept on server	
	//3(removed). if the version on client was bigger/newer than that on server
	if((clientOsVer == 0) ||(verDiff > maxVerKept)) { // || verDiff < 0))	{
		char image_file[MAX_STRING_LEN] = {0};
		char image_size[MAX_STRING_LEN] = {0};
		ret = get_data_image_size(userId, image_file, 
					image_size, sizeof(image_size)/sizeof(char));
		if(ret != 0)
			snprintf(outlist, MAX_STRING_LEN, "%ld%s%s%s%s", curVerOnServer, 
							INFO_SEP_STR, INFO_SEP_STR, 
							INFO_SEP_STR, INFO_SEP_STR);
		else
			snprintf(outlist, MAX_STRING_LEN, "%ld%s%s%s%s%s%s", curVerOnServer, 
							INFO_SEP_STR, INFO_SEP_STR,
							INFO_SEP_STR, image_file,
							INFO_SEP_STR, image_size);
	}
	else {
		ret = get_data_image_update_list(userId, osId, clientOsVer, 
				curVerOnServer, outlist);
		if(ret == 1) {
			verDiff = maxVerKept + 1;
			goto GetDataUpdateListMain;
		} else if(ret != 0) {
			ERR("Failed to get image udpate list\n");
			ret = -1;
			goto error;
		}
	}

	ret = 0;

error:
	if(maxVerKeptStr)
		free(maxVerKeptStr);

	return ret;
}

int sync_and_get_update_list(Client *client, 
	tc_pkt_t *p,
	char needSetClient,
	P_GET_IAMGE_CUR_VER pGetImageCurVer,
	P_GET_IAMGE_ID_BY_ID pGetImageIdById,
	P_GET_UPDATE_LIST_BY_VER pGetUpdateListByVer)
{
	uint8_t old_error;
	int ret = -1;
	long clientOsVer, curVerOnServer;
	char curVerOnServerStr[MAX_STRING_LEN] = {0};
	char *reply = NULL;
	char *pOsId = NULL;
	char *pClientOsVerStr = NULL;
	int userId;

	old_error = p->hdr.error;
	p->hdr.error = TC_ERR_CONFIG_ERR;

	if(NOT_IN_USE(client) || !client->username || client->username[0] == 0){
		ERR("client is not connected\n");
		p->hdr.error = TC_ERR_NOT_LOGINED;
		goto sync_and_get_update_exit;
	}

	ret = _parse_os_info((char *)p->tc_cmd.data, &pOsId, &pClientOsVerStr);
	if((ret != 0) || !pClientOsVerStr){
		ERR("Failed to parse os info %s\n", p->tc_cmd.data);
		goto sync_and_get_update_exit;
	}

	ret = str_to_long(pClientOsVerStr, &clientOsVer);
	if(ret == -1) {
		ERR("Failed to format client os version\n");
		goto sync_and_get_update_exit;			
	}
	if(clientOsVer < 0)
	{
		ERR("Invalid client os version: %ld\n", clientOsVer);
		ret = -1;
		goto sync_and_get_update_exit;
	}
	
	//BUGBUG: dummy function here
	//osimageid = pGetImageIdById(osId);
	/*
	if(str_to_long(osId, &osimageid) == -1) {
		ERR("Failed to format image id\n"); 
		goto error;			
	}
	if(osimageid < 0) {
		ERR("Failed to find osimageid by id %s\n", osId);
		p->hdr.error = TC_ERR_DATABASE;
		goto error;
	}*/

	userId = get_user_id(client->username);

	if(userId < 0) {
		ERR("Failed to find userId by id %s\n", client->username);
		p->hdr.error = TC_ERR_DATABASE;
		ret = -1;
		goto sync_and_get_update_exit;
	}

	ret = pGetImageCurVer(userId, pOsId, curVerOnServerStr);
	if(ret != 0){
		ERR("Failed to get image version info\n");
		p->hdr.error = TC_ERR_GET_VERSION;
		goto sync_and_get_update_exit;
	} 

	ret = str_to_long(curVerOnServerStr, &curVerOnServer);
	if(ret == -1) {
		ERR("Failed to format current version on server\n"); 
		p->hdr.error = TC_ERR_GET_VERSION;
		goto sync_and_get_update_exit;
	}
	if(curVerOnServer < 0) {
		ERR("Invalid os version on server: %ld\n", curVerOnServer); 
		p->hdr.error = TC_ERR_GET_VERSION;
		ret = -1;
		goto sync_and_get_update_exit;
	}

	reply = (char *)calloc(MAX_BUF_SIZE, sizeof(char));

	if(!reply) {
		ERR("Failed to allocate image udpate list\n");
		p->hdr.error = TC_ERR_CONFIG_ERR;
		ret = -1;
		goto sync_and_get_update_exit;
	}				

	ret = pGetUpdateListByVer(clientOsVer,
			curVerOnServer,
			userId,
			pOsId,
			reply);

	if(ret != 0) {
		ERR("Failed to get image udpate list\n");
		p->hdr.error = TC_ERR_DATABASE;
		goto sync_and_get_update_exit;
	}

	//DEBUG("OS image update list is: %s\n", reply);

	if(needSetClient) {
		long osimageid;
		
	    memset(p->tc_cmd.data, 0x0, MAX_STRING_LEN);
		ret = get_osimagename_by_id(pOsId, p->tc_cmd.data, MAX_STRING_LEN);
	    if(ret != 0)
	        goto sync_and_get_update_exit;

		// we should set the os first
		ret = set_os_conf(client, p);
		if(ret != 0)
			goto sync_and_get_update_exit;

		ret = str_to_long(pOsId, &osimageid);
		if(ret == -1) {
			ERR("Failed to format image id\n"); 
			goto sync_and_get_update_exit;
		}
		
		if(osimageid < 0) {
			ERR("Failed to find osimageid by id %s\n", pOsId);
			p->hdr.error = TC_ERR_DATABASE;
			ret = -1;
			goto sync_and_get_update_exit;
		}

		ret = update_user_status(client->username, osimageid, SYNCING); 
		if(ret != 0){
			ERR("failed to update online sync status\n");
			p->hdr.error = TC_ERR_DATABASE;
			goto sync_and_get_update_exit;
		} 

		MSG("user %s update online sync status on image %d\n", client->username, 
				client->image_id);
	}

	p->tc_cmd.len = strlen(reply) + 1;
	strncpy(p->tc_cmd.data, reply, p->tc_cmd.len);	
	
	p->hdr.error = old_error;
	
	ret = 0;

sync_and_get_update_exit:
	
	if(ret != 0) {
		p->hdr.ver_flags |= TC_FL_ERROR;
	}
	if(reply)
		free(reply);

	if(pOsId)
		free(pOsId);

	if(pClientOsVerStr)
		free(pClientOsVerStr);
	
	return ret;  
}

int sync_conf(Client *client, tc_pkt_t *p)
{
	return sync_and_get_update_list(client, 
				p, 
				1, 
				get_image_cur_ver,
				get_osimageid_by_id,
				get_update_list_by_ver);
}

int sync_data_conf(Client *client, tc_pkt_t *p)
{
	return sync_and_get_update_list(client, 
				p, 
				0, 
				get_data_image_cur_ver,
				get_osimageid_by_id,
				get_data_update_list_by_ver);
}


