/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <sys/stat.h>
#include <errno.h>

#define MAX_STRING_LEN 		256
#define ENABLE_SYSLOG		0
#define UDF_PAGE_SIZE  		(2*1024*1024)
#define TC_MDF_FORMAT		"%s%s-%s.mdf"
#define TC_UDF_FORMAT		"%s%s-%s.udf"
#define TC_CDF_FORMAT		"%s%s-%s.cdf"


#if ENABLE_SYSLOG
#define ERR(fmt,arg...) 	do {\
			    				syslog(LOG_ERR,fmt,##arg);\
	                        	printf(fmt,##arg);\
	                     	} while(0)

#define DEBUG(fmt,arg...)	syslog(LOG_ERR, fmt,##arg) 
#else                       
#define ERR(fmt,arg...)     do {\
	                        	printf(fmt,##arg);\
	                        } while(0) 

#define DEBUG(fmt,arg...)   do {\
	                            printf(fmt,##arg);\
	                        } while(0)                               
#endif

static int g_udf_page_size = UDF_PAGE_SIZE;
char* get_filename(char* str)
{
	int i, slen;
	int pos = -1;	
	char* tmp_char;

	if(str == NULL)
		return NULL;

	slen = strlen(str);
	for(i=0;i < slen;i++)
	{
		if(str[i] == '/')
		{
			pos = i;
		}
	}

	tmp_char = str + pos + 1;
	return tmp_char;
}

char *get_cur_dir()
{
	char *cur_dir;
	cur_dir = (char *)malloc(MAX_STRING_LEN);
	if(!cur_dir)
		return NULL;
	getcwd(cur_dir, MAX_STRING_LEN);
	//DEBUG("cur dir is :%s\n",cur_dir);
	return cur_dir;
}

off_t get_file_size(int fd)
{
	off_t size;
	struct stat fs_stat;
	int ret;

	if (ioctl(fd, BLKGETSIZE64, &size) == -1) {
		ret = fstat(fd, &fs_stat);
		if (ret == -1) {
			ERR("getsize error %d\n", errno);
			exit(1);
		}
		size = fs_stat.st_size;
	}

	return size;
}

int createMdfUdf(char *imageFilePath, 
	char *username, 
	char *pOutDir, 
	char isCdfOnly)
{  
	uint64_t i, total_pg_nr, udf_pg_nr; 
	uint64_t raw_file_size, mdf_len;
	uint32_t  *mdf_cache = NULL;
	char  *raw_image;
	char  md_file_path[MAX_STRING_LEN] = {0};
	char  ud_file_path[MAX_STRING_LEN] = {0};
	char  out_dir_path[MAX_STRING_LEN] = {0};
	int raw_fd = -1, udf_fd = -1, mdf_fd = -1;
	int outDirLen = -1;
	int ret = -1;

	raw_fd = open(imageFilePath, O_RDONLY);
	if (raw_fd < 0) {
		ret = errno;
		ERR("Failed to open image file \"%s\": %s\n", imageFilePath, strerror(errno));
		goto COW_ERROR;
	}

	raw_file_size = get_file_size(raw_fd);
	raw_image = get_filename(imageFilePath);
	if(!raw_image) {
		ret = -1;
		ERR("Failed to get image filename \"%s\"\n", imageFilePath);
		goto COW_ERROR;
	}

	//DEBUG("exportfile = %s\n", raw_image);

	if(pOutDir){
		//the length should not exceed the MAX_STRING_LEN
		strncpy(out_dir_path, pOutDir, MAX_STRING_LEN);
		out_dir_path[MAX_STRING_LEN - 1] = '\0';
		
		outDirLen = strlen(out_dir_path);
		if(outDirLen >= (MAX_STRING_LEN - 1)) {
			outDirLen = MAX_STRING_LEN - 2;
		}
		
		if(pOutDir[outDirLen - 1] != '/') {
			out_dir_path[outDirLen] = '/';
			out_dir_path[outDirLen+1] = 0;
		}
	}

	if(isCdfOnly) {
		snprintf(md_file_path, MAX_STRING_LEN, TC_CDF_FORMAT, out_dir_path, raw_image, username);
		DEBUG("About to create cdf file %s\n", md_file_path) ;
	} else {
		snprintf(md_file_path, MAX_STRING_LEN, TC_MDF_FORMAT, out_dir_path, raw_image, username);
		DEBUG("About to create mdf file %s\n", md_file_path) ;
	}

	mdf_fd = open(md_file_path, O_CREAT|O_RDWR|O_TRUNC, 0666);
	if (mdf_fd < 0) {
	  ret = errno;
	  if(isCdfOnly)
		ERR("Failed to create new cdf file (%s)", strerror(errno));
	  else
	  	ERR("Failed to create new mdf file (%s)", strerror(errno));
	  goto COW_ERROR;
	} 

	if(!isCdfOnly) {
		snprintf(ud_file_path, MAX_STRING_LEN, TC_UDF_FORMAT, out_dir_path, raw_image, username);
		DEBUG("About to create udf file %s", ud_file_path);

		udf_fd = open(ud_file_path, O_CREAT|O_RDWR|O_TRUNC, 0666);
		if (udf_fd < 0) {
			ret = errno;
			ERR("Could not open udf file (%s)", strerror(errno));
			goto COW_ERROR;
		}
	}

	udf_pg_nr = 0;

	total_pg_nr = (raw_file_size / g_udf_page_size) + 1;
	DEBUG("Udf file pages is 0x%llx\n", total_pg_nr);

	mdf_cache = malloc(total_pg_nr * sizeof(mdf_cache[0]));
	if (mdf_cache == NULL) {
		ret = errno;
		ERR("Could not allocate memory (%s)", strerror(errno)) ;
		goto COW_ERROR;
	}

	for (i=0; i<total_pg_nr; i++) 
		mdf_cache[i] = -1 ;

	write(mdf_fd, mdf_cache, sizeof(mdf_cache[0]) * total_pg_nr);
	//this is for sync client only since the client will not create udf file automatically
	//write(udf_fd, mdf_cache, sizeof(mdf_cache[0]));


	mdf_len = total_pg_nr * sizeof(mdf_cache[0]);

	if(isCdfOnly) {
		DEBUG("cdf_path = %s, cdf_len = 0x%llx\n", md_file_path, mdf_len);
	} else {
		DEBUG("udf_path = %s, udf_pg = 0x%llx\n", ud_file_path, udf_pg_nr);
		DEBUG("mdf_path = %s, mdf_len = 0x%llx\n", md_file_path, mdf_len);
	}

	ret = 0;

COW_ERROR:
	if(raw_fd >= 0)
		close(raw_fd);
	if(mdf_fd >= 0)
		close(mdf_fd);
	if(udf_fd >= 0)
		close(udf_fd);

	if(mdf_cache)
		free(mdf_cache);

	return ret;

}

void usage(){
	printf("createMdfUdf [-c] imageFilePath username [AbsOutputPath]\n");
}

int main(int argc, char *argv[]) {	
	int ret;
	int index = 1;
	char *pOutPath = NULL;
	char needFree = 0;
	char IsCdfOnly = 0;

	if(argc < 3 || argc > 5)
		goto error;

	if(strcmp(argv[1], "-c") == 0){
		argc--;
		IsCdfOnly = 1;
		index++;
	}

	if(argc < 3 || argc > 4)
		goto error;
	else if(argc == 4) {
		pOutPath = argv[index+2];
		if(pOutPath[0] != '/') {
			printf("Error: output dir should be absolute path started from '/'\n");
			goto error;
		}
	} else {
		pOutPath = get_cur_dir();
		needFree = 1;
	}

	ret = createMdfUdf(argv[index], argv[index+1], pOutPath, IsCdfOnly);	
	if(needFree && pOutPath)
		free(pOutPath);

	return ret;
	
error:
	usage();
	return -1;
}
