/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <openssl/evp.h>
#include "predefine.h"
#include "db.h"
#include "debug.h"

#if  USE_MYSQL
#include <mysql/mysql.h>	//mysql header file
#include <mysql/mysqld_error.h>	//mysql header file


static MYSQL g_db_connection; 		//used in db connection
static int g_db_connected=0;		//db connection status
static int g_need_clean_status=1;
static char *g_db_username=NULL,*g_db_password=NULL;
static int g_max_login_failure = 5;

char* getFilename(char* str)
{
	int i;
	int pos = -1;	
	char* temp;
	size_t len;	

	if(!str)
		return " ";
	
	len = strlen(str);
	for(i=0; i<len; i++)
	{
		if(str[i] == '/')
		{
			pos = i;
		}
	}
			
	temp = str + pos + 1;
	return temp;
}

int escape_parameter(char **pEscapedStr,
	char *pSrcStr)
{
	int ret = -1;
	size_t srcStrLen;

	if(!pEscapedStr || !pSrcStr) {
		ERR("esms error: invalid parameter\n");
		return ret;
	}

	srcStrLen = strlen(pSrcStr);
	*pEscapedStr = malloc(srcStrLen*2 + 1);
	if(!(*pEscapedStr)) {
		ERR("esms error: failed to alloc mem for uname\n");
		goto escape_str_exit;
	}

	mysql_escape_string(*pEscapedStr, 
							pSrcStr, 
							srcStrLen);
	
	ret = 0;

escape_str_exit:

	if(ret < 0) {
		if(pEscapedStr && (*pEscapedStr))
			free(*pEscapedStr);
	}
	
	return ret;
}

static int safe_mysql_query(MYSQL *sql_conn,const char *query)
{
	int mysql_error_code,errno;
	int retry;
	
	retry=0;
	while(1)
	{
		mysql_error_code = mysql_query(sql_conn,query);
		errno = mysql_errno(sql_conn);		
		if(retry || errno!=2006)  //server has gone away
		   break;		
		
		DEBUG("sql error %d: %s\n", errno, mysql_error(sql_conn));
		DEBUG("try to restart db...\n");
		g_db_connected = 0;
		g_need_clean_status = 0;
		if(start_db(g_db_username,g_db_password)!=0){
			g_need_clean_status=1;
			break;
		}
		g_need_clean_status = 1;
		retry = 1;	
	}
	return mysql_error_code;
}

void reset_all_status()
{  
  char pQueryStr[MAX_STRING_LEN];

#if UPDATE_USR_STATUS_IN_DB == 0
	  return;
#else  
  if(!g_need_clean_status)
  	return;	

  //DEBUG("reset all status\n");
  snprintf(pQueryStr, MAX_STRING_LEN, "delete from userstatus");

  safe_mysql_query(&g_db_connection, pQueryStr);
  g_need_clean_status=0;  
#endif
}

int set_conn_timeout(unsigned int timeout)
{
	int ret = -1;
	char pQueryStr[MAX_STRING_LEN];
	
	snprintf(pQueryStr,
		sizeof(pQueryStr),
		"set wait_timeout=%d",
		timeout);	
	if(mysql_real_query(&g_db_connection, pQueryStr, strlen(pQueryStr))) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return ret;
	}

	snprintf(pQueryStr,
		sizeof(pQueryStr),
		"set interactive_timeout=%d",
		timeout);	
	if(mysql_real_query(&g_db_connection, pQueryStr, strlen(pQueryStr))) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return ret;
	}

	ret = 0;
	
	return ret;
}

int save_authen_info(char *name,char *pwd)
{
	int ret = -1;
	char *tmpname,*tmppwd;
	size_t namelen,pwdlen;
	
	if(!name || !pwd)
		return ret;
	
	tmpname = g_db_username;
	tmppwd = g_db_password;		
	
	namelen = strlen(name) + 1;
	pwdlen = strlen(pwd) + 1;

	g_db_username = (char *)calloc(namelen, sizeof(char));
	if(!g_db_username) {
		ERR("save auth error: failed to alloc mem for uname\n");
		return ret;
	}
	
	g_db_password = (char *)calloc(pwdlen, sizeof(char));
	if(!g_db_password) {
		ERR("save auth error: failed to alloc mem for p\n");
		free(g_db_username);
		return ret;
	}

	strncpy(g_db_username, name, namelen);
	g_db_username[namelen - 1] = '\0';
	
	strncpy(g_db_password, pwd, pwdlen);
	g_db_password[pwdlen - 1] = '\0';

	ret = 0;

	return ret;
}

/** 
 * There are some functions used to connect db and search infomation in db
**/
//start the database connection
int start_db(char *name, char *pwd)
{	
	int ret = -1;
	char *max_login_failure_str;
	
	if(g_db_connected) 
	  return 0;

	//DEBUG("init mysql\n");
	mysql_init(&g_db_connection);

	if(!mysql_real_connect(&g_db_connection,"127.0.0.1",name,pwd,"tc",0,NULL,0)){
		//DEBUG("db: name = %s , pwd = %s\n",name,pwd);
		DEBUG("Database connection failure: %d, %s\n",mysql_errno(&g_db_connection),mysql_error(&g_db_connection));
		return ret;
	}else{
		;//DEBUG("DB connection established\n");
	}	
	
	ret = set_conn_timeout(30*24*3600);
	if(ret != 0) {
		return ret;
	}
	
	ret = save_authen_info(name, pwd);
	if(ret != 0) {
		return ret;
	}

	max_login_failure_str = get_default_server_config(TC_SERVER_ID, 
							"max_login_failure");
	if(max_login_failure_str) {
		sscanf(max_login_failure_str, "%d", &g_max_login_failure);
		free(max_login_failure_str);
	}
	
	reset_all_status();

	g_db_connected = 1;

	ret = 0;
	
	return ret;
}

//close the database connection
void end_db()
{
	if(g_db_connected) 
		mysql_close(&g_db_connection);
		
	g_db_connected = 0;
	
	if(g_db_username)
		free(g_db_username);
		
	if(g_db_password)
		free(g_db_password);

	g_db_username = NULL;
	g_db_password = NULL;
}

int is_autoboot_user(char *username) 
{
  char pQueryStr[DB_MAX_STRING_LEN];
  int  res;
  MYSQL_RES *res_ptr = NULL;
  MYSQL_ROW mysqlrow;
  int isABUser = 0;

  //DEBUG("db: is_autoboot_user \n");
  
  snprintf(pQueryStr, DB_MAX_STRING_LEN, 
  			"select A.id, A.groupName from usergroups A left join users B "
  			  "on A.id=B.groupId WHERE B.username = '%s'", username);

  res = safe_mysql_query(&g_db_connection, pQueryStr); 
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return isABUser;
		
  }else{
	res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				if(strcmp(mysqlrow[1], AUTO_BOOT_GROUPNAME) == 0)
					isABUser = 1;
			} else {
		        DEBUG("autoboot: sql returns no result\n");
		        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		} 
		
	}

  return isABUser;
  
sql_error:
  //DEBUG("SQL:%s\n", pQueryStr);
  if(res_ptr)
 	mysql_free_result(res_ptr);
  return isABUser;
}


int get_user_status(char *username, int osimage_id) 
{
  char pQueryStr[DB_MAX_STRING_LEN];
  int  res;
  MYSQL_RES *res_ptr = NULL;
  MYSQL_ROW mysqlrow;
  int isOnline = 0;

  //DEBUG("db: get_user_status \n");
  
#if UPDATE_USR_STATUS_IN_DB == 0
	  return 0;
#else
  
  snprintf(pQueryStr, 
  	DB_MAX_STRING_LEN,
  	"select isOnline from userstatus, users WHERE "
  	"userstatus.userid=users.id and users.username = '%s' and osimageid=%d",
  	username, 
  	osimage_id);

  res = safe_mysql_query(&g_db_connection, pQueryStr); 
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return isOnline;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d",&isOnline);
				DEBUG("usersignin : %d\n",isOnline);
			} else {
        DEBUG("user status: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} 
		
	}

  return isOnline;
  
sql_error:
  //DEBUG("SQL:%s\n", pQueryStr);
  if(res_ptr)
  	mysql_free_result(res_ptr);
  return isOnline;
#endif  
}

int get_user_status_by_mac(char *username) 
{
  char pQueryStr[DB_MAX_STRING_LEN];
  int  res;
  MYSQL_RES *res_ptr = NULL;
	MYSQL_ROW mysqlrow;
	int isOnline = 0;

  //DEBUG("db: get_user_status \n");
  
  snprintf(pQueryStr, 
  	DB_MAX_STRING_LEN,
  	"select isOnline from userstatus, users WHERE userstatus.userid=users.id "
  	"and users.username = '%s'", 
  	username);

  res = safe_mysql_query(&g_db_connection, pQueryStr);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return isOnline;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d",&isOnline);
				DEBUG("usersignin : %d\n",isOnline);
			} else {
        DEBUG("user status(mac): sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} 
		
	}

  return isOnline;
  
sql_error:
  //DEBUG("SQL:%s\n", pQueryStr);
  if(res_ptr)
  	mysql_free_result(res_ptr);
  return isOnline;
  
}


int insert_user_status(char *username, int osimage_id, char *ip, 
	int imageaccess, int online)
{	
	char pQueryStr[DB_MAX_STRING_LEN];	
	char pSubQueryStr[MAX_STRING_LEN];
	char signtime[25];
	time_t timep;
	struct tm *pGmTime;
	int ret;

	//DEBUG("db: insert_user_status\n");

#if UPDATE_USR_STATUS_IN_DB == 0
	return 0;
#else
  	snprintf(pSubQueryStr,
  		MAX_STRING_LEN, 
  		"select id from users where username = '%s'", 
  		username);
  
	time(&timep);
	pGmTime = gmtime(&timep);
	snprintf(signtime,
		"%d-%d-%d %d:%d:%d",
		(1900 + pGmTime->tm_year),
		(1 + pGmTime->tm_mon),
		pGmTime->tm_mday,
		pGmTime->tm_hour,
		pGmTime->tm_min,
		pGmTime->tm_sec);
	snprintf(pQueryStr,
		DB_MAX_STRING_LEN,
		"insert into userstatus(userid,isOnline, osImageId,ip,osRunmodeId,lastSignInTime)" 
        "values((%s),'%d',%d,'%s',%d,'%s')", 
        pSubQueryStr,
        online,
        osimage_id,
        ip,
        imageaccess,
        signtime);
		

	//DEBUG("update sql:%s\n", pQueryStr);

	ret = safe_mysql_query(&g_db_connection, pQueryStr);
	if(ret != 0) {
		//DUP_ENTRY
		if(mysql_errno(&g_db_connection) == ER_DUP_ENTRY) {
			return ERR_DUP_ENTRY_IN_DB;
		}
	}
	
	return ret;
#endif
}

int delete_user_status(char *username, int osimage_id)
{
	char pQueryStr[DB_MAX_STRING_LEN]={0};	

  	//DEBUG("db: delete_user_status\n");

#if UPDATE_USR_STATUS_IN_DB == 0
	  return 0;
#else  
  	snprintf(pQueryStr, 
  		DB_MAX_STRING_LEN,
  		"delete from userstatus where userid = "
  		"(select id from users where username='%s') and osimageid ='%d'",
		username,
		osimage_id);
  
	//DEBUG("update sql:%s\n",pQueryStr);

	return safe_mysql_query(&g_db_connection, pQueryStr);
#endif
}

int update_user_status(char *username, int osimage_id, int status)
{
	char pQueryStr[DB_MAX_STRING_LEN];	

  //DEBUG("db: update_user_status\n");

#if UPDATE_USR_STATUS_IN_DB   
	snprintf(pQueryStr, 
		DB_MAX_STRING_LEN,
		"update userstatus set isOnline='%d' where userid = "
		"(select id from users where username='%s') and osimageid ='%d'",
		status,
		username,
		osimage_id);
  
	//DEBUG("update sql:%s",pQueryStr);
	return safe_mysql_query(&g_db_connection, pQueryStr);
#else
	return 0;
#endif		
}

//to update image path after merge the udf into the old image
int update_image_status_by_id(char* osid, int needupdate)
{
	char pQueryStr[DB_MAX_STRING_LEN];

  //DEBUG("db: update_image_status\n");

	snprintf(pQueryStr,
		DB_MAX_STRING_LEN,
		"update osimages set needUpdate='%d' WHERE id = '%s'", 
		needupdate, 
		osid);
//	DEBUG("update sql:%s",pQueryStr);

	return safe_mysql_query(&g_db_connection, pQueryStr);	
}



//to update image path after merge the udf into the old image
int update_image_status(char* rawfile, char* newfile, int needupdate)
{
	char pQueryStr[DB_MAX_STRING_LEN];	
	char updatetime[25];
	time_t timep;
	struct tm *pGmTime;

  //DEBUG("db: update_image_status\n");
  
	if(newfile != NULL && needupdate == 0) {
  		time(&timep);
		pGmTime = gmtime(&timep);
		if(!pGmTime) {
			return -1;
		}
		snprintf(updatetime,
			sizeof(updatetime),
			"%d-%d-%d %d:%d:%d",
			(1900 + pGmTime->tm_year),
			(1 + pGmTime->tm_mon),
			pGmTime->tm_mday,
			pGmTime->tm_hour,
			pGmTime->tm_min,
			pGmTime->tm_sec);
		
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"update osimages set path='%s',lastUpdateTime='%s',"
			"needUpdate='0' WHERE path = '%s'",
			newfile,
			updatetime,
			rawfile);
	}	else if (newfile == NULL && needupdate){
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"update osimages set needUpdate='%d' WHERE path = '%s'", 
			needupdate, 
			rawfile);
	} else {
		return 0;
	}
  
	//DEBUG("update sql:%s",pQueryStr);

	return safe_mysql_query(&g_db_connection, pQueryStr);	
}

int update_server_status(int server_id,int status)
{
	char pQueryStr[DB_MAX_STRING_LEN];	
	char signtime[25];
	time_t timep;
	struct tm *pGmTime;

#if UPDATE_SRV_STATUS_IN_DB == 0
	return 0;
#else
	switch(status) {
	case SERVER_STATUS_ONLINE:
		time(&timep);
		pGmTime = gmtime(&timep);
		snprintf(signtime,
			sizeof(signtime),
			"%d-%d-%d %d:%d:%d",
			(1900 + pGmTime->tm_year),
			(1 + pGmTime->tm_mon),
			pGmTime->tm_mday,
			pGmTime->tm_hour,
			pGmTime->tm_min,
			pGmTime->tm_sec);
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"update servers set status='%d',lastPowerOnTime='%s' WHERE id = '%d'",
			status,
			signtime,
			server_id);
	  break;

	case SERVER_STATUS_OVERLOAD:
	  snprintf(pQueryStr,
	  	DB_MAX_STRING_LEN,
	  	"update servers, serverconfigs set servers.status='%d', "
	  	"serverconfigs.configvalue='%d'" 
         "WHERE servers.id=serverconfigs.serverid and "
         "serverconfigs.configname = 'maxClientNum' and servers.id=%d",
         status,
         MAX_CLIENTS,
         server_id);
	  break;
	  
	case SERVER_STATUS_NORMAL:
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"update servers set status='%d' WHERE id = '%d'",
			SERVER_STATUS_ONLINE,
			server_id);
		break;
		
	case SERVER_STATUS_OFFLINE:
	  snprintf(pQueryStr,
	  	DB_MAX_STRING_LEN,
	  	"update servers set status='%d' WHERE id = '%d'",
	  	SERVER_STATUS_OFFLINE,
	  	server_id);
	  break;
	  
	}

	//DEBUG("update sql:%s\n",pQueryStr);

	return safe_mysql_query(&g_db_connection, pQueryStr);	
#endif
}

int get_user_id_and_lf_times_and_salt(char *username,
	int *user_id,
	int *login_failure_times,
	unsigned char *salt)
{
	MYSQL_RES *res_ptr = NULL;
	//	MYSQL_RES *res_ptr1;
	MYSQL_ROW mysqlrow;
	int res, ret = -1;
	char pQueryStr[DB_MAX_STRING_LEN];

	if(!username){
		ERR("get_user_id: username is NULL!\n");
		return ret;
	}

	if(!user_id && !login_failure_times && !salt){
		ERR("get_user_id: all params is NULL!\n");
		return ret;
	}
	
	snprintf(pQueryStr,
		DB_MAX_STRING_LEN,
		"SELECT id,login_failure,salt FROM users WHERE userName = '%s' limit 1",
		username);
	res = safe_mysql_query(&g_db_connection, pQueryStr);
	if(res){
		ERR("get_ulf_s->select error %d: %s\n", res, mysql_error(&g_db_connection));
		return ret;
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				while((mysqlrow = mysql_fetch_row(res_ptr))){
					if(user_id)
						sscanf(mysqlrow[0],"%d",user_id);
					if(login_failure_times)
						sscanf(mysqlrow[1],"%d",login_failure_times);
					if(salt) {
						memcpy(salt, 
							mysqlrow[2], 
							//mysql_field_len(res_ptr, 2));
							SALT_LEN);
					}
					break;
					//MSG("user id : %d\n",userId);					
				}
			} else {
				DEBUG("get_ulf_s: sql returns no result\n");
				goto sql_error;
			}

			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} else {
			goto sql_error;
		}
	}

	ret = 0;

sql_error:
	//DEBUG("SQL:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return ret;
}

int get_user_id(char *username)
{
	int ret, user_id = -1;
	
	ret = get_user_id_and_lf_times_and_salt(username, 
			&user_id, 
			NULL,
			NULL);
	return user_id;
}

int insert_user_login_log(int userid, 
	char *username, 
	char *event, 
	int result)
{	
	char pQueryStr[DB_MAX_STRING_LEN];
	int ret = -1;  
	
	snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"insert into user_auth_history(user_id,user_name,event,result)" 
        	"values('%d','%s','%s',%d)",
			userid,
			username,
			event,
			result);
	
	ret = safe_mysql_query(&g_db_connection, pQueryStr);
	if(ret != 0) {
		if(mysql_errno(&g_db_connection) == ER_DUP_ENTRY) {
			return ERR_DUP_ENTRY_IN_DB;
		}
	}
	
	return ret;
}

int check_user_login_timestamp(int userid, int interval)
{
	int res;
	int ret = -1;
	MYSQL_RES *res_ptr = NULL;
	//	MYSQL_RES *res_ptr1;
	MYSQL_ROW mysqlrow;	
	char pQueryStr[MAX_STRING_LEN];
	time_t timep, d_now, last_failure_time;
	struct tm *p_now;
	int diff_time;

	if(userid < 0){
		ERR("check_user_login_timestamp: userid < 0!\n");
		ret = -1;
		return ret;
	}

	time(&timep);
	p_now = (struct tm *)localtime(&timep);
	if(!p_now) {
		ERR("check_user_login_timestamp: failed to get time!\n");
		ret = -1;
		return ret;
	}
	d_now = mktime(p_now);

	if(interval > d_now) {
		ERR("check_user_login_timestamp: invalid interval!\n");
		return ret;
	}

	//d_now -= interval;
	
	snprintf(pQueryStr,
		MAX_STRING_LEN,
		"select unix_timestamp(create_time) from user_auth_history WHERE user_id = '%d' and " \
		"result != 0 order by id desc limit 1", 
		userid);
	//ERR("auth_history:%s\n", pQueryStr);
	res = safe_mysql_query(&g_db_connection, pQueryStr);
	if(res){
		ERR("check_user_login_timestamp select error %d: %s\n", res, 
			mysql_error(&g_db_connection));
		return ret;
	} else {
			
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				while((mysqlrow = mysql_fetch_row(res_ptr))){
					sscanf(mysqlrow[0], "%d", (int *)(&last_failure_time));
					diff_time = (last_failure_time + interval) - d_now;
					if(diff_time <= 0) {
						ret = 0;
						break;
					}
				}
			} else {
				//DEBUG("check_user_login_failure: sql returns no result\n");
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		} else {
			goto sql_error;
		}
	}

sql_error:
	//DEBUG("SQL:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return ret;
}

int check_user_login_failure_times(int userid,
	char *username, 
	int isFailed,
	int *pLoginLeftTimes)
{
	int res;
	int ret = -1;
	MYSQL_RES *res_ptr = NULL;
	//	MYSQL_RES *res_ptr1;
	MYSQL_ROW mysqlrow;	
	char pQueryStr[DB_MAX_STRING_LEN];
	int login_faiulre = 0;

	if(!username){
		ERR("check_user_login_failure: username is NULL!\n");
		return ret;
	}

	if(isFailed == 0) {
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"update users set login_failure='0' WHERE username = '%s'", 
			username);
	} else {	
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"SELECT login_failure FROM users WHERE userName = '%s' limit 1",
			username);
	}
	res = safe_mysql_query(&g_db_connection, pQueryStr);
	if(res){
		ERR("check_user_login_failure select error %d: %s\n", res, 
			mysql_error(&g_db_connection));
		return ret;
	} else {
		if(isFailed == 0) {
			ret = 0;
			goto sql_error;
		}
	
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				while((mysqlrow = mysql_fetch_row(res_ptr))){
					sscanf(mysqlrow[0],"%d",&login_faiulre);
					login_faiulre++;

					if(login_faiulre <= g_max_login_failure) {
						snprintf(pQueryStr,
							DB_MAX_STRING_LEN,
							"update users set login_failure='%d' WHERE username = '%s'", 
							login_faiulre,
							username);
						
						res = safe_mysql_query(&g_db_connection, pQueryStr);
						if(res){
							ERR("check_user_login_failure: update error %d: %s\n", 
								res,
								mysql_error(&g_db_connection));
							ret = -1;
							goto sql_error;
						}
					}
					
					if(login_faiulre >= g_max_login_failure) {
						ERR("check_user_login_failure: maximum times(%d) reached\n",
							g_max_login_failure);

						if(pLoginLeftTimes)
							*pLoginLeftTimes = 0;

						ret = -2;

						goto insert_user_log_entry;
					} else {
						if(pLoginLeftTimes)
							*pLoginLeftTimes = (g_max_login_failure - login_faiulre);
					}

					//snprintf(pQueryStr,
					//		DB_MAX_STRING_LEN,
					//		"update users set isApproved='0' WHERE username = '%s'", 
					//		username);
					
					break;
					//MSG("user id : %d\n",userId);					
				}
			} else {
				//DEBUG("check_user_login_failure: sql returns no result\n");
				goto sql_error;
			}

			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} else {
			goto sql_error;
		}
	}

	ret = 0;

insert_user_log_entry:
	if(userid > 0) {
		int tmp;
		tmp = insert_user_login_log(userid, username, "login", isFailed);
	}

sql_error:
	//DEBUG("SQL:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return ret;
}

unsigned char *hex_to_str(unsigned char *input_array, 
	unsigned int array_len,
	unsigned int max_array_size)
{
    unsigned char high_part;
	unsigned char low_part;
    unsigned int i;
	unsigned char *outputStr = NULL;

	if(!input_array || (array_len == 0))
		return outputStr;

	if(array_len > max_array_size)
		return outputStr;

	outputStr = (unsigned char*)calloc((2*array_len + 1), sizeof(char));
	if(!outputStr)
		return outputStr;
	
	for(i = 0; i < array_len; i++)
	{
		high_part = (*(input_array+i)&0xf0) >> 4;
		low_part = *(input_array+i)& 0x0f;
		
		if (high_part > 9) 
			outputStr[2*i] = 'a' + high_part - 10;
		else 
			outputStr[2*i] = '0' + high_part;
		
		if (low_part > 9) 
			outputStr[2*i+1] = 'a' + low_part - 10;
		else    
			outputStr[2*i+1] = '0' + low_part;
	}
	
	outputStr[2*i] = 0;

	return outputStr;
}

unsigned char *encrypt_user_password(char *password,
	unsigned char *salt, 
	int salt_len)
{
	int ret;
	int iteration = ENCRYPT_ITERATION_TIMES; 
	unsigned int encrypt_pwd_len = PWD_LEN;
	unsigned char *encrypted_pwd = NULL;
	unsigned char *returnStr = NULL;
	
	encrypted_pwd = (unsigned char*)calloc(encrypt_pwd_len, sizeof(char));
	if(!encrypted_pwd)
		return NULL;
	
	ret = PKCS5_PBKDF2_HMAC(password, 
			strlen(password), 
			salt,
			salt_len,
			iteration,
			PBKDF2_DIG_ALGORITHM,
			encrypt_pwd_len, 
			encrypted_pwd);
	if(ret == 0) {
		free(encrypted_pwd);
		encrypted_pwd = NULL;
	}

	returnStr = hex_to_str(encrypted_pwd, encrypt_pwd_len, PWD_LEN);
	
	//ERR("returnStr is %s\n", returnStr);
	
	return returnStr;
}

//get client and ospath in db
// return:
// 0 on success, -1 if common error
// -2 if no image, -3 if user is blocked
// -4 if wrong uname/pwd, -5 if login failure interval not met
int get_user_oslist(char *username,
	char *pwd, 
	char **ppOslist,
	int *pLoginLeftTimes,
	int failure_interval)
{	
	MYSQL_RES *res_ptr = NULL;
	//	MYSQL_RES *res_ptr1;
	MYSQL_ROW mysqlrow;
	int res, firsttime = 1, oldfirsttime;
	char pQueryStr[DB_MAX_STRING_LEN];
	char tmpstr[32], accstr[32];
	int userid, groupid = -1, imageaccess, personalization;
	int ret = -1;
	int isApproved = 0;
	int loginFailureTimes = 0;
	char *encrypt_pwd = NULL;
	unsigned char salt[SALT_LEN] = {0};
	char *oslist = NULL;
	char *tmpFilename;
	unsigned int oslistLen = 32;
	size_t infoSepStrLen;

	//DEBUG("db: get_user_oslist\n");  

	if(!ppOslist){
		ERR("get_user_oslist: outlist is NULL!\n");
		return -1;
	}

	*ppOslist = NULL;

	ret = get_user_id_and_lf_times_and_salt(username,
				&userid,
				&loginFailureTimes,
				salt);
	if(ret != 0) {
		ERR("get_user_oslist: failed to get uid!\n");
		ret = -4;
		return ret;
	}

	encrypt_pwd = encrypt_user_password(pwd, salt, SALT_LEN);
	if(encrypt_pwd == NULL) {
		ERR("get_user_oslist: failed to encrypt pwd!\n");
		ret = -1;
		return ret;
	}

	snprintf(pQueryStr,
		DB_MAX_STRING_LEN,
		//"SELECT id, groupid, isApproved, loginFailureTimes FROM users WHERE userName = '%s' AND "
		"SELECT id, groupid, isApproved FROM users WHERE userName = '%s' AND "		
		"password = '%s' limit 1",
		username, 
		encrypt_pwd);
	
	free(encrypt_pwd);
#if 1
	res = safe_mysql_query(&g_db_connection, pQueryStr);
	if(res){
		ERR("select error %d: %s\n",res,mysql_error(&g_db_connection));
		return -1;
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				while((mysqlrow = mysql_fetch_row(res_ptr))){
					
					sscanf(mysqlrow[2], "%d", &isApproved);
					if(isApproved == 0) {
						DEBUG("os list(username): user is blocked\n");
						ret = -3;
						goto sql_error;
					}

					//sscanf(mysqlrow[3], "%d", &loginFailureTimes);
					if(loginFailureTimes >= g_max_login_failure) {
						
						sscanf(mysqlrow[0], "%d", &userid);
						//check whether it exceeds the interval
						ret = check_user_login_timestamp(userid, 
									failure_interval);

						if(ret != 0) {
							DEBUG("os list(username): maximum failure times reached\n");
							ret = -5;
							if(pLoginLeftTimes)
								*pLoginLeftTimes = 0;
							goto sql_error;
						}
					}

					ret = check_user_login_failure_times(userid, username, 0, NULL);
					if(ret != 0) {
						DEBUG("failed calling check_failure: %s\n", username);
						ret = -1;
						goto sql_error;
					}

					sscanf(mysqlrow[1], "%d", &groupid);
					break;
					//MSG("group id : %d\n",groupid);
				}
			} else {
				DEBUG("check user(%s): sql returns no result\n", username);
				/*ret = get_user_id_and_login_failure_times(username, 
							&userid,
							&loginFailureTimes);
				if(ret != 0)
					goto sql_error;*/
				
				if(loginFailureTimes >= g_max_login_failure) {
					//check whether it exceeds the interval
					ret = check_user_login_timestamp(userid, 
								failure_interval);

					if(ret != 0) {
						DEBUG("os list1(username): maximum failure times reached\n");
						ret = -5;
						if(pLoginLeftTimes)
							*pLoginLeftTimes = 0;
						goto sql_error;
					}

					DEBUG("clear the login failure times\n");
					snprintf(pQueryStr,
						DB_MAX_STRING_LEN,
						"update users set login_failure='0' WHERE username = '%s'", 
						username);
					ret = safe_mysql_query(&g_db_connection, pQueryStr);
					if(ret) {
						ERR("failed to update lft: %s\n", mysql_error(&g_db_connection));
						goto sql_error;
					}
				}
				ret = check_user_login_failure_times(userid, username, 1, pLoginLeftTimes);
				if(ret == -2)
					ret = -5;
				else
					ret = -4;
				goto sql_error;
			}

			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		} else {
			goto sql_error;
		}
	}
#endif

	//oslist[0]=0;

	snprintf(pQueryStr,
		DB_MAX_STRING_LEN,
		"SELECT A.id, A.imagename, A.path, A.iconPath, A.revision, B.newPath, "
		"A.groupRead, A.groupDiffWrite, A.groupInstall, A.groupRawWrite, "
		"A.personalization, A.ostype "
       "FROM osimages A left join imageupdatehistory B on (A.id=B.imageId and B.revision=A.revision) "
       "WHERE A.major = %d and A.minor=%d and (groupread like '%%,%d,%%' " 
        "or groupdiffwrite like '%%,%d,%%' "
        "or grouprawwrite like '%%,%d,%%' "
        "or groupinstall like '%%,%d,%%' )",
        0,
        0,
        groupid,
        groupid,
        groupid,
        groupid);
	res = safe_mysql_query(&g_db_connection, pQueryStr);
	if(res) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		goto sql_error;
	} else {
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if(mysql_num_rows(res_ptr)>0) {
				snprintf(tmpstr, sizeof(tmpstr), "%d", groupid);

				infoSepStrLen = strlen( INFO_SEP_STR);
				oldfirsttime = firsttime;
				
				while((mysqlrow = mysql_fetch_row(res_ptr))) {
					if(firsttime == 0) {
						oslistLen++;
					}
					else
						firsttime = 0;

					oslistLen += strlen( mysqlrow[0]);
					oslistLen += infoSepStrLen;
					oslistLen += strlen( mysqlrow[1]);
					oslistLen += infoSepStrLen;
					oslistLen += strlen( getFilename(mysqlrow[2]));
					oslistLen += infoSepStrLen;
					oslistLen += strlen( getFilename(mysqlrow[3]));
					oslistLen += infoSepStrLen;
					oslistLen += strlen( mysqlrow[4]);
					oslistLen += infoSepStrLen;
					oslistLen += strlen( getFilename(mysqlrow[5]));
					oslistLen += infoSepStrLen;

					if(mysqlrow[10] && strstr(mysqlrow[10],tmpstr))	{
						//personalization
						personalization = IMAGE_ENABLE_PERS;
					} else {
						personalization = IMAGE_DISABLE_PERS;
					}
					snprintf(accstr, sizeof(accstr), "%d", personalization);

					oslistLen += strlen(accstr);

					if(mysqlrow[9] && strstr(mysqlrow[9],tmpstr))	//if in groupInstall
						imageaccess=IMAGE_RAW_READWRITE;
					else if(mysqlrow[8] && strstr(mysqlrow[8],tmpstr))
						imageaccess=IMAGE_INSTALL_RWRITE;
					else if(mysqlrow[7] && strstr(mysqlrow[7],tmpstr))
						imageaccess=IMAGE_DIFF_READWRITE;
					else if(mysqlrow[6] && strstr(mysqlrow[6],tmpstr))
						imageaccess=IMAGE_READONLY;
					else
						imageaccess=IMAGE_NO_RIGHTS;
					
					snprintf(accstr, sizeof(accstr), "%d", imageaccess);
					oslistLen += strlen(accstr);					
					oslistLen += infoSepStrLen;

					if(mysqlrow[11]) {
						oslistLen += strlen(mysqlrow[11]);
					}
				}

				oslist = (char *)malloc(oslistLen);
				if(!oslist){
					ERR("failed to allocate mem with size %d for oslist\n", 
						oslistLen);
					ret = -1;
					goto sql_error;
				}
				oslist[0] = 0;				
				firsttime = oldfirsttime;
				mysql_data_seek(res_ptr, 0); 
				
				while((mysqlrow = mysql_fetch_row(res_ptr))) {
					if(firsttime == 0) {
						strcat(oslist, "\n");
					}
					else
						firsttime = 0;
						
					strcat(oslist, mysqlrow[0]);
					strcat(oslist, INFO_SEP_STR);
					strcat(oslist, mysqlrow[1]);
					strcat(oslist, INFO_SEP_STR);
					strcat(oslist, getFilename(mysqlrow[2]));
					strcat(oslist, INFO_SEP_STR);
					strcat(oslist, getFilename(mysqlrow[3]));
					strcat(oslist, INFO_SEP_STR);
					strcat(oslist, mysqlrow[4]);
					strcat(oslist, INFO_SEP_STR);
					strcat(oslist, getFilename(mysqlrow[5]));

					strcat(oslist, INFO_SEP_STR);

					if(mysqlrow[10] && strstr(mysqlrow[10],tmpstr))	{
						//personalization
						personalization = IMAGE_ENABLE_PERS;
					} else {
						personalization = IMAGE_DISABLE_PERS;
					}
					snprintf(accstr, sizeof(accstr), "%d", personalization);
					strcat(oslist, accstr);

					if(mysqlrow[9] && strstr(mysqlrow[9],tmpstr))	//if in groupInstall
						imageaccess=IMAGE_RAW_READWRITE;
					else if(mysqlrow[8] && strstr(mysqlrow[8],tmpstr))
						imageaccess=IMAGE_INSTALL_RWRITE;
					else if(mysqlrow[7] && strstr(mysqlrow[7],tmpstr))
						imageaccess=IMAGE_DIFF_READWRITE;
					else if(mysqlrow[6] && strstr(mysqlrow[6],tmpstr))
						imageaccess=IMAGE_READONLY;
					else
						imageaccess=IMAGE_NO_RIGHTS;
					
					snprintf(accstr, sizeof(accstr), "%d", imageaccess);
					strcat(oslist, accstr);

					strcat(oslist, INFO_SEP_STR);
					if(mysqlrow[11])
						strcat(oslist, mysqlrow[11]);
				}
			} else {
				//DEBUG("%s\n",pQueryStr);
				ERR("os list(image): sql returns no result\n");
				ret = -2;
				goto sql_error;
			}

			if(mysql_errno(&g_db_connection)){
				ERR("Retrive error: %s\n",mysql_error(&g_db_connection));
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} else {
			ERR("failed to find oslist for user %s\n", username);
			ret = -1;
			goto sql_error;
		}
	}

	*ppOslist = oslist;

	return 0;	

sql_error:
	//DEBUG("SQL:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);

	if(oslist)
		free(oslist);

	if(ppOslist)
		*ppOslist = NULL;
	
	return ret;
}

int is_image_seed_ready(int osimage_id) 
{
  char pQueryStr[DB_MAX_STRING_LEN];
  int  res;
  MYSQL_RES *res_ptr = NULL;
  MYSQL_ROW mysqlrow;
  int isReady = 0;

  //DEBUG("db: get_user_status \n");
  
  snprintf(pQueryStr, 
  	DB_MAX_STRING_LEN,
  	"SELECT A.id from osimages A left join seed B "
   	"on (B.file_path=A.path) where A.id=%d AND status='active' limit 1",
	osimage_id);

  res = safe_mysql_query(&g_db_connection, pQueryStr); 
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return isReady;
		
  }else{
		res_ptr = mysql_store_result(&g_db_connection);
    	if(res_ptr){

			if((mysql_num_rows(res_ptr)) > 0){
				isReady = 1;
			} else {
        		DEBUG("image seed status: sql returns no result\n");
        		goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		} 
		
	}

  return isReady;
  
sql_error:
  //DEBUG("SQL:%s\n", pQueryStr);
  if(res_ptr)
  	mysql_free_result(res_ptr);
  return isReady;  
}

//get client and ospath in db
int get_image_update_list(
		int user_id, 
		char *osimage_id, 
		int start_ver,
		int end_ver,
		char *outlist)
{	
	MYSQL_RES *res_ptr = NULL;
	//	MYSQL_RES *res_ptr1;
	MYSQL_ROW mysqlrow;
	int res, firsttime = 1;
	char pQueryStr[DB_MAX_STRING_LEN];
	char tmpstr[32]={0}, accstr[32] = {0};
	int totalVerNum;
	int realVerNum = 0;

	//DEBUG("db: get_image_update_list\n");  

	if(!outlist){
		ERR("get_image_update_list: outlist is NULL!\n");
		return -1;
	}
/*
	if(start_ver == end_ver) {
		ERR("get_image_update_list: start ver == end_ver!\n");
		return -1;
	} else*/	if(start_ver > end_ver) {		
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"SELECT (revision-1) as rev, rstMetaDataPath, rstUserDataPath FROM imageupdatehistory "
			"WHERE imageId='%s' and (revision>%d and revision<=%d) order by revision desc",
            osimage_id,
            end_ver,
            start_ver);
		totalVerNum = start_ver - end_ver;
	} else {
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"SELECT revision, metaDataPath, userDataPath FROM imageupdatehistory "
			"WHERE imageId='%s' and (revision>%d and revision<=%d ) order by revision asc",
            osimage_id,
            start_ver,
            end_ver);
		totalVerNum = end_ver - start_ver;
	}
	
	res = safe_mysql_query(&g_db_connection, pQueryStr);
	if(res) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;
	} else {
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if(mysql_num_rows(res_ptr)>=0) {
				while((mysqlrow = mysql_fetch_row(res_ptr))) {
					if(firsttime == 0)
						strcat(outlist, "\n");						
					else
						firsttime = 0;
						
					strcat(outlist, mysqlrow[0]);
					strcat(outlist, INFO_SEP_STR);
					strcat(outlist, getFilename(mysqlrow[1]));
					strcat(outlist, INFO_SEP_STR);
					strcat(outlist, getFilename(mysqlrow[2]));
					realVerNum++;
				}
			} else {
				//DEBUG("%s\n",pQueryStr);
				ERR("version: sql returns no result\n");
				goto sql_error;
			}

			if(mysql_errno(&g_db_connection)){
				ERR("Retrive error: %s\n",mysql_error(&g_db_connection));
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} else {
			ERR("failed to find outlist for os image id %s\n", osimage_id);
		}
	}

	if(totalVerNum != realVerNum) {
		outlist[0] = 0;
		return 1;
	}

	return 0;	

sql_error:
	//DEBUG("SQL:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return -1;
}

int get_data_image_update_list(
		int user_id, 
		char *osimage_id, 
		int start_ver,
		int end_ver,
		char *outlist)
{	
	MYSQL_RES *res_ptr = NULL;
	//	MYSQL_RES *res_ptr1;
	MYSQL_ROW mysqlrow;
	int res, firsttime = 1;
	char pQueryStr[DB_MAX_STRING_LEN];
	char tmpstr[32]={0}, accstr[32] = {0};
	int totalVerNum;
	int realVerNum = 0;

	//DEBUG("db: get_data_image_update_list\n");  

	if(!outlist){
		ERR("get_data_image_update_list: outlist is NULL!\n");
		return -1;
	}

	/*if(start_ver == end_ver) {
		ERR("get_data_image_update_list: start ver == end_ver!\n");
		return -1;
	} else*/	if(start_ver > end_ver) {		
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"SELECT (revision-1) as rev, rst_mdf_file, rst_udf_file, image_file, image_size FROM private_storage "
			"WHERE user_id=%d and (revision>%d and revision<=%d) order by revision desc",
            user_id,
            end_ver,
            start_ver);
		totalVerNum = start_ver - end_ver;
	} else {
		snprintf(pQueryStr,
			DB_MAX_STRING_LEN,
			"SELECT revision, mdf_file, udf_file, image_file, image_size FROM private_storage "
			"WHERE user_id=%d and (revision>%d and revision<=%d ) order by revision asc",
            user_id,
            start_ver,
            end_ver);
		totalVerNum = end_ver - start_ver;
	}

	res = safe_mysql_query(&g_db_connection, pQueryStr);
	if(res) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;
	} else {
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if(mysql_num_rows(res_ptr)>=0) {
				while((mysqlrow = mysql_fetch_row(res_ptr))) {
					if(firsttime == 0)
						strcat(outlist, "\n");						
					else
						firsttime = 0;
						
					strcat(outlist, mysqlrow[0]);
					strcat(outlist, INFO_SEP_STR);
					strcat(outlist, mysqlrow[1]);
					strcat(outlist, INFO_SEP_STR);
					strcat(outlist, mysqlrow[2]);
					strcat(outlist, INFO_SEP_STR);
					strcat(outlist, mysqlrow[3]);
					strcat(outlist, INFO_SEP_STR);
					strcat(outlist, mysqlrow[4]);
					realVerNum++;
				}
			} else {
				//DEBUG("%s\n",pQueryStr);
				ERR("version: sql returns no result\n");
				goto sql_error;
			}

			if(mysql_errno(&g_db_connection)){
				ERR("Retrive error: %s\n",mysql_error(&g_db_connection));
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} else {
			ERR("failed to find outlist for os image id %s\n", osimage_id);
		}
	}

	if(totalVerNum != realVerNum) {
		outlist[0] = 0;
		return 1;
	}

	return 0;	

sql_error:
	//DEBUG("SQL:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return -1;
}

int get_data_image_size(int user_id, char* p_image_file, 
	char *p_image_size, size_t image_size_len)
{
	MYSQL_RES *res_ptr = NULL;
	//	MYSQL_RES *res_ptr1;
	MYSQL_ROW mysqlrow;
	int res;
	char pQueryStr[MAX_STRING_LEN];
		
	snprintf(pQueryStr,
		MAX_STRING_LEN,
		"SELECT image_file, image_size FROM private_storage WHERE "
		"user_id=%d and revision=1 limit 1", 
		user_id);
	res = safe_mysql_query(&g_db_connection, pQueryStr);
	if(res){
		ERR("select error %d: %s\n", res, mysql_error(&g_db_connection));
		return -1;
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				while((mysqlrow = mysql_fetch_row(res_ptr))){
					strcat(p_image_file, getFilename(mysqlrow[0]));
					if((strlen(mysqlrow[1]) + 1) > image_size_len) {
						DEBUG("os list: buffer too small\n");
						goto sql_error;
					}
					sscanf(mysqlrow[1], "%s", p_image_size);
					break;
					//MSG("user id : %d\n",userId);					
				}
			} else {
				DEBUG("os list(username): sql returns no result\n");
				goto sql_error;
			}

			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} else {
			goto sql_error;
		}
	}

	return 0;

sql_error:
	//DEBUG("SQL:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return -1;
}

/*
int get_ospath_by_name(char *name, char *path, size_t path_len)
{
  char pQueryStr[DB_MAX_STRING_LEN];
  int  res;
  MYSQL_RES *res_ptr = NULL;
  MYSQL_ROW mysqlrow;

  //DEBUG("db: get_ospath_by_name\n");

  snprintf(pQueryStr,
  	DB_MAX_STRING_LEN,
  	"select path from osimages WHERE imagename = '%s'", 
  	name);

  res = safe_mysql_query(&g_db_connection, pQueryStr);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    	if(res_ptr){

			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				sscanf_s(mysqlrow[0], "%s", path, path_len);
				//DEBUG("osimage path %s\n",path);
			} else {
		        DEBUG("image path: sql returns no result\n");
		        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} 
		
	}

  return 0;
  
sql_error:
  //DEBUG("SQL:%s\n", pQueryStr);
  if(res_ptr)
 	mysql_free_result(res_ptr);
  return -1;
}
*/

int get_osimagename_by_id(char *osid, char *imagename, int imagenameLen) 
{
  char pQueryStr[DB_MAX_STRING_LEN];
  int  res;
  MYSQL_RES *res_ptr = NULL;
  MYSQL_ROW mysqlrow;

  //DEBUG("db: get_osimagename_by_id\n");

  snprintf(pQueryStr,
  	DB_MAX_STRING_LEN,
  	"select imagename from osimages WHERE id = '%s'", 
  	osid);

  res = safe_mysql_query(&g_db_connection, pQueryStr);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    	if(res_ptr){

			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				//sscanf(mysqlrow[0],"%s", imagename);
				strncpy(imagename, mysqlrow[0], imagenameLen);
				imagename[imagenameLen - 1] = '\0';
				//DEBUG("osimage name is %s\n",imagename);
			} else {
        		DEBUG("image name: sql returns no result\n");
        		goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} 
		
	}

  return 0;
  
sql_error:
  //DEBUG("SQL:%s\n", pQueryStr);
  if(res_ptr)
  	mysql_free_result(res_ptr);
  return -1;
}

int get_osimageid_by_name(char *name) 
{
  char pQueryStr[DB_MAX_STRING_LEN];
  int  res;
  int  osimageid = -1;
  MYSQL_RES *res_ptr = NULL;
	MYSQL_ROW mysqlrow;

  //DEBUG("db: get_osimageid_by_name\n");

  snprintf(pQueryStr,
  	DB_MAX_STRING_LEN,
  	"select id from osimages WHERE imagename = '%s'", 
  	name);

  res = safe_mysql_query(&g_db_connection, pQueryStr);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d", &osimageid);
				//DEBUG("osimage id is %d\n",osimageid);
			} else {
		        DEBUG("image id: sql returns no result\n");
		        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;

		} 
		
	}

  return osimageid;
  
sql_error:
  //DEBUG("SQL:%s\n", pQueryStr);
  if(res_ptr)
  	mysql_free_result(res_ptr);
  return -1;
}

int get_osimageid_by_id(char *id) 
{
  char pQueryStr[DB_MAX_STRING_LEN];
  int  res;
  int  osimageid = -1;
  MYSQL_RES *res_ptr = NULL;
	MYSQL_ROW mysqlrow;

  //DEBUG("db: get_osimageid_by_id\n");

  snprintf(pQueryStr,
  	DB_MAX_STRING_LEN,
  	"select id from osimages WHERE id = '%s'", 
  	id);

  res = safe_mysql_query(&g_db_connection, pQueryStr);
  if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
    if(res_ptr){

			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0],"%d", &osimageid);
				//DEBUG("osimage id is %d\n",osimageid);
			} else {
        DEBUG("image id: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		} 
		
	}

  return osimageid;
  
sql_error:
  //DEBUG("SQL:%s\n", pQueryStr);
  if(res_ptr)
  	mysql_free_result(res_ptr);
  return -1;
}

int is_image_locked(char* osimage_id)
{
	char pQueryStr[MAX_STRING_LEN];
	int  res;
	MYSQL_RES *res_ptr = NULL;
	MYSQL_ROW mysqlrow;
	int is_locked = 0;

	//DEBUG("db: is_image_locked \n");

	snprintf(pQueryStr,
		MAX_STRING_LEN,
		"select needUpdate from osimages WHERE id='%s'", 
		osimage_id);

	res = safe_mysql_query(&g_db_connection, pQueryStr); 
	if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				if(strcmp(mysqlrow[0], "2") == 0)
					is_locked = 1;
				//DEBUG("image status: %s\n",mysqlrow[0]);
			} else {
		    	DEBUG("image status: sql returns no result\n");
		    	goto sql_error;
			}
				
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		}		
	}

	return is_locked;

sql_error:
	//DEBUG("SQL error:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return -1;
  
}

int is_data_image_locked(char* osimage_id, int user_id)
{
	char pQueryStr[MAX_STRING_LEN];
	char pid_file[MAX_STRING_LEN];
	int  res;
	MYSQL_RES *res_ptr = NULL;
	MYSQL_ROW mysqlrow;
	int is_locked = 0;
	int pid_fd = -1;

	//DEBUG("db: is_image_locked \n");

	snprintf(pQueryStr,
		MAX_STRING_LEN,
		"SELECT image_file FROM private_storage "
		"WHERE user_id=%d and revision=1 limit 1", 
		user_id);

	res = safe_mysql_query(&g_db_connection, pQueryStr); 
	if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				snprintf(pid_file, 
					MAX_STRING_LEN, 
					"%s.merging", 
					mysqlrow[0]);
				pid_fd = open(pid_file, O_RDONLY);
				if(pid_fd >= 0) {
					is_locked = 1;
					close(pid_fd);
				}
				//DEBUG("image status: %s\n",mysqlrow[0]);
			} else {
		    	DEBUG("image status: sql returns no result\n");
		    	goto sql_error;
			}
				
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		}		
	}

	return is_locked;

sql_error:
	//DEBUG("SQL error:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return -1;
}

int get_data_image_cur_ver(int userId,
		char *osimage_id, 
		char *cur_ver) 
{
	char pQueryStr[MAX_STRING_LEN];
	int  res;
	MYSQL_RES *res_ptr = NULL;
	MYSQL_ROW mysqlrow;

	//DEBUG("db: get_image_cur_ver \n");
	
	snprintf(pQueryStr,
		MAX_STRING_LEN,
		"select A.revision from private_storage A " 
		" where A.user_id =%d order by id desc limit 1", 
		userId);

	res = safe_mysql_query(&g_db_connection, pQueryStr); 
	if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				if(mysqlrow && mysqlrow[0]) {
					strcat(cur_ver, mysqlrow[0]);
					//DEBUG("cur ver : %s\n",cur_ver);
				} else {
		    			DEBUG("Error: max version is NULL\n");
		    			goto sql_error;
				}
			} else {
		    	DEBUG("cur version: sql returns no result\n");
		    	goto sql_error;
			}
				
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		}		
	}

	return 0;

sql_error:
	//DEBUG("SQL error:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return -1;  
}

int get_image_cur_ver(int userId, char *osimage_id, char *cur_ver) 
{
	char pQueryStr[MAX_STRING_LEN];
	int  res;
	MYSQL_RES *res_ptr = NULL;
	MYSQL_ROW mysqlrow;

	//DEBUG("db: get_image_cur_ver \n");

	snprintf(pQueryStr,
		MAX_STRING_LEN,
		"select revision from osimages WHERE id='%s'", 
		osimage_id);

	res = safe_mysql_query(&g_db_connection, pQueryStr); 
	if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				strcat(cur_ver, mysqlrow[0]);
				//DEBUG("cur ver : %s\n",cur_ver);
			} else {
		    	DEBUG("cur version: sql returns no result\n");
		    	goto sql_error;
			}
				
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		}		
	}

	return 0;

sql_error:
	//DEBUG("SQL error:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	return -1;  
}


int get_imageaccess(char *osid, char *username, IMAGE_ACCESS *imageaccess) 
{
	char pQueryStr[DB_MAX_STRING_LEN];
	int  res;
	int  groupid = -1;
	char tmpstr[32];
	MYSQL_RES *res_ptr = NULL;
	MYSQL_ROW mysqlrow;

  //DEBUG("db: get_imageaccess\n");

  	snprintf(pQueryStr,
		DB_MAX_STRING_LEN,
		"select groupid from users WHERE username = '%s'", 
		username);

  	res = safe_mysql_query(&g_db_connection, pQueryStr);
  	if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;
		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				sscanf(mysqlrow[0], "%d", &groupid);
				//DEBUG("groupid is %d\n", groupid);
			} else {
				DEBUG("group id: sql returns no result\n");
				goto sql_error;
			}

			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}

			mysql_free_result(res_ptr);
			res_ptr = NULL;
		}
	}

  	snprintf(tmpstr,
		sizeof(tmpstr),
		"%d",
		groupid);
	snprintf(pQueryStr,
		DB_MAX_STRING_LEN,
		"SELECT groupRead,groupDiffWrite,groupInstall,groupRawWrite "
		"FROM osimages WHERE id = '%s' limit 1",
		osid);
	res = safe_mysql_query(&g_db_connection, pQueryStr);
	if(res) {
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return -1;
	} else {
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if(mysql_num_rows(res_ptr)>0) {
		
				while((mysqlrow = mysql_fetch_row(res_ptr))) {
					if(mysqlrow[3] && strstr(mysqlrow[3],tmpstr))	//if in groupInstall
						*imageaccess=IMAGE_RAW_READWRITE;
					else if(mysqlrow[2] && strstr(mysqlrow[2],tmpstr))
						*imageaccess=IMAGE_INSTALL_RWRITE;
					else if(mysqlrow[1] && strstr(mysqlrow[1],tmpstr))
						*imageaccess=IMAGE_DIFF_READWRITE;
					else if(mysqlrow[0] && strstr(mysqlrow[0],tmpstr))
						*imageaccess=IMAGE_READONLY;
					else
						*imageaccess=IMAGE_NO_RIGHTS;		//no rights to access the image
        }	

        //DEBUG("imageaccess = %d\n", (int) *imageaccess);        
			} else {
        DEBUG("image access: sql returns no result\n");
        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				DEBUG("Retrive error: %s\n",mysql_error(&g_db_connection));
        goto sql_error;
			}

      mysql_free_result(res_ptr);
	  res_ptr = NULL;
			
		} 

	}

  return 0;
  
sql_error:
  //DEBUG("SQL:%s\n", pQueryStr);
  if(res_ptr)
  	mysql_free_result(res_ptr);
  return -1;
}

char *get_default_server_config(int serverId,char *configName)
{	
	char pQueryStr[MAX_STRING_LEN];
	int  res;
	MYSQL_RES *res_ptr = NULL;
	MYSQL_ROW mysqlrow;
	char *configValue = NULL;
	size_t configValueLen;

	//DEBUG("db: get_default_server_config \n");

	snprintf(pQueryStr,
		MAX_STRING_LEN,
		"select configValue from serverconfigs where serverId=%d "
		"and configName='%s' limit 1",
		serverId,
		configName);
	res = safe_mysql_query(&g_db_connection, pQueryStr); 
	if(res){
		ERR("select error: %s\n", mysql_error(&g_db_connection));
		return NULL;		
	}else{
		res_ptr = mysql_store_result(&g_db_connection);
		if(res_ptr){
			if((mysql_num_rows(res_ptr)) > 0){
				mysqlrow = mysql_fetch_row(res_ptr);
				configValueLen = strlen(mysqlrow[0]) + 1;
				if(!(configValue=(char *)calloc(configValueLen, sizeof(char))))
				{
					ERR("get_default_server_config error:could not allocate space for configValue!");
					return NULL;
				}
				// dest str is longer than src
				strncpy(configValue, mysqlrow[0], configValueLen);
				//DEBUG("configname: %s,configValue : %s\n", configName, configValue);
			} else {
			        DEBUG("configValue: sql returns no result\n");
			        goto sql_error;
			}
			
			if(mysql_errno(&g_db_connection)){
				goto sql_error;
			}
			mysql_free_result(res_ptr);
			res_ptr = NULL;
		} 		
	}

	return configValue;

sql_error:
	//DEBUG("SQL:%s\n", pQueryStr);
	if(res_ptr)
		mysql_free_result(res_ptr);
	if(configValue)
		free(configValue);
	
	return NULL;
}

char *get_udf_dir()
{
	return get_default_server_config(TC_SERVER_ID, "userDataPath");
}

char *get_home_tc_dir()
{
	return get_default_server_config(TC_SERVER_ID, "home");
}

char *get_home_var_dir()
{
	return get_default_server_config(TC_SERVER_ID, "home_var");
}

char *get_lic_file()
{
	return get_default_server_config(TC_SERVER_ID, "lic_file");
}

char *get_ca_dir()
{
	char *ca_dir;

	ca_dir = get_default_server_config(TC_SERVER_ID, "ca_dir");
	if(!ca_dir)
		return DFT_TC_CA_DIR;

	return ca_dir;
}

char *get_srv_cert_dir()
{
	return get_ca_dir();
}

char *get_srv_key_dir()
{
	return get_ca_dir();
}

int get_login_failure_interval()
{
	int ret = -1;
	char *pInterval = NULL;
	
	pInterval = get_default_server_config(TC_SERVER_ID, "lf_retry_interval");
	if(pInterval != NULL) {
		sscanf(pInterval, "%d", &ret);
		free(pInterval);
	} else {
		ret = DFT_LOGIN_FAILURE_INTERVAL;
	}

	return ret;
}

#else

struct userinfo {
  char  name[64];
  char  passwd[64];
  int   images[8];
  IMAGE_ACCESS   rwmode;
};

struct imageinfo {
  int   id;
  char  name[64];
  char  path[128];
};

struct imageinfo images[] = { {1, "test", "/work/diskimage/test.img"}, 
                            {2, "winxp", "/work/diskimage/winxp.img"},
                            {3, "vista", "/work/diskimage/vista.img"},
                            {4, "linux", "/work/diskimage/linux.img"},
                          };
                          
struct userinfo users[] = { {"root", "123456",{1,2,3,4,0,0,0,0}, IMAGE_RAW_READWRITE},
                            {"1", "123456",{1,2,3,4,0,0,0,0}, IMAGE_DIFF_READWRITE},
                            {"2", "123456",{1,2,3,4,0,0,0,0}, IMAGE_DIFF_READWRITE},
                            {"3", "123456",{1,2,3,4,0,0,0,0}, IMAGE_DIFF_READWRITE}
                          };
                          

int start_db(char *name, char *pwd)
{
  return 0;
}

void reset_all_status()
{

}

void end_db()
{

}

//
// without DB, server has no way to trace user's online status
// so simply returns offline
//
int get_user_status(char *username, int osimage_id) 
{
  return 0;
}

int get_user_status_by_mac(char *username) 
{
  return 0;
}

int insert_user_status(char *username, int osimage_id, char *mac, int imageaccess, int online)
{
  return 0;
}

int delete_user_status(char *username, int osimage_id) 
{
  return 0;
}

int update_image_status(char* rawfile, char* newfile, int needupdate)
{
  return 0;
}

int update_server_status(int server_id,int status)
{
  return 0;
}

int get_user_oslist(char *username, char *pwd, char *oslist)
{
  int i,j;

  for(i = 0; i < sizeof(users) / sizeof(struct userinfo); i++) {

    if(strcmp(username, users[i].name) == 0 && strcmp(pwd, users[i].passwd) == 0) {

      oslist[0] = '\0';
      
      if(users[i].images[0] != 0) {
        strcat(oslist, images[0].name);       
      }

      for(j = 1; j < 8; j++) {
        if(users[i].images[j] != 0) {
          strcat(oslist, ";");
          strcat(oslist, images[j].name);
        }  
      }

    }

  }  

  return 0;
}
/*
int get_ospath_by_name(char *name, char *path) 
{
  int i;

  for(i = 0; i < sizeof(images) / sizeof(struct imageinfo); i++) {
    if(strcmp(name, images[i].name) == 0) {
      strncpy(path, images[i].path, xxx);
      return 0;
    }  
  }

  return -1;
}
*/
int get_osimageid_by_name(char *name) 
{
  int i;

  for(i = 0; i < sizeof(images) / sizeof(struct imageinfo); i++) {
    if(strcmp(name, images[i].name) == 0) {
      return images[i].id;
    }
  }

  return -1;
}

int get_imageaccess(char *osid, char *username, IMAGE_ACCESS *imageaccess) 
{
  int i;
  
  for(i = 0; i < sizeof(users) / sizeof(struct userinfo); i++) {
    if(strcmp(username, users[i].name) == 0) {
      *imageaccess = users[i].rwmode;
      return 0;
    }
  }

  return -1;
}

char *get_udf_dir()
{
  int udfDirLen = 128;  
  char *pUdfDir;

  pUdfDir = calloc(udfDirLen, 1);

  if(!pUdfDir) {
  	return pUdfDir;
  }

  strncpy(pUdfDir, DFT_UDF_DIR, udfDirLen);
  pUdfDir[udfDirLen - 1] = '\0';
  
  return pUdfDir;
}

#endif   //end USE_MYSQL

