/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <error.h>
#include <linux/fs.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include "predefine.h"
#include "fileops.h"
#include "debug.h"
#include "db.h"

#define NEW_DIR_MODE	0666

//extern char *g_app_fullname;


/* Functions start with '_' were for interal use only */

int _create_dir_ifnot_exists(char *src_dir)
{
	char *tmp_dir;
	char dst_dir[MAX_STRING_LEN];

	if(!src_dir) {
		ERR("create dir: param NULL\n");
		return -1;
	}
	
	tmp_dir = (char *)strtok(src_dir, "/");
	if(!tmp_dir) {
		ERR("Failed to parse src dir: %s\n", src_dir);
		return -1;
	}
	snprintf(dst_dir, MAX_STRING_LEN, "/%s", tmp_dir);
	
	if(mkdir(dst_dir, NEW_DIR_MODE)<0)
		;//DEBUG("%s already exist.\n", dst_dir);
	else 
		;//DEBUG("Dir created : %s\n", dst_dir);
	  
	if(chdir(dst_dir) == -1) {
		ERR("Failed switching into dir: %s\n", strerror(errno));
		return -1;
  	}
  
	while((tmp_dir = (char *)strtok(NULL,"/"))){
		snprintf(dst_dir, MAX_STRING_LEN, "%s",tmp_dir);
		
		if(mkdir(dst_dir, NEW_DIR_MODE) < 0)
			;//DEBUG("%s already exist.\n", dst_dir);
		else 
		  	;//DEBUG("Dir created: %s\n", dst_dir);
		  
		if(chdir(dst_dir) == -1) {
			ERR("Failed switching into dir: %s\n",dst_dir);
			return -1;
		}
	}
	return 0;
}

/* end of internal functions */

char* get_last_filename(char* str)
{
	int i, slen;
	int pos = -1;	
	char* tmp_char;

	if(str == NULL)
		return NULL;

	slen = strlen(str);
	for(i=0;i < slen;i++)
	{
		if(str[i] == '/')
		{
			pos = i;
		}
	}
	
	tmp_char = str+pos+1;
	return tmp_char;
}

char* get_parent_dir(char* str_path)
{
	int i, slen;
	char* parent_dir;

	if(str_path == NULL)
		return NULL;

	parent_dir = strdup(str_path);
	if(!parent_dir) {
		return NULL;
	}
	slen = strlen(parent_dir);
	for(i=(slen-1);i >= 0;i--)
	{
		if(parent_dir[i] == '/')
		{
			parent_dir[i] = '\0';
			break;
		}
	}	
	
	return parent_dir;
}

void create_dir(char *dst_dir)
{
  char dst_path[MAX_STRING_LEN];

  if(!dst_dir) {
  	ERR("setup_dir: dst_dir is NULL\n");
  	return;
  }else{
	strncpy(dst_path, dst_dir, MAX_STRING_LEN);
	dst_path[MAX_STRING_LEN - 1] = '\0';
	
  	_create_dir_ifnot_exists(dst_path);
  }
}

void setup_udf_dir()
{
  char udf_path[MAX_STRING_LEN];
  char *udf_dir;

  udf_dir = (char *)get_udf_dir();
  if(!udf_dir)
  {
	//to do this because strtok() in _create_dir_ifnot_exists() need the diffpath a non-const string.
  	strncpy(udf_path, DFT_UDF_DIR, MAX_STRING_LEN);
	udf_path[MAX_STRING_LEN - 1] = '\0';
	
  	_create_dir_ifnot_exists(udf_path);
  }else{
	strncpy(udf_path, udf_dir, MAX_STRING_LEN);
	udf_path[MAX_STRING_LEN - 1] = '\0';
	
  	_create_dir_ifnot_exists(udf_path);
	free(udf_dir);
  }
}

char *get_cur_dir()
{
    char *cur_dir;
    cur_dir = (char *)malloc(MAX_STRING_LEN);
    getcwd(cur_dir, MAX_STRING_LEN);
    //DEBUG("%s",file_path_getcwd);
    return cur_dir;
}

char *get_app_path()
{
	char *app_dir;
	int dir_len;
	int max_dir_len = MAX_STRING_LEN*2;
	int i;

	app_dir = (char *)malloc(max_dir_len);
	if(!app_dir)
		return NULL;
	
	dir_len = readlink("/proc/self/exe", app_dir, max_dir_len);
	if (dir_len < 0 || dir_len >= max_dir_len)
	{
	    ERR("Failed to get app path\n");
		free(app_dir);
	    return NULL;
	}
	app_dir[dir_len] = 0;
		
	for (i = (dir_len - 1); i >=0; --i)
	{
	    if (app_dir[i] == '/')
	    {
	        app_dir[i+1] = 0;
	        break;
	    }
	}
	return app_dir;
}

off_t
get_file_size(int fd)
{
	off_t size;
	struct stat fs_stat;
	int ret;

	if (ioctl(fd, BLKGETSIZE64, &size) == -1) {
		ret = fstat(fd, &fs_stat);
		if (ret == -1) {
			ERR("getsize error %d\n", errno);
			exit(1);
		}
		size = fs_stat.st_size;
	}

	return size;
}

int fileReadLine(int file_handle, uint64_t* pOffset,
		unsigned char** readBuffer, ulong maxTryBufLen)
{
	unsigned char*		tmpBuf;
	uint64_t	oldOffset;
	uint32_t	i = 0;
	ulong		realReadLen = 0;

	*readBuffer = NULL;
	oldOffset = *pOffset;
	if(maxTryBufLen == 0)
		maxTryBufLen = 2048;

	tmpBuf = (unsigned char*)malloc(maxTryBufLen);
	if (tmpBuf == NULL)
		return -1;

	if(lseek(file_handle, *pOffset, SEEK_SET) == -1) {
	  ERR("failed to seek pos 0x%lx (%d)\n", *pOffset, errno);
	  free(tmpBuf);
	  return -1;
	}

	if((realReadLen = read(file_handle, tmpBuf, maxTryBufLen)) == -1) {
	  ERR("failed to read 0x%lx bytes (%d)\n", maxTryBufLen, errno);
	  free(tmpBuf);
	  return -1;
	}

	*pOffset = oldOffset;
	//search for line terminator
	for(i=0; i< realReadLen; i++)
	{
		*pOffset += 1;
		if(tmpBuf[i] == '\n')
		{
			//isFound= TRUE;
			break;
		}
		else if((i+1 < realReadLen) && tmpBuf[i] == '\r' && tmpBuf[i+1] == '\n')
		{
			//isFound= TRUE;
			*pOffset += 1;
			break;
		}
	}

	if(i == 0)
	{
		free(tmpBuf);
		// it means we reached the end of the file
		if(realReadLen == 0)
			return -1;
		else
			return i;
	}

	*readBuffer = (unsigned char*)malloc(i+1);
	if(!(*readBuffer)) {
		ERR("failed to alloc read buffer\n");
	  	free(tmpBuf);
		return -1;
	}
	memset(*readBuffer, 0x0, i+1);
	memcpy(*readBuffer, tmpBuf, i);
	free(tmpBuf);
	return i+1;
}

static int
getConfigInfo(unsigned char* readBuffer, long bufLen,
		char* keyword, uint32_t keywordLen, char** pUFileName)
{
    long		nameLen = 0;

	if(readBuffer != NULL && (memcmp(readBuffer, keyword, keywordLen) == 0))
	{
		nameLen = bufLen-keywordLen;
		if(nameLen == 0)
		{
			*pUFileName = NULL;
			return -1;
		}
		else
			*pUFileName = malloc(nameLen);
		
		memset(*pUFileName, 0x0, nameLen);
		memcpy(
			*pUFileName,
			readBuffer+keywordLen,
			nameLen
			);

		return 0;
	}

	return -1;
}

static int
searchKeyword(KEYWORD_ENTRY* keywordEntries, uint32_t entryNum,
		unsigned char* readBuffer, long readLen, char** pFilename)
{
	int ntStatus;
	uint32_t i = 0;

	for(i=0; i < entryNum; i++)
	{
		if(keywordEntries[i].isMatched == 0)
		{
			ntStatus = getConfigInfo(readBuffer, readLen, keywordEntries[i].keyword,
					keywordEntries[i].len, pFilename);
			if (ntStatus == 0)
			{
				keywordEntries[i].isMatched = 1;
				return i;
			}
		}
	}

	return -1;
}

int
parseConfigFile (char* pConfigFilename,
	char* pConfigData)
{
    int							file_handle;
    unsigned char*						readBuffer;
    char*						filename, *newFilename;
    uint64_t					offset;
    uint32_t 					fileNum;

    uint32_t					i = 0;
	uint32_t					configuredFileNum = 0;
    long						readLen = 0;
    long						bufLen = 2048;
    int							keywordIndex = 0;
    char*						filenames[KEYWORD_NUM];
    KEYWORD_ENTRY 				keywordEntries[KEYWORD_NUM] = {
									{KEYWORD_SYNC_PASS, strlen(KEYWORD_SYNC_PASS), 0}
    								};

	if(!pConfigFilename || !pConfigData) {
		ERR("Wrong parameters\n");
    	return -1;
	}

    file_handle = open(pConfigFilename, O_RDWR);
    if(file_handle < 0)
    {
    	ERR("Error parsing %s: %d\n", pConfigFilename, errno);
    	return -1;
    }

    i 		= 0;
    offset 	= 0;

    fileNum = sizeof(keywordEntries)/sizeof(KEYWORD_ENTRY);

    filenames[0] = pConfigData;

    while(TRUE)
    {
    	if((readLen = fileReadLine(file_handle,
    			&offset, &readBuffer, bufLen)) != -1)
    	{
    		//"readLen == 0" indicates it meets a single line terminator
    		//so just continue to read next line...
    		if(readLen == 0)
    		{
    			//do not need to free readBuffer here since it's NULL
    			i++;
				continue;
    		}

			filename = NULL;

    		//compare to see whether the data in buffer starts with the keyword.
    		keywordIndex = searchKeyword(keywordEntries, fileNum,
    								readBuffer, readLen, &filename);

			free(readBuffer);

    		if(keywordIndex == -1 || keywordIndex >= fileNum)
    		{
				//close(file_handle);
				//free(readBuffer);
				//return -1;
				i++;
				if(filename)
					free(filename);
    			continue;
			}

    		if(filename != NULL)
    		{
				snprintf(filenames[keywordIndex], 
					MAX_STRING_LEN, 
					"%s", 
					filename);
				free(filename);
			} else
				memset(filenames[keywordIndex], 0, MAX_STRING_LEN);

			configuredFileNum++;
	    	i++;

			if(configuredFileNum == fileNum)
				break;
		}
    	else
    	{
    		break;
    	}
    }
    if(configuredFileNum != fileNum)
    {
    	ERR("No enough or duplicate configuration entry in configuration file found! \n ");

    	close(file_handle);
    	return -1;
    }

    close(file_handle);
    return 0;
}


