/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <linux/fs.h>
#include <sys/statfs.h> 

//typedef unsigned long long u64;
//typedef unsigned int uint32_t;

#define UDF_PAGE_SIZE 		(2*1024*1024) 
#define MAX_STRING_LEN 		256
#define DB_MAX_STRING_LEN 	(MAX_STRING_LEN*2)
#define FAILED_FILE_EXT		"merge_failed"
#define PID_FILE_EXT		".merging"
#define PROGRESS_FLAG		"-p"
#define RESTORE_FLAG		"-r"
#define VERIFY_ONLY_FLAG	"-v"
#define RESTORE_FILE_EXT	".rst"

void usage()
{
  printf("merge [%s] [%s] [%s] raw_file md_file ud_file [new_file]\n", 
  	PROGRESS_FLAG,
  	RESTORE_FLAG,
  	VERIFY_ONLY_FLAG);
  printf("\t%s: update merging progress in pid file\n", PROGRESS_FLAG);
  printf("\t%s: create image restoration files before merging\n", RESTORE_FLAG);
  printf("\t%s: verify image only and DO NOT merge the image\n", 
  	VERIFY_ONLY_FLAG);
  printf("Note: if parameter \"new_file\" is absent, " \
  		"all changes will be merged into the raw_file !!!\n\n");
}

off_t
get_file_size(int fd)
{
	off_t size;
	struct stat fs_stat;
	int ret;

	if (ioctl(fd, BLKGETSIZE64, &size) == -1) {
		ret = fstat(fd, &fs_stat);
		if (ret == -1) {
			printf("Error getting file size(fd %d): %s\n", fd, strerror(errno));
			exit(1);
		}
		size = fs_stat.st_size;
	}

	return size;
}

int is_disk_full(int fd, off_t expectedSize)
{
	struct statfs diskInfo = {0}; 
	//unsigned long long freeSize;
	unsigned long long availableSize;
	unsigned long long blkSize = diskInfo.f_bsize;
	
	if(fd < 0) {
		printf("Error: fd < 0 in is_disk_full\n");
		return 1;
	}

	if(fstatfs(fd, &diskInfo) != 0)
	{
		printf("Failed to get fs(fd %d) stat: %s\n", fd, strerror(errno));
		return 1;
	}

	blkSize = diskInfo.f_bsize;
	//availableSize = diskInfo.f_bfree * blocksize;
    availableSize = diskInfo.f_bavail * blkSize;

	if(expectedSize > availableSize)
		printf("is_disk_full: expected size is %llu MB, avaliable size is %llu MB\n",
			expectedSize/(1024*1024), availableSize/(1024*1024));

	return (expectedSize > availableSize);
}

char *check_path_safe(char *path)
{
	char *retStr;
	size_t pathLen;
	int i;
	
	if(!path)
		return NULL;

	retStr = path;
	pathLen = strlen(path);
	for(i=0; i<pathLen; i++) {
		if(path[i] == '.') {
			path[i] = '.';
			break;
		}
	}

	return retStr;
}

int main(int argc, char *argv[])
{  
	int raw_fd = -1, mdf_fd = -1;
	int udf_fd = -1, new_fd = -1;
	int mdf_rst_fd = -1, udf_rst_fd = -1;
	int pid_fd = -1;
	off_t required_disk_size;
	off_t raw_file_len,seek_len;
	off_t i,mdf_cache_num;
	int ret = 0, index;
	uint64_t realRdLen;
	uint32_t *mdf_cache = NULL;
	off_t mdf_cache_len,udf_len;
	off_t rdlen,offset,pages_nr;
	off_t pid_offset;
	unsigned char *tmpbuf = NULL;
	char *raw_file = NULL, *ud_file = NULL;
	char *md_file = NULL, *new_file = NULL;
	char *pid_file = NULL;
	//char failed_file_name[MAX_STRING_LEN] = {0};
	char calc_progress = 0;
	char enable_restore = 0;
	char verify_image_only = 0;
	int  last_index_parsed = 0;
	int argCount;
	char progressBuf[8] = {0};
	int progressBar = 0;
	size_t pb_len;
	char *p_mdf_rst_file = NULL;
	char *p_udf_rst_file = NULL;

	argCount = argc;
	index = 1;

	if(argCount < 4 || argCount > 8) {
		usage();
		return -1;
	}
	
parse_ext_options:
	if( strcmp(argv[index], PROGRESS_FLAG) == 0) {
		if(calc_progress == 1) {
			usage();
			return -1;
		}
		index++;
		argCount--;
		calc_progress = 1;
	}

	if( strcmp(argv[index], RESTORE_FLAG) == 0) {
		if(enable_restore == 1) {
			usage();
			return -1;
		}
		index++;
		argCount--;
		enable_restore = 1;
	}

	if( strcmp(argv[index], VERIFY_ONLY_FLAG) == 0) {
		if(verify_image_only == 1) {
			usage();
			return -1;
		}
		index++;
		argCount--;
		verify_image_only = 1;
	}

	// the option may comes like: -p || -r || -p -r || -r -p || NULL
	//if(index == 2 && ext_opts_reparsed == 0) {
	//	ext_opts_reparsed = 1;
	//	goto parse_ext_options;
	//}

	if((index > 1 && index < 4) && 
		last_index_parsed != index) {
		last_index_parsed = index;
		goto parse_ext_options;
	}

	if(argCount < 4 || argCount > 5) {
		usage();
		return -1;
	}

	if(enable_restore) {
		p_mdf_rst_file = (unsigned char *)calloc(DB_MAX_STRING_LEN, sizeof(char));
		if(!p_mdf_rst_file) {
			printf("Failed to allocate mem(%d) for mdf rst file: %s\n", 
				DB_MAX_STRING_LEN, 
				strerror(errno));
			return -1;
		}
		
		p_udf_rst_file = (unsigned char *)calloc(DB_MAX_STRING_LEN, sizeof(char));
		if(!p_udf_rst_file) {
			printf("Failed to allocate mem(%d) for udf rst file: %s\n", 
				DB_MAX_STRING_LEN, 
				strerror(errno));
			free(p_mdf_rst_file);
			return -1;
		}
	}
	
	tmpbuf = (unsigned char *)malloc(UDF_PAGE_SIZE);
	if(!tmpbuf) {
		printf("Failed to allocate mem(%d) for tmpbuf: %s\n", 
			UDF_PAGE_SIZE, 
			strerror(errno));
		if(p_mdf_rst_file)
			free(p_mdf_rst_file);
		if(p_udf_rst_file)
			free(p_udf_rst_file);
		return -1;
	}

	raw_file = (char *)malloc(MAX_STRING_LEN);
	if(!raw_file) {
		printf("Failed to allocate mem(%d) for raw_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	ud_file = (char *)malloc(MAX_STRING_LEN);
	if(!ud_file) {
		printf("Failed to allocate mem(%d) for ud_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	md_file = (char *)malloc(MAX_STRING_LEN);
	if(!md_file) {
		printf("Failed to allocate mem(%d) for md_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	new_file = (char *)malloc(MAX_STRING_LEN);
	if(!new_file) {
		printf("Failed to allocate mem(%d) for new_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	pid_file = (char *)malloc(MAX_STRING_LEN);
	if(!pid_file) {
		printf("Failed to allocate mem(%d) for pid_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	pb_len = sizeof(progressBuf);

	//printf("off_t %d\n", sizeof(off_t));

	strncpy(raw_file, argv[index], MAX_STRING_LEN);
	raw_file[MAX_STRING_LEN - 1] = '\0';
	index++;
	strncpy(md_file, argv[index], MAX_STRING_LEN);
	md_file[MAX_STRING_LEN - 1] = '\0';
	index++;
	strncpy(ud_file, argv[index], MAX_STRING_LEN);
	ud_file[MAX_STRING_LEN - 1] = '\0';
	index++;

	if(enable_restore) {
		snprintf(p_mdf_rst_file, DB_MAX_STRING_LEN, "%s%s",
			md_file, RESTORE_FILE_EXT);
		snprintf(p_udf_rst_file, DB_MAX_STRING_LEN, "%s%s", 
			ud_file, RESTORE_FILE_EXT);
	}
	
	snprintf(pid_file, MAX_STRING_LEN, "%s%s", raw_file, PID_FILE_EXT);
	pid_fd = open(pid_file, O_CREAT|O_TRUNC|O_RDWR, 0666); 
	if (pid_fd < 0) {
		printf("Failed to create pid file (%s)\n", strerror(errno));
		ret = -1;
		goto error;
	}
	
	if(argCount == 5) {
		strncpy(new_file, argv[index], MAX_STRING_LEN);
		new_file[MAX_STRING_LEN - 1] = '\0';
		//snprintf(failed_file_name, MAX_STRING_LEN, "%s_%s", new_file, FAILED_FILE_EXT);

		raw_fd = open(raw_file, O_RDONLY);

		if(raw_fd >= 0) {
			raw_file_len = get_file_size(raw_fd);

			if(!verify_image_only) {
				new_fd = open(new_file, O_CREAT|O_TRUNC|O_RDWR, 0666); 
				if (new_fd < 0) {
					printf("Failed to open raw file (%s)\n", strerror(errno));
					ret = -1;
					goto error;
				}
				
				if(is_disk_full(new_fd, raw_file_len)) {
					printf("Error: no avaliable disk size for the new image file\n");
					ret = -1;
					goto error;
				}
				
	  		} else
				new_fd = raw_fd;
		}
	}else{
		raw_fd = open(raw_file, O_RDWR);
		new_fd = raw_fd;
		if(raw_fd >= 0) {
			raw_file_len = get_file_size(raw_fd);
		}
	}

	if (raw_fd < 0) {   
		printf("Failed to open raw file \"%s\": %s\n", raw_file, strerror(errno)); 
		ret = -1;
		goto error;
	}    

	mdf_fd = open(md_file, O_RDONLY);
	if (mdf_fd < 0) {
		printf("Failed to open meta data file \"%s\": %s\n", md_file, strerror(errno)); 
		ret = -1;
		goto error;
	}

	udf_fd = open(ud_file, O_RDONLY);
	if (udf_fd < 0) {
		printf("Failed to open user data file \"%s\": %s\n", ud_file, strerror(errno)); 
		ret = -1;
		goto error;
	}

	if(enable_restore) {
		mdf_rst_fd = open(p_mdf_rst_file, O_CREAT|O_TRUNC|O_RDWR, 0666);
		if (mdf_rst_fd < 0) {
			printf("Failed to open meta rst data file \"%s\": %s\n", 
				p_mdf_rst_file,
				strerror(errno)); 
			ret = -1;
			goto error;
		}

		udf_rst_fd = open(p_udf_rst_file, O_CREAT|O_TRUNC|O_RDWR, 0666);
		if (udf_rst_fd < 0) {
			printf("Failed to open user rst data file \"%s\": %s\n", 
				p_udf_rst_file, 
				strerror(errno)); 
			ret = -1;
			goto error;
		}
	}

	//raw_file_len = get_file_size(raw_fd);

	//
	// if ud file's len is 0,do not create new file
	//
	udf_len = get_file_size(udf_fd); 
	if(udf_len ==-1 || udf_len == 0){
		printf("Error: udf file len is 0\n");
		ret = -1;
		goto error;
	}

	//
	// init mdf_cache array
	// this is for compatibility with previous version,
	// we do not round the mdf_cache_len
	// //
	mdf_cache_len = raw_file_len / UDF_PAGE_SIZE  + 1;	
	printf("raw_file_len is %llu,mdf_cache_len is %llu\n",(unsigned long long)raw_file_len, (unsigned long long)mdf_cache_len);

	if(enable_restore) {
		required_disk_size = (udf_len + mdf_cache_len*sizeof(uint32_t));
		if((new_fd >= 0) && new_fd != raw_fd)
			required_disk_size += raw_file_len;
		
		if(is_disk_full(new_fd, required_disk_size)) {
			printf("Error: no avaliable disk size for the restore files\n");
			ret = -1;
			goto error;
		}
	}
	
	mdf_cache = (uint32_t *)calloc(mdf_cache_len, sizeof(uint32_t));
	if (mdf_cache == NULL) {
		printf("Failed to allocate %d memory\n", mdf_cache_len); 
		ret = -1;
		goto error;
	}  

	//
	// fill mdf_cache with content from map file.
	//
	rdlen = 0;
	offset=0;  

	printf("Start to read metadata into mdf_cache...\n");

	if(read(mdf_fd, mdf_cache, mdf_cache_len*sizeof(uint32_t)) != 
			mdf_cache_len*sizeof(uint32_t)) {
		printf("Error: could not read all metadata file (%s)\n", 
			strerror(errno));
		ret = -1;
		goto error;
	}

	if(enable_restore) {
		if (write(mdf_rst_fd, mdf_cache, mdf_cache_len*sizeof(uint32_t)) !=	
				mdf_cache_len*sizeof(uint32_t)){
		    printf("Failed to write mdf_rst_fd: %s\n", 
				strerror(errno));
		    ret = -1;
			goto error;
		}
	}

	//
	// create a merged file
	//
	if(!verify_image_only)
		printf("Start to merge file...\n");
	else
		printf("Start to verify file...\n");
	
	if(calc_progress) {
		progressBar = 0;
		snprintf(progressBuf, 8, "%d", progressBar);
		pid_offset = lseek(pid_fd, (off_t)0, SEEK_SET);
		if (pid_offset == -1) {
			printf("Failed to lseek udf_fd %llu: %s\n",(unsigned long long)0, strerror(errno)); 
		} else {
			if (write(pid_fd, progressBuf, pb_len) != pb_len){
		    		printf("Failed to write pid_fd: %s\n", strerror(errno));
			}
		}
	}

	pages_nr = raw_file_len/UDF_PAGE_SIZE + 1;
	for (i=0; i<pages_nr; i++){
		realRdLen = UDF_PAGE_SIZE;
		if(i == (pages_nr - 1)){
		  	realRdLen = UDF_PAGE_SIZE - ((pages_nr*UDF_PAGE_SIZE) - raw_file_len);
		}
		if(mdf_cache[i] == (uint32_t)-1) {
			rdlen = 0;
			if(new_fd != raw_fd) {
				offset = lseek(raw_fd,(off_t)(i*UDF_PAGE_SIZE),SEEK_SET);    
				if (offset == -1) {
					printf("Failed to lseek raw_fd %llu: %s\n", (unsigned long long)(i*UDF_PAGE_SIZE), strerror(errno)); 
					ret = -1;
					goto error;
				}

				rdlen = read(raw_fd, tmpbuf, realRdLen);
				if (rdlen != realRdLen) {
					printf("Failed to read raw_fd: %s\n", strerror(errno));
					ret = -1;
					goto error;
				}

				offset = lseek(new_fd,(off_t)(i*UDF_PAGE_SIZE),SEEK_SET);    
				if (offset == -1) {
					printf("Failed to lseek new_fd0 %llu: %s\n", (unsigned long long)(i*UDF_PAGE_SIZE), strerror(errno)); 
					ret = -1;
					goto error;
				}
				
				if (write(new_fd, tmpbuf, realRdLen) !=	realRdLen){
				    printf("Failed to write new_fd0: %s\n", strerror(errno));
				    ret = -1;
					goto error;
				}
			}
		} else {
			mdf_cache_num = mdf_cache[i];

			if(enable_restore) {				
				offset = lseek(raw_fd,(off_t)(i*UDF_PAGE_SIZE),SEEK_SET);    
				if (offset == -1) {
					printf("Failed to lseek raw_fd1 %llu: %s\n", 
						(unsigned long long)(i*UDF_PAGE_SIZE), 
						strerror(errno)); 
					ret = -1;
					goto error;
				}

				rdlen = read(raw_fd, tmpbuf, realRdLen);
				if (rdlen != realRdLen) {
					printf("Failed to read raw_fd1 %s, real_rd_len=%llx, expect_rd_len=%llx, i=%llx, seek=%llx\n", 
						strerror(errno), 
						rdlen, 
						realRdLen,
						i,
						((unsigned long long)(i*UDF_PAGE_SIZE)));
					ret = -1;
					goto error;
				}

				// We need to backup the data in raw image file as restoration data
				offset = lseek(udf_rst_fd, 
							(off_t)(mdf_cache_num*UDF_PAGE_SIZE), 
							SEEK_SET);
				if (offset == -1) {
					printf("Failed to lseek udf_rst_fd %llu: %s\n",
						(unsigned long long)(mdf_cache_num*UDF_PAGE_SIZE), 
						strerror(errno)); 
					ret = -1;
					goto error;
				}

				if (write(udf_rst_fd, tmpbuf, realRdLen) !=	realRdLen){
				    printf("Failed to write udf_rst_fd: %s\n", strerror(errno));
				    ret = -1;
					goto error;
				}
			}

			if(!verify_image_only) {

				offset = lseek(udf_fd, (off_t)(mdf_cache_num*UDF_PAGE_SIZE), SEEK_SET);
				if (offset == -1) {
					printf("Failed to lseek udf_fd %llu: %s\n",
						(unsigned long long)(mdf_cache_num*UDF_PAGE_SIZE),
						strerror(errno)); 
					ret = -1;
					goto error;
				}

				rdlen = read(udf_fd, tmpbuf, realRdLen);
				if (rdlen != realRdLen) {
					printf("Failed to read udf_fd %s, real_rd_len=%llx, expect_rd_len=%llx, i=%llx, seek=%llx\n", 
						strerror(errno), 
						rdlen, 
						realRdLen,
						i,
						(off_t)(mdf_cache_num));
					ret = -1;
					goto error;
				}
				//if(new_fd == raw_fd) {
					offset = lseek(new_fd,(off_t)(i*UDF_PAGE_SIZE),SEEK_SET);    
					if (offset == -1) {
						printf("Failed to lseek new_fd %llu: %s\n", 
							(unsigned long long)(i*UDF_PAGE_SIZE), 
							strerror(errno)); 
						ret = -1;
						goto error;
					}
				//}				
				
				if (write(new_fd, tmpbuf, realRdLen) !=	realRdLen){
				    printf("Failed to write new_fd: %s\n", strerror(errno));
				    ret = -1;
					goto error;
				}
			}
		}

		if(calc_progress) {
			progressBar = ((i+1)*100)/pages_nr;
			snprintf(progressBuf, 8, "%d", progressBar);
			pid_offset = lseek(pid_fd, (off_t)0, SEEK_SET);
			if (pid_offset == -1) {
				printf("Failed to lseek udf_fd %llu: %s\n",(unsigned long long)0, strerror(errno)); 
			} else {
				if (write(pid_fd, progressBuf, pb_len) != pb_len){
			    		printf("Failed to write pid_fd: %s\n", strerror(errno));
				}
			}
		}
	}  

	fsync(new_fd);

	if(enable_restore)
		printf("Image restoration file(%s, %s) created successfully!\n",
			p_mdf_rst_file, p_udf_rst_file);
	
	if(!verify_image_only)
		printf("merge image file successfully!\n");
	else
		printf("Image file verification completed successfully!\n");

	ret = 0;
	
	goto sys_exit;

error:	
	if ((new_fd != raw_fd) && new_fd >= 0) {
		close (new_fd);
		new_fd = -1;
		//rename(new_file, failed_file_name);
		new_file = check_path_safe(new_file);
		if(new_file)
			unlink(new_file);
	}

sys_exit:

	if (udf_fd >= 0)
		close (udf_fd);
	if (mdf_fd >= 0)
		close (mdf_fd);
	
	if (udf_rst_fd >= 0) {
		close (udf_rst_fd);
		if(ret != 0) {
			p_udf_rst_file = check_path_safe(p_udf_rst_file);
			if(p_udf_rst_file)
				unlink(p_udf_rst_file);
		}
	}
	
	if (mdf_rst_fd >= 0) {
		close (mdf_rst_fd);
		if(ret != 0) {
			p_mdf_rst_file = check_path_safe(p_mdf_rst_file);
			if(p_mdf_rst_file)
				unlink(p_mdf_rst_file);
		}
	}
	
	if ((new_fd != raw_fd) && new_fd >= 0)
		close (new_fd);

	if (raw_fd >= 0)
		close (raw_fd);
	
	if (pid_fd >= 0) {
		close (pid_fd);
		pid_file = check_path_safe(pid_file);
		if(pid_file)
			unlink(pid_file);
	}

	if (mdf_cache != NULL)
		free(mdf_cache);

	if(p_mdf_rst_file)
		free(p_mdf_rst_file);

	if(p_udf_rst_file)
		free(p_udf_rst_file);

	if(tmpbuf)
		free(tmpbuf);

	if(raw_file)
		free(raw_file);

	if(ud_file)
		free(ud_file);

	if(md_file)
		free(md_file);

	if(new_file)
		free(new_file);

	if(pid_file)
		free(pid_file);

	return ret;
}
