/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <errno.h>
#include "predefine.h"
#include "tcprotocol.h"
#include "client.h"
#include "netcfg.h"
#include "debug.h"


#define DFT_MTU	1500

int
rcv_pkt(Client *p_client, uint8_t *buf, int pkt_len)
{
#ifdef SDL_BUILD_ENABLE
	return SSL_read(p_client->p_ssl, buf, pkt_len);
#else
	return read(p_client->sock, buf, pkt_len);
#endif
}

int
snd_pkt(Client *p_client, uint8_t *buf, int pkt_len)
{
#ifdef SDL_BUILD_ENABLE
	return SSL_write(p_client->p_ssl, buf, pkt_len);
#else
	return write(p_client->sock, buf, pkt_len);
#endif
}

int get_eth_mac(int eth_fd, char *eth_name, uint8_t *eth_mac)
{
	struct ifreq req;

	if(eth_fd == -1)
		return -1;

  	strncpy(req.ifr_name, eth_name, IFNAMSIZ);
	req.ifr_name[IFNAMSIZ - 1] = '\0';
	if (ioctl(eth_fd, SIOCGIFHWADDR, &req) == -1) {
		ERR("Failed to get mac addr of %s: %s\n", eth_name, strerror(errno));
		return -1;
	}
	
	memcpy(eth_mac, req.ifr_hwaddr.sa_data, 6);

	return 0;
}

 
/**
 * Connect a server's socket.
 *
 * @param serve the server we want to connect.
 **/
int init_socket(SERVER_CONFIG *s_config) {
	struct sockaddr_in addrin;

	int addrinlen = sizeof(addrin);
	int sock_flags;
	int af, yes=1;

	if(!s_config)
		return -1;
		
	af = AF_INET;

	if ((s_config->socket = socket(af, SOCK_STREAM, IPPROTO_TCP)) < 0)
	{
		ERR("Error calling socket(): %s\n", strerror(errno));
		return errno;
	}

	/* lose the pesky "Address already in use" error message */
	if (setsockopt(s_config->socket,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int)) == -1) {
		ERR("setsockopt SO_REUSEADDR");
	        return -1;
	}
	if (setsockopt(s_config->socket,SOL_SOCKET,SO_KEEPALIVE,&yes,sizeof(int)) == -1) {
		ERR("setsockopt SO_KEEPALIVE");
		return -1;
	}

	/* make the listening socket non-blocking */
	if ((sock_flags = fcntl(s_config->socket, F_GETFL, 0)) == -1) {
		ERR("setsockopt F_GETFL");
		return -1;
	}
	if (fcntl(s_config->socket, F_SETFL, sock_flags | O_NONBLOCK) == -1) {
		ERR("setsockopt F_GETFL F_SETFL O_NONBLOCK");
		return -1;
	}

	addrin.sin_family = AF_INET;
	addrin.sin_port = htons(s_config->port);
	if(!inet_aton(s_config->ip, &(addrin.sin_addr)))
	{
		ERR("could not parse listen address");
		return -1;
	}
	if (bind(s_config->socket, (struct sockaddr *) &addrin, addrinlen) < 0)
	{
		ERR("bind: %s\n", strerror(errno));
		return errno;
	}
	
	if (listen(s_config->socket, 1) < 0)
	{
		ERR("listen: %s\n", strerror(errno));
		return errno;
	}

	DEBUG("Listening on %s:%ld\n", s_config->ip, s_config->port);
	return 0;
}

