/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <getopt.h>
#include "predefine.h"
#include "process.h"
#include "db.h"
#include "fileops.h"
#include "debug.h"

static int  g_single_inst_lock_fd = -1;
static SERVER_CONFIG g_s_config = {0};

SERVER_CONFIG *get_server_config()
{
	return &g_s_config;
}

#ifdef SDL_BUILD_ENABLE

int 
setup_cert_params(char **pp_cert_full_file_path,
	char *cert_dir,
	char *cert_file
	) {
	int ret;
	char *tmp_file = NULL;
	char *sn_format;
	size_t str_len, cert_dir_len;
	
	if(!cert_dir || 
		!pp_cert_full_file_path || 
		!cert_file)
		return -1;

	*pp_cert_full_file_path = NULL;
	cert_dir_len = strlen(cert_dir);
	str_len = cert_dir_len + strlen(cert_file) + 4;
	tmp_file = (char *)calloc(str_len, sizeof(char));
	if(!tmp_file) {
		ERR("Error: failed to alloc mem for ca file\n");
		ret = -1;
		goto invalid;
	}

	if(cert_dir[cert_dir_len - 1] != '/')
		sn_format = "%s/%s";
	else
		sn_format = "%s%s";
	
	snprintf(tmp_file, str_len, sn_format, cert_dir, cert_file);	
		
	*pp_cert_full_file_path = tmp_file;
	ret = 0;

invalid:
	if(ret != 0) {
		if(tmp_file)
			free(tmp_file);
	}
	return ret;
}
#endif

int parse_params(int argc, char *argv[], SERVER_CONFIG *s_config) {
	int ret;
	int c,i, sum;
	size_t str_len;
	char *home_var, *tmp_path = NULL;
	struct option long_options[] = {
		{"ip", required_argument, NULL, 'i'},
		{"port", required_argument, NULL, 'p'},
		{"nic", required_argument, NULL, 'n'},
		{"lic_file", required_argument, NULL, 'l'},
#ifdef SDL_BUILD_ENABLE
		{"ca_dir", required_argument, NULL, 'c'},
#endif
		{0, 0, 0, 0}
	};	

	if(!s_config) {
		ERR("ERROR: server configuration object is NULL\n");
		return -1;
	}

#ifdef SDL_BUILD_ENABLE	
	while((c = getopt_long(argc, argv, "i:p:n:l:c:", 
#else
	while((c = getopt_long(argc, argv, "i:p:n:l:", 
#endif
		long_options, &i))>=0) {		
		switch (c) {		
		case 'i':
			strncpy(s_config->ip, optarg, 64);
			s_config->ip[63] = '\0';
			break;
		case 'p':
			//s_config->port = strtol(optarg, NULL, 0);
			if(str_to_long(optarg, &(s_config->port)) == -1) {
				ERR("Incorrect parameter for port: %s\n", optarg);
				goto invalid;
			}
			break;
		case 'n':
			strncpy(s_config->ethName, optarg, 64);
			s_config->ethName[63] = '\0';
			break;
#ifdef SDL_BUILD_ENABLE
		case 'c':
			str_len = strlen(optarg) + 2;
			if(tmp_path)
				free(tmp_path);
			tmp_path = (char *)calloc(str_len, sizeof(char));
			if(!tmp_path) {
				ERR("Failed to allocate mem for cert dir\n");
				goto invalid;
			}
			strncpy(tmp_path, optarg, str_len);
			break;
#endif
		case 'l':
			strncpy(s_config->license_file, optarg, MAX_STRING_LEN);
			s_config->license_file[MAX_STRING_LEN - 1] = '\0';
			break;
		case '?' :  
            if ( optopt == 'i' || optopt == 'p' || optopt == 'n' )  
                ERR("Error: option -%c requires an argument\n", optopt);  
            //else if ( isprint(optopt) )  
            //    ERR("Error: unknown option '-%c'\n", optopt);  
            else  
                ERR("Error: unknown option character '0x%x'\n", optopt);
			goto invalid;
		default:
			ERR("Unsupported command '%d'!\n", c);
			goto invalid;
		}
	}
	
	if(s_config->ip[0] == 0)
	{
		DEBUG("Ip is not set, will use \"%s\" by default\n", SRV_IP);
		strncpy(s_config->ip, SRV_IP, MAX_IP_LEN);
		s_config->ip[MAX_IP_LEN - 1] = '\0';
	}
	if(!s_config->port)
	{
		DEBUG("Port is not set, will use \"%d\" by default\n", SRV_PORT);
		s_config->port = SRV_PORT;
	}
	if(s_config->ethName[0] == 0)
	{
		DEBUG("Nic is not set, will use \"%s\" by default\n", NIC_NAME);
		strncpy(s_config->ethName, NIC_NAME, MAX_NIC_NAME_LEN);
		s_config->ethName[MAX_NIC_NAME_LEN - 1] = '\0';
	}

	if(!s_config->license_file || 
		s_config->license_file[0] == 0)
	{
		char *tmp_file;
		tmp_file = get_lic_file();
		if(!tmp_file) {
			str_len = strlen(DFT_LIC_FILE) + 2;
			tmp_file = (char *)calloc(str_len, sizeof(char));
			if(!tmp_file) {
				ERR("Error: failed to alloc mem for lic file\n");
				goto invalid;
			}
			s_config->license_file = tmp_file;
			snprintf(s_config->license_file, str_len, "%s", DFT_LIC_FILE);
		} else {
			s_config->license_file = tmp_file;
			//snprintf(s_config->license_file, MAX_STRING_LEN, "%s", tmp_file);
			//free(tmp_file);
		}
	}

#ifdef SDL_BUILD_ENABLE
	
	if(!tmp_path || 
		tmp_path[0] == 0){
		tmp_path = get_ca_dir();
	}

	if(!tmp_path) {
		ERR("Failed to get cert dir\n");
		goto invalid;
	}

#ifdef ENABLE_TC_MUTUAL_TLS
	ret = setup_cert_params(&s_config->ca_file,
					tmp_path,
					DFT_CA_FILE);
	if(ret != 0) {
		ERR("Failed to set up ca file path: %s\n");
		goto invalid;
	}
#endif
	
	ret = setup_cert_params(&s_config->cert_file,
			tmp_path,
			DFT_CERT_FILE);
	if(ret != 0) {
		ERR("Failed to set up cert file path: %s\n", tmp_path);
		goto invalid;
	}

	ret = setup_cert_params(&s_config->key_file,
			tmp_path,
			DFT_KEY_FILE);
	if(ret != 0) {
		ERR("Failed to set up key file path: %s\n", tmp_path);
		goto invalid;
	}
	
#endif

	home_var = get_home_var_dir();
	if(!home_var)
		snprintf(s_config->single_inst_file, 
			MAX_SI_FILE_LEN, 
			SINGLE_INST_LOCK_FORMAT, 
		    DFT_HOME_VAR, DFT_APP_NAME,
		    s_config->ethName, 
			s_config->port);
	else {
		snprintf(s_config->single_inst_file, 
			MAX_SI_FILE_LEN, 
			SINGLE_INST_LOCK_FORMAT, 
		    home_var, 
		    DFT_APP_NAME, 
		    s_config->ethName, 
			s_config->port);

		free(home_var);
	}

	s_config->login_failure_interval = get_login_failure_interval();
	
	return 0;
	
invalid:
	if(tmp_path)
		free(tmp_path);
	
	return -1;
}

int str_to_long(char *sourceStr, long *retVal)
{
	char * endStr = NULL;

	if(!retVal)
		return -1;
	
	errno = 0;
	*retVal = strtol(sourceStr, &endStr, 10);
	
	if (endStr != NULL && endStr[0] != 0) {
		ERR("Failed to convert str to long: source is %s, stop at: %s\n", 
			sourceStr, endStr); 
		return -1;
	}
	
	if ((errno == ERANGE && (*retVal == LONG_MAX || *retVal == LONG_MIN))    
	 	 || (errno != 0 && *retVal == 0)) {
		ERR("Failed to format string to num: %s\n", strerror(errno)); 
		return -1;
	}
	return 0;
}

void single_inst_lockfile_cleanup(char *proc_file_name)
{
    if (g_single_inst_lock_fd > 0) {
        close(g_single_inst_lock_fd);		
        g_single_inst_lock_fd = -1;
  		if(proc_file_name && unlink(proc_file_name) != 0) {
  			ERR("Failed to delete file (%s): %s\n", proc_file_name, strerror(errno));
		}
    }
}

int is_instance_running(char *proc_file_name)
{
	char tmp_file[MAX_SI_FILE_LEN + 1];
	char *tmp_dir;
    if(!proc_file_name)
		return 0;

	//BUGBUG: need to remove this in future...
	snprintf(tmp_file, MAX_SI_FILE_LEN, "%s", proc_file_name);

    g_single_inst_lock_fd = open(tmp_file, O_CREAT|O_RDWR, 0644);
    if (g_single_inst_lock_fd < 0) {
		//if the path does not exsit
		if(errno == 2){
	        DEBUG("Failed opening lock file (%s): %s, try creating parent dir to fix it\n", 
				tmp_file, strerror(errno));
			tmp_dir = get_parent_dir(tmp_file);
			if(tmp_dir) {
				create_dir(tmp_dir);
				free(tmp_dir);
				g_single_inst_lock_fd = open(tmp_file, O_CREAT|O_RDWR, 0644);			
			}
		}
		if (g_single_inst_lock_fd < 0) {
	        ERR("Failed opening lock file (%s): %s\n", tmp_file, strerror(errno));
	        return 1;
		}
    }

    if (flock(g_single_inst_lock_fd, LOCK_EX | LOCK_NB) != 0) {
		close(g_single_inst_lock_fd);
    	g_single_inst_lock_fd = -1;
        return 1;
    }

    return 0;
}

#if RUN_IN_DAEMON_MODE
void daemonize() {
	if(daemon(1, 0)<0) {
		ERR("Error in daemonize: %s\n", strerror(errno));
	}
}
#else
void daemonize() {
}
#endif

void setup_sighandler(P_SIG_HANDLER psh)
{
	if(signal(SIGINT, psh) == SIG_ERR)
		ERR("signal error (%s)", strerror(errno));

	if(signal(SIGTERM, psh) == SIG_ERR)
		ERR("signal error (%s)", strerror(errno));	

}

