/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <linux/fs.h>
#include <sys/statfs.h> 

//typedef unsigned long long u64;
//typedef unsigned int uint32_t;

#define UDF_PAGE_SIZE 		(2*1024*1024) 
#define MAX_STRING_LEN 		256
#define PID_FILE_EXT		".merging"
#define PROGRESS_FLAG		"-p"

void usage()
{
  printf("tc-disk-ar [%s <pid_file>] target_md_file target_ud_file source_md_file source_ud_file" \
  		" [output_md_file output_ud_file]\n\n", 
  	PROGRESS_FLAG);
  printf("\tmerge the content in source mdf/udf to target or output mdf/udf files\n");
  printf("\t%s <pid_file>: update merging progress in pid file\n", PROGRESS_FLAG);
   printf("Note: if parameter \"output_md_file output_ud_file\" is absent, " \
  		"all changes will be merged into the target mdf/udf files !!!\n\n");
}

off_t
get_file_size(int fd)
{
	off_t size;
	struct stat fs_stat;
	int ret;

	if (ioctl(fd, BLKGETSIZE64, &size) == -1) {
		ret = fstat(fd, &fs_stat);
		if (ret == -1) {
			printf("Error getting file size(fd %d): %s\n", fd, strerror(errno));
			exit(1);
		}
		size = fs_stat.st_size;
	}

	return size;
}

char *check_path_safe(char *path)
{
	char *retStr;
	size_t pathLen;
	int i;
	
	if(!path)
		return NULL;

	retStr = path;
	pathLen = strlen(path);
	for(i=0; i<pathLen; i++) {
		if(path[i] == '.') {
			path[i] = '.';
			break;
		}
	}

	return retStr;
}

int main(int argc, char *argv[])
{  
	int new_mdf_fd = -1, new_udf_fd = -1;
	int old_mdf_fd = -1, old_udf_fd = -1;
	int output_mdf_fd = -1, output_udf_fd = -1;
	int pid_fd = -1;
	off_t seek_len;
	off_t i,mdf_cache_num;
	off_t max_new_mdf_cache_num;
	int ret = 0, index;
	uint64_t realRdLen;
	uint32_t *new_mdf_cache = NULL;
	uint32_t *old_mdf_cache = NULL;
	off_t old_mdf_cache_len;
	off_t new_mdf_cache_len,udf_len;
	off_t rdlen,offset,pages_nr;
	off_t pid_offset;
	unsigned char *tmpbuf = NULL;
	char *new_md_file = NULL, *new_ud_file = NULL; 
	char *old_md_file = NULL, *old_ud_file = NULL;
	char *output_md_file = NULL, *output_ud_file = NULL;
	//char newfile[MAX_STRING_LEN]={0};
	char pid_file[MAX_STRING_LEN]={0};
	char calc_progress = 0;
	int argCount;
	char progressBuf[8] = {0};
	int progressBar = 0;
	size_t pb_len;

	argCount = argc;
	index = 1;

	if(argCount < 5 || argCount > 9) {
		usage();
		return -1;
	}
	
	if( strcmp(argv[index], PROGRESS_FLAG) == 0) {
		if(calc_progress == 1) {
			usage();
			return -1;
		}
		index++;
		argCount--;
		strncpy(pid_file, argv[index], MAX_STRING_LEN);
		pid_file[MAX_STRING_LEN - 1] = '\0';
		index++;
		argCount--;
		calc_progress = 1;
	}
	
	if(argCount != 5 && argCount != 7) {
		usage();
		return -1;
	}

	tmpbuf = (unsigned char *)malloc(UDF_PAGE_SIZE);
	if(!tmpbuf) {
		printf("Failed to allocate mem(%d) for tmpbuf: %s\n", 
			UDF_PAGE_SIZE, 
			strerror(errno));
		return -1;
	}

	new_md_file = (char *)malloc(MAX_STRING_LEN);
	if(!new_md_file) {
		printf("Failed to allocate mem(%d) for new_md_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	new_ud_file = (char *)malloc(MAX_STRING_LEN);
	if(!new_ud_file) {
		printf("Failed to allocate mem(%d) for new_ud_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	old_md_file = (char *)malloc(MAX_STRING_LEN);
	if(!old_md_file) {
		printf("Failed to allocate mem(%d) for old_md_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	old_ud_file = (char *)malloc(MAX_STRING_LEN);
	if(!old_ud_file) {
		printf("Failed to allocate mem(%d) for old_ud_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	output_md_file = (char *)malloc(MAX_STRING_LEN);
	if(!output_md_file) {
		printf("Failed to allocate mem(%d) for output_md_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	output_ud_file = (char *)malloc(MAX_STRING_LEN);
	if(!output_ud_file) {
		printf("Failed to allocate mem(%d) for output_ud_file: %s\n", 
			MAX_STRING_LEN, 
			strerror(errno));
		goto error;
	}

	pb_len = sizeof(progressBuf);

	//printf("off_t %d\n", sizeof(off_t));

	strncpy(new_md_file, argv[index], MAX_STRING_LEN);
	new_md_file[MAX_STRING_LEN - 1] = '\0';
	index++;
	strncpy(new_ud_file, argv[index], MAX_STRING_LEN);
	new_ud_file[MAX_STRING_LEN - 1] = '\0';
	index++;
	strncpy(old_md_file, argv[index], MAX_STRING_LEN);
	old_md_file[MAX_STRING_LEN - 1] = '\0';
	index++;
	strncpy(old_ud_file, argv[index], MAX_STRING_LEN);
	old_ud_file[MAX_STRING_LEN - 1] = '\0';
	index++;

	if(argCount == 7) {
		strncpy(output_md_file, argv[index], MAX_STRING_LEN);
		output_md_file[MAX_STRING_LEN - 1] = '\0';
		index++;
		strncpy(output_ud_file, argv[index], MAX_STRING_LEN);
		output_ud_file[MAX_STRING_LEN - 1] = '\0';
		index++;

		output_mdf_fd = open(output_md_file, O_CREAT|O_TRUNC|O_RDWR, 0666); 
		if (output_mdf_fd < 0) {
			printf("Failed to open output meta data file \"%s\": %s\n", 
				output_md_file, 
				strerror(errno)); 
			ret = -1;
			goto error;
		}

		output_udf_fd = open(output_ud_file, O_CREAT|O_TRUNC|O_RDWR, 0666); 
		if (output_udf_fd < 0) {
			printf("Failed to open output user data file \"%s\": %s\n", 
				output_ud_file, 
				strerror(errno)); 
			ret = -1;
			goto error;
		}
		
		new_mdf_fd = open(new_md_file, O_RDONLY);
		if (new_mdf_fd < 0) {
			printf("Failed to open new meta data file \"%s\": %s\n", 
				new_md_file, 
				strerror(errno)); 
			ret = -1;
			goto error;
		}

		new_udf_fd = open(new_ud_file, O_RDONLY);
		if (new_udf_fd < 0) {
			printf("Failed to open new user data file \"%s\": %s\n", 
				new_ud_file, 
				strerror(errno)); 
			ret = -1;
			goto error;
		}
	}else {
		new_mdf_fd = open(new_md_file, O_RDWR);
		if (new_mdf_fd < 0) {
			printf("Failed to open new meta data file \"%s\": %s\n", 
				new_md_file, 
				strerror(errno)); 
			ret = -1;
			goto error;
		}

		new_udf_fd = open(new_ud_file, O_RDWR);
		if (new_udf_fd < 0) {
			printf("Failed to open new user data file \"%s\": %s\n", 
				new_ud_file, 
				strerror(errno)); 
			ret = -1;
			goto error;
		}		
		output_mdf_fd = new_mdf_fd;
		output_udf_fd = new_udf_fd;
	}	
	
	//snprintf(pid_file, MAX_STRING_LEN, "%s%s", new_md_file, PID_FILE_EXT);
	pid_fd = open(pid_file, O_CREAT|O_TRUNC|O_RDWR, 0666); 
	if (pid_fd < 0) {
		printf("Failed to create pid file (%s)\n", strerror(errno));
		ret = -1;
		goto error;
	}

	old_mdf_fd = open(old_md_file, O_RDONLY);
	if (old_mdf_fd < 0) {
		printf("Failed to open old meta data file \"%s\": %s\n", 
			old_md_file, 
			strerror(errno)); 
		ret = -1;
		goto error;
	}

	old_udf_fd = open(old_ud_file, O_RDONLY);
	if (old_udf_fd < 0) {
		printf("Failed to open old user data file \"%s\": %s\n", 
			old_ud_file, 
			strerror(errno)); 
		ret = -1;
		goto error;
	}

	old_mdf_cache_len = get_file_size(old_mdf_fd);
	new_mdf_cache_len = get_file_size(new_mdf_fd);

	if (new_mdf_cache_len != old_mdf_cache_len) {
		printf("Error: newSize(%llu) != oldSize(%llu) of meta data file\n", 
			(unsigned long long)new_mdf_cache_len, 
			(unsigned long long)old_mdf_cache_len); 
		ret = -1;
		goto error;
	}

	udf_len = (old_mdf_cache_len % sizeof(uint32_t));
	if(udf_len != 0) {
		printf("Error: new/old_mdf_cache_len(%llu) was not multipe of sizeof(uint32_t) %d\n", 
			(unsigned long long)old_mdf_cache_len, 
			sizeof(uint32_t)); 
		ret = -1;
		goto error;
	}

	old_mdf_cache_len = (old_mdf_cache_len / sizeof(uint32_t));
	new_mdf_cache_len = old_mdf_cache_len;	

	printf("mdf_cache_len is %llu\n",(unsigned long long)new_mdf_cache_len);
	
	//
	// if ud file's len is 0,do not create new file
	//
	udf_len = get_file_size(new_udf_fd); 
	if(udf_len == -1 || udf_len == 0){
		printf("Error: new udf file len is 0\n");
		ret = -1;
		goto error;
	}

	udf_len = get_file_size(old_udf_fd); 
	if(udf_len == -1 || udf_len == 0){
		printf("Error: old udf file len is 0\n");
		ret = -1;
		goto error;
	}
	
	new_mdf_cache = (uint32_t *)calloc(new_mdf_cache_len, sizeof(uint32_t));
	if (new_mdf_cache == NULL) {
		printf("Failed to allocate %d memory for new mdf cache\n", new_mdf_cache_len); 
		ret = -1;
		goto error;
	}

	old_mdf_cache = (uint32_t *)calloc(old_mdf_cache_len, sizeof(uint32_t));
	if (old_mdf_cache == NULL) {
		printf("Failed to allocate %d memory for new mdf cache\n", old_mdf_cache_len); 
		ret = -1;
		goto error;
	}

	//
	// fill mdf_cache with content from map file.
	//
	rdlen = 0;
	offset = 0;  

	printf("Start to read metadata into mdf_cache...\n");
	printf("new mdf is: %s, new udf is: %s\n", new_md_file, new_ud_file);
	printf("old mdf is: %s, old udf is: %s\n", old_md_file, old_ud_file);
	if(output_mdf_fd != new_mdf_fd)
		printf("output mdf is: %s, output udf is: %s\n", output_md_file, output_ud_file);
	else
		printf("output mdf is: %s, output udf is: %s\n", new_md_file, new_ud_file);

	if(read(new_mdf_fd, new_mdf_cache, new_mdf_cache_len*sizeof(uint32_t)) != 
			new_mdf_cache_len*sizeof(uint32_t)) {
		printf("Error: could not read all new metadata file (%s)\n", 
			strerror(errno));
		ret = -1;
		goto error;
	}

	if(read(old_mdf_fd, old_mdf_cache, old_mdf_cache_len*sizeof(uint32_t)) != 
			old_mdf_cache_len*sizeof(uint32_t)) {
		printf("Error: could not read all old metadata file (%s)\n", 
			strerror(errno));
		ret = -1;
		goto error;
	}

	//
	// create a merged file
	//
	printf("Start to merge file...\n");
	
	if(calc_progress) {
		progressBar = 0;
		snprintf(progressBuf, 8, "%d", progressBar);
		pid_offset = lseek(pid_fd, (off_t)0, SEEK_SET);
		if (pid_offset == -1) {
			printf("Failed to lseek old_udf_fd %llu: %s\n",(unsigned long long)0, strerror(errno)); 
		} else {
			if (write(pid_fd, progressBuf, pb_len) != pb_len){
		    		printf("Failed to write pid_fd: %s\n", strerror(errno));
			}
		}
	}

	pages_nr = new_mdf_cache_len;
	max_new_mdf_cache_num = 0;

	// Get the max block number of new mdf
	for (i=0; i<pages_nr; i++){
		// If the new mdf cache was dirty
		if((new_mdf_cache[i] != (uint32_t)-1) && 
			(new_mdf_cache[i] > max_new_mdf_cache_num)) {
			max_new_mdf_cache_num = new_mdf_cache[i];
		}
	}
	//prinf("max_new_mdf_cache_num is %llu\n", (unsigned long long)max_new_mdf_cache_num);

	max_new_mdf_cache_num++;
	
	for (i=0; i<pages_nr; i++){
		realRdLen = UDF_PAGE_SIZE;
		// If the old mdf cache was not dirty
		if(old_mdf_cache[i] == (uint32_t)-1) {
			rdlen = 0;
			mdf_cache_num = new_mdf_cache[i];
			// If the output mdf file was not the new mdf file
			// and the new mdf file was dirty
			// copy the content from new to output
			if((output_mdf_fd != new_mdf_fd) && 
				(mdf_cache_num != (uint32_t)-1)) {
				offset = lseek(new_udf_fd, 
							(off_t)(mdf_cache_num*UDF_PAGE_SIZE), 
							SEEK_SET);
				if (offset == -1) {
					printf("Failed to lseek new udf file %llu: %s\n",
						(unsigned long long)(mdf_cache_num*UDF_PAGE_SIZE),
						strerror(errno)); 
					ret = -1;
					goto error;
				}

				rdlen = read(new_udf_fd, tmpbuf, realRdLen);
				if (rdlen != realRdLen) {
					printf("Failed to read new udf file %s, real_rd_len=%llx, expect_rd_len=%llx, i=%llx, seek=%llx\n", 
						strerror(errno), 
						rdlen, 
						realRdLen,
						i,
						(off_t)(mdf_cache_num));
					ret = -1;
					goto error;
				}

				
				//prinf("old 0 new 1: copy from new mdf is %llu\n", (unsigned long long)mdf_cache_num);
			
				offset = lseek(output_udf_fd,
							(off_t)(mdf_cache_num*UDF_PAGE_SIZE),
							SEEK_SET);    
				if (offset == -1) {
					printf("Failed to lseek output udf file %llu: %s\n", 
						(unsigned long long)(mdf_cache_num*UDF_PAGE_SIZE), 
						strerror(errno)); 
					ret = -1;
					goto error;
				}						
			
				if (write(output_udf_fd, tmpbuf, realRdLen) !=	realRdLen){
				    printf("Failed to write output udf file: %s\n", strerror(errno));
				    ret = -1;
					goto error;
				}
			}
		} else {
			rdlen = 0;
			mdf_cache_num = old_mdf_cache[i];
			
			offset = lseek(old_udf_fd, 
						(off_t)(mdf_cache_num*UDF_PAGE_SIZE), 
						SEEK_SET);
			if (offset == -1) {
				printf("Failed to lseek old_udf_fd %llu: %s\n",
					(unsigned long long)(mdf_cache_num*UDF_PAGE_SIZE),
					strerror(errno)); 
				ret = -1;
				goto error;
			}

			rdlen = read(old_udf_fd, tmpbuf, realRdLen);
			if (rdlen != realRdLen) {
				printf("Failed to read old_udf_fd %s, real_rd_len=%llx, expect_rd_len=%llx, i=%llx, seek=%llx\n", 
					strerror(errno), 
					rdlen, 
					realRdLen,
					i,
					(off_t)(mdf_cache_num));
				ret = -1;
				goto error;
			}

			if(new_mdf_cache[i] != (uint32_t)-1) {
				mdf_cache_num = new_mdf_cache[i];
				//prinf("old 1 new 1: copy from new mdf is %llu\n", 
				//	(unsigned long long)mdf_cache_num);
			} else {
				mdf_cache_num = max_new_mdf_cache_num;
				max_new_mdf_cache_num++;
				//prinf("old 1 new 0: copy from new mdf is(max) %llu\n", 
				//	(unsigned long long)mdf_cache_num);
			}
		
			offset = lseek(output_udf_fd,
						(off_t)(mdf_cache_num*UDF_PAGE_SIZE),
						SEEK_SET);    
			if (offset == -1) {
				printf("Failed to lseek output udf file %llu: %s\n", 
					(unsigned long long)(mdf_cache_num*UDF_PAGE_SIZE), 
					strerror(errno)); 
				ret = -1;
				goto error;
			}
		
			if (write(output_udf_fd, tmpbuf, realRdLen) !=	realRdLen){
			    printf("Failed to write output udf file: %s\n", strerror(errno));
			    ret = -1;
				goto error;
			}
			new_mdf_cache[i] =	mdf_cache_num;
		}	

		if(calc_progress) {
			progressBar = ((i+1)*100)/pages_nr;
			snprintf(progressBuf, 8, "%d", progressBar);
			pid_offset = lseek(pid_fd, (off_t)0, SEEK_SET);
			if (pid_offset == -1) {
				printf("Failed to lseek old_udf_fd %llu: %s\n",(unsigned long long)0, strerror(errno)); 
			} else {
				if (write(pid_fd, progressBuf, pb_len) != pb_len){
			    		printf("Failed to write pid_fd: %s\n", strerror(errno));
				}
			}
		}
	}  

	offset = lseek(output_mdf_fd, (off_t)0, SEEK_SET);    
	if (offset == -1) {
		printf("Failed to lseek output udf file %llu: %s\n", 
			(unsigned long long)(0), 
			strerror(errno)); 
		ret = -1;
		goto error;
	}

	if (write(output_mdf_fd,
			new_mdf_cache, 
			new_mdf_cache_len*sizeof(uint32_t)) !=	
			new_mdf_cache_len*sizeof(uint32_t)){
	    printf("Failed to write output mdf file: %s\n", 
			strerror(errno));
	    ret = -1;
		goto error;
	}

	fsync(output_mdf_fd);
	fsync(output_udf_fd);
	printf("MDF/UDF file merged successfully!\n");

	ret = 0;
	
	goto sys_exit;

error:	
	goto sys_exit;

sys_exit:

	if ((output_mdf_fd != new_mdf_fd) && 
		(output_mdf_fd >= 0)) {
		close (output_mdf_fd);
		if(ret != 0) {
			output_md_file = check_path_safe(output_md_file);
			if(output_md_file)
				unlink(output_md_file);
		}
	}
	
	if ((output_udf_fd != new_udf_fd) &&
		 output_udf_fd >= 0) {
		close (output_udf_fd);
		if(ret != 0) {
			output_ud_file = check_path_safe(output_ud_file);
			if(output_ud_file)
				unlink(output_ud_file);
		}
	}
	
	if (new_udf_fd >= 0)
		close (new_udf_fd);
	if (new_mdf_fd >= 0)
		close (new_mdf_fd);
	if (old_udf_fd >= 0)
		close (old_udf_fd);
	if (old_mdf_fd > 0)
		close (old_mdf_fd);
	
	if (pid_fd >= 0) {
		close (pid_fd);
		//unlink(pid_file);
	}

	if (new_mdf_cache != NULL)
		free(new_mdf_cache);

	if (old_mdf_cache != NULL)
		free(old_mdf_cache);

	if(tmpbuf)
		free(tmpbuf);

	if(new_md_file)
		free(new_md_file);
	if(new_ud_file)
		free(new_ud_file);
	if(old_md_file )
		free(old_md_file);
	if(old_ud_file )
		free(old_ud_file);
	if(output_md_file)
		free(output_md_file);
	if(output_ud_file )
		free(output_ud_file);

	return ret;
}

