/******************************************************************************
Copyright 2012 - 2020 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <time.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include "predefine.h"
#include "tcprotocol.h"
#include "client.h"
#include "db.h"
#include "netcfg.h"
#include "tctimer.h"
#include "debug.h"
#include "process.h"
#include "conf_cmd.h"
#include "fileops.h"
#include "license.h"

#define MAX_FILE_PATH	256

Client g_running_clients[MAX_CLIENTS] = {0};
fd_set g_mset;

int
do_conf_req(
	Client *client, 
	tc_pkt_t *p_tc_pkt, 
	int pkt_len, 
	char *p_free_client)
{
	//int len;
	int ret;
	int tag=0;
	__uint8_t conf_cmd;

	conf_cmd = SUB_CMD(p_tc_pkt);

	//len = ntohs(p_tc_pkt->tc_cmd.len);
	if(p_free_client)
	  *p_free_client = FALSE;
	
	GET_TAG(p_tc_pkt->hdr.tag, tag);

	if(conf_cmd != AUTH_CONF &&
	   (tag < 0 || tag >= MAX_CLIENTS)){
		ERR("do_conf_req: cmd(%d) with invalid tag %d", conf_cmd, tag);
		p_tc_pkt->hdr.ver_flags |= TC_FL_ERROR;
		p_tc_pkt->hdr.error = TC_ERR_BAD_PARAMETER;
		goto conf_req_cleanup;
	} 

	switch (conf_cmd) {
	case HB_CONF:
		update_hb(client, p_tc_pkt);
		return 0;
	case AUTH_CONF:
	  	//MSG("Qauth [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
	  	MSG("Qauth [%d ]\n", tag);
		ret = auth_conf(client, p_tc_pkt);
		if(ret != 0) {
			if(p_free_client)
				*p_free_client = TRUE;
		}
		break;
	case GET_USER_ID_CONF:
	  	//MSG("Qauth [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
	  	MSG("QgetUId [%d ]\n", tag);
		get_user_id_conf(client, p_tc_pkt);
		break;
	case SET_OS_CONF:		
		MSG("Qsetos [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
		set_os_conf(client, p_tc_pkt);
		break;		
	case RESET_CONF:		
		MSG("Qpoweroff [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
		reset_conf(client, p_tc_pkt);
		if(p_free_client)
			*p_free_client = TRUE;
		break;
	case UPLOAD_CHECK_CONF:		
		MSG("Qupload check [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
		upload_check_conf(client, p_tc_pkt,
			_check_upload_conditions);
		break;
	case UPLOAD_CONF:		
		MSG("Qupload [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
		upload_conf(client, p_tc_pkt);
		break;
	case SYNC_CONF:
		MSG("Qsync [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
		sync_conf(client, p_tc_pkt);
		break;
	case SYNC_DATA_CONF:
		MSG("QsyncData [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
		sync_data_conf(client, p_tc_pkt);
		break;
	case UPLOAD_CHECK_DATA_CONF:		
		MSG("Qupload check data [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
		upload_check_conf(client, p_tc_pkt,
			_check_upload_data_conditions);
		break;
	case UPLOAD_DATA_CONF:		
		MSG("Qupload data [%d %s]\n", tag, p_tc_pkt->tc_cmd.data);
		upload_data_conf(client, p_tc_pkt);
		break;
	default:
		p_tc_pkt->hdr.ver_flags |= TC_FL_ERROR;
		p_tc_pkt->hdr.error = TC_ERR_BAD_PARAMETER;
	}

conf_req_cleanup:
	
	p_tc_pkt->tc_cmd.bufcnt = htons(RW_BUF_CNT);
	p_tc_pkt->tc_cmd.scnt = 2;
	p_tc_pkt->tc_cmd.vercmd = 0x10;

	return p_tc_pkt->tc_cmd.len + (sizeof(*p_tc_pkt) - sizeof(p_tc_pkt->tc_cmd.data));
}

void do_req(Client* client, tc_hdr_t *p_tc_hdr, int pkt_len, char *p_free_client)
{
	int len;
	tc_pkt_t *p_tc_pkt;
	char *outputStr = "";

	switch (p_tc_hdr->cmd) {	
		
	case TC_CMD_CONFIG:
		if (pkt_len < sizeof(tc_cmd_t))
			return;

		len = do_conf_req(client, 
				(tc_pkt_t *)p_tc_hdr, pkt_len, 
				p_free_client);
		if (len == 0)
			return;
		break;
	default:
		p_tc_hdr->error = TC_ERR_BAD_COMMAND;
		
		p_tc_pkt = (tc_pkt_t *)p_tc_hdr;

		p_tc_pkt->tc_cmd.len = 
			strlen(outputStr) + 1;
		strncpy(p_tc_pkt->tc_cmd.data, outputStr, p_tc_pkt->tc_cmd.len);	
		
		len = p_tc_pkt->tc_cmd.len + 
			(sizeof(*p_tc_pkt) - sizeof(p_tc_pkt->tc_cmd.data));
		break;
	}	
	
	p_tc_hdr->ver_flags |= TC_FL_RESPONSE;
	p_tc_hdr->len = len;
	
	if (snd_pkt(client, (uint8_t *) p_tc_hdr, len) == -1) {
		ERR("write to network: %s\n", strerror(errno));
	}
	
}

void cleanup()
{
	int i;
	SERVER_CONFIG *s_config;

  	s_config = get_server_config();

	for(i = 0; i < MAX_CLIENTS; i++) {
		free_client(&(g_running_clients[i]));
	}

	//MSG("update server status in DB to 0\n");
	update_server_status(TC_SERVER_ID,0);

	//MSG("reset all status\n");
	reset_all_status();

	//MSG("end database connection\n");
	end_db();

	if(s_config) {
		if(s_config->socket > 0) {
			close(s_config->socket);
			s_config->socket = -1;
		}
	}

	single_inst_lockfile_cleanup(s_config->single_inst_file);
  
#if VERIFY_LICENSE
  if (is_license_invalid(s_config->license_file, s_config->eth_mac, 
  		sizeof(s_config->eth_mac), s_config->runTime, FALSE)) {
	ERR("Invalid license\n");
  }
#endif

  	free(s_config->cur_dir);
	
	MSG("Signal received, stop serving clients\n");

	closelog();

	exit(EXIT_SUCCESS);
}

int setup_server()
{
	int ret;

	setup_udf_dir();

  	if(update_server_status(TC_SERVER_ID, 1) != 0) {
		ERR("Failed to update server status\n");
		return -1;
  	}

	client_array_init(g_running_clients, sizeof(g_running_clients)/sizeof(Client));

  	setup_sighandler(cleanup);
  	//MSG("Setup signal handlers\n");

#if ENABLE_TC_TIMER  
  	setup_watchdog();
  	//MSG("Setup watchdog\n");
#endif

	daemonize();
    
  	return 0;
}

int
get_and_verify_request(
	Client *p_Client, 
	uint8_t *buf, 
	int bufsz,
	int *p_pkt_len)
{
	int receLen, sh;
	tc_hdr_t *p_tc_hdr;

	if(p_pkt_len)
		*p_pkt_len = 0;
 
	receLen = rcv_pkt(p_Client, buf, bufsz);
	if (receLen <= 0) {
		if(receLen == 0)
			ERR("Connection has been closed by peer\n");
		else
			ERR("Read network error (%s)\n", strerror(errno));
		// network error
		return -2;
	}

	if (receLen < sizeof(tc_hdr_t)){
		ERR("Packet size:%d is less than header size: %lu\n", receLen, sizeof(tc_hdr_t));
		return -1;
	}

	if (receLen > bufsz){
		ERR("Packet size:%d is greater than buf size: %d\n", receLen, bufsz);
		return -1;
	}
		
	p_tc_hdr = (tc_hdr_t *) buf;
		
	if (p_tc_hdr->ver_flags & TC_FL_RESPONSE){
		ERR("Packet is a request instead of a response\n");
		return -1;
	}

	if(p_pkt_len)
		*p_pkt_len = receLen;
	
	return 0;
}

Client* 
select_correct_client(
	Client* p_old_client, 
	tc_hdr_t *p_tc_hdr,
	char checkIp)
{
	int ret = 0;
	int tagIndex = 0; 
	Client* p_new_client = NULL;

	GET_TAG(p_tc_hdr->tag, tagIndex);

	//DEBUG("client ip is(tag->%d): %d.%d.%d.%d\n", 
	//	tagIndex,
	//	p_old_client->addr[0],
	//	p_old_client->addr[1],
	//	p_old_client->addr[2], 
	//	p_old_client->addr[3]);

	if(tagIndex >= 0 && tagIndex < MAX_CLIENTS) {
		p_new_client = g_running_clients + tagIndex;
		if(p_new_client != p_old_client) {
			if(NOT_IN_USE(p_new_client)) {
				ERR("select_client error: Packet(tag %d) is not in use\n", tagIndex);
				ret = -1;
				p_new_client = NULL;
				p_tc_hdr->error = TC_ERR_INVALID_TAG;
				goto select_client_cleanup;
			} else if(checkIp &&
					(memcmp(p_new_client->addr, 
						p_old_client->addr, 4) != 0)) {
				ERR("select_client_error: old client ip(tag %d-%d.%d.%d.%d) diffs from the new client(tag:%d-%d.%d.%d.%d)\n",
				p_old_client->usedIdx,
				p_old_client->addr[0],
				p_old_client->addr[1],
				p_old_client->addr[2], 
				p_old_client->addr[3],
				tagIndex,
				p_new_client->addr[0],
				p_new_client->addr[1],
				p_new_client->addr[2], 
				p_new_client->addr[3]);
				
			
				ret = -1;
				p_new_client = NULL;
				p_tc_hdr->error = TC_ERR_INVALID_MAC;
				goto select_client_cleanup;
			}
		}
	} else {
		if(p_tc_hdr->cmd == TC_CMD_CONFIG) {
			__uint8_t conf_cmd;
			tc_pkt_t *p_tc_pkt;

			p_tc_pkt = (tc_pkt_t *)p_tc_hdr;
			conf_cmd = SUB_CMD(p_tc_pkt);

			// Make sure we don't occupy the wrong client
			if(conf_cmd == AUTH_CONF &&
				(p_old_client->username && p_old_client->username[0] != 0)) {
				size_t cmd_data_len;
				char *pUsername = NULL;
				char *pPassword = NULL;
				char *cmd_data = NULL;

				cmd_data_len = strlen((char *)p_tc_pkt->tc_cmd.data) + 1;
				cmd_data = malloc(cmd_data_len);

				if(cmd_data == NULL) {
					ERR("select_client: failed to alloc mem for cmd\n");
					p_tc_hdr->error = TC_ERR_BAD_PARAMETER;
					goto select_client_cleanup;
				}
				
				strncpy(cmd_data, (char *)p_tc_pkt->tc_cmd.data, cmd_data_len);
				cmd_data[cmd_data_len - 1] = '\0';
				ret = _parse_user_info((char *)cmd_data, 
							&pUsername, 
							&pPassword);

				free(cmd_data);
				cmd_data = NULL;
				if((ret != 0) ||
					!pUsername){
					ERR("select_client: failed to parse user info %s\n", 
						p_tc_pkt->tc_cmd.data);
					p_tc_hdr->error = TC_ERR_PWD_UNAME;
					goto select_client_cleanup;
				}
				//DEBUG("old %s, new %s\n", p_old_client->username, username);

				// if the user name not equals
				if(memcmp(p_old_client->username, 
						pUsername, strlen(pUsername)) != 0) {
					int i;
					
					i = get_client(g_running_clients, 
							p_old_client->addr,
							pUsername);
					if(i != -1) {
						//DEBUG("select client: i %d\n", i);
						p_new_client = g_running_clients + i;
						memcpy(p_new_client->addr, p_old_client->addr, 4);
					}
				}

				if(pUsername)
					free(pUsername);
				if(pPassword)
					free(pPassword);
			}
		}
		if(!p_new_client)
			p_new_client = p_old_client;
	}

	if(p_new_client != p_old_client) {
		
		p_new_client->sock = p_old_client->sock;
#ifdef SDL_BUILD_ENABLE
		p_new_client->p_ssl = p_old_client->p_ssl;
#endif
		
		//DEBUG("select_client: old client(%s) != tagIndex(%d, %s)\n",
		//	p_old_client->username,
		//	tagIndex,
		//	p_new_client->username);
	}

select_client_cleanup:
	if(ret != 0) {
		tc_pkt_t *p_tc_pkt;
		
		p_tc_hdr->ver_flags |= TC_FL_ERROR;
		p_tc_hdr->ver_flags |= TC_FL_RESPONSE;

		p_tc_pkt = (tc_pkt_t *)p_tc_hdr;
		
		p_tc_pkt->tc_cmd.len = 0;	
		p_tc_hdr->len = (sizeof(*p_tc_pkt) - sizeof(p_tc_pkt->tc_cmd.data));

		if (snd_pkt(p_old_client, 
				(uint8_t *) p_tc_hdr, 
				512) == -1) {
			ERR("select_client failed to send package: %s\n", 
				strerror(errno));
		}
	}

	return p_new_client;
}

int handle_request(Client* p_client, tc_hdr_t *p_tc_hdr, int pkt_len, char *p_free_client)
{
	tc_pkt_t *tPkt;

	if(!p_client) {
		ERR("Connection is NULL\n");
		return -1;
	}
	tPkt = (tc_pkt_t *)p_tc_hdr;
	
	if(tPkt->tc_cmd.len > pkt_len) {
		ERR("data len is wrong\n");
		return -1;
	}
	
	tPkt->tc_cmd.data[tPkt->tc_cmd.len] = 0;

	//manually increase the heart beat of the client
	if(p_client->username && p_client->username[0] != 0)
		update_hb(p_client, NULL);

	do_req(p_client, p_tc_hdr, pkt_len, p_free_client);
	
	return 0;
}

#ifdef SDL_BUILD_ENABLE

int setup_client_ssl(SERVER_CONFIG *s_config, 
	Client *client)
{
	int ret;
	
	if(!client)
		return -1;
	
	client->p_ssl = SSL_new(s_config->p_tc_ssl_ctx);
	if(client->p_ssl == NULL) {
		ERR("setup_client_ssl: failed to new SSL\n");
		ret = -1;
		goto setup_client_ssl_cleanup;
	}

	SSL_set_fd(client->p_ssl, client->sock);
	 
	if(SSL_accept(client->p_ssl) == -1) {
		ERR("setup_client_ssl: failed to accept fd\n");
		ret = -1;
		goto setup_client_ssl_cleanup;
	}
	
	//DBG("Server with %s encryption\n", SSL_get_cipher(client->p_ssl));

	ret = 0;

setup_client_ssl_cleanup:
	if(ret != 0) {
		if(client->p_ssl) {
			SSL_shutdown(client->p_ssl);
			SSL_free(client->p_ssl);
			client->p_ssl = NULL;
		}
	}
	
	return ret;
}
#endif

void serve_clients(SERVER_CONFIG *s_config)
{
	struct sockaddr_in addrin;
	socklen_t addrinlen = sizeof(addrin);
	int max = 0, i;
	int sock,net;
	int bufsz = 1024*5;
	fd_set rset;
	struct timeval tv;
	Client *client;

	tc_hdr_t *p_tc_hdr;
	uint8_t     *buf = NULL;
	int       ret;
	int pkt_len;
	
	sock = s_config->socket;
	max  = sock > max ? sock : max;
	
  	MSG("Start to serve clients\n");
	if ((buf = malloc(bufsz)) == NULL) {
		ERR("malloc failed (%s)\n",strerror(errno));
		return;
	}
	
	for(;;) {
		FD_ZERO(&g_mset);
		FD_SET(sock, &g_mset);

		for(i=0; i<MAX_CLIENTS; i++)
		{
			client = g_running_clients + i;
			if(client->sock > 0)
				FD_SET(client->sock, &g_mset);
		}
		
		memcpy(&rset, &g_mset, sizeof(fd_set));
		tv.tv_sec  = 0;
		tv.tv_usec = 500;
		
		if(select(max+1, &rset, NULL, NULL, &tv)>0) {
			//TODO: we should try to handle client requests in another thread
			for(i=0; i<MAX_CLIENTS; i++)
			{
				client = g_running_clients + i;
				if(client->sock > 0 && FD_ISSET(client->sock, &rset))
				{
					memset(buf, 0x0, bufsz);

					ret = get_and_verify_request(client, 
							buf, 
							bufsz, 
							&pkt_len);
					if(ret == 0) {
						Client *p_new_client;
						p_tc_hdr = (tc_hdr_t *) buf;
						p_new_client = select_correct_client(client, 
											p_tc_hdr,
											FALSE);

						if(p_new_client) {
							char need_free_client = FALSE;
							
							if(p_new_client != client) {
								//if((!client->username || (client->username[0] == 0))) {
									if(p_new_client->sock == client->sock) {
										client->sock = -1;
									#ifdef SDL_BUILD_ENABLE
										client->p_ssl = NULL;
									#endif
									}
									free_client(client);
									client = NULL;
								/*} else {
									client->sock = -1;
								#ifdef SDL_BUILD_ENABLE
									client->p_ssl = NULL;
								#endif
								}*/
							}

							if(pkt_len > bufsz) {
								DEBUG("handle_request failed with pktsize/bufsz %d/%d\n",
										pkt_len, 
										bufsz);
							} else {								
								ret = handle_request(p_new_client, p_tc_hdr, pkt_len, &need_free_client);
								if(ret != 0)
								{
									DEBUG("handle_request failed with(%d) %d: %s\n", 
										ret,
										i, 
										p_new_client->username);
								} else if(need_free_client){
									if(p_new_client->sock)
										FD_CLR(p_new_client->sock, &g_mset);
									free_client(p_new_client);
									p_new_client = NULL;
								}
							}
						}
					} else {
						// network error
						if(ret == -2) {
							//DEBUG("Wait for the hb timer to kill the client %d: %s\n", i, 
								//client->username);
							if(IN_USE(client)) {
						#ifdef SDL_BUILD_ENABLE
								if(client->p_ssl) {
									SSL_shutdown(client->p_ssl);
									SSL_free(client->p_ssl);
									client->p_ssl = NULL;
								}
						#endif
								if(client->sock) {
									FD_CLR(client->sock, &g_mset);
									close(client->sock);
								}
								client->sock = -1;
	                        }
						}
					}		
				}
			}
			
			if(FD_ISSET(sock, &rset)) {
				if ((net = accept(sock, (struct sockaddr *) &addrin, &addrinlen)) < 0){
					ERR("accept: %s\n", strerror(errno));
					continue;
				}
				//DEBUG("Accept: %s\n", inet_ntoa(addrin.sin_addr));

				i = get_client(g_running_clients, 
						(unsigned char *)&addrin.sin_addr,
						NULL);

				// We use the tag to indentify the client, so no need to track addr here
				// i = get_client(g_running_clients, NULL);
				if(i != -1 )
				{					
					client = g_running_clients + i;
					client->sock = net;
					memcpy(client->addr, &addrin.sin_addr, 4);
					DEBUG("client ip is: %d.%d.%d.%d\n", client->addr[0], client->addr[1],
						client->addr[2], client->addr[3]);
					max = max > net ? max : net;
					ret = 0;

#ifdef SDL_BUILD_ENABLE
					ret = setup_client_ssl(s_config, client);
					if(ret != 0) {						
						ERR("failed to set up client ssl\n");
						free_client(client);
						client = NULL;
					}
#endif
				}else{
					ERR("no more available running slot\n");
					update_server_status(TC_SERVER_ID, SERVER_STATUS_OVERLOAD);
					ret = -1;					
  				}

				if(ret != 0) {
					tc_pkt_t *p_tc_pkt;

					p_tc_pkt = (tc_pkt_t *)buf;
					memset(p_tc_pkt, 0x0, sizeof(tc_pkt_t));					
											
					p_tc_hdr = &p_tc_pkt->hdr;
					p_tc_hdr->ver_flags |= TC_FL_ERROR;
					p_tc_hdr->ver_flags |= TC_FL_RESPONSE;
					p_tc_hdr->error = TC_ERR_UNAVAILABLE;
					p_tc_pkt->tc_cmd.len = 0;			

					if (write(net, (uint8_t *) p_tc_pkt, 512) == -1) {
						ERR("serve_client failed to send package: %s\n", 
							strerror(errno));
					}

					close(net);
				}
				net = -1;
			}				
		}
	}

	if(buf)
		free(buf);
}
 
void usage() {
	printf("tc sync server version %s\n", "1.0.0");
	printf("Usage:tc-sync-server [-i ip] [-p port] [-n|nic ethName] [-l licFile]"
#ifdef SDL_BUILD_ENABLE
		   " [-c certDir]"
#endif
		   "\n"
	       "\t-i|--ip\t\tip address to listen on\n"
	       "\t-p|--port\tport to listen at\n"
	       "\t-n|--nic\tnic name\n"
	       "\t-l|--lic_file\tlicense file path\n"
#ifdef SDL_BUILD_ENABLE
	       "\t-c|--ca_dir\tdirectory where file \"" DFT_CA_FILE "\" & \"" \
	       DFT_CERT_FILE "\" & \"" DFT_KEY_FILE "\" resides\n"
#endif
	       "Notes:\n\tif ip is not set, it will listen on ip: " SRV_IP " by default.\n\t"
	       "if port is not ,it will listen at port: %d by default.\n\t"
	       "if nic is not set, it will use: " NIC_NAME " by default.\n ", SRV_PORT);
}

#ifdef SDL_BUILD_ENABLE

int init_ssl_work(SERVER_CONFIG *p_s_config)
{	
	int ret;
	SSL_CTX *p_tc_ssl_ctx = NULL;
	const SSL_METHOD *p_tc_ssl_method;

	if(!p_s_config ||
#ifdef ENABLE_TC_MUTUAL_TLS	
		!p_s_config->ca_file ||
#endif
		!p_s_config->cert_file || 
		!p_s_config->key_file) {
		ERR("Error: ssl params are NULL\n");
		return -1;
	}

	p_s_config->p_tc_ssl_ctx = NULL;

	SSL_library_init();
	OpenSSL_add_all_algorithms();
	SSL_load_error_strings();
	
	p_tc_ssl_method = TLSv1_2_server_method();	
	p_tc_ssl_ctx = SSL_CTX_new(p_tc_ssl_method);
	if (p_tc_ssl_ctx == NULL) {
		ERR("Failed to create ssl ctx\n");
		ret = -1;
		goto init_ssl_cleanup;
	}
	
#ifdef ENABLE_TC_MUTUAL_TLS	
	SSL_CTX_set_verify(p_tc_ssl_ctx, SSL_VERIFY_PEER, 0);
	
	if (SSL_CTX_load_verify_locations(p_tc_ssl_ctx, 
			p_s_config->ca_file, 
			0) != 1) {
		ERR("Failed to load CA file: %s\n", p_s_config->ca_file);
		ret = -2;
		goto init_ssl_cleanup;
	}
#endif
	
	if (SSL_CTX_use_certificate_file(p_tc_ssl_ctx, 
			p_s_config->cert_file, 
			SSL_FILETYPE_PEM) <= 0) {
		ERR("Failed to load cert file: %s\n", p_s_config->cert_file);
		ret = -3;
		goto init_ssl_cleanup;
	}
	
	if (SSL_CTX_use_PrivateKey_file(p_tc_ssl_ctx, 
			p_s_config->key_file, 
			SSL_FILETYPE_PEM) <= 0) {
		ERR("Failed to load key file: %s\n", p_s_config->key_file);
		ret = -4;
		goto init_ssl_cleanup;;
	}
	
	if (!SSL_CTX_check_private_key(p_tc_ssl_ctx)) {
		ERR("Failed to check private key\n");
		ret = -1;
		goto init_ssl_cleanup;
	}

#ifdef ENABLE_TC_MUTUAL_TLS		 
	SSL_CTX_set_mode(p_tc_ssl_ctx, SSL_MODE_AUTO_RETRY);
#endif
	p_s_config->p_tc_ssl_ctx  = p_tc_ssl_ctx;
	
	ret = 0;
	
init_ssl_cleanup:

	if(ret != 0) {	
		if(p_tc_ssl_ctx)
			SSL_CTX_free(p_tc_ssl_ctx);
	}
	
	return ret;
}

#endif

int tc_main(int argc, char *argv[]) {
	int ret = TC_MAIN_ERR_COMMON;
	SERVER_CONFIG *s_config = NULL;
	char db_password[MAX_STRING_LEN];
	char db_username[64] = {0};
	size_t usernamelen;
	
	logging_init();
	//MSG("Log initialized\n");

	ret = parseConfigFile(DATA_CONFIG_FILE, db_password);
	if (ret != 0) {
  		ERR("Failed to parse configuration file: %d\n", ret);
		ret = TC_MAIN_ERR_DB;
  		goto init_failed;
  	}

	usernamelen = 64;
	snprintf(db_username, usernamelen, "%s", DFT_DB_UNAME);
	ret = start_db(db_username, db_password);
  	if (ret != 0) {
  		ERR("Failed to start database connection: %d\n", ret);
		ret = TC_MAIN_ERR_DB;
  		goto init_failed;
  	}

	s_config = get_server_config();
	s_config->socket = -1;

	if(parse_params(argc, argv, s_config) != 0) {
		usage();
		ret = TC_MAIN_ERR_BAD_PARAMETER;
		goto init_failed;
	}

	if(is_instance_running(s_config->single_inst_file)) {
		ERR("ERROR: Another TC sync server is running with same parameters\n");
		ret = TC_MAIN_ERR_INST_RUNNING;
		goto init_failed;
	}

#ifdef SDL_BUILD_ENABLE
	ret = init_ssl_work(s_config);
	if (ret != 0) {
		ERR("Failed to init ssl\n");
		ret = TC_MAIN_ERR_SSL;
	  	goto init_failed;
	}
	MSG("TLS enabled\n");
#endif

	ret = init_socket(s_config);
	if (ret != 0) {
		ERR("Failed to init socket1\n");
		ret = TC_MAIN_ERR_SOCK;
	  	goto init_failed;
	}
	//MSG("Network initialized\n");

	if(get_eth_mac(s_config->socket, 
		s_config->ethName, s_config->eth_mac) != 0){
		ERR("Failed to init socket2\n");
		ret = TC_MAIN_ERR_MAC;
	  	goto init_failed;
	}

#if VERIFY_LICENSE
	s_config->cur_dir = (char *)get_app_path();
	if(!s_config->cur_dir) {
		ret = TC_MAIN_ERR_COMMON;
		goto init_failed;
	}

	if (is_license_invalid(s_config->license_file, s_config->eth_mac, 
							sizeof(s_config->eth_mac), s_config->runTime, TRUE)) {
		ERR("Failed to verify license\n");
		ret = TC_MAIN_ERR_LICENSE;
		goto init_failed;
	}
#endif

	if (setup_server() != 0) {
    	ERR("Failed to setup server\n");
		ret = TC_MAIN_ERR_COMMON;
    	goto init_failed;
  	}

	serve_clients(s_config);

	cleanup();

	MSG("TC sync server halt\n");
	return 0 ;

init_failed:
	if(s_config) {
		single_inst_lockfile_cleanup(s_config->single_inst_file);
		
		if(s_config->socket > 0) {
			close(s_config->socket);
			s_config->socket = -1;
		}

		if(s_config->license_file)
			free(s_config->license_file);

#ifdef SDL_BUILD_ENABLE
#ifdef ENABLE_TC_MUTUAL_TLS
		if(s_config->ca_file)
			free(s_config->ca_file);
#endif

		if(s_config->cert_file)
			free(s_config->cert_file);

		if(s_config->key_file)
			free(s_config->key_file);

		if(s_config->p_tc_ssl_ctx)
			SSL_CTX_free(s_config->p_tc_ssl_ctx);
#endif

		if(s_config->cur_dir)
			free(s_config->cur_dir);
	}

	end_db();
	
	closelog();

	return ret;
}
