/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <QGraphicsLinearLayout>
#include <QTimer>
#include <MLabel>
#include <MProgressIndicator>
#include <MDebug>

#include "misc.h"
#include "initdiskpage.h"


#define INITDISK_PROG           "/tcdscript/inithd.sh"

const ERROR_CODE_MSG  ErrMsg[] =
{
    {1,  QT_TRID_NOOP("syntax or usage error")},
    {2,  QT_TRID_NOOP("illegal partition size")},
    {3,  QT_TRID_NOOP("interrupted by user")},
    {4,  QT_TRID_NOOP("hard disk not exist")},
    {5,  QT_TRID_NOOP("umount system partition failed")},
    {6,  QT_TRID_NOOP("umount cache partition failed")},
    {7,  QT_TRID_NOOP("clean MBR failed")},
    {8,  QT_TRID_NOOP("tag MBR failed")},
    {9,  QT_TRID_NOOP("hook MBR failed")},
    {10, QT_TRID_NOOP("fdisk hard disk failed")},
    {11, QT_TRID_NOOP("mount system partition failed")},
    {12, QT_TRID_NOOP("mount cache partition failed")},
    {13, QT_TRID_NOOP("format system partition failed")},
    {14, QT_TRID_NOOP("format cache partition failed")},
    {15, QT_TRID_NOOP("format user partition failed")},
    {16, QT_TRID_NOOP("system partition not exist")},
    {17, QT_TRID_NOOP("cache partition not exist")},
    {18, QT_TRID_NOOP("user partition not exist")},
    {19, QT_TRID_NOOP("remove boot files failed")},
    {20, QT_TRID_NOOP("copy boot files failed")},
    {21, QT_TRID_NOOP("create initrd image failed")},
    {22, QT_TRID_NOOP("compress initrd image failed")},
    {23, QT_TRID_NOOP("initrd image not exist")},
    {24, QT_TRID_NOOP("install boot loader failed")},
    {25, QT_TRID_NOOP("create VD failed")}
};


InitDiskPage::InitDiskPage(int nImgPartMB)
  : nImgPart(nImgPartMB)
{
    mDebug("InitDiskPage::InitDiskPage()");

    setObjectName("initdisk_page");
}

InitDiskPage::~InitDiskPage()
{
    mDebug("InitDiskPage::~InitDiskPage()");
}

void InitDiskPage::createContent()
{
    mDebug("InitDiskPage::createContent()");

//    MApplicationPage::createContent();
//    setStyleName("CommonApplicationPage");

    setComponentsDisplayMode(NavigationBar, MApplicationPageModel::Hide);
    setComponentsDisplayMode(HomeButton, MApplicationPageModel::Hide);
    setComponentsDisplayMode(EscapeButton, MApplicationPageModel::Hide);

    QGraphicsLinearLayout *layout = new QGraphicsLinearLayout(Qt::Vertical, centralWidget());

    layout->addStretch();

    spinner = new MProgressIndicator(this, MProgressIndicator::spinnerType);
    spinner->setObjectName("initdisk_spinner");
    spinner->setUnknownDuration(true);
    spinner->setVisible(true);
    layout->addItem(spinner);
    layout->setAlignment(spinner, Qt::AlignCenter);

    lbPrompt = new MLabel(qtTrId("Initializing the harddisk, please wait..."));
    lbPrompt->setObjectName("initdisk_lb_prompt");
    layout->addItem(lbPrompt);

    layout->addStretch();

    connect(&psTcSync, SIGNAL(error(QProcess::ProcessError)), this, SLOT(onPsError(QProcess::ProcessError)));
    connect(&psTcSync, SIGNAL(finished(int,QProcess::ExitStatus)), this, SLOT(onPsFinished(int,QProcess::ExitStatus)));


    QStringList lstArgs;
    if ( nImgPart != 0 )
    {
        lstArgs.append(QString::number(nImgPart));
    }

    psTcSync.start(INITDISK_PROG, lstArgs);
}

void InitDiskPage::ShowMessage(const QString &msg)
{
    displayBanner(this, msg);
    QTimer::singleShot(1500, this, SLOT(dismiss()));
}

void InitDiskPage::onPsError(QProcess::ProcessError errCode)
{
    QString msg;

    mDebug("InitDiskPage::onPsError: ") << errCode;

    msg = qtTrId("Execute program failed: %1.").arg(errCode);
    ShowMessage(msg);
}

void InitDiskPage::onPsFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    QString msg;
    const char *errmsg = QT_TRID_NOOP("unknown");

    mDebug("InitDiskPage::onPsFinished: ") << exitCode << exitStatus;

    if ( exitCode != 0 )
    {
        for ( uint i = 0; i < (sizeof(ErrMsg)/sizeof(ERROR_CODE_MSG)); i++ )
        {
            if ( ErrMsg[i].nCode == exitCode )
            {
                errmsg = ErrMsg[i].msg;
                break;
            }
        }
        msg = qtTrId("Init disk failed: %1.").arg(qtTrId(errmsg));
    }
    else
    {
        msg = qtTrId("Init disk successfully.");
    }
    ShowMessage(msg);
}
