/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <QtDBus>
#include <QGraphicsLinearLayout>
#include <QGraphicsGridLayout>
#include <QSwipeGesture>
#include <MImageWidget>
#include <MTextEdit>
#include <MButton>
#include <MLabel>
#include <MProgressIndicator>
#include <MMessageBox>
#include <MDebug>

#include "misc.h"
#include "inputdialog.h"
#include "powerdialog.h"
#include "indicatordialog.h"
#include "initdiskpage.h"
#include "updatefwpage.h"
#include "regmachpage.h"
#include "cleandiskpage.h"
#include "oslistpage.h"
#include "processpage.h"
#include "profilepage.h"
#include "settingpage.h"
#include "loginpage.h"


#define CLIENT_VERSION_FILE     "/usr/sbin/tc/VERSION"

#define TCAUTH_NET_PROG         "/usr/sbin/tc/tc-auth.sh.net"
#define TCAUTH_LOCAL_PROG       "/usr/sbin/tc/tc-auth.sh.local"

#define ADMINAUTH_PROG          "/usr/sbin/tc/admin-auth.sh"

#define TC_QUICKBOOT_PROG       "/usr/sbin/tc/get_img.sh"

#define USERINFO_FILE           "/mnt/rsvd/tc-user.info"
#define HOSTIP_FILE             "/mnt/rsvd/tc-serv.addr"

const ERROR_CODE_MSG  ErrMsg[] =
{
    {1, QT_TRID_NOOP("syntax or usage error")},
    {2, QT_TRID_NOOP("network not ready")},
    {3, QT_TRID_NOOP("machine not registered")},
    {4, QT_TRID_NOOP("invalid user name or password")},
    {5, QT_TRID_NOOP("no OS image available")},
    {6, QT_TRID_NOOP("connect server failed")},
    {7, QT_TRID_NOOP("get user ID failed")},
    {8, QT_TRID_NOOP("user is locked")}
};


LoginPage::LoginPage()
    : nDiskStatus(1)
    , bDiskBusy(true)
    , nBootOsId(-1)
{
    mDebug("LoginPage::LoginPage");

    setObjectName("login_page");

    QDBusConnection::sessionBus().registerObject("/LoginPage", this, QDBusConnection::ExportAllSlots);
}

LoginPage::~LoginPage()
{
    mDebug("LoginPage::~LoginPage");

    QDBusConnection::sessionBus().unregisterObject("/LoginPage");
}

void LoginPage::createContent()
{
    mDebug("LoginPage::createContent");

    setComponentsDisplayMode(NavigationBar, MApplicationPageModel::Hide);
    setComponentsDisplayMode(HomeButton, MApplicationPageModel::Hide);
    setComponentsDisplayMode(EscapeButton, MApplicationPageModel::Hide);

    QGraphicsLinearLayout *lyOuter = new QGraphicsLinearLayout(Qt::Vertical, centralWidget());
    QGraphicsGridLayout *lyGrid = new QGraphicsGridLayout();
    QGraphicsLinearLayout *lyStatus = new QGraphicsLinearLayout(Qt::Vertical);
    QGraphicsLinearLayout *lyLogin = new QGraphicsLinearLayout(Qt::Horizontal);

    readLine(CLIENT_VERSION_FILE, clientVer); ;
    readLine(HOSTIP_FILE, servIP);

    lbClientVer = new MLabel(clientVer);
    lbClientVer->setObjectName("login_lb_logo1");

    lbMachineName = new MLabel();
    lbMachineName->setObjectName("login_lb_logo2");

    imgQRCode = new MImageWidget();

#if 0
    MButton *btnQRCode = new MButton();
    btnQRCode->setObjectName("login_btn_status");
    btnQRCode->setViewType(MButton::iconType);
    btnQRCode->setIconID("icon-qrcode");
    lyStatus->addItem(btnQRCode);
#endif

    btnHdStatus = new MButton();
    btnHdStatus->setObjectName("login_btn_status");
    btnHdStatus->setViewType(MButton::iconType);
    btnHdStatus->setIconID("icon-disk_error");
    lyStatus->addItem(btnHdStatus);

    btnNetStatus = new MButton();
    btnNetStatus->setObjectName("login_btn_status");
    btnNetStatus->setViewType(MButton::iconType);
    btnNetStatus->setIconID("icon-net_noline");
    lyStatus->addItem(btnNetStatus);

    lyGrid->addItem(lyStatus, 0, 4);


    MLabel *lbLogo1 = new MLabel(qtTrId("Welcome to"));
    lbLogo1->setObjectName("login_lb_logo1");
    lyGrid->addItem(lbLogo1, 1, 0);

    MLabel *lbLogo2 = new MLabel(qtTrId("Transparent Computing"));
    lbLogo2->setObjectName("login_lb_logo2");
    lyGrid->addItem(lbLogo2, 2, 0);

    spinner = new MProgressIndicator(this, MProgressIndicator::spinnerType);
    spinner->setObjectName("login_spinner");
//    spinner->setStyleName("CommonLargeSpinner");
    spinner->setUnknownDuration(true);
    spinner->setVisible(false);
    lyGrid->addItem(spinner, 1, 1, 2, 1);

    btnRemID = new MButton();
    btnRemID->setObjectName("login_btn_remid");
    btnRemID->setViewType(MButton::switchType);
    btnRemID->setCheckable(true);
    lyGrid->addItem(btnRemID, 1, 2);

    MLabel *lbRemID = new MLabel(qtTrId("Remember user name"));
    lbRemID->setObjectName("login_lb_remid");
    lyGrid->addItem(lbRemID, 1, 3);

    btnOffMode = new MButton();
    btnOffMode->setObjectName("login_btn_locallogin");
    btnOffMode->setViewType(MButton::switchType);
    btnOffMode->setCheckable(true);
    btnOffMode->setChecked(false);
    lyGrid->addItem(btnOffMode, 2, 2);

    MLabel *lbLocalLogin = new MLabel(qtTrId("Offline mode"));
    lbLocalLogin->setObjectName("login_lb_locallogin");
    lyGrid->addItem(lbLocalLogin, 2, 3);

    btnProfile = new MButton();
    btnProfile->setObjectName("login_btn_profile");
    btnProfile->setViewType(MButton::iconType);
    btnProfile->setIconID("icon-profile");
    lyGrid->addItem(btnProfile, 1, 4);

    btnSetting = new MButton();
    btnSetting->setObjectName("login_btn_setting");
    btnSetting->setViewType(MButton::iconType);
    btnSetting->setIconID("icon-settings");
    lyGrid->addItem(btnSetting, 2, 4);

    btnPower= new MButton();
    btnPower->setObjectName("login_btn_power");
    btnPower->setViewType(MButton::iconType);
    btnPower->setIconID("icon-power");
    lyGrid->addItem(btnPower, 3, 4);

    teUser = new MTextEdit;
    teUser->setObjectName("login_te_user");
    teUser->setMaxLength(30+3);  // autoboot: xxx-ab
    // setText() then setPrompt() otherwise text and prompt will both show
    teUser->setPrompt(qtTrId("user name"));
    teUser->setPreferredWidth(1);
    teUser->setTextInteractionFlags(Qt::TextEditable);
    lyLogin->addItem(teUser);

    tePass = new MTextEdit;
    tePass->setObjectName("login_te_pass");
    tePass->setMaxLength(32);
    tePass->setPrompt(qtTrId("password"));
    tePass->setEchoMode(MTextEditModel::Password);
    tePass->setPreferredWidth(1);
    tePass->setTextInteractionFlags(Qt::TextEditable);
    lyLogin->addItem(tePass);

    lyGrid->addItem(lyLogin, 3, 0, 1, 2);

    if (teUser->text().isEmpty())
    {
        scene()->setFocusItem(teUser, Qt::ActiveWindowFocusReason);
    }
    else
    {
        scene()->setFocusItem(tePass, Qt::ActiveWindowFocusReason);
    }

    btnLogin = new MButton(qtTrId("Sign in"));
    btnLogin->setObjectName("login_btn_login");

    lyGrid->addItem(btnLogin, 3, 2, 1, 2);

    lyGrid->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);

    lyOuter->addItem(lbClientVer);
    lyOuter->addItem(lbMachineName);
    lyOuter->addStretch();
    lyOuter->addItem(imgQRCode);
    lyOuter->addItem(lyGrid);


    connect(btnLogin, SIGNAL(clicked()), this, SLOT(onBtnLoginClicked()));
#if SDL_BUILD_ENABLE != 1
    connect(btnHdStatus, SIGNAL(clicked()), this, SLOT(onBtnDiskClicked()));
#endif
    connect(btnProfile, SIGNAL(clicked()), this, SLOT(onBtnProfileClicked()));
    connect(btnSetting, SIGNAL(clicked()), this, SLOT(onBtnSettingClicked()));
    connect(btnPower, SIGNAL(clicked()), this, SLOT(onBtnPowerClicked()));
    connect(teUser, SIGNAL(tabPressed()), this, SLOT(onTabPressed()));
    connect(tePass, SIGNAL(tabPressed()), this, SLOT(onTabPressed()));
    connect(teUser, SIGNAL(returnPressed()), btnLogin, SIGNAL(clicked()));
    connect(tePass, SIGNAL(returnPressed()), btnLogin, SIGNAL(clicked()));
    connect(&psTcSync, SIGNAL(error(QProcess::ProcessError)), this, SLOT(onPsTcSyncError(QProcess::ProcessError)));
    connect(&psTcSync, SIGNAL(finished(int,QProcess::ExitStatus)), this, SLOT(onPsTcSyncFinished(int,QProcess::ExitStatus)));

    connect(this, SIGNAL(AutoBootEvent(const QString &, const QString &, int, const QString &, int)), this, SLOT(onAutoBootEvent(const QString &, const QString &, int, const QString &, int)), Qt::QueuedConnection);
    connect(this, SIGNAL(QuickBootEvent(int, const QString &, const QString &, const QString &)), this, SLOT(onQuickBootEvent(int, const QString &, const QString &, const QString &)), Qt::QueuedConnection);
    connect(this, SIGNAL(HwStatusEvent(int, int)), this, SLOT(onHwStatusEvent(int, int)), Qt::QueuedConnection);
    connect(this, SIGNAL(ShowQRImgEvent(const QString &)), this, SLOT(onShowQRImgEvent(const QString &)), Qt::QueuedConnection);
    connect(this, SIGNAL(InitDiskEvent(int)), this, SLOT(onInitDiskEvent(int)), Qt::QueuedConnection);
    connect(this, SIGNAL(UpdateFwEvent()), this, SLOT(onUpdateFwEvent()), Qt::QueuedConnection);
    connect(this, SIGNAL(RegMachineEvent()), this, SLOT(onRegMachineEvent()), Qt::QueuedConnection);
    connect(this, SIGNAL(RegMachineAutoEvent(const QString &)), this, SLOT(onRegMachineAutoEvent(const QString &)), Qt::QueuedConnection);
    connect(this, SIGNAL(MachineNameEvent(const QString &)), this, SLOT(onMachineNameEvent(const QString &)), Qt::QueuedConnection);

    connect(this, SIGNAL(appeared()), this, SLOT(onAppeared()));
}

bool LoginPage::event(QEvent *e)
{
    if (e->type() == QEvent::WindowUnblocked)
    {
        teUser->setFocus();
    }

    return MApplicationPage::event(e);
}

void LoginPage::enableInput()
{
    mDebug("enableInput");

    spinner->setVisible(false);
    teUser->setEnabled(true);
    tePass->setEnabled(true);
    btnLogin->setEnabled(true);
}

void LoginPage::disableInput()
{
    mDebug("disableInput");

    spinner->setVisible(true);
    teUser->setEnabled(false);
    tePass->setEnabled(false);
    btnLogin->setEnabled(false);
}

void LoginPage::OnRsvdStatus(int nStatus)
{
    if ( nStatus != 0 )
    {
        displayBanner(this, qtTrId("System partition not ready, settings will be lost after reboot!"));
    }
}

void LoginPage::OnDiskStatus(int nStatus)
{
    if ( nStatus == nDiskStatus )
    {
        return;
    }

    switch ( nStatus )
    {
        case 0:
            btnHdStatus->setIconID("icon-disk_ok");
            break;

        case 1:
            btnHdStatus->setIconID("icon-disk_fix");
            break;

        case 2:
            displayBanner(this, qtTrId("TC's partition is full, please cleanup disk."));
            btnHdStatus->setIconID("icon-disk_full");
            break;

        case 3:
            displayBanner(this, qtTrId("TC's partition is corrupted!"));
            btnHdStatus->setIconID("icon-disk_error");
            break;

        case 13:
            displayBanner(this, qtTrId("TC's partition is fixed, please reboot system."));
            break;

    }

    nDiskStatus = nStatus;

    bDiskBusy = (nStatus == 1) ? true : false;
}

void LoginPage::OnNetStatus(int nStatus)
{
    const char *str;

    switch ( nStatus )
    {
        case 0:
          str = "icon-net_ready";
          break;

        case 1:
          str = "icon-net_noline";
          break;

        case 2:
          str = "icon-net_noip";
          break;

        case 3:
          str = "icon-net_noserver";
          break;

        case 4:
          str = "icon-net_noservice";
          break;

    }

    btnNetStatus->setIconID(str);
}

int LoginPage::AutoBootOS(const QString &uname, const QString &passwd, int delay, const QString &osname, int osid)
{
    mDebug("AutoBootOS") << uname << passwd << osname << osid;

    if ( isOnDisplay() )
    {
        emit AutoBootEvent(uname, passwd, delay, osname, osid);
    }

    return 0;
}

int LoginPage::QuickBootOS(int osid, const QString &orgos, const QString &curos, const QString &uname)
{
    mDebug("QuickBootOS") << osid << orgos << curos << uname;

    if ( isOnDisplay() )
    {
        emit QuickBootEvent(osid, orgos, curos, uname);
    }

    return 0;
}

int LoginPage::HwStatus(int nHw, int nStatus)
{
    mDebug("HwStatus") << nHw << nStatus;

    emit HwStatusEvent(nHw, nStatus);

    return 0;
}

int LoginPage::ShowQRImg(const QString &imgfile)
{
    mDebug("ShowQRImg") << imgfile;

    emit ShowQRImgEvent(imgfile);

    return 0;
}

int LoginPage::InitDisk(int nImgPartMB)
{
    mDebug("InitDisk") << nImgPartMB;

    if ( isOnDisplay() )
    {
        emit InitDiskEvent(nImgPartMB);
    }

    return 0;
}

int LoginPage::UpdateFw()
{
    mDebug("UpdateFw");

    if ( isOnDisplay() )
    {
        emit UpdateFwEvent();
    }

    return 0;
}

int LoginPage::RegMachine()
{
    mDebug("RegMachine");

    if ( isOnDisplay() )
    {
        emit RegMachineEvent();
    }

    return 0;
}

int LoginPage::RegMachineAuto(const QString &cname)
{
    mDebug("RegMachineAuto");

    if ( isOnDisplay() )
    {
        emit RegMachineAutoEvent(cname);
    }

    return 0;
}

int LoginPage::MachineName(const QString &cname)
{
    mDebug("MachineName");

    emit MachineNameEvent(cname);

    return 0;
}

void LoginPage::onTabPressed()
{
    QGraphicsItem *item;

    mDebug("onTabPressed");

    item = centralWidget()->focusItem();
    if ( item == teUser )
    {
        mDebug("onTabPressed: teUser -> tePass");
        scene()->setFocusItem(tePass, Qt::TabFocusReason);
    }
    else if ( item == tePass )
    {
        mDebug("onTabPressed: tePass -> teUser");
        scene()->setFocusItem(teUser, Qt::TabFocusReason);
    }
}

void LoginPage::onBtnLoginClicked()
{
    QString strError;
    QStringList lstArgs;
    QString cmd;

    userName = teUser->text();

    if ( nDiskStatus )
    {
        strError = qtTrId("TC's partition is full or not ready.");
    }
    else if (userName.isEmpty())
    {
        strError = qtTrId("Please input user name.");
    }
    else if (tePass->text().isEmpty())
    {
        strError = qtTrId("Please input password.");
    }

    if (!strError.isEmpty())
    {
        displayBanner(this, strError);
        return;
    }

    disableInput();

    if (btnRemID->isChecked())
    {
        writeLine(USERINFO_FILE, userName);
    }
    else
    {
        writeLine(USERINFO_FILE, QString(""));
    }

    cmd = btnOffMode->isChecked() ? TCAUTH_LOCAL_PROG : TCAUTH_NET_PROG;

    lstArgs.append(userName);
    lstArgs.append(tePass->text());
    lstArgs.append(servIP);
    psTcSync.start(cmd, lstArgs);
}

void LoginPage::onBtnDiskClicked()
{
    int exitCode;
    QString cmd(ADMINAUTH_PROG);
    QStringList lstArgs;

    InputDialog dlg(qtTrId("Confirm"), qtTrId("Please input admin's password"));

    if ( dlg.exec() != M::OkButton )
    {
        mDebug("onBtnSettingClicked: cancel");
        return;
    }

    lstArgs.append(dlg.edit->text());
    exitCode = QProcess::execute(cmd, lstArgs);
    mDebug("call") << cmd << exitCode;

    if ( exitCode != 0 )
    {
        displayBanner(this, qtTrId("Wrong admin's password."));
        return;
    }

    pgCleanDisk = new CleanDiskPage();
    pgCleanDisk->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void LoginPage::onBtnProfileClicked()
{
    pgProfile = new ProfilePage(servIP);
    pgProfile->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void LoginPage::onBtnSettingClicked()
{
    int exitCode;
    QString cmd(ADMINAUTH_PROG);
    QStringList lstArgs;

    InputDialog dlg(qtTrId("Confirm"), qtTrId("Please input admin's password"));

    if ( dlg.exec() != M::OkButton )
    {
        mDebug("onBtnSettingClicked: cancel");
        return;
    }

    lstArgs.append(dlg.edit->text());
    exitCode = QProcess::execute(cmd, lstArgs);
    mDebug("call") << cmd << exitCode;

    if ( exitCode != 0 )
    {
        displayBanner(this, qtTrId("Wrong admin's password."));
        return;
    }

    pgSetting = new SettingPage(servIP);
    connect(pgSetting, SIGNAL(disappeared()), this, SLOT(onSettingPageFinished()));
    pgSetting->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void LoginPage::onBtnPowerClicked()
{
    if ( bDiskBusy )
    {
        displayBanner(this, qtTrId("checking the disk, please wait..."));
        return;
    }

    // see issues_compile.txt
    PowerDialog dlg;//();

    dlg.exec();
}

void LoginPage::onSettingPageFinished()
{
    mDebug("onSettingPageFinished");

    if (pgSetting->bIPChanged)
    {
        servIP = pgSetting->servIP;
        mDebug("new server IP: ") << servIP;
    }
}

void LoginPage::onPsTcSyncError(QProcess::ProcessError errCode)
{
    Q_UNUSED(errCode);

    mDebug("LoginPage::onPsError: ") << errCode;

    displayBanner(this, qtTrId("Can not run login program."));

    enableInput();
}

void LoginPage::onPsTcSyncFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    QString msg;
    const char *errmsg = QT_TRID_NOOP("unknown");

    mDebug("LoginPage::onPsFinished: ") << exitCode << exitStatus;

    enableInput();

    if ((exitStatus != QProcess::NormalExit) || (exitCode != 0))
    {
        for ( uint i = 0; i < (sizeof(ErrMsg)/sizeof(ERROR_CODE_MSG)); i++ )
        {
            if ( ErrMsg[i].nCode == exitCode )
            {
                errmsg = ErrMsg[i].msg;
                break;
            }
        }
        msg = qtTrId("login failed: %1.").arg(qtTrId(errmsg));
        displayBanner(this, msg);
        return;
    }

    OSListPage *pgOS = new OSListPage(userName, servIP, nBootOsId);
    nBootOsId = -1;

    pgOS->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void LoginPage::onAutoBootEvent(const QString &uname, const QString &passwd, int delay, const QString &osname, int osid)
{
    int  result;
    QString msg, sname(osname);

    sname.truncate(10);
    msg = qtTrId("Boot to OS(%1) automatically?").arg(sname);

    IndicatorDialog dlg(qtTrId("Confirm"), msg, delay*1000);
    result = dlg.exec();
    if ( result != MDialog::Accepted )
    {
        mDebug("onAutoBootEvent: cancel");
        nBootOsId = -1;
        return;
    }

    mDebug("onAutoBootEvent: accept");

    teUser->setText(uname);
    tePass->setText(passwd);
    nBootOsId = osid;

    emit onBtnLoginClicked();
}

void LoginPage::onQuickBootEvent(int osid, const QString &orgos, const QString &curos, const QString &uname)
{
    QStringList lstCmd;

    lstCmd.append(curos);
    lstCmd.append(uname);
    lstCmd.append(QString::number(osid));
    lstCmd.append(orgos);
    lstCmd.append(servIP);
    mDebug("LoginPage::onQuickBootEvent") << lstCmd;

    ProcessPage *pgProcess = new ProcessPage(qtTrId("Launch the OS"), QString(TC_QUICKBOOT_PROG), lstCmd);
    pgProcess->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void LoginPage::onHwStatusEvent(int nHw, int nStatus)
{
    switch ( nHw )
    {
        case 1:
            OnDiskStatus(nStatus);
            break;

        case 2:
            OnNetStatus(nStatus);
            break;

        case 3:
            OnRsvdStatus(nStatus);
            break;

        default:
            break;
    }
}

void LoginPage::onShowQRImgEvent(const QString &imgfile)
{
    QImage img = QImage(imgfile);
    imgQRCode->setImage(img);
}

void LoginPage::onInitDiskEvent(int nImgPartMB)
{
    pgInitDisk = new InitDiskPage(nImgPartMB);
    pgInitDisk->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void LoginPage::onUpdateFwEvent()
{
    pgUpdateFw = new UpdateFwPage();
    pgUpdateFw->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void LoginPage::onRegMachineEvent()
{
    pgRegMach = new RegMachPage(servIP, QString(""));
    pgRegMach->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void LoginPage::onRegMachineAutoEvent(const QString &cname)
{
    pgRegMach = new RegMachPage(servIP, cname);
    pgRegMach->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void LoginPage::onMachineNameEvent(const QString &cname)
{
    lbMachineName->setText(cname);
}

void LoginPage::onAppeared()
{
    QString userinfo;

    mDebug("LoginPage::onAppeared");

    readLine(USERINFO_FILE, userinfo);
    if (userinfo.isEmpty())
    {
        btnRemID->setChecked(false);
        teUser->clear();
    }
    else
    {
        btnRemID->setChecked(true);
        teUser->setText(userinfo);
    }
    tePass->clear();
}

