/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <QFile>
#include <QTimer>
#include <QGraphicsLinearLayout>
#include <MGridLayoutPolicy>
#include <QSwipeGesture>
#include <MApplicationWindow>
#include <MSceneManager>
#include <MLabel>
#include <MImageWidget>
#include <QState>
#include <QStateMachine>
#include <QParallelAnimationGroup>
#include <QPropertyAnimation>
#include <QSignalTransition>
#include <MAction>
#include <MLayout>
#include <MDebug>

#include "misc.h"
#include "processpage.h"
#include "oslistpage.h"


#define OSLIST_FILE             "/usr/sbin/tc/os.lst"
#define ACTLIST_FILE            "/usr/sbin/tc/act.lst"

#define TC_BASE_DIR             "/usr/sbin/tc"
#define TC_BASE_PIC_DIR         "/usr/sbin/tc/pic/"
#define TC_DEFAULT_OS_ICON      "/usr/sbin/tc/pic/default_os_icon.png"


OsListObjectMenu::OsListObjectMenu()
  : MObjectMenu(NULL)
{
    setObjectName("objectMenu");
}

OsListObjectMenu::~OsListObjectMenu()
{
    qDeleteAll(actListEntries);
}

bool OsListObjectMenu::Init(const QString &name)
{
    MAction *action;

    LoadFromFile(name);

    for (int i = 0; i < actListEntries.count(); i++)
    {
        action = new MAction(actListEntries[i]->desc, this);
        action->setObjectName("osAction");
        addAction(action);
        connect(action, SIGNAL(triggered()), this, SLOT(actListItem()));

        actListEntries[i]->action = action;
    }

    return true;
}

void OsListObjectMenu::checkPermission(int cur_per)
{
    ActListEntry *pEntry;
    int  hi_cper, lo_cper, hi_rper, lo_rper;

    hi_cper = cur_per & 0xF0;
    lo_cper = cur_per & 0x0F;
    mDebug("OSListPage::checkPermission(cur_per)") << cur_per << hi_cper << lo_cper;

    for (int i = 0; i < actListEntries.count(); i++)
    {
        pEntry = actListEntries[i];

        hi_rper = pEntry->permission & 0xF0;
        lo_rper = pEntry->permission & 0x0F;
        mDebug("OSListPage::checkPermission(permission)") << pEntry->permission << hi_rper << lo_rper;

        if ( (hi_cper >= hi_rper) && (lo_cper >= lo_rper) )
        {
            pEntry->action->setEnabled(true);
        }
        else
        {
            pEntry->action->setEnabled(false);        
        }
    }
}

void OsListObjectMenu::clickDefaultAct()
{
    mDebug("OsListObjectMenu::clickDefaultAct()");

    if (actListEntries.count() < 1)
    {
        return;
    }

    QString desc = actListEntries[0]->desc;
    QString cmd = actListEntries[0]->cmd;
    mDebug("OsListObjectMenu::actListItem()") << cmd;
    emit onActClicked(desc, cmd);
}

void OsListObjectMenu::actListItem()
{
    MAction *action = qobject_cast<MAction *>(sender());
    QString text = action->text();
    QString desc, cmd;

    mDebug("OsListObjectMenu::actListItem()") << text;

    for (int i = 0; i < actListEntries.count(); i++)
    {
        if (!text.compare(actListEntries[i]->desc))
        {
            desc = actListEntries[i]->desc;
            cmd = actListEntries[i]->cmd;
            break;
        }
    }

    if (cmd.isEmpty())
    {
        mWarning("OsListObjectMenu::actListItem()") << "not found";
        return;
    }

    mDebug("OsListObjectMenu::actListItem()") << cmd;
    emit onActClicked(desc, cmd);
}

bool OsListObjectMenu::LoadFromFile(const QString &name)
{
    QFile fileAct(name);
    ActListEntry  *pEntry;

    mDebug("OsListObjectMenu::LoadFromFile()");

    if (fileAct.open(QIODevice::ReadOnly))
    {
        QTextStream txtAct(&fileAct);
        while (!txtAct.atEnd())
        {
            QString s = txtAct.readLine();
            QStringList sl = s.split(";");
            mDebug("OsListObjectMenu::LoadFromFile()") << sl;
            if (sl.count() >=2)
            {
                pEntry = new ActListEntry;
                pEntry->desc = sl.at(0);
                pEntry->cmd = sl.at(1);
                pEntry->permission = sl.at(4).toInt(NULL, 16);

                actListEntries.append(pEntry);
            }
        }
    }

    return true;
}


OSListPage::OSListPage(const QString &strName, const QString &serv, int osid)
	: userName(strName)
	, server(serv)
	, curIndex(-1)
	, nBootOsId(osid)
	, aniState(0)
{
    mDebug("OSListPage::OSListPage()");

    setObjectName("oslist_page");
}

OSListPage::~OSListPage()
{
    mDebug("OSListPage::~OSListPage()");

    qDeleteAll(osListEntries);
}

void OSListPage::createContent()
{
    mDebug("OSListPage::createContent()");

//    MApplicationPage::createContent();
//    setStyleName("CommonApplicationPage");

    setComponentsDisplayMode(NavigationBar, MApplicationPageModel::Hide);
    setComponentsDisplayMode(HomeButton, MApplicationPageModel::Hide);
    setComponentsDisplayMode(EscapeButton, MApplicationPageModel::Hide);

    applicationWindow()->setNavigationBarOpacity(0.9);

#if 0
    MAction *actChgPwd = new MAction("icon-chgpwd", "Change a password", centralWidget());
    actChgPwd->setObjectName("oslist_act_chgpwd");
    actChgPwd->setLocation(MAction::ToolBarLocation);
    addAction(actChgPwd);
    connect(actChgPwd, SIGNAL(triggered()), this, SLOT(onActChgPwd()));
#endif

    LoadFromFile(QString(OSLIST_FILE));

    objMenu = new OsListObjectMenu();
    objMenu->Init(QString(ACTLIST_FILE));

    MLayout *layout = new MLayout;
    MGridLayoutPolicy *policy = new MGridLayoutPolicy(layout);
//    policy->setSpacing(10);
    policy->setHorizontalSpacing(10);
    policy->setVerticalSpacing(25);
    centralWidget()->setLayout(layout);

    stRootOs = new QState;
    stShowOs = new QState(stRootOs);
    stHideOs = new QState(stRootOs);
    stRootOs->setInitialState(stHideOs);

    aniGroupOs = new QParallelAnimationGroup;

    for ( int i = 0; i < osListEntries.count(); i++ )
    {
        MImageWidget *imgWidget = new MImageWidget(this);
        imgWidget->setImage(osListEntries[i]->icon);
        lstOsWidget.append(imgWidget);

        MLabel *hint = new MLabel(osListEntries[i]->desc);
        hint->setAlignment(Qt::AlignCenter);
        hint->setTextElide(true);

        QGraphicsLinearLayout *lyItem = new QGraphicsLinearLayout(Qt::Vertical);
        lyItem->addItem(imgWidget);
        lyItem->addItem(hint);
        lyItem->setPreferredWidth(1);

        policy->addItem(lyItem, i / 3, i % 3);

        stShowOs->assignProperty(imgWidget, "opacity", 1.0);
        stHideOs->assignProperty(imgWidget, "opacity", 0.0);

        stShowOs->assignProperty(hint, "opacity", 1.0);
        stHideOs->assignProperty(hint, "opacity", 0.0);
    
        QPropertyAnimation *ani = new QPropertyAnimation(imgWidget, "opacity");
        ani->setDuration(1500);
        aniGroupOs->addAnimation(ani);
    }

    stMachOs = new QStateMachine;
    stMachOs->addState(stRootOs);
    stMachOs->setInitialState(stRootOs);
    stMachOs->start();

    connect(objMenu, SIGNAL(onActClicked(const QString&, const QString&)), this, SLOT(actClicked(const QString&, const QString&)));
    connect(aniGroupOs, SIGNAL(finished()), this, SLOT(aniFinished()));
    connect(aniGroupOs, SIGNAL(stateChanged(QAbstractAnimation::State,QAbstractAnimation::State)), this, SLOT(aniStateChanged(QAbstractAnimation::State,QAbstractAnimation::State)));

    if ( nBootOsId < 0 )
    {
        QTimer *timer = new QTimer;
        timer->start(2000);
        timer->setSingleShot(true);
        QAbstractTransition *trans = stRootOs->addTransition(timer, SIGNAL(timeout()), stShowOs);
        trans->addAnimation(aniGroupOs);
    }
    else
    {
        // skip the "show" animation
        QTimer::singleShot(1500, this, SLOT(delayAutoBoot()));
    }
}

void OSListPage::aniFinished()
{
    mDebug("anifinished");

    if ( aniState == 1 )
    {
        aniState = 0;
        objMenu->clickDefaultAct();
    }
}

void OSListPage::aniStateChanged(QAbstractAnimation::State newState, QAbstractAnimation::State oldState)
{
    mDebug("aniStateChanged") << newState << oldState;
}

void OSListPage::enterDisplayEvent()
{
    mDebug("enterDisplayEvent");
}

void OSListPage::swipeGestureEvent(QGestureEvent *event, QSwipeGesture *gesture)
{
    mDebug("OSListPage::swipeGestureEvent()") << "State:" << gesture->state() << "HDir:" << gesture->horizontalDirection() << "VDir:" << gesture->verticalDirection();

    event->accept(gesture);

    if (gesture->state() == Qt::GestureFinished)
    {
        if (gesture->horizontalDirection() == QSwipeGesture::Right)
        {
            disappear();
        }
    }
}

void OSListPage::tapGestureEvent(QGestureEvent *event, QTapGesture* gesture)
{
    mDebug("OSListPage::tapGestureEvent()") << "State:" << gesture->state() << "Position:" << gesture->position();

    event->accept(Qt::TapGesture);

    if ( (nBootOsId >= 0) || (aniState > 0) )
    {
        mDebug("OSListPage::tapGestureEvent()") << "ignore";
        return;
    }

    if (gesture->state() == Qt::GestureStarted)
    {
        curIndex = -1;
        return;
    }

    if ((gesture->state() != Qt::GestureFinished) || (curIndex != -1))
    {
        return;
    }

    curIndex = GetTappedIndex(gesture->position());
    if (curIndex == -1 )
    {
        return;
    }

    BootOsAni();
}

void OSListPage::tapAndHoldGestureEvent(QGestureEvent *event, QTapAndHoldGesture *gesture)
{
    mDebug("OSListPage::tapAndHoldGestureEvent()") << "State:" << gesture->state() << "Position:" << gesture->position();

    event->accept(Qt::TapAndHoldGesture);

    if ( nBootOsId >= 0 || (aniState > 0) )
    {
        mDebug("OSListPage::tapAndHoldGestureEvent()") << "ignore";
        return;
    }

    if (gesture->state() == Qt::GestureStarted)
    {
        curIndex = -1;
        return;
    }

    if ((gesture->state() != Qt::GestureFinished) || (curIndex != -1))
    {
        return;
    }

    curIndex = GetTappedIndex(gesture->position());
    if (curIndex == -1 )
    {
        return;
    }

    objMenu->checkPermission(osListEntries[curIndex]->cur_per);

    objMenu->setCursorPosition(gesture->position());
    objMenu->setTitle(osListEntries[curIndex]->desc);

    sceneManager()->appearSceneWindow(objMenu);
}

bool OSListPage::LoadFromFile(const QString &name)
{
    QFile fileOs(name);
    OsListEntry  *pEntry;

    mDebug("OSListPage::LoadFromFile()");

    osDefaultIcon = QImage(QString(TC_DEFAULT_OS_ICON));

    if (fileOs.open(QIODevice::ReadOnly))
    {
        QTextStream txtOs(&fileOs);
        while (!txtOs.atEnd())
        {
            QString s = txtOs.readLine();
            QStringList sl = s.split(":");
            mDebug("OsListModel::LoadFromFile()") << sl;
            if (sl.count() >= 4)
            {
                pEntry = new OsListEntry;
                pEntry->id = sl.at(0);
                pEntry->desc = sl.at(1);
                pEntry->cur_img = sl.at(2);
                pEntry->org_img = sl.at(5);
                pEntry->icon = QImage(QString(TC_BASE_PIC_DIR) + sl.at(3));
                if (pEntry->icon.isNull())
                {
                    pEntry->icon = osDefaultIcon;
                }
                pEntry->latest_ver = sl.at(4);
                pEntry->cur_per = sl.at(6).toInt(NULL, 16);
                mDebug("cur_per=") << pEntry->cur_per;

                osListEntries.append(pEntry);
            }
        }
    }

    return true;
}

int OSListPage::GetTappedIndex(const QPointF &pos)
{
    QRectF rt;
    int index = 0;
    foreach (MImageWidget* imgWidget, lstOsWidget)
    {
        rt = imgWidget->sceneBoundingRect();
        if (rt.contains(pos))
        {
            mDebug("OSListPage::GetTappedIndex()") << "index=" << index;
            return index;
        }
        index++;
    }
    return -1;
}

void OSListPage::delayAutoBoot()
{
    mDebug("OSListPage::delayAutoBoot()");

    for ( int i = 0; i < osListEntries.count(); i++ )
    {
        if ( osListEntries[i]->id.toInt() == nBootOsId )
        {
            mDebug("OSListPage::delayAutoBoot()") << "BootOSId=" << nBootOsId << "Index=" << i;
            curIndex = i;
            BootOsAni();
            break;
        }
    }

    nBootOsId = -1;
}

void OSListPage::actClicked(const QString &desc, const QString &cmd)
{
    QStringList lstCmd;

    mDebug("OSListPage::actClicked()") << cmd;

    if (curIndex == -1)
    {
        mWarning("OSListPage::actClicked()") << "index not valid";
        return;
    }

    OsListEntry *entry = osListEntries[curIndex];

    lstCmd.append(entry->cur_img);
    lstCmd.append(userName);
    lstCmd.append(entry->id);
    lstCmd.append(entry->org_img);
    lstCmd.append(server);
    mDebug("MListPage::actClicked()") << lstCmd;

    ProcessPage *pgProcess = new ProcessPage(desc, QString(TC_BASE_DIR) + '/' + cmd, lstCmd);
    connect(pgProcess, SIGNAL(disappeared()), this, SLOT(dismiss()));
    pgProcess->appear(scene(), MSceneWindow::DestroyWhenDismissed);
}

void OSListPage::BootOsAni()
{
    mDebug("OSListPage::BootOsAni()");

    foreach (MImageWidget* imgWidget, lstOsWidget)
    {
        stHideOs->assignProperty(imgWidget, "opacity", 0.0);
    }
    stHideOs->assignProperty(lstOsWidget[curIndex], "opacity", 1.0);

    QAbstractTransition *transOs = stRootOs->addTransition(this, SIGNAL(aniTriggerHide()), stHideOs);
    transOs->addAnimation(aniGroupOs);

    aniState = 1;
    emit aniTriggerHide();
}

