/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <QtDBus>
#include <QSwipeGesture>
#include <QGraphicsLinearLayout>
#include <QFile>
#include <QTimer>
#include <QRegExp>
#include <MTextEdit>
#include <MLabel>
#include <MButton>
#include <MProgressIndicator>
#include <MDialog>
#include <MDebug>

#include "indicatordialog.h"
#include "misc.h"
#include "processpage.h"


#define SKIP_UPDATE_PIPE        "/tmp/skip-update.fifo"


// rsync-3.0.7/errcode.h
const ERROR_CODE_MSG  ErrMsg[] =
{
    {1, QT_TRID_NOOP("syntax or usage error")},
    {2, QT_TRID_NOOP("permission denied")},
    {3, QT_TRID_NOOP("communicate with server timeout")},

    {11, QT_TRID_NOOP("OS image not ready")},
    {12, QT_TRID_NOOP("create user delta files failed")},

    {21, QT_TRID_NOOP("download OS image failed")},
    {22, QT_TRID_NOOP("download upgrade data file failed")},
    {23, QT_TRID_NOOP("verify upgrade data file failed")},
    {24, QT_TRID_NOOP("download upgrade map file failed")},
    {25, QT_TRID_NOOP("upgrade OS image failed")},
    {26, QT_TRID_NOOP("verify OS image failed")},

    {31, QT_TRID_NOOP("server refused to accept user delta files")},
    {32, QT_TRID_NOOP("get mdf/udf version failed")},
    {33, QT_TRID_NOOP("mdf/udf version not match")},
    {34, QT_TRID_NOOP("upload mdf/udf files failed")},
    {35, QT_TRID_NOOP("update mdf/udf version failed")},

    {41, QT_TRID_NOOP("missing user data version file")},
    {42, QT_TRID_NOOP("missing user data file")},
    {43, QT_TRID_NOOP("user data is frozen")},
    {44, QT_TRID_NOOP("no new user data need upload")},
    {45, QT_TRID_NOOP("lock user data failed")},
    {46, QT_TRID_NOOP("upload user data failed")},
    {47, QT_TRID_NOOP("auto merge user data failed")},

    {51, QT_TRID_NOOP("query user data info failed")},
    {52, QT_TRID_NOOP("call web to create user data failed")},
    {53, QT_TRID_NOOP("sync user data info(v=0) failed")},
    {54, QT_TRID_NOOP("invalid user data file length")},
    {55, QT_TRID_NOOP("init user data file failed")},
    {56, QT_TRID_NOOP("sync user data info(v=N) failed")},
    {57, QT_TRID_NOOP("create user data mdf/udf failed")},

    {61, QT_TRID_NOOP("reset server's delta version failed")},
    {62, QT_TRID_NOOP("verify checksum failed")},
    {63, QT_TRID_NOOP("OS image's ID not match")},

    {71, QT_TRID_NOOP("download user data failed")},
    {72, QT_TRID_NOOP("download user upgrade file failed")},
    {73, QT_TRID_NOOP("upgrade user data failed")},
    {75, QT_TRID_NOOP("change password failed")}
};

#define TC_CALCFLOW_OUT             "/tmp/tc-calcflow.out"


ProcessPage::ProcessPage(const QString &title, const QString &program, const QStringList &arguments)
  : desc(title)
  , cmd(program)
  , args(arguments)
  , lineStart(0)
  , lineEnd(0)
  , curPos(0)
{
    mDebug("ProcessPage::ProcessPage()");

    setObjectName("process_page");

    QDBusConnection::sessionBus().registerObject("/ProcessPage", this, QDBusConnection::ExportAllSlots);
}

ProcessPage::~ProcessPage()
{
    mDebug("ProcessPage::~ProcessPage()");

    QDBusConnection::sessionBus().unregisterObject("/ProcessPage");
}

void ProcessPage::createContent()
{
    mDebug("ProcessPage::createContent()");

//    MApplicationPage::createContent();
//    setStyleName("CommonApplicationPage");

    setComponentsDisplayMode(NavigationBar, MApplicationPageModel::Hide);
    setComponentsDisplayMode(HomeButton, MApplicationPageModel::Hide);
    setComponentsDisplayMode(EscapeButton, MApplicationPageModel::Hide);

    connect(this, SIGNAL(SkipUpdatesEvent(int)), this, SLOT(onSkipUpdatesEvent(int)), Qt::QueuedConnection);

    process.setProcessChannelMode(QProcess::MergedChannels);
    process.start(cmd, args);

    QGraphicsLinearLayout *layout = new QGraphicsLinearLayout(Qt::Vertical, centralWidget());

    layout->addStretch();

    spinner = new MProgressIndicator(this, MProgressIndicator::spinnerType);
    spinner->setObjectName("process_spinner");
    spinner->setUnknownDuration(true);
    spinner->setVisible(true);
    layout->addItem(spinner);
    layout->setAlignment(spinner, Qt::AlignCenter);

    lbPrompt = new MLabel(qtTrId("processing, please wait..."));
    lbPrompt->setObjectName("process_lb_prompt");
    layout->addItem(lbPrompt);

    lbNetFlow = new MLabel();
    lbNetFlow->setObjectName("process_lb_netflow");
    layout->addItem(lbNetFlow);

    btnDetail = new MButton();
    btnDetail->setObjectName("process_btn_detail");
    btnDetail->setViewType(MButton::iconType);
    btnDetail->setIconID("icon-finger");
    btnDetail->setVisible(false);
    layout->addItem(btnDetail);

    txOutput = new MTextEdit(MTextEditModel::MultiLine);
    txOutput->setObjectName("process_te_output");
    txOutput->setReadOnly(true);
    txOutput->setVisible(false);
    layout->addItem(txOutput);
    bOutEnabled = false;

    process.closeWriteChannel();

    connect(btnDetail, SIGNAL(clicked()), this, SLOT(onBtnDetailClicked()));
    connect(&process, SIGNAL(error(QProcess::ProcessError)), this, SLOT(onPsError(QProcess::ProcessError)));
    connect(&process, SIGNAL(finished(int,QProcess::ExitStatus)), this, SLOT(onPsFinished(int,QProcess::ExitStatus)));
    connect(&process, SIGNAL(readyRead()), this, SLOT(onReadyRead()));
}

void ProcessPage::tapAndHoldGestureEvent(QGestureEvent *event, QTapAndHoldGesture *gesture)
{
    mDebug("ProcessPage::tapAndHoldGestureEvent()");

    event->accept(Qt::TapAndHoldGesture);

    if ((gesture->state() != Qt::GestureFinished))
    {
        return;
    }

    emit onBtnDetailClicked();
}

void ProcessPage::ShowMessage(const QString &title, const QString &msg)
{
    MDialog *dlg;
    MLabel *label;

    dlg = new MDialog(title, M::NoStandardButton);
    dlg->setObjectName("systemDialog");
    label = new MLabel(msg);
    label->setObjectName("systemDialogLabel");
    label->setStyleName("CommonBodyTextInverted");
    label->setWordWrap(true);
    dlg->setCentralWidget(label);

    connect(dlg, SIGNAL(disappeared()), this, SLOT(dismiss()));
    dlg->appear(MSceneWindow::DestroyWhenDone);
}

int ProcessPage::SkipUpdates(int delay)
{
    mDebug("SkipUpdates") << delay;

    emit SkipUpdatesEvent(delay);

    return 0;
}

void ProcessPage::onBtnDetailClicked()
{
    bOutEnabled = !bOutEnabled;
    if ( bOutEnabled )
    {
        txOutput->setVisible(true);
    }
    else
    {
        txOutput->setVisible(false);
    }
}

void ProcessPage::onPsError(QProcess::ProcessError errCode)
{
    QString msg;

    mDebug("ProcessPage::onPsError: ") << errCode;

    msg = qtTrId("Execute program failed: %1.").arg(errCode);
    ShowMessage(qtTrId("Error"), msg);
}

void ProcessPage::onPsFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    QString msg;
    const char *errmsg = QT_TRID_NOOP("unknown");

    mDebug("ProcessPage::onPsFinished: ") << exitCode << exitStatus;

    if ( exitCode != 0 )
    {
        for ( uint i = 0; i < (sizeof(ErrMsg)/sizeof(ERROR_CODE_MSG)); i++ )
        {
            if ( ErrMsg[i].nCode == exitCode )
            {
                errmsg = ErrMsg[i].msg;
                break;
            }
        }
        msg = qtTrId("%1 failed: %2.").arg(desc).arg(qtTrId(errmsg));
    }
    else
    {
        msg = qtTrId("%1 successfully.").arg(desc);
    }

    QString logfile = QString("/mnt/disk/tcentry.log.") + QDateTime::currentDateTime().toString("MM.dd.HH.mm.ss");
    writeLine(logfile.toLatin1().data(), QString::fromUtf8(fakecon.constData()));

    ShowMessage(qtTrId("Result"), msg);
}

void ProcessPage::onReadyRead()
{
    char ch, old;
    int room, len;
    QByteArray msg = process.readAll();
    QRegExp rx1("<(.*)>");
    QRegExp rx2("(100|[1-9]?\\d(\\.\\d+)?)%");
    QRegExp rx3("d\\[\\s*(\\d+(\\.\\d+)?MB/s)\\]");
    int pos;

    len = msg.size();
    room = fakecon.size() - curPos;
    if ( room < len )
    {
        fakecon.resize(fakecon.size() + len - room);
    }

    for (int i = 0; i < len; i++)
    {
        ch = msg[i];
        old = fakecon[curPos];
        fakecon[curPos] = ch;

        if ((ch == '\r') || (ch == '\n'))
        {
            if (lineEnd == curPos)
            {
                lineEnd += 1;
            }
            else
            {
                fakecon[curPos] = old;
            }

            const char *buf = fakecon.constData();

            txOutput->setText(QString::fromUtf8(buf));

            QString str(QString::fromUtf8(&buf[lineStart]));
            pos = rx1.indexIn(str);
            if ( pos >= 0 )
            {
                strPrompt = rx1.capturedTexts().at(1);
                lbPrompt->setText(strPrompt);
            }
            if ( strPrompt.size() )
            {
                pos = rx2.indexIn(str);
                if ( pos >= 0 )
                {
                    QString strProgress(strPrompt + ": ");
                    pos = rx3.indexIn(str);
                    if ( pos >= 0 )
                    {
                        strProgress += rx3.capturedTexts().at(1) + ", ";
                    }
                    strProgress += rx2.capturedTexts().at(0);
                    lbPrompt->setText(strProgress);

                    QFile fileFlow(TC_CALCFLOW_OUT);
                    if (fileFlow.open(QIODevice::ReadOnly))
                    {
                        QTextStream txtFlow(&fileFlow);
                        txtFlow.setCodec("UTF-8");
                        QString strFlow = txtFlow.readLine();
                        lbNetFlow->setText(strFlow);
                    }
                }
                else
                {
                    lbNetFlow->setText("");
                }
            }

            if (ch == '\r')
            {
                curPos = lineStart;
            }
            else
            {
                curPos = lineEnd;
                lineStart = lineEnd;
            }
        }
        else
        {
            curPos++;
            if ( lineEnd < curPos )
            {
                lineEnd = curPos;
            }
        }
    }
}

void ProcessPage::onSkipUpdatesEvent(int delay)
{
    int  result;
    const char  *noskip;
    QString msg(qtTrId("will update OS image..."));

    IndicatorDialog dlg(qtTrId("Confirm"), msg, delay*1000);
    result = dlg.exec();
    noskip = (result == MDialog::Accepted) ? "1" : "0";

    writeLine(SKIP_UPDATE_PIPE, QString(noskip));
}
