/******************************************************************************
Copyright 2016 - 2019 Intel Corporation
 
For licensing information, see the file 'LICENSE' in the root folder of
  this software module. 
******************************************************************************/

#include <QGraphicsLinearLayout>
#include <QTimer>
#include <MMessageBox>
#include <MTextEdit>
#include <MButton>
#include <MProgressIndicator>
#include <MDebug>

#include "misc.h"
#include "regmachpage.h"


#define REGMACH_PROG            "/usr/sbin/tc/reg-client.sh"

const ERROR_CODE_MSG  ErrMsg[] =
{
    {1, QT_TRID_NOOP("registration service is off")},
    {3, QT_TRID_NOOP("duplicate MAC address")},
    {4, QT_TRID_NOOP("duplicate machine name")},
    {5, QT_TRID_NOOP("illegal characters in name")}
};


RegMachPage::RegMachPage(const QString &serverIP, const QString &clientName)
    : servIP(serverIP)
    , cltName(clientName)
{
    mDebug("RegMachPage::RegMachPage()");

    setObjectName("regmach_page");
}

RegMachPage::~RegMachPage()
{
    mDebug("RegMachPage::~RegMachPage");
}

void RegMachPage::createContent()
{
    mDebug("RegMachPage::createContent()");

//    MApplicationPage::createContent();
//    setStyleName("CommonApplicationPage");

    setComponentsDisplayMode(NavigationBar, MApplicationPageModel::Hide);
    setComponentsDisplayMode(HomeButton, MApplicationPageModel::Hide);
    setComponentsDisplayMode(EscapeButton, MApplicationPageModel::Hide);

    QGraphicsLinearLayout *layout = new QGraphicsLinearLayout(Qt::Vertical, centralWidget());
    QGraphicsLinearLayout *lyRegMach = new QGraphicsLinearLayout(Qt::Horizontal);

    layout->addStretch();

    spinner = new MProgressIndicator(this, MProgressIndicator::spinnerType);
    spinner->setObjectName("regmach_spinner");
    spinner->setUnknownDuration(true);
    spinner->setVisible(false);
    layout->addItem(spinner);
    layout->setAlignment(spinner, Qt::AlignCenter);

    teMachName = new MTextEdit;
    teMachName->setObjectName("regmach_te_name");
    teMachName->setMaxLength(30);
    // setText() then setPrompt() otherwise text and prompt will both show
    teMachName->setPrompt(qtTrId("machine name(letters and numbers, max length 30)"));
    teMachName->setText(cltName);
    teMachName->setPreferredWidth(2);
    teMachName->setTextInteractionFlags(Qt::TextEditable);
    lyRegMach->addItem(teMachName);

    btnRegMach = new MButton(qtTrId("Register"));
    btnRegMach->setObjectName("regmach_btn_reg");
//    btnRegMach->setPreferredWidth(1);
    lyRegMach->addItem(btnRegMach);

    layout->addItem(lyRegMach);

    layout->addStretch();

    scene()->setFocusItem(teMachName, Qt::ActiveWindowFocusReason);

    connect(teMachName, SIGNAL(returnPressed()), btnRegMach, SIGNAL(clicked()));
    connect(btnRegMach, SIGNAL(clicked()), this, SLOT(onBtnRegMachClicked()));
    connect(&psTcSync, SIGNAL(error(QProcess::ProcessError)), this, SLOT(onPsError(QProcess::ProcessError)));
    connect(&psTcSync, SIGNAL(finished(int,QProcess::ExitStatus)), this, SLOT(onPsFinished(int,QProcess::ExitStatus)));

    if ( !cltName.isEmpty() )
    {
        displayBanner(this, qtTrId("Will register machine with the auto name."));
        QTimer::singleShot(2000, this, SLOT(onBtnRegMachClicked()));
    }
}

void RegMachPage::ShowMessage(bool bSucceed, const QString &msg)
{
    QString str;

    spinner->setVisible(false);
    btnRegMach->setEnabled(true);

    if ( bSucceed )
    {
        displayBanner(this, msg);
        QTimer::singleShot(1500, this, SLOT(dismiss()));
        return;
    }

    str = msg + qtTrId(" Click Abort button for reboot.");

    MMessageBox dlg(qtTrId("Confirm"), str, M::RetryButton|M::AbortButton);
    dlg.setObjectName("messageBox");
    if ( dlg.exec() == M::AbortButton )
    {
        QProcess::execute("reboot");
    }
}

void RegMachPage::onBtnRegMachClicked()
{
    QString name;
    QChar  ch;

    name = teMachName->text();

    if (name.isEmpty())
    {
        displayBanner(this, qtTrId("Please input machine name."));
        return;
    }

    for (int i = 0; i < name.length(); i++)
    {
        ch = name.at(i);
        if ((!ch.isLetterOrNumber()) && (ch != '-') && (ch != '_'))
        {
            displayBanner(this, qtTrId("Machine name is illegal."));
            return;
        }
    }

    btnRegMach->setEnabled(false);
    spinner->setVisible(true);

    QStringList lstArgs;
    lstArgs.append(name);
    lstArgs.append(servIP);
    psTcSync.start(REGMACH_PROG, lstArgs);
}

void RegMachPage::onPsError(QProcess::ProcessError errCode)
{
    QString msg;

    mDebug("RegMachPage::onPsError: ") << errCode;

    msg = qtTrId("Execute program failed: %1.").arg(errCode);
    ShowMessage(false, msg);
}

void RegMachPage::onPsFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    QString msg;
    const char *errmsg = QT_TRID_NOOP("unknown");

    mDebug("RegMachPage::onPsFinished: ") << exitCode << exitStatus;

    if ( exitCode != 0 )
    {
        for ( uint i = 0; i < (sizeof(ErrMsg)/sizeof(ERROR_CODE_MSG)); i++ )
        {
            if ( ErrMsg[i].nCode == exitCode )
            {
                errmsg = ErrMsg[i].msg;
                break;
            }
        }
        msg = qtTrId("Register machine failed: %1.").arg(qtTrId(errmsg));
        ShowMessage(false, msg);
    }
    else
    {
        ShowMessage(true, qtTrId("Register machine successfully."));
    }
}
