#!/bin/bash
# update centos packages
#

THIS_SCRIPT=centos_update.sh
LOGFILE="centos_update.log"
STEP=4
istep=0

exec 3>&1 4>&2
touch $LOGFILE
exec 1>$LOGFILE 2>&1

step(){
  istep=$(($istep + 1))
  echo "[$istep/$STEP] $@"
  echo "[$istep/$STEP] $@" >&3
}

prn(){
  echo "$@" >&3
}

exitv(){
  exec 1>&3 2>&4
  exec 3>&- 4>&-
  exit $1
}

tcinfo(){ # green
  echo -e "\e[32m[INFO]\e[0m $*" >&3
  echo -e "[INFO] $*" 
}

tcerr(){ # red
  echo -e "\e[31m[ERROR]\e[0m $*" >&3
  echo -e "[INFO] $*" 
}

tcwarn(){ # magenta
  echo -e "\e[35m[WARNING]\e[0m $*" >&3
  echo -e "[INFO] $*" 
}


#
# check the working directory
#
if [ ! ${BASH_SOURCE[0]} -ef $THIS_SCRIPT ] ; then
  prn ""
  prn "Please run $THIS_SCRIPT in its root directory"
  prn ""
  prn "e.g.,"
  prn "  If absolute path of file $THIS_SCRIPT is \"/Path/to/script/root/directory/$THIS_SCRIPT\","
  prn "  then run the following commands."
  prn "  cd /Path/to/script/root/directory"
  prn "  bash centos_update.sh"
  prn ""
  exitv 10
fi

#
# -i, --interactive
#
#"ab:c::"
shortopts="i"
#"a-long,b-long:,c-long::"
longopts="interactive"
unset interactive_flag

newargs=`getopt --options $shortopts --longoptions $longopts -n $THIS_SCRIPT -- "$@"`
if [ $? != 0 ] ; then
  prn "illegal arguments"
  exit 1
fi

eval set -- "$newargs"
while : ; do
  case "$1" in
    -i|--interactive)
      # echo "centos-update"
      interactive_flag=1
      shift
      ;;

    --)
      shift
      break;
      ;;

    *)
      echo "error parsing arguments"
      exit 1;
      ;;
  esac
done

#
# update kernel version to
# 3.10.0.957.21.3
#
kernel_update(){
  KRNSTR="3.10.0-957.21.3.el7.x86_64"
  KRN_DL=`pwd`"/../../3rdParty/server/centos7/krn957"
  KRNPKGS=(
    libsepol-2.5-10.el7.x86_64.rpm
    libselinux-2.5-14.1.el7.x86_64.rpm
    libsemanage-2.5-14.el7.x86_64.rpm
    libselinux-utils-2.5-14.1.el7.x86_64.rpm
    policycoreutils-2.5-29.el7_6.1.x86_64.rpm
    selinux-policy-3.13.1-229.el7_6.12.noarch.rpm
    kernel-tools-libs-3.10.0-957.21.3.el7.x86_64.rpm
    linux-firmware-20180911-69.git85c5d90.el7.noarch.rpm
    kernel-3.10.0-957.21.3.el7.x86_64.rpm
    kernel-tools-3.10.0-957.21.3.el7.x86_64.rpm
    selinux-policy-targeted-3.13.1-229.el7_6.12.noarch.rpm
    libselinux-python-2.5-14.1.el7.x86_64.rpm
    kernel-headers-3.10.0-957.21.3.el7.x86_64.rpm
  )

  tcinfo "check system kernel release"
  syskrnver=`uname -r`
  tcinfo "$syskrnver"
  if [ "${syskrnver}" == "${KRNSTR}" ] ; then
    tcinfo "kernel upgrade is not necessary"
    return 1
  else
    tcinfo "current kernel release: ${syskrnver}"
    tcinfo "will be upgraded to version: ${KRNSTR}"
    tcwarn "DO REBOOT SYSTEM AFTER KERNEL UPGRADATION, OTHERWISE YOU MIGHT GET A BROKEN SYSTEM"
  fi
  
  tcinfo "check directory"
  if [ ! -d $KRN_DL ] ; then
    tcerr "$KRN_DL: no such directory"
    return 20
  fi
  
  tcinfo "check packages"
  pkgtotal=${#KRNPKGS[@]}
  for ((i=0; i<$pkgtotal; i++)) ; do
    currpkg="${KRN_DL}/${KRNPKGS[$i]}"
  
    if [ ! -f ${currpkg} ] ; then
      tcerr "${currpkg}: no such file"
      return 21
    fi
  done
  
  tcinfo "install packages"
  yum -y localupdate $KRN_DL/*.rpm
  if [ $? -eq 0 ] ; then
    tcinfo "finish upgrading kernel"
    return 0
  fi
  
  tcwarn "fail to upgrade kernel, try installing without checking dependencies"
  for pf in "${KRNPKGS[@]}" ; do
    rpm --query ${pf%.rpm}
    if [ $? -eq 0 ] ; then
      continue
    elif [ "$pf" == "kernel-3.10.0-957.21.3.el7.x86_64.rpm" ] ; then
      rpm --install -v --force --nodeps $pf
      if [ $? -ne 0 ] ; then
        tcerr "fail to install $pf"
        return 2
      fi
    else
      rpm --upgrade -v --force --nodeps $pf
      if [ $? -ne 0 ] ; then
        tcerr "fail to install $pf"
        return 2
      fi
    fi
  done
  
  tcinfo "finish upgrading kernel"
  return 0
}

kernel_update
code=$?

if [ $code -eq 0 ] ; then
  if [ "x$interactive_flag" != "x1" ] ; then
    reboot
  fi
fi

exitv $code

