#!/bin/bash

. /etc/init.d/functions

export TC_HOME="/home/tc"
export IN_TOP="$(pwd)"
export TC_UP="$IN_TOP/upgrade"
export MIN_REV=9700
export TC_REV=

function update_firmware() {
    TC_BOOT="/home/tc/tftpboot"
    cd "$IN_TOP/client"
    fw_en="$IN_TOP/client/tc_firmware"
    fw_de="$fw_en.tar"
    if [ -f "$fw_de" ]; then
        rm -rf "$fw_de"
    fi
    gpg2 --no-tty --batch --quiet --passphrase "tc-firmware" --output "$fw_de" --decrypt "$fw_en"    
    tar xmf "$fw_de"
    chmod +x rootfs.cpio.gz
    cp rootfs.cpio.gz "$TC_BOOT"
    cp rootfs_log "$TC_BOOT"
    gunzip -f rootfs.cpio.gz
    cpio -i --quiet --to-stdout usr/share/boot/bzImage < rootfs.cpio > bzImage
    chmod +x bzImage
    cp bzImage "$TC_BOOT"
    cd "$IN_TOP"
}

function extract_tarball() {
    if [ -d "$TC_UP" ]; then
        rm -rf "$TC_UP"
    fi
    mkdir "$TC_UP"
    tar xmf "$IN_TOP/centos7/tcroot.tar.gz" -C "$TC_UP"
}

function update_binaries() {
    /usr/bin/cp -a "$TC_UP/home/tc/bin" "$TC_HOME"
    chmod -R +x "$TC_HOME/bin"
    chown root:root "$TC_HOME/bin/tc-perm"
    chmod +s "$TC_HOME/bin/tc-perm"

    cp "$TC_UP/home/tc/etc/setup_services" "$TC_HOME/etc/setup_services"
}

function update_webconsole() {
    /usr/bin/cp -a --remove-destination "$TC_UP/opt/lampp/webconsole" /opt/lampp
}

function stop_services() {
    if [ "$TC_REV" -gt "9140" ]; then
        systemctl stop tcs-disk-server
    fi
    systemctl stop tcs-ccontrol-server
    systemctl stop tcs-delivery-tracker
    systemctl stop tcs-delivery-server
    systemctl stop dnsmasq
    systemctl stop rsyncd
}

function start_services() {
    curl_opt='-H "Content-Type:application/json" -X POST -s'
    rest_url='http://localhost/tc/rest/controller.php'
    curl $curl_opt --data '{"rpc":"start-system"}' $rest_url -o start.res
    systemctl start tcs-delivery-tracker
    systemctl start tcs-delivery-server
    systemctl start tcs-mudf-server
    systemctl start dnsmasq
    systemctl start rsyncd
}

function read_installation() {
    conf="$TC_HOME/etc/tc.ini"
    r=$(grep build_revision $conf | awk -F= '{print $2}' | xargs)
    if [ -z $r ]; then
        # use new issue file here
        r=$(grep minor /etc/tc/issue | awk -F= '{print $2}' | xargs)
    fi
    if [ "$r" -lt "$MIN_REV" ]; then
        echo "$r"
        echo -n "Current revision ($r) < minimal ($MIN_REV)"
        return 1
    fi
    TC_REV=$r
}

function update_config() {
    if [ ! -d /home/tcharbor ]; then
        mkdir -p /home/tcharbor
    fi
    # update rsync configurations
    sed -i "s@/tc/upload/backup@/tcharbor@g" /etc/rsyncd.conf
    sed -i "/^ignore.*/d" /etc/rsyncd.conf

    chmod 755 /etc/rsyncd.conf
    chown root /etc/rsync.passwd

    # other older TC service files
    cp -a $TC_UP/etc/systemd/system/* /etc/systemd/system/
    systemctl daemon-reload
    systemctl enable tcs-mudf-server

    # New in 1910 pre release
    firewall-cmd --permanent --add-port=4824/tcp --quiet
    firewall-cmd --permanent --add-port=4826/tcp --quiet
    # end of 1910 pre release
    firewall-cmd --set-log-denied=all --quiet
    firewall-cmd --reload --quiet
    echo 1 > /proc/sys/net/netfilter/nf_conntrack_log_invalid        

    if [ ! -d /opt/lampp/htdocs/v1 ]; then
        mkdir /opt/lampp/htdocs/v1
    fi
    ln -sf /home/tc/userdata /opt/lampp/htdocs/v1/ci
}

function update_database() {
    mysql_bin="/opt/lampp/bin/mysql"
    # add 'role_id' field into 'role' table
    cmd1='ALTER TABLE `role` ADD `role_id` CHAR(4) NOT NULL DEFAULT "" AFTER `id`;'
    cmd2='ALTER TABLE `role` ADD UNIQUE KEY `role_id` (`role_id`);'    
    cmd3='ALTER TABLE `role_permission` MODIFY `role_id` CHAR(4);'
    cmd4='CREATE TABLE IF NOT EXISTS `role_group_binds` (`id` int(4) NOT NULL auto_increment, `role_id` char(4) NOT NULL DEFAULT "", `group_id` int(4) NOT NULL DEFAULT "1", `create_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP, UNIQUE KEY `id` (`id`)) ENGINE=InnoDB;'
    # if [ "$TC_REV" -lt "9582" ]; then
    #     echo $cmd1 | $mysql_bin -psa tc
    #     echo $cmd2 | $mysql_bin -psa tc
    #     echo $cmd3 | $mysql_bin -psa tc
    #     echo $cmd4 | $mysql_bin -psa tc
    # fi
    cmd=' ALTER TABLE `users` ADD `storage_lock` CHAR(32) NOT NULL DEFAULT "" AFTER `storage_frozen`;'
    cmd+='ALTER TABLE `usergroups` ADD `boot_ciid` INT(4) NOT NULL DEFAULT "0" AFTER `memo`;'
    cmd+='ALTER TABLE `usergroups` ADD `ps_background_upload` TINYINT(1) NOT NULL DEFAULT "-1" AFTER `memo`;'
    cmd+='ALTER TABLE `client_group` ADD `cg_autoboot_ciid` int(4) NOT NULL DEFAULT "0" AFTER `cg_create_ts`;'
    cmd+='DROP TABLE IF EXISTS `role`, `role_permission`;'
    cmd+='ALTER TABLE `users` CHANGE `userName` `userName` VARCHAR(40) COLLATE utf8_bin NOT NULL DEFAULT "";'

    if [ "$TC_REV" -lt "9583" ]; then
        echo $cmd | $mysql_bin -psa tc
    fi
}

function update_dependencies() {
    IN_DEP="$IN_TOP/3rdparty/web"
    TC_WEB="/opt/lampp/webconsole/assets"
    cp -a "$IN_DEP/chart.bundle.min.js" "$TC_WEB/chart.bundle.min.js"

    IN_DEP="$IN_TOP/3rdparty/centos7"
}

action "Read previous TC system revision" read_installation
action "Stop TC background services" stop_services
action "Extract upgrade package file" extract_tarball
action "Update TC server dependencies" update_dependencies
action "Update web management console" update_webconsole
action "Update database structure" update_database
action "Update internal binaries" update_binaries
action "Update TC server configuration" update_config
action "Update client firmware" update_firmware
action "Start TC background services" start_services

###
# TC release history revisions
# 
# * TC 1902 8754
# * TC 1902 8801
# * TC 1902 8987
# * TC 1902P1 9073
# * TC 1908   9543
# * TC 1910Pre   9980
# * TC 2003   wip
